#include "BufferInfoTable.h"

// initialize static data members ..
int BufferInfoTable::bitIndex = -1;

BufferInfoTable:: BufferInfoTable(void){
	// initialize ..
	
	tableCountUnique = 0;
	
	numberHitRequests = 0;		// "# of hits" = "# block requests" - "# requests read-from-disk" ..
	
	totalNumberBlockRequests = 0;	// "total # block requests" ..
	
	numberRequestsFromFile = 0;		// "# requests read-from-disk" ..
	
	
	tableName = new vector<string>();
	
	bufferPageFieldCount = new vector<int>();
	
	bufferPageRowCount = new vector<int>();		
											
	recordSize = new vector<int>();
	
	pageID = new vector<int>();
	
	pin_count = new vector<int>();			
	
	dirty = new vector<bool>();						

	empty = new vector<bool>();					

	selectData = new vector<bool>();			
	
	rowAvailable = new vector< vector<bool>* >();
	
	LRUBlocks = new list<int>();
	
}

BufferInfoTable:: ~BufferInfoTable(void){
}


int BufferInfoTable:: getBufferInfoTableIndex( int pageIDToLocate ){
	
	//cout << "BufferInfoTable:: getBufferInfoTableIndex() start .. " << endl << endl;
	
	int bitIndex = -1;
	
	// DEBUGGING ..
	//cout << "bufferPoolSize: " << bufferPoolSize << endl;
	
	for(int i=0; i<bufferPoolSize; i++){
		
		// DEBUGGING ..
		//cout << "pageID["<<i<<"]("<<pageID[i]<<") == pageIDToLocate("<<pageIDToLocate<<"): " << ( pageID[i] == pageIDToLocate ) << endl;
		
		if( pageID->at(i) == pageIDToLocate ){
			
			// DEBUGGING ..
			//cout << "bitIndex_1: " << bitIndex << endl;
			
			bitIndex = i;
			break;
			
		}
		
	}
	
	// DEBUGGING ..
	//cout << "bitIndex_2: " << bitIndex << endl;
	
	return bitIndex;
	
}  // end "getBufferInfoTableIndex( int pageID )" ..


int BufferInfoTable:: computeMaxRowCount_Index( int blockIndex ){
	
	// "N" : max number of rows ( for this blockIndex ) ..
	// "F" : number of fields ( for this blockIndex ) ..
	// "i" : size of an "int" variable, in bytes ( i = 4 bytes ) ..
	// "B" : size of a page, in bytes ( B = 48 bytes ) ..
	//  B - ( i * N * F ) = 0
	//          ( N * F ) = ( B / i )
	// 48 - ( 4 * N * F ) = 0
	//          ( N * F ) = 48 / 4 = 12
	//                  N = ( B / i ) / F
	//                  N = 12 / F
	
	
	// DEBUGGING ..
	//cout << "BufferInfoTable:: blockIndex: " << blockIndex << endl;
	
	// "F" ..
	int fieldCount = bufferPageFieldCount->at(blockIndex);
	
	// DEBUGGING ..
	//cout << "BufferInfoTable:: numberOfBytesPerBlock: " << numberOfBytesPerBlock << endl;
	//cout << "BufferInfoTable:: intSize: " << intSize << endl;
	//cout << "BufferInfoTable:: fieldCount: " << fieldCount << endl;
	
	// "( B / i )" ..
	int maxRowsPossiblePerBlock = ( numberOfBytesPerBlock / intSize );
	
	// DEBUGGING ..
	//cout << "BufferInfoTable:: maxRowsPossiblePerBlock: " << maxRowsPossiblePerBlock << endl;
	
	// "N" ..
	int maxRowCount = -1;
	
	if ( fieldCount > 0 ) {
		
		maxRowCount = ( maxRowsPossiblePerBlock / fieldCount );
		
	}
	
	// DEBUGGING ..
	//cout << "BufferInfoTable:: maxRowCount: " << maxRowCount << endl;
	
	// F = 1, N = 12
	// F = 2, N = 6
	// F = 3, N = 4
	// F = 4, N = 3
	// F = 5, N = 2.4 [2]*
	// F = 6, N = 2
	// F = 7, N = 1.714285 [1]*
	// F = 8, N = 1.5 [1]*
	
	// * : division operator should round these down to the nearest whole number shown ..
	
	return maxRowCount;
	
	
} // end "computeMaxRowCount( int blockIndex )" ..


int BufferInfoTable:: computeMaxRowCount_Attributes( int numberOfAttributes ){
	
	// "N" : max number of rows ( for this blockIndex ) ..
	// "F" : number of fields ( for this blockIndex ) ..
	// "i" : size of an "int" variable, in bytes ( i = 4 bytes ) ..
	// "B" : size of a page, in bytes ( B = 48 bytes ) ..
	//  B - ( i * N * F ) = 0
	//          ( N * F ) = ( B / i )
	// 48 - ( 4 * N * F ) = 0
	//          ( N * F ) = 48 / 4 = 12
	//                  N = ( B / i ) / F
	//                  N = 12 / F
	

	// DEBUGGING ..	
	//cout << "BufferInfoTable:: blockIndex: " << blockIndex << endl;
	
	// "F" ..
	int fieldCount = numberOfAttributes;
	
	// DEBUGGING ..
	//cout << "BufferInfoTable:: numberOfBytesPerBlock: " << numberOfBytesPerBlock << endl;
	//cout << "BufferInfoTable:: intSize: " << intSize << endl;
	//cout << "BufferInfoTable:: fieldCount: " << fieldCount << endl;
	
	// "( B / i )" ..
	int maxRowsPossiblePerBlock = ( numberOfBytesPerBlock / intSize );
	
	// DEBUGGING ..
	//cout << "BufferInfoTable:: maxRowsPossiblePerBlock: " << maxRowsPossiblePerBlock << endl;
	
	// "N" ..
	int maxRowCount = -1;
	
	if ( fieldCount > 0 ) {
		
		maxRowCount = ( maxRowsPossiblePerBlock / fieldCount );
		
	}
	
	// DEBUGGING ..
	//cout << "BufferInfoTable:: maxRowCount: " << maxRowCount << endl;
	
	// F = 1, N = 12
	// F = 2, N = 6
	// F = 3, N = 4
	// F = 4, N = 3
	// F = 5, N = 2.4 [2]*
	// F = 6, N = 2
	// F = 7, N = 1.714285 [1]*
	// F = 8, N = 1.5 [1]*
	
	// * : division operator should round these down to the nearest whole number shown ..
	
	return maxRowCount;
	
	
} // end "computeMaxRowCount( int blockIndex )" ..


void BufferInfoTable:: setTableName( vector<string>* newTableName ){
	
	tableName = newTableName;
	
}


void BufferInfoTable:: updateTableCountUnique() {
	
	// update tracking on "tableCount" ..
	
	int uniqueCount = 0;
	
	bool tableNameNotUnique = false;
	
	vector<string>* uniqueTableNames = new vector<string>();
	
	if ( tableName->size() > 0 ){
		
		// only count unique "tableName" if tracking at least one tableName ..
		
		for(int i=0; i<tableName->size(); i++){
			
			// loop through each "tableName", checking to see if it is "unique" ..
			
			for(int j=0; j<uniqueTableNames->size(); j++){
				
				// loop through all "uniqueTableNames", see if one matches "tableName[ i ]" ..
				
				if ( uniqueTableNames->at( j ).compare( tableName->at(i) ) == 0 ){
			
					// if "uniqueTableNames == tableName[ i ]", then
					// "tableName[ i ]" is a repeat, so don't count it ..
					
					tableNameNotUnique = true;
					break;
			
				}	
				
			}
				
			if ( !tableNameNotUnique ) {
					
				// the "tableName[ i ]" is new and unique, so count it ..
				uniqueCount++;
				uniqueTableNames->push_back( tableName->at(i) );
					
			} else {
				
				// reset this parameter for the next "tableName" ..
				tableNameNotUnique = false;
				
			}

		} // end "for()" ..
		
		// adjust tableNameUnique against "uniqueCount"
		// to reflect its latest updated value ..
		// a) tableCountUnique > uniqueCount
		// b) tableCountUnique == uniqueCount
		// c) tableCountUnique < uniqueCount
		if ( tableCountUnique != uniqueCount ) {
			
			tableCountUnique = uniqueCount;
			
		}
		
	} // end "if ( tableName.size() > 0 )" ..	
	
	
} // end "updateTableCountUnique()" ..
	
void BufferInfoTable:: setBufferPageFieldCount( vector<int>* newBufferPageFieldCount ){
	
	bufferPageFieldCount = newBufferPageFieldCount;
	
}
	
void BufferInfoTable:: setBufferPageRowCount( vector<int>* newBufferPageRowCount ){
	
	bufferPageRowCount = newBufferPageRowCount;
	
}
	
void BufferInfoTable:: setRecordSize( vector<int>* newRecordSize ){
	
	recordSize = newRecordSize;
	
}
	
void BufferInfoTable:: setPageID( vector<int>* newPageID ){
	
	pageID = newPageID;
	
}
	
void BufferInfoTable:: setPinCount( vector<int>* newPinCount ){
	
	pin_count = newPinCount;
	
}
	
void BufferInfoTable:: setDirty( vector<bool>* newDirty ){
	
	dirty = newDirty;
	
}
	
void BufferInfoTable:: setEmpty( vector<bool>* newEmpty ){
	
	empty = newEmpty;
	
}
	
void BufferInfoTable:: setSelectData( vector<bool>* newSelectData ){
	
	selectData = newSelectData;
	
}
	
void BufferInfoTable:: setRowAvailable( vector< vector<bool>* >* newRowAvailable ){
	
	rowAvailable = newRowAvailable;
	
}
	
void BufferInfoTable:: setNumberHitRequests( float newTotalNumberHitRequests ){
	
	numberHitRequests = newTotalNumberHitRequests;
	
}
	
void BufferInfoTable:: setTotalNumberBlockRequests( float newTotalNumberBlockRequests ){
	
	totalNumberBlockRequests = newTotalNumberBlockRequests;
	
}
	
void BufferInfoTable:: setNumberRequestsFromFile( float newNumberRequestsFromFile ){
	
	numberRequestsFromFile = newNumberRequestsFromFile;
	
}
	
	
void BufferInfoTable:: updateHitRate(){
	
/*
	// "# of hits" = "total # block requests" - "# requests read-from-disk" ..
	numberHitRequests = ( totalNumberBlockRequests - numberRequestsFromFile );
	
	// "hit rate" = "# of hits" / "total # block requests" ..
	hitRate = ( numberHitRequests / totalNumberBlockRequests );
	
*/

	// "total # block requests" = "# of hits" + "# requests read-from-disk" ..
	totalNumberBlockRequests = ( numberHitRequests + numberRequestsFromFile );
	
	// "hit rate" = "# of hits" / "total # block requests" ..
	if( totalNumberBlockRequests > 0 )
		hitRate = ( numberHitRequests / totalNumberBlockRequests );
	
}
	
void BufferInfoTable:: setLRUBlocks( list<int>* newLRUBlocks ){
		
	LRUBlocks = newLRUBlocks;
		
}
	
void BufferInfoTable:: setBufferPoolSize( int newBufferPoolSize ) {
	
	// DEBUGGING ..
	//cout << "BufferManager():: setBufferPoolSize start .." << endl;
	
	if ( bufferPoolSize == 0 ){
		
		// DEBUGGING ..
		//cout << "Point A .." << endl;
		
		bufferPoolSize = newBufferPoolSize;
		
		// initial setting at startup of DBMS ..
		if ( newBufferPoolSize > 0 ){
			
			for( int i=0; i<newBufferPoolSize; i++ )
				tableName->push_back( "" );
			
			for( int i=0; i<newBufferPoolSize; i++ )
				bufferPageFieldCount->push_back( 0 );
			
			for( int i=0; i<newBufferPoolSize; i++ )
				bufferPageRowCount->push_back( 0 );
			
			for( int i=0; i<newBufferPoolSize; i++ )
				recordSize->push_back( 0 );

			for( int i=0; i<newBufferPoolSize; i++ )
				pageID->push_back( -1 );
			
			for( int i=0; i<newBufferPoolSize; i++ )
				pin_count->push_back( 0 );
			
			for( int i=0; i<newBufferPoolSize; i++ )
				dirty->push_back( false );
			
			for( int i=0; i<newBufferPoolSize; i++ )
				empty->push_back( true );
				
			for( int i=0; i<newBufferPoolSize; i++ )
				selectData->push_back( true );
			

			// initialize "slots" with max possible per block,
			// even if in practice it will usually be less than this ..
			vector<bool> slots;
			
			int maxRowsPossiblePerBlock = ( BufferInfoTable::numberOfBytesPerBlock / BufferInfoTable::intSize );

			for(int i=0; i < maxRowsPossiblePerBlock; i++){
							
				slots.push_back( true );
							
			}

			for(int i=0; i < newBufferPoolSize; i++){
							
				rowAvailable->push_back( new vector<bool>( slots ) );
	
			}
			
		} else {
			
			// default size of data member arrays set to "1" ..
			
			//tableName[0] = "";
			
			bufferPageFieldCount->at(0) = 0;
			
			bufferPageRowCount->at(0) = 0;
			
			recordSize->at(0) = 0;
			
			pageID->at(0) = 0;
			
			pin_count->at(0) = 0;
			
			dirty->at(0) = false;
			
			empty->at(0) = true;
			
			selectData->at(0) = true;

		}
	
	} else {
		
		// do nothing (cannot change size except at beginning of program) ..
		
	}
	
	// DEBUGGING ..
	//cout << "BufferManager():: setBufferPoolSize end .." << endl;
	
} // end "setBufferPoolSize( int bpSize )" ..
	
	
	
	// getters ..
	
vector<string>* BufferInfoTable:: getTableName(){
	
	return tableName;
	
}
	
int BufferInfoTable:: getTableCountUnique(){
	
	return tableCountUnique;
	
}
	
vector<int>* BufferInfoTable:: getBufferPageFieldCount(){
	
	return bufferPageFieldCount;
	
}
	
vector<int>* BufferInfoTable:: getBufferPageRowCount(){
	
	return bufferPageRowCount;
	
}
	
vector<int>* BufferInfoTable:: getRecordSize(){
	
	return recordSize;
	
}
	
vector<int>* BufferInfoTable:: getPageID(){
	
	return pageID;
	
}
	
vector<int>* BufferInfoTable:: getPinCount(){
	
	return pin_count;
	
}
	
vector<bool>* BufferInfoTable:: getDirty(){
	
	return dirty;
	
}
	
vector<bool>* BufferInfoTable:: getEmpty(){
	
	return empty;
	
}
	
vector<bool>* BufferInfoTable:: getSelectData(){
	
	return selectData;
	
}
	
vector< vector<bool>* >* BufferInfoTable:: getRowAvailable(){
	
	return rowAvailable;
	
}
	
float BufferInfoTable:: getNumberHitRequests(){
	
	return numberHitRequests;
	
}
	
float BufferInfoTable:: getTotalNumberBlockRequests(){
	
	return totalNumberBlockRequests;
	
}
	
float BufferInfoTable:: getNumberRequestsFromFile(){
	
	return numberRequestsFromFile;
	
}
	
	
float BufferInfoTable:: getHitRate(){
	
	return hitRate;
	
}

	
list<int>* BufferInfoTable:: getLRUBlocks(){
		
	return LRUBlocks;
		
}

	
int BufferInfoTable:: getBufferPoolSize(){
	
	return bufferPoolSize;
	
}

void BufferInfoTable:: updateBufferInfoTable( int bitIndex, int bits ){
	
	if( bits == BufferInfoTableState::FLUSH_BLOCK ){
		
		tableName->at( bitIndex ) = "";
	
		bufferPageFieldCount->at( bitIndex ) = 0;
	
		bufferPageRowCount->at( bitIndex ) = 0;		
											
		recordSize->at( bitIndex ) = 0;
	
		pageID->at( bitIndex ) = -1;
	
		pin_count->at( bitIndex ) = 0;			
	
		dirty->at( bitIndex ) = false;						

		empty->at( bitIndex ) = true;					

		selectData->at( bitIndex ) = true;			
	
		int rowAvailableSize = rowAvailable->at( bitIndex )->size();
		for(int i=0; i<rowAvailableSize; i++)
			rowAvailable->at( bitIndex )->at(i) = true;
		
	}
	
}

bool BufferInfoTable:: isBufferPoolCommit(){

	// checks buffer pool to see if any pages in it are 
	// eligible to be flushed to file ..

	bool canBufferPoolBeCommited = false;

	if( ( dirty != NULL ) && 
		( empty != NULL ) ){
			
		// check the buffer pool for at least one
		// non-empty, flushable page ..
		for(int i=0; i<bufferPoolSize; i++){
		
			if( ( dirty->at(i) == true ) && 
				( empty->at(i) == false ) ){
				
				// this page is flushable ("dirty" == true)
				// and non-empty ("empty" = false) ..
				
				canBufferPoolBeCommited = true;
				break;
				
			}
		
		} // end "for()" ..
			
	}
	
	return canBufferPoolBeCommited;
	
}

