#ifndef BUFFERINFOTABLE_H_
#define BUFFERINFOTABLE_H_

#pragma once

#include "BufferInfoTableState.h"

#include <string>
#include <list>
#include <vector>
#include <iostream>

using namespace std;

class BufferInfoTable {
	
public:

	BufferInfoTable(void);
	~BufferInfoTable(void);

	// static data members ..
	
	int static bitIndex;
	
	int static const numberOfBytesPerBlock = 48;
	
	int static const intSize = 4;


	// data members ..
	
	
	// member functions ..
	
	int getBufferInfoTableIndex( int pageIDToLocate );
	
	int computeMaxRowCount_Index( int blockIndex );
	
	int computeMaxRowCount_Attributes( int numberOfAttributes );
	
	
	// setters ..
	
	void setTableName( vector<string>* newTableName );
	
	void updateTableCountUnique();
	
	void setBufferPageFieldCount( vector<int>* newBufferPageFieldCount );
	
	void setBufferPageRowCount( vector<int>* newBufferPageRowCount );
	
	void setRecordSize( vector<int>* newRecordSize );
	
	void setPageID( vector<int>* newPageID );
	
	void setPinCount( vector<int>* newPinCount );
	
	void setDirty( vector<bool>* newDirty );
	
	void setEmpty( vector<bool>* newEmpty );
	
	void setSelectData( vector<bool>* newSelectData );
	
	void setRowAvailable( vector< vector<bool>* >* newRowAvailable );
	
	void setNumberHitRequests( float newNumberOfHitRequests );
	
	void setNumberRequestsFromFile( float newNumberRequestsFromFile );
	
	void setTotalNumberBlockRequests( float newTotalNumberOfBlockRequests );
	
	void updateHitRate();
	
	void setLRUBlocks( list<int>* newLRUBlocks );
	
	void setBufferPoolSize( int newBufferPoolSize );
	
	
	
	// getters ..
	
	vector<string>* getTableName();
	
	int getTableCountUnique();
	
	vector<int>* getBufferPageFieldCount();
	
	vector<int>* getBufferPageRowCount();
	
	vector<int>* getRecordSize();
	
	vector<int>* getPageID();
	
	vector<int>* getPinCount();
	
	vector<bool>* getDirty();
	
	vector<bool>* getEmpty();
	
	vector<bool>* getSelectData();
	
	vector< vector<bool>* >* getRowAvailable();
	
	float getNumberHitRequests();
	
	float getNumberRequestsFromFile();
	
	float getTotalNumberBlockRequests();
	
	float getHitRate();
	
	list<int>* getLRUBlocks();
	
	int getBufferPoolSize();

	// "bits" is a value from "BufferInfoTableState" ..
	void updateBufferInfoTable( int bitIndex, int bits );
	
	// checks whether any pages in bp are eligible to be flushed to disk ..
	bool isBufferPoolCommit();

private:

	// data members ..
	
	vector<string>* tableName;				// tableName, one per BP block (may not be unique) ..
	
	int tableCountUnique;					// count of the "unique" tableName ..
	
	vector<int>* bufferPageFieldCount;
	
	vector<int>* bufferPageRowCount;		// need "bit.bufferPageRowCount" for "bm.findBlockForInsert()"
											// and for "recIndex = bit.bufferPageRowCount[ flushBlockIndex ];" ..
	vector<int>* recordSize;
	
	vector<int>* pageID;					// for insert this will be "-1" or new pageID .. 
	
	vector<int>* pin_count;					// for insert this will be "0" ..
	
	vector<bool>* dirty;					// for insert this is "true" ..
	
	
	

	vector<bool>* empty;					// initially the blocks of the buffer are emtpy ..
											// until the BP block has its very first insert, this will be "true" / "1" ..
											// when the block is "flushed" or "committed", this is also "true" / "1"

	vector<bool>* selectData;				// for select this is "true", for insert this is "false" ..
											// used for setting locks for "insert" statements, and
											// for not setting locks for "select" statements ..

	
	vector< vector<bool>* >* rowAvailable;	// tracks whether an individual row in the page is
											//    "free / available" (true): the row is empty or full of garbage values
											// "full / unavailable" (false): the row is occupied by valid data
											// 
											// 1 byte * 12 cells = 12 bytes ..
											// each entry of the vector is 1 byte (bool)
											// the max number of cells for a 48-byte block is 12 (12 * 4 = 48 bytes) ..
											//
											// format: bp blocks < rows in each bp block < row is (free / full) ? > > ..
											
											
	// "hit rate" parameters ..
	float numberHitRequests;			// numberHitRequests = ( totalNumberBlockRequests - numberRequestsFromFile ) ..
	
	float numberRequestsFromFile;			// all the request for a chunk of data from file ("page-from-file") ..
	
	float totalNumberBlockRequests;			// all the requests for a chunk of data, "block-from-bp" + "page-from-file" ..
											// (aka "numberHitRequests" + "numberRequestsFromFile" ) ..
	
	float hitRate;							// hitRate = ( numberHitRequests / totalNumberBlockRequests )
	
	
	// "LRU" parameter ..
	list<int>* LRUBlocks;
	
	
	// "BP" parameter ..
	int bufferPoolSize;
	
};

#endif /*BUFFERINFOTABLE_H_*/
