#include "Catalog.h"

Catalog:: Catalog() {
	
	// default constructor ..
	
	numberOfTables = 0;
	
	numberOfIndices = 0;
	
	fileCatalogTimestamp = new CatalogTimestamp();
	
	managerCatalogTimestamp = new CatalogTimestamp();
	
	fileCatalogTimestamp->setTimestamp(0);
	
	managerCatalogTimestamp->setTimestamp(0);
	
	catalogTables = new vector<CatalogTable*>();
	
	catalogIndices = new vector<CatalogIndex*>();
	
}


Catalog:: ~Catalog() {
}


string Catalog:: addCatalogTable( CatalogTable* catalogTable ){
	
	
	catalogTables->push_back( catalogTable );

	numberOfTables++;
	
	string returnMessage = "The table ";
	returnMessage += catalogTable->getTableName();
	returnMessage += " was added to the catalog.";
	
	return returnMessage;
	
	
} // end "addCatalogTable( CatalogTable )" ..


string Catalog:: removeCatalogTable( string tableName ){
	
	string returnMessage = "Attempt to remove table from catalog initiated.";
	
	//vector<CatalogTable>::iterator it;
	
	// DEBUGGING ..
	//cout << "catalog.catalogTables.size(): " << catalogTables.size() << endl;
	
	CatalogTable* catalogTable = NULL;
	
	int catalogTablesSize = catalogTables->size();
	
	//for( it = catalogTables.begin(); it != catalogTables.end(); it++){
	for(int i=0; i<catalogTablesSize; i++){
		
		catalogTable = catalogTables->at(i);
		
		// if you find the "tableName" to remove, remove it and stop looking ..
		//if ( (*it).getTableName().compare( tableName ) == 0 ){
		if ( catalogTable->getTableName().compare( tableName ) == 0 ){
			
			//catalogTables->erase( it );
			catalogTables->erase( catalogTables->begin() + i );
			
			numberOfTables--;
			
			// DEBUGGING ..
			//cout << "catalog.catalogTables.size(): " << catalogTables.size() << endl;
			
			returnMessage.empty();
			returnMessage += "The table ";
			returnMessage += tableName;
			returnMessage += " was removed from the catalog.";
			
			break;
			
		}
		
	}
	
	return returnMessage;
	
	
} // end "removeCatalogTable( string tableName )" ..


string Catalog:: addCatalogIndex( CatalogIndex* catalogIndex ){
	
	
	catalogIndices->push_back( catalogIndex );
	
	numberOfIndices++;
	
	string returnMessage = "The index ";
	returnMessage += catalogIndex->getIndexName();
	returnMessage += " was added to the catalog.";
	
	return returnMessage;
	
	
} // end "addCatalogIndex( CatalogIndex )" ..


string Catalog:: removeCatalogIndex( string indexName ){
	
	string returnMessage = "Attempt to remove index from catalog initiated.";
	
	//vector<CatalogIndex>::iterator;
	
	// DEBUGGING ..
	//cout << "catalog.catalogTables.size(): " << catalogTables.size() << endl;
	
	CatalogIndex* catalogIndex = NULL;
	
	int catalogIndicesSize = catalogIndices->size();
	
	//for( it = catalogIndices->begin(); it != catalogIndices->end(); it++){
	for( int i=0; i<catalogIndicesSize; i++){
		
		catalogIndex = catalogIndices->at(i);
		
		// if you find the "tableName" to remove, remove it and stop looking ..
		//if ( (*it).getIndexName().compare( indexName ) == 0 ){
		if ( catalogIndex->getIndexName().compare( indexName ) == 0 ){
			
			//this->catalogIndices->erase( it );
			catalogIndices->erase( catalogIndices->begin() + i );
			
			numberOfIndices--;
			
			// DEBUGGING ..
			//cout << "catalog.catalogIndices.size(): " << catalogIndices.size() << endl;
			
			returnMessage.empty();
			returnMessage += "The table ";
			returnMessage += indexName;
			returnMessage += " was removed from the catalog.";
			
			break;
			
		}
		
	}
	
	return returnMessage;
	
	
} // end "removeCatalogIndex( string indexName )" ..


/*
 * SETTERS ..
 */
 

void Catalog:: setNumberOfTables( int numberOfTables_ ){
	
	numberOfTables = numberOfTables_;
	
}

void Catalog:: setNumberOfIndices( int numberOfIndices_ ){
	
	numberOfIndices = numberOfIndices_;
	
}

void Catalog:: setCatalogTables( vector<CatalogTable*>* tables_ ){
	
	catalogTables = tables_;
	
}

void Catalog:: setCatalogIndices( vector<CatalogIndex*>* indices_ ){
	
	catalogIndices = indices_;
	
}

void Catalog:: setFileCatalogTimestamp( time_t fileTimestamp ){
	
	fileCatalogTimestamp->setTimestamp( fileTimestamp );
	
}

void Catalog:: setManagerCatalogTimestamp( time_t managerTimestamp ){
	
	managerCatalogTimestamp->setTimestamp( managerTimestamp );
	
}


/*
 * GETTERS ..
 */


int Catalog:: getNumberOfTables(){
	
	return numberOfTables;
	
}

int Catalog:: getNumberOfIndices(){
	
	return numberOfIndices;
	
}

vector<CatalogTable*>* Catalog:: getCatalogTables(){
	
	return catalogTables;
	
}


CatalogTable* Catalog:: getCatalogTable( string catalogTableName ){
	
	CatalogTable* catalogTable = NULL;
	
	CatalogTable* catalogTableFound = NULL;
	
	int catalogTablesSize = catalogTables->size();
	
	for(int i=0; i<catalogTablesSize; i++){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTableName.compare( catalogTable->getTableName() ) == 0 ){
			
			catalogTableFound = catalogTable;
			
		}
		
	}
	
	return catalogTableFound;
	
} // end "getCatalogTable( string catalogTableName )" ..


vector<CatalogIndex*>* Catalog:: getCatalogIndices(){
	
	return catalogIndices;
	
}


CatalogIndex* Catalog:: getCatalogIndex( string catalogIndexName ){
	
	CatalogIndex* catalogIndexFound;
	
	for(int i=0; i<catalogIndices->size(); i++){
		
		if( catalogIndexName.compare( catalogIndices->at(i)->getIndexName() ) == 0 ){
			
			catalogIndexFound = catalogIndices->at(i);
			
		}
		
	}
	
	return catalogIndexFound;
	
} // end "getCatalogIndex_indexName( string catalogIndexName )" ..


CatalogTimestamp* Catalog:: getFileCatalogTimestamp(){
	
	return fileCatalogTimestamp;
	
}

CatalogTimestamp* Catalog:: getManagerCatalogTimestamp(){

	return managerCatalogTimestamp;
	
}

void Catalog:: resetCatalog(){
	
	// reset all the values of the catalog ..
	
	numberOfTables = 0;
	
	numberOfIndices = 0;
	
	if( catalogTables != NULL )
		catalogTables->clear();
	else
		catalogTables = new vector<CatalogTable*>();

	if( catalogIndices != NULL )
		catalogIndices->clear();
	else
		catalogIndices = new vector<CatalogIndex*>();

	fileCatalogTimestamp = new CatalogTimestamp();
	
	managerCatalogTimestamp = new CatalogTimestamp();
	
	fileCatalogTimestamp->setTimestamp(0);
	
	managerCatalogTimestamp->setTimestamp(0);
	
}
