#ifndef CATALOG_H_
#define CATALOG_H_

#pragma once

#include <iostream>
#include <string>
#include <vector>

#include "CatalogTable.h"
#include "CatalogIndex.h"
#include "CatalogTimestamp.h"


using namespace std;

class Catalog {
	
public:

	// data members:

	
	// member functions:
	
	Catalog();
	~Catalog();
	
	string addCatalogTable( CatalogTable* catalogTable );
	
	string removeCatalogTable( string catalogIndex );
	
	
	string addCatalogIndex( CatalogIndex* catalogIndex );
	
	string removeCatalogIndex( string indexName );
	
	/*
	 * SETTERS ..
	 */
	 
	 void setNumberOfTables( int numberOfTables_ );
	 
	 void setNumberOfIndices( int numberOfIndices_ );
	 
	 void setCatalogTables( vector<CatalogTable*>* tables_ );
	 
	 void setCatalogIndices( vector<CatalogIndex*>* indices_ );
	 
	 void setFileCatalogTimestamp( time_t fileTimestamp );
	 
	 void setManagerCatalogTimestamp( time_t managerTimestamp );

	
	/*
	 * GETTERS ..
	 */
	 
	 int getNumberOfTables();
	 
	 int getNumberOfIndices();
	 
	 vector<CatalogTable*>* getCatalogTables();
	 
	 CatalogTable* getCatalogTable( string catalogTableName );
	 
	 vector<CatalogIndex*>* getCatalogIndices();
	 
	 CatalogIndex* getCatalogIndex( string catalogIndexName );
	 
	 CatalogTimestamp* getFileCatalogTimestamp();
	 
	 CatalogTimestamp* getManagerCatalogTimestamp();
	 
	 void resetCatalog();


private:

	// data members:
	
	int numberOfTables;								// count of the No. of non-volatile tables stored ..
	
	int numberOfIndices;							// count of the No. of non-volatile indices stored ..
	
	vector<CatalogTable*>* catalogTables;			// each represents a single table from the Catalog file ..

	vector<CatalogIndex*>* catalogIndices;			// each represents a single index from the Catalog file ..

	CatalogTimestamp* fileCatalogTimestamp;			// [value read in from DB] latest / current value read in from DB, indicating the last time that the DB was updated / changed ..
	
	CatalogTimestamp* managerCatalogTimestamp;		// [value written to DB] value stored in CatalogManager object, indicating the last time that the DB was updated / changed ..


	// member functions:
	
	
}; // end "Catalog" class ..

#endif /*CATALOG_H_*/
