#ifndef DATAHEADER_H_
#define DATAHEADER_H_

#pragma once

#include <vector>

#include "DataHeaderEntry.h"

using namespace std;

class DataHeader {

public:

	DataHeader(void);
	~DataHeader(void);

	// "DataHeader" contains:
	// 1) int totalNumberOfDataPages	4 bytes
	
	// "DataHeaderEntries" contains:
	// 1) int pageID					4 bytes
	// 2) int freeRowsCount			    4 bytes
	// 3) int filledRowsCount			4 bytes
	// 4) vector<bool> rowAvailable   12 bytes max
	// 5) int lockNumber				4 bytes
	
	// total:
	// "DataHeader" file size = 4 + 28*N
	// where "N" = number of "DataHeaderEntries"
	// 4 + 28*N	  || 	data bytes	||	filler bytes	||	total file bytes
	// ==================================================================
	// 4 + 28*(1) = 		32		+		16			=		48
	// 4 + 28*(2) = 		60		+		36			=		96
	// 4 + 28*(3) = 		86		+		 8			=		96
	// 4 + 28*(4) =    	   116		+		28			=	   144
	// etc ..
	
	
	// member methods ..
	
	// compute the max number of rows per data page, 
	// based on "filePageFieldCount" ..
	int computeMaxRowCount( int fieldCount_ );
	
	// setters ..
	void setTotalNumberOfDataPages( int newTotalNumberOfDataPages );
	
	void setHeaderEntries( vector<DataHeaderEntry*>* newHeaderEntries );
	
	
	// getters ..
	int getTotalNumberOfDataPages();
	
	vector<DataHeaderEntry*>* getHeaderEntries();
	
	
private:
	
	// data members ..
	
	// number of data pages for this table ..
	int totalNumberOfDataPages; 						
	
	// tracks each data page of the table, one entry per data page ..
	vector<DataHeaderEntry*>* headerEntries; 

};

#endif /*DATAHEADER_H_*/
