#ifndef DATAHEADERENTRY_H_
#define DATAHEADERENTRY_H_

#pragma once

#include <vector>

using namespace std;

class DataHeaderEntry {
	
public:

	DataHeaderEntry( int maxRowCount );
	~DataHeaderEntry( void );
	
	// setters ..
	
	void setPageID( int newPageID );
	
	void setFreeRowsCount( int newFreeRowsCount );
	
	void setFilledRowsCount( int newFilledRowsCount );
	
	void setRowAvailable( vector<bool>* newRowAvailable );
	
	void setLockNumber( int newLockNumber );
	
	
	// getters ..
	
	int getPageID();
	
	int getFreeRowsCount();
	
	int getFilledRowsCount();
	
	vector<bool>* getRowAvailable();		// getRowAvailable()->at() ..
	
	int getLockNumber();
	
private:

	// data members ..

	// Data page page id .. 
	// (This is the "page-from-file" / "buffer-pool-block" index number) ..
	int pageID;
	
	// Number of "free rows" (empty) in a data page .. 
	int freeRowsCount;
	
	// Number of "filled rows" (non-empty) in a data page ..
	// (this will be less than the "maxRowCount" calculated in
	// "computeMaxRowCount()" from either "DataHeader" or "BufferInfoTable") ..
	int filledRowsCount;

	// tracks whether an individual row in the page is
	//    "free / available" (true): the row is empty or full of garbage values
	// "full / unavailable" (false): the row is occupied by valid data
	// 
	// 1 byte * 12 cells = 12 bytes ..
	// each entry of the vector is 1 byte (bool)
	// the max number of cells for a 48-byte block is 12 (12 * 4 = 48 bytes) ..
	vector<bool>* rowAvailable;

	/*
	 * When an instantiation of this program (P1) reads a page from the file,
	 * it must "lock" the page to insure that other instantiations of this program (P2)
	 * don't attempt to change the locked page, making P1's copy of the page invalid,
	 * (not the latest valueof the page of the file).
	 * 
	 * Possible States for "lockNumber":
	 * lockNumber = 0								..	page-in-file is "unlocked" ..
	 * lockNumber > 0								..	page-in-file is "locked" ..
	 * lockNumber increasing every X minutes		..	page-in-file is "active locked" ( lock can only be unlocked by "lock-initiator" while in this state ) ..
	 * lockNumber not increasing every X minutes	..	page-in-file is "inactive locked" ( ok to be unlocked by non-"lock-initiator" ) ..
	 * NOTE: because no easy way to implement threads in C++, not implementing active / inactive locks,
	 * 	     but only lock / unlock ( where lockNumber > 0 indicates locked page, 
	 *       lockNumber = 0 indicates unlocked page ) ..
	 */
	int lockNumber;

};

#endif /*DATAHEADERENTRY_H_*/
