#include "DataPage.h"

DataPage:: DataPage(){
	
	pageID = -1;
	
	// default standard block size ..
	maxPageSize = 48;
	
	// DataPage starts out empty ..
	numberOfEntries = 0;
	
	// default value, all entries are type "int" ..
	sizeofEntries = sizeof(int);
	
	sizeOfLongestRow = 0;
	
	isNewRow = false;
	
	// contains the data in a "M x N" data structure ..
	manyRows = new vector< vector<int>* >();

}

DataPage:: DataPage( DataPage* dataPage ){
	
	// copy constructor ..
	
	pageID = dataPage->getPageID();
	
	maxPageSize = dataPage->getMaxPageSize();
	
	numberOfEntries = dataPage->getNumberOfEntries();
	
	sizeofEntries = dataPage->getSizeofEntries();
	
	sizeOfLongestRow = dataPage->getSizeOfLongestRow();
	
	isNewRow = true;
	
	manyRows = dataPage->getEntries();
	
}

DataPage:: ~DataPage(){
}

int DataPage:: getPageID(){
	
	return pageID;	
	
}

void DataPage:: setPageID( int pageID_ ){
	
	pageID = pageID_;
	
}

int DataPage:: getMaxPageSize(){
	
	return maxPageSize;	
	
}

void DataPage:: setMaxPageSize( int maxPageSize_ ){
	
	// maxPageSize_ ~ in # of bytes (not # of values) ..
	
	// this method sets the max size in bytes allowed for this page,
	// "maxPageSize" controls the number of entries allowed to be added ..
	
	// check that "maxPageSize" is at least
	// as big as the size of a single entry ~ sizeof(type) ..
	if( maxPageSize_ < sizeofEntries ){
		
		maxPageSize_ = sizeofEntries;
		
	} else {
		
		// check if "maxPageSize" is
		// a multiple of "sizeof(type)" ..
		int maxPageSizeDiff = maxPageSize_ % sizeofEntries;
	
		// if "maxPageSize" is a multiple of "sizeof(type)",
		// the result of the modulus operation should be zero ..
		// (and the given value is ok to use as-is) ..
		if( maxPageSizeDiff != 0 ){
		
			// if the modulus operation result is not zero,
			// simply add the result to the given
			// "maxPageSize" in order to arrive at the closest
			// value which _is_ a multiple of "sizeof(type)"..
		
			maxPageSize_ += maxPageSizeDiff;
		
		}	
		
	}
	
	// update the new "maxPageSize" ..
	
	maxPageSize = maxPageSize_;
	
}

int DataPage:: getNumberOfEntries(){
	
	return numberOfEntries;	
	
}

int DataPage:: getSizeofEntries(){
	
	return sizeofEntries;	
	
}

int DataPage:: getSizeOfLongestRow(){
	
	return sizeOfLongestRow;	
	
}

bool DataPage:: getIsNewRow(){
	
	return isNewRow;
	
}

int DataPage:: getEntry( int rowNumber, int columnNumber ){
	
	// if "rowNumber" or "columnNumber" are out-of-range,
	// return error flag ..
	if( rowNumber < 0 || columnNumber < 0 )
		return -1;
		
	int entry = -1;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		//    rowNumber: 0 .. N
		// numberOfRows: 1 .. N+1
		if( rowNumber < numberOfRows ){
			
			vector<int>* row = manyRows->at( rowNumber );
			
			if( row != NULL ){
				
				int numberOfColumns = row->size();
					
				//    columnNumber: 0 .. M
				// numberOfColumns: 1 .. M+1
				if( columnNumber < numberOfColumns ){
				
					entry = row->at( columnNumber );
				
				}
				
			}
			
		}
		
	}
	
	return entry;
	
}

vector< vector<int>* >* DataPage:: getEntries(){
	
	// return the entire DataPage as a collection copy ..
	
	// collection copy to return ..
	vector< vector<int>* >* entries = NULL;
	
	if( manyRows != NULL ){
		
		entries = new vector< vector<int>* >( manyRows->begin(), manyRows->end() );
		
	}
	
	return entries;
	
}

vector<int>* DataPage:: getEntries( int rowNumber ){
	
	// return the row of entries located at "rowNumber" ..
	
	// if "rowNumber" is out-of-range, exit ..
	if( rowNumber < 0 )
		return NULL;
	
	// single-dim collection to return ..
	vector<int>* rowEntries = NULL;
	
	if( manyRows != NULL ){
		
		int manyRowsSize = manyRows->size();
		
		if( manyRowsSize > 0 ){
			
			// if "rowNumber" is out-of-range, exit ..
			if( rowNumber >= manyRowsSize )
				return NULL;
		
			rowEntries = manyRows->at( rowNumber );	
			
		}
		
	}
	
	return rowEntries;
	
}

/*
int* DataPage:: getEntries(){
	
	// index for "entries" ..
	// for non-jagged arrays:
	// k = (i*numberOfRows) + j ..
	// since "manyRows" could be a jagged array:
	// k = (x+1) + j .. (see below, this is what I implemented) ..
	int k = -1;
	
	// for first row, x = -1 ..
	// for remainder of rows, "x"
	// is equal to the last "k index" 
	// of the previous row (see below) ..
	int x = -1;
	
	// single-dim array to return ..
	int* entries = new int[numberOfEntries];
	
	if( manyRows != NULL ){
		
		int numberOfColumns = -1;
		
		vector<int>* row = NULL;
		
		int numberOfRows = manyRows->size();
		
		// loop through each row ..
		for(int i=0; i<numberOfRows; i++){
			
			row = manyRows->at(i);
			
			if( row != NULL ){
				
				numberOfColumns = row->size();
				
				// loop through each attribute ..
				for(int j=0; j<numberOfColumns; j++){
					
					// "k" is last entry of previous row, +1, +j ..
					k = (x+1) + j;

					// convert multi-dim array 
					// as stored in this "DataPage"
					// into a single-dim array for return ..
					entries[k] = row->at(j);
					
					// DEBUGGING ..
					//cout << "entries[k("<<k<<")]: " << entries[k] << endl;
					
					// for first row, x = -1 ..
					// for all other rows, "x" is
					// equal to the last "k index" of
					// the previous row ..
					if( j == (numberOfColumns-1) )
						x = k;
						
					// DEBUGGING ..
					//cout << "x( j("<<j<<") ): " << x << endl;
					
				}
				
			}
			
		}
		
	}
	
	return entries;
	
}
*/

void DataPage:: updateSizeOfLongestRow_Add( int rowSize ){
	
	// helper method for "addXXX" ..
	
	if( sizeOfLongestRow < rowSize )
		sizeOfLongestRow = rowSize;
	
}

void DataPage:: updateSizeOfLongestRow_Remove( int rowSize ){
	
	// helper method for "removeXXX", "replaceXXX" ..
	
	// previous operation reduced the size of a row by 1 element (rowSize),
	// so check if the update is for the row with the max size (rowSize + 1) ..
	if( sizeOfLongestRow == ( rowSize + 1 ) )
		sizeOfLongestRow = rowSize;
		
	// if the update is NOT for the row with the max size,
	// do nothing, since the max row is unchanged; still the max size ..
	
}

void DataPage:: updateSizeOfLongestRow(){
	
	// helper method for "removeEntries()" ..
	
	vector<int>* row = NULL;
	
	int maxSize = 0;
	
	if( manyRows != NULL ){
		
		int manyRowsSize = manyRows->size();
		
		for(int i=0; i<manyRowsSize; i++){
			
			row = manyRows->at( i );
			
			if( row != NULL ){
				
				if( maxSize < row->size() )
					maxSize = row->size();
				
			}

		}
		
	}
	
	sizeOfLongestRow = maxSize;
	
}

bool DataPage:: setNewRowEntry( int entry ){
	
	// helper method for "addEntry()" ..
	
	//if( entry == NULL )
	//	return;
	
	bool isSetNewRowEntry = false;
	
	if( manyRows != NULL ){
		
		// make a new row ..
		vector<int>* row = new vector<int>();
	
		// add the entry to a new row ..
		row->push_back( entry );
		
		// add the new entry/new row to the set of rows ..
		manyRows->push_back( row );
		
		// now that a new row was added,
		// update the "new row" flag ..
		isNewRow = false;
		
		// return a flag indicating that the new row entry was created ..
		isSetNewRowEntry = true;
		
		// track the longest row ..
		updateSizeOfLongestRow_Add( 1 );
	
	}
	
	return isSetNewRowEntry;
	
}

bool DataPage:: setCurrentRowEntry( int entry ){
	
	// helper method for "addEntry()" ..
	
	bool isSetCurrentRowEntry = false;
		
	if( manyRows != NULL ){
	
		int rowNumber = manyRows->size() - 1;
		
		vector<int>* row = manyRows->at( rowNumber );
		
		if( row != NULL ){
			
			row->push_back( entry );
			
			isSetCurrentRowEntry = true;
			
			// track the longest row ..
			updateSizeOfLongestRow_Add( row->size() );
			
		}
		
	}
	
	return isSetCurrentRowEntry;

}


bool DataPage:: addEntry( int entry ){
	
	// avoid adding entries to a full DataPage ..
	if( ( numberOfEntries * sizeofEntries ) >= maxPageSize )
		return false;
	
	bool isEntryAdded = false;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		if( numberOfRows <= 0 ){
			
			// no rows yet, need to initialize ..
			
			isEntryAdded = setNewRowEntry( entry );
			
		} else {
			
			// page has been initialized,
			// so see how to add to it ..
			
			if( isNewRow == true ){
				
				// this means we should start a new row ..
				
				isEntryAdded = setNewRowEntry( entry );

			} else {
				
				// this means to add the entry to the current row ..	
				
				isEntryAdded = setCurrentRowEntry( entry );
				
			}
			
		}
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( isEntryAdded )
		numberOfEntries++;
	
	return isEntryAdded;
	
}

void DataPage:: toggleNewRow(){
	
	isNewRow = true;
	
}

bool DataPage:: insertEntry( int rowNumber, int columnNumber, int entry ){
	
	// attempt to insert the "entry" at the "rowNumber" and "columnNumber",
	// insert always appears 1 column prior to "columnNumber" ..
	// (to append after the "columnNumber", use "addEntry()" ..
	// if this fails, return false ..
	
	// if "rowNumber" or "columnNumber" are out-of-range,
	// return false ..
	if( rowNumber < 0 || columnNumber < 0 )
		return false;
	
	// avoid adding entries to a full DataPage ..
	if( ( numberOfEntries * sizeofEntries ) >= maxPageSize )
		return false;
	
	bool isEntryInserted = false;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		// if "rowNumber" is out-of-range, exit ..
		if( rowNumber >= numberOfRows )
			return false;
			
		vector<int>* row = manyRows->at( rowNumber );
			
		if( row != NULL ){
			
			int rowSize = row->size();
			
			// exit if "columnNumber" is out-of-range ..
			if( columnNumber >= rowSize )
				return false;
			
			// "insert" is not "zero-based",
			// so account for this ..
			//int insertPosition = columnNumber + 1;
			
			// perform the insert ..
			//row->insert( row->begin() + insertPosition, entry );
			row->insert( row->begin() + columnNumber, entry );
			
			// track the longest row ..
			updateSizeOfLongestRow_Add( row->size() );
			
			isEntryInserted = true;
			
		}
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( isEntryInserted )
		numberOfEntries++;
	
	return isEntryInserted;
	
}

bool DataPage:: addEntries( vector<int>* entries ){
	
	if( entries == NULL )
		return false;
	
	int entriesSize = entries->size();
	
	if( entriesSize <= 0 )
		return false;
	
	// avoid adding entries to 
	// what would become a full DataPage ..
	if( ( ( numberOfEntries + entriesSize ) * sizeofEntries ) > maxPageSize )
		return false;
	
	bool areEntriesAdded = false;
	
	if( manyRows != NULL ){
		
		manyRows->push_back( entries );
		
		// track the longest row ..
		updateSizeOfLongestRow_Add( entries->size() );
		
		areEntriesAdded = true;
		
	}
	
		// track the number of entries in this "DataPage" ..
	if( areEntriesAdded )
		numberOfEntries += entries->size();
	
	return areEntriesAdded;
	
}

bool DataPage:: insertEntries( int rowNumber, vector<int>* entries ){
	
	// attempt to insert the "entries" at the "rowNumber",
	// if this fails, return false ..
	// insert always appears 1 row prior to "rowNumber" ..
	// (to append after the "rowNumber", use "addEntries()" ..
	
	// if "rowNumber" is out-of-range,
	// return false ..
	if( rowNumber < 0 )
		return false;
	
	if( entries == NULL )
		return false;
		
	int entriesSize = entries->size();
		
	if( entriesSize <= 0 )
		return false;
	
	// avoid adding entries to a full DataPage ..
	if( ( ( numberOfEntries + entriesSize ) * sizeofEntries ) > maxPageSize )
		return false;
	
	bool areEntriesInserted = false;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		// if "rowNumber" is out-of-range, exit ..
		if( rowNumber >= numberOfRows )
			return false;
		
		manyRows->insert( manyRows->begin() + rowNumber, entries );
		
		// track the longest row ..
		updateSizeOfLongestRow_Add( entries->size() );
			
		areEntriesInserted = true;
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( areEntriesInserted )
		numberOfEntries += entries->size();
	
	return areEntriesInserted;
	
}


void DataPage:: eraseRow( int rowPosition ){
	
	// helper method for "removeEntry()" and "removeEntries()" ..
	
	// if "rowPosition" is out-of-range (min), exit ..
	if( rowPosition < 0 )
		return;
		
	int manyRowsSize = manyRows->size();
	
	// if the collection is empty, exit ..
	if( manyRowsSize <= 0 )
		return;
	
	// if "rowPosition" is out-of-range (max), exit ..
	if( rowPosition >= manyRowsSize )
		return;
	
	manyRows->erase( manyRows->begin() + rowPosition );
	
}


bool DataPage:: removeEntry(){
	
	// helper method for "addEntry()" ..
	// removes the "latest" entry in the vector ..
	
	bool isEntryRemoved = false;
		
	if( manyRows != NULL ){
		
		if( manyRows->size() > 0 ){
			
			int rowNumber = manyRows->size() - 1;
		
			vector<int>* row = manyRows->at( rowNumber );
		
			if( row != NULL ){
				
				int removePosition = row->size() - 1;
			
				row->erase( row->begin() + removePosition );
				
				// if the row is empty, delete it ..
				if( row->size() <= 0 )
					eraseRow( manyRows->size() - 1 );
				
				// track the longest row ..
				updateSizeOfLongestRow_Remove( row->size() );
			
				isEntryRemoved = true;
			
			}	
			
		}
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( isEntryRemoved )
		numberOfEntries--;
	
	return isEntryRemoved;

}

bool DataPage:: removeEntry( int rowNumber, int columnNumber ){
	
	// removes the entry located at the "rowNumber" and "columnNumber" ..
	
	// if "rowNumber" or "columnNumber" are out-of-range,
	// return false ..
	if( rowNumber < 0 || columnNumber < 0 )
		return false;
	
	bool isEntryRemoved = false;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		// if "rowNumber" is out-of-range, exit ..
		if( rowNumber >= numberOfRows )
			return false;
			
		vector<int>* row = manyRows->at( rowNumber );
			
		if( row != NULL ){
			
			int rowSize = row->size();
			
			// exit if "columnNumber" is out-of-range ..
			if( columnNumber >= rowSize )
				return false;
			
			// perform the removal ..
			row->erase( row->begin() + columnNumber );
			
			// if the row is empty, delete it ..
			if( row->size() <= 0 )
				eraseRow( rowNumber );
			
			// track the longest row ..
			updateSizeOfLongestRow_Remove( row->size() );
			
			isEntryRemoved = true;
			
		}
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( isEntryRemoved )
		numberOfEntries--;
	
	return isEntryRemoved;

}


bool DataPage:: removeEntries(){
	
	// helper method for "addEntries()" ..
	// removes the "latest" row of entries in the vector ..
	
	bool areEntriesRemoved = false;
	
	int rowToRemoveSize = 0;
	
	if( manyRows != NULL ){
		
		//cout << "manyRows->: " << manyRows->size() << endl << endl;
		
		// size of row to be removed ..
		rowToRemoveSize = manyRows->at( ( manyRows->size() - 1 ) )->size();
		
		eraseRow( manyRows->size() );
		
		// track the longest row ..
		// (removed a row, so re-calculate max) ..
		updateSizeOfLongestRow();
		
		areEntriesRemoved = true;
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( areEntriesRemoved )
		numberOfEntries -= rowToRemoveSize;
	
	return areEntriesRemoved;
	
}


bool DataPage:: removeEntries( int rowNumber ){
	
	// removes the row of entries located at the "rowNumber" ..
	
	// if "rowNumber" is out-of-range,
	// return false ..
	if( rowNumber < 0 )
		return false;
	
	bool areEntriesRemoved = false;
	
	int rowToRemoveSize = 0;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		// if "rowNumber" is out-of-range, exit ..
		if( rowNumber >= numberOfRows )
			return false;
			
		// "erase" is not "zero-based",
		// so account for this ..
		//int removePosition = rowNumber + 1;
			
		// size of row to be removed ..
		rowToRemoveSize = manyRows->at( rowNumber )->size();
			
		//manyRows->erase( manyRows->begin() + removePosition );
		manyRows->erase( manyRows->begin() + rowNumber );
			
		// track the longest row ..
		// (removed a row, so re-calculate max) ..
		updateSizeOfLongestRow();
			
		areEntriesRemoved = true;
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( areEntriesRemoved )
		numberOfEntries -= rowToRemoveSize;
	
	return areEntriesRemoved;

}


bool DataPage:: replaceEntries( int rowNumber, vector<int>* entries ){
	
	// if "rowNumber" is out-of-range,
	// return false ..
	if( rowNumber < 0 )
		return false;
	
	// if "entries" is null, exit ..
	if( entries == NULL )
		return false;
	
	int entriesSize = entries->size();
	
	// if "entries" are emtpy, exit ..
	if( entriesSize <= 0 )
		return false;
	
	bool areEntriesReplaced = false;
	
	int oldRowSize = 0;
	int newRowSize = 0;
	
	if( manyRows != NULL ){
		
		// if "rowNumber" is out-of-bounds, exit ..
		if( rowNumber >= manyRows->size() )
			return false;
		
		vector<int>* row = manyRows->at( rowNumber );
		
		// size of old row ..
		oldRowSize = row->size();
		
		row->swap( (*entries) );
		
		// size of new row ..
		newRowSize = entries->size();
		
		// track the longest row ..
		// (replaced a row, so re-calculate max) ..
		updateSizeOfLongestRow();
		
		areEntriesReplaced = true;
		
	}
	
	// track the number of entries in this "DataPage" ..
	if( areEntriesReplaced ){
		
		numberOfEntries -= oldRowSize;
		numberOfEntries += newRowSize;
		
	}
	
	return areEntriesReplaced;
	
}


int DataPage:: findFirstEntryInstance( int entry, int columnNumber ){
	
	// looks for an "entry" in a specific "columnNumber",
	// if it finds the "entry",
	// returns the first "rowNumber" with this "entry" / "columnNumber" ..
	
	// if "entry" is out-of-range, exit ..
	if( entry == NULL )
		return -1;
	
	// if "columnNumber" is out-of-range (min), exit ..
	if( columnNumber < 0 )
		return -1;
	
	// if "columnNumber" is out-of-range (max), exit ..
	if( columnNumber >= sizeOfLongestRow )
		return -1;
	
	int rowNumber = -1;
		
	if( manyRows != NULL ){
		
		int manyRowsSize = manyRows->size();
		
		if( manyRowsSize > 0 ){
			
			vector<int>* row = NULL;
			
			for(int i=0; i<manyRowsSize; i++){
				
				row = manyRows->at( i );
				
				// if "columnNumber" is out-of-range,
				// go on to the next row ..
				if( columnNumber >= row->size() )
					continue;
				
				// if row is found,
				// return its rowNumber ..
				if( row->at( columnNumber ) == entry ){
					
					rowNumber = i;
					
					break;
					
				}
				
			}
			
		}
		
	}
	
	return rowNumber;
	
}

vector<int>* DataPage:: findAllEntryInstances( int entry, int columnNumber ){
	
	// looks for an "entry" in a specific "columnNumber",
	// if it finds the "entry", stores it in a collection,
	// continues until it finds every instance in this DataPage,
	// returns the collections of "rowNumbers" for this "entry" / "columnNumber" ..
	
	//cout << "findAllEntryInstances() Point A .." << endl << endl;
	
	// if "entry" is out-of-range, exit ..
	if( entry == NULL )
		return NULL;
	
	// if "columnNumber" is out-of-range (min), exit ..
	if( columnNumber < 0 )
		return NULL;
	
	// if "columnNumber" is out-of-range (max), exit ..
	if( columnNumber >= sizeOfLongestRow )
		return NULL;
		
	vector<int>* rowEntries = NULL;
	
	int rowNumber = -1;
		
	if( manyRows != NULL ){
		
		int manyRowsSize = manyRows->size();
		
		if( manyRowsSize > 0 ){
			
			vector<int>* row = NULL;

			for(int i=0; i<manyRowsSize; i++){
				
				row = manyRows->at( i );
				
				// if "columnNumber" is out-of-range,
				// go on to the next row ..
				if( columnNumber >= row->size() )
					continue;
				
				// if row is found,
				// store it and,
				// continue to the next row ..
				if( row->at( columnNumber ) == entry ){
					
					// initialize "rowEntries" if it is not ..
					if( rowEntries == NULL )
						rowEntries = new vector<int>();
					
					rowEntries->push_back( i );
					
				}
				
			}
			
		}
		
	}
	
	return rowEntries;
	
}

int DataPage:: getNumberOfRows(){
	
	int numberOfRows = -1;
	
	if( manyRows != NULL ){
		
		numberOfRows = manyRows->size();
		
	}
	
	return numberOfRows;
	
}

int DataPage:: getNumberOfColumns( int rowNumber ){
	
	if( rowNumber < 0 )
		return -1;
	
	int numberOfColumns = -1;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		//    rowNumber: 0 .. N
		// numberOfRows: 1 .. N+1
		if( rowNumber < numberOfRows ){
			
			vector<int>* row = manyRows->at( rowNumber );
			
			if( row != NULL ){
				
				numberOfColumns = row->size();	
				
			}
			
		}
		
	}
	
	return numberOfColumns;
	
}

void DataPage:: printDataPage(){
	
	// debugging method ..
	
	int entry = -1;
	
	if( manyRows != NULL ){
		
		int numberOfRows = manyRows->size();
		
		if( numberOfRows > 0 ){
			
			int numberOfColumns = -1;
			
			// check each row of the DataPage ..
			for(int i=0; i<numberOfRows; i++){
			
				vector<int>* row = manyRows->at(i);
				
				if( row != NULL ){
					
					numberOfColumns = row->size();
					
					if( numberOfColumns > 0 ){
						
						// check each column for this row of the DataPage ..
						for(int j=0; j<numberOfColumns; j++){
							
							entry = row->at(j);
							 
							// print the current entry ..
							// 1) if its the last entry in the row,
							//    don't add a space ..
							// 2) if its NOT the last entry in the row,
							//    need a space between this entry and the next ..
							if( j == (numberOfColumns-1) )
								cout << entry;
							else
								cout << entry << " ";
							
						}
						
						// end of this row ..
						cout << endl;
						
					}
					
				}
			
			}	
			
		} else {
			
			cout << "EMPTY DATAPAGE .." << endl << endl;
			
		}
		
	}	
	
}


void DataPage:: clearPage(){
	
	// erase the contents of the page ..
	
	numberOfEntries = 0;
	
	sizeOfLongestRow = 0;
	
	isNewRow = true;
	
	manyRows->clear();
	
}


/*

#include "DataPage.h"

DataPage:: DataPage(void){
}

DataPage:: DataPage( int recordCount ){
	
	// need to figure out a way to limit "recordCount" to a number that
	// is less than the maximum that would be allowed to be written to the page-from-file
	// ( i.e.: whatever the max turns out to be from "hp.computeMaxRowCount()" or
	//        "bit.computeMaxRowCount( flushBlockIndex )" ) ..
	// NOTE: this is done is code prior to instantiation of this object - pre-condition for use
	//       ("programming by contract" condition) ..
	
	for (int i=0; i < recordCount ; i++)
		records.push_back( ( * new DataPageRecord() ) );
	
}

DataPage:: ~DataPage(void){
}

vector<DataPageRecord>* DataPage:: getRecords(){
	
	return &records;
	
}

*/
