#ifndef DATAPAGE_H_
#define DATAPAGE_H_

using namespace std;

#include <vector>
#include <iostream>

class DataPage {
	
public:

	DataPage();
	DataPage( DataPage* dataPage );
	virtual ~DataPage();
	
	int getPageID();
	void setPageID( int pageID_ );
	
	int getMaxPageSize();
	void setMaxPageSize( int maxPageSize_ );
	
	int getNumberOfEntries();
	
	int getSizeofEntries();
	
	int getSizeOfLongestRow();
	
	bool getIsNewRow();
	
	//    rowNumber: 0 .. N
	// columnNumber: 0 .. M
	int getEntry( int rowNumber, int columnNumber );
	
	vector< vector<int>* >* getEntries();
	vector<int>* getEntries( int rowNumber );

	void updateSizeOfLongestRow_Add( int rowSize );
	void updateSizeOfLongestRow_Remove( int rowSize );
	void updateSizeOfLongestRow();

	bool addEntry( int entry );
	void toggleNewRow();	// used with "addEntry" to specify beginning of a new row ..
	
	bool insertEntry( int rowNumber, int columnNumber, int entry );
	bool addEntries( vector<int>* entries );
	bool insertEntries( int rowNumber, vector<int>* entries );
	
	void eraseRow( int rowPosition );
	bool removeEntry();
	bool removeEntry( int rowNumber, int columnNumber );
	bool removeEntries();
	bool removeEntries( int rowNumber );
	
	bool replaceEntries( int rowNumber, vector<int>* entries );
	
	int findFirstEntryInstance( int entry, int columnNumber );
	vector<int>* findAllEntryInstances( int entry, int columnNumber );
	
	int getNumberOfRows();
	
	//       rowNumber: 0 .. N
	//    numberOfRows: 1 .. N+1
	// numberOfColumns: 1 .. M+1
	int getNumberOfColumns( int rowNumber );
	
	void printDataPage();
	
	void clearPage();
	
private:

	// helper methods ..
	bool setNewRowEntry( int entry );
	bool setCurrentRowEntry( int entry );
	
	// id / location in DataPage file to move file pointer
	// in order to locate this DataPage ..
	int pageID;
	
	// max size of the page, in bytes ..
	// ( maxNumberEntries = maxPageSize / sizeofEntries ) ..
	int maxPageSize;
	
	// current number of entries in the DataPage ..
	int numberOfEntries;
	
	// since all entries are type "int",
	// track the size of the entries for use during DataPage read/write ..
	// (NOTE: future extensions with attributes of other types, char, string, float,
	//        etc. would need an array that tracks each attribute type, here) ..
	int sizeofEntries;
	
	// track the length of the longest row,
	// useful to avoid out-of-range operations ..
	int sizeOfLongestRow;
	
	// track whether or not "addEntry" or "addEntries"
	// is starting with a "new row" ..
	bool isNewRow;
	
	// collection of the rows of entries ..
	vector< vector<int>* >* manyRows;
	
};

#endif /*DATAPAGE_H_*/


/*
#ifndef DATAPAGE_H_
#define DATAPAGE_H_

#pragma once

#include "DataPageRecord.h"
#include "DataPageField.h"
#include <vector>

using namespace std;
	
class DataPage {
	
public:

	DataPage(void);
	DataPage(int recordCount);
	~DataPage(void);
	
	vector<DataPageRecord>* getRecords();
	
private:
			
	vector<DataPageRecord> records;   // number of records possible in a page
									  // depends upon the number of fields ..
					        		  // length determined and limited using 
					        		  // "computeMaxRowCounts()" in either 
					        		  // "DataHeader" or "BufferInfoTable" ..
					   
};


#endif /*DATAPAGE_H_*/

