#include "IndexBucketHeader.h"

IndexBucketHeader:: IndexBucketHeader(){
	
	_indexBucketIDsForEachSearchKey = new map<int, vector<int>* >();
	
	_headerEntries = new vector<IndexBucketHeaderEntry*>();
	
}

IndexBucketHeader:: ~IndexBucketHeader(){
	
	if( _indexBucketIDsForEachSearchKey != NULL ) delete _indexBucketIDsForEachSearchKey;
	
	if( _headerEntries != NULL ) delete _headerEntries;
	
}


bool IndexBucketHeader:: hasSearchKey( int searchKey_ ){
	
	bool bucketHasSearchKey = false;
	
	map<int, vector<int>* >:: iterator it_hsk;
	for(it_hsk = _indexBucketIDsForEachSearchKey->begin(); it_hsk != _indexBucketIDsForEachSearchKey->end(); it_hsk++){
		
		if( it_hsk->first == searchKey_ ){
			
			bucketHasSearchKey = true;
			break;
			
		}
		
	}
	
	
	return bucketHasSearchKey;
	
	
} // end "hasSearchKey()" ..


void IndexBucketHeader:: addBucketIDsForOneSearchKey( int searchKey_, vector<int>* indexBucketIDsForOneSearchKey_ ){
		
	/*
	 * If the new "bucketIDs" are not already in the bucketIDs collection,
	 * this method adds the new bucketIDs to the collection.
	 * This method does NOT replace old bucketIDs with new bucketIDs ..
	 */
		
	vector<int>* newBucketIDs = NULL;
	
	vector<int>* currentBucketIDs = NULL;
	
	bool searchKeyFound = false;
	bool bucketIDFound = false;
	bool addedBucketID = false;

	
	// check to see if the searchKey / bucketIDs pair already exists in the collection,
	// if it does, update the bucketIDs with the passed in set ..
	map<int, vector<int>* >:: iterator it_bifosk;
	for(it_bifosk = _indexBucketIDsForEachSearchKey->begin(); it_bifosk != _indexBucketIDsForEachSearchKey->end(); it_bifosk++){
		
		
		if( it_bifosk->first == searchKey_ ){
			
			searchKeyFound = true;
			
			newBucketIDs = indexBucketIDsForOneSearchKey_;
			currentBucketIDs = ( *_indexBucketIDsForEachSearchKey )[ searchKey_ ];
			
			// iterate through "currentBucketIDs", see if it has any "newBucketIDs" ..
			for(int i=0; i<newBucketIDs->size(); i++){
				
				for(int j=0; j<currentBucketIDs->size(); j++){
				
					if( currentBucketIDs->at(j) == newBucketIDs->at(i) ){
						
						bucketIDFound = true;
						break;
						
					}
					
				}
				
				// if "currentBucketIDs" doesnt have the "newBucketID", ok to add it ..
				if( bucketIDFound == false ) {
					
					currentBucketIDs->push_back( newBucketIDs->at(i) );
					
					addedBucketID = true;
					
				} else
					bucketIDFound = false;	// reset for next "newBucketID" iteration ..
				
			} // end "for()" ..
			
		} // end "if" ..
		
	} // end "for()" ..
	
	if( addedBucketID == true ){
		
		// remove the current collection of "bucketIDs" ..
		
		_indexBucketIDsForEachSearchKey->erase( searchKey_ );
		
		// add the new collection of "bucketIDs",
		// "currentBucketIDs" should now contain both old & newly added bucketIDs ..
		_indexBucketIDsForEachSearchKey->insert( pair< int, vector<int>* >( searchKey_, currentBucketIDs ) );
		
	}
	
	// if searchKey isn't already in the collection, ok to add bucketIDs ..
	else if (searchKeyFound == false)
		_indexBucketIDsForEachSearchKey->insert( pair< int, vector<int>* >( searchKey_, indexBucketIDsForOneSearchKey_ ) );


} // end "addBucketIDsForOneSearchKey()" ..


void IndexBucketHeader:: replaceBucketIDsForOneSearchKey( int searchKey_, vector<int>* indexBucketIDsForOneSearchKey_ ){
		
	/*
	 * This method removes the old bucketIDs and replaces them with new bucketIDs ..
	 */
		
	vector<int>* newBucketIDs = NULL;
	
	vector<int>* currentBucketIDs = NULL;

	
	// check to see if the searchKey / bucketIDs pair already exists in the collection,
	// if it does, update the bucketIDs with the passed in set ..
	map<int, vector<int>* >:: iterator it_bifosk;
	for(it_bifosk = _indexBucketIDsForEachSearchKey->begin(); it_bifosk != _indexBucketIDsForEachSearchKey->end(); it_bifosk++){
		
		
		if( it_bifosk->first == searchKey_ ){
			
			// remove the current collection of "bucketIDs" ..
			_indexBucketIDsForEachSearchKey->erase( searchKey_ );
			
			// add the relacement "bucketIDs" to the collection ..
			_indexBucketIDsForEachSearchKey->insert( pair< int, vector<int>* >( searchKey_, indexBucketIDsForOneSearchKey_ ) );
			
		} // end "if" ..
		
	} // end "for()" ..


} // end "replaceBucketIDsForOneSearchKey()" ..


void IndexBucketHeader:: removeBucketIDsForOneSearchKey( int searchKey_ ){
		
	// remove the current collection of "bucketIDs" ..
	_indexBucketIDsForEachSearchKey->erase( searchKey_ );


} // end "removeBucketIDsForOneSearchKey()" ..


/*
 * GETTERS .. 
 */
	
int IndexBucketHeader:: getTotalNumberBuckets(){
	
	return _totalNumberOfBuckets;	
	
}

map<int, vector<int>* >* IndexBucketHeader:: getBucketIDsForEachSearchKey(){
	
	return _indexBucketIDsForEachSearchKey;
	
}

	
int IndexBucketHeader:: getNumberOfBucketsForOneSearchKey( int searchKey_ ){
	
	int numberOfBuckets = -1;
	
	vector<int>* bucketIDs = NULL;
	
	map<int, vector<int>* >:: iterator it_nobfosk;
	for(it_nobfosk = _indexBucketIDsForEachSearchKey->begin(); it_nobfosk != _indexBucketIDsForEachSearchKey->end(); it_nobfosk++){
		
		if( it_nobfosk->first == searchKey_ ){
			
			bucketIDs = it_nobfosk->second;
			
			numberOfBuckets = bucketIDs->size();
			
			break;
			
		}
		
	}
	
	
	return numberOfBuckets;
	
	
} // end "getNumberOfBucketsForOneSearchKey()" ..

	
vector<int>* IndexBucketHeader:: getBucketIDsForOneSearchKey( int searchKey_ ){
	
	vector<int>* bucketIDs = NULL;
	
	map<int, vector<int>* >:: iterator it_bifosk;
	for(it_bifosk = _indexBucketIDsForEachSearchKey->begin(); it_bifosk != _indexBucketIDsForEachSearchKey->end(); it_bifosk++){
		
		if( it_bifosk->first == searchKey_ ){
			
			bucketIDs = it_bifosk->second;
			break;
			
		}
		
	}
	
	return bucketIDs;
	
} // end "getBucketIDsForOneSearchKey()" ..

	
vector<IndexBucketHeaderEntry*>* IndexBucketHeader:: getHeaderEntries(){
	
	return _headerEntries;
	
}
	
	
/*
 * SETTERS .. 
 */
	
void IndexBucketHeader:: setTotalNumberBuckets( int totalNumberOfBuckets_ ){
	
	_totalNumberOfBuckets = totalNumberOfBuckets_;
	
}
	
void IndexBucketHeader:: setBucketIDsForEachSearchKey( map<int, vector<int>* >* indexBucketIDsForEachSearchKey_ ){
	
	_indexBucketIDsForEachSearchKey = indexBucketIDsForEachSearchKey_;
	
}

	
void IndexBucketHeader:: setBucketIDsForOneSearchKey( int searchKey_, vector<int>* indexBucketIDsForOneSearchKey_ ){
		
	/*
	 * This method finds the "bucketIDs" in the collection for the "searchKey",
	 * and over-writes them with the new "bucketIDs".
	 * If the "bucketIDs" are not found for the "searchKey",
	 * the "bucketIDs" are just added to the collection ..
	 */
	
	int counter = 0;
	
	bool searchKeyFound = false;
	
	// check to see if the searchKey / bucketIDs pair already exists in the collection,
	// if it does, update the bucketIDs with the passed in set ..
	map<int, vector<int>* >:: iterator it_bifosk;
	for(it_bifosk = _indexBucketIDsForEachSearchKey->begin(); it_bifosk != _indexBucketIDsForEachSearchKey->end(); it_bifosk++){
		
		counter++;
		
		if( it_bifosk->first == searchKey_ ){
			
			searchKeyFound = true;
			
			( *_indexBucketIDsForEachSearchKey )[ searchKey_ ] = indexBucketIDsForOneSearchKey_;
			
		}
		
	}
	
	// if searchKey / bucketIDs pair isn't already in the collection, add it ..
	if (searchKeyFound == false)
		_indexBucketIDsForEachSearchKey->insert( pair< int, vector<int>* >( searchKey_, indexBucketIDsForOneSearchKey_ ) );

} // end "setNumberOfIndexBucketsForEachSearchKey()" ..

	
void IndexBucketHeader:: setHeaderEntries( vector<IndexBucketHeaderEntry*>* headerEntries_ ){
	
	_headerEntries = headerEntries_;
	
}
