#ifndef INDEXBUCKETHEADERENTRY_H_
#define INDEXBUCKETHEADERENTRY_H_

using namespace std;

#include <vector>


class IndexBucketHeaderEntry {
	
public:

	IndexBucketHeaderEntry();
	~IndexBucketHeaderEntry();
	
	/*
	 * GETTERS .. 
	 */
	
	int getBucketID();
	
	int getSearchKey();
	
	int getFreePointerCount();
	
	//vector<bool>* getPointersAvailable();
	
	
	/*
	 * SETTERS .. 
	 */
	
	void setBucketID( int bucketID_ );
	
	void setSearchKey( int searchKey_ );
	
	void setFreePointerCount( int freePointerCount_ );
	
	//void setPointersAvailable( vector<bool>* pointersAvailable_ );
	
	
private:


	int bucketID; 							// Page id of an index bucket page that is being pointed to ..
											// This is the "page-from-file" index number..

	int searchKey;							// This is the search key associated with this bucket id ..

	int freePointerCount;					// Analogous to "freeRowCount" in "DataHeaderEntry",
											// this indicates the number of free pageID-pointers available in the bucket. 
											// This enables us to find out the free space in a bucket without 
											// actually reading the bucket ..
											// NOTE:
											// (always have at least 2 leaf ptrs - left and right) ..
											// i.e.: 
											//		 "P" = leaf pointer ..
											//		 "K" = search key ..
											//		 P0 K0 P1
											//       P0 K0 P1 K1 P2
											//       P0 K0 P1 K1 P2 K2 P3
											//		 P0 K0 P1 K1 P2 K2 P3 K3 P4
											//		 P0 K0 P1 K1 P2 K2 P3 K3 P4 K4 P5
	
	
	//vector<bool>* pointersAvailable;		// Analogous to "rowAvailable" in "DataHeaderEntry",
											// tracks whether an individual pageID-pointer in a IndexBucket is empty or occupied:
											//    "free / available" (true): the pageID-pointer is empty or has garbage value
											// "full / unavailable" (false): the pageID-pointer is occupied by valid data
											// 
											// 1 bytes * 12 cells = 12 bytes ..
											// each entry of the vector is 1 byte (bool) ..
	
};

#endif /*INDEXBUCKETHEADERENTRY_H_*/
