#ifndef INDEXHEADERENTRY_H_
#define INDEXHEADERENTRY_H_

using namespace std;

#include <vector>

class IndexHeaderEntry {
	
public:

	IndexHeaderEntry();
	~IndexHeaderEntry();
	
	/*
	 * GETTERS .. 
	 */
	
	int getPageID();
	
	int getNumberOfSearchKeys();
	
	int getNumberOfChildPageIDs();
	
	int getParentPageID();
	
	int getFreeSearchKeyCount();
	
	vector<bool>* getSearchKeysAvailable();
	
	
	/*
	 * SETTERS .. 
	 */
	
	void setPageID( int pageID_ );
	
	void setNumberOfSearchKeys( int numberOfSearchKeys_ );
	
	void setNumberOfChildPageIDs( int numberOfChildPageIDs_ );
	
	void setParentPageID( int parentPageID_ );
	
	void setFreeSearchKeyCount( int freeSearchKeyCount_ );
	
	void setSearchKeysAvailable( vector<bool>* searchKeysAvailable_ );
	
	
private:

	int pageID; 							// Page id of an index page that is being pointed to ..
											// This is the "page-from-file" index number..

	int numberOfSearchKeys;
	
	int numberOfChildPageIDs;


	int parentPageID;						// each B+ Tree node has 1 parent node, this stores each node's parent PageID ..
											// The root node doesn't have a parent (parentPageID = -1) ..
											
											
	int freeSearchKeyCount;					// Analogous to "freeRowCount" in "DataHeaderEntry",
											// this indicates the number of free search keys available in the page. 
											// This enables us to find out the free space on a page without 
											// actually reading the page ..
											// NOTE:
											// (always have at least 2 leaf ptrs - left and right) ..
											// i.e.: 
											//		 "P" = leaf pointer ..
											//		 "K" = search key ..
											//		 P0 K0 P1
											//       P0 K0 P1 K1 P2
											//       P0 K0 P1 K1 P2 K2 P3
											//		 P0 K0 P1 K1 P2 K2 P3 K3 P4
											//		 P0 K0 P1 K1 P2 K2 P3 K3 P4 K4 P5
	
	
	vector<bool>* searchKeysAvailable;		// Analogous to "rowAvailable" in "DataHeaderEntry",
											// tracks whether an individual search key in an IndexPage is empty or occupied:
											// "free / available" (true): the search key is empty or has garbage value
											// "full / unavailable" (false): the search key is occupied by valid data
											// 
											// 1 byte * 4 (search key) cells = 4 bytes ..
											// each entry of the vector is 1 byte (bool) ..

	
};

#endif /*INDEXHEADERENTRY_H_*/
