#include "BufferManager.h"

	
BufferManager:: BufferManager( int bufferPoolSize, StorageManager* sm_ ) {
	
	//cout << "BufferManager:: BufferManager() start .." << endl << endl;
	
	sm = sm_;
	
	lockReactivationOn = false;
	
	locks = new map< string, vector<int>* >();
	
	// this initializes the "bp" (BufferPool) ..
	//bp.setBufferPoolSize( 3 );
	bp = new BufferPool();
	bp->setBufferPoolSize( bufferPoolSize );
	for(int i=0; i<bufferPoolSize; i++){
		
		bp->getDataPages()->at(i)->setMaxPageSize( blockSize );
		bp->getDataPages()->at(i)->setPageID( -1 );
		
	}
	
	// this initializes the "bit" (BufferInfoTable) ..
	//bit.setBufferPoolSize( 3 );
	bit = new BufferInfoTable();
	bit->setBufferPoolSize( bufferPoolSize );
	
	bitSnapShot = new BITSnapShot();
	
	lastFlushedBlocks = new vector<DataPage*>();
	
	lastFlushedBlockPageIDs = new vector<int>();
	
	//cout << "BufferManager:: BufferManager() end .." << endl << endl;

}


BufferManager:: ~BufferManager(void) {
	
	//cout << "~BufferManager() start .." << endl << endl;
	
	// clear all the current locks that this app has on table pages,
	// tracked in the BufferInfoTable ("bit") ..
	// NOTE: if this doesn't work properly, pages will be locked in the file
	// permanently as soon as the user quits the app.
	// restarting the app, the user would not have the key to unlock the page ..

	// DEBUGGING ..
	//cout << "~BufferManager():: is locks empty?: " << locks.empty() << endl << endl;
	
	if ( !locks->empty() ) {
		
		// this indicates that the BP for this app has locks on 
		// data pages that need to be removed ..
		
		// this app is closing, and other users need access to 
		// the pages-from-file ..
		
		vector<int>* pageIDs = new vector<int>();
		
		string tableNameString = "";
	
		map< string, vector<int>* >::iterator it_locks;
		for( it_locks = locks->begin(); it_locks != locks->end(); it_locks++ ){
			
/*			DEBUGGING ..
			cout << "(*it_locks).first: \"" << ((char*)(*it_locks).first) << "\"" << endl << endl;
			vector<int> vectorInts( (*it_locks).second );
			vector<int>::iterator it_vectorInts;
			for( it_vectorInts = vectorInts.begin(); it_vectorInts != vectorInts.end(); it_vectorInts++ ){
				cout << "(*it_locks).second: \"" << (int)(*it_vectorInts) << "\"" << endl << endl;
			}
			cout << "locks.size(): " << locks.size() << endl;
			cout << "~BM():: is locks empty?: " << locks.empty() << endl << endl;
*/
		
			// workaround:
			// ignore "locks" for "tableName" that aren't recorded in the catalog,
			// because they aren't in the catalog anyway!
			// NOTE: see "purgeLocksMap()" for further explanation ..
			// this workaround prevents creation of "bad" writes to "DataHeader" file
			// via the "sm->clearTablePageLocks()" method call, which occur when
			// attempt made to clear a lock for a "tableName" that is in "locks" but
			// not in the catalog ..
			tableNameString = (*it_locks).first;

			if ( getCatalogManager()->doesCatalogTableNameExist( tableNameString ) ) {
				
				// DEBUGGING ..
				//cout << "sm->clearTablePageLocks( " << ( (char*)(*it_locks).first ) << ", pageID ) called .." << endl << endl;
				
				// clear the locks in the catalog that this user / app had on tables prior to shutting down ..
				getStorageManager()->clearTablePageLocks( it_locks->first, it_locks->second );
		
			}
		
		} // end "for()" ..
		
		
	} // end "if ( !locks.empty() )" ..
	
	
	// DEBUGGING ..
	//cout << "~BufferManager(void) called .." << endl << endl;
	
	// cleanup nested objects ..
	delete sm->getCatalogManager();					// CatalogManager is nested inside StorageManager, responsible for destruction here, though ..
	//delete sm;					// this is done in QueryManager ..
	
	
	//cout << "~BufferManager() end .." << endl << endl;


} // end "BufferManager:: ~BufferManager()" ..


StorageManager* BufferManager:: getStorageManager(){
	
	return sm;
	
}


CatalogManager* BufferManager:: getCatalogManager(){
	
	return sm->getCatalogManager();
	
}


MessageManager* BufferManager:: getMessageManager(){
	
	return sm->getCatalogManager()->getMessageManager();
	
}


void BufferManager:: purgeLocksMap( map< string, vector<int>* >* locks ){
	
	if( locks == NULL )
		return;
	
	/*
	 * This method should be invoked immediately prior to performing
	 * _any_ and _every_ operation on the "locks" map, 
	 * in order to prevent "garbage" keys or values from creeping into the map ..
	 * (this is a workaround, see below).
	 */
	
	// DEBUGGING ..
	//cout << "start purgeLocksMap() .." << endl << endl;
	
	string tableNameString = "";
	
	map< string, vector<int>* >::iterator it_locks;
	
	// DEBUGGING ..
	//cout << "locks.size(): " << locks.size() << endl << endl;
	
	int counter = locks->size();
	for( it_locks = locks->begin(); it_locks != locks->end(); it_locks++ ){
	
		if ( counter <= 0 ){
			// Because invoking "locks.erase()" inside of an iteration through "locks"
			// messes with the "iterator" algorithm, leave this block once the "counter"
			// decrements to 0 (or just in case, less than 0), which indicates that the "iterator"
			// has reached the end of the original "locks" map ..
			// (avoids performing an infinite loop when the last element in the map is erased) ..

			break;
			
		}
		
		tableNameString = (*it_locks).first;
		
		// DEBUGGING ..
		//cout << endl << "typeid( (*it_locks).second ).name(): " << typeid( (*it_locks).second ).name() << endl << endl;
		//cout << endl << "typeid( vector<int> ).name(): " << typeid( vector<int> ).name() << endl << endl;
		
		if ( !getCatalogManager()->doesCatalogTableNameExist( tableNameString ) || 
		     ( ( it_locks->second )->size() == 0 ) ) {	
			
			
				
			// if the "tableName" in "locks" doesn't exist in the catalog, then
			// we should remove the key/value pair associated with the "tableName" from "locks" ..
			// OR
			// if the vector<int> map "value" that holds the pageID (locked pages associated with "tableName" key),
			// is "size == 0" (no pageID), then this is an error, so remove this key/value pair ..
			
			// NOTE: this is a workaround, due to a malfunction in "locks" ..
			// for some unknown reason, when "locks" is empty, following the "writeBlock()" method call,
			// "locks" acquires the garbage value of key="", value=NULL; 
			// this workaround (along with workaround in destructor) is implemented in order to circumvent this problem.
			
			// DEBUGGING ..
			//cout << endl << "erasing lock on table: " << (*it_locks).first << endl << endl;
			
			// this erases both the "key" and its "value" from the "locks" map ..
			locks->erase( (*it_locks).first );
			
			if ( locks->size() <= 0 ) {
				
				// Because invoking "locks.erase()" inside of an iteration through "locks"
				// messes with the "iterator" algorithm, leave this block once the size of
				// "locks" is reduced to 0 (or just in case, less than 0) ..
				break;
				
			}
			
			counter--;
			
			// DEBUGGING ..
			//cout << "point A" << endl << endl;
			
		}
		
		// DEBUGGING ..
		//cout << "point B" << endl << endl;
		
		
	} // end "for()" ..
	
	
	// DEBUGGING ..
	//cout << "end purgeLocksMap() .." << endl << endl;
	
	
} // end "purgeLocksMap( map< char*, vector<int> > &locks )" .


bool BufferManager:: writeBlock( DataPage* dataPage, map< string, vector<int>* >* locks ) {
	
	if( dataPage == NULL )
		return false;
		
	if( locks == NULL )
		return false;
	
	// a) If "pg" is composed of a page loaded in from file, it had a lock on it ..
	// b) Otherwise, "pg" is a completely new block, which doesn't have a lock
	//    (b/c a new block doesnt have space for it allocated in the file yet) ..
	
	// DEBUGGING ..
	//cout << "BufferManager():: writeBlock_1: is locks empty?: " << locks.empty() << endl << endl;
	
	int blockIndex = BufferInfoTable::bitIndex;
	
	string tableName = "";
	
	if( blockIndex >= 0 ) 
		tableName = bit->getTableName()->at( blockIndex );
	
	//cout << "BufferManager:: writeBlock() blockIndex: " << blockIndex << endl << endl;
	
	//cout << "BufferManager:: writeBlock() tableName: " << tableName << endl << endl;
	
	int pageID = -1;
	
	if( blockIndex >= 0 )
		pageID = bit->getPageID()->at( blockIndex );
	
	//cout << "IndexPages:: writeBlock() Point A .." << endl << endl;

	bool writePagePerformed = getStorageManager()->writeDataPage( dataPage, bit );
	
	// if "writeDataPage()" fails, abort "writeBlock()" ..
	if ( writePagePerformed == false ) {
		
	 	string errorString = "\nFailed to write data page [ BM::writeBlock() ], data may not have been written to file.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );
		
		return false;
		
	}
/*

	// This is done in "flushBPBlock()" ..
	
	//int maxRowCount = bit.computeMaxRowCount( blockIndex );
	int numberOfRowsPossible = bit->getRowAvailable()->at( blockIndex )->size();
	
	// NOTE: if "numberOfSlots" changes, say because the data in this bp block
	// is from a different table, which has different number of slots, 
	// is "bit.rowAvailable[ blockIndex ][i]" update properly elsewhere???
	
	//for(int i=0; i<maxRowCount; i++){	
	for(int i=0; i<numberOfRowsPossible; i++){
		
		// reset the "bit"
		//(* bit.getRowAvailable() )[ blockIndex ][i] = true;
		bit->getRowAvailable()->at( blockIndex )->at(i) = true;
		
	}
*/
	
	// DEBUGGING ..
	//cout << "BufferManager():: writeBlock_2: is locks empty?: " << locks.empty() << endl << endl;
	
	//cout << "IndexPages:: writeBlock() Point B .." << endl << endl;

	// now that the data is written to file, release any locks that were
	// held on this bp block / data file ..
	removeLock( tableName, pageID );
	
	return true;
	
} // end "writeBlock( DataPage &pg, map< char*, vector<int> > &locks )" ..


bool BufferManager:: loadBlock( string tableName, DataPage* dataPage, map< string, vector<int>* >* locks, bool lockPage ){
		
		
	// DEBUGGING ..
	//cout << "BufferManager:: loadBlock() start .." << endl << endl;
		
	if( tableName.length() <= 0 )
		return false;
		
	if( dataPage == NULL )
		return false;
		
	if( locks == NULL )
		return false;
		
	//cout << "BufferManager:: loadBlock() Point A .." << endl << endl;

	if ( lockPage ) {
		
		// upon loading a block from file, the page-from-file needs to be "locked"
		// in order to ensure that concurrent users dont update the page before/while
		// this user does ..
	
		int blockIndex = BufferInfoTable::bitIndex;
	
		int pageID = -1;
		
		if( blockIndex >= 0 )
			pageID = bit->getPageID()->at( blockIndex );

		// DEBUGGING ..
		//cout << "lock activated, tableName: " << tableName << ", pageID: " << pageID << endl << endl << endl;

		addLock( tableName, pageID );
		
		
	} // end "if()" ..

	//cout << "BufferManager:: loadBlock() Point B.." << endl << endl;
	
	// DEBUGGING ..
	//dataPage->printDataPage();
	
	//cout << "BufferManager:: loadBlock() tableName: " << tableName << endl << endl;

	bool readPagePerformed = getStorageManager()->readDataPage( tableName, dataPage, bit );
	
	//dataPage->printDataPage();
	
	//cout << "BufferManager:: loadBlock() Point C.." << endl << endl;
	
	//cout << "BufferManager:: loadBlock() Point B.." << endl << endl;
	
	// DEBUGGING ..
	//dataPage->printDataPage();
	
	// DEBUGGING ..
	//cout << "BufferManager:: loadBlock() readPagePerformed: " << readPagePerformed << endl << endl;
	//cout << "BufferManager:: loadBlock() dataPage.records.size(): " << dataPage.records.size() << endl << endl;
	
	
	// if "readDataPage" not performed, then remove the lock (if locked) ..
	if( readPagePerformed == false ) {
		
	 	string errorString = "\nFailed to read data page [ BM::loadBlock() ], data may not have been read from file.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );
		
		if ( lockPage ) {
		
			int blockIndex = BufferInfoTable::bitIndex;
	
			int pageID = -1;
			
			if( blockIndex >= 0 )
				pageID = bit->getPageID()->at( blockIndex );
		
			removeLock( tableName, pageID );
			
		}
		
	}
	
	//cout << "BufferManager:: loadBlock() Point D.." << endl << endl;
	
	// DEBUGGING ..
	//cout << "BufferManager:: loadBlock() end .." << endl << endl;
	
	return readPagePerformed;
	
	
} // end "loadBlock( string tableName, DataPage &pg, map< char*, vector<int> > &locks, bool lockPage )" ..


void BufferManager:: removeLock( string tableName, int pageID ){
	
	if( tableName.length() <= 0 )
		return;
		
	if( pageID < 0 )
		return;
	
	// DEBUGGING ..
	//cout << "BM:: removeLock() start .." << endl << endl;

	if ( ( pageID >= 0 ) &&
		 ( getStorageManager()->getTablePageLock( tableName, pageID ) > 0 ) ) {
		
		// DEBUGGING ..
		//cout << "deactivate-lock" << endl << endl << endl;
		
		// if the page has a lock, you can deactivate it now
		// (once block is written to file, the lock is released) ..
		getStorageManager()->deactivateTablePageLock( tableName, pageID );
		
		// DEBUGGING ..
		//cout << "BM:: removeLock() Point A1 .." << endl << endl;
		
		// clear "locks" map of invalid "tableName" prior to operating on "locks" ..
		// not necessary ?
		//purgeLocksMap( locks );
		
		if ( !locks->empty() ) {
			
			// DEBUGGING ..
			//cout << "BM:: removeLock() Point A2 .." << endl << endl;
			
			bool keyFound = false;
			map< string, vector<int>* >::iterator it_locks;
			string tableNameString = tableName;
			string locksString = "";
			for(it_locks = locks->begin(); it_locks != locks->end(); it_locks++){
	
				locksString = ( *it_locks ).first;
				
				if ( tableNameString.compare( locksString ) == 0 ){
		
					keyFound = true;
					break;
		
				}
		
			}
			
			// DEBUGGING ..
			//cout << "BM:: removeLock() Point A2 .." << endl << endl;
		
			if ( keyFound ) {
				
				// DEBUGGING ..
				//cout << "BM:: removeLock() Point A3 .." << endl << endl;
				//cout << "locks[ tableName ].size(): " << locks[ tableName ].size() << endl << endl;
			
				// must stop tracking the lock for this PageID for this tableName ..
				
				int lockPageIDToRemove = 0;
				
				bool lockPageIDToRemoveFound = false;
				
				vector<int>* tablePageIDs = (*locks)[ tableName ];
				
				//int pageIDSize = tablePageIDs->size();
				
				//int tablePageID = -1;
				
				vector<int>::iterator it;
				//for(it = locks->at( tableName )->begin(); it != locks->at( tableName )->end(); it++){
				for(it = tablePageIDs->begin(); it != tablePageIDs->end(); it++){
				//for(int i=0; i<pageIDSize; i++){
					
					//tablePageID = tablePageIDs->at(i);

					if ( (*it) == pageID ){
					//if ( tablePageID == pageID ){
				
						// designate that subsequent code should
						// remove the "pageID" from the "vector<int>" 
						// which came from the "locks" map ..
						
						lockPageIDToRemoveFound = true;
						
						// DEBUGGING ..
						//cout << "BM:: writeBlock() locks[ tableName("<<tableName<<") ].erase( it("<<(*it)<<") ) called .." << endl << endl;
				
						break;
				
					}
					
					lockPageIDToRemove++;
			
					// DEBUGGING ..
					//cout << "it("<<(*it)<<") == pageID("<<(pageID)<<"): " << ((*it) == pageID) << endl << endl;
			
				} // end "for()" ..
				
				// DEBUGGING ..
				//cout << "BM:: removeLock() Point A4 .." << endl << endl;
				
				if ( lockPageIDToRemoveFound ) {
					
					// DEBUGGING ..
					//cout << "BM:: removeLock() Point A5 .." << endl << endl;
					
					// if the pageID to remove was found above, remove it from "locks" ..
					//locks->at( tableName )->erase( ( locks->at( tableName )->begin() + lockPageIDToRemove ) );
					tablePageIDs->erase( ( tablePageIDs->begin() + lockPageIDToRemove ) );
					//tablePageIDs->erase( lockPageIDToRemove );
					
				}
				
				// DEBUGGING ..
				//cout << "BM:: removeLock() Point A6 .." << endl << endl;
			
			} // end "if ( keyFound )" ..
		
		} // end "if ( !locks.empty() )" ..
			
	} // end "if( pg.pageID >= 0 )" ..
	
	
	// DEBUGGING ..
	//cout << "BM:: removeLock() end .." << endl << endl;
	
	
} // end "removeLock( string tableName, int pageID )" ..


void BufferManager:: addLock( string tableName, int pageID ){
	
		//cout << "BufferManager:: addLock() start .." << endl << endl;
	
		if( tableName.length() <= 0 )
			return;
		
		if( pageID < 0 )
			return;
			
		//cout << "BufferManager:: addLock() Point A .." << endl << endl;
	
		// "activate" the lock in "tableName", for this "pageID", using "lockNumberIncrement" ..
		// "lockNumberIncrement = 5" is arbitrary .. it should be greater than zero though
		// (originally, this was going to be a "timed" lock, where "lockNumberIncrement"
		//  was the number of minutes before the lock would be automatically released, but
		//  this required threads, which was too difficult to locate information on for C++) ..
		getStorageManager()->activateTablePageLock( tableName, pageID, 5 );
		
		//cout << "BufferManager:: addLock() Point B .." << endl << endl;
	
		// clear "locks" map of invalid "tableName" prior to operating on "locks" ..
		// not necessary ?
		//purgeLocksMap( locks );
		
		// check "locks" to see if this "tableName" has any locks 
		// on any "pageID"s yet ..
		bool keyFound = false;
		
		map< string, vector<int>* >::key_compare compare = locks->key_comp();
		map< string, vector<int>* >::iterator it_locks;
		for(it_locks = locks->begin(); it_locks != locks->end(); it_locks++){
	
			//cout << "BufferManager:: addLock() compare( ( *it_locks ).first("<<( *it_locks ).first<<"), tableName("<<tableName<<") == 0 ): " << ( tableName.compare( ( *it_locks ).first ) == 0 )   << endl << endl;
	
			if ( tableName.compare( ( *it_locks ).first ) == 0 ){
		
				// "tableName" was found in "locks" ..
				keyFound = true;
				break;
		
			}
		
		}
		
		//cout << "BufferManager:: addLock() Point C .." << endl << endl;
		
		bool pageIDAlreadyLocked = false;
	
		if ( keyFound == false ) {
			
			//cout << "BufferManager:: addLock() Point C1 .." << endl << endl;
			
			// "tableName" hasn't been entered into the "locks" map yet, so
			// add a new key / value pair to the "locks" map ..
		
			locks->insert( pair< string, vector<int>* >( tableName, new vector<int>() ) );
			
			//cout << "BufferManager:: addLock() Point C2 .." << endl << endl;
	
		} else if ( keyFound == true ) {
			
			//cout << "BufferManager:: addLock() Point C3 .." << endl << endl;
			
			// check "locks" with this "tableName" and 
			// make sure that there isn't already a lock for this "pageID" ..
			
			vector<int>* tablePageIDs = (*locks)[ tableName ];
			
			//cout << "BufferManager:: addLock() Point C4 .." << endl << endl;
			
			int pageIDsSize = 0;
			
			if( tablePageIDs != NULL )
				pageIDsSize = tablePageIDs->size();
				
			//cout << "BufferManager:: addLock() Point C5 .." << endl << endl;
			
			int tablePageID = -1;
	
			for(int i=0; i<pageIDsSize; i++){
				
				tablePageID = tablePageIDs->at(i);
			
				//if( locks->at( tableName )->at(i) == pageID ){
				if( tablePageID == pageID ){
				
					pageIDAlreadyLocked = true;
					break;
				
				}
			
			}
			
			//cout << "BufferManager:: addLock() Point C6 .." << endl << endl;
			
		}
	
		if ( pageIDAlreadyLocked == false ) {
			
			//cout << "BufferManager:: addLock() Point C7 .." << endl << endl;
	
			// must track which "pageID"s have locks for this "tableName",
			// so insert the "pageID" at the end of the "vector<int>" within the "locks" map ..
			(*locks)[ tableName ]->push_back( pageID );
			//tablePageIDs->push_back( pageID );
			
			//cout << "BufferManager:: addLock() Point C8 .." << endl << endl;
			
		}
		
		//cout << "BufferManager:: addLock() Point D .." << endl << endl;

} // end "addLock( string tableName, int pageID )" ..


FreeAndFlushBlockIndexAggregation* BufferManager:: getFreeAndFlushBlockIndexProcessing( int newFreeBlockIndex ){

	//if( newFreeBlockIndex < 0 )
	//	return NULL;
	
	// helper method(s):
	// 1) findBlockForInsert()
	// 2) findLRUAvailableBlockForFlushing()
	
	FreeAndFlushBlockIndexAggregation* fafbip = new FreeAndFlushBlockIndexAggregation();
	
	bool insertPerformed = true; 
	
	bool freeBlockFound = false;
	int freeBlockIndex = newFreeBlockIndex;
	
	bool flushBlockFound = false;
	int flushBlockIndex = -1;
	
	
	if ( freeBlockIndex >= 0 ) {
		
		// block available in BP for insert of new row 
		// (block either empty, partially full or can be overwritten) ..
		
		freeBlockFound = true;
		// flushBlockFound == false;
		// insertPerformed = true;
		
		BufferInfoTable::bitIndex = freeBlockIndex;
		
	} else {
		
		// no free block in the BP available, 
		// will need to flush a block in the BP to file first ..
		
		freeBlockFound = false;
		// flushBlockFound == false;
		// insertPerformed = true;
		
		flushBlockIndex = findLRUAvailableBlockForFlushing();
		
		if ( flushBlockIndex >= 0 ) {
			
			// block available for flushing ..
			
			flushBlockFound = true;
			// freeBlockFound = false;
			// insertPerformed = true;
			
			BufferInfoTable::bitIndex = flushBlockIndex;
			
		}  else { 
			
			// no free block; no block available for flushing; abort insert operation ..
			
			insertPerformed = false;
			// flushBlockFound = false;
			// freeBlockFound = false;
			
			BufferInfoTable::bitIndex = -1;
			
		} // end "if ( flushBlockIndex >= 0 )" ..
		
	} // end "if ( freeBlockIndex <= 0 )" ..
	
	fafbip->setInsertPerformed( insertPerformed );
	
	fafbip->setFreeBlockFound( freeBlockFound );
	//fafbip->setFreeBlockIndex( freeBlockIndex );
	
	fafbip->setFlushBlockFound( flushBlockFound );
	fafbip->setFlushBlockIndex( flushBlockIndex );
	
	return fafbip;
	
	
} // end "freeAndFlushBlockProcessing()" ..


bool BufferManager:: flushBPBlock( int flushBlockIndex, DataHeader* header ){
	
	//cout << "BufferManager:: flushBPBlock() start .." << endl << endl;
	
	if( flushBlockIndex < 0 )
		return false;
		
	if( header == NULL )
		return false;


	
	//cout << "BufferManager:: flushBPBlock() flushBlockIndex: " << flushBlockIndex << endl << endl;
	
	// helper method used by "insertRecordIntoBlock()" ..

	// 1) this indicates that the block should be "flushed" by
	//    saving / writing it to file, prior to inserting a new
	//    record into the block ..
				
	int makePageID = -1;
				
	if ( bit->getPageID()->at( flushBlockIndex ) == -1 ){
		
		//cout << "BufferManager:: flushBPBlock() Point A .." << endl << endl;
					
		// (a) this indicates that a new page has to be ALLOCATED,
		//     and a new "PageID" needs to be created 
		//     (see "freeBlockFound" section below, which uses the PageID) ..
					
		int currentMaxPageID = -1;
		int numberOfDataPages = header->getTotalNumberOfDataPages();
		for( int i=0; i < numberOfDataPages; i++ ){
				
			// find the max PageID in the DataHeader ..
			if ( header->getHeaderEntries()->at( i )->getPageID() > currentMaxPageID ) {
					
				currentMaxPageID = header->getHeaderEntries()->at( i )->getPageID();
					
			}
				
		} // end "for()" ..
		
		//cout << "BufferManager:: flushBPBlock() Point B .." << endl << endl;
				
		// (b) the "PageID" for the allocated page should be 
		//     1 greater than the current max in the DataHeader ..
		makePageID = ( currentMaxPageID + 1 );
					
		// "bit" tracks the new pageID for use in "flush" operation,
		//  in the "writeBlock()" method, just below here ..
		bit->getPageID()->at( flushBlockIndex ) = makePageID;
			
		// (c) allocate a new page to the file for this table,
		//     where the block in the BP will be 
		//     "flushed" / saved / written to this allocated file page ..
		bool isMakeDataPagePerformed = getStorageManager()->makeDataPage( bit->getTableName()->at( flushBlockIndex ), makePageID );
		
		//cout << "BufferManager:: flushBPBlock() Point C .." << endl << endl;
		
		// if "makeDataPage()" not performed,
		// "flushBPBlock()" should not be performed, so abort ..
		if ( isMakeDataPagePerformed == false ) {
			
	 		string errorString = "\nFailed to allocate data page [ BM::flushBPBlock() ], data may not have been written to file.";
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );
			
			return false;
			
		}
					
	} // end "if ( bit.pageID[ flushBlockIndex ] == -1 )" ..
	
	//cout << "BufferManager:: flushBPBlock() Point D .." << endl << endl;

	DataPage* dataPage = bp->getDataPages()->at( flushBlockIndex );
	
	if( makePageID >= 0 )
		dataPage->setPageID( makePageID );
	else
		dataPage->setPageID( bit->getPageID()->at( flushBlockIndex ) );
	
	// DEBUGGING ..
	//dataPage->printDataPage();

	// 2) need to write this current block to file 
	//    before "flushing" / overwriting it ..
	bool isWriteBlockPerformed = writeBlock( dataPage, locks );
				
	// if "writeBlock()" not performed,
	// "flushBPBlock()" should not be perfoemd, so abort ..
	// (ok if "makeDataPage()" was performed) ..
	if ( isWriteBlockPerformed == false ) {
		
	 	string errorString = "\nFailed to write data page [ BM::flushBPBlock() ], data may not have been written to file.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );
			
		return false;
			
	}
	
	//cout << "BufferManager:: flushBPBlock() Point E .." << endl << endl;
			
				
	// clear "locks" map of invalid "tableName" after operating on "locks" (in writeBlock()); 
	// seems to need this immediately following the function call to "writeBlock()" ..
				
	//purgeLocksMap( locks );

/*			
	char* tableName = (char*) bit.tableName[ flushBlockIndex ].c_str();

	// DEBUGGING ..
	cout << endl << "locks .." << endl << endl;
	vector<int> pageID;
	vector<int>::iterator it_pageIDs;
	map< char*, vector<int> >::iterator it_locks;
	cout << "locks.count( tableName("<<tableName<<") ): " << locks.count( tableName ) << endl;
	for(it_locks = locks.begin(); it_locks != locks.end(); it_locks++){
	
		cout << "lock: " << ( *it_locks ).first << endl;
		cout << "(( *it_locks ).second).size(): " << (( *it_locks ).second).size() << endl;
		pageID = ( *it_locks ).second;
		for(it_pageIDs = pageID.begin(); it_pageIDs != pageID.end(); it_pageIDs++){
			
			cout << "pageID: " << ( *it_pageIDs ) << endl;
			
		}
			
	}
	cout << endl << "end locks .." << endl << endl;
*/

	string tableName = bit->getTableName()->at( flushBlockIndex );
	
	//cout << "BufferManager:: flushBPBlock() Point F .." << endl << endl;
	
	//cout << "BufferManager:: flushBPBlock() tableName: " << tableName << endl << endl;

	// 3) must UPDATE the CATALOG ..
	// get the existing table size ..
	int newTableSize = getCatalogManager()->getTableSize( tableName );
			
	//cout << "BufferManager:: flushBPBlock() old tableSize: " << newTableSize << endl << endl;
			
	// add size of new rows to existing table size ..
	newTableSize += blockSize;
	//newTableSize += ( ( bit->getRecordSize()->at( flushBlockIndex ) ) * ( bit->getBufferPageRowCount()->at( flushBlockIndex ) ) );
			
	// update the catalog with the new table size ..
	//cout << "BufferManager:: flushBPBlock() updated tableSize: " << newTableSize << endl << endl;
	getCatalogManager()->setTableSize( tableName, newTableSize );
				
	//cout << "BufferManager:: flushBPBlock() Point F .." << endl << endl;
				
	// Track the latest flush  ..
	DataPage* copyDataPage = new DataPage( dataPage );
	getLastFlushedBlocks()->push_back( copyDataPage );
	getLastFlushedBlockPageIDs()->push_back( makePageID );
	
	// clear the current bp block that was flushed ..
	bp->getDataPages()->at( flushBlockIndex )->clearPage();
	
	// 4) need updated version of DataHeader ..
	//    (sm->makeDataPage() seen above may have changed it) ..
	bool readHeaderPerformed = getStorageManager()->readHeader( tableName, header );
	
	//cout << "BufferManager:: flushBPBlock() Point G .." << endl << endl;
	
	if ( readHeaderPerformed == false ) {
		
	 	string errorString = "\nFailed to read Header [ BM::flushBPBlock() ], may need to restart application for proper functioning.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );
		
		return false;
		
	}
	
	// reset "bit" values, now that this block was written ..
	getBufferInfoTable()->updateBufferInfoTable( flushBlockIndex, BufferInfoTableState::FLUSH_BLOCK );
	
	
	//cout << "BufferManager:: flushBPBlock() Point H .." << endl << endl;
	
	return true;
	
} // end "flushBPBlock( int flushBlockIndex, DataHeader* hp )" ..


int BufferManager:: findLRUAvailableBlockForFlushing(){
	
	// use this method to locate a block to flush ( LRU, pin_count = 0 ) ..
	
	// returns "-1" if it can't find an available block ..
	int bpBlock = -1;
	
	bool foundLRUAvailableBlock = false;
	
	list<int>::iterator it_lru;
	
	// loop through the BP ..
	for( it_lru = bit->getLRUBlocks()->begin(); it_lru != bit->getLRUBlocks()->end(); it_lru++ ){
		
		// get the index of the next LRU BP block ..
		bpBlock = ( *it_lru );
		
		// DEBUGGING ..
		//cout << "BM:: findLRUAvailableBlockForFlushing() bit.pin_count[ bpBlock ]("<<bit.pin_count[ bpBlock ]<<") == 0:  " << ( bit.pin_count[ bpBlock ] == 0 ) << endl << endl;
		
		if ( bit->getPinCount()->at( bpBlock ) == 0 ) {
			
			// DEBUGGING ..
			//cout << "BM:: findLRUAvailableBlockForFlushing() bpBlock: " << bpBlock << endl << endl;
			
			// the block is not pinned, 
			// so its the LRU available block ..
			foundLRUAvailableBlock = true;
			break;
			
		}
		
	}

	
	if ( !foundLRUAvailableBlock ){
		
		// returns "-1" if it can't find an available block ..
		bpBlock = -1;
		
	} 
	
	return bpBlock;
	
} // end "findLRUAvailableBlockForFlushing()" ..

vector<int>* BufferManager:: getDirtyBufferPoolBlocks(){
	
	vector<int>* bpPageIDs = new vector<int>();
	
	int bufferPoolSize = bit->getBufferPoolSize();
	
	//vector<DataPage*>* bpDataPages = bp->getDataPages();
	
	// look through the buffer pool for "dirty" pages ..
	for(int i=0; i<bufferPoolSize; i++){
		
		// "dirty" pages are "latest" copies, or
		// pages that have not yet been written to file,
		// and these are the pageIDs to return ..
		if( bit->getDirty()->at(i) == true ){
			
			bpPageIDs->push_back( bit->getPageID()->at(i) );	
			
		}
		
	}
	
	return bpPageIDs;
	
}

HeaderEntryResultsAggregation* BufferManager:: getHeaderEntryResults( DataHeader* header, vector<int>* bpPageIDs ){
	
	if( header == NULL )
		return NULL;
		
	if( bpPageIDs == NULL )
		return NULL;
	
	// "header" corresponds to meta-data on disk,
	// for the table associated with a tuple
	// to be inserted into the main memory buffer pool ..
	
	// "bpPageIDs" are the pageIDs of blocks in the
	// main memory buffer pool which may have
	// "older" copies residing on disk,
	// and therefore shouldn't be loaded
	// from the disk back into the MM buffer pool ..
	
	int loadPageID = -1;
			
	int headerEntryIndex = -1;
			
	bool hpEntryFound = false;
	
	// if "header" is NOT NULL proceed,
	// else just return default values ..
	if( header != NULL ){
		
		int headerEntryPageID = -1;
		int bpPageIDsSize = -1;
		if( bpPageIDs != NULL )
			bpPageIDsSize = bpPageIDs->size();
		bool isAvoidHeaderEntryPageID = false;
		
		DataHeaderEntry* dataHeaderEntry = NULL;
	
		int numberOfDataPages = -1;
		if( header != NULL )
			numberOfDataPages = header->getTotalNumberOfDataPages();

		for(int i=0; i<numberOfDataPages; i++){
				
			dataHeaderEntry = header->getHeaderEntries()->at( i );
		
			if( dataHeaderEntry == NULL )
				continue;
				
			//cout << "BufferManager:: getHeaderEntryResults() ( dataHeaderEntry->getFreeRowsCount()("<<dataHeaderEntry->getFreeRowsCount()<<") > 0 )["<<i<<"]): " << ( dataHeaderEntry->getFreeRowsCount() > 0 ) << endl;
					
			if ( ( dataHeaderEntry->getFreeRowsCount() > 0 ) &&
				 ( dataHeaderEntry->getLockNumber() == 0 ) ) {
				 	
				 //cout << "BufferManager:: getHeaderEntryResults() Point A["<<i<<"].." << endl << endl;
			
				// find the first "PageID" that has a free slot 
				// for this "tableNameX",
				// (but only if the page is not "locked") ..
				
				// next, check that its "pageID"
				// is not among the list to avoid ..
				// (file has "old copy", 
				//  list from bp has "latest" copy) ..
				headerEntryPageID = dataHeaderEntry->getPageID();
				
				// only take into consideration of "bpPageIDs" was provided ..
				if( bpPageIDs != NULL ){
					
					//cout << "BufferManager:: getHeaderEntryResults() Point B["<<i<<"].." << endl << endl;
					
					// loop through the collection of pageIDs ..
					for( int j=0; j<bpPageIDsSize; j++){
						
						if( headerEntryPageID == bpPageIDs->at(j) ){
							
							// this indicates that this headerEntry
							// should be avoided ..
							isAvoidHeaderEntryPageID = true;
							break;
							
						} // end "if()" ..
						
					} // end "for()" ..
					
					//cout << "BufferManager:: getHeaderEntryResults() Point C["<<i<<"].." << endl << endl;
					
					if( isAvoidHeaderEntryPageID == true ){
						
						// reset the flag and
						// continue to examine the next
						// headerEntry ..
						isAvoidHeaderEntryPageID = false;
						continue;	
						
					}
					
				} // end "if()" ..
				
				//cout << "BufferManager:: getHeaderEntryResults() Point D["<<i<<"].." << endl << endl;
				
				// if this point is reached,
				// this headerEntry has passed the tests above,
				// so continue ..
				
				loadPageID = headerEntryPageID;
		
				// track the index in the "HP" for this headerEntry,
				// to be used below ..
				headerEntryIndex = i;
						
				// track that a headerEntry was found ..
				hpEntryFound = true;
					
				break;
					
			} // end "if()" ..
				
		} // end "for()" ..
	
		cout << endl;	
		
	} // end "if()" ..
	
	HeaderEntryResultsAggregation* herp = new HeaderEntryResultsAggregation();
	
	herp->setLoadPageID( loadPageID );
	herp->setHeaderEntryIndex( headerEntryIndex );
	herp->setHpEntryFound( hpEntryFound );
	
	//cout << "BufferManager:: getHeaderEntryResults() Point E.." << endl << endl;
	
	return herp;
	
} // end "getHeaderEntryResults()" ..


void BufferManager:: loadAndInsertIntoPartial( LoadAndInsertIntoPartialAggregation* laiipp ){
	
	if( laiipp == NULL )
		return;
	
	int recIndex = -1;
	
	DataHeader* header = laiipp->getHeader();
	int headerEntryIndex = laiipp->getHeaderEntryIndex();
	int blockIndex = laiipp->getBlockIndex();
	int loadPageID = laiipp->getLoadPageID();
	string tableNameX = laiipp->getTableNameX();
	int numberOfAttributes = laiipp->getNumberOfAttributes();
	vector<int>* insertValues = laiipp->getInsertValues();
	
	if( header == NULL )
		return;
		
	if( headerEntryIndex < 0 )
		return;
	
	if( blockIndex < 0 )
		return;
	
	//if( loadPageID < 0 )
	//	return;
	
	if( tableNameX.length() <= 0 )
		return;
	
	if( numberOfAttributes <= 0 )
		return;
	
	if( insertValues == NULL )
		return;
	
	int fieldCount = getCatalogManager()->getNumberOfAttributes( tableNameX );
	
	int maxRowCount = header->computeMaxRowCount( fieldCount );
				
	DataHeaderEntry* hpEntry = new DataHeaderEntry( maxRowCount );
				
	hpEntry = header->getHeaderEntries()->at( headerEntryIndex );
				
			
	// 1) get "filledRowsCount" from the DataHeader,
	//    load it into the bit that corresponds to the bp page to be
	//    overwritten ..
	if( blockIndex >= 0 )
		bit->getBufferPageRowCount()->at( blockIndex ) = hpEntry->getFilledRowsCount();
					

	// 2) get "rowAvailable" from the DataHeader ..
	for(int i=0; i<maxRowCount; i++){
							
		if( blockIndex >= 0 )
			bit->getRowAvailable()->at( blockIndex )->at(i) = hpEntry->getRowAvailable()->at( i );
							
		// set "recIndex" ..
		// find the first free slot in the loaded block 
		// that is available for insert of a new record ..
		if ( ( hpEntry->getRowAvailable()->at( i ) == true ) &&
			 ( recIndex < 0 ) ){
							
			recIndex = i;
			//break;
							
		}
							
	} // end "for()" ..
					
					
	// 3) load in the page-from-file, then 
	//    insert the new row into this block ..
	//    ("recIndex" is reset to "-1" everytime "insertRecordIntoBlock()" is called,
	//     so finding a non-negative "recIndex" means "hpEntryFound == true") ..
	if( ( recIndex >= 0 ) &&
		( recIndex < maxRowCount ) ){
				
		// II & V: "LOAD" and "insert-into-partial" ..
				
		DataPage* loadedDataPage = new DataPage();
		
		loadedDataPage->setPageID( loadPageID );
				
				
		// II: "LOAD" ..
				
		// (a) read the page in from file ..
		bool loadBlockPerformed = loadBlock( tableNameX, loadedDataPage, locks, true );
		
		//cout << "BufferManager:: loadAndInsertIntoPartial() loadBlock() .." << endl << endl;
		//loadedDataPage->printDataPage();
						

		// if this fails, call "insertIntoEmpty()" instead and 
		// return from (stop doing) "loadAndInsertIntoPartial()" ..
		if ( loadBlockPerformed == false ) {
			
			InsertIntoEmptyAggregation* iiep = new InsertIntoEmptyAggregation();
			iiep->setBlockIndex( blockIndex );
			iiep->setNumberOfAttributes( numberOfAttributes );
			iiep->setInsertValues( insertValues );
			
			insertIntoEmpty( iiep );
			
			return;
			
		}

					
		// V: "insert-into-partial" ..
		
		int numberOfRows = 0;
		
		//numberOfRows = bp->getDataPages()->at( blockIndex )->getNumberOfRows();
		numberOfRows = loadedDataPage->getNumberOfRows();
	
		if( recIndex < numberOfRows ){
		
			// this means that the new row will replace an existing row ..
			//bp->getDataPages()->at( blockIndex )->replaceEntries( recIndex, insertValues );
			loadedDataPage->replaceEntries( recIndex, insertValues );
		
		} else {
		
			// this means that the new row is appended to the existing rows ..
			//bp->getDataPages()->at( blockIndex )->addEntries( insertValues );
			loadedDataPage->addEntries( insertValues );
		
		}
		
/*
		// DEBUGGING ..
		cout << "insert .." << endl << endl;
		for(int i=0; i<insertValues->size(); i++){
			
			cout << insertValues->at(i) << " ";	
			
		}
		cout << endl << endl;
		cout << "loadedDataPage + insert .." << endl << endl;
		loadedDataPage->printDataPage();
*/


/*	
	
		// (b) initialize the "loadedDataPage" with empty Records and Fields ..
		vector<DataPageRecord>* currentRecords = loadedDataPage->getRecords();
							
		int newNumberOfRecords = ( loadedDataPage->getRecords()->size() + 1 );
				
		loadedDataPage->getRecords()->clear();
							
		for(int i=0; i<newNumberOfRecords; i++){
								
			// fill "loadedDataPage" with empty Records, including
			// a new additional empty DataPageRecord for holding the insert row ..
			loadedDataPage->getRecords()->push_back( (*new DataPageRecord()) );
							
			if ( i == recIndex ){
							
				for(int j=0; j<numberOfAttributes; j++){

					// fill the empty Records with empty Fields ..
					loadedDataPage->getRecords()->at(i).getFields()->push_back( ( *new DataPageField() ) );
								
				}
								
			}
								
		}
		
		// (c) Fill "loadedDataPage" with data, including new row ..				
		for(int i=0; i<newNumberOfRecords; i++){
							
			if ( i < recIndex ) {

				loadedDataPage->getRecords()->at(i) = currentRecords->at(i);
								
			} else if ( i == recIndex ) {
	
				// (ii) loop through and assign each of the fields in the record ..
				for(int j=0; j < numberOfAttributes; j++){
							
					loadedDataPage->getRecords()->at( recIndex ).getFields()->at( j ).setValue( insertValues->at(j) );

				} // end "for()" ..
								
								
			} else if ( i > recIndex ) {

				// "loadedDataPage.records" has 1 more row than "currentRecords", 
				// so the indices are offset from each other for both by 1 ..
				loadedDataPage->getRecords()->at(i) = currentRecords->at(i-1);
								
			}
							
		}
		
*/
					
		// (d) overwrite/replace the "BP" page with the loaded-from-file page ..
		bp->getDataPages()->at( blockIndex ) = loadedDataPage;
								
		// (e) UPDATE "bit" entries ..
		bit->getBufferPageRowCount()->at( blockIndex )++;
			
		bit->getRowAvailable()->at( blockIndex )->at( recIndex ) = false;

		// track the pageID of the page-from-file that is 
		// about to be loaded into the "bp" ..
		bit->getPageID()->at( blockIndex ) = loadPageID;

		// "LOAD" + "insert-into-partial" is a request for a page-from-file,
		// so update this parameter in the "bit" ..
		// this is a "miss" ..
		bit->setNumberRequestsFromFile( bit->getNumberRequestsFromFile() + 1 );
		bit->updateHitRate();

	} // end "if( ( recIndex >= 0 ) &&" ..
	
	
} // end "loadAndInsertIntoPartial( LoadAndInsertIntoPartialAggregation* laiipp )" ..


//int BufferManager:: findBlockForInsert( string tableNameX, int numberOfAttributes ){
BlockForInsert* BufferManager:: findBlockForInsert( string tableNameX, int numberOfAttributes ){
	
	//cout << "BufferManager:: findBlockForInsert() start .." << endl << endl;
	
	//cout << "BufferManager:: findBlockForInsert() tableNameX: " << tableNameX << endl << endl;
	
	//cout << "BufferManager:: findBlockForInsert() numberOfAttributes: " << numberOfAttributes << endl << endl;
	
	if( tableNameX.length() <= 0 )
		return NULL;
		
	if( numberOfAttributes <= 0 )
		return NULL;
		
	//cout << "BufferManager:: findBlockForInsert() Point A .." << endl << endl;
	
	// use this method to find a block to insert into ..
	
	// returns "-1" if it can't find an available block ..
	int bpBlock = -1;
	
	int state = BlockForInsert::NONE;
	
	bool bpBlockFound = false;
	
	int maxRowCount = -1;
	
	string tableName = "";
	
	int bufferPoolSize = bp->getBufferPoolSize();
	
	// #1: first look for any available "partially full, same table" blocks ..
	for( int i = 0; i<bufferPoolSize; i++ ){
		
		maxRowCount = bit->computeMaxRowCount_Index( i );
		
		if ( bit->getTableName() != NULL )
			tableName = bit->getTableName()->at( i );
		
		//cout << "BufferManager:: findBlockForInsert() ( tableName("<<tableName<<").compare( tableNameX("<<tableNameX<<") ) == 0 ): " << ( tableName.compare( tableNameX ) == 0 ) << endl;
		//cout << "BufferManager:: findBlockForInsert() ( bit->getBufferPageFieldCount()->at( i("<<i<<") )("<<bit->getBufferPageFieldCount()->at( i )<<") == numberOfAttributes("<<numberOfAttributes<<") ): " << ( bit->getBufferPageFieldCount()->at( i ) == numberOfAttributes ) << endl;
		//cout << "BufferManager:: findBlockForInsert() ( bit->getBufferPageRowCount()->at( i("<<i<<") )("<<bit->getBufferPageRowCount()->at( i )<<") < maxRowCount("<<maxRowCount<<") ): " << ( bit->getBufferPageRowCount()->at( i ) < maxRowCount ) << endl;
		//cout << endl;	
		
		if ( ( tableName.compare( tableNameX ) == 0 ) &&
			 ( bit->getBufferPageFieldCount()->at( i ) == numberOfAttributes ) &&
		     ( bit->getBufferPageRowCount()->at( i ) < maxRowCount ) ){
			
			bpBlock = i;
			state = BlockForInsert::PARTIALLY_FULL_SAME_TABLE;
			bpBlockFound = true;
			break;
			
		}
		
	}
	
	//cout << "BufferManager:: findBlockForInsert() Point B .." << endl << endl;
	
	//cout << "BufferManager:: findBlockForInsert() bpBlockFound: " << bpBlockFound << endl << endl;
	
	if( bpBlockFound == false ){
	
		// #2: next look for any available "empty" blocks ..
		for( int i = 0; i < bufferPoolSize; i++ ){
		
			//cout << "BufferManager:: findBlockForInsert() ( bit->getEmpty()->at( i("<<i<<") )("<<bit->getEmpty()->at( i )<<") == true ): " << ( bit->getEmpty()->at( i ) == true ) << endl;
			//cout << endl;
		
			if ( bit->getEmpty()->at( i ) == true ){
			
				bpBlock = i;
				state = BlockForInsert::EMPTY;
				bpBlockFound = true;
				break;
			}
		
		}
		
		// this last bit is not done here,
		// instead its handled by "flush" portion,
		// and not in "LRU" method, 
		// but rather in "insert" calling method ..
/*
		if( bpBlockFound == false ){
			
			// #3: finally look for any (partially full/full, same/different table) "Not Dirty" blocks ..
			for( int i = 0; i < bp->getBufferPoolSize(); i++ ){
		
				cout << "BufferManager:: findBlockForInsert() ( bit->getEmpty()->at( i("<<i<<") )("<<bit->getEmpty()->at( i )<<") == true ): " << ( bit->getEmpty()->at( i ) == true ) << endl;
				cout << endl;
		
				if ( bit->getDirty()->at( i ) == false ){
			
					bpBlock = i;
					state = BlockForInsert::NOT_DIRTY;
					bpBlockFound = true;
					break;
				}
		
			}	
			
		}
*/
		
	} // end "if()" ..
	

	
/*
	// if an available "empty" bp block was not found, proceed ..
	if( emptyBPBlockFound == false ){
		
		// loop through the BP ..
		for( int i = 0; i < bp->getBufferPoolSize(); i++ ){
		
			int maxRowCount = bit->computeMaxRowCount_Index( i );
		
			string tableName = "";
			if ( bit->getTableName() != NULL )
				tableName = bit->getTableName()->at( i );
			

			//cout << "BufferManager:: findBlockForInsert() ( bit->getEmpty()->at( i("<<i<<") )("<<bit->getEmpty()->at( i )<<") == true ): " << ( bit->getEmpty()->at( i ) == true ) << endl;
			cout << "BufferManager:: findBlockForInsert() ( bit->getDirty()->at( i("<<i<<") )("<<bit->getDirty()->at( i )<<") == false ): " << ( bit->getDirty()->at( i ) == false ) << endl;
			
			if ( ( ( tableName.compare( tableNameX ) == 0 ) &&
			   	   ( bit->getBufferPageFieldCount()->at( i ) == numberOfAttributes ) &&
		       	   ( bit->getBufferPageRowCount()->at( i ) < maxRowCount ) ) ||
		     	 ( bit->getDirty()->at( i ) == false ) ){
		     
				// the block should either (a) match the table that the record is being inserted into,
				// or (b) the block is "empty" (not holding a record for any table) .. 
				// or (c) the block is not "dirty" so can be overwritten,
				// so its the "block available for insert" ..
				bpBlock = i;
			
				if ( ( tableName.compare( tableNameX ) == 0 ) &&
			 	 	 ( bit->getBufferPageFieldCount()->at( i ) == numberOfAttributes ) &&
		     	 	 ( bit->getBufferPageRowCount()->at( i ) < maxRowCount ) ){
		     
					state = BlockForInsert::PARTIALLY_FULL_SAME_TABLE;
			
				}
		
				if ( bit->getDirty()->at( i ) == false ){
		     
					state = BlockForInsert::EMPTY_OR_NOT_DIRTY;
			
				}
			
				break;
			
			}
		
		} // end "for()" ..
		
	}
*/
	
	// DEBUGGING ..
	//cout << "bpBlock: " << bpBlock << endl;
	
	//cout << "BufferManager:: findBlockForInsert() bpBlock: " << bpBlock << endl << endl;
	
	//cout << "BufferManager:: findBlockForInsert() state: " << state << endl << endl;
	
	//cout << "BufferManager:: findBlockForInsert() Point C .." << endl << endl;
	
	BlockForInsert* blockForInsert = new BlockForInsert();
	blockForInsert->setBpBlock( bpBlock );
	blockForInsert->setState( state );
	
	//return bpBlock;
	return blockForInsert;
	
} // end "findBlockForInsert( int numberOfAttributes )" ..


int BufferManager:: findBlockForLoad(){
	
	// use this method to find a bp block to load a page-from-file into ..
	
	// helper method for "returnFilePage()" ..
	
	// returns "-1" if it can't find an available block ..
	int bpBlock = -1;
	
	int bufferPoolSize = bp->getBufferPoolSize();
	
	// loop through the BP ..
	for( int i = 0; i<bufferPoolSize; i++ ){
		
		// "dirty" bit should be used,
		// but the project isn't implementing it ..
		//if ( ( bit->getEmpty()->at( i ) == true ) || 
		//     ( bit->getDirty()->at( i ) == false ) ){
		     	
		if ( bit->getEmpty()->at( i ) == true ){
			
			// the block should either (a) match the table that the record is being inserted into,
			// or (b) the block is "empty" (not holding a record for any table) .. 
			// so its the "block available for insert" ..
			bpBlock = i;
			break;
			
		}
		
	} // end "for()" ..
	
	return bpBlock;
	
} // end "findBlockForLoad()" ..


void BufferManager:: insertIntoEmpty( InsertIntoEmptyAggregation* iiep ){
	
	//cout << "BufferManager:: insertIntoEmpty() start .." << endl << endl;
	
	if( iiep == NULL )
		return;
	
	int blockIndex = iiep->getBlockIndex();
	
	int numberOfAttributes = iiep->getNumberOfAttributes();
	
	vector<int>* insertValues = iiep->getInsertValues();
	
	//cout << "BufferManager:: insertIntoEmpty() Point A .." << endl << endl;
	
	if( blockIndex < 0 )
		return;
	
	//cout << "BufferManager:: insertIntoEmpty() Point B .." << endl << endl;
		
	if( numberOfAttributes < 0 )
		return;
	
	//cout << "BufferManager:: insertIntoEmpty() Point C .." << endl << endl;
		
	if( insertValues == NULL )
		return;
	
	//cout << "BufferManager:: insertIntoEmpty() Point D .." << endl << endl;
	
	
	int maxRowCount = bit->computeMaxRowCount_Index( blockIndex );
	
	if( maxRowCount == -1 ){
		
		maxRowCount = bit->computeMaxRowCount_Attributes( numberOfAttributes );
		
	}
	
	//cout << "BufferManager:: insertIntoEmpty() maxRowCount: " << maxRowCount << endl << endl;
					
	// 1) set equal to "0" in preparation for insert of new record into an empty slot .. 
	bit->getBufferPageRowCount()->at( blockIndex ) = 0;
						
	// 2) no "hpEntry" found, so set all "rowAvailable" in "bit" at this index to "false" ..
	for(int i=0; i<maxRowCount; i++){

		bit->getRowAvailable()->at( blockIndex )->at(i) = false;
							
	}
	
	//cout << "BufferManager:: insertIntoEmpty() Point E .." << endl << endl;
				
				
	// 3) This block of code handles when "LOAD" operation is not 
	//    successful / necessary ..
				
	//if ( ( pageID < 0 ) || ( recIndex < 0 ) || ( ( recIndex >= maxRowCount ) && ( recIndex != -1 ) && ( maxRowCount != -1 ) ) ) {
	if ( maxRowCount > 0 ) {

		// (a) flag indicating that "sm->AllocateDataPage()" will be 
		// 	   required if (later on) this block is flushed to file ..
		bit->getPageID()->at( blockIndex ) = -1;
				
		// (b) initialize the block prior to the insert operation, 
		// 		since it has no data in it yet ..
		//bp->getDataPages()->at( blockIndex )->getRecords()->clear();

		//bp->getDataPages()->at( blockIndex )->getRecords()->push_back( ( *new DataPageRecord() ) );

		//bp->getDataPages()->at( blockIndex )->getRecords()->at( 0 ). getFields()->clear();
		//for(int i=0; i<numberOfAttributes; i++)
		//	bp->getDataPages()->at( blockIndex )->getRecords()->at( 0 ). getFields()->push_back( *new DataPageField() );

		// (c) need to insert row into the (now) empty / flushed BP block ..
		// 		 loop through and assign each of the fields in the record ..
		//for(int i=0; i < numberOfAttributes; i++){

		//	bp->getDataPages()->at( blockIndex )->getRecords()->at( 0 ).getFields()->at( i ).setValue( insertValues->at(i) );
				
		//}
		
		//int insertValuesSize = insertValues->size();
		//for(int i=0; i<insertValuesSize; i++)
		//	cout << "BufferManager:: insertIntoEmpty() insertValues->at(i("<<i<<")): " << insertValues->at(i) << endl;
	
		bp->getDataPages()->at( blockIndex )->clearPage();
		bp->getDataPages()->at( blockIndex )->addEntries( insertValues );
		
		//DEBUGGING ..
		//bp->getDataPages()->at( blockIndex )->printDataPage();

		// (d) UPDATE "bit" entries ..
		bit->getBufferPageRowCount()->at( blockIndex ) = 1;
				
		bit->getRowAvailable()->at( blockIndex )->at( 0 ) = false;
				
		for(int i=1; i<maxRowCount; i++)
			bit->getRowAvailable()->at( blockIndex )->at( i ) = true;

		// this was a "hit", so update the bit "hit rate" ..
		bit->setNumberHitRequests( bit->getNumberHitRequests() + 1 );
		bit->updateHitRate();

	} // end "if ( maxRowCount > 0 )" ..	
	
	
} // end "insertIntoEmpty(  InsertIntoEmptyAggregation* iiep  )" ..


void BufferManager:: insertIntoPartial( InsertIntoPartialAggregation* iipp ){
	
	// V: "insert-into-partial" ..
	
	if( iipp == NULL )
		return;
	
	int blockIndex = iipp->getBlockIndex();
	
	int numberOfAttributes = iipp->getNumberOfAttributes();
	
	vector<int>* insertValues = iipp->getInsertValues();
	
	
	if( blockIndex < 0 )
		return;
		
	if( numberOfAttributes < 0 )
		return;
		
	if( insertValues == NULL )
		return;
				
	//int insertValuesSize = insertValues->size();
	//for(int i=0; i<insertValuesSize; i++)
	//	cout << "BufferManager:: insertIntoPartial() insertValues->at(i("<<i<<")): " << insertValues->at(i) << endl;
	
				
	int maxRowCount = bit->computeMaxRowCount_Index( blockIndex );
					
	if( maxRowCount == -1 ){
						
		maxRowCount = bit->computeMaxRowCount_Attributes( numberOfAttributes );
						
	}
				
	// add a record with "numberOfAttributes" fields to the BP ..
/*
	// 1) save the current records in the block ..
	vector<DataPageRecord>* currentRecords = bp->getDataPages()->at( blockIndex )->getRecords();
			
	// get the current number of records for this block in the BP ..
	int currentNumberOfRecords = bit->getBufferPageRowCount()->at( blockIndex );
			
	// drop all the elements in BP "records" (for this index) to prepare to resize BP block ..
	bp->getDataPages()->at( blockIndex )->getRecords()->clear();
			
	// 2) assign space for an extra row to the BP block ..
	for(int i=0; i<( currentNumberOfRecords + 1 ); i++)
		bp->getDataPages()->at( blockIndex )->getRecords()->push_back( ( *new DataPageRecord() ) );	
*/
	// 3) fill the BP with the current block, leaving space for the new row ..
	int firstFreeRecIndex = -1;
				
	for(int i=0; i<maxRowCount; i++){
					
		//if ( bit->getRowAvailable()->at( blockIndex )->at( i ) == false ) {
						
			// (a) fill the BP with the current block ..
			//bp->getDataPages()->at( blockIndex )->getRecords()->at( i ) = currentRecords->at(i);
						
		//} else 
		
		if ( ( bit->getRowAvailable()->at( blockIndex )->at(i) == true ) && 
					( firstFreeRecIndex < 0 ) ) {
						
			// (b) track first slot available for the new row ..
			firstFreeRecIndex = i;
						
		}
					
				
	} // end "for()" ..
	
	//cout << "BufferManager:: insertIntoPartial() firstFreeRecIndex: " << firstFreeRecIndex << endl << endl;
				
/*	
	// 4) assign the final "new" record at the location of the first free row in the BP block ..
	bp->getDataPages()->at( blockIndex )->getRecords()->at( firstFreeRecIndex ) = ( * new DataPageRecord() );
			
	// 5) assign the "fields" array to the final "new" record ..
	bp->getDataPages()->at( blockIndex )->getRecords()->at( firstFreeRecIndex ). getFields()->clear();
				
	//for(int i=0; i<numberOfAttributes; i++)
		bp->getDataPages()->at( blockIndex )->getRecords()->at( firstFreeRecIndex ). getFields()->push_back( *new DataPageField() );
			
	// 6) assign "Value"s to each fields in the new record in the bp block ..
	    loop through each of the fields in the record ..
	for( int i=0; i < numberOfAttributes; i++ ){
			
		bp->getDataPages()->at( blockIndex )->getRecords()->at( firstFreeRecIndex ). getFields()->at( i ).setValue( insertValues->at(i) );
			
	} // end "for()" ..
*/
	
	int numberOfRows = bp->getDataPages()->at( blockIndex )->getNumberOfRows();
	
	bool areEntriesReplaced = false;
	bool areEntriesAdded = false;
	
	//cout << "BufferManager:: insertIntoPartial() numberOfRows: " << numberOfRows << endl << endl;
	
	if( firstFreeRecIndex < numberOfRows ){
		
		// this means that the new row will replace an existing row ..
		areEntriesReplaced = bp->getDataPages()->at( blockIndex )->replaceEntries( firstFreeRecIndex, insertValues );
		
	} else {
		
		// this means that the new row is appended to the existing rows ..
		areEntriesAdded = bp->getDataPages()->at( blockIndex )->addEntries( insertValues );
		
		//cout << "BufferManager:: insertIntoPartial() areEntriesAdded: " << areEntriesAdded << endl << endl;
		
		//int insertValuesSize = insertValues->size();
		//for(int i=0; i<insertValuesSize; i++)	
		//	cout << "BufferManager:: insertIntoPartial() insertValues->at(i("<<i<<")): " << insertValues->at(i) << endl << endl;
		
		// DEBUGGING ..
		//bp->getDataPages()->at( blockIndex )->printDataPage();
		
	}
		
	// UPDATE "bit" entries ..
	bit->getBufferPageRowCount()->at( blockIndex )++;
						
	bit->getRowAvailable()->at( blockIndex )->at( firstFreeRecIndex ) = false;
	
	// this was a "hit", so update the bit "hit rate" ..
	bit->setNumberHitRequests( bit->getNumberHitRequests() + 1 );
	bit->updateHitRate();
	
} // end "insertIntoPartial()" ..


bool BufferManager:: insertRecordIntoBlock( string tableNameX, vector<int>* insertValues, int numberOfAttributes ){
	
	//cout << "BufferManager:: insertRecordIntoBlock() start .." << endl << endl;
	
	//cout << "BufferManager:: insertRecordIntoBlock()               tableNameX: " << tableNameX << endl << endl;
	
	//cout << "BufferManager:: insertRecordIntoBlock() ( insertValues == NULL ): " << ( insertValues == NULL ) << endl << endl;
	
	//cout << "BufferManager:: insertRecordIntoBlock()       numberOfAttributes: " << numberOfAttributes << endl << endl;
	
	if( tableNameX.length() < 0 )
		return false;
	
	if( insertValues == NULL )
		return false;
		
	if( numberOfAttributes < 0 )
		return false;
	
	// helper methods:
	// 1) getFreeAndFlushBlockIndexProcessing()
	//		a) findBlockForInsert()
	// 2) flushBPBlock()[ I ]
	// 3) findLRUAvailableBlockForFlushing()
	// 4) getHeaderEntryResults() 
	// 5) loadAndInsertIntoPartial() [ II, V ]
	// 6) insertIntoEmpty() [ IV ]
	// 7) insertIntoPartial() [ V ]
	
	//cout << "BufferManager:: insertRecordIntoBlock() Point A .." << endl << endl;

	int intSize = 4;
	
	//int recIndex = -1;
	//int maxRowCount = -1;

	
	// 1) Determine whether to insert into a "free" block in the bp,
	//    load in a "free" page-from-file into a "flushed" bp block, or
	//    insert into an empty "flushed" block ..
	//int freeBlockIndex = findBlockForInsert( tableNameX, numberOfAttributes );
	BlockForInsert* bfi = findBlockForInsert( tableNameX, numberOfAttributes );
	
	//if( bfi == NULL )
	//	return false;
	
	int freeBlockIndex = -1;
	int blockForInsertState = BlockForInsert::NONE;
	
	if( bfi != NULL ){
		
		freeBlockIndex = bfi->getBpBlock();
		
		blockForInsertState = bfi->getState();
		
	}
	
	//cout << "BufferManager:: insertRecordIntoBlock() freeBlockIndex: " << freeBlockIndex << ", blockForInsertState: " << blockForInsertState << endl << endl;
	
	
	FreeAndFlushBlockIndexAggregation* fafbip = getFreeAndFlushBlockIndexProcessing( freeBlockIndex );

	if( fafbip == NULL )
		return false;

	bool insertPerformed = fafbip->getInsertPerformed();
	
	bool freeBlockFound = fafbip->getFreeBlockFound();
	
	bool flushBlockFound = fafbip->getFlushBlockFound();
	int flushBlockIndex = fafbip->getFlushBlockIndex();
	
	//cout << "BufferManager:: insertRecordIntoBlock() flushBlockIndex: " << flushBlockIndex << endl << endl;
	
	//cout << "BufferManager:: insertRecordIntoBlock() insertPerformed: " << insertPerformed << endl << endl;
	
	// clean up "FreeAndFlushBlockIndexAggregation" object 
	// now that we have its data results ..
	delete fafbip;
	

	if ( insertPerformed ){
	
		// 2) only proceed if it was determined above ("insertPerformed == true")
		//    that the record can be inserted (empty-insert, load-and-insert, flush/free) ..
	
		// Operations:
		//   I: "FLUSH" - write bp block data to file, making room for new record insert ..
		//  II: "LOAD" - retrieve page-from-file, load into bp block, has free row for new record insert ..
		// III: "Overwrite(no flush)" - insert new record into bp block, destroying existing data ..
		//  IV: "insert-into-empty" - insert new record into bp empty / free bp block ..
		//   V: "insert-into-partial" - insert new record into partially full bp block ..
	
		// "( !freeBlockFound && flushBlockFound )":
		// I & II & V: "FLUSH" + "LOAD" + "insert-into-partial"
		//     I & IV: "FLUSH" + "insert-into-empty" ..
		//   III & IV: "Overwrite(no flush)" + "insert-into-empty" ..
		
		//cout << "BufferManager:: insertRecordIntoBlock() freeBlockFound: " << freeBlockFound << ", freeBlockIndex: " << freeBlockIndex << ", flushBlockFound: " << flushBlockFound << ", flushBlockIndex: " << flushBlockIndex << endl << endl;

		if ( !freeBlockFound && flushBlockFound ) {
			
			//cout << "BufferManager:: insertRecordIntoBlock() Point B .." << endl << endl;
		
			bool readHeaderPerformed = false;
		
			// 3) NO available block for insert, so attempt to "flush" a block ..
			
			// Note: a row can only be inserted into an empty block, or
			//       a block which is not full which has rows that belong to the same
			//		 table as the insert-row.  Therefore, a) "full" BP blocks (with rows 
			//		 from any table whatsoever as long as they are the same table), or 
			//		 b) partially full BP blocks ("partials") with rows that belong to 
			//		 tables that are different from the insert-row, cannot have this row 
			//		 inserted into a BP block.  Both the full and the partial BP block must be 
			//		 "flushed" first (block is written to the table-file).
			
			// determine how to flush the current block in the BP to the file ..
			

			// need DataHeader info on block in file for page in BP to be flushed to ..
			DataHeader* header = new DataHeader();
			readHeaderPerformed = getStorageManager()->readHeader( bit->getTableName()->at( flushBlockIndex ), header );
			
			// if "readHeader()" failed,
			// abort "insertRecordIntoBlock()" ..
			if ( readHeaderPerformed == false ) {
				
	 			string errorString = "\nFailed to read Header [ BM::insertRecordIntoBlock() ], data may not have been inserted.";
	 	
	 			// update the "MessageManager" to track this error and output to user ..
				getMessageManager()->setErrorStringQM( errorString );
				
				return false;
				
			}
			
			//cout << "BufferManager:: insertRecordIntoBlock() Point C .." << endl << endl;
			
			
			// 4) flush the block and insert the _new_ record into the 
			//    free/flushed block space of the BP ..
			//    found an available block to flush b/c 
			//    "if ( flushBlockFound )" means ( pin_count=0, LRU ) ..
			
			// I: "FLUSH" operation ..
			
			//cout << "BufferManager:: insertRecordIntoBlock() ( bit->getDirty()->at( flushBlockIndex("<<flushBlockIndex<<") )("<<( bit->getDirty()->at( flushBlockIndex ) == true )<<") == true ): " << ( bit->getDirty()->at( flushBlockIndex ) == true ) << endl << endl;
			
			if( bit->getDirty()->at( flushBlockIndex ) == true ){
				
				//cout << "BufferManager:: insertRecordIntoBlock() Point C1 .." << endl << endl;
				
				bool flushBPBlockPerformed = flushBPBlock( flushBlockIndex, header );
				
				// if "flushBPBlock()" failed, 
				// abort "insertRecordIntoBlock()" ..
				if ( flushBPBlockPerformed == false ) {
				
	 				string errorString = "\nFailed to flush buffer pool block [ BM::insertRecordIntoBlock() ], data may not have been inserted.";
	 	
	 				// update the "MessageManager" to track this error and output to user ..
					getMessageManager()->setErrorStringQM( errorString );
				
					return false;
				
				}
	
			} // "if( bit.dirty[ flushBlockIndex ] == true )" ..
			
			//cout << "BufferManager:: insertRecordIntoBlock() Point D .." << endl << endl;

			// workaround ..
			// for an unknown reason, C++ swaps in a garbage value into "flushBlockIndex"
			// whenever user: inserts enough rows to fill buffer, user subsequently flushes each 
			// page of the buffer, and user inserts another row into a flushed buffer;
			// so reset the value here to get around this non-linear bug ..
			//flushBlockIndex = findLRUAvailableBlockForFlushing();
		
			//cout << "BufferManager:: insertRecordIntoBlock() flushBlockIndex: " << flushBlockIndex << endl << endl;
		
			//cout << "BufferManager:: insertRecordIntoBlock() Point D4 .." << endl << endl;
			
			// I & II & V: "FLUSH" + "LOAD" + "insert-into-partial" ..
		
			// or
		
			// I & IV: "FLUSH" + "insert-into-empty" ..
			
			// or
			
			// III & IV: "Overwrite(no flush)" + "insert-into-empty" ..
		

			// "clear" / overwrite the block in the BP by 
			// inserting this new record into it ..
			// (should have already "flushed" the block if the data was to be saved) ..
			
			// get the "page-from-file" where this block will be "flushed" / written to ..
			
			// need HP info of the page-from-file to be loaded in ..
			//sm->readHeader( (char*) tableNameX.c_str(), hp );
			readHeaderPerformed = getStorageManager()->readHeader( tableNameX, header );
			
			// if "readHeader()" failed, 
			// abort "insertRecordIntoBlock()" ..
			if ( readHeaderPerformed == false ) {
				
	 			string errorString = "\nFailed to read Header [ BM::insertRecordIntoBlock() ], data may not have been inserted.";
	 	
	 			// update the "MessageManager" to track this error and output to user ..
				getMessageManager()->setErrorStringQM( errorString );
				
				return false;
				
			}
			
			// 5) Get the "pageID" to use in "loadBlock()" method,
			//	  "headerEntryIndex" going along with this pageID, and
			//    "hpEntryFound" for control logic ..
			
			// first, get the "pageIDs" that shouldnt be loaded from file,
			// because the bp has the "latest" copies,
			// and the file has "old" copies ..
			vector<int>* avoidTheseBPPageIDs = getDirtyBufferPoolBlocks();
			
			HeaderEntryResultsAggregation* herp = getHeaderEntryResults( header, avoidTheseBPPageIDs );
			
			if( herp == NULL )
				return false;
			
			// "pageID" to use in "loadBlock()" method, default = -1 ..
			int loadPageID = herp->getLoadPageID();

			// "headerEntryIndex" corresponding to the "loadPageID", default = -1 ..
			int headerEntryIndex = herp->getHeaderEntryIndex();

			// "hpEntryFound" for control logic below ..
			bool hpEntryFound = herp->getHpEntryFound();
			
			//if( loadPageID < 0 )
			//	return false;
		
			//if( headerEntryIndex < 0 )
			//	return false;
			
			// cleanup ..
			delete herp;
			
			//cout << "BufferManager:: insertRecordIntoBlock() hpEntryFound-flush: " << hpEntryFound << ", loadPageID: " << loadPageID << endl << endl;
			

			// 6) "free slot" in the file to insert this row into ..
			if ( hpEntryFound ) {
				
				// I & II & V: "FLUSH" + "LOAD" + "insert-into-partial" ..
				
				//cout << "BufferManager:: insertRecordIntoBlock() Point D5 .." << endl << endl;
				
				// this means there IS a "free slot" in the file to 
				// insert this row into ..
				// load the page corresponding to "pageID" into the BP,
				// (overwrite the current contents of the BP block with 
				//  the loaded page) ..
					
				LoadAndInsertIntoPartialAggregation* laiipp = new LoadAndInsertIntoPartialAggregation();
				
				laiipp->setHeader( header );
				laiipp->setHeaderEntryIndex( headerEntryIndex );
				laiipp->setBlockIndex( flushBlockIndex );
				laiipp->setLoadPageID( loadPageID );
				laiipp->setTableNameX( tableNameX );
				laiipp->setNumberOfAttributes( numberOfAttributes );
				laiipp->setInsertValues( insertValues );
				
				// this always performs an "insert" operation,
				// but if "load" operation fails (due to failure to open a file), 
				// "insertIntoEmpty()" method is called inside of 
				// "loadAndInsertIntoPartial()" ..
				loadAndInsertIntoPartial( laiipp );
				
				// cleanup; once "loadAndInsertIntoPartial" is finished, 
				// this object can be reclaimed ..
				delete laiipp;
			
					
			// 7) NO "free slot" in the file to insert this row into ..
			} else {
				
				//cout << "BufferManager:: insertRecordIntoBlock() Point D6 .." << endl << endl;
				
				// I & IV: "FLUSH" + "insert-into-empty" ..
			
				// or
			
				// III & IV: "Overwrite(no flush)" + "insert-into-empty" ..
				
				// this means there were no "free slots" in the file,
				// or that either (a) b/c there are no free slots, didn't load in
				// a page-from-file, or (b) there is at least one free slot
				// in the file, but failed to retrieve it correctly,
				// so will need to allocate a new page in the file 
				// to be inserted into upon "flush" 
				// (see "flush" section above) or "commit" ..
					
				InsertIntoEmptyAggregation* iiep = new InsertIntoEmptyAggregation();
					
				iiep->setBlockIndex( flushBlockIndex );
				iiep->setNumberOfAttributes( numberOfAttributes );
				iiep->setInsertValues( insertValues );
				
				// this is always performed 
				// (doesn't depend on reading / opening the "DataHeader" file or 
				//  datapage file) ..
				insertIntoEmpty( iiep );

				// cleanup; once "insertIntoEmpty" is finished, 
				// this object can be reclaimed ..
				delete iiep;
			
			} // end "if( hpEntryFound )" ..
			
			//cout << "BufferManager:: insertRecordIntoBlock() Point D7 .." << endl << endl;
			
			// 8) UPDATE bit entries now that the operation is complete ..
			

			bit->getTableName()->at( flushBlockIndex ) = tableNameX;

			bit->getBufferPageFieldCount()->at( flushBlockIndex ) = numberOfAttributes;

			bit->getRecordSize()->at( flushBlockIndex ) = ( intSize * numberOfAttributes );

			// don't need this; this was set up above in this section ..
			//bit.pageID[ flushBlockIndex ] = pageID;		// either (a) current pageID, or (b) allocated pageID ..

			bit->getPinCount()->at( flushBlockIndex ) = 0;			// default for insert operation ..

			//cout << "BufferManager:: insertRecordIntoBlock() flushBlockIndex " << flushBlockIndex << " is dirty." << endl << endl;

			bit->getDirty()->at( flushBlockIndex ) = true;			// mark for "write-to-file" upon "commit", default for insert operation ..

			bit->getEmpty()->at( flushBlockIndex ) = false;			// block is not empty (just inserted a record into it) ..

			bit->getLRUBlocks()->remove( flushBlockIndex );		// remove block position from list ( where ever it is )..
			bit->getLRUBlocks()->push_back( flushBlockIndex );		// set block position at end of list ( its MRU ) ..

			// this is done in "insertIntoXXX()" and "loadAndInsertIntoXXX()" ..
			//bit->setTotalNumberBlockRequests( bit->getTotalNumberBlockRequests() + 1 );		// this is a "block request" ..
			//bit->updateHitRate();							// hit rate changes (whether hit or miss) ..

			bit->updateTableCountUnique();					// count of unique "tableName" associated with blocks in the BP ..
	

			//cout << "BM:: insertIntoBlock()     hits: " << bit->getNumberHitRequests() << endl;
			//cout << "BM:: insertIntoBlock() requests: " << bit->getTotalNumberBlockRequests() << endl;
			//cout << "BM:: insertIntoBlock() hit rate: " << bit->getHitRate() << endl << endl;

		} else if ( freeBlockFound && !flushBlockFound ) {
			
			//cout << "BufferManager:: insertRecordIntoBlock() Point D8 .." << endl << endl;

			// ( freeBlockFound && !flushBlockFound ):
			//      	V: "insert-into-partial" ..
			// 	   II & V: "LOAD" & "insert-into-partial" ..
			//     	   IV: "insert-into-empty" ..
		
			// 3) found an available block for insert ..
		
			// BP has a "free" block available for insert of a row ..
			// BP block is either "partial" full (page _may_ have been loaded in from file) 
			// or "empty" (_may_ need to load in a page from file) ..
			
			DataPage* loadedDataPage = new DataPage();
			
			// 4) the bp block found for insert is partially full (not empty) ..
			if ( blockForInsertState == BlockForInsert::PARTIALLY_FULL_SAME_TABLE ) {
				
				//cout << "BufferManager:: insertRecordIntoBlock() Point D9 .." << endl << endl;
				
				// V: "insert-into-partial" ..
				
				InsertIntoPartialAggregation* iipp = new InsertIntoPartialAggregation();
				
				iipp->setBlockIndex( freeBlockIndex );
				iipp->setNumberOfAttributes( numberOfAttributes );
				iipp->setInsertValues( insertValues );
				

				// this is always performed 
				// (doesn't depend on reading / opening the "DataHeader" file or 
				//  datapage file) ..
				insertIntoPartial( iipp );
				
				// cleanup; once "insertIntoPartial" is finished, 
				// this object can be reclaimed ..
				delete iipp;
				
				
			// 5) the bp block is may be overwritten or it is empty, 
			//    so load and insert into a partially full (not empty) page-from-file,
			//    or insert directly into the empty bp block ..
			} else if ( ( blockForInsertState == BlockForInsert::EMPTY ) || 
						( blockForInsertState == BlockForInsert::NOT_DIRTY ) ){
				
				//cout << "BufferManager:: insertRecordIntoBlock() Point D10 .." << endl << endl;
			
				// HERE: bit.empty[ freeBlockIndex ] == true ..
				// this means that the BP block at "freeBlockIndex" is "EMPTY" ..

				// II & V: "LOAD" & "insert-into-partial" ..
				// see if we can load in a page-from-file at the "freeBlockIndex" and insert the new row
				
				// OR
				
				// IV: "insert-into-empty" ..
				// insert into an "empty" block at the "freeBlockIndex".
				// (A page-from-file needs to be "allocated" 
				//  [see "( !freeBlockIndex && flushBlockIndex )" section above],
				//  prior to (future) flushing the block to disk ) ..
			
				DataHeader* header = new DataHeader();
				bool readHeaderPerformed = getStorageManager()->readHeader( tableNameX, header );
			
				//cout << "BufferManager:: insertRecordIntoBlock() Point D10-1 .." << endl << endl;

				// if "readHeader()" failed, 
				// abort "insertRecordIntoBlock()" ..
				if ( readHeaderPerformed == false ) {
				
	 				string errorString = "\nFailed to read Header [ BM::insertRecordIntoBlock() ], data may not have been inserted.";
	 	
	 				// update the "MessageManager" to track this error and output to user ..
					getMessageManager()->setErrorStringQM( errorString );
				
					return false;
				
				}
				
				//cout << "BufferManager:: insertRecordIntoBlock() Point D10-2 .." << endl << endl;
			
				// (a) Get the "pageID" to use in "loadBlock()" method,
				//	   "headerEntryIndex" corresponding to the pageID, and
				//     "hpEntryFound" for control logic ..
				
				// first, get the "pageIDs" that shouldnt be loaded from file,
				// because the bp has the "latest" copies,
				// and the file has "old" copies ..
				vector<int>* avoidTheseBPPageIDs = getDirtyBufferPoolBlocks();
				
				HeaderEntryResultsAggregation* herp = getHeaderEntryResults( header, avoidTheseBPPageIDs );
								
				if( herp == NULL )
					return false;
			
				// "pageID" to use in "loadBlock()" method, default = -1 ..
				int loadPageID = herp->getLoadPageID();
			
				// "headerEntryIndex" corresponding to the "loadPageID", default = -1 ..
				int headerEntryIndex = herp->getHeaderEntryIndex();

				// "hpEntryFound" for control logic below ..
				bool hpEntryFound = herp->getHpEntryFound();
				
				//cout << "BufferManager:: insertRecordIntoBlock() hpEntryFound-free: " << hpEntryFound << ", loadPageID: " << loadPageID << endl << endl;
				
				//cout << "BufferManager:: insertRecordIntoBlock() Point D10-3 .." << endl << endl;
				
				//if( loadPageID < 0 )
				//	return false;
					
				//if( headerEntryIndex < 0 )
				//	return false;

				// (b) entry found in the DataHeader ..
				if( hpEntryFound ){
					
					//cout << "BufferManager:: insertRecordIntoBlock() Point D11 .." << endl << endl;
					
					// II & V: "LOAD" & "insert-into-partial" ..
				
					// this means there IS a "free slot" in the file to insert this row into ..
					// load the page corresponding to "pageID" into the BP,
					// overwrite the current contents of the BP block with the loaded page ..				
					
					
					LoadAndInsertIntoPartialAggregation* laiipp = new LoadAndInsertIntoPartialAggregation();
				
					laiipp->setHeader( header );
					laiipp->setHeaderEntryIndex( headerEntryIndex );
					laiipp->setBlockIndex( freeBlockIndex );
					laiipp->setLoadPageID( loadPageID );
					laiipp->setTableNameX( tableNameX );
					laiipp->setNumberOfAttributes( numberOfAttributes );
					laiipp->setInsertValues( insertValues );
					

					// this always performs an "insert" operation,
					// but if "load" operation fails (due to failure to open a file), 
					// "insertIntoEmpty()" method is called inside of 
					// "loadAndInsertIntoPartial()" ..
					loadAndInsertIntoPartial( laiipp );

					// cleanup; once "loadAndInsertIntoPartial" is finished, 
					// this object can be reclaimed ..
					delete laiipp;

				// (c) no entry found in the DataHeader, insert into empty block ..
				} else {
					
					//cout << "BufferManager:: insertRecordIntoBlock() Point D12 .." << endl << endl;
					
					InsertIntoEmptyAggregation* iiep = new InsertIntoEmptyAggregation();
					
					iiep->setBlockIndex( freeBlockIndex );
					iiep->setNumberOfAttributes( numberOfAttributes );
					iiep->setInsertValues( insertValues );
				
					// insert directly into an empty block ..
					insertIntoEmpty( iiep );
					
					// cleanup; once "insertIntoEmpty" is finished, 
					// this object can be reclaimed ..
					delete iiep;
	
				} // end "if( hpEntryFound )" ..

				
			
			}  // end "if ( bit.empty[ freeBlockIndex ] == false )" ..

			//cout << "BufferManager:: insertRecordIntoBlock() Point D13 .." << endl << endl;
		
			// track the update ..
		
			bit->getTableName()->at( freeBlockIndex ) = tableNameX;
				
			bit->getBufferPageFieldCount()->at( freeBlockIndex ) = numberOfAttributes;
			
			bit->getRecordSize()->at( freeBlockIndex ) = ( numberOfAttributes * intSize );

			// don't need this; this was set up above in this section ..
			//bit.pageID[ freeBlockIndex ] = pageID;		// either (a) current pageID, or (b) allocated pageID ..

			bit->getPinCount()->at( freeBlockIndex ) = 0;			// default for insert operation ..
			
			bit->getDirty()->at( freeBlockIndex ) = true;			// mark for "write-to-file" upon "commit" ..
			bit->getEmpty()->at( freeBlockIndex ) = false;			// block is not empty (just inserted a record into it) ..
			
			bit->getLRUBlocks()->remove( freeBlockIndex );			// remove block position from list ( where ever it is )..
			bit->getLRUBlocks()->push_back( freeBlockIndex );		// set block position at end of list ( its MRU ) ..

			// this is done in "insertIntoXXX()" and "loadAndInsertIntoXXX()" ..
			//bit->setTotalNumberBlockRequests( bit->getTotalNumberBlockRequests() + 1 );					// this function is a new "block request" ..	
			//bit->updateHitRate();							        // hit rate changes (whether hit or miss) ..

			//cout << "BM:: insertRecordIntoBlock()     hits: " << bit->getNumberHitRequests() << endl;
			//cout << "BM:: insertRecordIntoBlock() requests: " << bit->getTotalNumberBlockRequests() << endl;
			//cout << "BM:: insertRecordIntoBlock() hit rate: " << bit->getHitRate() << endl << endl;

			bit->updateTableCountUnique();					        // count of unique "tableName" associated with blocks in the BP ..

		
		} // end "if ( !freeBlockFound && flushBlockFound )" ..
	

	} // end "if (insertPerformed)" ..
	
	//cout << "BufferManager:: insertRecordIntoBlock() Point D14 .." << endl << endl;

	return insertPerformed;
	
} // end "insertRecordIntoBlock( DataPageRecord &record, int numberOfAttributes )" ..


//bool BufferManager:: commitBufferPool(){
bool BufferManager:: commitBufferPool( vector<string>* tableNames ){
	
	if( tableNames == NULL )
		return false;
	
	bool commitPerformed = false;
	
	string tableNameX = "";
	
	int bufferPoolSize = bp->getBufferPoolSize();
	
	for(int i=0; i<bufferPoolSize; i++){
		
		// 1) skip BP pages that are empty or that are _not_ dirty ..
		if ( bit->getEmpty()->at( i ) || !bit->getDirty()->at( i ) ){
			
			continue;
			
		}
		
		//cout << "BufferManager:: commitBufferPool() Point A .." << endl << endl;
		
		tableNameX = bit->getTableName()->at( i );
		
		//cout << "BufferManager:: commitBufferPool() Point B .." << endl << endl;
		
		// track the tables that are committed here ..
		tableNames->push_back( tableNameX );
	
		// 3) ask Catalog if the TABLE-to-insert-into EXISTS ..
		bool doesTableNameXExist = getCatalogManager()->doesCatalogTableNameExist( tableNameX );
		
		// 4) table exists, so proceed ..
		if( doesTableNameXExist ){
			
			int pageID = -1;
				
			DataHeader* header = new DataHeader();
			bool readHeaderPerformed = getStorageManager()->readHeader( tableNameX, header );
				
			// if "readHeader()" is not performed, 
			// abort "commitBufferPool()" ..
			if ( readHeaderPerformed == false ) {
				
	 			string errorString = "\nFailed to read Header [ BM::commitBufferPool() ], data may not have been written to file.";
	 	
	 			// update the "MessageManager" to track this error and output to user ..
				getMessageManager()->setErrorStringQM( errorString );
				
				return false;
				
			}
				
			// (a) this indicates that a new page has to be ALLOCATED in the file,
			//     and a new "PageID" needs to be created to identify the new page-in-file ..
			if ( bit->getPageID()->at( i ) == -1 ){
					
				// (i) find the current max PageID in the DataHeader ..
				int currentMaxPageID = -1;
				for( int j=0; j<header->getTotalNumberOfDataPages(); j++ ){
				
					if ( header->getHeaderEntries()->at( j )->getPageID() > currentMaxPageID ) {
					
						currentMaxPageID = header->getHeaderEntries()->at( j )->getPageID();
					
					}
				
				} // end "for()" ..
					
				
				// (ii) the "PageID" for the allocated page should be 1 greater than 
				//      the current max in the DataHeader ..
				pageID = ( currentMaxPageID + 1 );
				//pageID = ( hp.dataPageentries[ ( hp.totalDataPages - 1 ) ].PageID + 1 );
				
				// (iii) track the new pageID in the "bit" ..
				bit->getPageID()->at( i ) = pageID;
				
				//cout << "IndexPages:: commitBufferPool() Point A .." << endl << endl;
				
				// (iv) allocate a new page to the file for this table,
				//      where this block in the BP will be "flushed" / saved / written to ..
				bool allocateDataPagePerformed = getStorageManager()->makeDataPage( bit->getTableName()->at( i ), pageID );

				// if "makeDataPage()" not performed,
				// "flushBPBlock()" should not be performed, so abort ..
				if ( allocateDataPagePerformed == false ) {
			
	 				string errorString = "\nFailed to allocate data page [ BM::commitBufferPool() ], data may not have been written to file.";
	 	
	 				// update the "MessageManager" to track this error and output to user ..
					getMessageManager()->setErrorStringQM( errorString );
			
					return false;
			
				}

					
			} // end "if ( bit.pageID[ i ] == -1 )" ..
				
			// (b) update "bitIndex" just prior to writing this block to file (below),
			//     and when this block of code finishes,
			//     "bitIndex" will correspond to the final block written ..
			BufferInfoTable::bitIndex = i;
			
			//cout << "IndexPages:: commitBufferPool() Point B .." << endl << endl;
			
			if( pageID >= 0 )
				bp->getDataPages()->at( i )->setPageID( pageID );
			else
				bp->getDataPages()->at( i )->setPageID( bit->getPageID()->at( i ) );
			
			// (c) need to write this current block to file before "flushing" it ..
			bool writeBlockPerformed = writeBlock( bp->getDataPages()->at( i ), locks );
			
			// if "writeBlock()" not performed,
			// "flushBPBlock()" should not be perfoemd, so abort ..
			// (ok if "makeDataPage()" was performed) ..
			if ( writeBlockPerformed == false ) {
			
	 			string errorString = "\nFailed to write buffer pool block [ BM::commitBufferPool() ], data may not have been written to file.";
	 	
	 			// update the "MessageManager" to track this error and output to user ..
				getMessageManager()->setErrorStringQM( errorString );
			
				return false;
			
			}
				
/*
			// clear "locks" map of invalid "tableName" after operating on "locks" (in writeBlock()); 
			// seems to need this immediately following the function call to "writeBlock()" ..
				
			//purgeLocksMap( locks );
				
			// DEBUGGING ..
			//cout << endl << "locks .." << endl << endl;
				
			//char* tableName = (char*) bit.tableName[ i ].c_str();

			vector<int> pageID;
			vector<int>::iterator it_pageIDs;
			map< char*, vector<int> >::iterator it_locks;
			//cout << "locks.count( tableName("<<tableName<<") ): " << locks.count( tableName ) << endl;
			for(it_locks = locks.begin(); it_locks != locks.end(); it_locks++){
	
				//cout << "lock: " << ( *it_locks ).first << endl;
				//cout << "(( *it_locks ).second).size(): " << (( *it_locks ).second).size() << endl;
				pageID = ( *it_locks ).second;
				for(it_pageIDs = pageID.begin(); it_pageIDs != pageID.end(); it_pageIDs++){
			
					cout << "BM:: commitBufferPool() locks.pageID: " << ( *it_pageIDs ) << endl;
			
				}
			
			}

			//cout << endl << "end locks .." << endl << endl;
*/
				
			// (d) must UPDATE the CATALOG ..

			// get the existing table size ..
			int newTableSize = getCatalogManager()->getTableSize( bit->getTableName()->at( i ) );
			
			//cout << "BufferManager:: commitBufferPool() oldTableSize: " << newTableSize << endl << endl;
			
			//int dataSize = ( bit->getRecordSize()->at( i ) ) * ( bit->getBufferPageRowCount()->at( i ) );

			//int padSize = ( blockSize - dataSize );
			
			// add size of new rows to existing table size ..
			newTableSize += blockSize;
			
			// update the catalog with the new table size ..
			//cout << "BufferManager:: commitBufferPool() updatedTableSize: " << newTableSize << endl << endl;
			getCatalogManager()->setTableSize( bit->getTableName()->at( i ), newTableSize );
				
				
			// Track the latest commit ..
			getLastFlushedBlocks()->push_back( bp->getDataPages()->at( i ) );
			getLastFlushedBlockPageIDs()->push_back( pageID );
				
			// 5) UPDATE the "bit" ..
		
			// DEBUGGING ..
			//cout << "bitTableNames(i("<<i<<")): " << bit->getTableName()->at( i ) << endl;
/*
			bit->getTableName()->at( i ) = "";
			
			bit->getBufferPageFieldCount()->at( i ) = 0;
		
			bit->getBufferPageRowCount()->at( i ) = 0;
			
			bit->getRecordSize()->at( i ) = 0;
			
			bit->getPageID()->at( i ) = -1;

			bit->getPinCount()->at( i ) = 0;
		
			bit->getDirty()->at( i ) = false;				// "commit" occurred, so reset to default ..
			bit->getEmpty()->at( i ) = true;				// block is "empty" (just flushed block to file); free to be written into ..
*/
			bit->getPinCount()->at( i ) = 0;				// free to be overriden ..
			bit->getDirty()->at( i ) = false;				// "not dirty", ready to be overwritten ..
			bit->getEmpty()->at( i ) = false;				// "not empty", keep the value cached ..

			bit->getLRUBlocks()->remove( i );				// remove block position from list ( where ever it is )..
			bit->getLRUBlocks()->push_back( i );			// set block position at end of list ( its MRU ) ..
		
			// for "commit", this is neither a 
			// "block request" nor a "hit" ..
			//bit.totalNumberBlockRequests++;				// this function is a new "block request" ..
			//bit.updateHitRate();							// hit rate changes due to line above ..
			bit->updateTableCountUnique();					// count of unique "tableName" associated with blocks in the BP ..
		
			// 2) if bp block is tracked by the bit as non-empty + dirty,
			//    then at least one bp block will be committed ..
			if ( commitPerformed == false )
				commitPerformed = true;
		
				
			}  // end "if( doesTableNameXExist )" ..
		
		
	} // end "for()" ..
	
	// DEBUGGING ..
	//vector<string>* bitTableNames = bit->getTableName();
	//int bitTableNamesSize = bitTableNames->size();
	//for(int i=0; i<bitTableNamesSize; i++){
	//			
	//	cout << "bitTableNames: " << bitTableNames->at(i) << endl;	
	//			
	//}
	
	return commitPerformed;
	
} // end "commitBufferPool()" ..


DataPage* BufferManager:: returnBPBlock( int bitIndex ){
	
	if( bitIndex < 0 )
		return NULL;
	
	// invoking this method indicates that "pageId" was found in the BP ( "Block Hit" ) ..
	
	//cout << "BufferManager:: returnBPBlock() start .." << endl << endl;
	
	// 1) set this for use in QM::selectStarFromX() and QM::selectColumnsFromX() ..
	BufferInfoTable::bitIndex = bitIndex;

	DataPage* returnDataPage = new DataPage();
	
	// 2) copy the BP "dataPage" into the "returnDataPage",
	//    makes sure to only copy current rows,
	//    excluding any rows left over from a previous "flush" operation ..

/*
	for(int i=0; i<bit->getBufferPageRowCount()->at( bitIndex ); i++){
		
		returnDataPage->getRecords()->push_back( ( *new DataPageRecord() ) );
		
		returnDataPage->getRecords()->at( i ).getFields()->clear();
		
		for(int j=0; j<bit->getBufferPageFieldCount()->at( bitIndex ); j++) {
			
			returnDataPage->getRecords()->at( i ).getFields()->push_back( *new DataPageField() );
		
			// copy the BP "dataPage" into the "returnDataPage" ..
			returnDataPage->getRecords()->at( i ).getFields()->at( j ).setValue( bp->getDataPages()->at( bitIndex )->getRecords()->at( i ).getFields()->at( j ).getValue() );
		
		} // end "for()" ..
		
	} // end "for()" ..
*/
	
	//cout << "BufferManager:: returnBPBlock() Point A .." << endl << endl;
	
	vector<int>* entries = NULL;
	
	DataPage* bpDataPage = bp->getDataPages()->at( bitIndex );
	
	returnDataPage->setPageID( bpDataPage->getPageID() );
	returnDataPage->setMaxPageSize( bpDataPage->getMaxPageSize() );
	
	int numberOfRows = bpDataPage->getNumberOfRows();
	
	for(int i=0; i<numberOfRows; i++){
		
		entries = bp->getDataPages()->at( bitIndex )->getEntries( i );
		
		// DEBUGGING ..
		//for(int j=0; j<entries->size(); j++){
		//	cout << "BufferManager:: returnBPBlock() entries->at(j("<<j<<")): " << entries->at(j) << endl << endl;
		//}
		
		returnDataPage->addEntries( entries );
		
	}
	
	//cout << "BufferManager:: returnBPBlock() Point B .." << endl << endl;

	// 3) track the updated BP block ..
	// NOTE: tableName, bufferPageFieldCount, bufferPageRowCount, recordSize, pageID, dirty shouldn't have changed ..

	//bit.pin_count[ bitIndex ]++;							// default for select operation, not implemented ..
	bit->getEmpty()->at( bitIndex ) = false;				// block is not empty (just loaded a page into it) ..
		
	// since "bit.LRUBlocks" is used to figure 
	// out the order for printing of the BP,
	// do not update it here ..
	//bit.LRUBlocks.remove( bitIndex );						// remove block position from list ( where ever it is )..
	//bit.LRUBlocks.push_back( bitIndex );					// set block position at end of list ( its MRU ) ..
		
	//bit->setTotalNumberBlockRequests( bit->getTotalNumberBlockRequests() + 1 );				// this is a "block request" ..
	bit->setNumberHitRequests( bit->getNumberHitRequests() + 1 );// this is a "hit" ..
	bit->updateHitRate();									// hit rate changes (this is a hit) ..
	bit->updateTableCountUnique();							// count of unique "tableName" associated with blocks in the BP ..

	bit->getLRUBlocks()->remove( bitIndex );				// remove block position from list ( where ever it is )..
	bit->getLRUBlocks()->push_back( bitIndex );				// set block position at end of list ( its MRU ) ..

	//cout << "BM:: returnBPBlock()     hits: " << bit->getNumberHitRequests() << endl;
	//cout << "BM:: returnBPBlock() requests: " << bit->getTotalNumberBlockRequests() << endl;
	//cout << "BM:: returnBPBlock() hit rate: " << bit->getHitRate() << endl << endl;
	
	//cout << "BufferManager:: returnBPBlock() Point C .." << endl << endl;

	return returnDataPage;

	
} // end "returnBPBlock( string tableName, int bitIndex )" ..


DataPage* BufferManager:: returnFilePage( string tableName, int returnPageID ){
	
	//cout << "BufferManager:: returnFilePage() start.." << endl << endl;
	
	if( tableName.length() < 0 )
		return NULL;
		
	if( returnPageID < 0 )
		return NULL;
	
	// "pageId" was not found in the BP ( "Block Miss" ) ..
	// load page from file, Flush the BP block if necessary first ..
	
	// helper method(s):
	// 1) findBlockForLoad();
	
	// DEBUGGING ..
/*
	list<int>::iterator it_lru;
	for( it_lru = bit.LRUBlocks.begin(); it_lru != bit.LRUBlocks.end(); it_lru++ ){
		// get the index of the next LRU BP block ..
		cout << "BM:: returnFilePage() ( *it_lru ): " << ( *it_lru ) << endl << endl;
	}
*/

	bool blockReturned = true;
	
	DataPage* returnDataPage = new DataPage();
	
	int maxRowCount = 0;
	
	int intSize = 4;
	
	//cout << "BufferManager:: returnFilePage() Point A.." << endl << endl;
	
		
	getCatalogManager()->readMetaData();
	
	CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );
	
	
	// 1) Determine if the bp block has an empty space (free) in which to load,
	//    or need to flush a bp block prior to file (flush) prior to loading in a file page ..
	int freeBlockIndex = findBlockForLoad();
	
	FreeAndFlushBlockIndexAggregation* fafbip = getFreeAndFlushBlockIndexProcessing( freeBlockIndex );

	if( fafbip == NULL )
		return NULL;
		
	//cout << "BufferManager:: returnFilePage() Point B.." << endl << endl;

	bool insertPerformed = fafbip->getInsertPerformed();
	
	bool freeBlockFound = fafbip->getFreeBlockFound();
	
	bool flushBlockFound = fafbip->getFlushBlockFound();
	int flushBlockIndex = fafbip->getFlushBlockIndex();
	
	//cout << "BufferManager:: returnFilePage() insertPerformed: " << insertPerformed << endl << endl;
	
	//cout << "BufferManager:: returnFilePage() flushBlockFound: " << flushBlockFound << endl << endl;
	
	//cout << "BufferManager:: returnFilePage() flushBlockIndex: " << flushBlockIndex << endl << endl;
	
	if( ( flushBlockIndex < 0 ) && ( flushBlockFound ) )
		return NULL;
	
	//cout << "BufferManager:: returnFilePage() Point C.." << endl << endl;


	// 2) flush the bp block and load a page-from-file into the free block space ..
	//    ( found an available block to flush ( pin_count=0, LRU ) ) ..
	if ( insertPerformed && ( freeBlockFound || flushBlockFound ) ) {
			
		int bitIndex = BufferInfoTable::bitIndex;
		
		if( bitIndex < 0 )
			return NULL;
		
		//cout << "BufferManager:: returnFilePage() Point C1.." << endl << endl;
		
		//cout << "BufferManager:: returnFilePage() Point A.." << endl << endl;
		//cout << "BufferManager:: returnFilePage() Point A.." << endl << endl;
		//cout << "BufferManager:: returnFilePage() Point A.." << endl << endl;
		
		// (a) flushing a dirty bp block ..
		if( flushBlockFound && ( bit->getDirty()->at( flushBlockIndex ) == true ) ){
				
			// (i) need HP info on page-from-fle to be flushed to ..
			// NOTE: this could move into "flushBPBlock" ..
			DataHeader* header = new DataHeader();
			bool readHeaderPerformed = sm->readHeader( bit->getTableName()->at( flushBlockIndex ), header );
			
			// if "readHeader()" fails, return an empty DataPage ..
			if ( readHeaderPerformed == false ) {
				
	 			string errorString = "\nFailed to read Header [ BM::returnFilePage() ], data results may be less than expected.";
	 	
	 			// update the "MessageManager" to track this error and output to user ..
				getMessageManager()->setErrorStringQM( errorString );
		
			}
				
			//cout << "BufferManager:: returnFilePage() Point C2.." << endl << endl;
				

			// (ii) perform flush on bp block ..
			bool flushBPBlockPerformed = flushBPBlock( flushBlockIndex, header );
			
			// if "flushBPBlock()" fails, return an empty DataPage ..
			if ( flushBPBlockPerformed == false ) {
			
	 			string errorString = "\nFailed to flush buffer pool block [ returnFilePage() ], data results may be less than expected.";
	 	
	 			// update the "MessageManager" to track this error and output to user ..
				getMessageManager()->setErrorStringQM( errorString );

			}
				
				
		}
		
		// (b) set the pageID in "bit" for use in "loadBlock()" ..
		bit->getPageID()->at( bitIndex ) = returnPageID;
		
		bp->getDataPages()->at( bitIndex )->setPageID( returnPageID );

		//cout << "BufferManager:: returnFilePage() Point D.." << endl << endl;
		
		//cout << "BufferManager:: returnFilePage() tableName: " << tableName << endl << endl;

		// (c) flush this block by READING a block from file into it ..
		bool loadBlockPerformed = loadBlock( tableName, bp->getDataPages()->at( bitIndex ), locks, true );
		

		// if "loadBlock()" fails, return an empty DataPage ..
		if ( loadBlockPerformed == false ) {
			
	 		string errorString = "\nFailed to load page from file [ returnFilePage() ], data results may be less than expected.";
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );

		}

		returnDataPage = bp->getDataPages()->at( bitIndex );
		
		if( returnDataPage == NULL )
			return NULL;
		
		//cout << "BufferManager:: returnFilePage() Point E.." << endl << endl;
		
		// DEBUGGING ..
		//returnDataPage->printDataPage();
		
		// (d) get updated DataHeader for the page-from-file to be loaded in ..
		DataHeader* header = new DataHeader();
		bool readHeaderPerformed = getStorageManager()->readHeader( tableName, header );
		
		// if "readHeader()" fails, return an empty DataPage ..
		if ( readHeaderPerformed == false ) {
			
	 		string errorString = "\nFailed to open Header [ returnFilePage() ], data results may be less than expected.";
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );

		}
			
		// (e) get "DataHeaderEntry" for this page-from-file ..
		int fieldCount = catalogTable->getNumberOfAttributes();
		maxRowCount = header->computeMaxRowCount( fieldCount );
		DataHeaderEntry* hpEntry = new DataHeaderEntry( maxRowCount );
		bool hpEntryFound = false;
		for(int i=0; i<header->getTotalNumberOfDataPages(); i++){
					
			if ( returnPageID == header->getHeaderEntries()->at( i )->getPageID() ) {
						
				hpEntry = header->getHeaderEntries()->at( i );
				hpEntryFound = true;
				break;
						
			}
					
		}
			

		// (f) update "bit" with loaded file values ..
		if( hpEntryFound ){
					
			// hpEntryFound == true, so get its "filledRowsCount" ..
			bit->getBufferPageRowCount()->at( bitIndex ) = hpEntry->getFilledRowsCount();
					
		} else {
					
			// hpEntryFound == false, so set equal to "0" ..
			bit->getBufferPageRowCount()->at( bitIndex ) = 0;
					
		}
			
		//cout << "BufferManager:: returnFilePage() Point F.." << endl << endl;

		bit->getTableName()->at( bitIndex ) = tableName;
		
		bit->getBufferPageFieldCount()->at( bitIndex ) = fieldCount;
		
		bit->getRecordSize()->at( bitIndex ) = ( intSize * fieldCount );
		
		//bit.getBufferPageFieldCount()->at( bitIndex ) = header.getFilePageFieldCount();
		//bit.getRecordSize()->at( bitIndex ) = ( intSize * header.getFilePageFieldCount() );
				
		// need "bit.bufferPageRowCount" for "findBlockForInsert()" above ..
		//bit.bufferPageRowCount[ flushBlockIndex ] = hp.filledRowsCount;
			
		bit->getPageID()->at( bitIndex ) = returnPageID;	// loaded block with "returnPageID" into this block in the BP ..
		//bit.pin_count[ bitIndex ]++;						// default for select operation ..
		bit->getDirty()->at( bitIndex ) = false;			// dont mark for "write-to-file" upon "commit" ..
		bit->getEmpty()->at( bitIndex ) = false;			// block is not empty (just loaded a page into it) ..
		//bit.blockToBeReleased[ bitIndex ] = true;			// QM tracks this in order to decrement pin_count once it releases the block ..
	
		// ( Although "bit.LRUBlocks" is used to figure 
		// out the order for printing of the BP,
		// a copy is made prior to reaching this point 
		// see QM:: selectStarFromX() )
		// MUST update "bit.LRUBlocks" here 
		// ( unlike BM:: returnBPBlock() )
		// because the BP changes due to an insert
		// of a page-from-file into the BP during
		// this method ..
		bit->getLRUBlocks()->remove( bitIndex );		// remove block position from list ( where ever it is )..
		bit->getLRUBlocks()->push_back( bitIndex );		// set block position at end of list ( its MRU ) ..
		
		//bit->setTotalNumberBlockRequests( bit->getTotalNumberBlockRequests() + 1 );			// this is a "block request" ..
		bit->setNumberRequestsFromFile( bit->getNumberRequestsFromFile() + 1 );				// this is a "miss" ..
		bit->updateHitRate();						// hit rate changes (this is a miss) ..
		bit->updateTableCountUnique();				// count of unique "tableName" associated with blocks in the BP ..
		
		//cout << "BM:: returnFilePage()     hits: " << bit->getNumberHitRequests() << endl;
		//cout << "BM:: returnFilePage() requests: " << bit->getTotalNumberBlockRequests() << endl;
		//cout << "BM:: returnFilePage() hit rate: " << bit->getHitRate() << endl << endl;
		
		//cout << "BufferManager:: returnFilePage() Point G.." << endl << endl;
			
	// 3) abort operation; no available BP block ..
	} else {
			
		blockReturned = false;
			
	} // end "if ( freeBlockFound || flushBlockFound )" ..

	
	return returnDataPage;

	
} // end "returnFilePage( string tableName, int returnPageID )" ..


void BufferManager:: reactivateTablePageLock( int lockNumberIncrement ) {
	
	// NOTE: NOT using this method anymore, since I am not implementing
	// concurrent, multithreaded lock mechanism ..
	
	int pageID = -1;
	
	DataHeader* header = new DataHeader();
	
	time_t currentTime = time( NULL );
	
	time_t timeToWaitBeforeReactivating = 300; 	// 300 sec = 5 min
	
	time_t reactivateTime = currentTime + timeToWaitBeforeReactivating;
	
	while( lockReactivationOn ) {
		
		// infinite loop, handling reactivation of active locks every X minutes ..
		// ends when "BufferManager::writeBlock()" sets "lockReactivationOn = false" ..
		
		if ( currentTime == reactivateTime ){
			
			// the time has arrived to reactivate all current active locks ..
			
			int lockNumber = 0;
			
			for( int i=0; i < locks->size(); i++ ){
				
				pageID = (*locks)[ bit->getTableName()->at( i ) ]->at( i );
				
				getStorageManager()->readHeader( bit->getTableName()->at( pageID ), header );
				
				// lock is "active" only when lockNumber is _incremented_ every X minutes,
				// ( lockNumber > 0 = active lock )
				// and can only be removed by the initiator ..
				// when it becomes "inactive", the lock can be cleared by anyone ..
				lockNumber = header->getHeaderEntries()->at( pageID )->getLockNumber();
				lockNumber += lockNumberIncrement;
				header->getHeaderEntries()->at( pageID )->setLockNumber( lockNumber );

			}
	
			// save the new DataHeader info to file ..
			getStorageManager()->writeHeader( bit->getTableName()->at( pageID ), header );
			
			// calculate a new "activateTime" ..
			currentTime = time( NULL );
			reactivateTime = currentTime + timeToWaitBeforeReactivating;
			
			
			// loops, waits for X minutes until "currentTime == reactivateTime" above ..
			
			
		} // end "if ( currentTime == activateTime )" ..
		
		
	} // end "while( true )" ..

	
} // end "reactivateTablePageLocks( char* tableName )" ..


/*
 * GETTERS ..
 */

map< string, vector<int>* >* BufferManager:: getLocks(){
	
	return locks;
	
}

	
BufferPool* BufferManager:: getBufferPool(){
	
	return bp;
	
}

	
BufferInfoTable* BufferManager:: getBufferInfoTable(){
	
	return bit;
	
}

	
BITSnapShot* BufferManager:: getBitSnapShot(){
	
	return bitSnapShot;
	
}

vector<DataPage*>* BufferManager:: getLastFlushedBlocks(){
	
	return lastFlushedBlocks;
	
}

vector<int>* BufferManager:: getLastFlushedBlockPageIDs(){
	
	return lastFlushedBlockPageIDs;
	
}

	
/*
 * SETTERS ..
 */
 
void BufferManager:: setLocks( map< string, vector<int>* >* newLocks ){
	
	locks = newLocks;
	
}

	
void BufferManager:: setBufferPool( BufferPool* newBufferPool ){
	
	bp = newBufferPool;
	
}

	
void BufferManager:: setBufferInfoTable( BufferInfoTable* newBit ){
	
	bit = newBit;
	
}

	
void BufferManager:: setBitSnapShot( BITSnapShot* newBITSnapShot ){
	
	bitSnapShot = newBITSnapShot;
	
}


void BufferManager:: setLastFlushedBlocks( vector<DataPage*>* newLastFlushedBlocks ){
	
	lastFlushedBlocks = newLastFlushedBlocks;
	
}

void BufferManager:: setLastFlushedBlockPageIDs( vector<int>* newLastFlushedBlockPageIDs ){
	
	lastFlushedBlockPageIDs = newLastFlushedBlockPageIDs;
	
}
