#include "CatalogManager.h"


CatalogManager:: CatalogManager( MessageManager* mm_, CatalogFile* catalogFile_ ) {
	
	//cout << "CatalogManager:: CatalogManager() start .." << endl << endl;

	// default constructor ..
	
	mm = mm_;
	
	catalog = new Catalog();
	
	//catalogFile = new CatalogFile( "C:\\", "Temp\\", "catalog", ".bin" );
	catalogFile = catalogFile_;

}


CatalogManager:: ~CatalogManager() {
	
	//cout << "~CatalogManager() start .." << endl << endl;
	
	// cleanup ..
	
	delete mm;
	
	delete catalog;
	
	delete catalogFile;
	
	//cout << "~CatalogManager() end .." << endl << endl;
	
}

void CatalogManager:: setCatalogFile( CatalogFile* catalogFile_ ){
	
	catalogFile = catalogFile_;
	
}


void CatalogManager:: printCatalogData( vector<string>* tableData ){
	
	if( tableData == NULL )
		return;
	
	// helper method for "printCatalog()" ..
	
	// instantiate the "tableFormatter" to print a table ..
	TableFormatter tableFormatter;
	
	// define the default table values for if no data in the cell ..
	//tableFormatter.addDefaultValue( 0, "n/a" );
	tableFormatter.setDefaultFieldValue( "-" );
	
	// character used in the divider separating "header" from "data" ..
	tableFormatter.setDivider( '=' );
	
	// middle, left, right divider-column spacers ..
	tableFormatter.setDividerColumnSpacer( "||" );
	tableFormatter.setDividerColumnSpacerLeft( "=" );
	tableFormatter.setDividerColumnSpacerRight( "=" );
	
	// middle, left, right data-column spacers ..
	tableFormatter.setDataColumnSpacer( "  " );
	tableFormatter.setDataColumnSpacerLeft( " " );
	tableFormatter.setDataColumnSpacerRight( " " );
	
	// middle, left, right header-column spacers ..
	tableFormatter.setHeaderColumnSpacer( "  " );
	tableFormatter.setHeaderColumnSpacerLeft( " " );
	tableFormatter.setHeaderColumnSpacerRight( " " );
	
	// define the table "header" column title names ..
	tableFormatter.addColumnHeader( "Table" );
	tableFormatter.addColumnHeader( "Size" );
	tableFormatter.addColumnHeader( "# of Att." );
	
	// define the default table values for if no data in the cell ..
	tableFormatter.setDefaultFieldValue( "empty" );
	
	// "Attribute Name" column title should be for all columns in "tableData"
	// except for the 1st three columns of the table ..
	int numberOfAttributeNames = tableData->size() - 3;
	
	for(int i=0; i<numberOfAttributeNames; i++){
		
		tableFormatter.addColumnHeader( "Attr" );
		
	}
	
	int columnCount = 0;
	
	int tableDataSize = -1;
	
	if( tableData != NULL )
		tableDataSize = tableData->size();
	
	for(int i=0; i<tableDataSize; i++){
		
		// store the "DATA" fields ..
		tableFormatter.addColumnData( columnCount, ( *tableData )[i] );
		columnCount++;
		
	}
	
	// set parameters for next row of data ..
	tableFormatter.endDataRow();
		
	// format the data input in preparation for printing ..
	tableFormatter.formatTable();
	
	// print the formatted table ..
	tableFormatter.printTable();
	
	
} // end "printCatalogData( vector<string> tableData )" ..


void CatalogManager:: printCatalog() {
	
	
	readMetaData();
	
	Catalog* catalog = getCatalog();
	 
	int numberOfTablesInCatalog = getNumberOfTables();
	 
	cout << "# of Tables in Catalog: " << numberOfTablesInCatalog << endl << endl;
	
	//vector<CatalogTable*>::iterator it_ct;
	
	CatalogTable* catalogTable = NULL;
	 
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	//vector<CatalogTableAttribute*>::iterator it_cta;
	
	CatalogTableAttribute* catalogTableAttribute = NULL;
	 
	vector<CatalogTableAttribute*>* catalogTableAttributes;
	
	vector<string>* tableData = NULL;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	int catalogTableAttributesSize = -1;
	
	//for( it_ct = catalogTables->begin(); it_ct != catalogTables->end(); it_ct++ ){
	for(int i=0; i<catalogTablesSize; i++){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable != NULL ){
			
			tableData = new vector<string>();
			
			//tableData->push_back( (*it_ct)->getTableName() );
			tableData->push_back( catalogTable->getTableName() );
		
			//tableData->push_back( intToStr( (*it_ct)->getTableSize() ) );
			tableData->push_back( intToStr( catalogTable->getTableSize() ) );
		
			//tableData->push_back( intToStr( (*it_ct)->getNumberOfAttributes() ) );
			tableData->push_back( intToStr( catalogTable->getNumberOfAttributes() ) );
		
			//catalogTableAttributes = (*it_ct)->getAttributes();
			catalogTableAttributes = catalogTable->getAttributes();
	
			if( catalogTableAttributes != NULL )
				catalogTableAttributesSize = catalogTableAttributes->size();
		
			//for( it_cta = catalogTableAttributes->begin(); it_cta != catalogTableAttributes->end(); it_cta++ ){
			for(int j=0; j<catalogTableAttributesSize; j++){
			
				catalogTableAttribute = catalogTableAttributes->at(j);
			
				//tableData->push_back( (*it_cta)->getTableAttributeName() );
				if( catalogTableAttribute != NULL )
					tableData->push_back( catalogTableAttribute->getTableAttributeName() );
			
			}
		
			// print this table to standard output ..
			printCatalogData( tableData );
		
			// cleanup ..
			delete tableData;
		
			cout << endl;
			
		}

	}
	 
	
} // end "printCatalog()" ..


void CatalogManager:: oldPrintCatalog() {
	
	
	// DEBUGGING ..
	// cout << "# of Tables in Catalog: " << numberOfTablesInCatalog << endl << endl;
	
	readMetaData();
	
	Catalog* catalog = getCatalog();
	 
	string fileTimestamp = getFileTimestampString();
	 
	int numberOfTablesInCatalog = getNumberOfTables();
	 
	// DEBUGGING ..
	// cout << "# of Tables in Catalog: " << numberOfTablesInCatalog << endl << endl;
	 
	//vector<CatalogTable>::iterator it_ct;

	string tableName = "";
	 
	int tableSize = -1;
	 
	int numberOfAttributes = -1;
	 
	//vector<CatalogTableAttribute*>::iterator it_cta;
	
	CatalogTable* catalogTable = NULL;
	 
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	CatalogTableAttribute* catalogTableAttribute = NULL;
	 
	vector<CatalogTableAttribute*>* catalogTableAttributes;
	
	int catalogTableAttributesSize = -1;
	
	 
	string catalogTableAttributeName = "";
	 
	 
	 // find the maximum length of any row to print ..
	 
	 map<string, int> maxStringLengths;
	 
	 int maxStringLength = -1;
	 
	 map<string, int> maxAttributeLengths;
	 
	 int maxAttributeLength = -1;
	 
	 //for( it_ct = catalogTables->begin(); it_ct != catalogTables->end(); it_ct++ ){
	 for(int i=0; i<catalogTablesSize; i++){
	 	
	 	catalogTable = catalogTables->at(i);
	 	
	 	//tableName = (*it_ct)->getTableName();
	 	tableName = catalogTable->getTableName();
	 	
	 	updateMaxStringLength( tableName, maxStringLength );
	 	
	 	//tableSize = (*it_ct)->getTableSize();
	 	tableSize = catalogTable->getTableSize();
	 	
	 	updateMaxStringLength( intToStr( tableSize ), maxStringLength );
	 	
	 	//numberOfAttributes = (*it_ct)->getNumberOfAttributes();
	 	numberOfAttributes = catalogTable->getNumberOfAttributes();
	 	
	 	updateMaxStringLength( intToStr( numberOfAttributes ), maxStringLength );
	 	
	 	//catalogTableAttributes = (*it_ct)->getAttributes();
	 	catalogTableAttributes = catalogTable->getAttributes();

		if( catalogTableAttributes != NULL )
			catalogTableAttributesSize = catalogTableAttributes->size();
	 	
	 	//for( it_cta = catalogTableAttributes->begin(); it_cta != catalogTableAttributes->end(); it_cta++ ){
	 	for(int j=0; j<catalogTableAttributesSize; j++){
	 		
	 		catalogTableAttribute = catalogTableAttributes->at(i);
	 		
	 		//catalogTableAttributeName = (*it_cta)->getTableAttributeName();
	 		catalogTableAttributeName = catalogTableAttribute->getTableAttributeName();
	 		
	 		updateMaxStringLength( catalogTableAttributeName, maxStringLength );
	 		
	 		updateMaxStringLength( catalogTableAttributeName, maxAttributeLength );
	 		
	 	}
	 	
	 	// store the max values for each table ..
	 	maxStringLengths.insert( pair<string, int>( tableName, maxStringLength ) );
	 	
	 	maxAttributeLengths.insert( pair<string, int>( tableName, maxAttributeLength ) );
	 	
	 	// reset these parameters for the next table / iteration ..
	 	maxStringLength = -1;
	 	
	 	maxAttributeLength = -1;
	 	
	 }
	 
	 // DEBUGGING ..
	 //cout << "maxStringLength: " << maxStringLength << endl << endl;
	 
	int maxLengthPromptStrings = 14;
	 
	int tableCounter = 0;
	 
	int paddingSize;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	 
	 //for( it_ct = catalogTables->begin(); it_ct != catalogTables->end(); it_ct++ ){
	 for(int i=0; i<catalogTablesSize; i++){
	 	
	 	catalogTable = catalogTables->at(i);
	 	
	 	if ( tableCounter > 0 ) {
	 		
	 		cout << endl << endl;
	 		
	 	}
	 	
	 	for(int i=0; i<( ( maxStringLengths.find( ( (char*)tableName.c_str() ) )->second ) + maxLengthPromptStrings + 1); i++){
	 		
	 		cout << "=";
	 		
	 	}
	 	
	 	cout << endl;
	 	
	 	//tableName = (*it_ct)->getTableName();
	 	tableName = catalogTable->getTableName();
	 	
	 	cout << "       Table: " << tableName << endl;
	 	
	 	for(int j=0; j<( ( maxStringLengths.find( ( (char*)tableName.c_str() ) )->second ) + maxLengthPromptStrings + 1); j++){
	 		
	 		cout << "=";
	 		
	 	}
	 	
	 	cout << endl;
	 	
	 	//tableSize = (*it_ct)->getTableSize();
	 	tableSize = catalogTable->getTableSize();
	 	
	 	//numberOfAttributes = (*it_ct).getNumberOfAttributes();
	 	numberOfAttributes = catalogTable->getNumberOfAttributes();
	 	
	 	if ( ( intToStr( tableSize ) ).size() > ( intToStr( numberOfAttributes ) ).size() ) {
	 		
	 		paddingSize = ( ( intToStr( tableSize ) ).size() - ( intToStr( numberOfAttributes ) ).size() );
	 		
	 		cout << "        Size: " << tableSize << endl;
	 		
	 		cout << " # of Fields: ";
	 		
	 		for(int j=0; j<paddingSize; j++){
	 		
	 			cout << " ";
	 		
	 		}
	 		
	 		cout << numberOfAttributes << endl;
	 		
	 	} else {
	 		
	 		paddingSize = ( ( intToStr( numberOfAttributes ) ).size() - ( intToStr( tableSize ) ).size() );
	 		
	 		cout << "        Size: ";
	 	
	 		for(int j=0; j<paddingSize; j++){
	 		
	 			cout << " ";
	 		
	 		}
	 		
	 		cout << tableSize << endl;
	 		
	 		cout << " # of Fields: " << numberOfAttributes << endl;
	 		
	 	}
	 	
	 	//catalogTableAttributes = (*it_ct)->getAttributes();
	 	catalogTableAttributes = catalogTable->getAttributes();
	 	
	 	if( catalogTableAttributes != NULL )
			catalogTableAttributesSize = catalogTableAttributes->size();
	 	
	 	//for( it_cta = catalogTableAttributes->begin(); it_cta != catalogTableAttributes->end(); it_cta++ ){
	 	for(int j=0; j<catalogTableAttributesSize; j++){
	 		
	 		for(int k=0; k<( ( maxStringLengths.find( ( (char*)tableName.c_str() ) )->second ) + maxLengthPromptStrings + 1); k++){
	 		
	 			cout << "-";
	 		
	 		}
	 		
	 		cout << endl;
	 		
	 		catalogTableAttribute = catalogTableAttributes->at(i);
	 		
	 		//catalogTableAttributeName = (*it_cta)->getTableAttributeName();
	 		catalogTableAttributeName = catalogTableAttribute->getTableAttributeName();
	 		
	 		cout << "  Field Name: " << catalogTableAttributeName << endl;
	 		
	 	}
	 	
	 	for(int k=0; k<( ( maxStringLengths.find( ( (char*)tableName.c_str() ) )->second ) + maxLengthPromptStrings + 1); k++){
	 		
	 		cout << "=";
	 		
	 	}
	 	
	 	cout << endl;
	 	
	 	tableCounter++;
	 	
	 }
	 
	
} // end "oldPrintCatalog()" ..


void CatalogManager:: resetCatalog() {
	
	catalog->resetCatalog();
	
	// cleanup and reassign ..
	//delete catalog;
	//catalog = new Catalog();

	//delete catalogFile;
	//catalogFile = new CatalogFile();
	
} // end "resetCatalog()" ..


string CatalogManager:: addCatalogTableToCatalog( CatalogTable* catalogTable ){
	
	if( catalogTable == NULL )
		return "\"catalogTable\" was NULL, \"catalogTable\" was not added to the catalog.";
	
	// wrapper for "Catalog" ..
	string returnString = catalog->addCatalogTable( catalogTable );
	
	return returnString;
	
} // end "addCatalogTable( CatalogTable )" ..


string CatalogManager:: addCatalogIndexToCatalog( CatalogIndex* catalogIndex ){
	
	if( catalogIndex == NULL )
		return "\"catalogIndex\" was NULL, \"catalogIndex\" was not added to the catalog.";
	
	// wrapper for "Catalog" ..
	string returnString = catalog->addCatalogIndex( catalogIndex );
	
	return returnString;
	
} // end "addCatalogIndex( CatalogIndex )" ..


string CatalogManager:: removeCatalogTableFromCatalog( string tableName ){
	
	if( tableName.length() <= 0 )
		return "\"tableName\" was empty, table was not removed from the catalog.";
	
	// wrapper for "Catalog" ..
	string returnString = catalog->removeCatalogTable( tableName );
	
	return returnString;
	
} // end "RemoveCatalogTable( string )" ..


string CatalogManager:: removeCatalogIndexFromCatalog( string tableName ){
	
	if( tableName.length() <= 0 )
		return "\"tableName\" was empty, index was not removed from the catalog.";
	
	// wrapper for "Catalog" ..
	string returnString = catalog->removeCatalogIndex( tableName );
	
	return returnString;
	
} // end "RemoveCatalogIndex( string )" ..


void CatalogManager:: writeTimestamp(){
	
	//cout << "CatalogManager:: writeTimestamp() start .." << endl;
		
	// get the current system time ..
	time_t currentTime;
	
	time( &currentTime );
	
	char* outputFileName = (char *) catalogFile->getFullyQualifiedFileName().c_str();
	

	ofstream outputFile( outputFileName, ios::binary | ios::out );
	
		
	//if ( outputFile.is_open() ){
	if ( !outputFile.fail() ){
		
		outputFile.seekp( 0 );
		
		outputFile.write( (char*)&currentTime, sizeof( time_t ) );
	
		catalog->setManagerCatalogTimestamp( currentTime );
	
		catalog->setFileCatalogTimestamp( currentTime );
		
		outputFile.close();

	}
	
} // end "writeTimestamp()" ..


void CatalogManager:: writeTimestamp( ofstream *outputFile ){
	
	if( outputFile == NULL )
		return;
	
	//cout << "CatalogManager:: writeTimestamp() start .." << endl;
		
	// get the current system time ..
	time_t currentTime;
	
	time( &currentTime );

	
	//if ( outputFile->is_open() ) {
	if ( !outputFile->fail() ){
		
		outputFile->seekp( 0 );
	
		outputFile->write( (char*)&currentTime, sizeof( time_t ) );

		catalog->setManagerCatalogTimestamp( currentTime );
	
		catalog->setFileCatalogTimestamp( currentTime );

	}
	
} // end "writeTimestamp()" ..


void CatalogManager:: readTimestamp(){
	
	//cout << "CatalogManager:: readTimestamp() start .." << endl;
	
	char* inputFileName = (char *) catalogFile->getFullyQualifiedFileName().c_str();
	
	ifstream inputFile( inputFileName, ios::binary | ios::in );
	
		
	//if ( inputFile.is_open() ){
	if ( !inputFile.fail() ){
		
		inputFile.seekg( 0, ios::end );
		
		int fileSize = inputFile.tellg();
		
		inputFile.seekg( 0 );
		
		
		if ( fileSize >= sizeof( time_t ) ) {
			
			time_t timestamp;
			inputFile.read( (char*)&timestamp, sizeof( time_t ) );	
		
			catalog->setFileCatalogTimestamp( timestamp );
		
		} // end "if ( fileSize >= sizeof( time_t ) )" ..
		
		
		// cleanup ..
		inputFile.close();

	}
	
} // end "readTimestamp()" ..

/*
 * METADATA ..
 */

bool CatalogManager:: writeMetaData(){
	
	// DEBUGGING ..
	//cout << "CatalogManager:: writeMetaData() start .." << endl << endl;
	//cout << "CatalogManager:: writeMetaData() catalogFile->getFullyQualifiedFileName(): " << catalogFile->getFullyQualifiedFileName() << endl << endl;
	
	char* outputFileName = (char*) catalogFile->getFullyQualifiedFileName().c_str();
	
	ofstream outputFile( outputFileName, ios::binary | ios::out );
	
	string metaData;
	
	bool writeMetaDataPerformed = false;
	
	//if ( outputFile.is_open() ){
	if ( !outputFile.fail() ){
	
		/*
		 * unwrap the "Catalog" object into its constituent primitive types,
		 * which are the types that can be written to file ..
		 * 
		 * <%numberOfTables%> numberOfTables1
		 * <%numberOfIndices%> numberOfIndices1
		 * <%catalogTableList%>
		 * 		<%catalogTable%>
		 * 			<%tableName%> tableName1
		 * 			<%tableSize%> tableSize1
		 * 			<%numberOfAttributes%> numberOfAttributes1
		 * 			<%tableAttributeNameList%>
		 * 				<%tableAttributeName%> tableAttributeName1
		 * 				<%tableAttributeName%> tableAttributeName2
		 * 				<%tableAttributeName%> tableAttributeName3
		 * ...
		 * 				<%tableAttributeName%> tableAttributeName8
		 * 			<%/tableAttributeNameList%>
		 * 		<%/catalogTable%>
		 * ...
		 * 		<%catalogTable%>
		 * 			<%tableName%> tableName2
		 * ...
		 * 		<%/catalogTable%>
		 * <%/catalogTableList%>
		 * <%catalogIndexList%>
		 * 		<%catalogIndex%>
		 * 			<%indexName%> indexName1
		 * 			<%indexTableName%> indexTableName1
		 * 			<%indexSize%> indexSize1
		 * 			<%isNoDuplicates%> isNoDuplicates1
		 * 			<%indexAttributeName%> indexAttributeName1
		 * 		<%/catalogIndex%>
		 * ...
		 * 		<%catalogIndex%>
		 * 			<%indexName%> indexName2
		 * 			<%indexTableName%> indexTableName1
		 * 			<%indexSize%> indexSize2
		 * 			<%isNoDuplicates%> isNoDuplicates2
		 * 			<%indexAttributeName%> indexAttributeName2
		 * 		<%/catalogIndex%>
		 * ...
		 * <%/catalogIndexList%>
		 * <%end%>
		 * 
		 */
		
		// the file was opened in "wb" mode, so must update the "timestamp" at this point ..
		writeTimestamp( &outputFile );


		// "<%numberOfTables%>" top-level delimiter ..
		metaData += "<%numberOfTables%>";
		
		// "numberOfTables" ..
		metaData += intToStr( catalog->getNumberOfTables() );
		
		// "<%numberOfIndices%>" top-level delimiter ..
		metaData += "<%numberOfIndices%>";

		// "numberOfIndices" ..
		metaData += intToStr( catalog->getNumberOfIndices() );

		// "catalogTables" ..
		//vector<CatalogTable>::iterator it_ct;
		vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
		
		int catalogTablesSize = -1;
	
		if( catalogTables != NULL )
			catalogTablesSize = catalogTables->size();
		
		// only need a start-list-tag if list is non-zero ..
		if( catalogTablesSize > 0 ){
			
			// "<%catalogTableList%>" top-level delimiter ..
			metaData += "<%catalogTableList%>";
			
		}
		
		// "catalogTables >> catalogTableAttributes" ..
		//vector<CatalogTableAttribute>::iterator it_cta;
		vector<CatalogTableAttribute*>* catalogTableAttributes = NULL;
		string tableAttributeName;
		
		CatalogTable* catalogTable = NULL;
		
		int catalogTableAttributesSize = -1;
		
		CatalogTableAttribute* catalogTableAttribute = NULL;

		// "catalogTables" and "catalogTables >> catalogTableAttributes" ..
		//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
		for(int i=0; i<catalogTablesSize; i++){
			
			catalogTable = catalogTables->at(i);
			
			// "<%catalogTable%>" list delimiter .. 
			metaData += "<%catalogTable%>";
			
			// "<%tableName%>" list delimiter ..
			metaData += "<%tableName%>";

			// "catalogTables >> tableName" ..
			//metaData += it_ct->getTableName();
			if( catalogTable != NULL )
				metaData += catalogTable->getTableName();

			// "<%tableSize%>" list delimiter ..
			metaData += "<%tableSize%>";

			// "catalogTables >> tableSize" ..
			//metaData += intToStr( it_ct->getTableSize() );
			if( catalogTable != NULL )
				metaData += intToStr( catalogTable->getTableSize() );

			// "<%numberOfAttributes%>" list delimiter ..
			metaData += "<%numberOfAttributes%>";

			// "catalogTables >> numberOfAttributes" ..
			//metaData += intToStr( it_ct->getNumberOfAttributes() );
			if( catalogTable != NULL )
				metaData += intToStr( catalogTable->getNumberOfAttributes() );

			// "<%tableAttributeNameList%>" list delimiter ..
			metaData += "<%tableAttributeNameList%>";

			// "catalogTables >> catalogTableAttributes" ..
			//catalogTableAttributes = it_ct->getAttributes();
			if( catalogTable != NULL )
				catalogTableAttributes = catalogTable->getAttributes();
	
			if( catalogTableAttributes != NULL )
				catalogTableAttributesSize = catalogTableAttributes->size();
			
			//for( it_cta = catalogTableAttributes.begin(); it_cta != catalogTableAttributes.end(); it_cta++ ){
			for(int j=0; j<catalogTableAttributesSize; j++){
				
				catalogTableAttribute = catalogTableAttributes->at(j);
				
				// "<%tableAttributeName%>" list delimiter (NESTED) ..
				metaData += "<%tableAttributeName%>";

				// "catalogTables >> catalogTableAttributes >> tableAttributeName" ..
				//metaData += it_cta->getTableAttributeName();
				if( catalogTableAttribute != NULL )
					metaData += catalogTableAttribute->getTableAttributeName();
				
			}

			// "<%/tableAttributeNameList%>" list delimiter END ..
			metaData += "<%/tableAttributeNameList%>";

			// "<%/catalogTable%>" list delimiter END.. 
			metaData += "<%/catalogTable%>";

		}

		// only need a end-list-tag if list is non-zero ..
		if( catalogTablesSize > 0 ){
			
			// "<%/catalogTableList%>" top-level delimiter END ..
			metaData += "<%/catalogTableList%>";
			
		}
		
		// "catalogIndices" ..
		//vector<CatalogIndex>::iterator it_ci;
		vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();

		CatalogIndex* catalogIndex = NULL;

		int catalogIndicesSize = -1;
	
		if( catalogIndices != NULL )
			catalogIndicesSize = catalogIndices->size();

		// only need a start-list-tag if list is non-zero ..
		if( catalogIndicesSize > 0 ){
			
			// "<%catalogIndexList%>" top-level delimiter ..
			metaData += "<%catalogIndexList%>";
			
		}
		

		// "catalogIndices" and "catalogIndices >> catalogIndexAttributes" ..
		//for( it_ci = catalogIndices->begin(); it_ci != catalogIndices->end(); it_ci++ ){
		for( int i=0; i < catalogIndicesSize; i++ ){
			
			catalogIndex = catalogIndices->at(i);
			
			// "<%catalogIndex%>" list delimiter .. 
			metaData += "<%catalogIndex%>";
			
			
			// "<%indexName%>" list delimiter ..
			metaData += "<%indexName%>";

			// "catalogIndices >> indexName" ..
			//metaData += it_ci->getIndexName();
			//metaData += catalogIndices->at(i)->getIndexName();
			if( catalogIndex != NULL )
				metaData += catalogIndex->getIndexName();
			
			// "<%indexTableName%>" list delimiter ..
			metaData += "<%indexTableName%>";

			// "catalogIndices >> tableName" ..
			//metaData += it_ci->getIndexTableName();
			//metaData += catalogIndices->at(i)->getIndexTableName();
			if( catalogIndex != NULL )
				metaData += catalogIndex->getIndexTableName();

			// "<%indexSize%>" list delimiter ..
			metaData += "<%indexSize%>";

			// "catalogIndices >> indexSize1" ..
			//metaData += intToStr( it_ci->getIndexSize() );
			//metaData += intToStr( catalogIndices->at(i)->getIndexSize() );
			if( catalogIndex != NULL )
				metaData += intToStr( catalogIndex->getIndexSize() );

			// "<%isNoDuplicates%>" list delimiter ..
			metaData += "<%isNoDuplicates%>";

			// "catalogIndices >> isDuplicate" ..
			//metaData += intToStr( it_ci->getIsNoDuplicates() );
			//metaData += intToStr( catalogIndices->at(i)->getIsNoDuplicates() );
			if( catalogIndex != NULL )
				metaData += intToStr( catalogIndex->getIsNoDuplicates() );

			// "<%indexAttributeName%>" list delimiter ..
			metaData += "<%indexAttributeName%>";
		
			// "catalogIndice >> catalogTableAttribute >> indexAttributeName" ..
			//metaData += it_ci->getAttribute().getIndexAttributeName();
			//metaData += catalogIndices->at(i)->getAttribute().getIndexAttributeName();
			if( catalogIndex != NULL )
				metaData += catalogIndex->getAttribute()->getIndexAttributeName();

			// "<%/catalogIndex%>" list delimiter END.. 
			metaData += "<%/catalogIndex%>";

		}


		// only need an end-list-tag if list is non-zero ..
		if( catalogIndicesSize > 0 ){
			
			// "<%/catalogIndexList%>" top-level delimiter END ..
			metaData += "<%/catalogIndexList%>";
			
		}
		
		// "<%end%>" top-level delimiter END ..
	 	metaData += "<%/end%>";
	 	
	 	
		// DEBUGGING ..
		//cout << "CatalogManager:: writeMetaData()     size of data to write to file:" << metaData.size() << endl;
		//cout << "CatalogManager:: writeMetaData() here is the data to write to file:" << endl << endl;
		//cout << "CatalogManager:: writeMetaData() metaData: " << metaData << endl << endl;
	
	
		// write the "metaData" string to file ..
		char* metaDataBuffer = (char*) metaData.data();
	
		// "ofstream" requires the use of char[], 
		// which is why a vector is not used ..
		outputFile.write( (char*)metaDataBuffer, metaData.size() );
	
		// cleanup ..
		outputFile.close();
		
		writeMetaDataPerformed = true;
		
		// DEBUGGING ..
		//cout << "CatalogManager:: writeMetaData(): " << metaData << endl << endl;
		
	}  else {
				
		string errorString = "\nFailed to open Catalog file [ CM::writeMetaData ], data may not have been saved to file.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
	 	getMessageManager()->setErrorStringCM( errorString );
				
	}

	return writeMetaDataPerformed;
	
} // end "writeMetaData()" ..


bool CatalogManager:: readMetaData(){
	
	// DEBUGGING ..
	//cout << "CatalogManager:: readMetaData() start .." << endl;
	//cout << "CatalogManager:: readMetaData() catalogFile->getFullyQualifiedFileName(): " << catalogFile->getFullyQualifiedFileName() << endl << endl;
	
	char* inputFileName = (char*) catalogFile->getFullyQualifiedFileName().c_str();
	
	ifstream inputFile( inputFileName, ios::binary | ios::in );
	
	string metaData;
	
	bool readMetaDataPerformed = false;
	
	//if ( inputFile.is_open() ) {
	if ( !inputFile.fail() ){
		
		/*
		 * wrap the constituent primitive-type-parts of "Catalog" that were written to file,
		 * back into a single object "Catalog" ..
		 * 
		 * <%numberOfTables%> numberOfTables1
		 * <%numberOfIndices%> numberOfIndices1
		 * <%catalogTableList%>
		 * 		<%catalogTable%>
		 * 			<%tableName%> tableName1
		 * 			<%tableSize%> tableSize1
		 * 			<%numberOfAttributes%> numberOfAttributes1
		 * 			<%tableAttributeNameList%>
		 * 				<%tableAttributeName%> tableAttributeName1
		 * 				<%tableAttributeName%> tableAttributeName2
		 * 				<%tableAttributeName%> tableAttributeName3
		 * ...
		 * 				<%tableAttributeName%> tableAttributeName8
		 * 			<%/tableAttributeNameList%>
		 * 		<%/catalogTable%>
		 * ...
		 * 		<%catalogTable%>
		 * 			<%tableName%> tableName1
		 * ...
		 * 		<%/catalogTable%>
		 * <%/catalogTableList%>
		 * <%catalogIndexList%>
		 * 		<%catalogIndex%>
		 * 			<%indexName%> indexName1
		 * 			<%indexTableName%> indexTableName1
		 * 			<%indexSize%> indexSize1
		 * 			<%isNoDuplicates%> isNoDuplicates1
		 * 			<%indexAttributeName%> indexAttributeName1
		 * 		<%/catalogIndex%>
		 * ...
		 * 		<%catalogIndex%>
		 * 			<%indexName%> indexName2
		 * 			<%indexTableName%> indexTableName1
		 * 			<%indexSize%> indexSize2
		 * 			<%isNoDuplicates%> isNoDuplicates2
		 * 			<%indexAttributeName%> indexAttributeName2
		 * 		<%/catalogIndex%>
		 * ...
		 * <%/catalogIndexList%>
		 * <%end%>
		 * 
		 */
		
		inputFile.seekg( 0 );
		
		inputFile.seekg( 0, ios::end );
		
		long fileSize = inputFile.tellg();
		
		// DEBUGGING ..
		//cout << "CatalogManager:: readMetaData() fileSize: " << fileSize << endl;
		
		// move the file indicator forward beyond the "timestamp" portion of the file ..
		size_t dataPosition = sizeof( time_t ); 
		
		inputFile.seekg( dataPosition );
		
		
		// initialize some useful quantities ..
		CatalogTable* catalogTable = new CatalogTable();
		
		vector<CatalogTableAttribute*>* catalogTableAttributes = NULL; //= new vector<CatalogTableAttribute*>();
		
		vector<CatalogTable*>* catalogTables = new vector<CatalogTable*>();
		
		
		CatalogIndex* catalogIndex = NULL; //new CatalogIndex();
		
		//vector<CatalogIndexAttribute*>* catalogIndexAttributes = new vector<CatalogIndexAttribute*>();
		
		vector<CatalogIndex*>* catalogIndices = new vector<CatalogIndex*>();
		
		 
		string delimiterFound = "";
		
		string endDelimiter = "<%end%>";
		
		//cout << "CatalogManager:: readMetaData() Point C4-1 .." << endl << endl;
		
		
		if ( fileSize > dataPosition ) {
			
			char* metaDataBuffer = new char[ fileSize ];
		
			// "ifstream" requires the use of char[], 
			// which is why a vector is not used ..
			inputFile.read( metaDataBuffer, fileSize );
		
			// cleanup ..
			inputFile.close();
		
			// DEBUGGING ..
			//cout << "size of the data read from file: " << ( fileSize - sizeof(time_t) ) << endl;
			//cout << "here is the data read from file:" << endl << endl;
			//cout << metaDataBuffer << endl << endl;
	
		
			string metaData = metaDataBuffer;
		
			const char* delimiters = ">%<";
		 
			vector<string>* metaDataTokens = getInputStringTokens( metaData, delimiters );
			
			// DEBUGGING ..
			//cout << "metaDataTokens.size(): " << metaDataTokens.size() << endl;	
			//for(int i=0; i<metaDataTokens.size(); i++){
			//	cout << "metaDataToken: " << metaDataTokens[i] << endl;	
			//}
			//cout << endl;
		
			int tokenCounter = 0;
		
			if( metaDataTokens != NULL ){
				
				do {
		
					if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "numberOfTables" ) == 0 ) {
			
						// "numberOfTables" ..
						int numberOfTables;
				
						tokenCounter++;
						string dataFound = metaDataTokens->at( tokenCounter );
						numberOfTables = strToInt( dataFound );
				
						catalog->setNumberOfTables( numberOfTables );
				
						// DEBUGGING ..
						//cout << " numberOfTables:" << numberOfTables << endl;
				
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "numberOfIndices" ) == 0 ) {
			
						// "numberOfIndices" ..
						int numberOfIndices;
				
						tokenCounter++;
						string dataFound = metaDataTokens->at( tokenCounter );
						numberOfIndices = strToInt( dataFound );
				
						catalog->setNumberOfIndices( numberOfIndices );
				
						// DEBUGGING ..
						//cout << " numberOfIndices:" << numberOfIndices << endl;
				
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "catalogTableList" ) == 0 ) {
				
						// DEBUGGING ..
						//cout << "	<%catalogTableList%> found .." << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "catalogTable" ) == 0 ) {
				
						// DEBUGGING ..
						//cout << "	<%catalogTable%> found .." << endl;
				
						// instantiate a new "CatalogTable" object ..
						catalogTable = new CatalogTable;
				
						// DEBUGGING ..
						//cout << "	new CatalogTable .. " << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "tableName" ) == 0 ) {

						// "catalogTable >> tableName" ..
						string tableName;
						tokenCounter++;
						tableName = metaDataTokens->at( tokenCounter );
				
						if( catalogTable != NULL )
							catalogTable->setTableName( tableName );
				
						// DEBUGGING ..
						//cout << "	tableName: " << tableName << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "tableSize" ) == 0 ) {

						// "catalogTable >> tableSize" ..
						int tableSize;
						tokenCounter++;
						string dataFound = metaDataTokens->at( tokenCounter );
						tableSize = strToInt( dataFound );
				
						if( catalogTable != NULL )
							catalogTable->setTableSize( tableSize );
				
						// DEBUGGING ..
						//cout << "	tableSize: " << tableSize << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "numberOfAttributes" ) == 0 ) {

						// "catalogTable >> numberOfAttributes" ..
						int numberOfAttributes;
						tokenCounter++;
						string dataFound = metaDataTokens->at( tokenCounter );
						numberOfAttributes = strToInt( dataFound );
				
						if( catalogTable != NULL )
							catalogTable->setNumberOfAttributes( numberOfAttributes );
				
						// DEBUGGING ..
						//cout << "	numberOfAttributes: " << numberOfAttributes << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "tableAttributeNameList" ) == 0 ) {
			
						// instantiate a new "CatalogTableAttributes" object ..
						catalogTableAttributes = new vector< CatalogTableAttribute* >;
				
						// DEBUGGING ..
						//cout << "	<%tableAttributeNameList%> found .." << endl;
						//cout << "	new list<CatalogTableAttribute> .. " << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "tableAttributeName" ) == 0 ) {
				
						// "catalogTable >> catalogTableAttribute" ..
						tokenCounter++;
						string attributeName = metaDataTokens->at( tokenCounter );
				
						//CatalogTableAttribute catalogTableAttribute( attributeName );
						CatalogTableAttribute* catalogTableAttribute = new CatalogTableAttribute( attributeName );
					
						if( catalogTableAttributes != NULL )
							catalogTableAttributes->push_back( catalogTableAttribute );
				
						// DEBUGGING ..
						//cout << "	tableAttributeName: " << attributeName << endl;
				
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "/tableAttributeNameList" ) == 0 ) {

						// "tableAttributeNameList" END ..
						if( catalogTable != NULL )
							catalogTable->setAttributes( catalogTableAttributes );
				
						// DEBUGGING ..
						//cout << "	( *catalogTable ).getNumberOfAttributes(): " << ( *catalogTable ).getNumberOfAttributes() << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "/catalogTable" ) == 0 ) {

						// "catalogTable" END ..
						if( ( catalogTables != NULL ) && ( catalogTable != NULL ) )
							catalogTables->push_back( catalogTable );
				
						// DEBUGGING ..
						//cout << "	<%/catalogTable%> found .." << endl;
						//cout << "	catalogTables.size(): " << catalogTables.size() << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "/catalogTableList" ) == 0 ) {

						// "catalogTableList" END ..
						if( catalogTables != NULL )
							catalog->setCatalogTables( catalogTables );
				
						// DEBUGGING ..
						//cout << "	<%/catalogTableList%> found .." << endl;
						//cout << "	catalog->getNumberOfTables(): " << catalog->getNumberOfTables() << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "catalogIndexList" ) == 0 ) {
				
						// DEBUGGING ..
						//cout << "	<%catalogIndexList%> found .." << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "catalogIndex" ) == 0 ) {
				
						// DEBUGGING ..
						//cout << "	<%catalogIndex%> found .." << endl;
				
						// instantiate a new "CatalogIndex" object ..
						catalogIndex = new CatalogIndex;
				
						// DEBUGGING ..
						//cout << "	new CatalogIndex .. " << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "indexName" ) == 0 ) {

						// "catalogIndex >> indexName" ..
						string indexName;
						tokenCounter++;
						indexName = metaDataTokens->at( tokenCounter );
				
						if( catalogIndex != NULL )
							catalogIndex->setIndexName( indexName );
				
						// DEBUGGING ..
						//cout << "	indexName: " << indexName << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "indexTableName" ) == 0 ) {

						// "catalogIndex >> indexTableName" ..
						string indexTableName;
						tokenCounter++;
						indexTableName = metaDataTokens->at( tokenCounter );
				
						if( catalogIndex != NULL )
							catalogIndex->setIndexTableName( indexTableName );
				
						// DEBUGGING ..
						//cout << "	indexTableName: " << indexTableName << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "indexSize" ) == 0 ) {

						// "catalogIndex >> indexSize" ..
						string indexSize;
						tokenCounter++;
						indexSize = metaDataTokens->at( tokenCounter );
				
						if( catalogIndex != NULL )
							catalogIndex->setIndexSize( strToInt( indexSize ) );
				
						// DEBUGGING ..
						//cout << "	indexSize: " << indexSize << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "isNoDuplicates" ) == 0 ) {

						// "catalogIndex >> isDuplicates" ..
						string isDuplicates;
						tokenCounter++;
						isDuplicates = metaDataTokens->at( tokenCounter );
				
						if( catalogIndex != NULL )
							catalogIndex->setIsNoDuplicates( setBoolean( strToInt( isDuplicates ) ) );
				
						// DEBUGGING ..
						//cout << "	isDuplicates: " << isDuplicates << endl;

					}  else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "indexAttributeName" ) == 0 ) {

						// "catalogIndex >> indexAttributeName" ..
						string indexAttributeName;
						tokenCounter++;
						indexAttributeName = metaDataTokens->at( tokenCounter );
				
						if( catalogIndex != NULL )
							catalogIndex->setAttribute( indexAttributeName );
				
						// DEBUGGING ..
						//cout << "	indexAttributeName: " << indexAttributeName << endl;

					}  else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "/catalogIndex" ) == 0 ) {

						// "catalogIndex" END ..
						if( ( catalogIndices != NULL ) && ( catalogIndex != NULL ) )
							catalogIndices->push_back( catalogIndex );
				
						// DEBUGGING ..
						//cout << "	<%/catalogIndex%> found .." << endl;
						//cout << "	catalogIndices.size(): " << catalogIndices.size() << endl;
			
					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "/catalogIndexList" ) == 0 ) {

						// "catalogIndexList" END ..
						if( catalogIndices != NULL )
							catalog->setCatalogIndices( catalogIndices );
				
						// DEBUGGING ..
						//cout << "	<%/catalogIndexList%> found .." << endl;
						//cout << "	catalog->getNumberOfIndices(): " << catalog->getNumberOfIndices() << endl;

					} else if ( strcmp( metaDataTokens->at( tokenCounter ).c_str(), "/end" ) == 0 ) {

						// DEBUGGING ..
						//cout << "	END REACHED .." << endl; 

						// stop processing .. 
						break;
			
					} else {
			
						// its not a recognized delimiter ..
					
						// DEBUGGING ..
						//cout << "	delimiter not found .." << endl;
						//cout << "   metaDataTokens.at( tokenCounter("<<tokenCounter<<") ): " << metaDataTokens.at( tokenCounter ) << endl;
					
						break;
			
					}
			
					// loop / increment through the "metaDataTokens" vector ..
					tokenCounter++;
			
			
				} while( tokenCounter < metaDataTokens->size() );	
				
			}			
			
		} // end "if ( fileSize > dataPosition )" ..
		
		
		readMetaDataPerformed = true;
		

	} else {
				
		string errorString = "\nFailed to open Catalog file [ CM::readMetaData ], data may not have been retrieved.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
	 	getMessageManager()->setErrorStringCM( errorString );
				
	}
	
	//cout << "CatalogManager:: readMetaData() Point C4-2 .." << endl << endl;
	
	return readMetaDataPerformed;
	
} // end "readMetaData()" ..


time_t CatalogManager:: getFileTimestamp(){
	
	
	CatalogTimestamp* fileCatalogTimestamp = catalog->getFileCatalogTimestamp();
	
	time_t timestamp = NULL;
	
	if( fileCatalogTimestamp != NULL )
		timestamp = fileCatalogTimestamp->getTimestamp();
	
	return timestamp;
	
	
} // end "getFileTimestamp()" ..


time_t CatalogManager:: getManagerTimestamp(){
	
	
	CatalogTimestamp* managerCatalogTimestamp = catalog->getManagerCatalogTimestamp();
	
	time_t timestamp = NULL;
	
	if( managerCatalogTimestamp != NULL )
		timestamp = managerCatalogTimestamp->getTimestamp();
	
	return timestamp;
	
	
} // end "getManagerTimestamp()" ..


string CatalogManager:: getFileTimestampString(){
	
	CatalogTimestamp* fileCatalogTimestamp = catalog->getFileCatalogTimestamp();
	
	string timestampString = "";
	
	if( fileCatalogTimestamp != NULL )
		timestampString = fileCatalogTimestamp->getTimestampString();
	
	//char* fileTimestamp = ctime( &timestamp );
	
	//string fileTimestampString( fileTimestamp );
	
	// DEBUGGING ..
	//cout << "getFileTimestampString(): " << fileTimestampString << endl;
	
	return timestampString;
	
	
} // end "getFileTimestampString()" ..


string CatalogManager:: getManagerTimestampString(){
	
	
	CatalogTimestamp* managerCatalogTimestamp = catalog->getManagerCatalogTimestamp();
	
	string timestampString = "";
	
	if( managerCatalogTimestamp != NULL )
		timestampString = managerCatalogTimestamp->getTimestampString();
	
	//time_t timestamp = managerCatalogTimestamp->getTimestamp();
	
	//char* managerTimestamp = ctime( &timestamp );
	
	//string managerTimestampString( managerTimestamp );
	
	// DEBUGGING ..
	//cout << "getManagerTimestampString(): " << managerTimestampString << endl;
	
	return timestampString;
	
	
} // end "getManagerTimestampString()" ..


bool CatalogManager:: hasCatalogChanged(){
	
	//cout << "CatalogManager:: hasCatalogChanged() start .." << endl;
	
	bool hasCatalogChanged = false;

	// get the copy of the catalog timestamp that was
	CatalogTimestamp* managerCatalogTimestamp = catalog->getManagerCatalogTimestamp();

	time_t managerTimestamp = NULL;
	
	if( managerCatalogTimestamp != NULL )
		managerTimestamp = managerCatalogTimestamp->getTimestamp();

	// read the timestamp in the catalog, and 
	// update it into the "fileCatalogTimestamp" ..
	readTimestamp();

	// get the most recent copy of the catalog timestamp ..
	CatalogTimestamp* fileCatalogTimestamp = catalog->getFileCatalogTimestamp();

	time_t fileTimestamp = NULL;
	
	if( managerCatalogTimestamp != NULL )
		fileTimestamp = fileCatalogTimestamp->getTimestamp();
		
	// if timestamps are not the same, then the catalog has been updated ..
	if( managerTimestamp != fileTimestamp ) {

		hasCatalogChanged = true;
	
	}
	
	// DEBUGGING ..
	//cout << "managerTimestamp: " << managerTimestamp << endl;
	//cout << "   fileTimestamp: " << fileTimestamp << endl;
	
	
	return hasCatalogChanged;
	
	
} // end "hasCatalogChanged()" ..


string CatalogManager:: addCatalogTable( CatalogTable* catalogTable ){
	
	if( catalogTable == NULL )
		return "\"catalogTable\" was NULL, catalogTable was not added to the catalog.";
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// DEBUGGING ..
		//cout << "inside addCatalogTable().hasCatalogChanged() .. " << endl;
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	// add the new "catalogTable" to the copy here in memory ..
	string returnMessage = addCatalogTableToCatalog( catalogTable );
	
	// update the meta data to file ..
	writeMetaData();
	
	return returnMessage;
	
} // end "addCatalogTable( CatalogTable )" ..


string CatalogManager:: removeCatalogTable( string tableName ) {
	
	if( tableName.length() <= 0 )
		return "\"tableName\" was NULL, \"catalogTable\" was not removed from the catalog.";
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	// remove the chosen "catalogTable" from the copy here in memory ..
	string returnMessage = removeCatalogTableFromCatalog( tableName );
	
	// update the meta data to file ..
	writeMetaData();
	
	return returnMessage;
	
	
} // end "removeCatalogTable( string tableName )" ..


string CatalogManager:: replaceCatalogTable( string tableName, CatalogTable* catalogTable ){
	
	if( tableName.length() <= 0 )
		return "\"tableName\" was empty, \"catalogTable\" was not replaced in the catalog.";
		
	if( catalogTable == NULL )
		return "\"catalogTable\" was NULL, \"catalogTable\" was not replaced in the catalog.";
	
	// replace "tableName" with "catalogTable" ..
	
	bool catalogTableFound = false;
	
	// return message initialized ..
	string returnMessage = "The table was not found.  The table was not replaced.";
	
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager 
		// (in preparation for a "write" operation) ..
		readMetaData();
		
	}
	
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	//vector<CatalogTable*>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
		
	string catalogTableName = "";
	
	//for( it_ct = catalogTables->begin(); it_ct != catalogTables->end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTableName = catalogTables->at(i)->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			// remove the chosen "catalogTable" from the copy here in memory ..
			string returnMessageRemove = removeCatalogTableFromCatalog( tableName );
			
			// add the new "catalogTable" to the copy here in memory ..
			string returnMessageAdd = addCatalogTableToCatalog( catalogTable );
			
			// update the meta data to file ..
			writeMetaData();
			
			catalogTableFound = true;
			
		}
		
	}
	
	if( catalogTableFound ){
		
		returnMessage.empty();
				
		returnMessage = "The table was replaced successfully.";
		
	} else {
		
		// if the catalogTable isnt already present in the catalog, 
		// go ahead and add it ..
		addCatalogTable( catalogTable );
		
		returnMessage.empty();
				
		returnMessage = "The table was not \"found and replaced\"; however, it was added.";

	}
	
	return returnMessage;
	
	
	
} // end "replaceCatalogTable( CatalogTable catalogTable )" ..


string CatalogManager:: addCatalogIndex( CatalogIndex* catalogIndex ){
	
	if( catalogIndex == NULL )
		return "\"catalogIndex\" was NULL, \"catalogIndex\" was not added to the catalog.";
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// DEBUGGING ..
		//cout << "inside addCatalogIndex().hasCatalogChanged() .. " << endl;
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	// add the new "catalogIndex" to the copy here in memory ..
	string returnMessage = addCatalogIndexToCatalog( catalogIndex );
	
	// update the meta data to file ..
	writeMetaData();
	
	return returnMessage;
	
} // end "addCatalogIndex( CatalogIndex )" ..


string CatalogManager:: removeCatalogIndex( string indexName ) {
	
	if( indexName.length() <= 0 )
		return "\"indexName\" was empty, \"catalogIndex\" was not removed from the catalog.";
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	// remove the chosen "catalogTable" from the copy here in memory ..
	string returnMessage = removeCatalogIndexFromCatalog( indexName );
	
	// update the meta data to file ..
	writeMetaData();
	
	return returnMessage;
	
	
} // end "removeCatalogIndex( indexName )" ..


string CatalogManager:: replaceCatalogIndex( string indexName, CatalogIndex* catalogIndex ){
	
	if( indexName.length() <= 0 )
		return "\"indexName\" was empty, \"catalogIndex\" was not replaced in the catalog.";
		
	if( catalogIndex == NULL )
		return "\"catalogIndex\" was NULL, \"catalogIndex\" was not replaced in the catalog.";
	
	// replace "indexName" with "catalogIndex" ..
	
	bool catalogIndexFound = false;
	
	// return message initialized ..
	string returnMessage = "The index was not found.  The index was not added.";
	
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager 
		// (in preparation for a "write" operation) ..
		readMetaData();
		
	}
	
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	//vector<CatalogIndex*>::iterator it_ct;
	
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
		
	string catalogIndexName = "";
	
	//for( it_ci = catalogIndices->begin(); it_ci != catalogIndices->end(); it_ci++ ){
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndexName = catalogIndices->at(i)->getIndexName();
		
		if( catalogIndexName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		if ( indexName.compare( catalogIndexName ) == 0 ) {
			
			// remove the chosen "catalogTable" from the copy here in memory ..
			string returnMessageRemove = removeCatalogIndexFromCatalog( indexName );
			
			// add the new "catalogTable" to the copy here in memory ..
			string returnMessageAdd = addCatalogIndexToCatalog( catalogIndex );
			
			// update the meta data to file ..
			writeMetaData();
			
			catalogIndexFound = true;
			
		}
		
	}
	
	if( catalogIndexFound ){
		
		returnMessage.empty();
				
		returnMessage = "The index was replaced successfully.";
		
	} else {
		
		// if the catalogTable isnt already present in the catalog, 
		// go ahead and add it ..
		addCatalogIndex( catalogIndex );
		
		returnMessage.empty();
				
		returnMessage = "The index was not \"found and replaced\"; however, it was added.";

	}
	
	return returnMessage;
	
	
	
} // end "replaceCatalogIndex( CatalogIndex catalogIndex )" ..


string CatalogManager:: addCatalogTableAttribute( string tableName, CatalogTableAttribute* catalogTableAttribute ){
	
	if( tableName.length() <= 0 )
		return "\"tableName\" was empty, \"catalogTableAttribute\" was not added to the catalog.";
		
	if( catalogTableAttribute == NULL )
		return "\"catalogTableAttribute\" was NULL, \"catalogTableAttribute\" was not added to the catalog.";
	
	bool catalogTableFound = false;
	
	bool attemptToExceedAttributeMax = false;
	
	// return message initialized ..
	string returnMessage = "The table was not found.  The attribute was not added.";
	
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
		
	string catalogTableName = "";
	
	int numberOfAttributes = -1;
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
			
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			numberOfAttributes = catalogTable->getNumberOfAttributes();
			
			if( numberOfAttributes < 0 )
				continue;
			
			// numberOfAttributes cannot exceed 8 ..
			//if ( it_ct->getNumberOfAttributes() <= 7 ) {
			if ( numberOfAttributes <= 7 ) {
				
				//it_ct->addAttribute( catalogTableAttribute );
				catalogTable->addAttribute( catalogTableAttribute );
				
				// remove this table, and replace it with the updated version here in memory ..
				removeCatalogTable( tableName );
				
				// add the updated table to replace the old table to the "catalog" here in memory ..
				//addCatalogTable( (*it_ct) );
				addCatalogTable( catalogTable );
				
				// update the meta data to file ..
				writeMetaData();
				
				//cout << "addCatalogTableAttribute() added attribute " << catalogTableAttribute.getTableAttributeName() << endl;
				
			} else {
				
				attemptToExceedAttributeMax = true;
				//cout << "addCatalogTableAttribute() attribute NOT added." << endl;

			}
			
			catalogTableFound = true;
			
		}
		
	}
	

	if( catalogTableFound ){
		
		if ( attemptToExceedAttributeMax ) {
			
			returnMessage.empty();
				
			returnMessage = "The table has 8 attributes.  A table cannot have more than 8 attributes.";
				
		} else {
			
			returnMessage.empty();
				
			returnMessage = "The attribute was removed successfully.";
			
		}
		
	}
	
	return returnMessage;
	
	
} // end "addCatalogTableAttribute( string tableName, CatalogTableAttribute catalogTableAttribute )" ..


string CatalogManager:: removeCatalogTableAttribute( string tableName, string attributeName ){
	
	if( tableName.length() <= 0 )
		return "\"tableName\" was empty, \"catalogTableAttribute\" was not removed from the catalog.";
		
	if( attributeName.length() <= 0 )
		return "\"attributeName\" was empty, \"catalogTableAttribute\" was not removed from the catalog.";
	
	bool catalogTableFound = false;
	
	// return message initialized ..
	string returnMessage = "The attribute could not be removed, because the attribute could not be found.";
		
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
		
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//it_ct->removeAttribute( attributeName );
			catalogTable->removeAttribute( attributeName );
			
			// remove this table, and replace it with the updated version here in memory ..
			string returnMessageRemoveCatalogTable = removeCatalogTable( tableName );
			
			// add the updated table to replace the old table for the "catalog" here in memory ..
			//string returnMessageAddCatalogTable = addCatalogTable( ( *it_ct ) );
			string returnMessageAddCatalogTable = addCatalogTable( catalogTable );
			
			// update the meta data to file ..
			writeMetaData();
			
			catalogTableFound = true;
					
			break;
				
		}
			
	}
				
	if( catalogTableFound ){
		
		returnMessage.empty();
				
		returnMessage = "The attribute was removed successfully.";
		
	}
	
	return returnMessage;
	
} // end "removeCatalogTableAttribute( string tableName, string attributeName )" ..


string CatalogManager:: replaceCatalogTableAttribute( string tableName, string attributeName, CatalogTableAttribute* catalogTableAttribute ){
	
	if( tableName.length() <= 0 )
		return "\"tableName\" was empty, \"catalogTableAttribute\" was not replaced in the catalog.";
		
	if( attributeName.length() <= 0 )
		return "\"attributeName\" was empty, \"catalogTableAttribute\" was not replaced in the catalog.";
	
	if( catalogTableAttribute == NULL )
		return "\"catalogTableAttribute\" was NULL, \"catalogTableAttribute\" was not replaced in the catalog.";
	
	// replace the "tableName" having "attributeName", with "catalogTableAttribute" ..
	
	//cout << "CatalogManager::  replaceCatalogTableAttribute() start .." << endl;
	
	bool catalogTableFound = false;
	
	// return message initialized ..
	string returnMessage = "The table was not found.  The attribute was not replaced.";
	
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager (in preparation for a "write" operation) ..
		readMetaData();
		
	}
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
		
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//it_ct->removeAttribute( attributeName );
			catalogTable->removeAttribute( attributeName );
			
			//it_ct->addAttribute( catalogTableAttribute );
			catalogTable->addAttribute( catalogTableAttribute );
			
			// remove this table, and replace it with the updated version here in memory ..
			string returnMessageRemoveCatalogTable = removeCatalogTable( tableName );
				
			// add the updated table to replace the old table for the "catalog" here in memory ..
			//string returnMessageAddCatalogTable = addCatalogTable( ( *it_ct ) );
			string returnMessageAddCatalogTable = addCatalogTable( catalogTable );
				
			// update the meta data to file ..
			writeMetaData();
			
			catalogTableFound = true;
					
			break;
			
		}
		
	}  // end of "for()" ..

	if( catalogTableFound ){
		
		returnMessage.empty();
				
		returnMessage = "The attribute was replaced successfully.";
		
	} else {
		
		// if the catalogTable isnt already present in the catalog, 
		// go ahead and add it to the catalog ..
		addCatalogTableAttribute( tableName, catalogTableAttribute );
		
		returnMessage.empty();
		
		return returnMessage;
		
	}
	
	return returnMessage;
	

} // end "replaceCatalogTable( CatalogTable catalogTable )" ..


bool CatalogManager:: isCatalogTableNameUnique( string tableName ){
	
	if( tableName.length() <= 0 )
		return false;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	bool isTableNameUnique = true;
	
	int tableNameCounter = 0;
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
		
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			tableNameCounter++;
			
		}
		
	}
	
	// if there is a count of more than one table with "tableName",
	// then the tableName is NOT unique ..
	if ( tableNameCounter > 1 ) {
		
		isTableNameUnique = false;
		
	}
	
	return isTableNameUnique;
	
	
} // end "isCatalogTableNameUnique( string tableName )" ..


bool CatalogManager:: isCatalogTableAttributeNameUnique( string tableName, string attributeName ){
	
	if( tableName.length() <= 0 )
		return false;
		
	if( attributeName.length() <= 0 )
		return false;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	bool isAttributeNameUnique = true;
	
	int attributeNameCounter = 0;
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	CatalogTableAttribute* catalogTableAttribute = NULL;
	
	//vector<CatalogTable>::iterator it_ct;

	int catalogTableAttributesSize = -1;

	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	string catalogTableAttributeName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		catalogTableName = catalogTable->getTableName();
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//vector<CatalogTableAttribute*>* catalogTableAttributes = it_ct->getAttributes();
			vector<CatalogTableAttribute*>* catalogTableAttributes = catalogTable->getAttributes();
			
			//vector<CatalogTableAttribute>::iterator it_cta;
			
			if( catalogTableAttributes != NULL )
				catalogTableAttributesSize = catalogTableAttributes->size();
			
			//for( it_cta = catalogTableAttributes.begin(); it_cta != catalogTableAttributes.end(); it_cta++ ){
			for( int j=0; j<catalogTableAttributesSize; j++ ){
				
				catalogTableAttribute = catalogTableAttributes->at(j);
				
				if( catalogTableAttribute == NULL )
					continue;
				
				catalogTableAttributeName = catalogTableAttribute->getTableAttributeName();
				
				//cout << "attributeName: " << ( *it_cta ).getTableAttributeName() << endl;
				
				//if ( attributeName.compare( it_cta->getTableAttributeName() ) == 0 ) {
				if ( attributeName.compare( catalogTableAttributeName ) == 0 ) {
					
					attributeNameCounter++;
					
				}

			}
			
			if ( attributeNameCounter > 0 ){
				
				break;
				
			}
			
		}
		
	}
	
	// if there is a count of more than one table with "tableName",
	// then the tableName is NOT unique ..
	if ( attributeNameCounter > 1 ) {
		
		isAttributeNameUnique = false;
		
	}
	
	
	return isAttributeNameUnique;
	
	
} // end "isCatalogTableAttributeNameUnique( string tableName, string attributeName )" ..


bool CatalogManager:: doesCatalogTableNameExist( string tableName ){

	if( tableName.length() <= 0 )
		return false;
	
	//cout << "CatalogManager:: doesCatalogTableNameExist() start .." << endl;
	
	//cout << "CatalogManager:: doesCatalogTableNameExist() Point C4 .." << endl << endl;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}

	//cout << "CatalogManager:: doesCatalogTableNameExist() Point C4-1 .." << endl << endl;
	
	bool doesTableNameExist = false;
	
	int tableNameCounter = 0;
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	//cout << "CatalogManager:: doesCatalogTableNameExist() Point C5 .." << endl << endl;
	
	//vector<CatalogTable>::iterator it_ct;
	
	//cout << "Point D2.." << endl;
	//cout << "catalogTables.size(): " << catalogTables.size() << endl;

	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		//cout << "Point D2_["<< ++i <<"].." << endl;
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			tableNameCounter++;
			//cout << "tableName["<< tableName <<"].compare( ( *it_ct ).getTableName()["<< ( *it_ct ).getTableName() <<"] ) == 0: " << (tableName.compare( ( *it_ct ).getTableName() ) == 0) << endl;
			//cout << "tableNameCounter: " << tableNameCounter << endl;
			
		}
		
	}
	
	//cout << "CatalogManager:: doesCatalogTableNameExist() Point C6 .." << endl << endl;
	
	// if there is a count of more than zero tables with "tableName",
	// then the tableName DOES exist already ..
	if ( tableNameCounter > 0 ) {
		
		doesTableNameExist = true;
		
	}
	
	//cout << "CatalogManager:: doesCatalogTableNameExist() Point C7 .." << endl << endl;
	
	
	return doesTableNameExist;
	
	
} // end "doesCatalogTableNameExist( string tableName )" ..


bool CatalogManager:: doesCatalogIndexNameExist( string indexName ){

	if( indexName.length() <= 0 )
		return false;
	
	//cout << "CatalogManager:: doesCatalogIndexNameExist() start .." << endl;
	
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}

	
	bool doesIndexNameExist = false;
	
	int indexNameCounter = 0;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	CatalogIndex* catalogIndex = NULL;
	
	//vector<CatalogIndex>::iterator it_ci;

	// DEBUGGING ..
	//cout << "catalogIndices.size(): " << catalogIndices.size() << endl;

	//for( it_ci = catalogIndices->begin(); it_ci != catalogIndices->end(); it_ci++ ){
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
		
	string catalogIndexName = "";
		
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		catalogIndexName = catalogIndex->getIndexName();
		
		//cout << "Point D2_["<< ++i <<"].." << endl;
		
		// find the catalogTable with "tableName" ..
		//if ( indexName.compare( it_ci->getIndexName() ) == 0 ) {
		if ( indexName.compare( catalogIndexName ) == 0 ) {
			
			indexNameCounter++;
			// DEBUGGING ..
			//cout << "indexName["<< indexName <<"].compare( ( *it_ct ).getIndexName()["<< ( *it_ct ).getTableName() <<"] ) == 0: " << (tableName.compare( ( *it_ct ).getTableName() ) == 0) << endl;
			//cout << "indexNameCounter: " << indexNameCounter << endl;
			
		}
		
	}
	
	// if there is a count of more than zero tables with "indexName",
	// then the indexName DOES exist already ..
	if ( indexNameCounter > 0 ) {
		
		doesIndexNameExist = true;
		
	}
	
	
	return doesIndexNameExist;
	
	
} // end "doesCatalogIndexNameExist( string indexName )" ..


bool CatalogManager:: doesCatalogTableAttributeNameExist( string tableName, string attributeName ){
	
	if( tableName.length() <= 0 )
		return false;
		
	if( attributeName.length() <= 0 )
		return false;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	bool doesAttributeNameExist = false;
	
	int attributeNameCounter = 0;
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	vector<CatalogTableAttribute*>* catalogTableAttributes = NULL;
	
	CatalogTableAttribute* catalogTableAttribute = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	string catalogTableAttributeName = "";
	
	int catalogTableAttributesSize = -1;
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//vector<CatalogTableAttribute*>* catalogTableAttributes = it_ct->getAttributes();
			catalogTableAttributes = catalogTable->getAttributes();
			
			if( catalogTableAttributes == NULL )
					continue;
			
			//vector<CatalogTableAttribute>::iterator it_cta;
			
			catalogTableAttributesSize = catalogTableAttributes->size();
			
			//for( it_cta = catalogTableAttributes.begin(); it_cta != catalogTableAttributes.end(); it_cta++ ){
			for( int j=0; j<catalogTableAttributesSize; j++ ){
				
				catalogTableAttribute = catalogTableAttributes->at(j);
				
				if( catalogTableAttribute == NULL )
					continue;
		
				catalogTableAttributeName = catalogTableAttribute->getTableAttributeName();
				
				//cout << "attributeName: " << ( *it_cta ).getTableAttributeName() << endl;
				
				//if ( attributeName.compare( it_cta->getTableAttributeName() ) == 0 ) {
				if ( attributeName.compare( catalogTableAttribute->getTableAttributeName() ) == 0 ) {
					
					attributeNameCounter++;
					
				}

			}
			
			if ( attributeNameCounter > 0 ){
				
				break;
				
			}
			
		}
		
	}
	
	// if there is a count of more than zero attributes with "tableNameAttribute",
	// then the "tableNameAttribute" exists ..
	if ( attributeNameCounter > 0 ) {
		
		doesAttributeNameExist = true;
		
	}
	
	return doesAttributeNameExist;
	
} // end "doesCatalogTableAttributeNameExist( string tableName, string attributeName )" ..


bool CatalogManager:: doesCatalogTableHaveIndex( string tableName ){
	
	if( tableName.length() <= 0 )
		return false;
	
	//cout << "CatalogManager:: doesCatalogTableHaveIndex() Point A.." << endl << endl;
	
	//vector<string>* catalogIndexAttributeNames = getCatalogIndexAttributeNames( tableName );
	vector<string>* catalogIndexAttributeNames = getCatalogTableAttributeNames( tableName );
	
	bool doesCatalogHaveIndex = false;
	
	int catalogIndexAttributeNamesSize = -1;
	
	if( catalogIndexAttributeNames != NULL )
		catalogIndexAttributeNamesSize = catalogIndexAttributeNames->size();
		
	string catalogIndexAttributeName = "";
	
	for( int i=0; i<catalogIndexAttributeNamesSize; i++ ){
		
		if( catalogIndexAttributeNames != NULL )
			catalogIndexAttributeName = catalogIndexAttributeNames->at(i);
			
		if( catalogIndexAttributeName.length() <= 0 )
			continue;
		
		doesCatalogHaveIndex = doesCatalogTableAttributeNameHaveIndex( catalogIndexAttributeName, tableName );
		
		if( doesCatalogHaveIndex )
			break;
		
	}
	
	return doesCatalogHaveIndex;
	
}


bool CatalogManager:: doesCatalogTableAttributeNameHaveIndex( string tableName, string attributeName ){
	
	if( tableName.length() <= 0 )
		return false;
		
	if( attributeName.length() <= 0 )
		return false;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	bool doesAttributeNameHaveIndex = false;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	CatalogIndex* catalogIndex = NULL;
	
	//vector<CatalogIndex>::iterator it_ci;
	
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
	
	string indexTableName = "";
	CatalogIndexAttribute* catalogIndexAttribute = NULL;
	string indexAttributeName = "";
	
	//for( it_ci = catalogIndices->begin(); it_ci != catalogIndices->end(); it_ci++ ){
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		indexTableName = catalogIndex->getIndexTableName();
		
		if( indexTableName.length() <= 0 )
			continue;
		
		catalogIndexAttribute = catalogIndex->getAttribute();
		
		if( catalogIndexAttribute == NULL )
			continue;
		
		indexAttributeName = catalogIndexAttribute->getIndexAttributeName();
		
		if( indexAttributeName.length() <= 0 )
			continue;
		
		// find the catalogIndex with "tableName" ..
		if ( ( tableName.compare( indexTableName ) == 0 ) &&
		     ( attributeName.compare( indexAttributeName ) == 0 ) ) { 
			
			doesAttributeNameHaveIndex = true;
			break;
			
		}
		
	}
	
	return doesAttributeNameHaveIndex;
	
} // end "doesCatalogTableAttributeNameHaveIndex( string tableName, string attributeName )" ..


vector<string>* CatalogManager:: getCatalogIndexAttributeNames( string tableName ) {
	
	if( tableName.length() <= 0 )
		return NULL;
	
	vector<string>* indexAttributeNames = NULL;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	CatalogIndex* catalogIndex = NULL;
	
	//vector<CatalogIndex>::iterator it_ci;
	
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
	
	string indexTableName = "";
	CatalogIndexAttribute* catalogIndexAttribute = NULL;
	string indexAttributeName = "";
	
	//for( it_ci = catalogIndices->begin(); it_ci != catalogIndices->end(); it_ci++ ){
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		indexTableName = catalogIndex->getIndexTableName();
		
		// find the catalogIndex with "tableName" ..
		if ( tableName.compare( indexTableName ) == 0 ) { 
			
			if( indexAttributeNames == NULL)
				indexAttributeNames = new vector<string>();
			
			catalogIndexAttribute = catalogIndex->getAttribute();
			
			if( catalogIndexAttribute == NULL )
				continue;
			
			indexAttributeName = catalogIndexAttribute->getIndexAttributeName();
			
			if( indexAttributeName.length() <= 0 )
				continue;
			
			if( indexAttributeNames == NULL )
				indexAttributeNames = new vector<string>();
			
			indexAttributeNames->push_back( indexAttributeName );
			
		}
		
	}
	
	return indexAttributeNames;
	
} // end "getCatalogIndexAttributeNames()" ..


vector<string>* CatalogManager:: getCatalogIndexNames( string tableName ) {
	
	if( tableName.length() <= 0 )
		return NULL;
	
	vector<string>* indiceNames = NULL;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	CatalogIndex* catalogIndex = NULL;
	
	//vector<CatalogIndex>::iterator it_ci;
	
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
	
	string indexTableName = "";
	string indexName = "";
	
	//for( it_ci = catalogIndices->begin(); it_ci != catalogIndices->end(); it_ci++ ){
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		indexTableName = catalogIndex->getIndexTableName();
		
		if( indexTableName.length() <= 0 )
			continue;
		
		// collect the names of all the indices for this tableName ..
		if ( tableName.compare( indexTableName ) == 0 ){
			
			if( indiceNames == NULL)
				indiceNames = new vector<string>();

			indexName = catalogIndex->getIndexName();

			if( indexName.length() <= 0 )
				continue;

			if( indiceNames == NULL )
				indiceNames = new vector<string>();

			//indiceNames->push_back( it_ci->getIndexName() );
			indiceNames->push_back( indexName );	
			
		}

	}
	
	return indiceNames;
	
} // end "getCatalogIndexNames()" ..


string CatalogManager:: getCatalogIndexName( string tableName, string attributeName ) {
	
	// return the "indexName" associated with this "attributeName" ..
	
	//cout << "CatalogManager:: getCatalogIndexName() tableName: " << tableName << endl << endl;
	
	//cout << "CatalogManager:: getCatalogIndexName() attributeName: " << attributeName << endl << endl;
	
	if( tableName.length() <= 0 )
		return NULL;
		
	if( attributeName.length() <= 0 )
		return NULL;
	
	string indexName = "";
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	CatalogIndex* catalogIndex = NULL;
	
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
	
	string indexTableName = "";
	string indexAttributeName = "";
	CatalogIndexAttribute* catalogIndexAttribute = NULL;
	
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		indexTableName = catalogIndex->getIndexTableName();
		
		if( indexTableName.length() <= 0 )
			continue;
		
		// get the indexName for this tableName, attributeName ..
		if ( tableName.compare( indexTableName ) == 0 ){
			
			catalogIndexAttribute = catalogIndex->getAttribute();
			
			if( catalogIndexAttribute == NULL )
				continue;
				
			indexAttributeName = catalogIndexAttribute->getIndexAttributeName();
			
			if( indexAttributeName.length() <= 0 )
				continue;
			
			if ( attributeName.compare( indexAttributeName ) == 0 ){
				
				indexName = catalogIndex->getIndexName();
				break;
				
			}
			
		}

	}
	
	return indexName;
	
} // end "getCatalogIndexName()" ..


string CatalogManager:: getCatalogAttributeNameOfIndex( string tableName, string indexName ) {
	
	// return the "attributeName" associated with this "indexName" ..
	
	if( tableName.length() <= 0 )
		return NULL;

	if( indexName.length() <= 0 )
		return NULL;
	
	string attributeName = NULL;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	CatalogIndex* catalogIndex = NULL;
	
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
		
	string indexTableName = "";
	string indexAttributeName = "";
	CatalogIndexAttribute* catalogIndexAttribute = NULL;
	
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		indexTableName = catalogIndex->getIndexTableName();
		
		// get the indexName for this tableName, attributeName ..
		if ( tableName.compare( indexTableName ) == 0 ){
			
			indexTableName = catalogIndex->getIndexName();
			
			if( indexTableName.length() <= 0 )
				continue;
			
			if ( indexName.compare( indexTableName ) == 0 ){
				
				catalogIndexAttribute = catalogIndex->getAttribute();
				
				if( catalogIndexAttribute == NULL )
					continue;
				
				attributeName = catalogIndexAttribute->getIndexAttributeName();
				break;
				
			}
			
		}

	}
	
	return attributeName;
	
} // end "getCatalogAttributeNameOfIndex()" ..


vector<CatalogTable*>* CatalogManager:: getCatalogTables() {
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	return catalog->getCatalogTables();

	
} // end "getCatalogTables()" ..


CatalogTable* CatalogManager:: getCatalogTable( string tableName ) {
	
	if( tableName.length() <= 0 )
		return NULL;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	CatalogTable* foundCatalogTable = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for(int i=0; i<catalogTablesSize; i++){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		// find the catalogTable with "tableName" ..
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			foundCatalogTable = catalogTable;
			
			break;
			
		}
		
	}
	
	return foundCatalogTable;
	
} // end "getCatalogTable( string tableName )" ..


vector<CatalogIndex*>* CatalogManager:: getCatalogIndices() {
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	return catalog->getCatalogIndices();

	
} // end "getCatalogIndices()" ..


CatalogIndex* CatalogManager:: getCatalogIndex( string indexName ) {
	
	if( indexName.length() <= 0 )
		return NULL;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	CatalogIndex* catalogIndex = NULL;
	
	CatalogIndex* foundCatalogIndex = NULL;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	//cout << "CatalogManager:: getCatalogIndex() catalogIndices->size(): " << catalogIndices->size() << endl << endl;

/*
	vector<CatalogIndex>::iterator it_ci;
	
	for( it_ci = catalogIndices.begin(); it_ci != catalogIndices.end(); it_ci++ ){
		
		cout << "CatalogManager:: getCatalogIndex_indexName() ( indexName("<<indexName<<").compare( it_ci->getIndexName()("<<it_ci->getIndexName()<<") ) == 0 ): " << ( indexName.compare( it_ci->getIndexName() ) == 0 ) << endl << endl;
		
		// find the catalogIndex with "indexName" ..
		if ( indexName.compare( it_ci->getIndexName() ) == 0 ) {
			
			//foundCatalogIndex = ( *it_ci );
			foundCatalogIndex = &( *it_ci );
			
			cout << "CatalogManager:: getCatalogIndex_indexName() ( *it_ci ).getIndexName(): " << ( *it_ci ).getIndexName() << endl << endl;
			
			break;
			
		}
		
	}
*/

	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();

	string tableIndexName = "";

	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		tableIndexName = catalogIndex->getIndexName();
		
		if( tableIndexName.length() <= 0 )
			continue;
		
		// find the catalogIndex with "indexName" ..
		if ( indexName.compare( tableIndexName ) == 0 ) {
			
			//foundCatalogIndex = ( *it_ci );
			foundCatalogIndex = catalogIndex;
			
			//cout << "CatalogManager:: getCatalogIndex() catalogIndices->at(i).getIndexName(): " << catalogIndices->at(i)->getIndexName() << endl << endl;
			
			break;
			
		}
		
	}
	
	return foundCatalogIndex;
	
} // end "getCatalogIndex_indexName( string indexName )" ..


CatalogIndex* CatalogManager:: getCatalogIndex_indexName( string tableName, string indexName ) {
	
	if( tableName.length() <= 0 )
		return NULL;
		
	if( indexName.length() <= 0 )
		return NULL;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	CatalogIndex* catalogIndex = NULL;
	
	CatalogIndex* foundCatalogIndex = NULL;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	//cout << "CatalogManager:: getCatalogIndex() catalogIndices->size(): " << catalogIndices->size() << endl << endl;
	
	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();
	
	string catalogIndexName = "";
	string catalogIndexTableName = "";
	
	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		//cout << "CatalogManager:: getCatalogIndex() ( indexName("<<indexName<<").compare( it_ci->getIndexName()("<<catalogIndex->getIndexName()<<") ) == 0 ): " << ( indexName.compare( catalogIndex->getIndexName() ) == 0 ) << endl << endl;
		//cout << "CatalogManager:: getCatalogIndex() catalogIndices->at(i).getIndexTableName(): " << catalogIndex->getIndexTableName() << endl << endl;
		
		if( catalogIndex == NULL )
			continue;
		
		catalogIndexName = catalogIndex->getIndexName();
		
		if( catalogIndexName.length() <= 0 )
			continue;
		
		catalogIndexTableName = catalogIndex->getIndexTableName();
		
		if( catalogIndexTableName.length() <= 0 )
			continue;
		
		// find the catalogIndex with "indexName" ..
		if ( ( indexName.compare( catalogIndexName ) == 0 ) &&
		     ( tableName.compare( catalogIndexTableName ) == 0 ) ) {
			
			//foundCatalogIndex = ( *it_ci );
			foundCatalogIndex = catalogIndex;
			
			//cout << "CatalogManager:: getCatalogIndex() catalogIndices->at(i).getIndexName(): " << catalogIndex->getIndexName() << endl << endl;
			
			break;
			
		}
		
	}
	
	return foundCatalogIndex;
	
} // end "getCatalogIndex_indexName( string indexName )" ..


CatalogIndex* CatalogManager:: getCatalogIndex_attributeName( string tableName, string attributeName ) {
	
	if( tableName.length() <= 0 )
		return NULL;
		
	if( attributeName.length() <= 0 )
		return NULL;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	CatalogIndex* catalogIndex = NULL;
	
	CatalogIndex* foundCatalogIndex = NULL;
	
	vector<CatalogIndex*>* catalogIndices = catalog->getCatalogIndices();
	
	//cout << "CatalogManager:: getCatalogIndex() catalogIndices->size(): " << catalogIndices->size() << endl << endl;
	
/*
	vector<CatalogIndex>::iterator it_ci;
	
	for( it_ci = catalogIndices.begin(); it_ci != catalogIndices.end(); it_ci++ ){
		
		//cout << "CatalogManager:: getCatalogIndex_indexName() ( indexName("<<indexName<<").compare( it_ci->getIndexName()("<<it_ci->getIndexName()<<") ) == 0 ): " << ( indexName.compare( it_ci->getIndexName() ) == 0 ) << endl << endl;
		
		// find the catalogIndex with "indexName" ..
		if ( ( attributeName.compare( it_ci->getAttribute().getIndexAttributeName() ) == 0 ) &&
		     ( tableName.compare( it_ci->getIndexTableName() ) == 0 ) ) {
			
			//foundCatalogIndex = ( *it_ci );
			foundCatalogIndex = &( *it_ci );
			
			cout << "CatalogManager:: getCatalogIndex_indexName() ( *it_ci ).getIndexName(): " << ( *it_ci ).getIndexName() << endl << endl;
			
			break;
			
		}
		
	}
*/

	int catalogIndicesSize = -1;
	
	if( catalogIndices != NULL )
		catalogIndicesSize = catalogIndices->size();

	string catalogIndexTableName = "";
	string catalogIndexAttributeName = "";
	CatalogIndexAttribute* catalogIndexAttribute = NULL;

	for( int i=0; i<catalogIndicesSize; i++ ){
		
		catalogIndex = catalogIndices->at(i);
		
		if( catalogIndex == NULL )
			continue;
		
		//cout << "CatalogManager:: getCatalogIndex() catalogIndices->at(i).getAttribute().getIndexAttributeName(): " << catalogIndex->getAttribute()->getIndexAttributeName() << endl << endl;
		//cout << "CatalogManager:: getCatalogIndex() catalogIndices->at(i).getIndexTableName(): " << catalogIndex->getIndexTableName() << endl << endl;
		
		catalogIndexAttribute = catalogIndex->getAttribute();
		
		if( catalogIndexAttribute == NULL )
			continue;
			
		catalogIndexAttributeName = catalogIndexAttribute->getIndexAttributeName();
		
		if( catalogIndexAttributeName.length() <= 0 )
			continue;
		
		catalogIndexTableName = catalogIndex->getIndexTableName();
		
		if( catalogIndexTableName.length() <= 0 )
			continue;
		
		// find the catalogIndex with "indexName" ..
		if ( ( attributeName.compare( catalogIndexAttributeName ) == 0 ) &&
		     ( tableName.compare( catalogIndexTableName ) == 0 ) ) {
			
			//foundCatalogIndex = ( *it_ci );
			foundCatalogIndex = catalogIndex;
			
			//cout << "CatalogManager:: getCatalogIndex() catalogIndices->at(i).getIndexName(): " << catalogIndex->getIndexName() << endl << endl;
			
			break;
			
		}
		
	}
	
	return foundCatalogIndex;
	
} // end "getCatalogIndex_indexName( string indexName )" ..


CatalogTableAttribute* CatalogManager:: getCatalogTableAttribute( string tableName, string attributeName ){
	
	if( tableName.length() <= 0 )
		return NULL;

	if( attributeName.length() <= 0 )
		return NULL;
	
	bool attributeFound = false;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	vector<CatalogTableAttribute*>* catalogTableAttributes = NULL;
	
	CatalogTableAttribute* catalogTableAttribute = NULL;
	
	CatalogTableAttribute* foundCatalogTableAttribute = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	int catalogTableAttributesSize = -1;
	string catalogTableAttributeName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//vector<CatalogTableAttribute> catalogTableAttributes = it_ct->getAttributes();
			catalogTableAttributes = catalogTable->getAttributes();
			
			if( catalogTableAttributes == NULL )
				continue;
			
			//vector<CatalogTableAttribute>::iterator it_cta;
			
			catalogTableAttributesSize = catalogTableAttributes->size();
			
			//for( it_cta = catalogTableAttributes.begin(); it_cta != catalogTableAttributes.end(); it_cta++ ){
			for( int j=0; j<catalogTableAttributesSize; j++ ){
				
				catalogTableAttribute = catalogTableAttributes->at(j);
				
				if( catalogTableAttribute == NULL )
					continue;
				
				catalogTableAttributeName = catalogTableAttribute->getTableAttributeName();
				
				if( catalogTableAttributeName.length() <= 0 )
					continue;
				
				// find the catalogTableAttribute with "attributeName" ..
				//if ( attributeName.compare( it_cta->getTableAttributeName() ) == 0 ) {
				if ( attributeName.compare( catalogTableAttributeName ) == 0 ) {
					
					//foundCatalogTableAttribute = ( *it_cta );
					foundCatalogTableAttribute = catalogTableAttribute;
					
					attributeFound = true;
					
					break;
					
				}
				
			} // end "for()" ..
			
			if( attributeFound ){
				
				break;
				
			}
			
		}
		
	}
	
	return foundCatalogTableAttribute;
	
} // end "getCatalogTableAttributes( string tableName, string attributeName )" ..


vector<string>* CatalogManager:: getCatalogTableAttributeNames( string tableName ){
	
	if( tableName.length() <= 0 )
		return NULL;
	
	vector<CatalogTableAttribute*>* catalogTableAttributes = getCatalogTableAttributes( tableName );
	
	vector<string>* catalogTableAttributeNames = NULL;
	
	int catalogTableAttributesSize = -1;
	
	if( catalogTableAttributes != NULL )
		catalogTableAttributesSize = catalogTableAttributes->size();
	
	CatalogTableAttribute* catalogTableAttribute = NULL;
	string catalogTableAttributeName = "";
	
	for(int i=0; i<catalogTableAttributesSize; i++){
		
		catalogTableAttribute = catalogTableAttributes->at(i);
		
		if( catalogTableAttribute == NULL )
			continue;
		
		catalogTableAttributeName = catalogTableAttribute->getTableAttributeName();
		
		if( catalogTableAttributeName.length() <= 0 )
			continue;
		
		if( catalogTableAttributeNames == NULL )
			catalogTableAttributeNames = new vector<string>();
		
		catalogTableAttributeNames->push_back( catalogTableAttributeName );
		
	}
	
	return catalogTableAttributeNames;
	
}

vector<CatalogTableAttribute*>* CatalogManager:: getCatalogTableAttributes( string tableName ){
	
	if( tableName.length() <= 0 )
		return NULL;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	vector<CatalogTableAttribute*>* foundCatalogTableAttributes = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//foundCatalogTableAttributes = it_ct->getAttributes();
			foundCatalogTableAttributes = catalogTable->getAttributes();
			
			break;
			
		}
		
	}
	
	return foundCatalogTableAttributes;
	
} // end "getCatalogTableAttributes( string tableName, string attributeName )" ..



int CatalogManager:: getNumberOfTables(){
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	int numberOfTables = catalog->getNumberOfTables();
	
	return numberOfTables;
	
} // end "getNumberOfTables()" ..


int CatalogManager:: getTableSize( string tableName ){
	
	if( tableName.length() <= 0 )
		return -1;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	int tableSize = 0;
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//tableSize = it_ct->getTableSize();
			tableSize = catalogTable->getTableSize();
			
		}
		
	}
	
	return tableSize;
	
} // end "getTableSize( string tableName )" ..


void CatalogManager:: setTableSize( string tableName, int tableSize ){
	
	if( tableName.length() <= 0 )
		return;
	
	if( tableSize < 0 )
		return;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	//vector<CatalogTable>::iterator it_ct;
	
	CatalogTable* catalogTable = NULL;
	
	//CatalogTable* catalogTableToUpdate = NULL;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//( *it_ct ).setTableSize( tableSize );
			//catalogTableToUpdate = ( *it_ct );
			//catalogTableToUpdate.setTableSize( tableSize );
			//catalogTables.erase( it_ct );
			catalogTable->setTableSize( tableSize );
			break;
			
		}
		
	}
	
	// add the table with the updated "tableSize" ..
	//catalogTables.push_back( catalogTableToUpdate );
	
	// make sure that the "Catalog" gets set to have the updated table ..
	catalog->setCatalogTables( catalogTables );
	
	// update that Catalog with the latest "tableSize" ..
	writeMetaData();
	
} // end "CatalogManager::setTableSize( string tableName, int tableSize )" ..


int CatalogManager:: getNumberOfAttributes( string tableName ) {
	
	if( tableName.length() <= 0 )
		return -1;
	
	//cout << "CatalogManager:: getNumberOfAttributes() start .." << endl;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}

	
	int numberOfAttributes = 0;
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	CatalogTable* catalogTable = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
	
	string catalogTableName = "";
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
			
		catalogTableName = catalogTable->getTableName();
			
		if( catalogTableName.length() <= 0 )
			continue;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//numberOfAttributes = it_ct->getNumberOfAttributes();
			numberOfAttributes = catalogTable->getNumberOfAttributes();
			
			break;
			
		}
		
	}
	
	
	return numberOfAttributes;
	
	
} // end "getNumberOfAttributes( string tableName )" ..


int CatalogManager:: getAttributePosition( string tableName, string attributeName ) {
	
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point A .." << endl << endl;
	
	if( tableName.length() <= 0 )
		return -1;
		
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point A1 .." << endl << endl;
		
	if( attributeName.length() <= 0 )
		return -1;
	
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point A2 .." << endl << endl;
	
	// check for changes in the file ..
	if ( hasCatalogChanged() ){
		
		// if the file changed, then update the manager ..
		readMetaData();
		
	}
	
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point A3 .." << endl << endl;
	
	int attributePosition = -1;
	
	int attributeCounter = 0;
	
	vector<CatalogTable*>* catalogTables = catalog->getCatalogTables();
	
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point A4 .." << endl << endl;
	
	CatalogTable* catalogTable = NULL;
	
	vector<CatalogTableAttribute*>* catalogTableAttributes = NULL;
	
	CatalogTableAttribute* catalogTableAttribute = NULL;
	
	//vector<CatalogTable>::iterator it_ct;
	
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point A5 .." << endl << endl;
	
	int catalogTablesSize = -1;
	
	if( catalogTables != NULL )
		catalogTablesSize = catalogTables->size();
		
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point A6 .." << endl << endl;
	
	string catalogTableName = "";
	int catalogTableAttributesSize = -1;
	string catalogTableAttributeName = "";
	
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point B .." << endl << endl;
	
	//for( it_ct = catalogTables.begin(); it_ct != catalogTables.end(); it_ct++ ){
	for( int i=0; i<catalogTablesSize; i++ ){
		
		catalogTable = catalogTables->at(i);
		
		if( catalogTable == NULL )
			continue;
		
		catalogTableName = catalogTable->getTableName();
		
		if( catalogTableName.length() <= 0 )
			continue;
			
		//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point C .." << endl << endl;
		
		// find the catalogTable with "tableName" ..
		//if ( tableName.compare( it_ct->getTableName() ) == 0 ) {
		if ( tableName.compare( catalogTableName ) == 0 ) {
			
			//vector<CatalogTableAttribute*>* catalogTableAttributes = it_ct->getAttributes();
			catalogTableAttributes = catalogTable->getAttributes();
			
			if( catalogTableAttributes == NULL )
				continue;
			
			//vector<CatalogTableAttribute>::iterator it_cta;
			
			catalogTableAttributesSize = catalogTableAttributes->size();
			
			//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point D .." << endl << endl;
			
			// find the position of the attributeName ..
			//for( it_cta = catalogTableAttributes.begin(); it_cta != catalogTableAttributes.end(); it_cta++ ){
			for( int j=0; j<catalogTableAttributesSize; j++ ){
				
				catalogTableAttribute = catalogTableAttributes->at(j);
				
				if( catalogTableAttribute == NULL )
					continue;
				
				catalogTableAttributeName = catalogTableAttribute->getTableAttributeName();
				
				if( catalogTableAttributeName.length() <= 0 )
					continue;
				
				//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point E .." << endl << endl;
				
				//if ( it_cta->getTableAttributeName().compare( attributeName ) == 0 ) {
				if ( attributeName.compare( catalogTableAttributeName ) == 0 ) {
					
					// if this "attributeName" in the catalog matches the one passed in,
					// track this attribute position ..
					// ( i.e.: table(aa, bb, cc) => aa (position 0), bb (position 1), cc (position 2), etc ) ..
					
					attributePosition = attributeCounter;
					
					break;
					
				}
				
				attributeCounter++;
				
			} // end "for()" ..
			
			// as soon as the catalog table is found, stop looping ..
			break;
			
		} // end "if()" ..
		
	} // end "for()" ..
	
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() >> getAttributePosition() Point F .." << endl << endl;
	
	return attributePosition;
	
} // end "getAttributePosition( string tableName, string attributeName )" ..


/*
 * SETTERS ..
 */



/*
 * GETTERS ..
 */


MessageManager* CatalogManager:: getMessageManager(){
	
	return mm;
	
}


Catalog* CatalogManager:: getCatalog(){
	
	return this->catalog;
	
}

CatalogFile* CatalogManager:: getCatalogFile(){
	
	return this->catalogFile;
	
}
