#ifndef CATALOGMANAGER_H_
#define CATALOGMANAGER_H_

#pragma once

#include <cstdio>
#include <ctime>
#include <fstream>
#include <string>
#include <map>
#include <vector>

#include "../Catalog/Catalog.h"
#include "../Catalog/CatalogTimestamp.h"
#include "../Catalog/CatalogFile.h"

#include "../Utilities/String_Utilities.h"
#include "../Utilities/TableFormatter.h"

#include "MessageManager.h"

using namespace std;

class CatalogManager {
	
public:

	// data members:

	
	
	// member methods:
	
	CatalogManager( MessageManager* mm, CatalogFile* catalogFile_ );
	
	~CatalogManager();
	
	
	/*
	 * DEBUGGING ..
	 */
	
	void printCatalogData( vector<string>* tableData );			// helper method for "printCatalog()" ..
	void printCatalog();
	void oldPrintCatalog();
	
	/*
	 * UTILITY ..
	 */
	 
	void setCatalogFile( CatalogFile* catalogFile_ );
	void resetCatalog();

	
	/*
	 * TIMESTAMP ..
	 */
	 
	void writeTimestamp();
	
	void writeTimestamp( ofstream* outputFile );
	
	void readTimestamp();
	
	
	/*
	 * METADATA ..
	 */
	
	
	// write the entire catalog to file, from manager ..
	bool writeMetaData();
	
	// read the entire catalog from file, to manager ..
	bool readMetaData();
	
	
	
	// manage "timestamp" of last write-to-file ..

	// "fileTimestamp" last "timestamp" written, stored in the file ..
	time_t getFileTimestamp();
	
	// "managerTimestamp" last "timestamp" written, stored in the manager ..
	time_t getManagerTimestamp();
	
	string getFileTimestampString();
	
	string getManagerTimestampString();
	
	
	// compare "fileTimestamp" with "managerTimestamp", and see if the catalog file has changed (write operation performed on it) ..
	bool hasCatalogChanged();
	
	
	// add / remove / replace the table / attribute metadata ..
	string addCatalogTable( CatalogTable* catalogTable );
	
	string removeCatalogTable( string tableName );
	
	string replaceCatalogTable( string tableName, CatalogTable* catalogTable );
	
	string addCatalogIndex( CatalogIndex* catalogIndex );
	
	string removeCatalogIndex( string indexName );
	
	string replaceCatalogIndex( string indexName, CatalogIndex* catalogIndex );
	
	string addCatalogTableAttribute( string tableName, CatalogTableAttribute* catalogTableAttribute );
	
	string removeCatalogTableAttribute( string tableName, string attributeName );
	
	string replaceCatalogTableAttribute( string tableName, string attributeName, CatalogTableAttribute* catalogTableAttribute );
	
	
	// check table / attribute "uniqueness" ..
	bool isCatalogTableNameUnique( string tableName );
	
	bool isCatalogTableAttributeNameUnique( string tableName, string attributeName );
	
	bool doesCatalogTableNameExist( string tableName );
	
	bool doesCatalogIndexNameExist( string indexName );
	
	bool doesCatalogTableAttributeNameExist( string tableName, string attributeName );
	
	bool doesCatalogTableHaveIndex( string tableName );
	
	bool doesCatalogTableAttributeNameHaveIndex( string tableName, string attributeName );
	
	vector<string>* getCatalogIndexAttributeNames( string tableName );
	
	vector<string>* getCatalogIndexNames( string tableName );
	
	string getCatalogIndexName( string tableName, string attributeName );
	
	string getCatalogAttributeNameOfIndex( string tableName, string indexName );
	
	// "getters" for table / attribute metadata ..
	vector<CatalogTable*>* getCatalogTables();
	
	CatalogTable* getCatalogTable( string tableName );
	
	vector<CatalogIndex*>* getCatalogIndices();
	
	CatalogIndex* getCatalogIndex( string indexName );
	
	CatalogIndex* getCatalogIndex_indexName( string tableName, string indexName );
	
	CatalogIndex* getCatalogIndex_attributeName( string tableName, string attributeName );
	
	vector<CatalogTableAttribute*>* getCatalogTableAttributes( string tableName );
	
	CatalogTableAttribute* getCatalogTableAttribute( string tableName, string attributeName );

	vector<string>* getCatalogTableAttributeNames( string tableName );
	
	int getNumberOfTables();
	
	int getTableSize( string tableName );
	
	void setTableSize( string tableName, int tableSize );
	
	int getNumberOfAttributes( string tableName );
	
	int getAttributePosition( string tableName, string attributeName );
	
	/*
	 * SETTERS ..
	 */
	 

	/*
	 * GETTERS ..
	 */
	 
	 MessageManager* getMessageManager();
	 
	 Catalog* getCatalog();
	 
	 CatalogFile* getCatalogFile();
	 


private:

	// data members:
	
	MessageManager* mm;				// Tracks "error" and "result" messages produces in each "manager" component ..
	
	Catalog* catalog;				// represents all the tables of the Catalog file ..
	
	CatalogFile* catalogFile;
	
	
	// member methods:
	
	/*
	 * wrappers for adding "CatalogTable" to "Catalog" ..
	 */
	 
	string addCatalogTableToCatalog( CatalogTable* catalogTable );
	
	string addCatalogIndexToCatalog( CatalogIndex* catalogIndex );
	
	string removeCatalogTableFromCatalog( string tableName );
	
	string removeCatalogIndexFromCatalog( string tableName );
	
	
}; // end "CatalogManager" class ..

#endif /*CATALOGMANAGER_H_*/
