#ifndef INDEXMANAGER_H_
#define INDEXMANAGER_H_

#include "../DataPage/DataPage.h"

#include "../Database/DatabaseFile.h"

#include "../Index/IndexPage.h"
#include "../Index/IndexHeader.h"
#include "../Index/IndexRows.h"
#include "../Index/IndexNodeRow.h"
#include "../Index/IndexNode.h"
#include "../Index/IndexBucket.h"
#include "../Index/IndexBucketHeader.h"
#include "../Index/IndexBucketHeaderEntry.h"

#include "../Catalog/CatalogIndex.h"

#include "StorageManager.h"
#include "CatalogManager.h"
#include "MessageManager.h"

class IndexManager {
	
public:

	IndexManager();
	IndexManager( StorageManager* sm_ );
	
	virtual~IndexManager();
	
	StorageManager* getStorageManager();
	MessageManager* getMessageManager();
	CatalogManager* getCatalogManager();
	
	string constructPagePadding( int fillerSize );

	void setIsIndexDebug( bool isIndexDebug_ );
	bool getIsIndexDebug();
	
	int calculateIndexSize( string tableName_, string indexName_ );

	bool createIndex( string indexName_, string indexTableName_, bool isNoDuplicates_, string attributeName_ );
	bool createIndexWithData( string indexName_, string indexTableName_, bool isNoDuplicates_, string attributeName_ );
	void createIndexLeafNodes( string indexTableName_, string attributeName_, vector<DataPage*>* dataPages_, vector<int>* pageIDs, vector<int>* allSearchKeys_, vector<int>* searchKeyPageIDs_ );
	bool checkForDuplicates( vector<int>* allSearchKeys );
	
	void createUniqueSearchKeys( vector<int>* uniqueSearchKeys_, vector<int>* allSearchKeys_ );
	
	void createIndexBucketsAndUniqueSearchKeys( vector<IndexBucket*>* indexBuckets_, vector<int>* uniqueSearchKeys_, IndexBucketHeader* indexBucketHeader, vector<int>* allSearchKeys_, vector<int>* searchKeyPageIDs_ );
	
	void createParentNodes( vector<int>* uniqueSearchKeys_, vector< vector< vector<int>* >* >* parentNodes_, vector< bool >* extraParentNodeGeneratedPerRow_ );
	void createIndexPages( vector<int>* allSearchKeys, vector<int>* searchKeyPageIDs, vector<IndexPage*>* indexPages_, IndexHeader* indexHeader, vector< vector< vector<int>* >* >* parentNodes_, vector<bool>* extraParentNodeGeneratedPerRow_, bool isNoDuplicates_ );
	void addPageIDsToIndexBucket( vector<int>* searchKeyPageIDs_, IndexBucket* indexBucket_, int startPosition, int endPosition );

	bool writeIndexHeader( string tableName_, string indexName_, IndexHeader* header_ );
	bool writeIndexBucketHeader( string tableName_, string indexName_, IndexBucketHeader* indexBucketHeader_ );
	bool writeIndexBucket( string tableName, string indexName , IndexBucket* indexBucket, int bucketPageID );
	bool writeIndexBuckets( string indexTableName_, string indexName_, vector<IndexBucket*>* indexBuckets_ );
	//map<int, vector<int>* >* StorageManager:: writeIndexBuckets( vector<IndexBucket*>* indexBuckets_, char* indexTableName_, char* indexName_ );

	bool writeIndexPage( string tableName_, string indexName_, IndexPage* indexPage_, int pageID_, int parentPageID_, bool isNew_, bool isDuplicatesAllowed_ );
	bool writeIndexPages( string indexTableName_, string indexName_, vector<IndexPage*>* indexPages_, vector<bool>* isNew_, bool isDuplicatesAllowed_ );

	bool readIndexHeader( string tableName_, string indexName_, IndexHeader* header_ );
	bool readIndexBucketHeader( string tableName_, string indexName_, IndexBucketHeader* indexBucketHeader_ );
	bool readIndexBucket( string tableName_, string indexName_, IndexBucket* indexBucket_, int bucketPageID_ );

	bool readIndexPage( string tableName_, string indexName_, IndexPage* indexPage_, int pageID_ );
	bool readIndexPages( string indexTableName_, string indexName_, vector<IndexPage*>* indexPages_ );
	int getIndexSize( string tableName_, string indexName_ );

	bool updateIndex( string indexName_, string indexTableName_, string attributeName_, int searchKey_, int childPageID_ );

	void printIndex( string indexTableName_, string indexName_ );

	vector<int>* lookUpLeafNodeID( int searchKey_, string indexTableName_, string indexName_, bool &searchKeyFound_ );
	vector<int>* lookUpIDs( string indexTableName_, string indexName_, int searchKey_ );
	vector<int>* lookUpIDsForIndices( string indexTableName_, vector<string>* indexNames_ );
	int lookUpIDForInsert( string indexTableName_, string indexName_, int searchKey_ );

	//void setIsPrintIndex( bool isPrintIndex_ );
	//bool getIsPrintIndex();

private:

	static const int blockSize = 48;
	static const int intSize = sizeof(int);
	static const int boolSize = sizeof(bool);
	static const int indexBucketSize = 12;	// number of pageIDs per bucket (4b ints, 48 bytes per bucket) ..
	static const int fanout = 5;
	
	bool isIndexDebug;
	
	StorageManager* sm;
	
	// contains file directory information ..
	DatabaseFile* databaseFile;	

};

#endif /*INDEXMANAGER_H_*/
