#ifndef QUERYMANAGER_H_
#define QUERYMANAGER_H_

#pragma once

#include "../DataPage/DataPage.h"
#include "../DataHeader/DataHeader.h"
#include "../Buffer/BufferPool.h"
#include "../Buffer/BufferInfoTable.h"
#include "../Buffer/BITSnapShot.h"
#include "../Catalog/CatalogTable.h"
#include "../Catalog/CatalogFile.h"
#include "../Database/DatabaseFile.h"

#include "IndexManager.h"
#include "BufferManager.h"
#include "StorageManager.h"
#include "CatalogManager.h"
#include "MessageManager.h"

// helper classes ..
#include "../Aggregations/DataForWhereClauseAggregation.h"

#include "../Utilities/Vector_Utilities.h"

#include <string>
#include <vector>
#include <iostream>
#include <fstream>

using namespace std;


class QueryManager {
	
public:

	QueryManager( int bufferPoolSize, CatalogFile* catalogFile, DatabaseFile* databaseFile  );
	~QueryManager(void);
	
	// data members ..
	
	
	// member methods ..
	
	QueryManager* getQueryManager();
	
	IndexManager* getIndexManager();
	
	BufferManager* getBufferManager();
	
	StorageManager* getStorageManager();
	
	CatalogManager* getCatalogManager();
	
	MessageManager* getMessageManager();
	
	
	// "CREATE TABLE X" ..
	void createTable( string tableName, vector<string>* attributeNames, int numberOfAttributes );
	
	// "CREATE INDEX I" ..
	void createIndex( string indexName, string tableName, bool isNoDuplicates, string attributeName );
	
	// "INSERT INTO X VALUES( a1, a2, .. )" ..
	void insertIntoTableValues( string tableNameX, vector<string>* insertValues, int numberOfAttributes );
	
	// "COMMIT" ..
	void commit();
	
	// helper method for "insertIntoTableValues()" and "commit()" ..
	void updateIndex( vector<string>* tableNames );
	
	DataPage* getGoodRowsDataPage( DataPage* dataPage, vector< vector<string>* >* queryPairs, vector<bool>* isPairs, string tableNameX );
	
	// "PRINT BUFFER" ..
	map<int, DataPage* >* selectBuffer( string tableNameX );
	
	// helper method for "selectFromX()" ..
	DataPage* processDataForWhereClause( DataForWhereClauseAggregation* dfwcp );
	
	// "SELECT * FROM X where a = 1000 and a = bb" ..
	//DataPage* selectStarFromX( char* columnNameA, char* tableNameX );
	//vector<DataPage> selectStarFromX( char* tableNameX );
	//void selectStarFromX( char* tableNameX, vector<DataPage> &bpDataPages, vector<DataPage> &fileDataPages );
	//vector< vector<DataPage> > selectStarFromX( char* tableNameX );
	//vector< map<int, DataPage > > selectStarFromX( char* tableNameX );
	//vector< map<int, DataPage > > selectStarFromX( char* tableNameX, list<int> &LRUBlocks );
	//vector< map<int, DataPage > > selectStarFromX( char* tableNameX, BITSnapShot &bitSnapShot );
	//vector< map<int, DataPage > > selectStarFromX( char* tableNameX );
	//vector< map<int, DataPage > > selectStarFromX( char* tableNameX, multimap<string, string> whereColumnNamePairs, map<string, int> whereColumnNameValuePairs );
	vector< map<int, DataPage* >* >* selectFromX( string tableNameX, vector<string>* selectColumnNames, multimap<string, string>* whereColumnNamePairs, map<string, int>* whereColumnNameValuePairs );
	
	// helper method ..
	map<int, string>* groupSelectColumnNames( string tableNameX, vector<string>* selectColumnNames );

private:

	IndexManager* im;

	// present for "BufferPool", "BufferInfoTable", "CatalogManager", "StorageManager" ..
	BufferManager* bm;
	
	// sometimes QueryManager will access storage manager directly ..
	StorageManager* sm;
	
};

#endif /*QUERYMANAGER_H_*/
