#include "StorageManager.h"


StorageManager:: StorageManager() {

	//cout << "StorageManager:: StorageManager() start .." << endl << endl;

	headerSize = 0;
	
	fieldCount = 0;

	//cout << "StorageManager:: StorageManager() end .." << endl << endl;

}


StorageManager:: StorageManager( CatalogManager* cm_, DatabaseFile* databaseFile_ ){

	//cout << "StorageManager:: StorageManager() start .." << endl << endl;


	cm = cm_;

	databaseFile = databaseFile_;

	headerSize = 0;

	//cout << "StorageManager:: StorageManager() end .." << endl << endl;

}


StorageManager:: ~StorageManager(void){

	//cout << "~StorageManager() start .." << endl << endl;


	//delete mm;  // This is done in the CatalogManager ..

	//delete cm;  // This is done in the BufferManager ..


	//cout << "~StorageManager() end .." << endl << endl;

}

void StorageManager:: setDatabaseFile( DatabaseFile* databaseFile_ ){
	
	databaseFile = databaseFile_;
	
}


CatalogManager* StorageManager:: getCatalogManager(){

	return cm;

}


MessageManager* StorageManager:: getMessageManager(){

	return cm->getMessageManager();

}


string StorageManager:: constructPagePadding( int padSize ){

	// create the "filler" string that is used to fill up
	// the portion of a 48-byte page in the file
	// which is not occupied by data, so that
	// the file size is always an increment of 48 bytes

	// i.e.: data = 36 bytes, filler = 12 bytes, total = 48 bytes ..

	string pad;

	pad.assign( padSize, '%' );

	return pad;


} // end "constructPagePadding( int padSize )" ..


bool StorageManager:: createTable( string tableName, int fieldCount, vector<string>* attributeNames ){

	//cout << "StorageManager:: createTable() start .." << endl << endl;

	// DEBUGGING ..
	//cout << "StorageManager:: createTable() tableName: " << tableName << endl << endl;
	//cout << "StorageManager:: createTable() fieldCount: " << fieldCount << endl << endl;

	// creates the "meta-data" that tracks the table (size, # columns, etc)
	// in the "DataHeader" file (a file separate from the data file) ..

	// creating the table involves creating the header file,
	// but not creating the data file ..
	// (the data file is only created once the user issues the first "insert" command)

	// this method writes 1 4-byte integers to the "header" file
	// using the "writeHeader()" method ..

	// header->totalNumberOfDataPages: 4 bytes ..

	DataHeader* header = new DataHeader();
	header->setTotalNumberOfDataPages( 0 ); 			// initial # data pages is zero (no data yet) ..
	//header->setFilePageFieldCount( fieldCount );		// "fieldCount" is # of attributes or columns of the table ..
	
	// "writeHeader" needs this, since the catalog hasn't been updated yet ..
	this->fieldCount = fieldCount;

	//cout << "StorageManager:: createTable() Point A .." << endl << endl;

	bool writeHeaderPerformed = writeHeader( tableName, header );
	
	//cout << "StorageManager:: createTable() Point B .." << endl << endl;

	if ( writeHeaderPerformed == false ) {

	 	string errorString = "\nHeader file was not written to [ SM::createTable() ], table may not have been created properly.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );

	} else {

		// update the Catalog ..
		
		// DEBUGGING ..
		//cout << "StorageManager:: createTable() fieldCount: " << fieldCount << endl << endl;
		//for(int i=0; i<attributeNames->size(); i++){
		//	cout << "\"" << attributeNames->at(i) << "\" is the " << i << " field" << endl;
		//}

		CatalogTable* catalogTable = new CatalogTable();

		catalogTable->setTableName( tableName );

		// "addAttribute" increments "numberOfAttributes" ..
		catalogTable->setNumberOfAttributes( 0 );

		for( int i=0; i<fieldCount; i++ ) {

			catalogTable->addAttribute( attributeNames->at(i) );

		}

		// the table starts out without any records ( empty )..
		catalogTable->setTableSize( 0 );

		//catalogFile = getCatalogManager()->getCatalogFile();

		// write the new table to the Catalog file ..
	 	string returnMessageCatalog = getCatalogManager()->addCatalogTable( catalogTable );

	}

	// return whether or not "writeHeader()" worked ..
	return writeHeaderPerformed;


	//cout << "StorageManager:: createTable() end .." << endl << endl;


} // end "createTable( int fieldCount, string tableName )" ..



int StorageManager:: calculateNumberOfFreeRows( int rowNumberPerPage, int columnNumber ){

	// calculates the number of rows "free" (not occupied by data) in the file-page ..

	// "rowNumberPerPage":
	// is the number of rows of the file-page occupied by data
	// (remainder of the rows of the file-page are occupied by filler string,
	// in order for each file-page to have a size equal to 48 bytes;
	// i.e.: data = 36 bytes, filler = 12 bytes, total = 48 bytes) ..

	// "columnNumber":
	// the total number of rows per file-page depends on
	// the "columnNumber" (# attributes or columns) of the table ..

	// i.e.:
	// 1) empty table that can hold only "int" values, where size of each int is 4 bytes
	// 2) page-file size is 48 bytes
	// 3) assume columnNumber = 4 (4 columns)
	// 		a) size of each row =  4 columns * 4-byte-int = 16 bytes
	// 		b)    max # of rows = 48 bytes / 16 bytes per row = 3 rows
	// 4) assume columnNumber = 2 (2 columns)
	// 		a) size of each row =  2 columns * 4-byte-int = 8 bytes
	// 		b)    max # of rows = 48 bytes / 8 bytes per row = 6 rows

	// Thus, the "max # of rows" of a file-page depends on
	// the number of columns of the table (fieldCount) ..
	// However, this is not calculated here ..
	// instead, the number of "free" rows is calculated!

	int sizeOfInt = 4;

	int numberOfFreeRows = 0;

	if ( columnNumber > 0 ) {

		// "size of each row"
		// columnNumber * sizeOfInt ..
		int sizeOfSingleRow = ( columnNumber * sizeOfInt );

		// "page-file size"
		// ( rowNumberPerPage * columnNumber * sizeOfInt ) ..
		int sizeOfAllDataEntries = ( rowNumberPerPage * sizeOfSingleRow );

		// "rows of free space"
		// ( 48 bytes - ( rowNumberPerPage * columnNumber * sizeOfInt ) ) ..
		int sizeOfFreeSpace = ( blockSize - ( sizeOfAllDataEntries ) );

		// "# of free space rows"
		// ( blockSize - ( sizeOfAllDataEntries ) ) / ( columnNumber * sizeOfInt ) ..
		numberOfFreeRows = ( sizeOfFreeSpace / sizeOfSingleRow );

	}


	return numberOfFreeRows;


} // end "calculateNumberOfFreeRows( int rowNumberPerPage, int columnNumber )" ..


void StorageManager:: activateTablePageLock( string tableName, int pageID, int lockNumberIncrement ) {

	//cout << "BufferManager:: activateTablePageLock() start .." << endl << endl;

	// "locks" a data file-page of a table,
	// making it inaccessible to all but a single user,
	// until the user releases the lock ..

	//           "tableName": name of the table with a page to be "locked" ..

	//              "pageID": index of the file-page ..

	// "lockNumberIncrement": integer greater than zero, used as flag indicating that
	//						  the file-page should be "locked" ..
	//						  (note: originally this was going to be a "timed" lock,
	//						   where every-so-often the "lockNumberIncrement" would be decremented
	//						   and when it reached 0, the lock would be released unless the user's
	//						   process reactivated the lock .. but multi-threading in C++ turned out
	//						   to be too involved / hard to locate information on)


	DataHeader* header = new DataHeader();
	bool isHeaderRead = readHeader( tableName, header );
	
	//cout << "BufferManager:: activateTablePageLock() Point A .." << endl << endl;

	if ( pageID < header->getTotalNumberOfDataPages() ) {
		
		//cout << "BufferManager:: activateTablePageLock() Point B .." << endl << endl;

		// "lockNumberIncrement" must be greater than 0 to activate the lock on the file-page ..
		if( lockNumberIncrement > 0 ){
			
			//cout << "BufferManager:: activateTablePageLock() Point C .." << endl << endl;

			// Originally, it was planned that the lock was "active" only when lockNumber
			// would be _incremented_ every X minutes, ( lockNumber > 0 = active lock )
			// and would only be removed by the initiator ..
			// header._headerEntries[ pageID ].lockNumber += lockNumberIncrement;

			// now it is enough to invoke "activateTablePageLock()" a single time,
			// setting the "headerEntry" lockNumber to a value greater than 0 ..
			header->getHeaderEntries()->at( pageID )->setLockNumber( lockNumberIncrement );
			
			//cout << "BufferManager:: activateTablePageLock() Point C .." << endl << endl;

		} else {
			
			//cout << "BufferManager:: activateTablePageLock() Point D .." << endl << endl;

			// default value (arbitrary, but should be greater than zero)..
			header->getHeaderEntries()->at( pageID )->setLockNumber( 5 );
			
			//cout << "BufferManager:: activateTablePageLock() Point E .." << endl << endl;

		}
		
		//cout << "BufferManager:: activateTablePageLock() Point F .." << endl << endl;

		// update "DataHeader" ..
		writeHeader( tableName, header );
		
		//cout << "BufferManager:: activateTablePageLock() Point G .." << endl << endl;

	}


} // end "activateTablePageLock( int pageID, int lockNumberIncrement, string tableName )" ..


void StorageManager:: deactivateTablePageLock( string tableName, int pageID ){

	// "unlocks" a data file-page of a table,
	// making it accessible again to any user,
	// now that the user has released the lock ..

	//           "tableName": name of the table with a page to be "unlocked" ..

	//              "pageID": index of the file-page ..

	DataHeader* header = new DataHeader();
	readHeader( tableName, header );

	// only use a valid "pageID" ..
	if ( ( pageID < header->getTotalNumberOfDataPages() ) &&
	     ( pageID >= 0 ) ) {

		// ( lockNumber > 0 ) : active lock which
		// can only be removed by the initiator ..
		header->getHeaderEntries()->at( pageID )->setLockNumber( 0 );

	}

	// update "DataHeader" ..
	writeHeader( tableName, header );


} // end "deactivateTablePageLock( int pageID, string tableName )" ..


bool StorageManager:: isTableHasLocks( string tableName ) {

	// "tableName": name of the table with page(s) that may be "locked" ..

	DataHeader* header = new DataHeader();
	readHeader( tableName, header );

	bool locksFound = false;

	for (int i = 0; i < header->getTotalNumberOfDataPages(); i++){

		// track whether this tableName has any locks ..
		if ( header->getHeaderEntries()->at( i )->getLockNumber() > 0 ){

			locksFound = true;

			break;

		}

	}

	return locksFound;

} // end "tableHasLocks( string tableName )" ..


vector<int> StorageManager:: getTablePageLocks( string tableName ) {

	// "tableName": name of the table with pages "locked and "unlocked" ..

	DataHeader* header = new DataHeader();
	readHeader( tableName, header );

	vector<int> lockNumbers;

	for (int i = 0; i < header->getTotalNumberOfDataPages(); i++){

		// get all of the "lockNumbers" for every page
		// in the file of this "tableName" ..
		lockNumbers.push_back( header->getHeaderEntries()->at( i )->getLockNumber() );

	}

	return lockNumbers;

} // end "getTablePageLocks( string tableName )" ..


int StorageManager:: getTablePageLock( string tableName, int pageID ) {

	// "tableName": name of the table with a page to be checked for a "lock" ..

	// "pageID": index of the file-page ..

	DataHeader* header = new DataHeader();
	readHeader( tableName, header );

	int lockNumber = 0;

	if ( ( pageID < header->getTotalNumberOfDataPages() ) &&
	     ( pageID >= 0 ) ) {

		// get the current "lockNumber" for the given "pageID" ..
		lockNumber = header->getHeaderEntries()->at( pageID )->getLockNumber();

	}

	return lockNumber;

} // end "getTablePageLock( int pageID, string tableName )" ..


void StorageManager:: clearAllTablePageLocks( string tableName ) {

	// "tableName": name of the table with a page to be released of a "lock" ..

	DataHeader* header = new DataHeader();
	readHeader( tableName, header );

	for (int i = 0; i < header->getTotalNumberOfDataPages(); i++){

		// release the "lock" held on any and all pages for this "tableName" ..
		header->getHeaderEntries()->at( i )->setLockNumber( 0 );

	}

	// update "DataHeader" ..
	writeHeader( tableName, header );

} // end "clearAllTablePageLocks( string tableName )" ..


void StorageManager:: clearTablePageLocks( string tableName, int* pageIDs, int intArraySize ) {

	// "tableName": name of the table with page(s) to be "unlocked" ..

	// "pageID": indexes of the file-pages ..

	// NOTE: this method wont work,
	// because you will go off the edge of pageID int array ..
	// see other method ..

	DataHeader* header = new DataHeader();
	readHeader( tableName, header );

	for (int i = 0; i < header->getTotalNumberOfDataPages(); i++){

		for(int j=0; j<intArraySize; j++){

			if ( ( pageIDs[j] < header->getTotalNumberOfDataPages() ) &&
	        	 ( pageIDs[j] >= 0 ) &&
	         	 ( pageIDs[j] == i ) ) {

				// release the "lock" only on "pageID[i]",
				// skipping other locks not in the collection ..
				header->getHeaderEntries()->at( i )->setLockNumber( 0 );

	         }

		}

	}

	// update "DataHeader" ..
	writeHeader( tableName, header );


} // end "clearTablePageLocks( int* pageIDs, int intArraySize, string tableName )" ..


void StorageManager:: clearTablePageLocks( string tableName, vector<int>* pageIDs ) {

	// "tableName": name of the table with page(s) to be "unlocked" ..

	// "pageID": indexes of the file-pages ..

	DataHeader* header = new DataHeader();
	readHeader( tableName, header );

	for (int i = 0; i < header->getTotalNumberOfDataPages(); i++){

		for(int j=0; j<pageIDs->size(); j++){

			if ( pageIDs->at(j) == i ) {

				// "unlock" the lock on the pages, one for each "pageID" ..
				header->getHeaderEntries()->at( i )->setLockNumber( 0 );

			}

		}

	}

	// update "DataHeader" ..
	writeHeader( tableName, header );


} // end "clearTablePageLocks( vector<int> pageIDs, string tableName )" ..



bool StorageManager:: writeHeader( string tableName, DataHeader* header ) {

	// "tableName": name of the table associated with the header ..

	//    "header": header to write to file ..

	bool writeHeaderPerformed = false;

	string emptyString = "";


	if ( emptyString.compare( tableName ) != 0 ) {


		// handle the directory path of the file ..

		//string directoryPlusFileName = "C:\\Temp\\";

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		// append the file
		directoryPlusFileName += "DataHeader_";
		directoryPlusFileName += tableName;
		char* fileName = (char*)directoryPlusFileName.c_str();

		fstream headerFile( fileName, ios::in | ios::out | ios::binary );

		// this means the file doesn't exist yet,
		// so attempt to create the file ready for output ..
		if( headerFile.fail() ){

			// needed in order to "clear" errorflags left over from first "fail" ..
			headerFile.clear();

			headerFile.open( fileName, ios::out | ios::binary | ios::app );

		}

		if ( !headerFile.fail() ) {

			headerFile.seekp( ios::beg );

			// DEBUGGING ..
			//checkFileState( dataTable );

			// number of fields in the "header" ..
			int numberOfHeaderFields = 2;

			// number of fields in each "headerEntry" of the "header" ..
			int numberOfHeaderEntryFields = 4;

			int fieldCount = 0;
			
			CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );

			if( catalogTable != NULL )
				fieldCount = catalogTable->getNumberOfAttributes();

			int maxNumberOfRows = header->computeMaxRowCount( fieldCount );

			// size in bytes of the "rowsAvailable" array found in each "headerEntry" of the "header" ..
			// [ (2 bytes) * (N rows per page) ]
			int rowsAvailableSize = 12;
			
			if( maxNumberOfRows >= 0 )
				rowsAvailableSize = ( boolSize * maxNumberOfRows );

			// size in bytes of a single "headerEntry" of the "header" ..
			// [ [ (4 bytes) * (4 fields) ] +  ]
			int headerPageEntrySize =  ( ( intSize * numberOfHeaderEntryFields ) + rowsAvailableSize );

			// size in bytes of the header file ..
			// [ (4 bytes) * (2 fields) ] + [ (M data pages) * (single header entry size) ]
			// 8 bytes + M*16 bytes == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			headerSize = ( ( intSize * numberOfHeaderFields ) + ( header->getTotalNumberOfDataPages() * headerPageEntrySize ) );


			//int filePageFieldCount = header->getFilePageFieldCount();

			int totalNumberOfDataPages = header->getTotalNumberOfDataPages();

			// write the "header" fields to the header file ..
			//headerFile.write( (const char *) ( &( fieldCount ) ), intSize );

			headerFile.write( reinterpret_cast<const char*>( &totalNumberOfDataPages ), intSize );


			// write the "headerEntry" fields to the header file ..

			int pid = 0;

			int fsc = 0;

			int ln = 0;

			int fprc = 0;

			for(int i=0; i<header->getTotalNumberOfDataPages(); i++){


				pid = header->getHeaderEntries()->at( i )->getPageID();
				headerFile.write( reinterpret_cast<const char*>( &pid ), intSize );

				fsc = header->getHeaderEntries()->at( i )->getFreeRowsCount();
				headerFile.write( reinterpret_cast<const char*>( &fsc ), intSize );

				ln = header->getHeaderEntries()->at( i )->getLockNumber();
				headerFile.write( reinterpret_cast<const char*>( &ln ), intSize );

				fprc = header->getHeaderEntries()->at( i )->getFilledRowsCount();
				headerFile.write( reinterpret_cast<const char*>( &fprc ), intSize );


				// if "rowAvailable" does not have enough members to match the current
				// "headerPageEntry" size, then resize "rowAvailable" and set its defaults ..
				if( header->getHeaderEntries()->at( i )->getRowAvailable()->size() != maxNumberOfRows ){

					header->getHeaderEntries()->at( i )->getRowAvailable()->clear();

					for(int j=0; j<maxNumberOfRows; j++) {

						header->getHeaderEntries()->at( i )->getRowAvailable()->push_back( true );

						//cout << "StorageManager:: writeHeader() header.getHeaderEntries()->at( i("<<i<<") ).getRowAvailable()->push_back( true ) .." << endl << endl;

					}

				}

				// write the "rowAvailable" fields for this "headerEntry" to the header file ..
				bool slotAvailable = true;
				for(int j=0; j<maxNumberOfRows; j++) {

					slotAvailable = header->getHeaderEntries()->at( i )->getRowAvailable()->at( j );

					headerFile.write( reinterpret_cast<const char*>( &slotAvailable ), boolSize );

					//cout << "StorageManager:: writeHeader() dataTable.write( reinterpret_cast< const char * >( &( slotAvailable("<<slotAvailable<<") ) ), intSize("<<boolSize<<") ) .." << endl << endl;

				}

			}

			int pagePaddingSize = 0;

			int gaugePagePaddingSize = blockSize;


			// increase "gaugeTheFillerSize" by increments of "blockSize",
			// until "gaugeTheFillerSize >= headerSize" ..
			// i.e.:         headerSize == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			//		      exit the loop: (48 >= 24), (48 >= 40), (96 >= 56), (96 >= 72), (96 >= 88) ..
			while ( true ){

				if( gaugePagePaddingSize >= headerSize ){

					break;

				} else {

					gaugePagePaddingSize += blockSize;

				}

			} // end "while ( true )" ..

			// "fillerSize" should be less than 48 bytes,
			// such that adding "fillerSize" to "headerSize"
			// is an even multiple of "blockSize" ..
			// i.e.:         headerSize == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			//     gaugePagePaddingSize == 48, 96, 144, 192, 240 ..
			//          pagePaddingSize == (48 - 24), (48 - 40), (96 - 56), (96 - 72), (96 - 88), (144 - 104) ..
			pagePaddingSize = ( gaugePagePaddingSize - headerSize );


			if ( pagePaddingSize > 0 ) {

				string pagePadding = constructPagePadding( pagePaddingSize );

				int charSize = 1;

				// write the filler string to the DataHeader file,
				// so that the header file size is an even multiple of "blockSize" ..
				for(int i=0; i<pagePadding.size(); i++){

					// DEBUGGING ..
					//cout << filler[i];

					headerFile.write( reinterpret_cast<const char*>( &pagePadding[i] ), charSize );

				}

				// DEBUGGING ..
				//cout << endl << endl;


			}

			// DEBUGGING ..
			//checkFileState( tbl );

			// flag indicating the "writeHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			writeHeaderPerformed = true;

		} else {

	 		string errorString = "\nHeader file failed to open [ SM::writeHeader() ], so Header file was not written.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..


		headerFile.clear();
		headerFile.close();


	} else {

	 	string errorString = "\nHeader file table name was not valid [ SM::writeHeader() ], so Header file was not written.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	return writeHeaderPerformed;


} // end "writeHeader( DataHeader* header, string tableName )" ..


bool StorageManager:: readHeader( string tableName, DataHeader* header ) {

	//cout << "StorageManager:: readHeader() Point D10-1A .." << endl << endl;

	bool readHeaderPerformed = false;

	string emptyString = "";

	// DEBUGGING ..
	//cout << "StorageManager:: readHeader() tableName: " << tableName << endl;

	if ( emptyString.compare( tableName ) != 0 ) {

		//string directoryPlusFileName = "C:\\Temp\\";

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		directoryPlusFileName += "DataHeader_";
		directoryPlusFileName += tableName;
		char* fileName = (char*)directoryPlusFileName.c_str();

		//fstream headerFile(tableName, ios::in );
		//fstream headerFile( tableName, ios::in | ios::binary );
		fstream headerFile( fileName, ios::in | ios::binary );

		// if the constructor fails, give it a second shot ..
		if( headerFile.fail() ){

			// needed in order to "clear" errorflags left over from first "fail" ..
			headerFile.clear();

			headerFile.open( fileName, ios::in | ios::binary );

		}

		//cout << "StorageManager:: readHeader() Point D10-1B .." << endl << endl;

		if ( !headerFile.fail() ) {

			headerFile.seekg( ios::beg );

			// DEBUGGING ..
			//checkFileState( tbl );
			
			//cout << "StorageManager:: readHeader() Point D10-1C .." << endl << endl;

			int fieldCount = 0;
			
			CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );

			if( catalogTable != NULL )
				fieldCount = catalogTable->getNumberOfAttributes();

			int totalNumberOfDataPages = -1;

			//dataTable.read( (char *) ( &( fieldCount ) ), intSize );

			headerFile.read( reinterpret_cast<char*>( &totalNumberOfDataPages ), intSize );

			//header->setFilePageFieldCount( filePageFieldCount );

			header->setTotalNumberOfDataPages( totalNumberOfDataPages );

			//cout << "StorageManager:: readHeader() Point D10-1D .." << endl << endl;

			// DEBUGGING ..
			//cout << "StorageManager:: readHeader() header->totalNumberOfDataPages: " << header->totalNumberOfDataPages << endl;
			//cout << "StorageManager:: readHeader() header->filePageFieldCount: " << header->filePageFieldCount << endl;


			int pid = 0;

			int fsc = 0;

			int ln = 0;

			int fprc = 0;

			bool sa = false;


			header->getHeaderEntries()->clear();
			
			//cout << "StorageManager:: readHeader() Point D10-1D1 .." << endl << endl;

			int numberOfSlotsAvailable = header->computeMaxRowCount( fieldCount );

			//cout << "StorageManager:: readHeader() Point D10-1E .." << endl << endl;

			for (int i=0; i<header->getTotalNumberOfDataPages(); i++){

				header->getHeaderEntries()->push_back( new DataHeaderEntry( numberOfSlotsAvailable ) );

				headerFile.read( reinterpret_cast<char*>( &pid ), intSize );
				header->getHeaderEntries()->at( i )->setPageID( pid );

				headerFile.read( reinterpret_cast<char*>( &fsc ), intSize );
				header->getHeaderEntries()->at( i )->setFreeRowsCount( fsc );

				headerFile.read( reinterpret_cast<char*>( &ln ), intSize );
				header->getHeaderEntries()->at( i )->setLockNumber( ln );

				headerFile.read( reinterpret_cast<char*>( &fprc ), intSize );
				header->getHeaderEntries()->at( i )->setFilledRowsCount( fprc );


				for(int j=0; j<numberOfSlotsAvailable; j++) {

					headerFile.read( reinterpret_cast<char*>( &sa ), boolSize );
					header->getHeaderEntries()->at( i )->getRowAvailable()->at( j ) = sa;

				}

				// DEBUGGING ..
				//cout << "SM:: readHeader() header.headerPageEntries["<<i<<"].pageID: " << header._headerEntries[i].pageID << endl;
				//cout << "SM:: readHeader() header.headerPageEntries["<<i<<"].freeRowsCount: " << header._headerEntries[i].freeRowsCount << endl;
				//cout << "SM:: readHeader() header.headerPageEntries["<<i<<"].lockNumber: " << header._headerEntries[i].lockNumber << endl;
				//cout << "SM:: readHeader() header.headerPageEntries["<<i<<"].filledRowsCount: " << header._headerEntries[i].filledRowsCount << endl;
				//for(int j=0; j<numberOfSlotsAvailable; j++) {
					//cout << "SM:: readHeader() header.headerPageEntries["<<i<<"].rowAvailable["<<j<<"]: " << header._headerEntries[i].rowAvailable[j] << endl;
				//}

			}
			
			//cout << "StorageManager:: readHeader() Point D10-1F .." << endl << endl;

			// DEBUGGING ..
			//checkFileState( tbl );

			// flag indicating the "readHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			readHeaderPerformed = true;

		} else {

	 		string errorString = "\nHeader file failed to open [ SM::readHeader() ], so Header file was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..


		headerFile.clear();
		headerFile.close();


	} else {

	 	string errorString = "\nHeader file table name was not valid [ SM::readHeader() ], so Header file was not read.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	//cout << "StorageManager:: readHeader() Point D10-1G .." << endl << endl;

	if ( !readHeaderPerformed ) {

		// "DataHeader" was not found, b/c the file was not opened and/or found, so
		// set the values of the "hp" to default values, and
		// indicate that the "DataHeader" was _not_ found ..

		//header->setFilePageFieldCount( 0 );

		header->setTotalNumberOfDataPages( 0 );

		header->getHeaderEntries()->clear();

	}


	// DEBUGGING ..
	//cout << "SM:: readHeader() headerPageFound: " << headerPageFound << endl;
	//cout << "SM:: readHeader() end .." << endl;


	return readHeaderPerformed;


} // end "readHeader( DataHeader* header, string tableName )" ..


bool StorageManager:: makeDataPage( string tableName, int newPageID ) {

	// "tableName": name of the table for which the new data file-page will be written ..

	// "newPageID": index of the new data file-page ..

	// This method assumes that "createTable()" was called prior, because it wont work unless there is a table.
	// Also, this method is called in order to create an empty 48-byte file-page at the end of
	// an existing data file.

	// This occurs when a new record is being inserted into the data file,
	// but there are no existing file-pages that have empty rows which are available to accept the new row.
	// Therefore, "makeDataPage()" creates the new 48-byte file-page,
	// and fills the file-page with a "pad" string.

	bool makeDataPagePerformed = false;

	string emptyString = "";


	if ( emptyString.compare( tableName ) != 0 ) {
		
		//cout << "StorageManager:: makeDataPage() Point A .." << endl << endl;


		// Empty data page, initialized with 0 records
		//DataPage* dataPage = new DataPage();


		// Get the header page information ..
		// DataHeader should be updated each time a new page is allocated.
		DataHeader* header = new DataHeader();
		bool readHeaderPerformed = readHeader( tableName, header );

		// if reading header failed, can't allocate data page ..
		if( readHeaderPerformed == false ) {

	 		string errorString = "\nHeader file failed to open [ SM::makeDataPage() ], so data file was not written to.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );

			return false;

		}
		
		//cout << "StorageManager:: makeDataPage() Point B .." << endl << endl;

		// "makeDataPage" adds 1 more data page to the table ..
		header->setTotalNumberOfDataPages( header->getTotalNumberOfDataPages() + 1 );
		
		//cout << "StorageManager:: makeDataPage() Point C .." << endl << endl;

		// use header to calculate the max # of rows per file-page for this table ..
		// (not to be confused with "StorageManager:: calculateNumberOfFreeRows()") ..
		
		int fieldCount = 0;
			
		CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );

		if( catalogTable != NULL )
			fieldCount = catalogTable->getNumberOfAttributes();
				
		int maxRowCount = 12;
		
		if( maxRowCount >= 0 )
			maxRowCount = header->computeMaxRowCount( fieldCount );
			
		//cout << "StorageManager:: makeDataPage() Point D .." << endl << endl;

		DataHeaderEntry* dataHeaderEntry = new DataHeaderEntry( maxRowCount );
		dataHeaderEntry->setPageID( newPageID );
		dataHeaderEntry->setFreeRowsCount( maxRowCount );
		dataHeaderEntry->setFilledRowsCount( 0 );
		dataHeaderEntry->setLockNumber( 0 );
		// "rowAvailable" is initialized inside constructor ..

		header->getHeaderEntries()->push_back( dataHeaderEntry );
		
		//cout << "StorageManager:: makeDataPage() Point E .." << endl << endl;

		bool writeHeaderPerformed = writeHeader( tableName, header );

		// if writing header failed, can't allocate data page ..
		if( writeHeaderPerformed == false ) {

	 		string errorString = "\nFailed to write to Header file [ SM::allocateDataPage() ], so data file was not written to.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );

			return false;

		}
		
		//cout << "StorageManager:: makeDataPage() Point F .." << endl << endl;

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		directoryPlusFileName += tableName;
		char* fileName = (char*)directoryPlusFileName.c_str();


		// open the data file for this "tableName" ..
		//        "ios::binary": binary format (so file is not human-readable)
		// "ios::in | ios::out": even though this is strictly a "write" operation,
		//                       I found that this was necessary in order for this to work ..
		//fstream dataTable( tableName,  ios::in | ios::out | ios::binary );
		fstream dataTable( fileName,  ios::in | ios::out | ios::binary );
		//fstream dataTable( fileName,  ios::in | ios::out | ios::ate | ios::binary );
		//fstream dataTable( fileName,  ios::out | ios::binary );
		//ofstream dataTable( fileName,  ios::out | ios::binary );


		// this means the file doesn't exist yet,
		// so attempt to create the file ready for output ..
		if( dataTable.fail() ){

			// needed in order to "clear" errorflags left over from first "fail" ..
			dataTable.clear();

			dataTable.open( fileName, ios::out | ios::binary | ios::app );

		}
		
		//cout << "StorageManager:: makeDataPage() Point G .." << endl << endl;

		// only able to allocate the new page if the file opens ..
		if ( !dataTable.fail() ) {

			// make sure file pointer starts at beginning of file ..
			dataTable.seekp( ios::beg );

			// move the file pointer to the start of the new file-page ..
			dataTable.seekp( newPageID * blockSize );

			// DEBUGGING ..
			//checkFileState( dataTable );

			// write the "pad" string, in order to create the 48-byte data file-page ..
			string pad = constructPagePadding( blockSize );
			
			//cout << "StorageManager:: makeDataPage() tableName: " << tableName << ", newPageID: " << newPageID << ", pad(" << pad.length() << "): " << pad << endl << endl;

			const char* padChar = pad.c_str();

			int padCharSize = pad.size();

			//int charSize = 1;

			//for(int i = 0; i < pad.size(); i++){

				// DEBUGGING ..
				//cout << filler[i] << endl;

				//dataTable.write( reinterpret_cast< const char * >( &( pad[i] ) ), charSize );
				//dataTable.write( reinterpret_cast<const char*>( &padChar ), padCharSize );
				dataTable.write( reinterpret_cast<const char*>( padChar ), padCharSize );

			//}
			
			//cout << "StorageManager:: makeDataPage() Point H .." << endl << endl;

			// DEBUGGING ..
			//cout << endl << endl;

			// DEBUGGING ..
			//checkFileState( dataTable );

			// flag indicating the "makeDataPage()" operation
			// was performed successfully (tableName ok, file opened) ..
			makeDataPagePerformed = true;

		} else {

	 		string errorString = "\nData file failed to open [ SM::makeDataPage() ], so data file was not written to.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !dataTable.fail() )" ..

		dataTable.clear();
		dataTable.close();


	} else {

	 	string errorString = "\nData file table name was not valid [ SM::allocateDataPage() ], so data file was not written to.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	//cout << "StorageManager:: makeDataPage() Point J .." << endl << endl;

	return makeDataPagePerformed;


} // end "makeDataPage( int newPageID, string tableName )" ..


bool StorageManager:: writeDataPage( DataPage* newDataPage, BufferInfoTable* bit ) {

	//cout << "SM:: writeDataPage() start .." << endl;

	int blockIndex = BufferInfoTable::bitIndex;

	string tableName = bit->getTableName()->at( blockIndex );

	int rowCount = bit->getBufferPageRowCount()->at( blockIndex );

	int recordSize = bit->getRecordSize()->at( blockIndex );

	//int pageID = bit->getPageID()->at( blockIndex );
	int pageID = newDataPage->getPageID();
	
	int fieldCount = bit->getBufferPageFieldCount()->at( blockIndex );

	// DEBUGGING ..
	//cout << "StorageManager:: writeDataPage() blockIndex: " << blockIndex << endl;
	//cout << "StorageManager:: writeDataPage() tableName: " << tableName << endl;
	//cout << "StorageManager:: writeDataPage() rowCount: " << rowCount << endl;
	//cout << "StorageManager:: writeDataPage() recordSize: " << recordSize << endl;
	//cout << "StorageManager:: writeDataPage() pageID: " << pageID << endl;
	
	//newDataPage->printDataPage();

	DataHeader* header = new DataHeader();
	bool readHeaderPerformed = readHeader( tableName, header );

	// if reading header failed, can't write page ..
	if( readHeaderPerformed == false ) {

	 	string errorString = "\nFailed to read Header file [ SM::writeDataPage() ], so data page was not written to.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );

		return false;

	}

	//string directoryPlusTableName = "C:\\Temp\\";

	string directoryPlusTableName = databaseFile->getFileDrive();

	directoryPlusTableName += databaseFile->getFileDirectory();

	directoryPlusTableName += tableName;
	char* fileName = (char*)directoryPlusTableName.c_str();


	int dataSize = ( rowCount * recordSize );

	int padSize = ( blockSize - dataSize );

	// DEBUGGING ..
	//cout << "StorageManager:: writeDataPage() fileName: " << fileName << endl << endl;
	//cout << "StorageManager:: writeDataPage() padSize( blockSize("<<blockSize<<") - dataSize("<<dataSize<<") ): " << padSize << endl << endl;

	// need "readDataPage()" here, in order to save the data page state
	// prior to a write operation.  If "writeDataPage()" fails at any time,
	// entire transaction is cancelled, and the data page previous
	// state should be restored ..
	DataPage* currentDataPage = new DataPage();
	bool readCurrentPagePerformed = readDataPage( tableName, currentDataPage, pageID );

	// if "readDataPage()" failed, can't write page ..
	if( readCurrentPagePerformed == false ) {

	 	string errorString = "\nFailed to read current data page [ SM::writeDataPage() ], so data page was not written to.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );

		return false;

	}

	// Workaround:
	// although "writeDataPage()" is intended for data output to the file only,
	// I found that leaving out "ios::in" results in unwanted behavior,
	// basically b/c without "ios::in" the output data is inserted at the "file cursor",
	// and all the data _preceeding_ it is cleared / overwritten.
	// for some reason, "ios::in | ios::out" prevents this unwanted behavior,
	// allowing the desired "random access" (unix = rb+) writes ..
	//fstream dataTable( fileName, ios::out | ios::ate | ios::binary );
	fstream dataTable( fileName, ios::in | ios::out | ios::binary );

	//cout << "StorageManager:: readDataPage() blockIndex: " << blockIndex << endl << endl;

	//cout << "StorageManager:: writeDataPage() fileName: " << fileName << endl << endl;

	// this means the file doesn't exist yet,
	// so attempt to create the file ready for output ..
	// (this should have been taken care of in "makeDataPage") ..
	if( dataTable.fail() ){

		// needed in order to "clear" errorflags left over from first "fail" ..
		dataTable.clear();

		dataTable.open( fileName, ios::out | ios::binary | ios::app );

	}

	bool writePagePerformed = false;

	if ( !dataTable.fail() ) {

		// make sure file pointer starts at beginning of file ..
		dataTable.seekp( ios::beg );

		// Put DataHeader into its own file by itself,
		// so locate page-to-write referenced from beginning of file ..
		dataTable.seekp( pageID * blockSize );

		// DEBUGGING ..
		//cout << "pageID * blockSize: " << (pageID * blockSize) << endl << endl;
		//checkFileState( dataTable );

		int fieldValue = -1;
		
		//cout << "StorageManager:: writeDataPage()   rowCount: " << rowCount << endl << endl;
		
		//cout << "StorageManager:: writeDataPage() fieldCount: " << fieldCount << endl << endl;
		
		// DEBUGGING ..
		//cout << "StorageManager:: writeDataPage() specialValuePageId: " << pageID << endl << endl;

		// write the values in the DataPage to file ..
		for (int i=0; i < rowCount; i++) {

			for (int j=0; j < fieldCount; j++) {

				// DEBUGGING ..
				//cout << "StorageManager:: writeDataPage() newDataPage->records["<<i<<"].fields["<<j<<"].Value: " << newDataPage->records[i]->fields[j]->Value << endl << endl;

				//fieldValue = newDataPage->getRecords()->at( i )->getFields()->at( j )->getValue();
				fieldValue = newDataPage->getEntry( i, j );
				
				// DEBUGGING ..
				//if( pageID == 6 )
				//cout << "StorageManager:: writeDataPage() fieldValue: " << fieldValue << endl << endl;

				dataTable.write( reinterpret_cast<const char*>( &fieldValue ), intSize );

			}

		}

		if ( padSize > 0 ) {

			// DEBUGGING ..
			//cout << "StorageManager:: writeDataPage() blockSize: " << blockSize << endl << endl;
			//cout << "StorageManager:: writeDataPage() dataSize: " << dataSize << endl << endl;
			//cout << "StorageManager:: writeDataPage() fillerSize ( blockSize("<<blockSize<<") - dataSize("<<dataSize<<") ): " << fillerSize << endl << endl;

			string pad = constructPagePadding( padSize );
			
			//cout << "StorageManager:: writeDataPage() pad: " << pad << endl << endl;
			//cout << "StorageManager:: writeDataPage() tableName: " << tableName << ", pageID: " << pageID << ", pad(" << pad.length() << "): " << pad << endl << endl;

			int charSize = 1;

			// write the filler string values to fill up a 48-byte-block-DataPage to file ..
			for(int i=0; i<pad.size(); i++){

				//cout << filler[i];

				dataTable.write( reinterpret_cast<const char*>( &( pad[i] ) ), charSize );

			}

			//cout << endl << endl;

		}


		// flag indicating the "writeDataPage()" operation
		// was performed successfully (header read, file created, file opened) ..
		writePagePerformed = true;


	} else {

	 	string errorString = "\nFailed to open data file [ SM::writePage() ], so data page was not written to.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( !tbl.fail() )" ..

	// DEBUGGING ..
	//checkFileState( tbl );

	dataTable.clear();
	dataTable.close();


	// DEBBUGGING ..
	//cout << "StorageManager:: writeDataPage() header.totalNumberOfDataPages: " << header.totalNumberOfDataPages << endl;
	//cout << "StorageManager:: writeDataPage() rowCount: " << rowCount << endl;
	//cout << "StorageManager:: writeDataPage() header.filePageFieldCount: " << header.filePageFieldCount << endl;

	//cout << "IndexPages:: writeDataPage() Point A .." << endl << endl;

	int freeRowsCount = calculateNumberOfFreeRows( rowCount, fieldCount );

	int savePreviousRowCount = -1;
	
	//int numberOfDataPages = header->getTotalNumberOfDataPages();
	
	
	// need to check whether or not the current entry
	// already exists in the file ..
	vector<DataHeaderEntry*>* headerEntries = header->getHeaderEntries();
	
	bool isHeaderEntryFound = false;
	int headerEntryIndex = -1;
	
	int numberOfEntries = headerEntries->size();
	for(int i=0; i<numberOfEntries; i++){
		
		if( headerEntries->at(i)->getPageID() == pageID ){
			
			isHeaderEntryFound = true;
			headerEntryIndex = i;
			break;
			
		}
		
	}
	
	//cout << "IndexPages:: writeDataPage() Point B .." << endl << endl;
	
	if( isHeaderEntryFound ){
		
		//cout << "IndexPages:: writeDataPage() Point C1 .." << endl << endl;
		
		// the entry already exists in the file,
		// so we just need to update the one that already exists ..
		
		//cout << "StorageManager:: writeDataPage() freeRowsCount: " << freeRowsCount << endl << endl;

		headerEntries->at( headerEntryIndex )->setFreeRowsCount( freeRowsCount );
		
		//cout << "IndexPages:: writeDataPage() Point C2 .." << endl << endl;

		// find "bit" info to write to "DataHeader" ..

		if ( ( blockIndex > -1 ) &&
			 ( blockIndex < bit->getBufferPoolSize() ) ) {
			 	
			//cout << "IndexPages:: writeDataPage() Point C3 .." << endl << endl;

			// save the "previousRowCount" for use in case "writeHeader()" fails ..
			savePreviousRowCount = headerEntries->at( headerEntryIndex )->getFilledRowsCount();
			
			//cout << "IndexPages:: writeDataPage() previousFilledRowsCount: " << savePreviousRowCount << endl << endl;
			
			//cout << "IndexPages:: writeDataPage() Point C4 .." << endl << endl;

			// set "HP" entry fields to the values in "bit" ..
			headerEntries->at( headerEntryIndex )->setFilledRowsCount( rowCount );
			
			//cout << "StorageManager:: writeDataPage()   freeRowsCount: " << freeRowsCount << endl << endl;
			//cout << "StorageManager:: writeDataPage() filledRowsCount: " << rowCount << endl << endl;

			//cout << "IndexPages:: writeDataPage() Point C5 .." << endl << endl;
/*
			// DEBUGGING ..
			//cout << "StorageManager:: writeDataPage() bit.bufferPageRowCount[ bitIndex("<<blockIndex<<") ]: " << bit.bufferPageRowCount[ blockIndex ] << endl << endl;

			//cout << "IndexPages:: writeDataPage() rowCount: " << rowCount << endl << endl;
			//cout << "IndexPages:: writeDataPage() headerEntryIndex: " << headerEntryIndex << endl << endl;
			//cout << "IndexPages:: writeDataPage() blockIndex: " << blockIndex << endl << endl;
			//cout << "IndexPages:: writeDataPage() headerEntries-pageID: " << headerEntries->at( headerEntryIndex )->getPageID() << endl << endl;
			//cout << "IndexPages:: writeDataPage() bit-pageID: " << pageID << endl << endl;
			//cout << "IndexPages:: writeDataPage() headerEntries-size: " << headerEntries->at( headerEntryIndex )->getRowAvailable()->size() << endl << endl;
			//cout << "IndexPages:: writeDataPage() bit-size: " << bit->getRowAvailable()->at( blockIndex )->size() << endl << endl;

			for(int i=0; i<headerEntries->at( headerEntryIndex )->getRowAvailable()->size(); i++){
				cout << "IndexPages:: writeDataPage() headerEntries-getRowAvailable: " << headerEntries->at( headerEntryIndex )->getRowAvailable()->at( i ) << endl << endl;
			}

			for(int i=0; i<bit->getRowAvailable()->at( blockIndex )->size(); i++){
				cout << "IndexPages:: writeDataPage() bit-getRowAvailable: " << bit->getRowAvailable()->at( blockIndex )->at( i ) << endl << endl;
			}
*/

			// assuming that "headerEntries" is smaller than "bit" ..
			//int lessSlots = headerEntries->at( headerEntryIndex )->getRowAvailable()->size();
			
			//int moreSlots = bit->getRowAvailable()->at( blockIndex )->size();

			// make sure that the min is what we expect ..
			//int numberSlots = min( lessSlots, moreSlots );
			
			//cout << "IndexPages:: writeDataPage() headerEntries-numberSlots: " << numberSlots << endl << endl;

			for(int i=0; i<rowCount; i++){
			//for(int i=0; i<numberSlots; i++){
				
				//cout << "IndexPages:: writeDataPage() i: " << i << endl << endl;

				headerEntries->at( headerEntryIndex )->getRowAvailable()->at( i ) = bit->getRowAvailable()->at( blockIndex )->at( i );

				// DEBUGGING ..
				//cout << "StorageManager:: writeDataPage() bit.rowAvailable[ bitIndex("<<blockIndex<<") ]["<<j<<"]: " << bit.rowAvailable[ blockIndex ][j] << endl << endl;

			}
			
			//cout << "IndexPages:: writeDataPage() bit-moreSlots: " << moreSlots << endl << endl;
			
			//for(int i=numberSlots; i<moreSlots; i++){
				
				//cout << "IndexPages:: writeDataPage() i: " << i << endl << endl;

			//	headerEntries->at( headerEntryIndex )->getRowAvailable()->push_back( bit->getRowAvailable()->at( blockIndex )->at( i ) );

				// DEBUGGING ..
				//cout << "StorageManager:: writeDataPage() bit.rowAvailable[ bitIndex("<<blockIndex<<") ]["<<j<<"]: " << bit.rowAvailable[ blockIndex ][j] << endl << endl;

			//}
			
			//cout << "IndexPages:: writeDataPage() Point C6 .." << endl << endl;
			
			// release the lock on this page ..
			headerEntries->at( headerEntryIndex )->setLockNumber( 0 );

		}
		
		//cout << "IndexPages:: writeDataPage() Point D .." << endl << endl;
		
	} else {
		
		// the entry does not yet exist in the file,
		// so we need to add a new entry ..
		
		int maxRowCount = bit->computeMaxRowCount_Attributes( fieldCount );
		
		DataHeaderEntry* headerEntry = new DataHeaderEntry( maxRowCount );
		headerEntry->setPageID( pageID );
		headerEntry->setFreeRowsCount( freeRowsCount );
		headerEntry->setFilledRowsCount( rowCount );
		
		vector<bool>* rowAvailable = new vector<bool>();
		for(int i=0; i<rowCount; i++)
			rowAvailable->push_back( bit->getRowAvailable()->at( blockIndex )->at( i ) );
		
		headerEntry->setRowAvailable( rowAvailable );
		headerEntry->setLockNumber( 0 );
		
		headerEntries->push_back( headerEntry );
		
		// increment the number of data pages,
		// since a new page was just written to file ..
		header->setTotalNumberOfDataPages( header->getTotalNumberOfDataPages() + 1 );
		
	}
	
/*
	for (int i = 0; i<numberOfDataPages; i++) {

		// DEBBUGGING ..
		//cout << "SM:: writeDataPage() hp.headerPageEntries["<<i<<"].PageID: " << hp._headerEntries[i].pageID << endl << endl;

		if ( pageID == header->getHeaderEntries()->at( i )->getPageID() ) {

			cout << "StorageManager:: writeDataPage() freeRowsCount: " << freeRowsCount << endl << endl;

			header->getHeaderEntries()->at( i )->setFreeRowsCount( freeRowsCount );

			// find "bit" info to write to "DataHeader" ..

			if ( ( blockIndex > -1 ) &&
			     ( blockIndex < bit->getBufferPoolSize() ) ) {

				// release the lock on this page ..
				header->getHeaderEntries()->at( i )->setLockNumber( 0 );

				// save the "previousRowCount" for use in case "writeHeader()" fails ..
				savePreviousRowCount = header->getHeaderEntries()->at( i )->getFilledRowsCount();

				// set "HP" entry fields to the values in "bit" ..
				header->getHeaderEntries()->at( i )->setFilledRowsCount( bit->getBufferPageRowCount()->at( blockIndex ) );

				// DEBUGGING ..
				//cout << "StorageManager:: writeDataPage() bit.bufferPageRowCount[ bitIndex("<<blockIndex<<") ]: " << bit.bufferPageRowCount[ blockIndex ] << endl << endl;

				for(int j=0; j<totalNumberOfRows; j++){

					header->getHeaderEntries()->at( i )->getRowAvailable()->at( j ) = ( bit->getRowAvailable()->at( blockIndex ) )->at( j );

					// DEBUGGING ..
					//cout << "StorageManager:: writeDataPage() bit.rowAvailable[ bitIndex("<<blockIndex<<") ]["<<j<<"]: " << bit.rowAvailable[ blockIndex ][j] << endl << endl;

				}

			}

			// each "pageID" is unique (there should be no duplicate pageID in a headerEntry),
			// so once the pageID is located, finish tracking data corresponding to this pageID ..
			break;


		} // end "if ( pg.pageID == hp._headerEntries[i].pageID )" ..

	} // end "for()" ..
*/

	bool isWriteHeaderPerformed = writeHeader( tableName, header );



	// if writing header failed, can't write page,
	// so restore the data page to its previous value ..
	if( isWriteHeaderPerformed == false ) {


		int previousRowCount = savePreviousRowCount;

		int previousFieldCount = 0;
			
		CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );

		if( catalogTable != NULL )
			previousFieldCount = catalogTable->getNumberOfAttributes();
		
		//int previousFieldCount = header->getFilePageFieldCount();

		// Put DataHeader into its own file by itself,
		// so locate page-to-write referenced from beginning of file ..
		dataTable.seekp( pageID * blockSize );

		// DEBUGGING ..
		//cout << "pageID * blockSize: " << (pageID * blockSize) << endl << endl;
		//checkFileState( tbl );

		int fieldValue = -1;

		// write the values in the DataPage to file ..
		//for (int i=0; i < rowCount; i++) {
		for (int i=0; i < previousRowCount; i++) {

			//for (int j=0; j < fieldCount; j++) {
			for (int j=0; j < previousFieldCount; j++) {

				// DEBUGGING ..
				//cout << "StorageManager:: writeDataPage() currentDataPage.records["<<i<<"].fields["<<j<<"].Value: " << pg.records[i].fields[j].Value << endl << endl;

				//fieldValue = currentDataPage->getRecords()->at( i )->getFields()->at( j )->getValue();
				
				fieldValue = newDataPage->getEntry( i, j );

				dataTable.write( reinterpret_cast<const char*>( &fieldValue ), intSize );

			}

		}

		int previousDataSize = ( previousRowCount * recordSize );

		int previousPaddingSize = ( blockSize - previousDataSize );

		if ( previousPaddingSize > 0 ) {

			// DEBUGGING ..
			//cout << "StorageManager:: writeDataPage() blockSize: " << blockSize << endl << endl;
			//cout << "StorageManager:: writeDataPage() dataSize: " << dataSize << endl << endl;
			//cout << "StorageManager:: writeDataPage() paddingSize ( blockSize("<<blockSize<<") - dataSize("<<dataSize<<") ): " << paddingSize << endl << endl;

			//string pad = constructPagePadding( paddingSize );
			string pad = constructPagePadding( previousPaddingSize );

			int charSize = 1;

			// write the pad string values to fill up a 48-byte-block-DataPage to file ..
			for(int i=0; i<pad.size(); i++){

				//cout << filler[i];

				dataTable.write( reinterpret_cast<const char*>( &( pad[i] ) ), charSize );

			}

			//cout << endl << endl;

		}


	 	string errorString = "\nFailed to write Header file [ SM::writeDataPage() ], so data page was not written to.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );


		return false;

	}



	// NOTE: this is a problem [RESOLVED], because if "dataTable.write()" was
	//       successful above, then if "writeHeader()" here fails,
	//       how do we UNDO "dataTable.write()"??  need transaction??
	//       assume not a problem? need a way to store previous state
	//		 so that if the entire transaction isn't completed then the
	//		 datafile can be restored .. 
	//       OK - this is done in "duplicated" code above ..
	//		 need to change this to function calls ??

	//cout << "SM:: writeDataPage() end .." << endl;

	return writePagePerformed;

} // end "writeDataPage( DataPage* newDataPage, BufferInfoTable* bit )" ..


bool StorageManager:: readDataPage( string tableName, DataPage* dataPage, BufferInfoTable* bit ) {


	//cout << "StorageManager:: readDataPage() start .." << endl << endl;


	bool readPagePerformed = false;

	string emptyString = "";


	if ( emptyString.compare( tableName ) != 0 ) {


		DataHeader* header = new DataHeader();
		bool readHeaderPerformed = readHeader( tableName, header );

		// DEBUGGING ..
		//cout << "StorageManager:: readDataPage() readHeaderPerformed: " << readHeaderPerformed << endl << endl;

		// if reading header failed, can't read page ..
		if( readHeaderPerformed == false ) {

	 		string errorString = "\nFailed to read Header file [ SM::readDataPage() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );

			return false;

		}


		int blockIndex = BufferInfoTable::bitIndex;

		//int pageID = bit->getPageID()->at( blockIndex );
		int pageID = dataPage->getPageID();


		// DEBUGGING ..
		//cout << "StorageManager:: readDataPage() pageID: " << pageID << endl << endl;
		//cout << "StorageManager:: readDataPage() blockIndex: " << blockIndex << endl << endl;


		//string directoryPlusTableName = "C:\\Temp\\";

		string directoryPlusTableName = databaseFile->getFileDrive();

		directoryPlusTableName += databaseFile->getFileDirectory();

		directoryPlusTableName += tableName;
		char* fileName = (char*)directoryPlusTableName.c_str();

		//fstream dataTable(tableName,  ios::in );
		//fstream dataTable( tableName,  ios::in | ios::binary );
		fstream dataTable( fileName,  ios::in | ios::binary );
		//ifstream dataTable( fileName, ios::binary );
		
		// if the constructor failed, give it a second shot ..
		if( dataTable.fail() ){

			// needed in order to "clear" errorflags left over from first "fail" ..
			dataTable.clear();

			dataTable.open( fileName, ios::in | ios::binary );

		}
		
		//cout << "StorageManager:: readDataPage() blockIndex: " << blockIndex << endl << endl;
		
		//cout << "StorageManager:: readDataPage() fileName: " << fileName << endl << endl;

		int headerEntryIndex = -1;
		
		int fieldCount = 0;
			
		CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );

		if( catalogTable != NULL )
			fieldCount = catalogTable->getNumberOfAttributes();

		if ( !dataTable.fail() ) {

			// DataHeader is written into its own file (by itself)
			// so locate "page-to-read", referenced from beginning of file ..
			dataTable.seekg( ( pageID * blockSize ) );

			// DEBUGGING ..
			//cout << "StorageManager:: readDataPage() (pageID * blockSize): " << (pageID * blockSize) << endl << endl;
			//cout << "StorageManager:: readDataPage() header.totalNumberOfDataPages: " << header.totalNumberOfDataPages << endl << endl;

			// DEBUGGING ..
			//checkFileState( tbl );


			int rowCount = 0;
			//int fieldCount = 0;

			for(int i=0; i<header->getTotalNumberOfDataPages(); i++){

				// DEBUGGING ..
				//cout << "StorageManager:: readDataPage() pageID("<<pageID<<") == header._headerEntries["<<i<<"].pageID("<<header._headerEntries[i].pageID<<"): " << ( pageID == header._headerEntries[i].pageID ) << endl << endl;

				if ( pageID == header->getHeaderEntries()->at( i )->getPageID() ){

					rowCount = header->getHeaderEntries()->at( i )->getFilledRowsCount();

					//fieldCount = header->getFilePageFieldCount();

					headerEntryIndex = i;

					// DEBUGGING ..
					//cout << "StorageManager:: readDataPage() rowCount: " << rowCount << endl << endl;
					//cout << "StorageManager:: readDataPage() fieldCount: " << fieldCount << endl << endl;

					break;

				}

			}


/*
			dataPage->getRecords()->clear();

			for(int i=0; i < rowCount; i++) {

				dataPage->getRecords()->push_back( ( *new DataPageRecord() ) );

			}

			int fileValue = -1;


			for (int i=0; i < rowCount; i++) {

				dataPage->getRecords()->at( i )->getFields()->clear();

				for(int j=0; j < fieldCount; j++) {

					dataPage->getRecords()->at( i )->getFields()->push_back( *new DataPageField() );

				}

				for (int k =0; k < fieldCount; k++) {

					dataTable.read( (char *) ( &( fileValue ) ), intSize );

					dataPage->getRecords()->at( i )->getFields()->at( k ).setValue( fileValue );

					// DEBUGGING ..
					//cout << "StorageManager:: readDataPage() fileValue: " << fileValue << endl;
					//cout << "StorageManager:: readDataPage() dataPage.records["<<i<<"].fields["<<k<<"].Value: " << dataPage.records[i].fields[k].Value << endl << endl;

				}

			}
			
*/

			// make sure that the dataPage (to read into) is empty ..
			dataPage->clearPage();

			dataPage->setPageID( pageID );
			dataPage->setMaxPageSize( 48 );

			int fileValue = -1;
			
			for (int i=0; i < rowCount; i++) {

				for (int i=0; i < fieldCount; i++) {

					dataTable.read( reinterpret_cast<char *>( &fileValue ), intSize );
	
					dataPage->addEntry( fileValue );

					// DEBUGGING ..
					//cout << "StorageManager:: readDataPage() fileValue: " << fileValue << endl;
					//cout << "StorageManager:: readDataPage() dataPage.records["<<i<<"].fields["<<k<<"].Value: " << dataPage.records[i].fields[k].Value << endl << endl;

				}
				
				dataPage->toggleNewRow();
				
			}

			readPagePerformed = true;


		} else {

	 		string errorString = "\nFailed to open data file [ SM::readDataPage() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		}  // end "if ( !tbl.fail() )" ..

		// DEBUGGING ..
		//checkFileState( tbl );

		dataTable.clear();
		dataTable.close();

		int maxSlotsAvailable = header->computeMaxRowCount( fieldCount );

		// DEBUGGING ..
		//cout << "StorageManager:: readDataPage() maxSlotsAvailable: " << maxSlotsAvailable << endl << endl;


		if ( ( blockIndex > -1 ) &&
		     ( headerEntryIndex > -1 ) &&
			 ( blockIndex < bit->getBufferPoolSize() ) ) {

			// set "bit" fields to the values in "DataHeader" entry ..
			bit->getBufferPageRowCount()->at( blockIndex ) = header->getHeaderEntries()->at( headerEntryIndex )->getFilledRowsCount();

			// DEBUGGING ..
			//cout << "StorageManager:: readDataPage() hp.headerPageEntries["<<headerEntryIndex<<"].filledRowsCount: " << hp._headerEntries[ headerEntryIndex ].filledRowsCount << endl;

			for(int j=0; j<maxSlotsAvailable; j++) {

				bit->getRowAvailable()->at( blockIndex )->at( j ) = header->getHeaderEntries()->at( headerEntryIndex )->getRowAvailable()->at( j );

				// DEBUGGING ..
				//cout << "StorageManager:: readDataPage() hp.headerPageEntries["<<headerEntryIndex<<"].rowAvailable["<<j<<"]: " << hp._headerEntries[ headerEntryIndex ].rowAvailable[j] << endl;

			}

		} // end "if ( ( blockIndex > -1 ) &&" ..


	} // end "if ( emptyString.compare( tableName ) != 0 )" ..


	return readPagePerformed;


} // end "bool StorageManager:: readDataPage( DataPage* dataPage, BufferInfoTable &bit, string tableName )" ..


bool StorageManager:: readDataPage( string tableName, DataPage* dataPage, int pageID ) {

	// This method is essentially the same as "readDataPage()", except that
	// "bit.getBufferPageRowCount()" and "bit.getRowAvailable()"
	// are not updated ..

	// This method is used by "writeDataPage()" in order to save the state of the page-from-file
	// prior to performing the "writeDataPage()" operation on the page-from-file; if "writeDataPage()"
	// fails at any point, the operation is treated like an atomic transaction and rolled back.
	// In other words, if the page-from-file was changed during the failed "writeDataPage()" operation,
	// the info saved using data returned from the "readCurrentPage" is used to reset the
	// page-from-file to its previous state ..


	//cout << "StorageManager:: readCurrentPage() start .." << endl << endl;


	bool readPagePerformed = false;

	string emptyString = "";


	if ( emptyString.compare( tableName ) != 0 ) {

		DataHeader* header = new DataHeader();
		bool readHeaderPerformed = readHeader( tableName, header );

		// DEBUGGING ..
		//cout << "StorageManager:: readCurrentPage() readHeaderPerformed: " << readHeaderPerformed << endl << endl;
		//cout << "StorageManager:: readCurrentPage() tableName: " << tableName << endl << endl;

		// if reading header failed, can't read page ..
		if( readHeaderPerformed == false ) {

	 		string errorString = "\nFailed to read Header file [ SM::readCurrentPage() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );

			return false;

		}

		// DEBUGGING ..
		//cout << "StorageManager:: readDataPage() pageID: " << pageID << endl << endl;

		//string directoryPlusTableName = "C:\\Temp\\";

		string directoryPlusTableName = databaseFile->getFileDrive();

		directoryPlusTableName += databaseFile->getFileDirectory();

		directoryPlusTableName += tableName;
		char* fileName = (char*)directoryPlusTableName.c_str();

		//fstream dataTable(tableName,  ios::in );
		//fstream dataTable( tableName,  ios::in | ios::binary );
		fstream dataTable( fileName,  ios::in | ios::binary );
		//ifstream dataTable( fileName, ios::binary );

		// if the constructor failed, give it a second shot ..
		if( dataTable.fail() ){

			// needed in order to "clear" errorflags left over from first "fail" ..
			dataTable.clear();

			dataTable.open( fileName, ios::in | ios::binary );

		}

		int headerEntryIndex = -1;
		
		int fieldCount = 0;
			
		CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );

		if( catalogTable != NULL )
			fieldCount = catalogTable->getNumberOfAttributes();

		if ( !dataTable.fail() ) {


			// DataHeader is written into its own file (by itself)
			// so locate "page-to-read", referenced from beginning of file ..
			dataTable.seekg( ( pageID * blockSize ) );

			// DEBUGGING ..
			//cout << "StorageManager:: readDataPage() (pageID * blockSize): " << (pageID * blockSize) << endl << endl;
			//cout << "StorageManager:: readDataPage() header.totalNumberOfDataPages: " << header.totalNumberOfDataPages << endl << endl;

			// DEBUGGING ..
			//checkFileState( tbl );


			int rowCount = 0;
			//int fieldCount = 0;

			for(int i=0; i<header->getTotalNumberOfDataPages(); i++){

				// DEBUGGING ..
				//cout << "StorageManager:: readDataPage() pageID("<<pageID<<") == header._headerEntries["<<i<<"].pageID("<<header._headerEntries[i].pageID<<"): " << ( pageID == header._headerEntries[i].pageID ) << endl << endl;

				if ( pageID == header->getHeaderEntries()->at( i )->getPageID() ){

					rowCount = header->getHeaderEntries()->at( i )->getFilledRowsCount();

					//fieldCount = header->getFilePageFieldCount();

					headerEntryIndex = i;

					// DEBUGGING ..
					//cout << "StorageManager:: readDataPage() rowCount: " << rowCount << endl << endl;
					//cout << "StorageManager:: readDataPage() fieldCount: " << fieldCount << endl << endl;

					break;

				}

			}



/*
			dataPage->getRecords()->clear();

			for(int i=0; i < rowCount; i++) {

				dataPage->getRecords()->push_back( ( *new DataPageRecord() ) );

			}

			int fileValue = -1;


			for (int i=0; i < rowCount; i++) {

				dataPage->getRecords()->at( i )->getFields()->clear();

				for(int j=0; j < fieldCount; j++) {

					dataPage->getRecords()->at( i )->getFields()->push_back( *new DataPageField() );

				}

				for (int k =0; k < fieldCount; k++) {

					dataTable.read( (char *) ( &( fileValue ) ), intSize );

					dataPage->getRecords()->at( i )->getFields()->at( k ).setValue( fileValue );

					// DEBUGGING ..
					//cout << "StorageManager:: readDataPage() fileValue: " << fileValue << endl;
					//cout << "StorageManager:: readDataPage() dataPage.records["<<i<<"].fields["<<k<<"].Value: " << dataPage.records[i].fields[k].Value << endl << endl;

				}

			}
			
*/

			dataPage->setPageID( pageID );
			dataPage->setMaxPageSize( 48 );
		
			int fileValue = -1;
			
			for (int i=0; i < rowCount; i++) {

				for (int i=0; i < fieldCount; i++) {

					dataTable.read( reinterpret_cast<char *>( &fileValue ), intSize );

					dataPage->addEntry( fileValue );

					// DEBUGGING ..
					//cout << "StorageManager:: readDataPage() fileValue: " << fileValue << endl;
					//cout << "StorageManager:: readDataPage() dataPage.records["<<i<<"].fields["<<k<<"].Value: " << dataPage.records[i].fields[k].Value << endl << endl;

				}
				
			}


			readPagePerformed = true;


		} else {

	 		string errorString = "\nFailed to open data file [ SM::readCurrentPage() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..

		// DEBUGGING ..
		//checkFileState( tbl );

		dataTable.clear();
		dataTable.close();


	} // end "if ( emptyString.compare( tableName ) != 0 )" ..


	return readPagePerformed;


} // end "readDataPage( DataPage* dataPage, int pageID, string tableName )" ..

/*
 * GETTERS
 */


DatabaseFile* StorageManager:: getDatabaseFile(){

	return this->databaseFile;

}


void StorageManager:: checkFileState( fstream &dataTable ){

	// DEBUGGING method .. doesn't seem to work effectively,
	// always outputs "dataTable.fail() == true .. ",
	// even when the file successully opens .. NOT USED ..

	if( dataTable.rdstate() ){
		if ( dataTable.fail() )
			cout << "dataTable.fail() == true .. " << endl << endl;
		else if ( dataTable.good() )
			cout << "dataTable.good() == true .. " << endl << endl;
		else if ( dataTable.bad() )
			cout << "dataTable.bad() == true .. " << endl << endl;
		else if ( dataTable.eof() )
			cout << "dataTable.eof() == true .. " << endl << endl;
	}

} // end "checkFileState()" ..

bool StorageManager:: readBatchFile( string fileName, vector<string>* commands ) {


	//cout << "StorageManager:: readBatchFile() start .." << endl << endl;
	

	if( commands == NULL )
		return false;

	vector<string>* parsedCommands = NULL;

	bool readBatchFile = false;
	
	char* fileBuffer;

	string emptyString = "";


	if ( emptyString.compare( fileName ) != 0 ) {


		string directoryPlusTableName = databaseFile->getFileDrive();

		directoryPlusTableName += databaseFile->getFileDirectory();

		directoryPlusTableName += fileName;
		
		char* batchFileName = (char*)directoryPlusTableName.c_str();
		
		//cout << "StorageManager:: readBatchFile() batchFileName: " << batchFileName << endl << endl;

		fstream batchFile( batchFileName,  ios::in | ios::out );
		
		if ( batchFile.fail() ) {
			
			// needed in order to "clear" errorflags left over from first "fail" ..
			batchFile.clear();
			
			// if the constructor failed to open a file,
			// try again, but this time with the "open()" method ..
			batchFile.open( batchFileName,  ios::in | ios::out );
			
		}

		if ( !batchFile.fail() ) {

			// DataHeader is written into its own file (by itself)
			// so locate "page-to-read", referenced from beginning of file ..
			batchFile.seekg( 0, ios::end );
			int fileLength = batchFile.tellg();
			batchFile.seekg( 0, ios::beg );
			
			fileBuffer = new char[ fileLength ];
			
			//cout << "StorageManager:: readBatchFile() fileLength: " << fileLength << endl << endl;
			
			//cout << "StorageManager:: readBatchFile() fileBuffer: " << ( (string) fileBuffer ) << endl << endl;

			batchFile.read( fileBuffer, fileLength );

			// parse the "fileBuffer" into individual "string" commands .. 
			parsedCommands = processBatchFileChars( fileBuffer, fileLength );
			
			if( ( commands != NULL ) &&
			    ( parsedCommands != NULL ) ){
				
				// add the "parsedCommands" into the "commands" collection ..
				
				int parsedCommandsSize = parsedCommands->size();
				for(int i=0; i<parsedCommandsSize; i++){
					
					commands->push_back( parsedCommands->at(i) );
					
				}
				
			}
			
			readBatchFile = true;


		} else {

	 		string errorString = "\nFailed to open batch file [ SM::readBatchFile() ], so batch file was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		}  // end "if ( !batchFile.fail() )" ..

		batchFile.clear();
		batchFile.close();


	} // end "if ( emptyString.compare( tableName ) != 0 )" ..


	return readBatchFile;


} // end "bool StorageManager:: readBatchFile()" ..


vector<string>* StorageManager:: processBatchFileChars( char* fileBuffer, int fileLength ){
	
	//cout << "StorageManager:: processBatchFileChars() start .." << endl << endl;
	
	// helper method for "readBatchFile()" ..
	
	string batchFileCommand = "";
	
	vector<string>* batchFileCommands = new vector<string>();
	
	for(int i=0; i<fileLength; i++){
		
		if( ( fileBuffer[i] == '\n' ) || 								// linux ..
			( fileBuffer[i-1] == '\r' ) && ( fileBuffer[i] == '\n' ) ){	// windows ..
			
			// DEBUGGING ..
			//if( batchFileCommand.size() > 0 )
			//	cout << "StorageManager:: processBatchFileChars() batchFileCommand: " << batchFileCommand << endl << endl;
			
			if( batchFileCommand.size() > 0 )
				batchFileCommands->push_back( batchFileCommand );
			
			// clear the temp parameter for the next command ..
			batchFileCommand = "";
			
		}
		
		// add the next char of this command to the temp parameter,
		// avoiding "\r" and "\n" characters ..
		if( ( fileBuffer[i] != '\r' ) && ( fileBuffer[i] != '\n' ) )
			batchFileCommand += fileBuffer[i];
		
	}
	
	return batchFileCommands;
	
}

