#ifndef STORAGEMANAGER_H_
#define STORAGEMANAGER_H_

#pragma once

#include "../DataPage/DataPage.h"
#include "../DataHeader/DataHeader.h"
#include "../Buffer/BufferInfoTable.h"
#include "../Utilities/String_Utilities.h"
#include "../Database/DatabaseFile.h"

#include "CatalogManager.h"
#include "MessageManager.h"

#include <string>
#include <map>
#include <vector>
#include <queue>

#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>

using namespace std;

class StorageManager {

public:
	StorageManager();
	StorageManager( CatalogManager* cm_, DatabaseFile* databaseFile_ );
	~StorageManager(void);

	// data members ..



	// member methods ..

	void setDatabaseFile( DatabaseFile* databaseFile_ );

	CatalogManager* getCatalogManager();

	MessageManager* getMessageManager();
	
	string constructPagePadding( int fillerSize );

	bool createTable( string tableName, int fieldCount, vector<string>* attributeNames );

	void activateTablePageLock( string tableName, int pageID, int lockNumberIncrement );
	void deactivateTablePageLock( string tableName, int pageID );

	bool isTableHasLocks( string tableName );

	vector<int> getTablePageLocks( string tableName );
	int getTablePageLock( string tableName, int pageID );
	
	void clearAllTablePageLocks( string tableName );
	void clearTablePageLocks( string tableName, int* pageIDs, int intArraySize );
	void clearTablePageLocks( string tableName, vector<int>* pageIDs );

	bool readHeader( string tableName, DataHeader* hp );
	bool writeHeader( string tableName, DataHeader* hp );

	bool makeDataPage( string tableName, int pageID );

	bool writeDataPage( DataPage* pg, BufferInfoTable* bit );
	bool readDataPage( string tableName, DataPage* dataPage, BufferInfoTable* bit );
	bool readDataPage( string tableName, DataPage* dataPage, int pageID );

	DatabaseFile* getDatabaseFile();

	void checkFileState( fstream &tbl );
	
	bool readBatchFile( string fileName, vector<string>* commands );
	vector<string>* processBatchFileChars( char* fileBuffer, int fileLength );


private:

	// data members ..

	static const int blockSize = 48;
	static const int intSize = sizeof(int);
	static const int boolSize = sizeof(bool);
	static const int fanout = 5;
	
	int fieldCount;

	// Tracks metadata stored in "Catalog.bin" file ..
	CatalogManager* cm;

	// contains file directory information ..
	DatabaseFile* databaseFile;				

	int headerSize;


	// member methods ..

	int calculateNumberOfFreeRows( int recordCountPerPage, int fieldCount );

	//void copyHeaderEntries( DataHeaderEntry oldSet[], int oldSetLength, DataHeaderEntry newSet[] );

};

#endif /*STORAGEMANAGER_H_*/
