#ifndef TESTER_H_
#define TESTER_H_

#pragma once

#include "../UserInterface/UserInterface.h"

#include "../Catalog/CatalogTable.h"
#include "../Catalog/CatalogTableAttribute.h"

#include "../Index/IndexBucketHeader.h"
#include "../Index/IndexHeader.h"
#include "../Index/IndexBucket.h"
#include "../Index/IndexNodeRow.h"
#include "../Index/IndexNode.h"
#include "../Index/IndexPage.h"
#include "../Index/IndexRows.h"


#include "../Managers/CatalogManager.h"

#include <iostream>
#include <vector>
#include <cmath>
#include <algorithm>

class Tester {
	
public:

	Tester();
	~Tester();
	
	void writeMetaData();
	
	void readMetaData();
	
	void testPowerOperator();
	
	void createIndexSearchKeys();
	
	void createIndexPages();
	
	bool writeIndexBucket();
	
	void writeIndexBuckets();
	
	bool writeIndexPage( char* tableName_, char* indexName_, IndexPage* indexPage_, int pageID_, int parentPageID_, bool isNew_ );
	
	bool writeIndexPages();
	
	bool writeIndexBucketHeader( char* tableName_, char* indexName_, IndexBucketHeader& indexBucketHeader_ );
	
	bool writeIndexHeader( char* tableName_, char* indexName_, IndexHeader& header_ );
	
	string createFillerString( int fillerSize );
	
	bool readIndexPage( char* tableName_, char* indexName_, IndexPage* indexPage_, int pageID_ );
	
	bool readIndexPages( vector<IndexPage*>* indexPages_ );
	
	bool readIndexHeader( char* tableName, char* indexName_, IndexHeader& header );
	
	bool readIndexBucketHeader( char* tableName_, char* indexName_, IndexBucketHeader& indexBucketHeader_ );
	
	bool readIndexBucket( char* tableName_, char* indexName_, IndexBucket* indexBucket_, int bucketPageID_ );
	
	void test_write();
	
	void test_read();
	
	int getNumberOfUniqueRows();
	
	void test_insert( UserInterface* ui, string tableName, int numberOfUniqueRows, int numberRepeatRows, int multiple, int numberOfAttributes );
	
	void test_createTables_insert_select();
	
	void test_createTable_createIndex_insert_printIndex();
	
	void test_createTable_insert_createIndex_printIndex();
	
	void test_createTable_insert_createIndex_selectData();
	
	void printMessages( UserInterface* ui );
	
	void test_testData();
	
private:

	// contains all the rows of searchKeys ..
	// first row are the leaf searchKeys ..
	// remaining rows (if there) are parent searchKeys ..
	vector< vector< vector<int>* >* >* parentNodes_;
	
	vector< bool >* extraParentNodeGeneratedPerRow_;		// For each row of nodes, in last set of nodes
															// if number of nodes = 1, then 
															// its parent is generated as an exceptional case, so
															// indicate that the parent should have only 1 child pageID
															// instead of the usual 2 or more .. 
															
	
};

#endif /*TESTER_H_*/
