#include "UserInterface.h"

UserInterface:: UserInterface( int maxColumnsPerTable, int bufferPoolSize ) : MAX_COLUMNS_PER_TABLE( maxColumnsPerTable ){

	// startup output text ..
	cout << "=================================" << endl << endl;
	
	cout << "Database Management System (DBMS)" << endl << endl;
	
	cout << "CS525 - Sec 2 - Steve Talbot" << endl << endl;
	
	time_t rawtime;
	
	time( &rawtime );
	
	cout << ctime( &rawtime );
	
	cout << endl;
	
	cout << "=================================" << endl << endl;

	// initialize running the program (do not yet quit) ..
	quitProgram = false;
	
	// used in create index statement ..
	// (default is "NO DUPLICATES" option not used) ..
	isNoDuplicates = false;
	
	numberOfColumns = 0;
	
	numberOfTables = 0;
	
	// this will need to be changed for the "unix/linux version" ..
	// i.e.: "/tmp/databasefiles" ..
	fileDrive = "/";
	fileDirectory = "home/";
	fileDirectoryStartPosition = 2;
	isWindows = true;
	
	CatalogFile* catalogFile = new CatalogFile( fileDrive, fileDirectory, "catalog", ".bin" );
	
	DatabaseFile* databaseFile = new DatabaseFile( fileDrive, fileDirectory );
	
	qm = new QueryManager( bufferPoolSize, catalogFile, databaseFile );
	
	// initialize these collection parameters ..
	inputTokens = new vector<string>();
	
	tableNames = new vector<string>();
	
	createColumnNames = new vector<string>();
	
	insertValues = new vector<string>();
	
	selectColumnNames = new vector<string>();
	
	whereColumnNamePairs = new multimap<string, string>();
	
	whereColumnNameValuePairs = new map<string, int>();
	
	commandInput = "";
	
	//cout << "UserInterface:: UserInterface() end .." << endl << endl;

}

UserInterface:: ~UserInterface(){
	
	//cout << "~UserInterface() start .." << endl << endl;
	
	delete qm;
	
	//cout << "~UserInterface() end .." << endl << endl;
	
}


QueryManager* UserInterface:: getQueryManager(){
	
	return qm;
	
}

IndexManager* UserInterface:: getIndexManager(){
	
	return qm->getIndexManager();
	
}

BufferManager* UserInterface:: getBufferManager(){
	
	return qm->getBufferManager();
	
}


StorageManager* UserInterface:: getStorageManager(){
	
	return qm->getStorageManager();
	
}


CatalogManager* UserInterface:: getCatalogManager(){
	
	return qm->getStorageManager()->getCatalogManager();
	
}


MessageManager* UserInterface:: getMessageManager(){
	
	return qm->getStorageManager()->getCatalogManager()->getMessageManager();
	
}


void UserInterface:: promptUserForCommand(){
	
	cout << "> ";
	
	int inputBufferSize = 1000;
	
	char inputChars[ inputBufferSize ];
	cin.getline( inputChars, 1000, '\n' );
	//cin.get(inputChars, 1000, '\n');
	
	if( inputChars != NULL )
		commandInput += inputChars;
	
	preProcessCommand();

} // end "promptUserForCommand()" ..


void UserInterface:: preProcessCommand(){
	
	// helper method for "promptUserForCommand()" and "runBatchFile()" ..
	
	string endline = "\n";
	
	string carriageReturn = "\r";
	
	if( ( commandInput == endline ) || ( commandInput == carriageReturn ) || ( commandInput.length() <= 0 ) ||
	    ( commandInput.at(0) == '\n' ) || ( commandInput.at(0) == '\r' ) ) {
		
	 	string errorString = "\nUser entered invalid input.  Please try again.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringP( errorString );	// string detailing the error in "UserInterface" ..
		
		return;
		
	}
	
	// strip off the '\n' at the end of the commandInput ..
	int commandInputSize = commandInput.size();
	if( ( commandInput[ ( commandInputSize - 1 ) ] == '\n' ) || 
		( commandInput[ ( commandInputSize - 1 ) ] == '\r' ) )
		commandInput = commandInput.substr( 0, ( commandInputSize - 1 ) );
	
	if( ( stringToUpper( commandInput ) == "END" ) ||
	    ( stringToUpper( commandInput ) == "EXIT" ) ||
	    ( stringToUpper( commandInput ) == "LOGOFF" ) || 
	    ( stringToUpper( commandInput ) == "LOG OFF" ) ||
	    ( stringToUpper( commandInput ) == "LOGOUT" ) || 
	    ( stringToUpper( commandInput ) == "LOG OUT" ) ||
	    ( stringToUpper( commandInput ) == "Q" ) ||
		( stringToUpper( commandInput ) == "QUIT" ) ||
	    ( stringToUpper( commandInput ) == "STOP" ) ){
		
		quitProgram = true;
		
	}
	
	// DEBUGGING ..
	//cout << "commandInput: \"" << commandInput << "\"" << endl;
	
	return;
	
} // end "preprocessCommand()" ..


bool UserInterface:: isQuitProgram(){
	
	return quitProgram;
	
} // end "isQuitProgram()" ..


void UserInterface:: addTableName( string tableName_ ){
	
	if( tableNames != NULL ){
		
		tableNames->push_back( tableName_ );
		
	}
	
}


void UserInterface:: clearTableNames(){
	
	tableNames->clear();
	
}


void UserInterface:: parseCommand(){
	
	//cout << "UserInterface:: parseCommand() start .." << endl << endl;
	
	// only proceed processing the query statement if the parser
	// has not yet generated an errorMessage ..
	if ( getMessageManager()->hasErrorMessage() == false ){
		
		// "strtok()" method called inside "getInputStringTokens()" mangles "commandInput".
		// that's ok, except that its used in "parseSetStatement()" and it needs to be not-mangled.
		// so, need to make a copy and pass the copy into "getInputStringTokens()",
		// allowing the copy to get mangled instead of "commandInput" ..
		string commandInputCopy = "";
		int commandInputSize = commandInput.size();
		for(int i=0;i<commandInputSize;i++)
			commandInputCopy.push_back( commandInput[i] );

		const char* delimiters = "),;=( :\\/";
	
		//vector<char>* charTokens = getInputStringTokens( commandInputCopy, delimiters );
		
		// change tokens to vector<string> ..
		//int charTokensSize = charTokens.size();
		//for(int i=0; i<charTokensSize; i++){
			
		//	inputTokens->push_back( charTokens->at(i) );
			
		//}
		
		inputTokens = getInputStringTokens( commandInputCopy, delimiters );
	
		//DEBUGGING ..
/*		for(int i=0; i<inputTokens.size(); i++){
			if ( inputTokens.size() > 0 ) {
				cout << "inputTokens.size(): " << inputTokens.size() << endl;
				cout << "inputTokens[" << i << "]: \"" << inputTokens[i] << "\"" << endl;
			}
		}
*/

		// validate that first fields of the commandInput is supported ..
		if ( !parseFirstField() ) {
		
			return;
		
		}
	
		// validate that second fields of the commandInput is supported ..
		if ( !hasSecondField() ){

			// stop parsing, move on to the "perform command" step ..
			return;
		
		}
	
		// validate that the remaining fields of the commandInput is supported ..
		if ( !parseRemainingFields() ) {

			return;
		
		}
		
	
	} // end "if ( getMessageManager()->hasErrorMessage() == false )" ..
	

} // end "parseCommand()" ..


void UserInterface:: performCommand(){
	
	// DEBUGGING ..
	//cout << "UserInterface:: performCommand() start .." << endl << endl;
	//cout << "UserInterface:: performCommand() hasParserError: " << hasParserError << endl << endl;
	//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
	//cout << "UserInterface:: performCommand() getMessageManager()->getErrorMessage(): " << getMessageManager()->getErrorMessage() << endl << endl;
	
	// only proceed processing the query statement if the parser
	// has not yet generated an errorMessage ..
	if ( getMessageManager()->hasErrorMessage() == false ) {
		
		// this is where the UI interacts with the other components ..
		if ( firstField.compare("create") == 0 ) {
			
			//cout << "UserInterface:: performCommand() create .." << endl << endl;
			
			if ( secondField.compare("table") == 0 ) {
				
				//cout << "UserInterface:: performCommand() create table .." << endl << endl;
				
				timer.startTimer();
			
				//          tableName -> thirdField
				//    attributeNames - > createColumnNames
				// numberOfAttributes -> createColumnNames.size()

				int numberOfAttributes = createColumnNames->size();
				
				// DEBUGGING ..
				//for(int i=0; i<createColumnNames->size(); i++){
				//	cout << "\"" << createColumnNames->at(i) << "\" is the " << i << " field" << endl;
				//}

				// resultMessage / errorMessage handled in "QueryManager" ..
				getQueryManager()->createTable( tableName, createColumnNames, numberOfAttributes );
		
				timer.stopTimer();
				
			} else if ( secondField.compare("index") == 0 ) {
				
				// DEBUGGING ..
				//cout << "UserInterface:: performCommand() create index .." << endl << endl;
				
				timer.startTimer();
			
				//          tableName -> thirdField
				//    attributeNames - > createColumnNames
				// numberOfAttributes -> createColumnNames.size()

				int numberOfAttributes = createColumnNames->size();

				// resultMessage / errorMessage handled in "QueryManager" ..
				getQueryManager()->createIndex( indexName, tableName, isNoDuplicates, createColumnNames->at(0) );
		
				// DEBUGGING ..
				//cout << "UserInterface:: performCommand() createIndex() called .." << endl << endl;
		
				timer.stopTimer();
				
			}
		
		
		} else if ( firstField.compare("insert") == 0 ) {
			
			timer.startTimer();
			
			int numberOfAttributes = insertValues->size();
			
			// resultMessage / errorMessage handled in "QueryManager" ..
			getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			timer.stopTimer();
		
		
		} else if ( firstField.compare("commit") == 0 ) {
			
			timer.startTimer();
			
			// resultMessage / errorMessage handled in "QueryManager" ..
			getQueryManager()->commit();
			
			timer.stopTimer();
		
		} else if ( firstField.compare("select") == 0 ) {
			
			// DEBUGGING ..
			//cout << "UserInterface:: \"select\" found .." << endl << endl;
			//cout << "UserInterface:: secondField: " << secondField << endl << endl;
				
			if ( ( secondField.compare("*") == 0 ) || 
				 ( selectColumnNames->size() > 0 ) ) {
				 	
				 timer.startTimer();
				
				// DEBUGGING ..
				//cout << "UserInterface:: \"*\" found .." << endl << endl;
				
				// resultMessage / errorMessage handled in "printDatabaseTable()" ..
				printDatabaseTable();
				
				timer.stopTimer();
				
			} // end "if ( secondField.compare("*") == 0 )" ..
		
		
		} else if ( firstField.compare("print") == 0 ) {
			
			// 4 "print" choices: catalog, table, buffer, hit rate / hitrate ..
			
			if ( secondField.compare("catalog") == 0 ) {
				
				// "print catalog" ..
				
				timer.startTimer();
				
				// resultMessage / errorMessage handled in "CatalogManager" ..
				getCatalogManager()->printCatalog();
				
				timer.stopTimer();
				
			} else if ( secondField.compare("table") == 0 ) {

				// "print table X" ..
				
				timer.startTimer();
				
				// resultMessage / errorMessage handled in "printDatabaseTable()" ..
				printDatabaseTable();

				timer.stopTimer();

				
			} else if ( secondField.compare("buffer") == 0 ) {

				// "print buffer [data]" ..
				
				
				if ( thirdField.compare("data") == 0 ) {
					
					// "print buffer data" ..
					
					timer.startTimer();
					
					// resultMessage / errorMessage handled in "printBufferDataResults()" ..
					printBufferDataResults();
					
					timer.stopTimer();
					
				} else {
				
					// "print buffer" ..
					
					timer.startTimer();
				
					// resultMessage / errorMessage handled in "printBufferResults()" ..
					printBufferResults();
					
					timer.stopTimer();
					
				} // end "if ( thirdField.compare("data") == 0 )" ..
				
				
			} else if ( ( ( secondField.compare("hit") == 0 ) && 
						  ( thirdField.compare("rate") == 0 ) ) ||
						( secondField.compare("hitrate") == 0 ) ) {

				// "print hit rate" / "print hit rate" ..
				
				timer.startTimer();
				
				int hitRateStringLength = ( floatToStr( getBufferManager()->getBufferInfoTable()->getHitRate() ) ).size();
				
				if( hitRateStringLength > 3 ){
					
					char resultString[10];  // the number should be no greater than 0.123 ..
					
					sprintf( resultString, "%.3f", getBufferManager()->getBufferInfoTable()->getHitRate() );
					
					cout << "\nhit rate: " << resultString << endl << endl;
					
				} else {
					
					cout << "\nhit rate: " << getBufferManager()->getBufferInfoTable()->getHitRate() << endl << endl;	
					
				}
				
				
				// DEBUGGING ..
				//cout << "              # of hits: " << getBufferManager()->getBufferInfoTable()->getNumberHitRequests() << endl << endl;
				//cout << "    # of block requests: " << getBufferManager()->getBufferInfoTable()->getTotalNumberBlockRequests() << endl << endl;
				//cout << "# of requests from file: " << getBufferManager()->getBufferInfoTable()->getNumberRequestsFromFile() << endl << endl;
				
				
				timer.stopTimer();
				
				
			} else if ( secondField.compare("index") == 0 ){

				// "print indexdebug" ..
				
				timer.startTimer();
				
				// set parameter for printing index ..
				getIndexManager()->printIndex( tableName, indexName );
				
				timer.stopTimer();

				
			}  // end "if ( secondField.compare("catalog") == 0 )" ..
		
		} else if ( firstField.compare("set") == 0 ) {
			
			
			if ( secondField.compare("timer") == 0 ) {
				
				// "set timer" ..
				
				if ( thirdField.compare("on") == 0 ) {
				
					// "set timer on" ..
					timer.setTimerOn();
					
					cout << endl << "timer on" << endl << endl;
				
				} else if ( thirdField.compare("off") == 0 ) {

					// "set timer off" ..
					timer.setTimerOff();
					
					cout << endl << "timer off" << endl << endl;
				
				} 
				
				
			} else if( secondField.compare("path") == 0 ) {
				
				timer.startTimer();
				
				// set file directory where database files are written / read ..
				// "fileDirectoryStartPosition" and "isWindows" set in "parseSetStatement()" ..
				updateFileDirectoryPaths();
				
				timer.stopTimer();
				
			} else if ( secondField.compare("indexdebug") == 0 ) {
				
				// "set indexdebug" ..
				
				timer.startTimer();
				
				if ( thirdField.compare("on") == 0 ) {
				
					// "set indexdebug on" ..
					getIndexManager()->setIsIndexDebug( true );
					
					cout << endl << "indexdebug on" << endl << endl;
				
				} else if ( thirdField.compare("off") == 0 ) {

					// "set indexdebug off" ..
					getIndexManager()->setIsIndexDebug( false );
					
					cout << endl << "indexdebug off" << endl << endl;
				
				} 
				
				timer.stopTimer();
				
			} // end "if ( secondField.compare("timer") == 0 )" ..
			
		} else if ( firstField.compare("show") == 0 ) {
			
			if ( secondField.compare("path") == 0 ) {
				
				timer.startTimer();
				
				// "show path" ..
				
				printPath();
				
				timer.stopTimer();
				
			}
			
		} else if ( firstField.compare("run") == 0 ) {
			
			if ( secondField.compare("batch") == 0 ) {
				
				// "run batch" ..
				
				runBatchCommands();
				
			}
			
		} // end "if ( firstField.compare("create") == 0 )" .. // end "if ( firstField.compare("create") == 0 )" ..
		
	} // end "if ( getMessageManager()->hasErrorMessage() == false )" ..

	
} // end "performCommand()" ..

void UserInterface:: printPath(){
	
	// helper method for "performCommand()" ..
	
	// "show path" ..
	string pathString = fileDrive + fileDirectory;
				
	cout << pathString << endl << endl;
	
}

void UserInterface:: runBatchCommands(){
	
	// helper method for "performCommand()" ..
	
	string returnString = "";
	
	vector<string>* commands = new vector<string>();
	
	string batchFileName = inputTokens->at( fileDirectoryStartPosition );
	
	// "run batch X" ..
	bool isReadBatchFile = getQueryManager()->getStorageManager()->readBatchFile( batchFileName, commands );
	
	if( isReadBatchFile ){
		
		//cout << "UserInterface:: runBatchCommands() Point A.." << endl << endl;
		
		// run each of the commands from the file ..
		int commandsSize = -1;
		if( commands != NULL )
			commandsSize = commands->size();
			
		//cout << "UserInterface:: runBatchCommands() commandsSize: " << commandsSize << endl << endl;
		
		//cout << "UserInterface:: runBatchCommands() Point B.." << endl << endl;
		
		if( commandsSize > 0 ){
			
			// there are commands from the batch file to run,
			// so proceed ..
			
			for(int i=0; i<commandsSize; i++){
		
				commandInput = commands->at(i);
			
				// DEBUGGING ..
				//cout << "UserInterface:: runBatchCommands() commandInput: " << commandInput << endl << endl;
		
				// show the user the batch command prior to running it ..
				cout << "batch command > \"" << commandInput << "\"" << endl << endl;
			
				// workaround .. "preProcessCommand()" strips these off
				// the end of each "commandInput" for windows OS,
				// which is OK for command-line input,
				// but for batch file these were stripped off;
				// put them back on here to avoid problems ..
				if( isWindows )
					commandInput += "\n";
		
				preProcessCommand();
		
				parseCommand();
			
				performCommand();
			
				printMessagesAndCleanUp();
		
			}
			
			// finished processing the batch file,
			// notify the user as such ..
			returnString = "\nProcessing batch file \"";
			returnString += batchFileName;
			returnString += "\" completed.";
			
			getMessageManager()->setResultMessage( returnString );
			
		} else {
			
			// there were no commands to process,
			// notify the user as such ..
			returnString = "\nUnable to process commands in batch file \"";
			returnString += batchFileName;
			returnString += "\" [ UI::runBatchCommands() ].";
		
			getMessageManager()->setResultMessage( returnString );
			
		}
		

		
	}

}


void UserInterface:: updateFileDirectoryPaths(){
			
	// if there is anything in the buffer,
	// perform a "commit()" operation
	// in order to save whatever is in the buffer to file
	// before changing to a different directory ..
	getQueryManager()->commit();
			
	fileDrive = "";
	fileDirectory = "";
	
	// "set path X", where "X" is in position "2" ..
	fileDirectoryStartPosition = 2;
	
	// "isWindows" should be set in "updatePathOperatingSystem()" ..
	
	// start after "set path ", looking for directory info ..
	// get "fileDirectory" in order to update the "CatalogFile" in the "cm" ..
	int inputTokensSize = inputTokens->size();

	if( isWindows ) {
		
		for(int i=fileDirectoryStartPosition; i<inputTokensSize; i++){
						
			if( i == fileDirectoryStartPosition ){
			
				// get "fileDrive" ..
				fileDrive = inputTokens->at( fileDirectoryStartPosition );
		
				fileDrive += ":\\";
			
			} else if( i > fileDirectoryStartPosition ) {
			
				fileDirectory += inputTokens->at(i);
			
				fileDirectory += "\\";
				
			}
						
			//cout << "UserInterface:: inputTokens[i("<<i<<")]: " << inputTokens[i] << endl << endl;
						
		}
		
	} else {
		
		for(int i=fileDirectoryStartPosition; i<inputTokensSize; i++){
						
			// set the "root" drive ..
			if( i == fileDirectoryStartPosition )
				fileDirectory += "/";

			fileDirectory += inputTokens->at(i);
			
			// get "fileDrive" ..
			fileDirectory += "/";
			
			// NOTE: first directory captured in "fileDrive" for windows, fileDrive == "" for unix/linux ..
						
			//cout << "UserInterface:: inputTokens[i("<<i<<")]: " << inputTokens[i] << endl << endl;
						
		}
		
	}

	
	// DEBUGGING ..
	//cout << "UserInterface:: updateFileDirectoryPaths() fileDrive: " << fileDrive << endl << endl;
	//cout << "UserInterface:: updateFileDirectoryPaths() fileDirectory: " << fileDirectory << endl << endl;
	
	// since the path may have changed,
	// clear the catalog and start out with default values ..
	getCatalogManager()->resetCatalog();
	
	// update the Catalog (CatalogFile) ..
	CatalogFile* newCatalogFile = getCatalogManager()->getCatalogFile();
	
	// DEBUGGING ..
	//cout << "UserInterface:: updateFileDirectoryPaths() newCatalogFile->getFileName(): " << newCatalogFile->getFileName() << endl << endl;
	//cout << "UserInterface:: updateFileDirectoryPaths() newCatalogFile->getFileExtension(): " << newCatalogFile->getFileExtension() << endl << endl;
	
	newCatalogFile->setFileDrive( fileDrive );
	
	newCatalogFile->setFileDirectory( fileDirectory );
	
	
	// update the Catalog (DatabaseFile) ..
	DatabaseFile* newDatabaseFile = getStorageManager()->getDatabaseFile();
	
	newDatabaseFile->setFileDrive( fileDrive );
	
	newDatabaseFile->setFileDirectory( fileDirectory );

	// show the result of this operation to the user ..
	printPath();
	
	// nothing in the buffer pool to "commit",
	// so inform the user that the operation was cancelled ..
	string returnString = "\nThe working directory path name was changed.";
		
	getMessageManager()->setResultMessage( returnString );

	
} // end "updateFileDirectoryPaths()"


void UserInterface:: printDatabaseTable() {
	
	// this is called in "performCommand()" for "select *" command ..
	
	//cout << "UserInterface:: printDatabaseTable() start .." << endl << endl;
	
	map<int, DataPage* >* bpDataPages = NULL;
				
	map<int, DataPage* >* fileDataPages = NULL;
				
	vector< map<int, DataPage* >* >* returnPages = NULL;
				
				
	//cout << "UserInterface:: printDatabaseTable() Point A1.." << endl << endl;
				
	if ( tableNames->size() == 1 ){
		
		//cout << "UserInterface:: printDatabaseTable() Point A2.." << endl << endl;
					
		tableName = tableNames->at(0);
					
		// DEBUGGING ..
		//cout << "UserInterface:: tableName: " << tableName << endl << endl;
					
		getBufferManager()->getBitSnapShot()->getLRUBlocks()->clear();
					
		getBufferManager()->getBitSnapShot()->getRowAvailable()->clear();
		
		//cout << "UserInterface:: printDatabaseTable() Point B.." << endl << endl;
				
		returnPages = getQueryManager()->selectFromX( tableName, selectColumnNames, whereColumnNamePairs, whereColumnNameValuePairs );
					
		// DEBUGGING ..
		//cout << "UserInterface:: selectFromX() done .. " << endl << endl;
		//cout << "( ( bpDataPages.size()("<<bpDataPages.size()<<") > 0 ) || ( fileDataPages.size()("<<fileDataPages.size()<<") > 0 ) ): " << ( ( bpDataPages.size() > 0 ) || ( fileDataPages.size() > 0 ) ) << endl << endl;
		
		//cout << "UserInterface:: printDatabaseTable() Point C.." << endl << endl;
		
		if ( ( returnPages != NULL ) && 
			 !getMessageManager()->hasErrorMessage() ) {
						
			bpDataPages = returnPages->at(0);
					
			fileDataPages = returnPages->at(1);
			
			// DEBUGGING ..
			//cout << "UserInterface:: printDatabaseTable() bpDataPages.size(): " << bpDataPages.size() << endl << endl;
			//cout << "UserInterface:: printDatabaseTable() fileDataPages.size(): " << fileDataPages.size() << endl << endl;
						
			if ( ( bpDataPages != NULL ) && 
				 ( fileDataPages != NULL ) &&
				 ( ( bpDataPages->size() > 0 ) || ( fileDataPages->size() > 0 ) ) ) {
						
				// print out any results that are returned ..
						
				//cout << "RESULTS start .." << endl << endl;
				
				//cout << "UserInterface:: printDatabaseTable() Point D.." << endl << endl;
								
				printResults( bpDataPages, fileDataPages, selectColumnNames );
				//oldPrintResults( bpDataPages, fileDataPages, selectColumnNames );

			} else {
		
				// indicate that the query of an existing table returned no results ..

				getMessageManager()->setResultMessage( "\nNO RESULTS ...\n" );
						
			}
						
		} // end "if ( !getMessageManager()->hasErrorMessage() )" ..
				
				
		// DEBUGGING ..
		//cout << "tableName.size(): " << tableName.size() << endl << endl;
		//cout << "tableName.size() == 1 end .." << endl << endl;
		
		//cout << "UserInterface:: printDatabaseTable() Point E.." << endl << endl;
					
					
	}
	
	//cout << "UserInterface:: printDatabaseTable() Point F.." << endl << endl;

	
} // end "printDatabaseTable()" ..


void UserInterface:: printMessagesAndCleanUp(){
	
	// DEBUGGING ..
	//cout << "getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl;
	//cout << "getMessageManager()->getErrorMessage(): " << getMessageManager()->getErrorMessage() << endl;
		
	//cout << "message start .." << endl << endl;

	if ( getMessageManager()->hasErrorMessage() ){
			
		//cout << "print error message .." << endl << endl;
				
		cout << getMessageManager()->getErrorMessage();
				
	}
			
	//cout << "print result message .." << endl << endl;
			
	cout << getMessageManager()->getResultMessage();
	
	cout << "\n\n";
	
	
	// reset the global values in "UI" and "MM",
	// including error string, error flags, etc..
	resetValues();
	
} // end "printMessagesAndCleanUp()" ..


void UserInterface:: printResults( map<int, DataPage* >* bpDataPages,  map<int, DataPage* >* fileDataPages, vector<string>* selectColumnNames ){
	
	//cout << "UserInterface:: printResults() start .." << endl << endl;
	
	// DEBUGGING ..
	//cout << "UserInterface:: printResults() for table \"" << tableName << "\":" << endl << endl;
	//cout << "UserInterface:: bpDataPages.size(): " << bpDataPages.size() << endl << endl;
	//cout << "UserInterface:: fileDataPages.size(): " << fileDataPages.size() << endl << endl;
	//cout << "UserInterface:: selectColumnNames.size(): " << selectColumnNames.size() << endl << endl;
	
	// DEBUGGING ..
	//list<int>::iterator it_LRU;
	//for(it_LRU = qm->bm->bit.LRUBlocks.begin(); it_LRU != qm->bm->bit.LRUBlocks.end(); it_LRU++){
			
		//cout << "UserInterface:: bit.LRUBlocks.(*it_LRU): " << (*it_LRU) << endl << endl;
			
	//} // end "for()" ..
	
	// DEBUGGING ..
	//for(it_LRU = qm->bm->bitSnapShot.LRUBlocks.begin(); it_LRU != qm->bm->bitSnapShot.LRUBlocks.end(); it_LRU++){
		//cout << "UserInterface:: bitSnapShot.LRUBlocks.(*it_LRU): " << (*it_LRU) << endl << endl;
	//} // end "for()" ..
	
	// DEBUGGING ..
	//vector< vector<bool> >::iterator it_saBlock;
	//vector<bool>::iterator it_sa;
	//for(it_saBlock = qm->bm->bit.rowAvailable.begin(); it_saBlock != qm->bm->bit.rowAvailable.end(); it_saBlock++){
		//for(it_sa = (*it_saBlock).begin(); it_sa != (*it_saBlock).end(); it_sa++){
			//cout << "UserInterface:: bit.rowAvailable.(*it_sa): " << (*it_sa) << endl << endl;	
		//}
		//cout << "---------------------" << endl << endl;
	//} // end "for()" ..
	
	// DEBUGGING ..
	//for(it_saBlock = qm->bm->bitSnapShot.rowAvailable.begin(); it_saBlock != qm->bm->bitSnapShot.rowAvailable.end(); it_saBlock++){
		//for(it_sa = (*it_saBlock).begin(); it_sa != (*it_saBlock).end(); it_sa++){
			//cout << "UserInterface:: bitSnapShot.rowAvailable.(*it_sa): " << (*it_sa) << endl << endl;	
		//}
		//cout << "---------------------" << endl << endl;
	//} // end "for()" ..
	
	//cout << "UserInterface:: printResults() Point A.." << endl << endl;
	
	// instantiate the "tableFormatter" to print a table ..
	TableFormatter tableFormatter;
	
	// define the default table values for if no data in the cell ..
	//tableFormatter.addDefaultValue( 0, "n/a" );
	tableFormatter.setDefaultFieldValue( "-" );
	
	// character used in the divider separating "header" from "data" ..
	tableFormatter.setDivider( '=' );
	
	// middle, left, right divider-column spacers ..
	tableFormatter.setDividerColumnSpacer( "||" );
	tableFormatter.setDividerColumnSpacerLeft( "=" );
	tableFormatter.setDividerColumnSpacerRight( "=" );
	
	// middle, left, right data-column spacers ..
	tableFormatter.setDataColumnSpacer( "  " );
	tableFormatter.setDataColumnSpacerLeft( " " );
	tableFormatter.setDataColumnSpacerRight( " " );
	
	//cout << "UserInterface:: printResults() Point A .." << endl << endl;
	
	if ( ( ( bpDataPages != NULL ) && ( bpDataPages->size() > 0 ) ) || 
		 ( ( fileDataPages != NULL ) && ( fileDataPages->size() > 0 ) ) )  {
	
		// DEBUGGING ..

		getCatalogManager()->readMetaData();
	
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point B .." << endl << endl;
		
		/*
		 * process the "header" rows ..
		 */ 
	
		CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );
		
		vector<string>* headerStrings = new vector<string>();
		
		string ctaString = "";
		
		map<int, string>* columnNames = new map<int, string>();
		
		//cout << "UserInterface:: printResults() Point C.." << endl << endl;
		
		if ( ( selectColumnNames != NULL ) && 
			 ( selectColumnNames->size() > 0 )  ) {
			
			// this handles "select a, bb, ccc from x" queries ..
			
			//cout << "UserInterface:: printResults() tableName: " << tableName << endl << endl;
			
			//cout << "UserInterface:: printResults() selectColumnNames->size(): " << selectColumnNames->size() << endl << endl;
			
			// match valid column names with their position order (from left to right)
			// as it is recorded in the catalog ..
			columnNames = getQueryManager()->groupSelectColumnNames( tableName, selectColumnNames );
			
			//cout << "UserInterface:: printResults() columnNames->size(): " << columnNames->size() << endl << endl;
			
			// DEBUGGING ..
			//cout << "UserInterface:: catalogTable.getNumberOfAttributes() (I): " << catalogTable.getNumberOfAttributes() << endl << endl;
			//cout << "UserInterface:: columnNames.size() (I): " << columnNames.size() << endl << endl;
			
			int numberOfAttributes = 0;
			if( catalogTable != NULL )
				numberOfAttributes = catalogTable->getNumberOfAttributes();
				
			//cout << "UserInterface:: printResults() Point D.." << endl << endl;
			
			for(int i=0; i<numberOfAttributes; i++){
				
				// DEBUGGING ..
				//cout << "UserInterface:: columnNames.find( "<<i<<" )->second (I): " << columnNames.find( i )->second << endl << endl;
				
				if( columnNames->find( i ) != columnNames->end() ){
					
					ctaString = columnNames->find( i )->second;
					
					// DEBUGGING ..
					//cout << "UserInterface:: ctaString (I): " << ctaString << endl << endl;
					
					headerStrings->push_back( ctaString );
					
					// DEBUGGING ..
					//cout << "UserInterface:: ctaString.size() (I): " << ctaString.size() << endl << endl;
		
					//headerLengths.push_back( ctaString.size() );
					
				}
				
			} // end "for()" ..
			
			//cout << "UserInterface:: printResults() Point E.." << endl << endl;
			
		} else {
			
			//cout << "UserInterface:: printResults() Point F.." << endl << endl;

			// this handles "select * from x" queries ..
	
			vector<CatalogTableAttribute*>* ctas = catalogTable->getAttributes();
	
			if( ctas != NULL ){
				
				vector<CatalogTableAttribute*>::iterator it_cta;
	
				for(it_cta = ctas->begin(); it_cta != ctas->end(); it_cta++){
	
					ctaString = (*it_cta)->getTableAttributeName();

					headerStrings->push_back( ctaString );


				}
				
			}
			
			//cout << "UserInterface:: printResults() Point G.." << endl << endl;

		} // end "if ( selectColumnNames.size() > 0  )" ..
		
		//cout << "UserInterface:: printResults() Point H.." << endl << endl;
		
		// define the table "header" column title names ..
		int headerStringsSize = headerStrings->size();
		for(int i=0; i<headerStringsSize;i++){
			
			tableFormatter.addColumnHeader( headerStrings->at(i) );

		}
		
		//cout << "UserInterface:: printResults() Point H1.." << endl << endl;
		
		// middle, left, right header-column spacers ..
		tableFormatter.setHeaderColumnSpacer( "  " );
		tableFormatter.setHeaderColumnSpacerLeft( " " );
		tableFormatter.setHeaderColumnSpacerRight( " " );

	
		DataHeader* header = new DataHeader();
		getStorageManager()->readHeader( tableName, header );
	
		vector<bool>* rowAvailable = new vector<bool>();
	
		int rowCount = 0;
	
		int headerEntryIndex = 0;
		
		int fieldCount = -1;
		
		//cout << "UserInterface:: printResults() Point H2.." << endl << endl;
	
		if ( ( selectColumnNames != NULL ) && 
			 ( selectColumnNames->size() > 0 )  ) {
			
			// this handles "select a, bb, ccc from x" queries ..
			
			//cout << "UserInterface:: printResults() (columnNames == NULL): " << (columnNames == NULL) << endl << endl;
			
			if( columnNames != NULL )
				fieldCount = columnNames->size();
			else
				fieldCount = 0;
			
		} else {
			
			// this handles "select * from x" queries ..
			
			fieldCount = catalogTable->getNumberOfAttributes();
			
		}
		
		//cout << "UserInterface:: printResults() Point H3.." << endl << endl;
	
		vector<int>* maxDataLengths = new vector<int>();
	
		int dataValue = -1;
	
		string dataString;
	
		map<int, DataPage* >::iterator it_dp;
	
		//cout << "UserInterface:: printResults() Point J.." << endl << endl;

		/*
		 * process the "data" rows ..
		 */ 
	
		headerEntryIndex = 0;
	
		rowCount = 0;
	
		// calculate the number of "DataPages" to print ..
		int numberOfDataPagesToPrint = 0;
		
		if( fileDataPages != NULL )
			numberOfDataPagesToPrint = fileDataPages->size();
		
		if( bpDataPages != NULL )
			numberOfDataPagesToPrint += bpDataPages->size();
		

		int minPageID = 0;
		
		int fileDataPageCounter = 0, bpDataPageCounter = 0;
		
		vector<int>* minFilePageIDs = new vector<int>();
		
		vector<int>* minBPPageIDs = new vector<int>();
		
		bool skipPageID = false;
		
		bool isMinFilePageID = false;
		
		bool minFilePageIDsPopped = false;
		
		//cout << "UserInterface:: printResults() Point K.." << endl << endl;

		// track whether the BP has pageID read in from file ..
		bool pageIDsFoundInBP = false;
			
		int numberPageIDsFoundInBP = 0;
		
		if( bpDataPages != NULL ){
			
			for(it_dp = bpDataPages->begin(); it_dp != bpDataPages->end(); it_dp++){
				
				// search results from BP for pageID that were read in from file ..
				
				if ( it_dp->first >= 0 ) {
					
					// this indicates that at least 1 DataPage in the BP
					// was read in from the file ..
				
					pageIDsFoundInBP = true;
					
					numberPageIDsFoundInBP++;
					
				}
				
			} // end "for()" ..
			
		}
		
		//cout << "UserInterface:: printResults() Point L.." << endl << endl;
		
		int minFilePageIDsSize = -1;

		// print out the "DataPages" ..
		for(int i=0; i<numberOfDataPagesToPrint; i++){
			
			// look through all the "DataPages" (both file and BP) ..
			
			
			// reset these parameters for current iteration ..
			minPageID = -1;
			
			fileDataPageCounter = 0;
			
			isMinFilePageID = false;
			
			if( fileDataPages != NULL ){
				
				// Handle "file DataPages" ..
				for(it_dp = fileDataPages->begin(); it_dp != fileDataPages->end(); it_dp++){
				
					// find the "minimum" pageID in "fileDataPages" map ..
					minFilePageIDsSize = minFilePageIDs->size();
					for(int j=0; j<minFilePageIDsSize; j++){
					
						// look through all the "file DataPages" ..

						if( minFilePageIDs->at(j) == it_dp->first ){
						
							// this indicates that we've used this pageID already,
							// so skip it and look for the next "minPageID" ..
							skipPageID = true;
							break;
						
						}
					
					} // end "for()" ..
				
					if ( !skipPageID ) {
					
						// (*it_dp).first ) > 0 ( != -1 ) ..
						
						// find the "minimum" pageID in "fileDataPages" map ..
						if ( ( minPageID >= it_dp->first ) ||
				    		 ( minPageID < 0 ) )  {

							// "minPageID >= ( (*it_dp).first )" ..
							// indicates that new "minPageID" found ..
						
							// "minPageID == -1" ..
							// indicates that search for "minPageID" just started for this iteration ..
					
							minPageID = it_dp->first;
							isMinFilePageID = true;
							break;
					
						}
					
					} else {

						// reset this parameter for next iteration ..
						skipPageID = false;	
					
					}
				
				
				} // end "for()" ..
				
			}


			
			// update "minFilePageIDs" with the current "minPageID" ..
			if ( minPageID >= 0 ){

				minFilePageIDs->push_back( minPageID );
					
			}	
			
			// reset these parameters for its upcoming reuse ..
			minFilePageIDsPopped = false;
			int bpBlockIndex = -1;


			// Handle "BP DataPages" ..
			
			if ( pageIDsFoundInBP ) {
				
				// handle a BP that has both DataPages
				// 1) loaded in from file, and
				// 2) yet to be allocated-and-written-to-file ..
				
				if ( numberPageIDsFoundInBP > 0 ) {

					// First, handle the DataPages loaded in from file;
					// these have pageID >= 0 ..
					
					bpDataPageCounter = 0;
					
					if( bpDataPages != NULL ){
						
						for(it_dp = bpDataPages->begin(); it_dp != bpDataPages->end(); it_dp++){


							if ( it_dp->first < 0 ) {
						
								// since this section of code is looking for positive pageID,
								// skip the pageID equal to "-1" for now ..
								continue;
						
							}
					
							// find the "minimum" pageID in "bpDataPages" map ..

							int minBPPageIDsSize = minBPPageIDs->size();
							for(int j=0; j<minBPPageIDsSize; j++){
							
								// look through all the "BP DataPages" ..
				
								if( minBPPageIDs->at(j) == it_dp->first ){
								
									// this indicates that we've used this pageID already,
									// so skip it and look for the next "minPageID" ..
									skipPageID = true;
									break;
								
								}
						
							}
				
							if ( !skipPageID ) {
							
								// find the "minimum" pageID in "bpDataPages" map ..
								if ( ( minPageID >= it_dp->first ) ||
				    	 	 		 ( minPageID < 0 ) )  {
				    	 	 	 	
									// "minPageID >= ( (*it_dp).first )" ..
									// indicates that new "minPageID" found ..
								
									// "minPageID == -1" ..
									// indicates that search for "minPageID" just started for this iteration, 
									// and that "fileDataPages" had no pageID ..		
						
									if ( !minFilePageIDsPopped && isMinFilePageID ) {
									
										// since the "min" pageID here is from "bpDataPages",
										// remove the "minPageID" found above for "fileDataPages"
										// ( but only when above found "isMinFilePageID == true";
										//   this insures that older "minFilePageIDs" aren't mistakenly popped ) ..
										// NOTE: this should only occur at most _once_ per iteration ..
										minFilePageIDs->pop_back();
										minFilePageIDsPopped = true;
							
									}
					
									minPageID = it_dp->first;
									isMinFilePageID = false;
								
									// this means that the "minPageID" found in the
									// bp was used instead of the one from the file, so
									// update this parameter for the next iteration ..
									// (this shouldn't be outside of the loop, because
									//  if the file "minPageID" is used, don't want to 
									//  decrement this parameter)
									numberPageIDsFoundInBP--;
					
					
								} // end "if ( ( minPageID >= ( (*it_dp).first ) )" ..
					
					
							} else {

								// reset this parameter for next iteration ..
								skipPageID = false;	
					
							} // end "if ( !skipPageID )" ..
				
				
						} // end "for()" ..
						
					}
	
				} else {
					
					// Second, handle the DataPages inserted into the BP
					// but that are not yet written to file;
					// these have pageID == -1 ..
					
					// "(*it_dp).first ) == -1"..
					// this indicates that block in BP hasn't been written to file yet,
					// so use the "LRU" block instead (as the current block to print),
					// because all "minPageIDs" that were "read-from-file" were already
					// handled above ..
					
					// NOTE: "minPageID >= 0" indicates that a "fileDataPage" minPageID
					// was located, so go with that instead.  Also, BP minPageID was 
					// all dealt with by this point ..
							
					if ( ( bpBlockIndex < 0 ) && ( minPageID < 0 ) ) {
							
						// "bpBlockIndex == -1" ..
						// this indicates that the "LRU" block is available for use ..
							
						// OR
							
						// "minPageID == -1" ..
						// this indicates that "minPageID" wasn't found yet,
						// so go ahead and use the "LRU" block .. 
						
						// get the index of the next LRU BP block, and update list ..
						bpBlockIndex = getBufferManager()->getBitSnapShot()->getLRUBlocks()->front();
						//bpBlockIndex = qm->bm->bitSnapShot.LRUBlocks.front();
						
						//cout << "UserInterface:: Point V_1 .." << endl << endl;
						//cout << "UserInterface:: bpBlockIndex: " << bpBlockIndex << endl << endl;
						
						getBufferManager()->getBitSnapShot()->getLRUBlocks()->pop_front();
						//qm->bm->bitSnapShot.LRUBlocks.pop_front();
						//cout << "UserInterface:: Point V_2 .." << endl << endl;
						
						getBufferManager()->getBitSnapShot()->getLRUBlocks()->push_back( bpBlockIndex );
						//qm->bm->bitSnapShot.LRUBlocks.push_back( bpBlockIndex );
						//cout << "UserInterface:: Point V_3 .." << endl << endl;
						
						isMinFilePageID = false;
						//cout << "UserInterface:: Point V_4 .." << endl << endl;
					
					} 

					
				} // end "if ( numberPageIDsFoundInBP > 0 )" ..
				
			} else {
				
				// There are _no_ DataPages loaded in from file, 
				// so only need to handle the DataPages inserted into the BP;
				// these have pageID == -1 ..
					
				// "(*it_dp).first ) == -1"..
				// this indicates that block in BP hasn't been written to file yet,
				// so use the "LRU" block instead (as the current block to print),
				// because all "minPageIDs" that were "read-from-file" were already
				// handled above ..
							
				// NOTE: "minPageID >= 0" indicates that a "fileDataPage" minPageID
				// was located, so go with that instead ..
					
				if ( ( bpBlockIndex < 0 ) && ( minPageID < 0 ) ) {
						
					// "bpBlockIndex == -1" ..
					// this indicates that the "LRU" block is available for use ..
							
					// OR
							
					// "minPageID == -1" ..
					// this indicates that "minPageID" wasn't found yet,
					// so go ahead and use the "LRU" block .. 
						
					// get the index of the next LRU BP block, and update list ..
					bpBlockIndex = getBufferManager()->getBitSnapShot()->getLRUBlocks()->front();

					getBufferManager()->getBitSnapShot()->getLRUBlocks()->pop_front();

					getBufferManager()->getBitSnapShot()->getLRUBlocks()->push_back( bpBlockIndex );
					
					isMinFilePageID = false;
					
				} 
				
				
			} // end "if ( pageIDsFoundInBP )" ..

			// update "minBPPageIDs" with the current "minPageID" ..
			if ( ( minPageID >= 0 ) && ( !isMinFilePageID ) ) {

				minBPPageIDs->push_back( minPageID );
					
			}


			// print out the "DataPage" for the current "minPageID" ..
			if( isMinFilePageID ){
				
				int columnCount = 0;
				
				DataPage* fileDataPage = fileDataPages->find( minPageID )->second;
				
				vector<int>* entries = NULL;
				
				int numberOfFields = -1;
				
				vector<DataHeaderEntry*>* headerEntries = header->getHeaderEntries();
				
				DataHeaderEntry* headerEntry = NULL;
				
				if( headerEntries != NULL )
					headerEntry = headerEntries->at( minPageID );
					
				if( headerEntry != NULL )
					rowAvailable = headerEntry->getRowAvailable();
				
				if( fileDataPage != NULL )
					rowCount = fileDataPage->getNumberOfRows();
				
				for(int i=0; i<rowCount; i++){
				
					if( ( rowAvailable == NULL ) || 
						( i == rowAvailable->size() ) )
						break;
				
					if ( rowAvailable->at(i) == false ) {
						
						if( fileDataPage != NULL )
							entries = fileDataPage->getEntries( i );
						
						if( entries != NULL )
							numberOfFields = entries->size();
					
						//for(int j=0; j<fieldCount; j++){
						for(int j=0; j<numberOfFields; j++){
						
							dataValue = entries->at( j );
						
							dataString = intToStr( dataValue );
						
							// store the "DATA" fields ..
							tableFormatter.addColumnData( columnCount, dataString );
							columnCount++;
	
						} // end "for()" ..
						
						// set parameters for next row of data ..
						tableFormatter.endDataRow();
						columnCount = 0;
					
					} // end "if()" .. 
				
				} // end "for()" ..
				
				
			} else {
			
				DataPage* bpDataPage = new DataPage();
				
				if ( bpBlockIndex >= 0 ){
					
					// the "bpDataPage" value is found using a key = "bpOffset" 
					// ( see QueryManager::selectStarFromX() ),
					// where 
					// "bpOffset" = [ (-1) * (i+1) ]
					//        "i" = 0 to bp.getBufferPoolSize()
					// so
					// "bpOffset" = [ (-1) * (bpBlockIndex+1) ]
					
					int bpOffset = ( (-1) * ( bpBlockIndex+1 ) );
					
					rowAvailable = getBufferManager()->getBufferInfoTable()->getRowAvailable()->at( bpBlockIndex );
					
					if( bpDataPages != NULL )
						bpDataPage = bpDataPages->find( bpOffset )->second;
					
				} else {
					
					// BP pages were captured in "bpDataPages" and file pages were captured in "fileDataPages",
					// The operation of capturing "fileDataPages" changes values in the "bit",
					// so in order to keep "bit" values that match the pages in "bpDataPages",
					// a "snap shot" of the "bit" that properly matches "bpDataPages" is stored in "qm->bm->bitSnapShot" ..
					
					int bpIndex = getBufferManager()->getBitSnapShot()->getBufferInfoTableIndex( tableName, minPageID );
					
					rowAvailable = getBufferManager()->getBitSnapShot()->getRowAvailable()->at( bpIndex );
					
					if( bpDataPages != NULL )
						bpDataPage = bpDataPages->find( minPageID )->second;
					
				}
				
				int columnCount = 0;
				
				int numberOfFields = -1;
				
				vector<int>* entries = NULL;
				
				//rowCount = bpDataPage->getRecords()->size();
				rowCount = bpDataPage->getNumberOfRows();
				
				for(int i=0; i<rowCount; i++){
					
					if( ( rowAvailable == NULL ) || 
						( i == rowAvailable->size() ) )
						break;
					
					if ( rowAvailable->at(i) == false ) {
				
						if( bpDataPage != NULL )
							entries = bpDataPage->getEntries( i );
						
						if( entries != NULL )
							numberOfFields = entries->size();
				
						//for(int j=0; j<fieldCount; j++){
						for(int j=0; j<numberOfFields; j++){

							//dataValue = bpDataPage->getRecords()->at( i ).getFields()->at( j ).getValue();
							dataValue = entries->at(j);
					
							dataString = intToStr( dataValue );

							// store the "DATA" fields ..
							tableFormatter.addColumnData( columnCount, dataString );
							columnCount++;
				
						} // end "for()" ..
						
						// set parameters for next row of data ..
						tableFormatter.endDataRow();
						columnCount = 0;

					} // end "if()" .. 
			
				} // end "for()" ..
				
			} // end "if( isMinFilePageID )" ..
			
			
		} // end "for(int i=0; i<numberOfDataPagesToPrint; i++)" ..
		
		//cout << "UserInterface:: printResults() Point M.." << endl << endl;
		
		// format the data input in preparation for printing ..
		tableFormatter.formatTable();
	
		// print the formatted table ..
		tableFormatter.printTable();

		
	} else {
		
		//cout << "NO RESULTS" << endl << endl;
		string resultString = "\nNO RESULTS ...\n";

		getMessageManager()->setResultMessage( resultString );
		
	}
	
	
} // end "printResults()" ..



void UserInterface:: printBufferResults(){

	// this is called in "performCommand()" for "print buffer" command ..
	
	// print buffer meta data ..
				
	// determine the max lengths of fields for printing 
	// buffer meta data ..
	
	// instantiate the "tableFormatter" to print a table ..
	TableFormatter tableFormatter;
	
	// define the table "header" column title names ..
	tableFormatter.addColumnHeader( "buffer #" );
	tableFormatter.addColumnHeader( "file / table" );
	tableFormatter.addColumnHeader( "block #" );
	
	// define the default table values for if no data in the cell ..
	tableFormatter.addDefaultValue( 0, "n/a" );
	tableFormatter.addDefaultValue( 1, "empty" );
	tableFormatter.addDefaultValue( 2, "none" );
	tableFormatter.setDefaultFieldValue( "E" );
	
	// middle, left, right header-column spacers ..
	tableFormatter.setHeaderColumnSpacer( "  " );
	tableFormatter.setHeaderColumnSpacerLeft( " " );
	tableFormatter.setHeaderColumnSpacerRight( " " );
	
	// character used in the divider separating "header" from "data" ..
	tableFormatter.setDivider( '=' );
	
	// middle, left, right divider-column spacers ..
	tableFormatter.setDividerColumnSpacer( "||" );
	tableFormatter.setDividerColumnSpacerLeft( "=" );
	tableFormatter.setDividerColumnSpacerRight( "=" );
	
	// middle, left, right data-column spacers ..
	tableFormatter.setDataColumnSpacer( "  " );
	tableFormatter.setDataColumnSpacerLeft( " " );
	tableFormatter.setDataColumnSpacerRight( " " );
				
	string bpNumber = "";
	
	string bitTableName = "";
				
	int bitPageID = -1;
	string pageIDString = "";
	
	string printString = "";
	
	int columnCount = 0;
	
	int bufferPoolSize = getBufferManager()->getBufferInfoTable()->getBufferPoolSize();
		
	for(int i=0; i<bufferPoolSize; i++){
					
					
		// column "0" ..
		bpNumber = intToStr( i );
		
		if( bpNumber.size() > 0 )
			tableFormatter.addColumnData( columnCount, bpNumber );
		
		columnCount++;
					
					
		// column "1" ..
		//bitTableName = qm->bm->bit.tableName[i];
		bitTableName = getBufferManager()->getBufferInfoTable()->getTableName()->at(i);
		
		if( bitTableName.size() > 0 )
			tableFormatter.addColumnData( columnCount, bitTableName );
		
		columnCount++;
		
		
		// column "2" ..
		//bitPageID = qm->bm->bit.pageID[i];
		bitPageID = getBufferManager()->getBufferInfoTable()->getPageID()->at(i);
					
		pageIDString = intToStr( bitPageID );
		
		if( pageIDString.size() > 0 )
			tableFormatter.addColumnData( columnCount, pageIDString );
		
		// update "tableFormatter" for next row ..
		tableFormatter.endDataRow();
		
		// reset parameter for next row ..
		columnCount = 0;
		
	
	} // end "for()" ..	
	
	// format the data input in preparation for printing ..
	tableFormatter.formatTable();
	
	// print the formatted table ..
	tableFormatter.printTable();
	
	// the code below is replaced by the code above using "tableFormatter" ..
	
	
} // end "printBufferResults()" ..


void UserInterface:: printBufferDataResults(){

	// this is called in "performCommand()" for "print buffer data" command ..
	
	// DEBUGGING ..
	//cout << "UserInterface:: printBufferDataResults() Point A .." << endl << endl;
	

	// print the rows in the buffer for each "tableName" in the bit,
	// but do not print duplicates 
	// (ex: assume the following for the bp blocks
	//		bp block 1: tableName = "abc"
	//		bp block 2: tableName = "az9"
	//		bp block 3: tableName = "abc"
				
	//      the first iteration will locate tableName = "abc" from
	//		bp block 1 and print up _all_ rows for "abc" from both
	//		bp block 1 & bp block 3.
				
	//		the second iteration will locate tableName = "az9" from
	//		bp block 2 and print up its rows.
				
	//		the third iteration will locate tableName = "abc" from
	//		bp block 3, but since this tableName was already located
	//		by bp block 1, the rows for this tableName will not be 
	//		duplicated.
				
	//bool foundResults = false;
				
	map<int, DataPage* >* returnPages = new map<int, DataPage* >();
				
	set<string> processedBitTableNames;
					
	string bitTableName = "";
	
	bool anyPagesReturned = false;
	
	// DEBUGGING ..
	//cout << "UserInterface:: printBufferDataResults() Point B .." << endl << endl;
	//cout << "UserInterface:: printBufferDataResults() qm->bm->bit.bufferPoolSize: " << qm->bm->bit.bufferPoolSize << endl << endl;
				
	int bufferPoolSize = getBufferManager()->getBufferInfoTable()->getBufferPoolSize();
				
	for(int i=0; i<bufferPoolSize; i++){
					
		// get the tableName for this bp block ..
		bitTableName = getBufferManager()->getBufferInfoTable()->getTableName()->at(i);
		
		// DEBUGGING ..
		//cout << "UserInterface:: printBufferDataResults() Point C .." << endl << endl;
		//cout << "UI:: print buffer: bitTableName["<<i<<"]: " << bitTableName << endl << endl;
		//cout << "processedBitTableNames.find( bitTableName("<<bitTableName<<") ) == processedBitTableNames.end()" << ( processedBitTableNames.find( bitTableName ) == processedBitTableNames.end() ) << endl << endl;
					
		// check the "processedBitTableNames" set if this tableName
		// was already encountered in a previous bp block ..
					
		if( ( processedBitTableNames.find( bitTableName ) == processedBitTableNames.end() ) &&
			( bitTableName.compare( "" ) != 0 ) ){
					    
			// DEBUGGING ..
			//cout << "UserInterface:: printBufferDataResults() Point D .." << endl << endl;
						
			// this tableName was not yet encountered in the bp,
			// so print up all rows associated with this tableName,
			// in all bp blocks (including this one) ..
			returnPages = getQueryManager()->selectBuffer( bitTableName );
						
			if ( !getMessageManager()->hasErrorMessage() ) {
				
			// DEBUGGING ..
			//cout << "UserInterface:: printBufferDataResults() Point E .." << endl << endl;
						
				if ( returnPages->size() > 0 ) {
					
					anyPagesReturned = true;
					
					// DEBUGGING ..
					//cout << "UserInterface:: printBufferDataResults() Point F .." << endl << endl;
						
						
					// print out any results that are returned ..
						
					tableName = bitTableName;
					
					cout << "table: \"" << tableName << "\"" << endl;
								
					printResults( returnPages, new map<int, DataPage* >(), new vector<string>() );
								               //map<int, DataPage > &fileDataPages, vector<string> selectColumnNames
				
					cout << endl;
				
				}
								
						
			} // end "if ( !getMessageManager()->hasErrorMessage() )" ..
						
			// track that this tableName was processed
			// and its rows should not be printed again ..
			processedBitTableNames.insert( bitTableName );
						
		}
					
	} // end "for()" ..


	if ( anyPagesReturned == false ) {
			
		// indicate that the query of the buffer returned no results ..

		// DEBUGGING ..
		//cout << "UserInterface:: printBufferDataResults() Point H .." << endl << endl;

		getMessageManager()->setResultMessage( "\nNO RESULTS ...\n" );
			
	} // end "( bitTableName.compare( "" ) != 0 )" ..

	
} // end "printBufferDataResults()" ..



void UserInterface:: resetValues(){
	
	//cout << "UserInterface:: resetValues() begin .." << endl << endl;
	
	firstField.clear();
	
	secondField.clear();
	
	thirdField.clear();
	
	fourthField.clear();
	
	inputTokens->clear();
	
	tableName.clear();
	
	tableNames->clear();
	
	numberOfTables = 0;
	
	createColumnNames->clear();					
	
	numberOfColumns = 0;									
	
	insertValues->clear();						
	
	selectColumnNames->clear();
	
	whereColumnNamePairs->clear();
	
	whereColumnNameValuePairs->clear();
	
	// clear out the error string to prepare for next command ..
	errorString.clear();
	
	hasParserError = false;
		
	// clear out the string holding user input to prepare for next command ..
	commandInput.clear();
	
	getMessageManager()->resetErrorStatus();
	
} // end "resetValues()" ..


void UserInterface:: changePath( vector<string>* inputTokens_, bool isWindows_ ){
	
	inputTokens = inputTokens_;
	
	isWindows = isWindows_;
	
	updateFileDirectoryPaths();
	
}


bool UserInterface:: parseFirstField(){
	
	//cout << "UserInterface:: parseFirstField() start .." << endl;
	
	bool firstFieldFound = false;
	
	string first = "create";
	string second = "insert";
	string third = "select";
	string fourth = "print";
	string fifth = "set";
	string sixth = "commit";
	string seventh = "show";
	string eighth = "run";
	vector<string> firstFieldTitles;
	firstFieldTitles.push_back( first );
	firstFieldTitles.push_back( second );
	firstFieldTitles.push_back( third );
	firstFieldTitles.push_back( fourth );
	firstFieldTitles.push_back( fifth );
	firstFieldTitles.push_back( sixth );
	firstFieldTitles.push_back( seventh );
	firstFieldTitles.push_back( eighth );
	
	int fieldNumber = 0;
		
	firstFieldFound = validateKeywordField( fieldNumber, firstFieldTitles );
	
	// DEBUGGING ..
	//cout << "firstFieldFound: " << firstFieldFound << endl;
	//if ( firstFieldFound == false ) {
		//string errorString = "\nKeyword missing.  Expected \"create\", \"insert\", \"select\", \"print\", \"set\" or \"commit\".";
		//qm->bm->sm->mm->setErrorStringP( errorString );
	//}
	//cout << "UserInterface:: parseFirstField() hasParserError: " << hasParserError << endl << endl;
	
	return firstFieldFound;
	
	//cout << "UserInterface:: parseFirstField() end .." << endl;

	
} // end "parseFirstField()" ..


bool UserInterface:: hasSecondField(){
	
	//cout << "UserInterface:: hasSecondField() start .." << endl;
	
	if ( inputTokens->size() >= 2 ) {
		
		return true;
		
	} else {
		
		if ( firstField.compare( "create" ) == 0 ){
			
			//string errorString = "\nKeyword missing.  Expected \"table\" or \"view\".";
			string errorString = "\nKeyword missing.  Expected \"table\".";
			getMessageManager()->setErrorStringP( errorString );
			
		} else if ( firstField.compare( "insert" ) == 0 ) {
			
			string errorString = "\nKeyword missing.  Expected \"into\".";
			getMessageManager()->setErrorStringP( errorString );
			
		} else if ( firstField.compare( "select" ) == 0 ) {
			
			string errorString = "\nKeyword or column name(s) missing.  Expected \"*\" or comma-delimited column name(s).";
			getMessageManager()->setErrorStringP( errorString );
		
		} else if ( firstField.compare( "print" ) == 0 ) {
			
			string errorString = "\nKeyword missing.  Expected \"buffer\", \"catalog\", \"table\", \"hit\", \"hitrate\".";
			getMessageManager()->setErrorStringP( errorString );
		
		} else if ( firstField.compare( "set" ) == 0 ) {
			
			string errorString = "\nKeyword missing.  Expected \"timer\", \"path\".";
			getMessageManager()->setErrorStringP( errorString );
			
		} else if ( firstField.compare( "show" ) == 0 ) {
			
			string errorString = "\nKeyword missing.  Expected \"path\".";
			getMessageManager()->setErrorStringP( errorString );
			
		} else if ( firstField.compare( "run" ) == 0 ) {
			
			string errorString = "\nKeyword missing.  Expected \"batch\".";
			getMessageManager()->setErrorStringP( errorString );
			
		}

		// DEBUGGING ..
		//cout << "UserInterface:: hasSecondField() hasParserError: " << hasParserError << endl << endl;
		//cout << "UserInterface:: hasSecondField() end .." << endl;

		return false;	
		
	}
	
} // end "hasSecondField()" ..


bool UserInterface:: parseRemainingFields(){
	
	// DEBUGGING ..
	//cout << "UserInterface:: parseRemainingFields() start .." << endl;
	
	bool remainingFieldsFound = false;
	
	string firstFieldAccepted;
	
	
	firstFieldAccepted = "create";
	
	// if FIRST fields is "create" ..
	if ( firstField.compare( firstFieldAccepted ) == 0 ) {
		
		parseCreateStatement();

	}
	
	
	firstFieldAccepted = "insert";
	
	// if FIRST fields is "insert" ..
	if ( firstField.compare( firstFieldAccepted ) == 0 ) {
		
		parseInsertStatement();
		
	}
	
	
	firstFieldAccepted = "select";
	
	// if FIRST fields is "select" ..
	if ( firstField.compare( firstFieldAccepted ) == 0 ) {
		
		parseSelectStatement();
		
	}
	
	
	firstFieldAccepted = "print";
	
	// if FIRST fields is "print" ..
	if ( firstField.compare( firstFieldAccepted ) == 0 ) {
		
		parsePrintStatement();
		
	}
	
	
	firstFieldAccepted = "set";
	
	// if FIRST fields is "set" ..
	if ( firstField.compare( firstFieldAccepted ) == 0 ) {
		
		parseSetStatement();
		
	}
	
	firstFieldAccepted = "show";
	
	// if FIRST fields is "show" ..
	if ( firstField.compare( firstFieldAccepted ) == 0 ) {
		
		parseShowStatement();
		
	}
	
	firstFieldAccepted = "run";
	
	// if FIRST fields is "run" ..
	if ( firstField.compare( firstFieldAccepted ) == 0 ) {
		
		parseRunStatement();
		
	}
	
	// DEBUGGING ..
	//cout << "UserInterface:: parseRemainingFields() hasParserError: " << hasParserError << endl << endl;
	//cout << "UserInterface:: parseRemainingFields() end .." << endl;
	
	return remainingFieldsFound;
	
} // end "parseRemainingFields()" ..


void UserInterface:: parseCreateStatement(){
	
	// DEBUGGING ..
	//cout << "UserInterface:: parseCreateStatement() start .." << endl << endl;
	
	int fieldNumber = 1;
	vector<string> secondFieldTitles;
	
	string first = "table";		// "create table xyz(aa,bb,cc)" ..
	//string second = "view";
	secondFieldTitles.push_back( first );
	//secondFieldTitles.push_back( view );
		
	bool secondFieldFound = validateKeywordField( fieldNumber, secondFieldTitles );
		
	// DEBUGGING ..
	//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
		
	if ( secondFieldFound ) {
			
		if( secondField.compare( "table" ) == 0 ) {
			
			// DEBUGGING ..
			//cout << "parseCreateStatement() create table .." << endl << endl;
			
			// "create table", get the tableName ..
			fieldNumber = 2;
			bool thirdFieldFound = validateTableName( fieldNumber );
			
			if ( thirdFieldFound ){
				
				// DEBUGGING ..
				//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
				//cout << "parseCreateStatement() validateColumnNamesCreate .." << endl << endl;
				
				// "create table", get the createColumnNames ..
				fieldNumber = 3;
				bool columnNamesValid = validateColumnNamesCreate( fieldNumber );
				
			}
			
		} else if ( secondField.compare( "index" ) == 0 ) {
			
			
			// DEBUGGING ..
			//cout << "parseCreateStatement() validateIndexName .." << endl << endl;
		
			
			// "create index" ..
			fieldNumber = 2;
			bool thirdFieldFound = validateIndexName( fieldNumber );
		
			if ( thirdFieldFound ) {
				
				// "create index" ..
				first = "on";
				vector<string> fourthFieldTitles;
				fourthFieldTitles.push_back( first );
				fieldNumber = 3;
		
				// DEBUGGING ..
				//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
				//cout << "parseCreateStatement() validateKeywordField - II .." << endl << endl;
		
				bool fourthFieldFound = validateKeywordField( fieldNumber, fourthFieldTitles );
			
				if( fourthFieldFound ){
					
					// DEBUGGING ..
					//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
					//cout << "parseCreateStatement() validateTableName .." << endl << endl;
					
					// "create index", get the tableName ..
					fieldNumber = 4;
					bool fifthFieldFound = validateTableName( fieldNumber );
					
					if( fifthFieldFound ){
						
						// DEBUGGING ..
						//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
						//cout << "parseCreateStatement() validateColumnNamesCreate .." << endl << endl;
						
						// "create index", get the columnName ..
						fieldNumber = 5;
						bool columnNamesValid = validateColumnNamesCreateIndex( fieldNumber );
						
						if ( columnNamesValid ) {
							
							// if the command has no more keywords / words, stop parsing ..
							// (NOTE: "create index I on xyz(aa)" == 6 tokens)
							// otherwise, check for "no duplicates" (optional keyword ==  2 tokens) ..
							if ( inputTokens->size() > 6 ) {
							
								// "create index", get the option "no duplicates" ..
								first = "no";
								vector<string> sixthFieldTitles;
								sixthFieldTitles.push_back( first );
								fieldNumber = 6;
		
								// DEBUGGING ..
								//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
								//cout << "parseCreateStatement() validateKeywordField() - III .." << endl << endl;
		
								bool sixthFieldFound = validateKeywordField( fieldNumber, sixthFieldTitles );
						
								if ( sixthFieldFound ) {
							
									// "create index", get the option "no duplicates" ..
									first = "duplicates";
									vector<string> seventhFieldTitles;
									seventhFieldTitles.push_back( first );
									fieldNumber = 7;
		
									// DEBUGGING ..
									//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
									//cout << "parseCreateStatement() validateKeywordField() - IV .." << endl << endl;
		
									bool seventhFieldFound = validateKeywordField( fieldNumber, seventhFieldTitles );
			
									// DEBUGGING ..
									//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
									//cout << "parseCreateStatement() set is Duplicates .." << endl << endl;
			
									if ( seventhFieldFound ) {
							
										// "create index, NO duplicates" ..
										isNoDuplicates = true;
							
										// DEBUGGING ..
										//cout << "commandInput( NO duplicates ): " << commandInput << endl << endl;
							
									} else {
							
										// "create index, duplicates" ..
										isNoDuplicates = false;
							
										// DEBUGGING ..
										//cout << "commandInput( duplicates ): " << commandInput << endl << endl;
							
									} // end "if( seventhFieldFound )" .. 
							
								} // end "if( sixthFieldFound )" ..
								
							} // end "if ( inputTokens.size() > 6 )" ..
							
						} // end "if( columnNamesValid )" .. 
						
					} // end "if( fifthFieldFound )" .. 
					
				} // end "if( fourthFieldFound )" .. 
				
			} // end "if ( thirdFieldFound )" .. 
		
		} // end "if( secondField.compare( "table" ) )" .. 
			
	} // end "if ( secondFieldFound )" .. 
	
		
} // end "parseCreateStatement()" ..


void UserInterface:: parseInsertStatement(){
	
	int fieldNumber = 1;
	vector<string> secondFieldTitles;
	
	string first = "into";
	secondFieldTitles.push_back( first );
		
	bool secondFieldFound = validateKeywordField( fieldNumber, secondFieldTitles );
		
	if ( secondFieldFound ){
			
		fieldNumber = 2;
		bool thirdFieldFound = validateTableName( fieldNumber );
		
		if ( thirdFieldFound ){
				
			first = "values";
			vector<string> fourthFieldTitles;
			fourthFieldTitles.push_back( first );
			fieldNumber = 3;
			bool fourthFieldFound = validateKeywordField( fieldNumber, fourthFieldTitles );
				
			if ( fourthFieldFound ){
					
				// get the insertValues ..
				fieldNumber = 4;
				bool insertValuesFound = validateInsertValues( fieldNumber );
				
				//if ( insertValuesFound = false ) {
					//string errorString = "\nMissing insert value(s).  Values not inserted.";
					//qm->bm->sm->mm->setErrorStringP( errorString );
				//} // end "if ( insertValuesFound = false )" ..
					
			} //else {
				//string errorString = "\nKeyword missing.  Expected \"values\".";
				//qm->bm->sm->mm->setErrorStringP( errorString );
			//} // end "if( fourthFieldFound )" ..
				
		} //else {
			//string errorString = "\nTable name missing.  Expected table name following \"into\" keyword.";
			//qm->bm->sm->mm->setErrorStringP( errorString );
		//} // end "if ( thirdFieldFound )" ..
			
	} //else {
		//string errorString = "\nKeyword missing.  Expected \"into\".";
		//qm->bm->sm->mm->setErrorStringP( errorString );
	//}
		
} // end "parseInsertStatement()" ..


void UserInterface:: parseSelectStatement(){
	
	//fields_t fieldNumber = two;
	int fieldNumber = 1;
	vector<string> secondFieldTitles;
	
	string first = "*";
	secondFieldTitles.push_back( first );
		
	// handles "select * from" statements ..
	bool secondFieldFound = validateKeywordFieldSelect( fieldNumber, secondFieldTitles );
		
	// DEBUGGING ..
	//cout << "UserInterface:: parseSelectStatement() hasParserError (I): " << hasParserError << endl << endl;
		
	if ( secondFieldFound == false ){
		
		// secondField ("*" keyword) not found, so 
		// proceed to check for secondField (column names) below ..
		
		// DEBUGGING ..
		//cout << "UserInterface:: Point A .." << endl << endl;
		
		//fieldNumber = 3;
		
		// handles "select aaa, bbb, ccc.. from" statements ..
		secondFieldFound = validateColumnNamesSelect( fieldNumber );
		
		// DEBUGGING ..
		//cout << "UserInterface:: parseSelectStatement() hasParserError (II): " << hasParserError << endl << endl;
		//cout << "UserInterface:: secondFieldFound: " << secondFieldFound << endl << endl;
		//cout << "numberOfColumns: " << numberOfColumns << endl;
		
		// handles "select a, b, .. 	..		..		  from"          statements ..
		//			   0   1  2  numberOfColumns  (1 + numberOfColumns)
		// prepare for "from" fields ..
		fieldNumber += numberOfColumns;	// 1 + numberOfColumns ..
	
		// DEBUGGING ..
		//cout << "UserInterface:: fieldNumber: " << fieldNumber << endl << endl;
		
	} else {
		
		// DEBUGGING ..
		//cout << "UserInterface:: Point B .." << endl << endl;
		//cout << "UserInterface:: parseSelectStatement() hasParserError (III): " << hasParserError << endl << endl;
		
		// handles "select * from" statements ..
		//			   0   1   2
		// prepare for "from" fields ..
		fieldNumber = 2;
		
		// "*" constitutes only a single column ..
		numberOfColumns = 1;
		
	}
	
	// DEBUGGING ..
	//cout << "UserInterface:: Point C .." << endl << endl;
	
	if ( secondFieldFound ) {
		
		// DEBUGGING ..
		//cout << "UserInterface:: Point D .." << endl << endl;
			
		vector<string> thirdFieldTitles;
	
		first = "from";
		thirdFieldTitles.push_back( first );
		
		bool thirdFieldFound = validateKeywordFieldSelect( fieldNumber, thirdFieldTitles );

		// DEBUGGING ..
		//cout << "UserInterface:: Point E .." << endl << endl;
		//cout << "UserInterface:: parseSelectStatement() hasParserError (IV): " << hasParserError << endl << endl;

		if ( thirdFieldFound ) {
			
			// DEBUGGING ..
			//cout << "UserInterface:: Point F .." << endl << endl;
				
			// get 1 or more tables ..
			fieldNumber += 1;
			bool fourthfieldFound = validateTableNames( fieldNumber );
			
			// DEBUGGING ..
			//cout << "UserInterface:: parseSelectStatement() hasParserError (V): " << hasParserError << endl << endl;
			//cout << "UserInterface:: Point G .." << endl << endl;

			if ( fourthfieldFound ) {
					
				fieldNumber += numberOfTables;
				vector<string> fifthFieldTitles;
	
				first = "where";
				string second = "having";
				fifthFieldTitles.push_back( first );
				fifthFieldTitles.push_back( second );
		
				bool fifthFieldFound = validateKeywordFieldSelect( fieldNumber, fifthFieldTitles );
					
				// DEBUGGING ..
				//cout << "UserInterface:: parseSelectStatement() hasParserError (VI): " << hasParserError << endl << endl;
					
				if ( fifthFieldFound ) {
						
					fieldNumber += 1;
					bool sixthFieldFound = validateColumnNamesWhere( fieldNumber );
					
					// DEBUGGING ..
					//cout << "UserInterface:: parseSelectStatement() hasParserError (VII): " << hasParserError << endl << endl;
					
					/*
					if ( sixthFieldFound == false ) {
						
						string errorString = "\nMissing expression pairs.  Expected pairs following \"where\" or \"having\" keyword.";
						string errorString = "\nMissing expression pairs.  Expected pairs following \"where\" keyword.";		
						qm->bm->sm->mm->setErrorStringP( errorString );
						
					} // end "if ( sixthFieldFound == false )" ..
					*/
					
				} else {
					
					// "fifthFieldFound == false" indicates that
					// query is of form "select .. from abc" with no
					// "where" or "having" clauses, so ok to do nothing here ..
					
				} // end "if ( fifthFieldFound )" ..
					
			} //else {
			
				//string errorString = "\nMissing table name(s) following \"from\" keyword.";	
				//getMessageManager()->setErrorStringP( errorString );
			
			//} // end "if ( fourthFieldFound )" ..
				
		} else {
						
			// DEBUGGING ..
			//cout << "no From Keyword .." << endl << endl;
			
			string errorString = "\nMissing keyword.  Expected \"from\".";	
			getMessageManager()->setErrorStringP( errorString );
			
		} // end "if ( thirdFieldFound )" ..
			
	} //else {
			
			//string errorString = "\nMissing keyword.  Expected \"*\" or column name(s) following \"select\" keyword.";	
			//qm->bm->sm->mm->setErrorStringP( errorString );
			
	//} // end "if ( secondFieldFound )" ..
		
	
} // end "parseSelectStatement()" ..


void UserInterface:: parsePrintStatement(){
	
	int fieldNumber = 1;
	vector<string> secondFieldTitles;
	
	string first = "catalog";
	string second = "table";
	string third = "buffer";
	string fourth = "hit";
	string fifth = "hitrate";
	string sixth = "index";			// print index I of table T ..
	secondFieldTitles.push_back( first );
	secondFieldTitles.push_back( second );
	secondFieldTitles.push_back( third );
	secondFieldTitles.push_back( fourth );
	secondFieldTitles.push_back( fifth );
	secondFieldTitles.push_back( sixth );
		
	bool secondFieldFound = validateKeywordField( fieldNumber, secondFieldTitles );
		
	if ( secondFieldFound ){
			
		string tableString = "table";
		string bufferString = "buffer";
		string hitString = "hit";
		string hitRateString = "hitrate";
		string indexString = "index";
			
		if ( secondField.compare( tableString ) == 0 ) {
				
			fieldNumber = 2;
			bool thirdFieldFound = validateTableNames( fieldNumber );
			
			//if ( thirdFieldFound == false ) {
				
				//string errorString = "\nMissing table name(s) following \"from\" keyword.";	
				//getMessageManager()->setErrorStringP( errorString );
				
			//}

				
		} else if ( secondField.compare( bufferString ) == 0 ) {
				
			first = "data";
			vector<string> thirdFieldTitles;
			thirdFieldTitles.push_back( first );
			fieldNumber = 2;
			
			// this indicates that query may be
			// "print buffer data" and not "print buffer" ..
			if( inputTokens->size() > fieldNumber ){
				
				bool thirdFieldFound = validateKeywordField( fieldNumber, thirdFieldTitles );
			
				
			}
			
			// this indicates that query may be
			// "print buffer" and not "print buffer data",
			// so ok to do nothing ( handled in "performCommand()" ) ..
			
				
		} else if ( secondField.compare( hitString ) == 0 ) {
				
			first = "rate";
			vector<string> thirdFieldTitles;
			thirdFieldTitles.push_back( first );
			fieldNumber = 2;
			bool thirdFieldFound = validateKeywordField( fieldNumber, thirdFieldTitles );
				
			//if ( thirdFieldFound == false ) {
					
				//string errorString = "\nKeyword missing.  Expected \"rate\".";
				//getMessageManager()->setErrorStringP( errorString );
					
			//} // end "if ( !thirdFieldFound )" ..
				
		} else if ( secondField.compare( indexString ) == 0 ) {
				
			// "print index" ..
			fieldNumber = 2;
			bool thirdFieldFound = validateIndexName( fieldNumber );
		
			if ( thirdFieldFound ) {
				
				// "create index" ..
				first = "of";
				vector<string> fourthFieldTitles;
				fourthFieldTitles.push_back( first );
				fieldNumber = 3;
		
				// DEBUGGING ..
				//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
				//cout << "parseCreateStatement() validateKeywordField - II .." << endl << endl;
		
				bool fourthFieldFound = validateKeywordField( fieldNumber, fourthFieldTitles );
			
				if( fourthFieldFound ){
					
					// DEBUGGING ..
					//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
					//cout << "parseCreateStatement() validateTableName .." << endl << endl;
					
					// "create index" ..
					first = "table";
					vector<string> fifthFieldTitles;
					fifthFieldTitles.push_back( first );
					fieldNumber = 4;
		
					// DEBUGGING ..
					//cout << "UserInterface:: performCommand() getMessageManager()->hasErrorMessage(): " << getMessageManager()->hasErrorMessage() << endl << endl;
					//cout << "parseCreateStatement() validateKeywordField - II .." << endl << endl;
		
					bool fifthFieldFound = validateKeywordField( fieldNumber, fifthFieldTitles );
			
					if( fifthFieldFound ){
					
						// "create index", get the tableName ..
						fieldNumber = 5;
						bool sixthFieldFound = validateTableName( fieldNumber );
						
					} // end "if( fifthFieldFound )" .. 
					
				} // end "if( fourthFieldFound )" .. 
				
			} // end "if ( thirdFieldFound )" .. 
				
		} else {
				
			// "print catalog", "print buffer" or "Print hitrate" so do nothing further ..
				
		} // end "if ( secondField.compare( tableString ) == 0 )" ..
			
	} //else {
			
		//string errorString = "\nKeyword missing.  Expected \"catalog\", \"table\", \"buffer\", \"hit rate\" or \"hitrate\".";
		//getMessageManager()->setErrorStringP( errorString );
			
	//} // end "if ( secondFieldFound )" ..
		
	
} // end "parsePrintStatement()" ..

void UserInterface:: updatePathOperatingSystem(){
	
	// helper method for "parseSetStatement()" ..
			
	if ( commandInput.size() >= 12 ) {
				
		// DEBUGGING ..
		//cout << "UserInterface:: Point B .." << endl << endl;
		//cout << "UserInterface:: commandInput: " << commandInput << endl << endl;

		//cout << endl << endl;
				
		// DEBUGGING ..
		//cout << "UserInterface:: ( setBoolean( isalpha( commandInput[9]("<<commandInput[9]<<") ) ): " << setBoolean( isalpha( commandInput[9] ) ) << endl << endl;
		//cout << "UserInterface:: ( commandInput[10] == ':' ): " << ( commandInput[10] == ':' ) << endl << endl;
		//cout << "UserInterface:: ( commandInput[11] == '\\' ): " << ( commandInput[11] == '\\' ) << endl << endl;
		
		string windowsSlash = "\\";
		
		size_t foundWindows = commandInput.find( windowsSlash );
		
		//cout << "UserInterface:: ( foundWindows != string::npos ): " << ( foundWindows != string::npos ) << endl << endl;
		
		if( foundWindows != string::npos ){
			
			// "windowsSlash" was found in "commandInput" ..
			
			isWindows = true;
			
			// assume that "commandInput" is windows-based ..
			return;
			
		}
			
/*
		// "set path" was issued for a Windows machine?
		// look for "C:\", "D:\", etc ..
		if ( ( setBoolean( isalpha( commandInput[9] ) ) ) &&
			 ( commandInput[10] == ':' ) && 
			 ( commandInput[11] == '\\' ) ) {
			     	 	
			 // DEBUGGING ..
			 //cout << "UserInterface:: Point C .." << endl << endl;
			
			isWindows = true;
			//updateFileDirectoryPaths( fileDirectoryStartPosition, true );
			
		}
*/
				
	}
	
	if ( commandInput.size() >= 10 ) {
		
		string unixSlash = "/";
		
		size_t foundUnix = commandInput.find( unixSlash );
		
		//cout << "UserInterface:: ( foundUnix != string::npos ): " << ( foundUnix != string::npos ) << endl << endl;
		
		if( foundUnix != string::npos ){
			
			// "unixSlash" was found in "commandInput" ..
			
			isWindows = false;
			
			// assume that "commandInput" is "unix" based ..
			
		}
/*
		// DEBUGGING ..
		//cout << "UserInterface:: Point D .." << endl << endl;
		//cout << "UserInterface:: ( commandInput[9] == '/' ): " << ( commandInput[9] == '/' ) << endl << endl;
				
		// "set path" was issued for a Unix/Linux machine?
		if ( commandInput[9] == '/' ) {
					
			// DEBUGGING ..
			//cout << "UserInterface:: Point E .." << endl << endl;
					
			isWindows = false;
			//updateFileDirectoryPaths( fileDirectoryStartPosition, false );
					
		}
*/
				
	}	
	
}

void UserInterface:: parseSetStatement(){
	
	//cout << "UserInterface:: parseSetStatement() start .." << endl;
	
	int fieldNumber = 1;
	vector<string> secondFieldTitles;
	
	string first = "timer";
	string second = "path";
	string third = "indexdebug";
	secondFieldTitles.push_back( first );
	secondFieldTitles.push_back( second );
	secondFieldTitles.push_back( third );
		
	bool secondFieldFound = validateKeywordField( fieldNumber, secondFieldTitles );
		
	if ( secondFieldFound ) {
			
		string timerString = "timer";
		string pathString = "path";
		string indexString = "indexdebug";
		
		if ( secondField.compare( timerString ) == 0 ) {
				
			first = "on";
			string second = "off";
			vector<string> thirdFieldTitles;
			thirdFieldTitles.push_back( first );
			thirdFieldTitles.push_back( second );
			fieldNumber = 2;
			bool thirdFieldFound = validateKeywordField( fieldNumber, thirdFieldTitles );
				
		} else if( secondField.compare( pathString ) == 0 ){
			
			// decide between Windows or Unix/Linux ..
			
			fieldNumber = 2;
			bool thirdFieldFound = validatePathName( fieldNumber );
			
			if( thirdFieldFound == true ){
				
				updatePathOperatingSystem();
				
			}
			
		} else if ( secondField.compare( indexString ) == 0 ) {
				
			first = "on";
			string second = "off";
			vector<string> thirdFieldTitles;
			thirdFieldTitles.push_back( first );
			thirdFieldTitles.push_back( second );
			fieldNumber = 2;
			bool thirdFieldFound = validateKeywordField( fieldNumber, thirdFieldTitles );
				
		} // end "if ( secondField.compare( timerString ) == 0 )" ..
			
	} // end "if ( secondFieldFound )" ..

	
} // end "parseSetStatement()" ..


void UserInterface:: parseShowStatement(){
	
	//cout << "UserInterface:: parseShowStatement() start .." << endl;
	
	int fieldNumber = 1;
	vector<string> secondFieldTitles;
	
	string first = "path";
	secondFieldTitles.push_back( first );
		
	bool secondFieldFound = validateKeywordField( fieldNumber, secondFieldTitles );

	
} // end "parseShowStatement()" ..

void UserInterface:: parseRunStatement(){
	
	//cout << "UserInterface:: parseRunStatement() start .." << endl;
	
	int fieldNumber = 1;
	vector<string> secondFieldTitles;
	
	string first = "batch";
	secondFieldTitles.push_back( first );
		
	bool secondFieldFound = validateKeywordField( fieldNumber, secondFieldTitles );
		
	if ( secondFieldFound ) {
			
		string batchString = "batch";
		
		if( secondField.compare( batchString ) == 0 ){
			
			// get the pathname location for the batch file
			// of commands to run ..
			
			fieldNumber = 2;
			bool thirdFieldFound = validateBatchFilePathName( fieldNumber );
			
		} // end "if ( secondField.compare( batchString ) == 0 )" ..
			
	} // end "if ( secondFieldFound )" ..

	
} // end "parseRunStatement()" ..

bool UserInterface:: validateKeywordField( int fieldNumber, vector<string> fieldTitles ){
	
	// DEBUGGING ..
	//cout << "now in parseCommand() >> validateKeywordField() .." << endl;
	
	string inputField;
	
	// DEBUGGING ..
	//cout << "fieldNumber: " << fieldNumber << endl;
	//cout << "fieldTitles.size(): " << fieldTitles.size() << endl;
	//string selectString = "select";
	
	if ( ( fieldNumber + 1 ) <= inputTokens->size() ){
		
		// DEBUGGING ..
		//cout << "( fieldNumber + 1 )[" << ( fieldNumber + 1 ) << "] <= inputTokens.size()[" << inputTokens.size() << "]: " << ( ( fieldNumber + 1 ) <= inputTokens.size() ) << endl;
		
		inputField = inputTokens->at( fieldNumber );
	
		inputField = stringToLower( inputField );
		
	}
	
	
	bool fieldFound = false;
	
	
	for(int i=0; i<fieldTitles.size(); i++){
		
		// DEBUGGING ..
		//cout << "fieldTitles[" << i << "]: " << fieldTitles[i] << endl;
		
		if( inputField.compare( fieldTitles[i] ) == 0 ){
		
			// DEBUGGING ..
			//cout << "\"" << fieldTitles[i] << "\" is the " << fieldNumber << " fields" << endl;

			if ( fieldNumber == 0 ){
				
				firstField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "firstField: " << firstField<< endl;
	
			} else if ( fieldNumber == 1 ) {

				secondField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "secondField: " << secondField << endl;
				
			} else if ( fieldNumber == 2 ) {

				thirdField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "thirdField: " << thirdField << endl;
				
			} else if ( fieldNumber == 3 ) {

				fourthField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "fourthField: " << fourthField << endl;
				
			} else if ( fieldNumber == 4 ) {

				fifthField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "fifthField: " << fifthField << endl;
				
			} else if ( fieldNumber == 5 ) {

				sixthField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "sixthField: " << sixthField << endl;
				
			} else if ( fieldNumber == 6 ) {

				seventhField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "seventhField: " << seventhField << endl;
				
			} else if ( fieldNumber == 7 ) {

				eighthField = inputField;
				
				fieldFound = true;
				
				// DEBUGGING ..
				//cout << "eighthField: " << eighthField << endl;
				
			} 
		
		}
		
	}
	
	// DEBUGGING ..
	//cout << "fieldFound: " << fieldFound << endl;
	//cout << "inputField: " << inputField << endl;
	
	if ( !fieldFound ){

		if ( inputField.size() > 0 ){
			
			string errorString = "\nInvalid command near \"";
		
			errorString += inputField;
			
			errorString += "\". Keyword missing.  Expected \"";
			
			if ( fieldTitles.size() == 1 ) {
				
				errorString += fieldTitles[0];
				
			} else {
				
				for(int i=0; i<(fieldTitles.size() - 1); i++){
				
					errorString += fieldTitles[i];
					
					errorString += "\", \"";
			
				}
				
				errorString += fieldTitles[ (fieldTitles.size() - 1) ];
				
				
				
			}
			
			errorString += "\".";

			getMessageManager()->setErrorStringP( errorString );
		
			
		} else {
			
			string errorString = "\nInvalid command; keyword missing.  Expected \"";
			
			if ( fieldTitles.size() == 1 ) {
				
				errorString += fieldTitles[0];
				
			} else {
				
				for(int i=0; i<(fieldTitles.size() - 1); i++){
				
					errorString += fieldTitles[i];
					
					errorString += "\", \"";
			
				}
				
				errorString += fieldTitles[ (fieldTitles.size() - 1) ];
				
				
				
			}
			
			errorString += "\".";
			
			getMessageManager()->setErrorStringP( errorString );
			
			
		} // end "if(inputField.size() > 0)" ..
		
	
		return false;
		
	} else {
		
		return true;
		
	}
	
} // end "validateKeywordField( fields_t fieldNumber, Array<string> fieldTitles )" ..


bool UserInterface:: validateKeywordFieldSelect( int fieldNumber, vector<string> fieldTitles ){
	
	// DEBUGGING ..
	//cout << "now in parseCommand() >> validateKeywordFieldSelect() .." << endl;
	
	string inputField;
	
	string expectedKeyword;
	
	// DEBUGGING ..
	//cout << "fieldNumber: " << fieldNumber << endl;
	//cout << "fieldTitles.size(): " << fieldTitles.size() << endl;
	//string selectString = "select";
	
	if ( ( fieldNumber + 1 ) <= inputTokens->size() ){
		
		// DEBUGGING ..
		//cout << "( fieldNumber + 1 )[" << ( fieldNumber + 1 ) << "] <= inputTokens.size()[" << inputTokens.size() << "]: " << ( ( fieldNumber + 1 ) <= inputTokens.size() ) << endl;
		
		inputField = inputTokens->at( fieldNumber );
	
		inputField = stringToLower( inputField );
		
	}
	
	
	bool fieldFound = false;
	
	
	for(int i=0; i<fieldTitles.size(); i++){
		
		// DEBUGGING ..
		//cout << "fieldTitles[" << i << "]: " << fieldTitles[i] << endl;
		
		if( inputField.compare( fieldTitles[i] ) == 0 ){
		
			// DEBUGGING ..
			//cout << "\"" << fieldTitles[i] << "\" is the " << fieldNumber << " fields" << endl;

			if ( fieldNumber == 0 ){
				
				// "select" keyword ..
				
				if ( inputField.compare("select") == 0 ){
				
					firstField = inputField;
				
					fieldFound = true;
				
				} else {
					
					expectedKeyword = "select";
					
				}
				
				// DEBUGGING ..
				//cout << "firstField: " << firstField<< endl;
	
			} else if ( fieldNumber == 1 ) {

				// "*" keyword ..
				
				if ( inputField.compare("*") == 0 ){

					secondField = inputField;
				
					fieldFound = true;
					
				} else {
					
					expectedKeyword = "*";	
					
				}
				
				// DEBUGGING ..
				//cout << "secondField: " << secondField << endl;
				
			} else if ( fieldNumber == ( 1 + numberOfColumns ) ) {

				// "from" keyword ..
				
				if ( inputField.compare("from") == 0 ){

					thirdField = inputField;
				
					fieldFound = true;
					
				} else {
					
					expectedKeyword = "from";
					
				}
				
				// DEBUGGING ..
				//cout << "secondField: " << secondField << endl;
				
			} else if ( fieldNumber == ( 2 + numberOfColumns + numberOfTables ) ) {

				// "where" keyword ..
				
				if ( inputField.compare("where") == 0 ){
					
					thirdField = inputField;
				
					fieldFound = true;
					
				}  else {
					
					expectedKeyword = "where";	
					
				}
				
				// DEBUGGING ..
				//cout << "thirdField: " << thirdField << endl;
				
			}
		
		}
		
	}
	
	// DEBUGGING ..
	//cout << "fieldFound: " << fieldFound << endl;
	
	if ( !fieldFound ){
	
/*
		if ( inputField.size() > 0 ){
			string errorString = "\nInvalid command near \"";
			errorString += inputField;
			errorString += "\".";
			errorString += "\nKeyword missing.  Expected \"";
			errorString += expectedKeyword;
			errorString += "\"";
			getMessageManager()->setErrorStringP( errorString );
		}
*/

		// DEBUGGING ..
		//cout << "fields is not found: " << inputField << endl << endl;
	
		return false;
		
	} else {
		
		return true;
		
	}
	
} // end "validateKeywordFieldSelect( fields_t fieldNumber, Array<string> fieldTitles )" ..



bool UserInterface:: validateField( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "now in parseCommand() >> validateField() .." << endl;
	
	string inputField;
	
	string selectString = "select";
	
	bool fieldFound = false;
	
	
	if ( ( fieldNumber + 1 ) <= inputTokens->size() ){
		
		inputField = inputTokens->at( fieldNumber );
	
		inputField = stringToLower( inputField );
		
	}
	
	
	// fields 2 of "select" statement can also be an "attributeName" ..
	if ( firstField.compare( selectString ) == 0 ) {
		
		if ( fieldNumber == 1 ) {
			
			// atribute names have max length = 3 ..
			if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){
				
				secondField = inputField;
				
				fieldFound = true;
			
			} else if ( inputField.size() > 0 ) {

				secondField = inputField.substr(0, 3);
				
				fieldFound = true;
				
			}
			
		} else if ( fieldNumber == 2 )  {
			
			// atribute names have max length = 3 ..
			if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){

				thirdField = inputField;
				
				fieldFound = true;
			
			} else {

				thirdField = inputField.substr(0, 3);
				
				fieldFound = true;
				
			}
			
		}
			
		// DEBUGGING ..
		//cout << "\"" << secondField << "\" is the " << fieldNumber << " fields" << endl;
			
	}
	
	// DEBUGGING ..
	//cout << "fieldFound: " << fieldFound << endl;
	
	if ( !fieldFound ){

		if ( inputField.size() > 0 ){
			
			string errorString = "\nInvalid command near \"";
		
			errorString += inputField;
			
			errorString += "\".";
		
			getMessageManager()->setErrorStringP( errorString );
			
		} else {
			
			string errorString = "\nInvalid command.";
		
			getMessageManager()->setErrorStringP( errorString );
			
		} // end "if ( inputField.size() > 0 )" ..


		return false;
		
		
	} else {
		
		return true;
		
	}
	
} // end "validateField( fields_t fieldNumber )" ..



bool UserInterface:: validateTableName( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "UserInterface:: parseCommand() >> validateTableName() .." << endl;
	
	string inputField;
	
	bool fieldTooLong = false;
	
	bool fieldNotThere = false;
	
	
	if ( ( fieldNumber + 1 ) <= inputTokens->size() ){
		
		inputField = inputTokens->at( fieldNumber );
	
		inputField = stringToLower( inputField );
		
	}
	
	
	// atribute names have max length = 3 ..
	if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){
		
		tableName = inputField;
			
	} else if ( inputField.size() > 0 ) {

		tableName = inputField.substr(0, 3);
		
		fieldTooLong = true;
				
	} else {

		fieldNotThere = true;
		
	}
	
	// DEBUGGING ..
	//cout << "\"" << tableName << "\" is the " << fieldNumber << " fields" << endl;
	
	
	if ( fieldNotThere ) {

		string errorString = "\nMissing table name.  Expected table name following keyword.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	} 
	
	if ( fieldTooLong ){

		string warningString = "\nTable name was too long (";
		
		warningString += intToStr( (int)inputField.size() );
		
		warningString += " chars).  Table name truncated: \"";
		
		warningString += inputField.substr(0, 3);
		
		warningString += "\".";
		
		getMessageManager()->setResultMessage( warningString );
		
	} 
	

	return true;

	
} // end "validateTableName( fields_t fieldNumber )" ..


bool UserInterface:: validateTableNames( int fieldNumber ){
	
	
	//cout << "now in parseCommand() >> validateTableNames() .." << endl;
	
	string inputField;
	
	bool fieldTooLong = false;
	
	//bool noWhereKeyword = false;
	
	int tableIndex = fieldNumber;
	
	int endTokenIndex = ( inputTokens->size() - 1 );
	
	//int numberOfTables = 0;
	
	if ( tableIndex <= endTokenIndex ) {
		
		// get out the tokens ..
		inputField = inputTokens->at( tableIndex );
	
		inputField = stringToLower( inputField );
		
	}
	
	
	while( ( inputField != "where" ) &&
		   ( inputField != "having" ) ) {
		
		// can't do this here; currently implemented below
		// after "numberOfTables" is calculated ..
		// MAX_COLUMNS_PER_TABLE is the limit
		// more than this is an error ..
		//if ( numberOfColumns > MAX_COLUMNS_PER_TABLE ){
			//tooManyColumns = true;
			//break;
		//}
		
		// DEBUGGING ..
		//cout << "inputField: " << inputField << endl;
		
		// attribute names have min length = 1, max length = 3 ..
		if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){
			
			tableNames->push_back( inputField );
			numberOfTables++;
			
		} else if ( inputField.size() > 0 ) {
				
			tableNames->push_back( inputField.substr(0, 3) );
			numberOfTables++;
		
			if ( !fieldTooLong )
				fieldTooLong = true;
				
		} 

		tableIndex++;
		
		if ( tableIndex <= endTokenIndex ){
		
			// get out the tokens ..
			inputField = inputTokens->at( tableIndex );
	
			inputField = stringToLower( inputField );
		
		} else {
			
			break;
			
		}
		
	}

	// DEBUGGING ..
	//for(int i=0; i<tableName.size(); i++){
		//cout << "\"" << tableName[i] << "\" is the " << i << " fields" << endl;
	//}
	
	if ( numberOfTables == 0 ) {

		errorString += "\nMissing table name(s).  Expected table name(s) following keyword.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	}
	
	// this is anticipating a "JOIN" operation (not yet implemented),
	// but works just as well with single tables ..
	if ( numberOfColumns > ( numberOfTables * MAX_COLUMNS_PER_TABLE ) ) {
		
		string errorString = "\nUser entered too many columns.  Maximum number of columns is ";
		
		errorString += intToStr( ( MAX_COLUMNS_PER_TABLE * numberOfTables ) );
		
		errorString += " for the ";
		
		errorString += intToStr( numberOfTables );
		
		errorString += " table(s) the user entered.";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( fieldTooLong ){
		
		string warningString = "\nTable name(s) too long.  Table name(s) truncated: ";
		
		for(int i=0; i < ( tableNames->size() - 1 ); i++){
		
			warningString += "\"";
			
			warningString += tableNames->at(i);
			
			warningString += "\", ";
		
		}
		
		warningString += "\"";
		
		warningString += tableNames->at( ( tableNames->size() - 1 ) );
		
		warningString += "\".";
		
		getMessageManager()->setResultMessage( warningString );
		
	}
	
	return true;

	
} // end "validateTableNames( fields_t fieldNumber )" ..



bool UserInterface:: validateIndexName( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "now in parseCommand() >> validateIndexName() .." << endl;
	
	string inputField;
	
	bool fieldTooLong = false;
	
	bool fieldNotThere = false;
	
	
	if ( ( fieldNumber + 1 ) <= inputTokens->size() ){
		
		inputField = inputTokens->at( fieldNumber );
	
		inputField = stringToLower( inputField );
		
	}
	
	
	// atribute names have max length = 3 ..
	if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){
		
		indexName = inputField;
			
	} else if ( inputField.size() > 0 ) {

		indexName = inputField.substr(0, 3);
		
		fieldTooLong = true;
				
	} else {

		fieldNotThere = true;
		
	}
	
	// DEBUGGING ..
	//cout << "\"" << indexName << "\" is the " << fieldNumber << " fields" << endl;
	
	
	if ( fieldNotThere ) {

		string errorString = "\nMissing index name.  Expected index name following keyword.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	} 
	
	if ( fieldTooLong ){

		string warningString = "\nIndex name was too long (";
		
		warningString += intToStr( (int)inputField.size() );
		
		warningString += " chars).  Index name truncated: \"";
		
		warningString += inputField.substr(0, 3);
		
		warningString += "\".";
		
		getMessageManager()->setResultMessage( warningString );
		
	} 
	

	return true;

	
} // end "validateIndexName( fields_t fieldNumber )" ..


bool UserInterface:: validateColumnNamesCreate( int fieldNumber ){
	
	//cout << "UserInterface::  parseCommand() >> validateColumnNamesCreate() .." << endl;
	
	string inputField;
	
	bool fieldTooLong = false;
	
	bool tooManyColumns = false;
	
	int colIndex = fieldNumber;
	
	int endTokenIndex = ( inputTokens->size() - 1 );
	
	// reset this parameter for current method invocation ..
	numberOfColumns = 0;
	
	
	while( colIndex <= endTokenIndex ) {
		
		// MAX_COLUMNS_PER_TABLE is the limit
		// more than this is an error ..
		
		// DEBUGGING ..
		//cout << "( numberOfColumns("<<numberOfColumns<<") > MAX_COLUMNS_PER_TABLE("<<MAX_COLUMNS_PER_TABLE<<") ): " << ( numberOfColumns > MAX_COLUMNS_PER_TABLE ) << endl << endl;
		
		// This is for the "create()" operation,
		// so this is different than the "JOIN" operation conditional statement
		// needed for the "select" statement ..
		// i.e. NOT "( numberOfTables * MAX_COLUMNS_PER_TABLE )" ..
		if ( numberOfColumns >= MAX_COLUMNS_PER_TABLE ){
			
			tooManyColumns = true;
			break;
			
		}
		
		// get out the tokens ..
		inputField = inputTokens->at( colIndex );
	
		inputField = stringToLower( inputField );
		
		// attribute names have min length = 1, max length = 3 ..
		if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){
			
			createColumnNames->push_back( inputField );
			numberOfColumns++;
			
		} else if ( inputField.size() > 0 ) {
				
			createColumnNames->push_back( inputField.substr(0, 3) );
			numberOfColumns++;
		
			if ( !fieldTooLong )
				fieldTooLong = true;
				
		} 
		
		colIndex++;
		
	}
	
	// DEBUGGING ..
	//for(int i=0; i<createColumnNames->size(); i++){
	//	cout << "\"" << createColumnNames->at(i) << "\" is the " << i << " field" << endl;
	//}
	
	if ( numberOfColumns == 0 ) {

		string errorString = "\nColumn name(s) missing.  Expected column name(s) following table name.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	}
	
	if ( tooManyColumns ) {
		
		string errorString = "\nUser entered too many columns.  Maximum number of columns is ";
		
		errorString += intToStr( MAX_COLUMNS_PER_TABLE );
		
		errorString += ".";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( fieldTooLong ){
			
		string warningString = "\nColumn name(s) too long.  Column name(s) truncated: ";
		
		for(int i=0; i < ( createColumnNames->size() - 1 ); i++){
		
			warningString += "\"";
			
			warningString += createColumnNames->at(i);
			
			warningString += "\", ";
		
		}
		
		warningString += "\"";
		
		warningString += createColumnNames->at( ( createColumnNames->size() - 1 ) );
		
		warningString += "\".";

		getMessageManager()->setErrorStringP( warningString );
		
		
	}
	
	return true;
	
} // end "validateColumnNamesCreate( fields_t fieldNumber )" ..



bool UserInterface:: validateColumnNamesCreateIndex( int fieldNumber ){
	
	//cout << "now in parseCommand() >> validateColumnNamesCreateIndex() .." << endl;
	
	string inputField;
	
	bool fieldTooLong = false;
	
	// reset this parameter for current method invocation ..
	numberOfColumns = 0;
	
	// MAX_COLUMNS_PER_TABLE is the limit
	// more than this is an error ..
		
	// DEBUGGING ..
	//cout << "( numberOfColumns("<<numberOfColumns<<") > MAX_COLUMNS_PER_TABLE("<<MAX_COLUMNS_PER_TABLE<<") ): " << ( numberOfColumns > MAX_COLUMNS_PER_TABLE ) << endl << endl;

		
	// get out the tokens ..
	inputField = inputTokens->at( fieldNumber );
	
	inputField = stringToLower( inputField );
		
	// attribute names have min length = 1, max length = 3 ..
	if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){
			
		createColumnNames->push_back( inputField );
		numberOfColumns++;
			
	} else if ( inputField.size() > 0 ) {
				
		createColumnNames->push_back( inputField.substr(0, 3) );
		numberOfColumns++;
		
		if ( !fieldTooLong )
			fieldTooLong = true;
				
	} 

	
	// DEBUGGING ..
	//for(int i=0; i<createColumnNames.size(); i++){
		//cout << "\"" << createColumnNames[i] << "\" is the " << i << " fields" << endl;
	//}
	
	if ( numberOfColumns == 0 ) {

		string errorString = "\nColumn name missing.  Expected column name following table name.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	}
	
	if ( fieldTooLong ){
			
		string warningString = "\nColumn name too long.  Column name truncated: ";
			
		warningString += createColumnNames->at(0);
		
		warningString += "\".";

		getMessageManager()->setErrorStringP( warningString );
		
		
	}
	
	return true;
	
} // end "validateColumnNamesCreateIndex( fields_t fieldNumber )" ..


bool UserInterface:: validateInsertValues( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "now in parseCommand() >> validateInsertValues() .." << endl;
	
	string inputField;
	
	string nonIntegerField;
	
	bool tooManyColumns = false;
	
	bool fieldNotInteger = false;
	
	int colIndex = fieldNumber;
	
	int endTokenIndex = ( inputTokens->size() - 1 );
	
	numberOfColumns = 0;
	
	// ask Catalog if the NUMBER of ATTRIBUTES is correct ..
	string tableName = getTableName();

	int catalogNumberOfAttributes = getCatalogManager()->getNumberOfAttributes( tableName );
	
	// ask Catalog if the table exists ..
	bool tableExists = getCatalogManager()->doesCatalogTableNameExist( tableName );
	
	// only bother processing the insert values if the table exists ..
	if ( tableExists ) {
		
		while( colIndex <= endTokenIndex ) {
		
			if ( numberOfColumns >= catalogNumberOfAttributes ){
			
				tooManyColumns = true;
				break;
			
			}
		
			// get out the tokens ..
			inputField = inputTokens->at( colIndex );
	
			inputField = stringToLower( inputField );
		
			// DEBUGGING ..
			//cout << "colIndex: " << colIndex << ", inputField: " << inputField.c_str() << endl;
		
			// attribute names have min length = 1, max length = 3 ..
			if ( isInteger( (char*)inputField.c_str() ) ){
			
				// DEBUGGING ..
				//cout << "inputField:" << inputField << endl;
			
				insertValues->push_back( inputField );
				numberOfColumns++;
			
			} else {
		
				nonIntegerField = inputField;
				fieldNotInteger = true;
				break;
		
			}
		
			colIndex++;
		
		}
		
		
	} // end "if ( tableExists )" ..
	
	
	// DEBUGGING ..
	//cout << "insertValues.size(): " << insertValues.size() << endl;
	
	// DEBUGGING ..
	//for(int i=0; i<insertValues.size(); i++){
		//cout << "\"" << insertValues[i] << "\" is the " << i << " fields" << endl;
	//}
	
	if ( ( numberOfColumns == 0 ) && ( !fieldNotInteger ) && ( tableExists ) ) {

		string errorString = "\nMissing insert value(s).  Values not inserted.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	}
	
	if ( tooManyColumns ) {
		
	 	string errorString = "\nTable \"";
	 		
	 	errorString += tableName;
	 		
	 	errorString += "\" has ";
	 		
	 	errorString += intToStr( catalogNumberOfAttributes );
	 		
	 	errorString += " attributes. User attempted to insert ";
	 		
	 	errorString += intToStr( ( numberOfColumns + 1 ) );
	 		
	 	errorString += " column values.  The record could not be inserted.";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( fieldNotInteger ){
		
		insertValues->clear();
			
		string errorString = "\nInsert value \"";
		
		errorString += nonIntegerField;
		
		errorString += "\" is not an integer.  Values not inserted.";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}

	
	return true;
	
} // end "validateInsertValues( int fieldNumber )" ..


bool UserInterface:: validateColumnNamesSelect( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "now in parseCommand() >> validateColumnNamesSelect() .." << endl;
	
	string inputField;
	
	bool fieldTooLong = false;
	
	bool tooManyColumns = false;
	
	bool noFromKeyword = false;
	
	int colIndex = fieldNumber;
	
	int endTokenIndex = ( inputTokens->size() - 1 );
	
	// reset this parameter for current method invocation ..
	numberOfColumns = 0;
	
	
	if ( colIndex <= endTokenIndex ) {
		
		// get out the tokens ..
		inputField = inputTokens->at( colIndex );
	
		inputField = stringToLower( inputField );
		
		// DEBUGGING ..
		//cout << "inputField: " << inputField << endl;
		
	}
	
	
	while( inputField != "from" ) {
		
		// NOTE: see "validateTableNames()" for consideration of "JOIN"
		//       on checking correct number of columns ..
		// MAX_COLUMNS_PER_TABLE is the limit
		// more than this is an error ..
		if ( numberOfColumns >= MAX_COLUMNS_PER_TABLE ){
			
			tooManyColumns = true;
			break;
			
		}
		
		// DEBUGGING ..
		//cout << "inputField: " << inputField << endl;
		
		// attribute names have min length = 1, max length = 3 ..
		if ( ( 0 < inputField.size() ) && ( inputField.size() < 4 ) ){
			
			selectColumnNames->push_back( inputField );
			numberOfColumns++;
			
		} else if ( inputField.size() > 0 ) {
				
			selectColumnNames->push_back( inputField.substr(0, 3) );
			numberOfColumns++;
		
			if ( !fieldTooLong )
				fieldTooLong = true;
				
		} 
		
		colIndex++;
		
		// DEBUGGING ..
		//cout << "validateColumnNamesSelect() ( colIndex("<<colIndex<<") <= endTokenIndex("<<endTokenIndex<<") ): " << ( colIndex <= endTokenIndex ) << endl << endl;
		
		if ( colIndex <= endTokenIndex ) {
			
			// get out the tokens ..
			inputField = inputTokens->at( colIndex );
	
			inputField = stringToLower( inputField );
			
		} else {
			
			// error, b/c got to the end of the statement without reaching the "from" keyword ..
			noFromKeyword = true;
			break;
			
		}
		
	}
	
	// DEBUGGING ..
	//for(int i=0; i<selectColumnNames.size(); i++){
		//cout << "\"" << selectColumnNames[i] << "\" is the " << i << " fields" << endl;
	//}
	
	if ( numberOfColumns == 0 ) {
		
		string errorString = "\nMissing keyword.  Expected \"*\" or column name(s) following \"select\" keyword.";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( noFromKeyword ) {
		
		cout << "validateColumnNamesSelect() noFromKeyword: " << noFromKeyword << endl << endl;
		
		string errorString = "\nKeyword missing.  Expected \"from\".";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	

	if ( tooManyColumns ) {
		
		errorString += "\nUser entered too many columns.  Maximum number of columns is ";
		
		errorString += intToStr( MAX_COLUMNS_PER_TABLE );
		
		errorString += ".";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	
	if ( fieldTooLong ){

		string warningString = "\nColumn name(s) too long.  Column name(s) truncated: ";
		
		for(int i=0; i < ( selectColumnNames->size() - 1 ); i++){
		
			warningString += "\"";
			
			warningString += selectColumnNames->at(i);
			
			warningString += "\", ";
		
		}
		
		warningString += "\"";
		
		warningString += selectColumnNames->at( ( selectColumnNames->size() - 1 ) );
		
		warningString += "\".";
		
		//qm->bm->sm->mm->setResultMessage( warningString );
		//qm->bm->sm->cm->mm->setResultMessage( warningString );
		getMessageManager()->setResultMessage( warningString );

		
	}
	
	// DEBUGGING ..
	//cout << "validateColumnNamesSelect() return true .." << endl << endl;
	
	return true;
	
} // end "validateColumnNamesSelect( int fieldNumber )" ..


bool UserInterface:: validateColumnNamesWhere( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "UserInterface:: validateColumnNamesWhere() start .." << endl;
	
	string inputField;
	
	int colIndex = fieldNumber;
	
	int endTokenIndex = ( inputTokens->size() - 1 );
	
	// track whether "columnName" is longer than the predefined max ..
	bool fieldTooLong = false;
	
	// track whether the number of condition "where" clauses is greater than 
	// max possible, given the number of tables, and the number of columns for each table ..
	bool tooManyConditions = false;
	
	// track whether the "where" clauses that are provided are complete
	// (i.e.: "aaa = 5 and bbb = 6", instead of "aaa = 5 bbb = 6") ..
	bool missingConjunction = false;
	
	// track whether the "where" clauses that are provided do not have extra conjunctions ..
	// (i.e.: "aaa = 5" instead of "aaa = 5 and")
	bool extraConjunction = false;
	
	// track whether the "where" clause has "missing clauses"
	// (i.e.: where aaa = 5 AND ..., etc)
	bool clauseMissingFollowingAndOr = false;
	
	// track whether the "columnName" exists in the catalog ..
	bool columnNameDoesNotExist = false;
	
	//track whether the LHS of the expression is a "columnName" ..
	bool columnNameIsNotAlphanumeric = false;
	
	// track whether 1 or more "columnNames" = "value" expressions are repeated incorrectly ..
	// ex: "a = 1000 and a = 2000" ..
	bool attributeRepeatedIncorrectly = false;
	
	// track whether 1 or more "columnNames" = "columnNames" expressions are repeated incorrectly ..
	// ex: "a = bb and a = bb" ..
	bool expressionRepeatedIncorrectly = false;
	
	// track whether "leftColumnName == inputField", which is an error ..
	// ex: "a = a" ..
	bool expressionColumnNamesEqual = false;
	
	// track whether a token is missing in the "where" clause expression ..
	int whereClauseTokenCount = ( ( endTokenIndex - fieldNumber ) + 1 );
	bool attributeMissing = false;
	
	// track "columnNames" in order to check for repeats ..
	vector<string> columnNames;
	
	// track "columnNames" in order to check for repeats ..
	//map<string, string> columnNamePairs;
	
	// track left "columnName" until it can be added along with right "columnName"
	// to the map "untruncatedColumnNamePairs" ..
	string leftColumnName = "";

	
	// max number of columns that any statement can operate on
	// is the predefined "max number of columns per table" times
	// the "number of tables" that were entered in the statement ..
	int MAX_NUMBER_OF_COLUMNS = MAX_COLUMNS_PER_TABLE * numberOfTables;
	
	// DEBUGGING ..
	//cout << "MAX_COLUMNS_PER_TABLE[" << MAX_COLUMNS_PER_TABLE << "] * numberOfTables[" << numberOfTables << "] = MAX_NUMBER_OF_COLUMNS[" << MAX_NUMBER_OF_COLUMNS << "]" << endl;
	
	// track the number of "where" clauses (separated by conjunctions) ..
	int conditionCounter = 1;
	
	// MAX number of simple "And" conditions ..
	// ( MAX number of simple "Or" conditions = [ ( n * (n-1) ) / 2 ] + (num of unique values)
	//   where left = aj=ak, and right = all the values possible for aj=x )
	// NOTE: _doesn't_ include non-simple (aj=ak and aj=al) or (aj=ak and aj=am) ..
	// max = [[ ( n * (n-1) ) / 2 ] + n ] ..
	// i.e: n = 2
	//    [r] = redundant
	// "where" clause columns: a, b, c
	//  .. where a = 1, b = 1, c = 1						--> "n"
	//		and       a = b, a = c
	//		and	b = a [r], b = c
	//		and	c = a [r], c = b [r]
	// 		-->   a = b, a = c, b = c						--> "((n-1) * n)/2"
	int MAX_NUMBER_OF_CONDITIONS = ( ( ( MAX_NUMBER_OF_COLUMNS * ( MAX_NUMBER_OF_COLUMNS - 1 ) ) * 0.5 ) + MAX_NUMBER_OF_COLUMNS );
	
	// DEBUGGING ..
	//cout << "( ( ( MAX_NUMBER_OF_COLUMNS[" << MAX_NUMBER_OF_COLUMNS << "] * ( MAX_NUMBER_OF_COLUMNS - 1 )[" << (MAX_NUMBER_OF_COLUMNS - 1) << "] ) * 0.5 ) + MAX_NUMBER_OF_COLUMNS[" << MAX_NUMBER_OF_COLUMNS << "] ) = MAX_NUMBER_OF_CONDITIONS[" << MAX_NUMBER_OF_CONDITIONS << "]" << endl;
	
	//// track whether parsing "columnName" or "columnNameValue" in the "where" clause ..
	//bool isColumnName = true;
	
	// track whether the current fields is "leftColumnName", "rightColumnName" or "rightColumnValue" ..
	int leftColumnIndex = colIndex; // + 0
	
	int rightColumnIndex = colIndex + 1;
	
	//int rightColumnValueIndex = 1;
	
	//int currentFieldPosition = leftColumnNameIndex;
	
	//if ( colIndex <= endTokenIndex ) {
		// get out the tokens ..
		//inputField = inputTokens.at( colIndex );
		//inputField = stringToLower( inputField );
		//cout << "inputField: " << inputField << endl;
		// track the number of clauses (separated by conjunctions) ..
		//conditionCounter++;
	//}
	
	// DEBUGGING ..
	//cout << "UI:: validateColumnNamesWhere() (Point A) .." << endl << endl;
	
	int expressionPairCounter = 0;
	
	while( colIndex <= endTokenIndex ) {
		
		// get out the tokens ..
		
		inputField = inputTokens->at( colIndex );
		
		// DEBUGGING ..
		//cout << "UI:: validateColumnNamesWhere() inputField: " << inputField << endl << endl;
		//cout << "UI:: validateColumnNamesWhere() ( colIndex("<<colIndex<<") <= endTokenIndex("<<endTokenIndex<<") ): " << ( colIndex <= endTokenIndex ) << endl << endl;
		
		//if ( isColumnName ) {
		if ( colIndex == leftColumnIndex ) {
			
			// track the number of attributes encountered in each expression pair ..
			expressionPairCounter++;
			
			// DEBUGGING ..
			//cout << "UI:: validateColumnNamesWhere() (Point B) ( colIndex("<<colIndex<<") == leftColumnIndex("<<leftColumnIndex<<") ): " << ( colIndex == leftColumnIndex ) << endl << endl;
			
			// parsing "columnName"		 (aaa* = 5) ..
			if ( !isAlphanumeric( ( (char*)inputField.c_str() ) ) ) {
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point C) BREAK !isAlphanumeric ( inputField("<<inputField<<") ) .. " << endl << endl;
				
				// the LHS of the expression must be a columnName ..
				columnNameIsNotAlphanumeric = true;
				break;
				
			}
			
			inputField = stringToLower( inputField );
			
			if ( inputField.size() >= 4 ) {
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point D) BREAK truncate( inputField("<<inputField<<") ) .. " << endl << endl;
				
				inputField = inputField.substr(0, 3);
				
			}
			
			if ( ( inputField.compare( "and" ) == 0 ) ||
			   ( ( inputField.compare( "or" ) == 0 ) ) ) {
			   	
			   // DEBUGGING ..
			   //cout << "UI:: validateColumnNamesWhere() (Point E) BREAK incomplete clause (and/or)( inputField("<<inputField<<") ) .. " << endl << endl;
			   	
				// do something about incomplete clauses ..
					
				// the "name" of a "where" clause expression 
				// pair is missing (leftColumnIndex shouldn't point to and/or) ..
				extraConjunction = true;
				break;
			   		
			}
			
			// check that the "columnName" exists in the catalog ..
			bool attributeExists = getCatalogManager()->doesCatalogTableAttributeNameExist( tableNames->at(0), inputField );
			
			if ( !attributeExists ){
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point F) BREAK doesNotExist( tableName("<<( (char*) tableName[0].c_str() )<<"), inputField("<<inputField<<") ) .. " << endl << endl;
				
				columnNameDoesNotExist = true;
				break;
				
			}
			
			// DEBUGGING ..
			//cout << "UI:: validateColumnNamesWhere() (Point G) .. " << endl << endl;
			
			// save left "columnName", in order to check against right "columnName"
			// below on next iteration, and also to avoid repeats of (aaa = 5) expressions ..
			leftColumnName = inputField;
			
			// increment index so that "colIndex" will eventually match 
			// with the next left "columnName" ..
			leftColumnIndex += 3;
			
			// DEBUGGING ..
			//cout << "inputField(A[" << colIndex << "]): " << inputField << endl;
			//cout << "whereColumnNames.size([" << whereColumnNames.size() << "]) > whereColumnValues.size([" << whereColumnValues.size() << "]): " << ( whereColumnNames.size() > whereColumnValues.size() ) << endl;
			
		} else if ( colIndex == rightColumnIndex ) {
			
			// DEBUGGING ..
			//cout << "UI:: validateColumnNamesWhere() (Point H) ( colIndex("<<colIndex<<") == rightColumnIndex("<<rightColumnIndex<<") ): " << ( colIndex == rightColumnIndex ) << endl << endl;
			
			// parsing "columnNameValue" (aaa = 5*) or "columnName" (aaa = bbb*) ..
			if ( isInteger( ( (char*)inputField.c_str() ) )  ) {
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point I) isInteger( inputField("<<inputField<<") ) .. " << endl << endl;
				
				// track the number of attributes encountered in each expression pair ..
				expressionPairCounter++;
				
				// parsing "columnNameValue" (aaa = 5*) ..
				
				// check whether the "columnName" is repeated ..
				for(int i=0; i<columnNames.size(); i++){
				
					if ( columnNames[i].compare( leftColumnName ) == 0 ){
						
						// DEBUGGING ..
						//cout << "UI:: validateColumnNamesWhere() (Point J) BREAK expressionRepeated (aaa = 1) ( leftColumnName ("<<leftColumnName<<") ) .. " << endl << endl;
					
						// the "columnName" = "value" expression is repeated incorrectly ..
						// ex: "a = 1000 and a = 2000" ..
						attributeRepeatedIncorrectly = true;
						break;
					
					}
				
				}
			
				if ( attributeRepeatedIncorrectly ) {
				
					// the "columnName" = "value" expression is repeated incorrectly ..
					// ex: "a = 1000 and a = 2000" ..
					break;
				
				} else {
					
					// track this columnName in order to compare it with
					// the LHS of expressions in the _LHS_ iterations that follow ..
					columnNames.push_back( leftColumnName );	
					
				}
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point K) insert( leftColumnName("<<leftColumnName<<"), inputField("<<inputField<<") ) .. " << endl << endl;
				
				// save the column Name / Value pairs ..
				// ( NOTE: repeats were handled above in "colIndex == leftColumnIndex" ) ..
				whereColumnNameValuePairs->insert( pair<string, int>( leftColumnName, strToInt( inputField ) ) );
				
			} else {
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point L) !isInteger( inputField("<<inputField<<") ) .. " << endl << endl;
				
				// parsing "columnName" (aaa = bbb*) ..
				if ( leftColumnName.compare( inputField ) == 0 ) {
					
					// "leftColumnName == inputField" is an error ..
					expressionColumnNamesEqual = true;
					break;
					
				}
				
				if ( !isAlphanumeric( ( (char*)inputField.c_str() ) ) ) {
					
					// DEBUGGING ..
					//cout << "UI:: validateColumnNamesWhere() (Point M) BREAK !isAlphanumeric ( inputField ("<<inputField<<") ) .. " << endl << endl;
				
					// the LHS of the expression must be a columnName ..
					columnNameIsNotAlphanumeric = true;
					break;
				
				}
			
				inputField = stringToLower( inputField );
			
				if ( inputField.size() >= 4 ) {
					
					// DEBUGGING ..
					//cout << "UI:: validateColumnNamesWhere() (Point N) BREAK truncate( inputField("<<inputField<<") ) .. " << endl << endl;
				
					inputField = inputField.substr(0, 3);
				
				}
				
				if ( ( inputField.compare( "and" ) == 0 ) ||
			   	   ( ( inputField.compare( "or" ) == 0 ) ) ) {
			   	
					// do something about incomplete clauses ..
					
					// DEBUGGING ..
					//cout << "UI:: validateColumnNamesWhere() (Point O) BREAK incomplete clause (and/or)( inputField("<<inputField<<") ) .. " << endl << endl;
					
					// the "value" of a "where" clause expression 
					// pair is missing (rightColumnIndex shouldn't point to and/or) ..
					extraConjunction = true;
					break;
			   		
				} else {
					
					// track the number of attributes encountered in each expression pair ..
					expressionPairCounter++;	
					
			   	}
			
				// check that the "columnName" exists in the catalog ..
				//bool attributeExists = qm->bm->cm->doesCatalogTableAttributeNameExist( ( (char*) tableName[0].c_str() ), inputField );
				bool attributeExists = getCatalogManager()->doesCatalogTableAttributeNameExist( tableNames->at(0), inputField );
			
				if ( !attributeExists ){
					
					// DEBUGGING ..
					//cout << "UI:: validateColumnNamesWhere() (Point P) BREAK doesNotExist( tableName("<<tableName<<"), inputField("<<inputField<<") ) .. " << endl << endl;
				
					columnNameDoesNotExist = true;
					break;
				
				}
				
				if( whereColumnNamePairs != NULL ){
					
					// check whether the "columnName" is repeated ..
					multimap<string, string>::iterator it_colNms;
					for( it_colNms = whereColumnNamePairs->begin(); it_colNms != whereColumnNamePairs->end(); it_colNms++ ){
						
								// if ( leftCol == leftCol && rightCol == rightCol ) || 
								// [i.e.: for "a=b and a=b", a==a && b==b] ..
							
						if ( ( ( ( it_colNms->first ).compare( leftColumnName ) == 0 ) &&
				       	   	 ( ( it_colNms->second ).compare( inputField ) == 0 ) ) ||
				       
				       		// if ( leftCol == rightCol && rightCol == leftCol ) {
				       		// [i.e.: for "a=b and b=a", a==a && b==b] ..
				       		
			 		 	 	 ( ( ( it_colNms->first ).compare( inputField ) == 0 ) &&
				       	   	   ( ( it_colNms->second ).compare( leftColumnName ) == 0 ) ) ) {
					
								// DEBUGGING ..
								//cout << "UI:: validateColumnNamesWhere() (Point Q) BREAK expressionRepeated (aaa = bbb) ( leftColumnName ("<<leftColumnName<<") == inputField ("<<inputField<<") ) .. " << endl << endl;
					
								// the "columnName" = "columnName" expression is repeated incorrectly ..
								// ex: "a = bb and a = bb" ..
								expressionRepeatedIncorrectly = true;
								break;

					
						} // end "if()" ..
			
			
					} // end "for()" ..
					
				} // end "if()" ..

			
				if ( expressionRepeatedIncorrectly ) {
				
					// the "columnName" = "value" expression is repeated incorrectly ..
					// ex: "a = bb and a = bb" ..
					break;
				
				} else {
				
					// the "columnName" = "value" expression NOT repeated incorrectly, so
					// track this columnName in order to compare it with the next expression ..
					if( whereColumnNamePairs != NULL )
						whereColumnNamePairs->insert( pair<string, string>( leftColumnName, inputField ) );

				}
				
				
			} // end "if ( isInteger( inputField ) )" ..
			
			
			// increment index so that "colIndex" will eventually match 
			// with the next right "columnNameValue" or "columnName" ..
			rightColumnIndex += 3;
			
			
		} // end "if ( colIndex == leftColumnIndex )" ..
		
		// DEBUGGING ..
		//cout << "UI:: validateColumnNamesWhere() (Point Q1) expressionPairCounter: " << expressionPairCounter << endl << endl;
		
		// increment in preparation for next token in parsed 
		// "where"-clause string ..
		colIndex++;
		
		
		if ( colIndex <= endTokenIndex ) {
			
			// DEBUGGING ..
			//cout << "UI:: validateColumnNamesWhere() (Point R) ( colIndex("<<colIndex<<") <= endTokenIndex("<<endTokenIndex<<") ): " << ( colIndex <= endTokenIndex ) << endl << endl;
		
			// get out the tokens ..
			inputField = inputTokens->at( colIndex );
			
			if ( !isInteger( ( (char*) inputField.c_str() ) ) ) {
				
				inputField = stringToLower( inputField );
				
			} else {
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point S) BREAK conjunctionMissedInsteadData( inputField("<<inputField<<") ) .. " << endl << endl;
				//conjunctionMissedInsteadData = true;
				//break;	
				
			}
	
			// if the query is entered properly,
			// the inputField here == either
			// 1) "and" or "or" ( when above was "aaa = bbb*" )..
			// 2) right "columnName" in "aaa = bbb*" ( when above was "aaa* = bbb" ) ..
			
			if ( ( inputField.compare( "and" ) != 0 ) &&
			   ( ( inputField.compare( "or" ) != 0 ) ) ) {
			   	
			   	// DEBUGGING ..
			   	//cout << "inputField(C[" << colIndex << "]): " << inputField << endl;
			   	//cout << "whereColumnNames.size([" << whereColumnNames.size() << "]) > whereColumnValues.size([" << whereColumnValues.size() << "]): " << ( whereColumnNames.size() > whereColumnValues.size() ) << endl;
			   	
			   	if ( colIndex != rightColumnIndex )  {
			   		
					// do something about incomplete clauses ..
					
					// DEBUGGING ..
					//cout << "UI:: validateColumnNamesWhere() (Point T) BREAK incomplete clause (and/or)( inputField("<<inputField<<") ) .. " << endl << endl;
					
					// the clause is missing an "and" or 
					// an "or" conjunction in the query ..
					missingConjunction = true;
					break;
			   		
			   	}
			   	
			} else {
				
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() (Point T1) expressionPairCounter: " << expressionPairCounter << endl << endl;
				
				if ( expressionPairCounter == 2 ) {
					
					// the number of attributes encountered in this "where" clause
					// expression pair must be equal to 2 ..
					
					// the inputField here == "and" or "or" ..
			   	
					// track the number of clauses ( separated by conjunctions ) ..
					conditionCounter++;
			   	
			   		if ( conditionCounter > MAX_NUMBER_OF_CONDITIONS ) {
			   		
			   			/* 
			   			 * User has entered more "where" clause conditions then 
			   			 * the maximum number possible ..
			   			 * 
			   			 * ( excluding redundancies and the fact that there is no
			   			 *   checking the catalog about whether attributes entered exist in the tables provided,
			   			 *   or whether the tables provided exist, yet ) ..
			   			 * 
			   			 * NOTE: this wont apply for simple "Or" clauses ..
			   			 * 
			   			 */
 
 						// DEBUGGING ..
 						//cout << "UI:: validateColumnNamesWhere() (Point U) BREAK tooManyConditions( inputField("<<inputField<<") ) .. " << endl << endl;
 
 						tooManyConditions = true;
					
			   			break;
			   		
			  		 }
				
					// skip the conjunction and get the next set of 
					// "columnName" and "columnNameValue" pairs ..
				
					// increment in preparation for next token in parsed 
					// "where"-clause string ..
					colIndex++;
				
					if ( colIndex > endTokenIndex ) {
					
						// get out the tokens ..
						//inputField = inputTokens.at( colIndex );
						//inputField = stringToLower( inputField );
					
						// DEBUGGING ..
						//cout << "inputField(C): " << inputField << endl;
						//cout << "UI:: validateColumnNamesWhere() (Point V) BREAK clauseMissingFollowingAndOr( inputField("<<inputField<<") ) .. " << endl << endl;
					
						// "and" or "or" keyword, without a following clause, is an error ..
						clauseMissingFollowingAndOr = true;
					
						break;
					
					
					} // end "if ( colIndex > endTokenIndex )" ..
					
				} else {
					
					// DEBUGGING ..
					//cout << "UI:: validateColumnNamesWhere() BREAK attributeMissing (Point V1) .." << endl << endl;
					
					// the number of attributes encountered in this "where" clause
					// expression pair must be equal to 2 ..
					attributeMissing = true;
					break;
					
				}  // end "" ..
				
				// reset this parameter for the next "where" clause expression
				// pair iterations ..
				expressionPairCounter = 0;
					
			} // end "if ( ( inputField.compare( "and" ) != 0 ) &&" ..
			
			
		} // end "if ( colIndex <= endTokenIndex )" ..
		
		
	} // end of "while()" ..
	
	// DEBUGGING ..
	//cout << "UI:: validateColumnNamesWhere() (Point W) .." << endl << endl;
	//cout << "UI:: validateColumnNamesWhere() whereClauseTokenCount: " << whereClauseTokenCount << endl << endl;
	//cout << "UI:: validateColumnNamesWhere() ( colIndex("<<colIndex<<") >= endTokenIndex("<<endTokenIndex<<") ): " << ( colIndex >= endTokenIndex ) << endl << endl;
		
		
	if ( ( attributeMissing == false ) && 
	     ( colIndex >= endTokenIndex ) ) {
			
		//cout << "UI:: validateColumnNamesWhere() (Point X) .." << endl << endl;
				
		// double-check if attributes are missing ..
		// also handles the case when attribute is missing from the 
		// final expression pair (in a set of multiple pairs) ..
		
		// check whether the "where" clause expressions are well-formed ..
		// ex: good = "where a = 1000 and a = bb"; 
		// ex: bad = "where a = 1000 and a = "
			
		// ex: select * from abc where a = 2000 and b = 2000 and ccc = 3000 and a = bb
		//				  token count: 1	2	 3	4	  5	  6	  7		 8	 9	10  11
		//                 first pair: ********
		//			  remaining pairs:          **************************************
				
		if ( ( whereClauseTokenCount <= 1 ) || 
		     ( whereClauseTokenCount == 3 ) || 
		     ( whereClauseTokenCount == 4 ) ){
				
			// DEBUGGING ..
			//cout << "UI:: validateColumnNamesWhere() BREAK attributeMissing (Point Y) .." << endl << endl;
				
			// this indicates that there are one or more attributes missing from the
			// "where" clause in the first clause attribute pair 
			// (should be "whereClauseTokenCount == 2") ..
			attributeMissing = true;
			
		} else if( whereClauseTokenCount > 4 ) {
			
			// DEBUGGING ..
			//cout << "UI:: validateColumnNamesWhere() (Point Z) .." << endl << endl;
			
			int numberTokensFirstPair = 2;				// ex: where "a = 1000" ..
		
			int numberTokensRemainingPairs_Each = 3;	// ex: where a = 1000 "and bb = 200" ..
		
			// b/c "(whereClauseTokenCount > 4)", must be whereClauseTokenCount == 5 or more, so
			// "numberTokensRemainingPairs_Total = (5 or more) - 2", which is "(3 or greater)" ..
			int numberTokensRemainingPairs_Total = ( whereClauseTokenCount - numberTokensFirstPair );
		
			if ( ( numberTokensRemainingPairs_Total % numberTokensRemainingPairs_Each ) != 0 ) {
			
				// DEBUGGING ..
				//cout << "UI:: validateColumnNamesWhere() BREAK attributeMissing (Point AA) .." << endl << endl;
			
				// since "numberTokensRemainingPairs_Total = (3 or greater)"
				// "( (3 or greater) % 3 != 0 )" will work correctly for remaining pairs ..
				
				// this indicates that there is an attribute missing from the "where"
				// clause following after the first clause attribute pair ..
				attributeMissing = true;
			}
			
		} // "if( whereClauseTokenCount > 2 )" ..
			
			
	} // end "if ( colIndex > endTokenIndex )" ..
	
	
	// DEBUGGING ..
	//for(int i=0, j=0; i<whereColumnNames.size(), j<whereColumnValues.size(); i++, j++){
		//cout << "\"" << whereColumnNames[i] << "\" is the " << i << " name fields" << endl;
		//cout << "\"" << whereColumnValues[i] << "\" is the " << j << " value fields" << endl;
	//}

	if ( extraConjunction ) {

		string errorString = "\nEncountered \"and\" or \"or\" conjunction in \"where\" clause pair inappropriately near \"";
					
		string inputFieldError = inputTokens->at( ( colIndex - 1 ) );
		inputFieldError = stringToLower( inputFieldError );
		errorString += inputFieldError;
					
		errorString += "\".";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( missingConjunction ) {

		errorString = "\nThe \"and\" or \"or\" keyword missing from \"where\" clause pair near \"";
					
		string inputFieldError = inputTokens->at( ( colIndex - 1 ) );
		inputFieldError = stringToLower( inputFieldError );
		errorString += inputFieldError;
					
		errorString += "\".";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( tooManyConditions ) {

		string errorString = "\nUser entered too many \"where\" clause conditions.  Maximum is \"";
					
		errorString += intToStr( MAX_NUMBER_OF_CONDITIONS );
					
		errorString += "\" clauses for the table(s) entered.";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	
	if ( expressionColumnNamesEqual ) {

		string errorString = "\n\"Where\" clause condition attribute set equal to itself incorrectly for attribute \"";
		
		errorString += leftColumnName;
		
		errorString += "\"\n";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( attributeRepeatedIncorrectly ) {

		string errorString = "\n\"Where\" clause condition(s) repeated incorrectly for attribute \"";
		
		errorString += leftColumnName;
		
		errorString += "\"\n";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( expressionRepeatedIncorrectly ) {

		string errorString = "\n\"Where\" clause condition(s) repeated incorrectly for condition \"";
		
		errorString += leftColumnName;
		
		errorString += " = ";
		
		errorString += inputField;
		
		errorString += "\"\n";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	
	if ( clauseMissingFollowingAndOr ) {

		string errorString = "\nClause missing following the keyword \"";
					
		errorString += inputField;
					
		errorString += "\".";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( columnNameIsNotAlphanumeric ) {

		string errorString = "\"where\" clause column name \"";
					
		errorString += inputField;
					
		errorString += "\" is not alphabetic.";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( columnNameDoesNotExist ) {

		string errorString = "\"where\" clause column name \"";
					
		errorString += inputField;
					
		errorString += "\" does not exist in the catalog for this table.";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	
	if ( attributeMissing ) {

		string errorString = "\nAn attribute is missing from the \"where\" clause near \"";
		
		errorString += leftColumnName;
		//errorString += inputField;
		
		errorString += "\"\n";
		
		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	
	if ( ( whereColumnNamePairs->size() == 0 ) &&
	     ( whereColumnNameValuePairs->size() == 0 ) ) {

		string errorString = "\nMissing clause(s) following the \"where\" keyword.";

		getMessageManager()->setErrorStringP( errorString );
		
		return false;
		
	}
	

	if ( fieldTooLong ){

		string warningString = "\n\"where\" clause column name(s) too long.  Column name(s) truncated: ";
		
		map<string, int>::iterator it_nameValue, it_checkNameValues;
		for( it_nameValue = whereColumnNameValuePairs->begin(); it_nameValue != whereColumnNameValuePairs->end(); it_nameValue++ ){
		
			warningString += "\"";
			
			warningString += ( it_nameValue->first );
			
			warningString += " = ";
			
			warningString += ( intToStr( ( (int) it_nameValue->second ) ) );
			
			it_checkNameValues = it_nameValue;
			it_checkNameValues++;
			if( it_checkNameValues != whereColumnNameValuePairs->end() ) {
				
				warningString += "\", ";
				
			} else if ( whereColumnNamePairs->size() == 0 ) {
				
				warningString += "\".";
				
			} // end "if()" ..
		
		
		} // end "for()" ..
		

		if( whereColumnNamePairs != NULL ){
			
			multimap<string, string>::iterator it_names, it_checkNames;
			for( it_names = whereColumnNamePairs->begin(); it_names != whereColumnNamePairs->end(); it_names++ ){
		
				warningString += "\"";
			
				warningString += ( it_names->first );
			
				warningString += " = ";
			
				warningString += ( it_names->second );
			
				it_checkNames = it_names;
				it_checkNames++;
				if( it_checkNames != whereColumnNamePairs->end() ) {
				
					warningString += "\", ";
				
				} else {
				
					warningString += "\".";
				
				} // end "if()" ..
		
		
			} // end "for()" ..
			
		} // end "if()" ..

		
		getMessageManager()->setResultMessage( warningString );
		

	} // end "if ( fieldTooLong )" ..

	
	return true;
	
	
} // end "validateColumnNamesWhere( int fieldNumber )" ..

bool UserInterface:: validatePathName( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "UserInterface:: parseCommand() >> validatePathName() .." << endl;
	
	string inputField;
	
	bool fieldNotThere = false;
	
	
	if ( ( fieldNumber + 1 ) <= inputTokens->size() ){
		
		inputField = inputTokens->at( fieldNumber );
	
		inputField = stringToLower( inputField );
		
	}
	
	
	// atribute names have max length = 3 ..
	if ( inputField.size() <= 0 ){

		fieldNotThere = true;
		
	}
	
	// DEBUGGING ..
	//cout << "\"" << tableName << "\" is the " << fieldNumber << " fields" << endl;
	
	
	if ( fieldNotThere ) {

		string errorString = "\nMissing path name.  Expected directory path name following keyword.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	}
	
	return true;

	
} // end "validatePathName( int fieldNumber )" ..


bool UserInterface:: validateBatchFilePathName( int fieldNumber ){
	
	// DEBUGGING ..
	//cout << "UserInterface:: parseCommand() >> validateBatchFilePathName() .." << endl;
	
	string inputField;
	
	bool fieldNotThere = false;
	
	
	if ( ( fieldNumber + 1 ) <= inputTokens->size() ){
		
		inputField = inputTokens->at( fieldNumber );
	
		inputField = stringToLower( inputField );
		
	}
	
	if ( inputField.size() <= 0 ){

		fieldNotThere = true;
		
	}
	
	if ( fieldNotThere ) {

		string errorString = "\nMissing path name.  Expected batch file path name following keyword.";
		
		getMessageManager()->setErrorStringP( errorString );

		return false;
		
	}
	
	return true;

	
} // end "validatePathName( int fieldNumber )" ..


/*
 * GETTERS ..
 */

string UserInterface:: getTableName(){
 	
 	return tableName;
 	
}

vector<string>* UserInterface:: getSelectColumnNames(){
 	
 	return selectColumnNames;
 	
}

multimap<string, string>* UserInterface:: getWhereColumnNamePairs(){
 	
 	return whereColumnNamePairs;
 	
}

map<string, int>* UserInterface:: getWhereColumnNameValuePairs(){
 	
 	return whereColumnNameValuePairs;
 	
}


/*
 * SETTERS ..
 */


void UserInterface:: setInputTokens( vector<string>* inputTokens_ ){
	
	inputTokens = inputTokens_;
	
}

void UserInterface:: setTableName( string tableName_ ){
	
	tableName = tableName_;
	
}

void UserInterface:: setSelectColumnNames( vector<string>* selectColumnNames_ ){
	
	selectColumnNames = selectColumnNames_;
	
}

void UserInterface:: setWhereColumnNamePairs( multimap<string, string>* whereColumnNamePairs_ ){
	
	whereColumnNamePairs = whereColumnNamePairs_;
	
}

void UserInterface:: setWhereColumnNameValuePairs( map<string, int>* whereColumnNameValuePairs_ ){
	
	whereColumnNameValuePairs = whereColumnNameValuePairs_;
	
}
	
/*
 * OLD / NOT USED ..
 */
 
 
 
 void UserInterface:: clearOldTablePageLocks() {
	
	// NOT CURRENTLY USING THIS METHOD ..
	// DIDN'T IMPLEMENT timed LOCKS ..
	
	int pageID = -1;
	
	DataHeader* header = new DataHeader();
	
	map<int, int> currentLocks;
	
	// store all of the old locks prior to waiting X minutes ..
	//for( int i=0; i < qm->bm->locks.size(); i++ ){
	for( int i=0; i < getBufferManager()->getLocks()->size(); i++ ){
				
		pageID = (*getBufferManager()->getLocks())[ getBufferManager()->getBufferInfoTable()->getTableName()->at( i ) ]->at( i );
				
		getStorageManager()->readHeader( (string) getBufferManager()->getBufferInfoTable()->getTableName()->at( pageID ).c_str(), header );
			
		// storing old locks ..
		currentLocks[ pageID ] = header->getHeaderEntries()->at( pageID )->getLockNumber();

	}
	
	time_t currentTime = time( NULL );
	
	time_t timeToWaitBeforeReactivating = 300; 	// 300 sec = 5 min
	
	time_t reactivateTime = currentTime + timeToWaitBeforeReactivating;
	
	while( true ) {
		
		// infinite loop, handling deactivation of active locks, 
		// following wait-time of X minutes ..
		// ends when "currentTime == reactivateTime" ..
		
		if ( currentTime == reactivateTime ){
			
			// the time has arrived to deactivate all old inactive locks ..
			
			for( int i=0; i < getBufferManager()->getLocks()->size(); i++ ){
				
				pageID = (*getBufferManager()->getLocks())[ getBufferManager()->getBufferInfoTable()->getTableName()->at( i ) ]->at( i );
				
				getStorageManager()->readHeader( getBufferManager()->getBufferInfoTable()->getTableName()->at( pageID ), header );
				
				if ( ( currentLocks[ pageID ] == header->getHeaderEntries()->at( pageID )->getLockNumber() ) &&
				     ( header->getHeaderEntries()->at( pageID )->getLockNumber() != 0 ) ){
					
					// if "hp.dataPageentries[ pageID ].lockNumber == 0", old lock is already deactivated ..
					// lockNumber has not changed in X minutes, so deactivate this lock ..
					header->getHeaderEntries()->at( pageID )->setLockNumber( 0 );
					
				}

			}
	
			// save the new DataHeader info to file ..
			getStorageManager()->writeHeader( (string)getBufferManager()->getBufferInfoTable()->getTableName()->at( pageID ).c_str(), header );
			
			// calculate a new "activateTime" ..
			currentTime = time( NULL );
			reactivateTime = currentTime + timeToWaitBeforeReactivating;
			
			
			// stop looping, old inactive locks cleared ..
			break;
			
		} // end "if ( currentTime == activateTime )" ..
		
		
	} // end "while( true )" ..

	
} // end "clearOldTablePageLocks()" ..


void UserInterface:: oldPrintBufferResults(){

	// this is called in "performCommand()" for "print buffer" command ..
	
	// print buffer meta data ..
				
	// determine the max lengths of fields for printing 
	// buffer meta data ..
	
				
	// last bp num printed out in column is "bit.bufferPoolSize" ..
	//int maxBPNumberSize = ( intToStr( qm->bm->bit.bufferPoolSize ) ).size();
	int maxBPNumberSize = ( intToStr( getBufferManager()->getBufferInfoTable()->getBufferPoolSize() ) ).size();
	
	string emptyTableName = "empty";
	
	int maxTableNameLength = -1;
				
	string bitTableName = "";
	
	string nonePageID = "none";
				
	int maxPageIDLength = -1;
				
	string bitPageIDString = "";
	
	int bitPageID = -1;
				
	for(int i=0; i<getBufferManager()->getBufferInfoTable()->getBufferPoolSize(); i++){
					
		bitTableName = getBufferManager()->getBufferInfoTable()->getTableName()->at( i );
		
		if ( bitTableName.size() == 0 ) {
			
			// this means bp block has no tableName yet ( tableName == "" ),
			// so output will be "empty" ..
			updateMaxStringLength( emptyTableName, maxTableNameLength );
			
		} else {
			
			updateMaxStringLength( bitTableName, maxTableNameLength );	
			
		}
					
		bitPageID = getBufferManager()->getBufferInfoTable()->getPageID()->at( i );
		
		if ( bitPageID < 0 ) {
			
			// this means bp block has no pageID yet ( bitPageID == -1 ),
			// so output will be "none" ..
			updateMaxStringLength( nonePageID, maxPageIDLength );
			
		} else {
			
			updateMaxStringLength( bitPageIDString, maxPageIDLength );	
			
		}
								
	} // end "for()" ..
	
				
	// print buffer meta data ..
				
	string bpNumberHeader = "buffer #";
				
	string fileHeader = "file / table";
				
	string pageIDHeader = "block #";
	
	int fieldPlusPaddingWidth = 0;
				
	int bpNumberColumnWidth = 0;
				
	int fileColumnWidth = 0;
				
	int pageIDColumnWidth = 0;
				
	if( maxBPNumberSize > bpNumberHeader.size() ) {
					
		//cout << bpNumberHeader;
		//for(int i=0; i<( maxBPNumberSize - bpNumberHeader.size() ); i++){
			//cout << " ";
		//}
					
		bpNumberColumnWidth = maxBPNumberSize;
					
	} else {
					
		// DEBUGGING ..
		//cout << bpNumberHeader;
					
		bpNumberColumnWidth = bpNumberHeader.size();
					
	}
	
	// "%-*s" : "-" is left justify, "*" is pad "paddingWidth" number of spaces,
	//          "s" is insert string ..
	printf("%-*s", bpNumberColumnWidth, ( (char*) bpNumberHeader.c_str() ) );
				
	// space between columns ..
	cout << "  ";
				
				
	if( maxTableNameLength > fileHeader.size() ) {
					
		//cout << fileHeader;
		//for(int i=0; i<( maxTableNameLength - fileHeader.size() ); i++){
			//cout << " ";
		//}
					
		fileColumnWidth = maxTableNameLength;
					
	} else {
				
		// DEBUGGING ..
		//cout << fileHeader;

		fileColumnWidth = fileHeader.size();
					
	}
	
	// "%-*s" : "-" is left justify, "*" is pad "paddingWidth" number of spaces,
	//          "s" is insert string ..
	printf("%-*s", fileColumnWidth, ( (char*) fileHeader.c_str() ) );
				
	// space between columns ..
	cout << "  ";
				
				
	if( maxPageIDLength > pageIDHeader.size() ) {
					
		// DEBUGGING ..
		//cout << pageIDHeader;
		//for(int i=0; i<( maxPageIDLength - pageIDHeader.size() ); i++){
			//cout << " ";
		//}

		pageIDColumnWidth = maxPageIDLength;
					
	} else {
					
		// DEBUGGING ..
		//cout << pageIDHeader;

		pageIDColumnWidth = pageIDHeader.size();
					
	}
	
	// "%-*s" : "-" is left justify, "*" is pad "paddingWidth" number of spaces,
	//          "s" is insert string ..
	printf("%-*s", pageIDColumnWidth, ( (char*) pageIDHeader.c_str() ) );
				
	cout << endl;
	
	
	
	// print the "divder" bar for this table ..
	
	for(int i=0; i<bpNumberColumnWidth; i++)
		cout << "=";	

	cout << "||";
	
	for(int i=0; i<fileColumnWidth; i++)
		cout << "=";	

	cout << "||";
	
	for(int i=0; i<pageIDColumnWidth; i++)
		cout << "=";	
	
	cout << endl;
	
	
	
	// begin printing the "print buffer" data ..
				
	string bpNumber = "";
				
	string pageIDString = "";
	
	string printString = "";
	
	//cout << "bpNumberColumnWidth: " << bpNumberColumnWidth << endl << endl;
				
	//for(int i=0; i<qm->bm->bit.bufferPoolSize; i++){
	for(int i=0; i<getBufferManager()->getBufferInfoTable()->getBufferPoolSize(); i++){
					
		bpNumber = intToStr( i );
					
		//cout << bpNumber;
					
		if( bpNumber.size() < bpNumberColumnWidth ) {
		
			// "%-*s" : "-" is left justify, "*" is pad "paddingWidth" number of spaces,
			//          "s" is insert string ..
			printf("%-*s", bpNumberColumnWidth, ( (char*) bpNumber.c_str() ) );
					
		} else {
			
			cout << bpNumber;
			
		}
					
		// space between columns ..
		cout << "  ";	
					
		//bitTableName = qm->bm->bit.tableName[i];
		bitTableName = getBufferManager()->getBufferInfoTable()->getTableName()->at( i );
		
		if ( bitTableName.size() == 0 ) {
			
			//cout << emptyTableName;
			printString = emptyTableName;
			
		} else {
					
			//cout << bitTableName;
			printString = bitTableName;
			
		}
		
		// "%-*s" : "-" is left justify, "*" is pad "paddingWidth" number of spaces,
		//          "s" is insert string ..
		printf("%-*s", fileColumnWidth, ( (char*) printString.c_str() ) );
					
		// space between columns ..
		cout << "  ";	
					
		bitPageID = getBufferManager()->getBufferInfoTable()->getPageID()->at( i );
					
		pageIDString = intToStr( bitPageID );
		
		if ( bitPageID < 0  ) {
			
			//cout << nonePageID;
			printString = nonePageID;
			
		} else {
					
			//cout << pageIDString;
			printString = pageIDString;
			
		}
					 
		// "%-*s" : "-" is left justify, "*" is pad "paddingWidth" number of spaces,
		//          "s" is insert string ..
		printf("%-*s", pageIDColumnWidth, ( (char*) printString.c_str() ) );
					 
 
		// end this row ..
		cout << endl;
					
	} // end "for()" ..	
	
	
} // end "UserInterface:: oldPrintBufferResults()" ..


void UserInterface:: oldPrintResults( map<int, DataPage* >* bpDataPages,  map<int, DataPage* >* fileDataPages, vector<string>* selectColumnNames ){
	
	// DEBUGGING ..
	//cout << "UserInterface:: printResults() for table \"" << tableName << "\":" << endl << endl;
	//cout << "UserInterface:: bpDataPages.size(): " << bpDataPages.size() << endl << endl;
	//cout << "UserInterface:: fileDataPages.size(): " << fileDataPages.size() << endl << endl;
	//cout << "UserInterface:: selectColumnNames.size(): " << selectColumnNames.size() << endl << endl;
	
	// DEBUGGING ..
	//list<int>::iterator it_LRU;
	//for(it_LRU = qm->bm->bit.LRUBlocks.begin(); it_LRU != qm->bm->bit.LRUBlocks.end(); it_LRU++){
			
		//cout << "UserInterface:: bit.LRUBlocks.(*it_LRU): " << (*it_LRU) << endl << endl;
			
	//} // end "for()" ..
	
	// DEBUGGING ..
	//for(it_LRU = qm->bm->bitSnapShot.LRUBlocks.begin(); it_LRU != qm->bm->bitSnapShot.LRUBlocks.end(); it_LRU++){
		//cout << "UserInterface:: bitSnapShot.LRUBlocks.(*it_LRU): " << (*it_LRU) << endl << endl;
	//} // end "for()" ..
	
	// DEBUGGING ..
	//vector< vector<bool> >::iterator it_saBlock;
	//vector<bool>::iterator it_sa;
	//for(it_saBlock = qm->bm->bit.rowAvailable.begin(); it_saBlock != qm->bm->bit.rowAvailable.end(); it_saBlock++){
		//for(it_sa = (*it_saBlock).begin(); it_sa != (*it_saBlock).end(); it_sa++){
			//cout << "UserInterface:: bit.rowAvailable.(*it_sa): " << (*it_sa) << endl << endl;	
		//}
		//cout << "---------------------" << endl << endl;
	//} // end "for()" ..
	
	// DEBUGGING ..
	//for(it_saBlock = qm->bm->bitSnapShot.rowAvailable.begin(); it_saBlock != qm->bm->bitSnapShot.rowAvailable.end(); it_saBlock++){
		//for(it_sa = (*it_saBlock).begin(); it_sa != (*it_saBlock).end(); it_sa++){
			//cout << "UserInterface:: bitSnapShot.rowAvailable.(*it_sa): " << (*it_sa) << endl << endl;	
		//}
		//cout << "---------------------" << endl << endl;
	//} // end "for()" ..
	
	
	if ( ( bpDataPages->size() > 0 ) || 
		 ( fileDataPages->size() > 0 ) )  {
	
		// DEBUGGING ..
		//cout << "UserInterface:: Point A" << endl;
	
		getCatalogManager()->readMetaData();
	
		// DEBUGGING ..
		//cout << "UserInterface:: Point B" << endl;
	
		CatalogTable* catalogTable = getCatalogManager()->getCatalogTable( tableName );
		
		vector<string>* headerStrings = new vector<string>();
		
		string ctaString;
		
		vector<int> headerLengths;
		
		map<int, string>* columnNames = new map<int, string>();
		
		if ( selectColumnNames->size() > 0  ) {
			
			// this handles "select a, bb, ccc from x" queries ..
			
			// match valid column names with their position order (from left to right)
			// as it is recorded in the catalog ..
			columnNames = getQueryManager()->groupSelectColumnNames( tableName, selectColumnNames );
			
			// DEBUGGING ..
			//cout << "UserInterface:: catalogTable.getNumberOfAttributes() (I): " << catalogTable.getNumberOfAttributes() << endl << endl;
			//cout << "UserInterface:: columnNames.size() (I): " << columnNames.size() << endl << endl;
			
			for(int i=0; i<catalogTable->getNumberOfAttributes(); i++){
				
				// DEBUGGING ..
				//cout << "UserInterface:: columnNames.find( "<<i<<" )->second (I): " << columnNames.find( i )->second << endl << endl;
				
				if( columnNames->find( i ) != columnNames->end() ){
					
					ctaString = columnNames->find( i )->second;
					
					// DEBUGGING ..
					//cout << "UserInterface:: ctaString (I): " << ctaString << endl << endl;
					
					headerStrings->push_back( ctaString );
					
					// DEBUGGING ..
					//cout << "UserInterface:: ctaString.size() (I): " << ctaString.size() << endl << endl;
		
					headerLengths.push_back( ctaString.size() );
					
				}
				
			} // end "for()" ..
			
		} else {

			// this handles "select * from x" queries ..
	
			// DEBUGGING ..
			//cout << "UserInterface:: Point C" << endl;
	
			vector<CatalogTableAttribute*>* ctas = catalogTable->getAttributes();
	
			// DEBUGGING ..
			//cout << "UserInterface:: Point D" << endl;
	
			vector<CatalogTableAttribute*>::iterator it_cta;
	
			for(it_cta = ctas->begin(); it_cta != ctas->end(); it_cta++){
	
				ctaString = ( *it_cta )->getTableAttributeName();
				
				// DEBUGGING ..
				//cout << "UserInterface:: ctaString (II): " << ctaString << endl << endl;
				
				headerStrings->push_back( ctaString );
				
				// DEBUGGING ..
				//cout << "UserInterface:: ctaString.size() (II): " << ctaString.size() << endl << endl;
		
				headerLengths.push_back( ctaString.size() );


			} // end "" ..
			
			
		} // end "if ( selectColumnNames.size() > 0  )" ..

		// DEBUGGING ..
		//cout << "UserInterface:: Point E" << endl;
	
		DataHeader* header = new DataHeader();
		getStorageManager()->readHeader( tableName, header );
	
		// DEBUGGING ..
		//cout << "UserInterface:: Point F" << endl << endl;
		//int totalNumberOfDataPages = hp.totalNumberOfDataPages;
		//for(int i=0; i<totalNumberOfDataPages; i++){
			//hp._headerEntries[i].filledRowsCount;
		//}
	
		vector<bool>* rowAvailable = new vector<bool>();
	
		int rowCount = 0;
	
		int headerEntryIndex = 0;
		
		int fieldCount = -1;
	
		if ( selectColumnNames->size() > 0  ) {
			
			// this handles "select a, bb, ccc from x" queries ..
			
			fieldCount = columnNames->size();
			
		} else {
			
			// this handles "select * from x" queries ..
			
			fieldCount = catalogTable->getNumberOfAttributes();
			
		}
		
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point A .." << endl << endl;
		//cout << "UserInterface:: fieldCount: " << fieldCount << endl;
		//cout << "UserInterface:: Point G" << endl;
		//int maxLength = -1;
		
		int bpIndex = 0;
	
		vector<int> maxDataLengths;
	
		int dataValue = -1;
	
		string dataString;
	
		map<int, DataPage* >::iterator it_dp;
	
		for(it_dp = bpDataPages->begin(); it_dp != bpDataPages->end(); it_dp++){
	
			// DEBUGGING ..
			//cout << "UserInterface:: Point G_1" << endl;
			
			if ( it_dp->first < 0 ) {
				
				// the "bpDataPage" value is found using a key = "bpIndex" 
				// ( see QueryManager::selectStarFromX() ),
				// where 
				// "bpIndex" = [ (-1) * (i+1) ]
				//        "i" = 0 to bp.getBufferPoolSize()
				// so
				// "bpIndex" = [ (-1) * ( (*it_dp).first + 1 ) ]
				// 	"(*it_dp).first" [no pageID] /\ "bpIndex"
				// 		        -1               /\     0
				//		        -2               /\     1
				//		        -3               /\     2, etc ..
				
				bpIndex = ( (-1) * ( it_dp->first + 1 ) );
				
				// the "bpDataPage" is from the BP ..
				rowAvailable = getBufferManager()->getBufferInfoTable()->getRowAvailable()->at( bpIndex );
				
				// DEBUGGING ..
				//cout << "UserInterface:: ( (*it_dp).first + 1 ) (I): " << ( (*it_dp).first + 1 ) << endl << endl;
				//cout << "UserInterface:: bpIndex (I): " << bpIndex << endl << endl;
				
			} else {
				
				// BP pages were captured in "bpDataPages" and file pages were captured in "fileDataPages",
				// The operation of capturing "fileDataPages" changes values in the "bit",
				// so in order to keep "bit" values that match the pages in "bpDataPages",
				// a "snap shot" of the "bit" that properly matches "bpDataPages" is stored in "qm->bm->bitSnapShot" ..
				// 	"(*it_dp).first" [pageID] /\ "bpIndex"
				//             0              /\     1
				//			   3              /\     2
				//             2              /\     0, etc ..

				bpIndex = getBufferManager()->getBitSnapShot()->getBufferInfoTableIndex( tableName, it_dp->first );
				
				// the "bpDataPage" is from the BP ..
				rowAvailable = getBufferManager()->getBitSnapShot()->getRowAvailable()->at( bpIndex );
				
				// DEBUGGING ..
				//cout << "UserInterface:: (*it_dp).first (II): " << (*it_dp).first << endl << endl;
				//cout << "UserInterface:: bpIndex (II): " << bpIndex << endl << endl;
				
			}
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point B .." << endl << endl;
	
			//rowCount = hp._headerEntries[ headerEntryIndex ].filledRowsCount;
			//rowCount = it_dp->second->getRecords()->size();
			rowCount = it_dp->second->getNumberOfRows();
		
			// DEBUGGING ..
			//cout << "UserInterface:: rowCount: " << rowCount << endl;
			//cout << "UserInterface:: fieldCount: " << fieldCount << endl << endl;
			//cout << "UserInterface:: bpDataPages.size(): " << bpDataPages.size() << endl << endl;
			//cout << "UserInterface:: bpDataPages[0].records.size(): " << bpDataPages[0].records.size() << endl << endl;
			//cout << "UserInterface:: bpDataPages[0].records[0].fields.size(): " << bpDataPages[0].records[0].fields.size() << endl << endl;
			//cout << "UserInterface:: bpIndex: " << bpIndex << endl << endl;
		
			// need to add this to DataHeaderEntry .. and update BM.insertInto() .. !
			//numberOfSlots = hp._headerEntries[ headerEntryIndex ].numberOfSlots;
		
			for(int i=0; i<rowCount; i++){

				// DEBUGGING ..
				//cout << "UserInterface:: Point G["<<i<<"]_2" << endl << endl;
				//cout << "UserInterface:: rowAvailable["<<i<<"]: " << rowAvailable[i] << endl << endl;
			
				if ( rowAvailable->at(i) == false ) {
				
					// DEBUGGING ..
					//cout << "UserInterface:: Point G["<<i<<"]_3" << endl << endl;
				
					for(int j=0; j<fieldCount; j++){
					
						// DEBUGGING ..
						//cout << "UserInterface:: Point G["<<i<<"]["<<j<<"]_4" << endl;
						//cout << "UserInterface:: ( *it_dp ).second.records["<<i<<"].fields["<<j<<"].Value: " << ( *it_dp ).second.records[i].fields[j].Value << endl << endl;

						//dataValue = it_dp->second->getRecords()->at( i ).getFields()->at( j ).getValue();
						dataValue = it_dp->second->getEntries( i )->at( j );
					
						dataString = intToStr( dataValue );
		
						// DEBUGGING ..
						//cout << "UserInterface:: dataString: " << dataString << endl << endl;
						//cout << "UserInterface:: maxDataLengths.size(): " << maxDataLengths.size() << endl << endl;
		
						if( maxDataLengths.size() < fieldCount ){
						
							// DEBUGGING ..
							//cout << "UserInterface:: Point G["<<i<<"]["<<j<<"]_5" << endl << endl;
							//cout << "UserInterface:: dataString.size(): " << dataString.size() << endl << endl;
					
							maxDataLengths.push_back( dataString.size() );
					
						} else if ( maxDataLengths[j] < dataString.size() ) {
						
							// DEBUGGING ..
							//cout << "UserInterface:: Point G["<<i<<"]["<<j<<"]_6" << endl << endl;
							//cout << "UserInterface:: dataString.size(): " << dataString.size() << endl << endl;
					
							maxDataLengths[j] = dataString.size();
					
						} // end "if()" ..
				
					} // end "for()" ..
				
				} // end "if()" .. 
			
			} // end "for()" ..
		
			// DEBUGGING ..
			//cout << "UserInterface:: bpIndex: " << bpIndex << endl << endl;
			//bpIndex++;
		
			
		} // end "for()" ..
	
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point C .." << endl << endl;
		//cout << "UserInterface:: Point H" << endl;
		
		
		for(it_dp = fileDataPages->begin(); it_dp != fileDataPages->end(); it_dp++){
			
			// DEBUGGING ..
			//cout << "UserInterface:: Point H_1" << endl;
	
			headerEntryIndex = it_dp->first;
	
			//rowCount = hp._headerEntries[ headerEntryIndex ].filledRowsCount;
			//rowCount = it_dp->second->getRecords()->size();
			rowCount = it_dp->second->getNumberOfRows();
		
			// DEBUGGING ..
			//cout << "UserInterface:: rowCount: " << rowCount << endl;
			//cout << "UserInterface:: fieldCount: " << fieldCount << endl << endl;
			//cout << "UserInterface:: fileDataPages.size(): " << fileDataPages.size() << endl << endl;
			//cout << "UserInterface:: fileDataPages[0].records.size(): " << fileDataPages[0].records.size() << endl << endl;
			//cout << "UserInterface:: fileDataPages[0].records[0].fields.size(): " << fileDataPages[0].records[0].fields.size() << endl << endl;
			//cout << "UserInterface:: headerEntryIndex: " << headerEntryIndex << endl << endl;
		
			// need to add this to DataHeaderEntry .. and update BM.insertInto() .. !
			//numberOfSlots = hp._headerEntries[ headerEntryIndex ].numberOfSlots;
		
			// the "selectDataPage" is a page-from-file ..
			rowAvailable = header->getHeaderEntries()->at( headerEntryIndex )->getRowAvailable();
			// rowAvailable = filePrintTable.rowAvailable[ headerEntryIndex ]; // make this work ..

			//for(int i=0; i<numberOfSlots; i++){
			for(int i=0; i<rowCount; i++){

				// DEBUGGING ..
				//cout << "UserInterface:: Point H["<<i<<"]_2" << endl << endl;
				//cout << "UserInterface:: rowAvailable["<<i<<"]: " << rowAvailable[i] << endl << endl;
			
				if ( rowAvailable->at(i) == false ) {
				
					// DEBUGGING ..
					//cout << "UserInterface:: Point H["<<i<<"]_3" << endl << endl;
				
					for(int j=0; j<fieldCount; j++){
					
						// DEBUGGING ..
						//cout << "UserInterface:: Point H["<<i<<"]["<<j<<"]_4" << endl;
						//cout << "UserInterface:: ( *it_dp ).second.records["<<i<<"].fields["<<j<<"].Value: " << ( *it_dp ).second.records[i].fields[j].Value << endl << endl;

						//dataValue = it_dp->second->getRecords()->at( i ).getFields()->at( j ).getValue();
						dataValue = it_dp->second->getEntries( i )->at( j );
					
						dataString = intToStr( dataValue );

						// DEBUGGING ..
						//cout << "UserInterface:: dataString: " << dataString << endl << endl;
						//cout << "UserInterface:: maxDataLengths.size(): " << maxDataLengths.size() << endl << endl;
		
						if( maxDataLengths.size() < fieldCount ){
						
							// DEBUGGING ..
							//cout << "UserInterface:: Point H["<<i<<"]["<<j<<"]_5" << endl << endl;
							//cout << "UserInterface:: dataString.size(): " << dataString.size() << endl << endl;
					
							maxDataLengths.push_back( dataString.size() );
					
						} else if ( maxDataLengths[j] < dataString.size() ) {
						
							// DEBUGGING ..
							//cout << "UserInterface:: Point H["<<i<<"]["<<j<<"]_6" << endl << endl;
							//cout << "UserInterface:: dataString.size(): " << dataString.size() << endl << endl;
					
							maxDataLengths[j] = dataString.size();
					
						} // end "if()" ..
				
					} // end "for()" ..
				
				} // end "if()" .. 
			
			} // end "for()" ..
		
			// DEBUGGING ..
			//headerEntryIndex++;
			//cout << "UserInterface:: headerEntryIndex: " << headerEntryIndex << endl << endl;


		} // end "for()" ..
		
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point D .." << endl << endl;
		//cout << "UserInterface:: Point J" << endl;
	
		vector<int> fieldWidths;
	
		// DEBUGGING ..
		//cout << "UserInterface:: maxDataLengths.size(): " << maxDataLengths.size() << endl << endl;
		//cout << "UserInterface:: headerLengths.size(): " << headerLengths.size() << endl << endl;
		//cout << "UserInterface:: fieldWidths.size(): " << fieldWidths.size() << endl << endl;
	
		for(int i=0; i<fieldCount; i++){
		
			// DEBUGGING ..
			//cout << endl << "UserInterface:: maxDataLengths["<<i<<"]: " << maxDataLengths[i] << endl << endl;
			//cout << endl << "UserInterface:: headerLengths["<<i<<"]: " << headerLengths[i] << endl << endl;
		
			if( maxDataLengths[i] < headerLengths[i] ) {
			
				// DEBUGGING ..
				//cout << "UserInterface:: Point J["<<i<<"]_1" << endl;
			
				// if (f1 < H1) fieldWidth = H1 ..
				fieldWidths.push_back( headerLengths[i] );
			
			} else if ( maxDataLengths[i] > headerLengths[i] ) {
			
				// DEBUGGING ..
				//cout << "UserInterface:: Point J["<<i<<"]_2" << endl;
				
				// if (f1 > H1) fieldWidth = f1 ..
				fieldWidths.push_back( maxDataLengths[i] );
			
			} else if ( maxDataLengths[i] == headerLengths[i] ) {
			
				// DEBUGGING ..
				//cout << "UserInterface:: Point J["<<i<<"]_3" << endl;
			
				// if (f1 == H1) fieldWidth = H1 or f1..
				fieldWidths.push_back( headerLengths[i] );
				//fieldWidths.push_back( maxDataLengths[i] );
			
			} // end "if()" ..
		
		} // end "for()" ..
	
		// DEBUGGING ..
		//cout << "UserInterface:: Point K" << endl << endl;
		//cout << "UserInterface:: printResults() Point E .." << endl << endl;
	
	
		/*
		 * print out the "header" row ..
		 */ 
	
		// DEBUGGING ..
		//cout << "UserInterface:: rowCount: " << rowCount << endl << endl;
	
		int backFill = -1;
	
		int backFillHalf = -1;
		
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() fieldCount: " << fieldCount << endl << endl;
	
		for(int i=0; i<fieldCount; i++){
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point E_1["<<i<<"] .." << endl << endl;
		
			// buffer whitespace ..
			cout << " ";
		
			backFill = ( fieldWidths[i] - headerLengths[i] );
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point E_2["<<i<<"] .." << endl << endl;
		
			if ( isEven( backFill ) ) {
				
				// DEBUGGING ..
				//cout << "UserInterface:: printResults() Point E_3["<<i<<"] .." << endl << endl;
			
				// i.e.: 2 = ( 4 / 2 ) ..
				backFillHalf = ( backFill / 2 );
			
			} else {
			
				// DEBUGGING ..
				//cout << "UserInterface:: printResults() Point E_4["<<i<<"] .." << endl << endl;
			
				// isOdd( backFill ) == true ..
			
				// i.e.: 3 = ( 5 / 2 ) + 1 = 2 + 1 ..
				backFillHalf = ( ( backFill / 2 ) + 1 );
			
			}
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point E_6["<<i<<"] .." << endl << endl;
		
			// print the "LEFT" backfill ..
			for(int j=0; j<backFillHalf; j++){
				
				cout << " ";
				
			}
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point E_7["<<i<<"] .." << endl << endl;
		
			// print the "HEADER" string ..
			cout << headerStrings->at(i);
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point E_8["<<i<<"] .." << endl << endl;
		
			// print the "RIGHT" backfill ..
			for(int j=0; j<(backFill - backFillHalf); j++){
				
				cout << " ";
				
			}
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point E_9["<<i<<"] .." << endl << endl;
		
			// buffer whitespace ..
			cout << " ";
		
		
		} // end "for()" ..
		
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point F .." << endl << endl;
		//cout << "Point K" << endl;
	
		cout << endl;
	
		int tableWidth = 0;
	
		for(int i=0; i<fieldWidths.size(); i++){
		
			tableWidth += fieldWidths[i];

		}
	
		int numberOfBufferSpaces = 2;
	
		tableWidth += ( numberOfBufferSpaces * fieldCount );
	
		for(int i=0; i<tableWidth; i++){
		
			cout << "=";

		}
	
		cout << endl;
	
	
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point G .." << endl << endl;
	
		/*
		 * print out the "data" rows ..
		 */ 



		backFill = -1;
	
		backFillHalf = -1;
	
		headerEntryIndex = 0;
	
		rowCount = 0;
	
		//fieldCount = catalogTable.getNumberOfAttributes(); // this is done much above ..
	
		// calculate the number of "DataPages" to print ..
		int numberOfDataPagesToPrint = 0;
		
		numberOfDataPagesToPrint = fileDataPages->size();
		
		numberOfDataPagesToPrint += bpDataPages->size();
		
		// DEBUGGING ..
		//cout << "UserInterface:: numberOfDataPagesToPrint: " << numberOfDataPagesToPrint << endl << endl;
		
		int minPageID = 0;
		
		int fileDataPageCounter = 0, bpDataPageCounter = 0;
		
		vector<int> minFilePageIDs;
		
		vector<int> minBPPageIDs;
		
		bool skipPageID = false;
		
		bool isMinFilePageID = false;
		
		bool minFilePageIDsPopped = false;
		
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point H .." << endl << endl;
		
		// track whether the BP has pageID read in from file ..
		bool pageIDsFoundInBP = false;
			
		int numberPageIDsFoundInBP = 0;
			
		for(it_dp = bpDataPages->begin(); it_dp != bpDataPages->end(); it_dp++){
				
			// search results from BP for pageID that were read in from file ..
				
			// DEBUGGING ..
			//cout << "(*it_dp).first: " << (*it_dp).first << endl << endl;
				
			if ( (*it_dp).first >= 0 ) {
					
				// this indicates that at least 1 DataPage in the BP
				// was read in from the file ..
				
				pageIDsFoundInBP = true;
					
				numberPageIDsFoundInBP++;
					
			}
				
		} // end "for()" ..
		
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point I .." << endl << endl;
		//cout << "UserInterface:: numberPageIDsFoundInBP: " << numberPageIDsFoundInBP << endl << endl;
		
		//int static const BP_PAGEID_COUNT = numberPageIDsFoundInBP;
		
		// print out the "DataPages" ..
		for(int i=0; i<numberOfDataPagesToPrint; i++){
			
			// look through all the "DataPages" (both file and BP) ..
			
			// DEBUGGING ..
			//cout << "UserInterface:: Point A .." << endl << endl;
			//cout << "-------------------------------------------" << endl << endl;
			//cout << "UserInterface:: DataPage["<<i<<"] start .." << endl << endl;
			
			// reset these parameter for current iteration ..
			minPageID = -1;
			
			fileDataPageCounter = 0;
			
			isMinFilePageID = false;
		
			// DEBUGGING ..
			//cout << "UserInterface:: Point B .." << endl << endl;
			//cout << "UserInterface:: printResults() Point J .." << endl << endl;
			
			// Handle "file DataPages" ..
			for(it_dp = fileDataPages->begin(); it_dp != fileDataPages->end(); it_dp++){
				
				// DEBUGGING ..
				//cout << "-------------------------------------------" << endl << endl;
				//cout << "UserInterface:: fileDataPages["<<fileDataPageCounter++<<"] start .." << endl << endl;
				//cout << "UserInterface:: Point C .." << endl << endl;
			
				// find the "minimum" pageID in "fileDataPages" map ..
				
				for(int j=0; j<minFilePageIDs.size(); j++){
					
					// look through all the "file DataPages" ..
					
					// DEBUGGING ..
					//cout << "UserInterface:: Point D .." << endl << endl;
					//cout << "UserInterface:: minFilePageIDs["<<j<<"]("<<minFilePageIDs[j]<<") == ( (*it_dp).first )("<<((*it_dp).first)<<"): " << ( minFilePageIDs[j] == ( (*it_dp).first ) ) << endl << endl;
					
					if( minFilePageIDs[j] == ( it_dp->first ) ){
						
						// DEBUGGING ..
						//cout << "UserInterface:: Point E .." << endl << endl;
						//cout << "UserInterface:: file pageID("<<( (*it_dp).first )<<") skipped .." << endl << endl;
						
						// this indicates that we've used this pageID already,
						// so skip it and look for the next "minPageID" ..
						skipPageID = true;
						break;
						
					}
					
				} // end "for()" ..
				
				if ( !skipPageID ) {
					
					// DEBUGGING ..
					//cout << "UserInterface:: Point F .." << endl << endl;
					//cout << "UserInterface:: ( minPageID("<<minPageID<<") >= ( (*it_dp).first )("<<((*it_dp).first)<<") ): " << ( minPageID >= ( (*it_dp).first ) ) << endl << endl;
					
					// (*it_dp).first ) > 0 ( != -1 ) ..
						
					// find the "minimum" pageID in "fileDataPages" map ..
					if ( ( minPageID >= ( it_dp->first ) ) ||
				    	 ( minPageID < 0 ) )  {
				    	 	
				    	// DEBUGGING ..
				    	//cout << "UserInterface:: Point G .." << endl << endl;
				    	//cout << "UserInterface:: minPageID: " << ( (*it_dp).first ) << endl << endl;
					
						// "minPageID >= ( (*it_dp).first )" ..
						// indicates that new "minPageID" found ..
						
						// "minPageID == -1" ..
						// indicates that search for "minPageID" just started for this iteration ..
					
						minPageID = ( it_dp->first );
						isMinFilePageID = true;
					
					}
					
				} else {
					
					// DEBUGGING ..
					//cout << "UserInterface:: Point H .." << endl << endl;
					
					// reset this parameter for next iteration ..
					skipPageID = false;	
					
				}
				
				
			} // end "for()" ..
			
			// update "minFilePageIDs" with the current "minPageID" ..
			if ( minPageID >= 0 ){
				
				// DEBUGGING ..
				//cout << "UserInterface:: Point I .." << endl << endl;
				
				minFilePageIDs.push_back( minPageID );
					
			}	
			
			// DEBUGGING ..
			//cout << "UserInterface:: Point J .." << endl << endl;
			
			// reset these parameters for its upcoming reuse ..
		
			minFilePageIDsPopped = false;
			
			int bpBlockIndex = -1;
			
			//numberPageIDsFoundInBP = BP_PAGEID_COUNT;
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point K .." << endl << endl;
			
			// Handle "BP DataPages" ..
			
			if ( pageIDsFoundInBP ) {
				
				// DEBUGGING ..
				//cout << "UserInterface:: Point K .." << endl << endl;
				
				// handle a BP that has both DataPages
				// 1) loaded in from file, and
				// 2) yet to be allocated and written to file ..
				
				if ( numberPageIDsFoundInBP > 0 ) {
					
					// DEBUGGING ..
					//cout << "UserInterface:: Point L .." << endl << endl;
					
					// First, handle the DataPages loaded in from file;
					// these have pageID >= 0 ..
					
					bpDataPageCounter = 0;
					
					for(it_dp = bpDataPages->begin(); it_dp != bpDataPages->end(); it_dp++){
					
						// DEBUGGING ..
						//cout << "UserInterface:: Point M .." << endl << endl;
						
						if ( it_dp->first < 0 ) {
						
							// since this section of code is looking for positive pageID,
							// skip the pageID equal to "-1" for now ..
							continue;
						
						}
					
						// find the "minimum" pageID in "bpDataPages" map ..
				
						// DEBUGGING ..
						//cout << "-------------------------------------------" << endl << endl;
						//cout << "UserInterface:: bpDataPages["<<bpDataPageCounter++<<"] start .." << endl << endl;
				
						for(int j=0; j<minBPPageIDs.size(); j++){
							
							// DEBUGGING ..
							//cout << "UserInterface:: Point N .." << endl << endl;
					
							// look through all the "BP DataPages" ..
				
							//cout << "UserInterface:: minBPPageIDs["<<j<<"]("<<minBPPageIDs[j]<<") == ( (*it_dp).first )("<<((*it_dp).first)<<"): " << ( minBPPageIDs[j] == ( (*it_dp).first ) ) << endl << endl;
					
							if( minBPPageIDs[j] == ( it_dp->first ) ){
								
								// DEBUGGING ..
								//cout << "UserInterface:: Point O .." << endl << endl;
								//cout << "UserInterface:: BP pageID("<<( (*it_dp).first )<<") skipped .." << endl << endl;
						
								// this indicates that we've used this pageID already,
								// so skip it and look for the next "minPageID" ..
								skipPageID = true;
						
							}
						
						}
				
						if ( !skipPageID ) {
							
							// DEBUGGING ..
							//cout << "UserInterface:: Point P .." << endl << endl;
							//cout << "UserInterface:: ( minPageID("<<minPageID<<") >= ( (*it_dp).first )("<<((*it_dp).first)<<") ): " << ( minPageID >= ( (*it_dp).first ) ) << endl << endl;

							// find the "minimum" pageID in "bpDataPages" map ..
							if ( ( minPageID >= ( it_dp->first ) ) ||
				    	 	 	 ( minPageID < 0 ) )  {
				    	 	 	 	
				    	 	 	// DEBUGGING ..
				    	 	 	//cout << "UserInterface:: Point Q .." << endl << endl;
				    	 	 	//cout << "UserInterface:: minPageID: " << ( (*it_dp).first ) << endl << endl;
					
								// "minPageID >= ( (*it_dp).first )" ..
								// indicates that new "minPageID" found ..
								
								// "minPageID == -1" ..
								// indicates that search for "minPageID" just started for this iteration, 
								// and that "fileDataPages" had no pageID ..		
						
								if ( !minFilePageIDsPopped && isMinFilePageID ) {
									
									// DEBUGGING ..
									//cout << "UserInterface:: Point R .." << endl << endl;
									//cout << "UserInterface:: minFilePageIDs popped .." << endl << endl;
							
									// since the "min" pageID here is from "bpDataPages",
									// remove the "minPageID" found above for "fileDataPages"
									// ( but only when above found "isMinFilePageID == true";
									//   this insures that older "minFilePageIDs" aren't mistakenly popped ) ..
									// NOTE: this should only occur at most _once_ per iteration ..
									minFilePageIDs.pop_back();
									minFilePageIDsPopped = true;
							
								}
					
								minPageID = ( it_dp->first );
								isMinFilePageID = false;
								
								// this means that the "minPageID" found in the
								// bp was used instead of the one from the file, so
								// update this parameter for the next iteration ..
								// (this shouldn't be outside of the loop, because
								//  if the file "minPageID" is used, don't want to 
								//  decrement this parameter)
								numberPageIDsFoundInBP--;
					
					
							} // end "if ( ( minPageID >= ( (*it_dp).first ) )" ..
					
					
						} else {
							
							// DEBUGGING ..
							//cout << "UserInterface:: Point S .." << endl << endl;
					
							// reset this parameter for next iteration ..
							skipPageID = false;	
					
						} // end "if ( !skipPageID )" ..
				
				
					} // end "for()" ..
				
					// DEBUGGING ..
					//cout << "UserInterface:: Point T .." << endl << endl;
					
					
				} else {
					
					// DEBUGGING ..
					//cout << "UserInterface:: Point U .." << endl << endl;
					
					// Second, handle the DataPages inserted into the BP
					// but that are not yet written to file;
					// these have pageID == -1 ..
					
					// "(*it_dp).first ) == -1"..
					// this indicates that block in BP hasn't been written to file yet,
					// so use the "LRU" block instead (as the current block to print),
					// because all "minPageIDs" that were "read-from-file" were already
					// handled above ..
					
					// NOTE: "minPageID >= 0" indicates that a "fileDataPage" minPageID
					// was located, so go with that instead.  Also, BP minPageID was 
					// all dealt with by this point ..
							
					if ( ( bpBlockIndex < 0 ) && ( minPageID < 0 ) ) {
						
						// DEBUGGING ..
						//cout << "UserInterface:: Point V .." << endl << endl;
						//cout << "UserInterface:: bpBlockIndex: " << bpBlockIndex << endl << endl;
						//cout << "UserInterface:: qm->bm->bitSnapShot.LRUBlocks.size(): " << qm->bm->bitSnapShot.LRUBlocks.size() << endl << endl;
							
						// "bpBlockIndex == -1" ..
						// this indicates that the "LRU" block is available for use ..
							
						// OR
							
						// "minPageID == -1" ..
						// this indicates that "minPageID" wasn't found yet,
						// so go ahead and use the "LRU" block .. 
						
						// get the index of the next LRU BP block, and update list ..
						bpBlockIndex = getBufferManager()->getBitSnapShot()->getLRUBlocks()->front();
						
						//cout << "UserInterface:: Point V_1 .." << endl << endl;
						//cout << "UserInterface:: bpBlockIndex: " << bpBlockIndex << endl << endl;
						getBufferManager()->getBitSnapShot()->getLRUBlocks()->pop_front();
						
						//cout << "UserInterface:: Point V_2 .." << endl << endl;
						getBufferManager()->getBitSnapShot()->getLRUBlocks()->push_back( bpBlockIndex );
						
						//cout << "UserInterface:: Point V_3 .." << endl << endl;
						isMinFilePageID = false;
						
						//cout << "UserInterface:: Point V_4 .." << endl << endl;
					
					} 

					
				} // end "if ( numberPageIDsFoundInBP > 0 )" ..
				
				// DEBUGGING ..
				//cout << "UserInterface:: Point W .." << endl << endl;
			
			} else {
				
				// DEBUGGING ..
				//cout << "UserInterface:: printResults() Point K .." << endl << endl;
				//cout << "UserInterface:: Point X .." << endl << endl;
					
				// There are _no_ DataPages loaded in from file, 
				// so only need to handle the DataPages inserted into the BP;
				// these have pageID == -1 ..
					
				// "(*it_dp).first ) == -1"..
				// this indicates that block in BP hasn't been written to file yet,
				// so use the "LRU" block instead (as the current block to print),
				// because all "minPageIDs" that were "read-from-file" were already
				// handled above ..
							
				// NOTE: "minPageID >= 0" indicates that a "fileDataPage" minPageID
				// was located, so go with that instead ..
					
				if ( ( bpBlockIndex < 0 ) && ( minPageID < 0 ) ) {
					
					// DEBUGGING ..
					//cout << "UserInterface:: Point Y .." << endl << endl;
							
					// "bpBlockIndex == -1" ..
					// this indicates that the "LRU" block is available for use ..
							
					// OR
							
					// "minPageID == -1" ..
					// this indicates that "minPageID" wasn't found yet,
					// so go ahead and use the "LRU" block .. 
						
					// get the index of the next LRU BP block, and update list ..
					//bpBlockIndex = qm->bm->bitSnapShot.LRUBlocks.front();
					bpBlockIndex = getBufferManager()->getBitSnapShot()->getLRUBlocks()->front();
					getBufferManager()->getBitSnapShot()->getLRUBlocks()->pop_front();
					getBufferManager()->getBitSnapShot()->getLRUBlocks()->push_back( bpBlockIndex );
					isMinFilePageID = false;
					
				} 
				
				
			} // end "if ( pageIDsFoundInBP )" ..
			
			// DEBUGGING ..
			//cout << "UserInterface:: Point Z .." << endl << endl;

			// update "minBPPageIDs" with the current "minPageID" ..
			if ( ( minPageID >= 0 ) && ( !isMinFilePageID ) ) {
				
				// DEBUGGING ..
				//cout << "UserInterface:: Point AA .." << endl << endl;
				
				minBPPageIDs.push_back( minPageID );
					
			}
			
			// DEBUGGING ..
			//cout << "UserInterface:: printResults() Point L .." << endl << endl;
			//cout << "UserInterface:: Point AB .." << endl << endl;
			//cout << "UserInterface:: minPageID: " << minPageID << endl << endl;
			//cout << "UserInterface:: bpBlockIndex: " << bpBlockIndex << endl << endl;
			
			// print out the "DataPage" for the current "minPageID" ..
			if( isMinFilePageID ){
				
				// DEBUGGING ..
				//cout << "UserInterface:: printResults() Point M .." << endl << endl;
				//cout << "UserInterface:: Point AC .." << endl << endl;
				//cout << "UserInterface:: isMinFilePageID: " << isMinFilePageID << endl << endl;
				
				DataPage* fileDataPage = fileDataPages->find( minPageID )->second;
				
				//rowCount = fileDataPage->getRecords()->size();
				rowCount = fileDataPage->getNumberOfRows();
				
				// DEBUGGING ..
				//cout << "UserInterface:: minPageID: " << minPageID << endl << endl;
				//cout << "UserInterface:: fileDataPages.size(): " << fileDataPages.size() << endl << endl;
				//cout << "UserInterface:: fileDataPage.records.size(): " << fileDataPage.records.size() << endl << endl;
				
				rowAvailable = header->getHeaderEntries()->at( minPageID )->getRowAvailable();
				
				int columnCount = 0;
				
				for(int i=0; i<rowCount; i++){
				
					// DEBUGGING ..
					//cout << "Point K_4["<<headerEntryIndex<<"]["<<i<<"]" << endl;
			
					if ( rowAvailable->at(i) == false ) {
					
						// DEBUGGING ..
						//cout << "Point K_5["<<headerEntryIndex<<"]["<<i<<"]" << endl;
				
						for(int j=0; j<fieldCount; j++){
						
							// DEBUGGING ..
							//cout << "Point K_6["<<headerEntryIndex<<"]["<<i<<"]["<<j<<"]" << endl;
					
							// buffer whitespace ..
							cout << " ";
					
							//dataValue = fileDataPage->getRecords()->at( i ).getFields()->at( j ).getValue();
							dataValue = fileDataPage->getEntries( i )->at( j );
						
							// DEBUGGING ..
							//cout << "Point K_7["<<headerEntryIndex<<"]["<<i<<"]["<<j<<"]" << endl;
					
							dataString = intToStr( dataValue );
						
							// DEBUGGING ..
							//cout << "Point K_8["<<headerEntryIndex<<"]["<<i<<"]["<<j<<"]" << endl;
					
							backFill = ( fieldWidths[j] - dataString.size() );
						
							// DEBUGGING ..
							//cout << "Point K_9["<<headerEntryIndex<<"]["<<i<<"]["<<j<<"]" << endl;
							//cout << "backFill: " << backFill << endl << endl;
					
							// print the backfill ..
							for(int k=0; k<backFill; k++){
				
								cout << " ";
				
							}
						
							// DEBUGGING ..
							//cout << "Point K_10["<<headerEntryIndex<<"]["<<i<<"]["<<j<<"]" << endl;
		
							// print the "DATA" fields ..
		
							cout << dataString;
						
							// DEBUGGING ..
							//cout << "Point K_11["<<headerEntryIndex<<"]["<<i<<"]["<<j<<"]" << endl;
		
							// buffer whitespace ..
							cout << " ";

				
						} // end "for()" ..
					
						// DEBUGGING ..
						//cout << "Point K_12["<<headerEntryIndex<<"]["<<i<<"]" << endl;
				
						// carriage return to next line ..
						cout << endl;
				
					} // end "if()" .. 
				
					// DEBUGGING ..
					//cout << "Point K_13["<<headerEntryIndex<<"]["<<i<<"]" << endl;
			
				} // end "for()" ..
				
				
			} else {
			
				// DEBUGGING ..
				//cout << "UserInterface:: printResults() Point N .." << endl << endl;
				//cout << "UserInterface:: !isMinFilePageID: " << !isMinFilePageID << endl << endl;
				
				DataPage* bpDataPage = new DataPage();
				
				if ( bpBlockIndex >= 0 ){
					
					// the "bpDataPage" value is found using a key = "bpOffset" 
					// ( see QueryManager::selectStarFromX() ),
					// where 
					// "bpOffset" = [ (-1) * (i+1) ]
					//        "i" = 0 to bp.getBufferPoolSize()
					// so
					// "bpOffset" = [ (-1) * (bpBlockIndex+1) ]
					
					int bpOffset = ( (-1) * (bpBlockIndex+1) );
					
					rowAvailable = getBufferManager()->getBufferInfoTable()->getRowAvailable()->at( bpBlockIndex );
					
					bpDataPage = bpDataPages->find( bpOffset )->second;
					
					// DEBUGGING ..
					//cout << "UserInterface:: bpBlockIndex (I): " << bpBlockIndex << endl << endl;
					//cout << "UserInterface:: bpOffset (I): " << bpOffset << endl << endl;
					
				} else {
					
					// BP pages were captured in "bpDataPages" and file pages were captured in "fileDataPages",
					// The operation of capturing "fileDataPages" changes values in the "bit",
					// so in order to keep "bit" values that match the pages in "bpDataPages",
					// a "snap shot" of the "bit" that properly matches "bpDataPages" is stored in "qm->bm->bitSnapShot" ..
					
					int bpIndex = getBufferManager()->getBitSnapShot()->getBufferInfoTableIndex( tableName, minPageID );
					
					rowAvailable = getBufferManager()->getBitSnapShot()->getRowAvailable()->at( bpIndex );
					
					bpDataPage = bpDataPages->find( minPageID )->second;
					
					// DEBUGGING ..
					//cout << "UserInterface:: bpIndex (II): " << bpIndex << endl << endl;
					//cout << "UserInterface:: minPageID (II): " << minPageID << endl << endl;
					
				}
				
				
				//rowCount = bpDataPage->getRecords()->size();
				rowCount = bpDataPage->getNumberOfRows();
				
				// DEBUGGING ..
				//cout << "UserInterface:: rowCount: " << rowCount << endl << endl;
				
				int columnCount = 0;
				
				for(int i=0; i<rowCount; i++){
					
					// DEBUGGING ..
					//cout << "UserInterface:: rowAvailable["<<i<<"] == false: " << ( rowAvailable[i] == false ) << endl << endl;
			
					if ( rowAvailable->at(i) == false ) {
				
						for(int j=0; j<fieldCount; j++){
					
							// buffer whitespace ..
							cout << " ";
					
							//dataValue = bpDataPage->getRecords()->at( i ).getFields()->at( j ).getValue();
							dataValue = bpDataPage->getEntries( i )->at( j );
					
							dataString = intToStr( dataValue );
					
							backFill = ( fieldWidths[j] - dataString.size() );
					
							// DEBUGGING ..
							//cout << "backFill: " << backFill << endl << endl;
					
							// print the backfill ..
							for(int k=0; k<backFill; k++){
					
								cout << " ";
				
							}
		
							// print the "DATA" fields ..
		
							cout << dataString;
		
							// buffer whitespace ..
							cout << " ";

				
						} // end "for()" ..
				
						// carriage return to next line ..
						cout << endl;
				
					} // end "if()" .. 
			
				} // end "for()" ..
				
			} // end "if( isMinFilePageID )" ..
			
			
		} // end "for(int i=0; i<numberOfDataPagesToPrint; i++)" ..
		
		// DEBUGGING ..
		//cout << "UserInterface:: printResults() Point O .." << endl << endl;

		
	} else {
		
		//cout << "NO RESULTS" << endl << endl;
		string resultString = "NO RESULTS ...\n";
		getMessageManager()->setResultMessage( resultString );
		
	}
	
	
} // end "oldPrintResults()" ..
