#ifndef TABLEFORMATTER_H_
#define TABLEFORMATTER_H_

#pragma once

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include "String_Utilities.h"

using namespace std;

class TableFormatter {
	
public:

	TableFormatter();
	
	void resetValues();
	
	void resetHeaders();
	
	void resetData();
	
	void resetHeaderSpacers();
	
	void resetDividerSpacers();
	
	void resetDataSpacers();
	
	bool addColumnHeader( string newHeader );
	
	bool addColumnData( int columnNumber, string newColumnData );
	
	bool addDefaultValue( int columnNumber, string newDefaultValue );
	
	bool endDataRow();
	
	void formatTable();
	
	void printTable();
	
	void printErrorString();
	
	void test( TableFormatter tableFormatter );

	
	/*
	 * SETTERS ..
	 */
	
	// header "spacer" setters ..
	void setHeaderColumnSpacer( string newHeaderColumnSpacer );
	
	void setHeaderColumnSpacerLeft( string newHeaderColumnSpacerLeft );
	
	void setHeaderColumnSpacerRight( string newHeaderColumnSpacerRight );
	
	
	// divider "spacer" setters ..
	void setDividerColumnSpacer( string newDividerColumnSpacer );
	
	void setDividerColumnSpacerLeft( string newDividerColumnSpacerLeft );
	
	void setDividerColumnSpacerRight( string newDividerColumnSpacerRight );

	
	// data "spacer" setters ..
	void setDataColumnSpacer( string newDataColumnSpacer );
	
	void setDataColumnSpacerLeft( string newDataColumnSpacerLeft );
	
	void setDataColumnSpacerRight( string newDataColumnSpacerRight );
	

	
	void setColumnHeaders( vector<string> columnHeaders );
	
	void setDivider( char newDivider );
	
	void setColumnData( vector< map<int, string> > columnData );
	
	void setDefaultFieldValues( map<int, string> defaultFieldValues );
	
	void setDefaultFieldValue( string defaultFieldValue );
	
	/*
	 * GETTERS ..
	 */
	
	
private:

	bool assertionsOn;

	// output to the application when the size of either "columnHeaders" 
	// or "columnData" is zero ..
	string errorString;
	
	// parameter that indicates that an error of some kind has occurred ..
	bool errorFlag;
	
	
	// character or string representing the space between column titles ..
	string headerColumnSpacer;
	
	// character or string representing the left external space prior to 
	// the left-most column title ..
	string headerColumnSpacerLeft;
	
	// tracks whether or not to put header "spacer" on left outside edge of table ..
	bool hasExternalHeaderSpacerLeft;
	
	// character or string representing the right external space after the 
	// the right-most column title ..
	string headerColumnSpacerRight;
	
	// tracks whether or not to put header "spacer" on right outside edge of table ..
	bool hasExternalHeaderSpacerRight;
	
	
	// character or string representing the space between columns divders..
	string dividerColumnSpacer;
	
	// character or string representing the left external space prior to
	// the left-most column divider ..
	string dividerColumnSpacerLeft;
	
	// tracks whether or not to put divider "spacer" on left  outside edge of table ..
	bool hasExternalDividerSpacerLeft;
	
	// character or string representing the left external space prior to
	// the left-most column divider ..
	string dividerColumnSpacerRight;
	
	// tracks whether or not to put divider "spacer" on right outside edge of table ..
	bool hasExternalDividerSpacerRight;
	
	
	// character or string representing the space between data columns ..
	string dataColumnSpacer;
	
	// character or string representing the left external space prior to
	// the left-most data column ..
	string dataColumnSpacerLeft;
	
	// tracks whether or not to but data "spacer" on left outside edge of table ..
	bool hasExternalDataSpacerLeft;
	
	// character or string representing the left external space prior to
	// the left-most data column ..
	string dataColumnSpacerRight;
	
	// tracks whether or not to but data "spacer" on right outside edge of table ..
	bool hasExternalDataSpacerRight;
	
	
	// default string char(s) used in place of missing data values ..
	string defaultFieldValue;

	// max number of "header" columns for this table 
	// (data could be a jagged array) ..
	int maxNumberOfHeaderColumns;
	
	// max number of "data" columns for this table
	// (header fields could be more or less than number of data columns) ..
	int maxNumberOfDataColumns;
	
	// whenever the method "formatTable" is called, this should be "true";
	// this is set to "false" via "resetValues" method ..
	bool isTableFormatted;
	
	// this parameter tracks whether "headerSizes" is already populated ..
	bool isHeaderPopulated;

	// these are the max lengths for each "header" fields ..
	// "m" columns < > ..
	vector<int> headerSizes;
	
	// this parameter tracks whether "dataSizes" is already populated ..
	bool isDataPopulated;
	
	// these are the lengths of each row fields in "header" + "data" columns ..
	// "m" columns < "n" rows > ..
	vector< vector<int> > dataSizes;
	
	// these are the max lengths out of each "header" + "data" column ..
	// columns < > ..
	vector<int> maxColumnSizes;
	
	// these are the "headers" ..
	vector<string> columnHeaders;
	
	// this is the character that represents the "divider" that separates
	// the "header" from the "data" in the table ..
	char dividerChar;
	
	// these are the rows of "data", of the table to print ..
	// "n" rows ..
	map<int, string> columnDataRow;
	
	// these are the rows of columns of "data", of the table to print ..
	// "n" rows < column number, column attribute > ..
	vector< map<int, string> > columnData;
	
	// these are the default values to print when the fields is empty ..
	// 1 row < column number, column attribute > ..
	map<int, string> defaultFieldValues;
	
	
	void findMaxNumberOfColumns();
	
	bool isHeaderEmpty();
	
	bool isDataEmpty();
	
	void populateHeaderSizes();
	
	void populateDataSizes();
	
	void populateMaxColumnSizes();
	
	void populateSizes();
	
	void printHeaders();
	
	void printDivider();
	
	void printData();
	
	
	/*
	 * SETTERS ..
	 */
	
	
	/*
	 * GETTERS ..
	 */
	
	
	
};

#endif /*TABLEFORMATTER_H_*/
