#include "IndexManager.h"

IndexManager:: IndexManager() {

	//cout << "StorageManager:: StorageManager() start .." << endl << endl;

	isIndexDebug = false;

	//cout << "StorageManager:: StorageManager() end .." << endl << endl;

}


IndexManager:: IndexManager( StorageManager* sm_ ){

	//cout << "StorageManager:: StorageManager() start .." << endl << endl;

	sm = sm_;

	isIndexDebug = false;

	//cout << "StorageManager:: StorageManager() end .." << endl << endl;

}

IndexManager::~IndexManager(){
}

StorageManager* IndexManager:: getStorageManager(){

	return sm;

}

MessageManager* IndexManager:: getMessageManager(){

	return sm->getCatalogManager()->getMessageManager();

}

CatalogManager* IndexManager:: getCatalogManager(){

	return sm->getCatalogManager();

}

void IndexManager:: setIsIndexDebug( bool isIndexDebug_ ){

	isIndexDebug = isIndexDebug_;

}

bool IndexManager:: getIsIndexDebug(){

	return isIndexDebug;

}

int IndexManager:: calculateIndexSize( string tableName_, string indexName_ ){

	//cout << "IndexManager:: calculateIndexSize() start.. " << endl << endl;

	// "Index Header" file size ..
	//string directoryPlusFileName = databaseFile->getFileDrive();
	DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();
	
	string directoryPlusFileName = databaseFile->getFileDrive();
	directoryPlusFileName += databaseFile->getFileDirectory();

	directoryPlusFileName += "IndexHeader_";
	directoryPlusFileName += tableName_;
	directoryPlusFileName += "_";
	directoryPlusFileName += indexName_;
	char* fileName = (char*)directoryPlusFileName.c_str();
	
	//cout << "IndexManager:: calculateIndexSize() Point A.. " << endl << endl;

	fstream dataTable1( fileName, ios::in |ios::binary );

	int indexSize = 0;

	if ( !dataTable1.fail() ) {

		dataTable1.seekg( 0, ios::end );
		indexSize += dataTable1.tellg();

	}

	dataTable1.clear();
	dataTable1.close();
	
	//cout << "IndexManager:: calculateIndexSize() Point B.. " << endl << endl;


	// "Index Bucket Header" file size ..
	directoryPlusFileName = databaseFile->getFileDrive();
	directoryPlusFileName += databaseFile->getFileDirectory();

	directoryPlusFileName += "IndexBucketHeader_";
	directoryPlusFileName += tableName_;
	directoryPlusFileName += "_";
	directoryPlusFileName += indexName_;
	fileName = (char*)directoryPlusFileName.c_str();
	
	//cout << "IndexManager:: calculateIndexSize() Point C.. " << endl << endl;


	fstream dataTable2( fileName, ios::in | ios::binary );

	if ( !dataTable2.fail() ) {

		dataTable2.seekg( 0, ios::end );
		indexSize += dataTable2.tellg();

	}

	dataTable2.clear();
	dataTable2.close();
	
	//cout << "IndexManager:: calculateIndexSize() Point D.. " << endl << endl;


	// "Index Buckets" file size ..
	directoryPlusFileName = databaseFile->getFileDrive();
	directoryPlusFileName += databaseFile->getFileDirectory();

	directoryPlusFileName += "IndexBuckets_";
	directoryPlusFileName += tableName_;
	fileName = (char*)directoryPlusFileName.c_str();

	//fstream dataTable3( fileName, ios::in | ios::out |ios::binary );
	fstream dataTable3( fileName, ios::in |ios::binary );

	if ( !dataTable3.fail() ) {

		dataTable3.seekg( 0, ios::end );
		indexSize += dataTable3.tellg();

	}

	dataTable3.clear();
	dataTable3.close();
	
	//cout << "IndexManager:: calculateIndexSize() Point E.. " << endl << endl;


	// "Index Pages" file size ..
	directoryPlusFileName = databaseFile->getFileDrive();
	directoryPlusFileName += databaseFile->getFileDirectory();
	
	directoryPlusFileName += "Index_";
	directoryPlusFileName += tableName_;
	fileName = (char*)directoryPlusFileName.c_str();

	//fstream dataTable4( fileName, ios::in | ios::out | ios::binary );
	fstream dataTable4( fileName, ios::in | ios::binary );

	if ( !dataTable4.fail() ) {

		dataTable4.seekg( 0, ios::end );
		indexSize += dataTable4.tellg();

	}

	dataTable4.clear();
	dataTable4.close();
	
	//cout << "IndexManager:: calculateIndexSize() Point F.. " << endl << endl;


	return indexSize;


} // end "calculateIndexSize()" ..


/*
 * INDEX METHODS
 */


 bool IndexManager:: createIndex( string indexName_, string indexTableName_, bool isNoDuplicates_, string attributeName_ ){

	// creates a B+ Tree (unclustered) index with "pointer buckets" ..


	// DEBUGGING ..
	//cout << "IndexManager:: createIndex() start .." << endl << endl;
	//cout << "IndexManager:: createIndex() indexTableName: " << indexTableName_ << endl << endl;
	//cout << "IndexManager:: createIndex() indexName: " << indexName_ << endl << endl;
	//cout << "IndexManager:: createIndex() attributeName: " << attributeName_ << endl << endl;
	//cout << "IndexManager:: createIndex() isNoDuplicates: " << isNoDuplicates_ << endl << endl;

	// creates the "meta-data" that tracks the index (size, # columns, etc)
	// in the "IndexHeader" file (a file separate from the data file) ..

	// creating the index involves creating the IndexHeader file,
	// but the data file may or may not exist yet ..

	// I If the table exists, but it has no data yet ("DataHeader_xyz" file, no "xyz" file) ..
	//    1) create the Index without data ..
	//			a) store Index metadata in the Catalog ..
	//          b) NOTE: because there isn't any data yet, the "Index_xyz" file isn't created yet,
	//					 so technically only the Catalog metadata is made in this step ..

	// II If the table exists, and it has data ("DataHeader_xyz" file, "xyz" file) ..
	//    1) create the Index ..
	//    2) "bulk load" Index with the search keys of Index's "attributeName" ..
	//			a) since there is data to index, "Index_xyz" file is created with "pointer buckets" ..
	//			b) "Pointer Buckets" created inside of "IndexBucket_xyz" file ..
	//          c) NOTE: future insertions / deletions will update these files ..
	//    3) store metadata ..
	//			a) store Index metadata in the Catalog ..
	//			b) store Index metadata in the IndexHeader ..


	// III if the table doesn't exist ..
	//    1) don't create an index; the table must exist first before an index can be created for it ..

	// this method writes 1 4-byte integers to the "header" file
	// using the "writeHeader()" method ..

	// header.totalNumberOfDataPages: 4 bytes ..

	bool createIndexPerformed = false;

	// check if the table exists in the catalog ..
	bool doesCatalogTableExist = getCatalogManager()->doesCatalogTableNameExist( indexTableName_ );

	// DEBUGGING ..
	//cout << "IndexManager:: createIndex() doesCatalogTableExist: " << doesCatalogTableExist << endl << endl;

	if ( doesCatalogTableExist == true ) {

		// check if the attribute already has an index on it ..
		bool doesCatalogTableAttributeNameHaveIndex = getCatalogManager()->doesCatalogTableAttributeNameHaveIndex( indexTableName_, attributeName_ );

		// DEBUGGING ..
		//cout << "IndexManager:: createIndex() doesCatalogTableAttributeNameHaveIndex: " << doesCatalogTableAttributeNameHaveIndex << endl << endl;

		if ( doesCatalogTableAttributeNameHaveIndex == false ) {

			int tableSize = getCatalogManager()->getTableSize( indexTableName_ );

			// DEBUGGING ..
			//cout << "IndexManager:: createIndex() tableSize: " << tableSize << endl << endl;

			// I table exists, but it may or may not have data yet ..

			if( tableSize > 0 ) {
				
				//cout << "IndexManager:: createIndex() Point A .." << endl << endl;

				createIndexPerformed = createIndexWithData( indexName_, indexTableName_, isNoDuplicates_, attributeName_ );

				//cout << "IndexManager:: createIndex() Point A1 .." << endl << endl;

				//if ( createIndexPerformed == false )
					//return createIndexPerformed;

			} // end "if( tableSize > 0 )" ..

			//cout << "IndexManager:: createIndex() Point B .." << endl << endl;

			// II create the index ( without data ) ..

			//    store Index metadata in the Catalog ..


			// a) store Index metadata in the Catalog ..
			CatalogIndex* catalogIndex = new CatalogIndex();

			catalogIndex->setIndexTableName( indexTableName_ );
			catalogIndex->setIndexName( indexName_ );
			
			//cout << "IndexManager:: createIndex() Point C .." << endl << endl;

			if( tableSize > 0 ) {

				// calculate the index size ..
				int indexSize = calculateIndexSize( indexTableName_, indexName_ );
				//int indexSize = getIndexSize( indexTableName_, indexName_ );

				// add the new "CatalogIndex" to the "Catalog" in the "CatalogManager" ..
				//getCatalogManager()->getCatalog()->getCatalogIndex_indexName( indexName_ )->setIndexSize( indexSize );
				catalogIndex->setIndexSize( indexSize );

			} else {

				catalogIndex->setIndexSize( 0 );							// index starts with zero size ..

			}
			
			//cout << "IndexManager:: createIndex() Point D .." << endl << endl;

			catalogIndex->setIsNoDuplicates( isNoDuplicates_ );
			catalogIndex->setAttribute( attributeName_ );

			// add the new "CatalogIndex" to the "Catalog" in the "CatalogManager" ..
			getCatalogManager()->getCatalog()->addCatalogIndex( catalogIndex );

			// update the "Catalog.bin" file with the new "CatalogIndex" ..
			getCatalogManager()->writeMetaData();

			createIndexPerformed = true;
			
			//cout << "IndexManager:: createIndex() Point E .." << endl << endl;

			//cout << "IndexManager:: createIndex() indexName: " << getCatalogManager()->getCatalogIndex_indexName( indexName_ )->getIndexName() << endl << endl;
			//cout << "IndexManager:: createIndex() indexTableName: " << getCatalogManager()->getCatalogIndex_indexName( indexName_ )->getIndexTableName() << endl << endl;
			//cout << "IndexManager:: createIndex() isNoDuplicates: " << getCatalogManager()->getCatalogIndex_indexName( indexName_ )->getIsNoDuplicates() << endl << endl;
			//cout << "IndexManager:: createIndex() indexSize: " << getCatalogManager()->getCatalogIndex_indexName( indexName_ )->getIndexSize() << endl << endl;


		} else {

			// the attribute already has an index on it, so don't create this new index ..
	 		// notify the user that the attribute already has an index on it ..

	 		string errorString;

	 		errorString += "\nThe table attribute already as an index [ SM::createIndex ].";

	 		// update the "MessageManager" to track this error and output to user ..
	 		getMessageManager()->setErrorStringSM( errorString );


		} // end "if ( doesCatalogTableAttributeNameHaveIndex == false )" ..


	} else {

		// III table doesn't exist, so don't create the index ..
	 	// notify the user that the table does not exist ..
	 	// (user is notified that the index cannot be created in QueryManager) ..

	 	string errorString;

	 	errorString += "\nThe table does not exist [ SM::createIndex ].";

	 	// update the "MessageManager" to track this error and output to user ..
	 	getMessageManager()->setErrorStringSM( errorString );


	} // end "if ( doesCatalogTableExist == true )" ..



	/*
	bool writeHeaderPerformed = writeIndexHeader( char* tableName, char* indexName, char* attributeName, bool isDuplicates, header );

	if ( writeHeaderPerformed == false ) {

	 	string errorString = "\nIndex Header file was not written to [ SM::createIndex() ], table may not have been created properly.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );

	}
	*/

	//cout << "IndexManager:: createIndex() Point F .." << endl << endl;

	// return whether or not "createIndexPerformed()" worked ..
	return createIndexPerformed;


	//cout << "IndexManager:: createIndex() end .." << endl << endl;


} // end "createIndex( char* tableName, char* indexName, char* attributeName, bool isDuplicates )" ..


bool IndexManager:: createIndexWithData( string indexName_, string indexTableName_, bool isNoDuplicates_, string attributeName_ ){

	//cout << "IndexManager:: createIndexWithData() start .." << endl << endl;

	// II table exists, and it has data ..

	//	1) create the Index (done above)..

	//	2) "bulk load" Index with the search keys of Index attribute ..
	//		a) create "Index_xyz" file using existing data ..

	bool createIndexPerformed = false;

	DataHeader* header = new DataHeader();
	bool isHeaderRead = getStorageManager()->readHeader( indexTableName_, header );

	// DEBUGGING ..
	//cout << "IndexManager:: createIndexWithData() isheaderRead: " << isHeaderRead << endl << endl;

	if ( isHeaderRead ){

		vector<DataHeaderEntry*>* headerEntries = header->getHeaderEntries();

		// get all the "pageIDs" from this "indexTableName",
		// for this index ..
		vector<int>* pageIDs = new vector<int>();
		int headerEntriesSize = headerEntries->size();
		for(int i=0; i<headerEntriesSize; i++){

			pageIDs->push_back( headerEntries->at(i)->getPageID() );

			// DEBUGGING ..
			//cout << "IndexManager:: createIndexWithData() headerEntries->at(i("<<i<<")).getPageID(): " << headerEntries->at(i).getPageID() << endl << endl;

		}

		// get all the "DataPages" for this "indexTableName",
		// for this index ..
		DataPage* dataPage = new DataPage();
		vector<DataPage*>* dataPages = new vector<DataPage*>();
		for(int i=0; i<pageIDs->size(); i++){

			dataPage = new DataPage();
			bool isPageRead = getStorageManager()->readDataPage( indexTableName_, dataPage, pageIDs->at(i) );

			// DEBUGGING ..
			//cout << "IndexManager:: createIndexWithData() isPageRead(i("<<i<<"), pageIDs[i]("<<pageIDs[i]<<")): " << isPageRead << endl << endl;

			if( isPageRead ){

				dataPages->push_back( dataPage );

			}

		}

		// A series of method-calls follows, filling data structures with values ..

		// 					"createIndexLeafNodes()": "pageIDs", "allSearchKeys", "searchKeyPageIDs"

		// "createIndexBucketsAndUniqueSearchKeys()": "indexBuckets", "uniqueSearchKeys"

		// 					   "createParentNodes()": "parentNodes", "extraParentNodeGeneratedPerRow"

		// 						"createIndexPages()": "indexPages"

		// DEBUGGING ..
		//cout << "IndexManager:: createIndexWithData() pageIDs.size(): " << pageIDs.size() << endl << endl;
		//cout << "IndexManager:: createIndexWithData() dataPages.size(): " << dataPages.size() << endl << endl;

		// get all the "allSearchKeys" and "searchKeyPageIDs" for this "indexTableName", for this index ..
		// "pageIDs", "allSearchKeys", "searchKeyPageIDs" filled with values ..
		vector<int>* allSearchKeys = new vector<int>();
		vector<int>* searchKeyPageIDs = new vector<int>();

		createIndexLeafNodes( indexTableName_, attributeName_, dataPages, pageIDs, allSearchKeys, searchKeyPageIDs );

		// check to see if existing data has "duplicate" searchKey values ..
		bool hasDuplicateSearchKeys = checkForDuplicates( allSearchKeys );


		if ( hasDuplicateSearchKeys && isNoDuplicates_ ) {

			// data has duplicates, but user specified no duplicates ..

			// set the error message ..
	 		string errorString;

	 		errorString += "\nThe table has duplicate data, but the user specified no duplicates [ SM::createIndex ].  Therefore, the index was not created.";

	 		// update the "MessageManager" to track this error and output to user ..
	 		getMessageManager()->setErrorStringSM( errorString );

			createIndexPerformed = false;

			return createIndexPerformed;

		}

		// used in "createParentNodes" ..
		// if buckets are used (isNoDuplicates_ == false), then these are paired with each bucket
		// else, these are simply the unique searchKeys used in the method above ..
		vector<int>* uniqueSearchKeys = new vector<int>();

		// if buckets are used (isNoDuplicates_ == false), then need to declare these here ..
		IndexBucketHeader* indexBucketHeader = new IndexBucketHeader();

		vector<IndexBucket*>* indexBuckets = new vector<IndexBucket*>();

		if ( isNoDuplicates_ == false ) {

			// NOTE: if "isNoDuplicates_ == true" this variable is not used ..


			//		b) create "IndexBuckets" file which holds the bucket pages (for each search key in the "Index_xyz" file) ..

			// each bucket has "unique pageIDs" for a specific "searchKey" ..
			// (no need to repeat pageIDs, since algorithm finds all duplicate searchKeys on a page) ..
			// "indexBuckets", "uniqueSearchKeys" filled with values ..
			// NOTE: buckets composed above in step "a)" ..


			createIndexBucketsAndUniqueSearchKeys( indexBuckets, uniqueSearchKeys, indexBucketHeader, allSearchKeys, searchKeyPageIDs );

		} else {

			createUniqueSearchKeys( uniqueSearchKeys, allSearchKeys );

		}


		// d) create "parent-node searchKeys" ..
		// "parentNodes", "extraParentNodeGeneratedPerRow" filled with values ..

		// contains all the rows of searchKeys ..
		// first row are the leaf searchKeys ..
		// remaining rows (if any) are parent searchKeys ..
		// root node is the last row ..
		// "parent< row < key > >"
		//vector< vector< vector<int>* >* >* parentNodes = NULL;
		vector< vector< vector<int>* >* >* parentNodes = new vector< vector< vector<int>* >* >();


		//cout << "IndexPages:: createIndexWithData() Point A .." << endl << endl;

		// For each row of nodes, in last set of nodes
		// if number of nodes = 1, then
		// its parent is generated as an exceptional case, so
		// indicate that the parent should have only 1 child pageID
		// instead of the usual 2 or more ..
		//vector< bool >* extraParentNodeGeneratedPerRow = NULL;
		vector< bool >* extraParentNodeGeneratedPerRow = new vector<bool>();


		createParentNodes( uniqueSearchKeys, parentNodes, extraParentNodeGeneratedPerRow );
		//cout << "                   (parentNodes == NULL)?: " << (parentNodes == NULL) << endl << endl;
		//cout << "(extraParentNodeGeneratedPerRow == NULL)?: " << (extraParentNodeGeneratedPerRow == NULL) << endl << endl;

		// "createIndexPages()" pre-conditions:

		//            index nodes: "parentNodes" -> "parentSearchKeysRow" -> "parentSearchKeys"

		//             leaf nodes: "uniqueSearchKeys" ..

		// buckets for leaf nodes: "indexBuckets" ..

		// need to make "IndexPages" for each node of the index ..
		// "indexPages" filled with values ..
		vector<IndexPage*>* indexPages = new vector<IndexPage*>();

		IndexHeader* indexHeader = new IndexHeader();
		//IndexHeader indexHeader;

		createIndexPages( allSearchKeys, searchKeyPageIDs, indexPages, indexHeader, parentNodes, extraParentNodeGeneratedPerRow, isNoDuplicates_ );

		// DEBBUGGING ..
		//cout << "IndexManager:: createIndexWithData() indexHeader->getTotalNumberOfIndexPages(): " << indexHeader->getTotalNumberOfIndexPages() << endl << endl;

		//for(int k=0; k<indexHeader->getHeaderEntries()->size(); k++){

		//	cout << "IndexManager:: createIndexWithData() indexHeader->getHeaderEntries()->at(k)->getPageID(): " << indexHeader->getHeaderEntries()->at(k)->getPageID() << endl << endl;

		//	cout << "IndexManager:: createIndexWithData() indexHeader->getHeaderEntries()->at(k)->getParentPageID(): " << indexHeader->getHeaderEntries()->at(k)->getParentPageID() << endl << endl;

		//}

		// now that index pages have been created,
		// update the header pages that track them ..
		bool isWriteIndexHeader = writeIndexHeader( indexTableName_, indexName_, indexHeader );

		//cout << "IndexManager:: createIndexWithData() I: indexPages == NULL: " << (indexPages == NULL) << endl << endl;

		// "writeIndexBuckets()" pre-conditions:

		//			  index nodes: indexPages[i]->getSearchKeys() ..
		//			  index nodes: indexPages[i]->getChildPageIDs() ..

		//             leaf nodes: "uniqueSearchKeys" ..

		// buckets for leaf nodes: "indexBuckets" ( vector<int>* pageIDs = indexBuckets[i]->getPageIDs() ) ..

		//  3) store metadata ..

		// if "buckets" were created above for the duplicate pageIDs per search key,
		// then all the "buckets" are written to file, here ..
		if ( isNoDuplicates_ == false ){

			//		a) store Index metadata in the IndexBucket ..
			bool isWriteIndexBuckets = writeIndexBuckets( indexTableName_, indexName_, indexBuckets );
			//map<int, vector<int>* >* bucketPageIDsPerSearchKey = writeIndexBuckets( &indexBuckets, indexTableName_, indexName_ );

			//cout << "IndexManager:: createIndexWithData() Point A .." << endl << endl;
			//indexBucketHeader.setBucketIDsForEachSearchKey( bucketPageIDsPerSearchKey );

			//cout << "IndexManager:: createIndexWithData() Point B .." << endl << endl;

			indexBucketHeader->setTotalNumberBuckets( indexBucketHeader->getHeaderEntries()->size() );

			// update the "IndexBucketHeader" file ..
			bool isWriteBucketHeader = writeIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );

			// make sure that each "indexPage" childPageID points to a correct bucketID ..
			// (since duplicates are allowed, childPageID points not to a DataPage, but rather to an IndexBucket) ..
			vector<IndexBucketHeaderEntry*>* headerEntries = indexBucketHeader->getHeaderEntries();

			int numberOfBuckets = indexBucketHeader->getTotalNumberBuckets();
			int numberOfIndexPages = indexPages->size();

			int bucketSearchKey = -1;
			int bucketID = -1;

			int numberOfSearchKeys = -1;

			//cout << "IndexManager:: createIndexWithData() Point B1 .." << endl << endl;

			// loop through each bucket ..
			for(int i=0; i<numberOfBuckets; i++){

				//cout << "IndexManager:: createIndexWithData() Point B2 .." << endl << endl;

				bucketSearchKey = headerEntries->at(i)->getSearchKey();

				// check each bucket against all of the indexPages ..
				for(int j=0; j<numberOfIndexPages; j++){

					//cout << "IndexManager:: createIndexWithData() Point B3 .." << endl << endl;

					// ignore non-leaf nodes ..
					if( indexPages->at(j)->getIsLeafNode() == true ){

						//cout << "IndexManager:: createIndexWithData() Point B4 .." << endl << endl;

						numberOfSearchKeys = indexPages->at(j)->getSearchKeys()->size();

						// locate the leaf node corresponding to this "bucketSearchKey" ..
						for(int k=0; k<numberOfSearchKeys; k++){

							//cout << "IndexManager:: createIndexWithData() Point B5 .." << endl << endl;

							//cout << "IndexManager:: createIndexWithData() ( bucketSearchKey("<<bucketSearchKey<<") == indexPages->at(j("<<j<<"))->getSearchKeys()->at(k("<<k<<"))("<<indexPages->at(j)->getSearchKeys()->at(k)<<") ): " << ( bucketSearchKey == indexPages->at(j)->getSearchKeys()->at(k) ) << endl << endl;

							if( bucketSearchKey == indexPages->at(j)->getSearchKeys()->at(k) ){

								// add the "bucketID" into the "childPageID" value
								// of the indexPage ..
								// (for the leaf-node, kth "searchKey" corresponds to the kth "childPageID") ..

								bucketID = headerEntries->at(i)->getBucketID();

								indexPages->at(j)->getChildPageIDs()->push_back( bucketID );

								//cout << "IndexManager:: createIndexWithData() Point B6 .." << endl << endl;

								//cout << "IndexManager:: createIndexWithData() bucketID: " << bucketID << endl << endl;

							}

						} // end "for()" ..

					}

				} // end "for()" ..

			} // end "for()" ..

		} // end "if()" ..



		//cout << "IndexManager:: createIndexWithData() Point C .." << endl << endl;

		//cout << "IndexManager:: createIndexWithData() II: indexPages == NULL: " << (indexPages == NULL) << endl << endl;

		//cout << "IndexManager:: createIndexWithData() indexPages->size(): " << indexPages->size() << endl << endl;

		// need to figure out if pages are NEW or not ...
		// vector<bool> areNew_ = new vector<bool>();
		
		//cout << "IndexPages:: createIndexWithData() Point B .." << endl << endl;

		//		b) store Index metadata in the IndexPages ..
		vector<bool>* isNew = new vector<bool>();
		for(int i=0; i<indexPages->size(); i++)
			isNew->push_back( false );

		//cout << "IndexManager:: createIndexWithData() Point D .." << endl << endl;

		//bool wereIndexPagesWritten = writeIndexPages( indexTableName_, indexName_, indexPages, isNew );
		bool wereIndexPagesWritten = writeIndexPages( indexTableName_, indexName_, indexPages, isNew, isNoDuplicates_ );
		// isNoDuplicates_


		//cout << "IndexManager:: createIndexWithData() Point E .." << endl << endl;

		//		c) store Index metadata in the IndexHeader ..

		//bool isWriteIndexBucketHeader = writeIndexBucketHeader( indexTableName, indexName, indexBucketHeader );

		//cout << "IndexManager:: createIndexWithData() Point F .." << endl << endl;

		//bool isWriteIndexHeader = writeIndexHeader( indexTableName, indexName, *indexHeader );



	} // end "if ( isheaderRead )" ..
	
	//cout << "IndexManager:: createIndexWithData() Point F .." << endl << endl;

	createIndexPerformed = true;

	return createIndexPerformed;

} // end "createIndexWithData()" ..


void IndexManager:: createIndexLeafNodes( string indexTableName_, string attributeName_, vector<DataPage*>* dataPages_, vector<int>* pageIDs, vector<int>* allSearchKeys_, vector<int>* searchKeyPageIDs_ ){

	// get all the "searchKeys" for this "indexTableName",
	// for this index ..

	//cout << "IndexManager:: createIndexLeafNodes() start .." << endl << endl;

	vector<int> allSearchKeysCopy;
	vector<int> searchKeyPageIDsCopy;

	//vector<DataPageRecord>* records;
	//vector<DataPageField>* fields;
	DataPage* dataPage = NULL;
	int numberOfRowsInPage = -1;
	vector<int>* entries = NULL;

	// also get the "column number" (fieldNumber) of the attribute (zero-based)
	// for use below ..
	string attributeNameString = attributeName_;
	string indexTableNameString = indexTableName_;
	int fieldNumber = getCatalogManager()->getAttributePosition( indexTableNameString, attributeNameString );

	int searchKeyValue = -1;

	for(int i=0; i<dataPages_->size(); i++){

		// get all the records for each DataPage ..
		//records = dataPages_->at(i)->getRecords();
		
		dataPage = dataPages_->at(i);
		
		numberOfRowsInPage = dataPage->getNumberOfRows();

		//for(int j=0; j<records->size(); j++){
		for(int j=0; j<numberOfRowsInPage; j++){

			// get all the fields for each record ..
			//fields = records->at( j ).getFields();
			entries = dataPage->getEntries( j );
			
			// get the "searchKey" value from this Field/Record ..
			//searchKeyValue = fields->at( fieldNumber ).getValue();
			if( fieldNumber < entries->size() )
				searchKeyValue = entries->at( fieldNumber );

			// DEBUGGING ..
			//cout << "IndexManager:: createIndexLeafNodes() allSearchKeys.push_back( searchKeyValue )(i="<<i<<", j="<<j<<"): " << searchKeyValue << endl << endl;

			// add this "searchKey" for this Field/Record to the total ..
			allSearchKeys_->push_back( searchKeyValue );

			// DEBUGGING ..
			//cout << "IndexManager:: createIndexLeafNodes() searchKeyPageIDs.push_back( pageIDs[i] )(i="<<i<<", j="<<j<<"): " << pageIDs->at(i) << endl << endl;

			// each "allSearchKeys" is paired with/linked to its "pageID" ..
			searchKeyPageIDs_->push_back( pageIDs->at( i ) );

		}

	}

	// need to pair searchKeys with their pageIDs,
	// for use in creating the "leaf" nodes of the index ..

	// copy "allSearchKeys" and "searchKeyPageIDs" prior to sorting them ..
	allSearchKeysCopy = (* allSearchKeys_ );
	searchKeyPageIDsCopy = (* searchKeyPageIDs_ );

	// i.e. (unsorted) ..
	// 2 3 5 2 4 3 1 3 6	= searchKeys_copy
	// | | | | | | | | |
	// 7 5 8 2 9 2 1 4 5 	= pageIDs_copy

	// DEBUGGING ..
	//cout << "IndexManager:: createIndexLeafNodes() allSearchKeys(unsorted): ";
	//for(int i=0; i<allSearchKeys_->size(); i++){

	//	cout << allSearchKeys_->at(i) << ", ";

	//}
	//cout << endl << endl;

	// DEBUGGING ..
	//cout << "IndexManager:: createIndexLeafNodes() searchKeyPageIDs(unsorted): ";
	//for(int i=0; i<searchKeyPageIDs_->size(); i++){

	//	cout << searchKeyPageIDs_->at(i) << ", ";

	//}
	//cout << endl << endl;

	// sort all of the "searchKeys" for all the pages ..
	sort( allSearchKeys_->begin(), allSearchKeys_->end() );

	// DEBUGGING ..
	//cout << "IndexManager:: createIndexLeafNodes() allSearchKeys(sorted): ";
	//for(int i=0; i<allSearchKeys_->size(); i++){

	//	cout << allSearchKeys_->at(i) << ", ";

	//}
	//cout << endl << endl;

	// i.e.:
	// 1 [2 2] [3 3 3] 4 5 6	= searchKeys .. but what about "pageIDs" .. see next ..

	// now you have to sort "searchKeyPageIDs" to still be paired with/linked to "allSearchKeys" ..
	// ("searchKeyPageIDs" is still paired with "allSearchKeysCopy") ..
	vector<int>* duplicatePageIDs;
	int currentKey = -1;
	int currentKeyPosition = 0;
	int searchKeyCount = 0;

	// DEBUGGING ..
	//cout << "IndexManager:: createIndexLeafNodes() allSearchKeys_->size(): " << allSearchKeys_->size() << endl << endl;


	// iterate through "allSearchKeys", pairing searchKeys with "pageIDs" ..
	while( currentKeyPosition < allSearchKeys_->size() ) {

		searchKeyCount = 0;
		currentKey = allSearchKeys_->at( currentKeyPosition );
		duplicatePageIDs = new vector<int>();

		// collect all the "pageIDs" for the currentKey ..
		for(int j=0; j<allSearchKeysCopy.size(); j++){

			// DEBUGGING ..
			//cout << "IndexManager:: createIndexLeafNodes() ( allSearchKeysCopy[j]("<<allSearchKeysCopy[j]<<") == currentKey("<<currentKey<<") ): " << ( allSearchKeysCopy[j] == currentKey ) << endl << endl;

			if( allSearchKeysCopy[j] == currentKey ){

				// collect "duplicatePageIDs" ..
				duplicatePageIDs->push_back( searchKeyPageIDsCopy[j] );

				searchKeyCount++;

			} // end "if()" ..

		} // end "for()" ..

		// DEBUGGING ..
		//cout << "IndexManager:: createIndexLeafNodes() searchKeyCount: " << searchKeyCount << endl << endl;

		// sort the duplicate "pageIDs" prior to pairing them with searchKeys ..
		sort( duplicatePageIDs->begin(), duplicatePageIDs->end() );

		// DEBUGGING ..
		//cout << "IndexManager:: createIndexLeafNodes() duplicatePageIDs(sorted): ";
		//for(int i=0; i<duplicatePageIDs->size(); i++){

		//	cout << duplicatePageIDs->at(i) << ", ";

		//}
		//cout << endl << endl;

		// link each duplicate "pageID" with its "searchKey" for the currentKey ..
		// 					"currentKeyPosition" = the position following the count of the set
		//						  				   of sorted search keys from the previous iteration ..
		// 						"searchKeyCount" = the count of search keys from the current iteration ..
		// "currentKeyPosition + searchKeyCount" = the end position for the current iteration ..
		for(int j=currentKeyPosition; j<(currentKeyPosition + searchKeyCount); j++){

			// DEBUGGING ..
			//cout << "searchKeyPageIDs[j]("<<j<<")("<<searchKeyPageIDs_->at(j)<<") = duplicatePageIDs->at( ( j-currentKeyPosition("<<currentKeyPosition<<") )("<<(j-currentKeyPosition)<<") )("<<duplicatePageIDs->at( (j-currentKeyPosition) )<<")" << endl << endl;

			searchKeyPageIDs_->at(j) = duplicatePageIDs->at( (j-currentKeyPosition) );

		}

		// DEBUGGING ..
		//cout << "IndexManager:: createIndexLeafNodes() searchKeyPageIDs(sorting): ";
		//for(int i=0; i<searchKeyPageIDs_->size(); i++){

		//	cout << searchKeyPageIDs_->at(i) << ", ";

		//}
		//cout << endl << endl;

		// translate index to next set of duplicate, sorted searchKeys ..
		// i.e.: 1 -> [2 2] -> [3 3 3] -> etc ..
		currentKeyPosition += searchKeyCount;

		// DEBUGGING ..
		//cout << "IndexManager:: createIndex() currentKeyPosition: " << currentKeyPosition << endl << endl;

	} // end "while()" ..


	// DEBUGGING ..
	//cout << "IndexManager:: createIndexLeafNodes() searchKeyPageIDs(sorted): ";
	//for(int i=0; i<searchKeyPageIDs_->size(); i++){

	//	cout << searchKeyPageIDs_->at(i) << ", ";

	//}
	//cout << endl << endl;

	// DEBUGGING ..
	//cout << "================================================" << endl << endl;

	// "pageIDs" (w/ duplicates) sorted and linked to "allSearchKeys" ..
	// 1 [2 2] [3 3 3] 4 5 6	= searchKeys
	// |  | |   | | |  | | |
	// 1  2 7   2 4 5  9 8 5	= pageIDs	.. sorted within duplicate "searchKeys" ..


} // end "createIndexLeafNodes( vector<DataPage*> dataPages_, vector<int>* pageIDs, vector<int>* allSearchKeys_, vector<int>* searchKeyPageIDs_, char* indexTableName_, char* attributeName_ )" ..


bool IndexManager:: checkForDuplicates( vector<int>* allSearchKeys ){

	//cout << "IndexManager:: checkForDuplicates() start .." << endl << endl;

	bool hasDuplicates = false;

	int allSearchKeysSize = allSearchKeys->size();

	int searchKey = -1;

	// compare each searchKey against all the others,
	// see if there are any duplicate values ..
	for(int i=0; i<allSearchKeysSize; i++){

		searchKey = allSearchKeys->at(i);

		for(int j=0; j<allSearchKeysSize; j++){

			// don't compare the searchKey against itself ..
			if( j == i )
				continue;

			if( searchKey == allSearchKeys->at(j) ) {

				hasDuplicates = true;
				break;

			}

		}

		if( hasDuplicates == true )
			break;

	}

	return hasDuplicates;

} // end "checkForDuplicates()" ..


void IndexManager:: createUniqueSearchKeys( vector<int>* uniqueSearchKeys_, vector<int>* allSearchKeys_ ){

	//cout << "IndexManager:: createUniqueSearchKeys() start .." << endl << endl;

	// this method is used to create "uniqueSearchKeys_"
	// for use with "createIndexPages()" method when
	// "isNoDuplicates_" is set to "true" ..

	int allSearchKeysSize = allSearchKeys_->size();

	int currentUniqueKey = -1;

	for(int i=0; i<allSearchKeys_->size(); i++){

		if( currentUniqueKey < 0  ){

			// assign the 1st value of "allSearchKeys" to a variable ..
			currentUniqueKey = allSearchKeys_->at( i );

		}

		if( ( (i+1) < allSearchKeys_->size() ) &&
			( currentUniqueKey == allSearchKeys_->at( (i+1) ) ) ){

			// DEBUGGING ..
			//cout << "(Point A) skip .. " << endl << endl;

			// found next duplicate value in set of values, so
			// continue until reaching next non-duplicate value ..
			continue;

		} else {

			uniqueSearchKeys_->push_back( currentUniqueKey );

			// reset this parameter so we assigne "currentUniqueKey" above ..
			currentUniqueKey = -1;

		}

	}

} // end "createUniqueSearchKeys()" ..


void IndexManager:: createIndexBucketsAndUniqueSearchKeys( vector<IndexBucket*>* indexBuckets_, vector<int>* uniqueSearchKeys_, IndexBucketHeader* indexBucketHeader, vector<int>* allSearchKeys_, vector<int>* searchKeyPageIDs_ ){

	//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() start .." << endl << endl;

	// each bucket has "unique pageIDs" for a specific "searchKey" ..
	// (no need to repeat pageIDs, since algorithm finds all duplicate searchKeys on a page) ..
	IndexBucket* indexBucket = NULL;

	// "allSearchKeys" and "searchKeyPageIDs" sorted and paired/linked, so
	// use them to create "buckets" ..
	int currentUniqueKey = -1;

	// create the "indexBucketHeader" associated with these "indexBuckets_" ..
	//vector<IndexBucketHeaderEntry>* indexBucketHeaderEntries = new vector<IndexBucketHeaderEntry>();
	IndexBucketHeaderEntry* indexBucketHeaderEntry = NULL;

	int currentBucketID = 0;

	//int blockSize = 48;
	//int intSize = 4;

	int startPosition = 0;
	int endPosition = 0;
	
	int allSearchKeysSize = -1;
	if( allSearchKeys_ != NULL )
		allSearchKeysSize = allSearchKeys_->size();

	for(int i=0; i<allSearchKeysSize; i++){

		// iterate through "allSearchKeys", handle sets of duplicate values ..

		if( currentUniqueKey < 0  ){

			// assign the 1st value of "allSearchKeys" to a variable ..
			currentUniqueKey = allSearchKeys_->at( i );

		}
		
		//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() i: " << i << endl << endl;

		// DEBUGGING ..
		//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() currentUniqueKey: " << currentUniqueKey << endl << endl;
		//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() ( (i+1)("<<(i+1)<<") < allSearchKeys.size()("<<allSearchKeys_->size()<<") ): " << ( (i+1) < allSearchKeys_->size() ) << endl << endl;
		//if( (i+1) < allSearchKeys_->size() ) cout << "IndexManager:: createIndex() createIndexBucketsAndUniqueSearchKeys() ( currentUniqueKey("<<currentUniqueKey<<") == allSearchKeys[ (i+1)("<<(i+1)<<") ]("<<allSearchKeys_->at( (i+1) )<<") ): " << ( currentUniqueKey == allSearchKeys_->at( (i+1) ) ) << endl << endl;

		if( ( (i+1) < allSearchKeys_->size() ) &&
			( currentUniqueKey == allSearchKeys_->at( (i+1) ) ) ){

			// DEBUGGING ..
			//cout << "(Point A) skip .. " << endl << endl;

			// found next duplicate value in set of values, so
			// continue until reaching next non-duplicate value ..
			continue;

		} else {


			if( ( (i+1) < allSearchKeys_->size() ) &&
				   ( currentUniqueKey != allSearchKeys_->at( (i+1) ) ) ) {

				//cout << "(Point B) .. " << endl << endl;

				// found next non-duplicate value in set of values ..
				endPosition = i;

				// get the ptr to the "pageIDs" for this bucket ..
				indexBucket = new IndexBucket();

				// create an "IndexBucket" using the "searchKeyPageIDs" for this "currentUniqueKey" ..
				addPageIDsToIndexBucket( searchKeyPageIDs_, indexBucket, startPosition, endPosition );
				
				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point A .." << endl << endl;

				// track the unique "searchKey" for this bucket ..
				uniqueSearchKeys_->push_back( currentUniqueKey );
				
				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point B .." << endl << endl;

				// add the new "indexBucket" to a collection (1 indexBucket per unique searchKey) ..
				indexBuckets_->push_back( indexBucket );
				
				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point C .." << endl << endl;

				// DEBUGGING ..
				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() uniqueSearchKeys->size(): " << uniqueSearchKeys_->size() << endl << endl;
				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBucket->getPageIDs()->size(): " << indexBucket->getPageIDs()->size() << endl << endl;
				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBuckets->size(): " << indexBuckets_->size() << endl << endl;

				// reset these parameters for the next "allSearchKeys" iteration(s) ..
				//startPosition = (endPosition+1);	// (i+1) ..
				//endPosition = startPosition;		// (i+1) ..
				//currentUniqueKey = -1;				// set above in next iteration ..
				//if ( (i+1) < allSearchKeys.size() )
				//	currentUniqueKey = allSearchKeys[ (i+1) ];


			} else if( (i+1) == allSearchKeys_->size() ) {

				// handle the final value of "allSearchKeys" ..

				//cout << "(Point C) .. " << endl << endl;

				// update "endPosition" to match current position ..
				endPosition = i;

				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() startPosition: " << startPosition << endl << endl;
				//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() endPosition: " << endPosition << endl << endl;

				if( startPosition != endPosition ){

					// final value is a duplicate ..

					// DEBUGGING ..
					//cout << "(Point C1) final value in set, currentUniqueKey(" << currentUniqueKey << ") IS a duplicate .." << endl << endl;

					// found next non-duplicate value in set of values ..
					endPosition = i;

					// get the ptr to the "pageIDs" for this bucket ..
					indexBucket = new IndexBucket();

					// create an "IndexBucket" using the "searchKeyPageIDs" for this "currentUniqueKey" ..
					addPageIDsToIndexBucket( searchKeyPageIDs_, indexBucket, startPosition, endPosition );
					
					//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point D .." << endl << endl;

					// track the unique "searchKey" for this bucket ..
					uniqueSearchKeys_->push_back( currentUniqueKey );
					
					//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point E .." << endl << endl;

					// add the new "indexBucket" to a collection (1 indexBucket per unique searchKey) ..
					indexBuckets_->push_back( indexBucket );
					
					//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point F .." << endl << endl;

					// DEBUGGING ..
					//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() uniqueSearchKeys->size(): " << uniqueSearchKeys_->size() << endl << endl;
					//cout << "IndexManager:: ccreateIndexBucketsAndUniqueSearchKeys() indexBucket->getPageIDs()->size(): " << indexBucket->getPageIDs()->size() << endl << endl;
					//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBuckets->size(): " << indexBuckets_->size() << endl << endl;

					// reset these parameters for the next "allSearchKeys" iteration(s) ..
					//startPosition = (endPosition+1);	// (i+1) ..
					//endPosition = startPosition;		// (i+1) ..
					//currentUniqueKey = -1;				// set above in next iteration ..
					//if ( (i+1) < allSearchKeys.size() )
					//	currentUniqueKey = allSearchKeys[ (i+1) ];

				} else {

					// final value is not a duplicate, so just add it to the collection ..

					// DEBUGGING ..
					//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() (Point C2) final value in set, currentUniqueKey(" << currentUniqueKey << ") is NOT a duplicate .." << endl << endl;

					// get the ptr to the "pageIDs" for this bucket ..
					indexBucket = new IndexBucket();
					indexBucket->getPageIDs()->push_back( searchKeyPageIDs_->at(i) );

					// track the unique "searchKey" for this bucket ..
					uniqueSearchKeys_->push_back( currentUniqueKey );

					// add the new "indexBucket" to a collection (1 indexBucket per unique searchKey) ..
					indexBuckets_->push_back( indexBucket );


				} // end "if()" ..


			} // end "if()" ..
			
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point G .." << endl << endl;
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point D .." << endl << endl;

			// set "IndexBucketHeaderEntry" values for this "IndexBucket" ..
			indexBucketHeaderEntry = new IndexBucketHeaderEntry();
			
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point H .." << endl << endl;
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point E .." << endl << endl;

			// each bucket has a "bucketID", must track it in the header ..
			indexBucketHeaderEntry->setBucketID( currentBucketID++ );
			
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point I .." << endl << endl;
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBucketHeaderEntry->setBucketID(): " << currentBucketID << endl << endl;

			// each bucket has a "currentUniqueKey" ..
			indexBucketHeaderEntry->setSearchKey( currentUniqueKey );
			
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point J .." << endl << endl;

			// reset these parameters for the next "allSearchKeys" iteration(s) ..
			startPosition = (endPosition+1);	// (i+1) ..
			endPosition = startPosition;		// (i+1) ..
			currentUniqueKey = -1;				// set above in next iteration ..

			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBucketHeaderEntry->setSearchKey(): " << currentUniqueKey << endl << endl;

			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point F .." << endl << endl;

			// each bucket has 1-12 "pageIDs", track how many are left to fill ..
			int indexBucketPageIDsSize = indexBucket->getPageIDs()->size();
			
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point K .." << endl << endl;
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point G .." << endl << endl;

			int freePointerCount = ( ( blockSize / intSize ) - indexBucketPageIDsSize );

			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point H .." << endl << endl;

			indexBucketHeaderEntry->setFreePointerCount( freePointerCount );
			
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point L .." << endl << endl;
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point I .." << endl << endl;

			// tracks the bucket slots that are free for pageIDs ..
			//vector<bool> pointersAvailable;
			//vector<int>* pageIDs = indexBucket->getPageIDs();
			//for(int i=0; i<indexBucketPageIDsSize; i++){

			//	if( pageIDs->at(i) > 0 )
			//		pointersAvailable.push_back( false );
			//	else
			//		pointersAvailable.push_back( true );

			//}

			//indexBucketHeaderEntry->setPointersAvailable( &pointersAvailable );

			// add this "entry" to the "indexBucketHeader" ..
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBucketHeader == NULL: " << (indexBucketHeader == NULL) << endl << endl;
			vector<IndexBucketHeaderEntry*>* entries = indexBucketHeader->getHeaderEntries();
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() entries == NULL: " << (entries == NULL) << endl << endl;
			
			indexBucketHeader->getHeaderEntries()->push_back( indexBucketHeaderEntry );
			
			//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point M .." << endl << endl;


		} // end "if( currentUniqueKey == NULL )" ..


	} // end "for()" ..
	
	//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() Point Z .." << endl << endl;
	//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() ME:: indexBuckets->size(): " << indexBuckets_->size() << endl << endl;

	// DEBUGGING ..
	//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() uniqueSearchKeys: ";
	//for(int i=0; i<uniqueSearchKeys_->size(); i++){

	//	cout << uniqueSearchKeys_->at(i) << ", ";

	//}
	//cout << endl << endl;

	// DEBUGGING ..
	//cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBuckets: " << endl;
	//for(int i=0; i<indexBuckets_->size(); i++){

	//	vector<int>* pageIDs = indexBuckets_->at(i)->getPageIDs();

	//	cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() pageIDs->size(): " << pageIDs->size() << endl;

	//	cout << "IndexManager:: createIndexBucketsAndUniqueSearchKeys() indexBucket[i("<<i<<")] pageIDs: ";

	//	for(int k=0; k<pageIDs->size(); k++){

	//		cout << pageIDs->at(k) << ", ";

	//	}

	//	cout << endl;

	//}
	//cout << endl << endl;


} // end "createIndexBucketsAndUniqueSearchKeys( vector<IndexBucket*> indexBuckets_, vector<int>* uniqueSearchKeys, vector<int>* allSearchKeys_, vector<int>* searchKeyPageIDs_ )" ..


void IndexManager:: createParentNodes( vector<int>* uniqueSearchKeys_, vector< vector< vector<int>* >* >* parentNodes_, vector< bool >* extraParentNodeGeneratedPerRow_ ){

	//cout << "IndexManager:: createParentNodes() start .." << endl << endl;

	// DEBUGGING ..
	//cout << "================================================" << endl << endl;

	//					|50 90 ..   |
	//				/	  	  |			\
	// |10 20 30 40|	|50 60 70 80|	|90 100 110 120| ..


	// this changes to become all the keys of the row of parent nodes ..
	// (starts out as all the leaf searchKeys) ..
	vector<int> currentRowAllSearchKeys = (* uniqueSearchKeys_ );


	// "vector< vector< vector<int>* >* >* parentNodes_" ..		// contains all the rows of searchKeys ..
																// "parent< row < key > >"
																// first row are the leaf searchKeys ..
																// remaining rows (if there) are parent searchKeys ..
																// root node is the last row ..
																// (this method has code to initialize this parameter,
																//  but when the method returned, this parameter == NULL, so
																//  initialized this parameter outside the method instead) ..




	//"vector< bool >* extraParentNodeGeneratedPerRow";			// For each row of nodes, in last set of nodes
																// if number of nodes = 1, then
																// its parent is generated as an exceptional case, so
																// indicate that the parent should have only 1 child pageID
																// instead of the usual 2 or more ..
																// (this method has code to initialize this parameter,
																//  but when the method returned, this parameter == NULL, so
																//  initialized this parameter outside the method instead) ..

	// contains all the "searchKeys" for a row ..
	vector< vector<int>* >* parentSearchKeysRow = NULL;

	// contains the "searchKeys" for a single parent node ..
	vector<int>* parentSearchKeys = NULL;

	//NOTE: parentNodes -> parentSearchKeysRow -> parentSearchKeys

	int numberSearchKeysPerNode = 4;	// i.e.: |10 20 30 40|

	int fanout = 5;						// i.e.: |P|10|P|20|P|30|P|40|P|

	int indexKeyOffset = 5;				// every "indexKeyOffset^L" searchKey is a pointer value for a parent node above ..
										// where "L" is the row number (leaf node is row number = 1) ..
										// i.e.: leaf node: every (6^1 = 6th) leaf node search key
										//       next level closer to root node: every (6^2 = 36th) leaf node search key

	bool atLeastOnce = true;

	bool extraParentNodeGenerated = false;	// true if last parent node in row is 6th node,
											// and a super-parent node needs to be generated for it ..
											// parent node was generated for a single child node ..



	if( currentRowAllSearchKeys.size() <= numberSearchKeysPerNode ){

		// handle the case of 4 or less search keys,
		// where the only parent node is a generated root node ..

		// need leaf nodes row ..
		parentSearchKeysRow = new vector< vector<int>* >();

		parentSearchKeysRow->push_back( uniqueSearchKeys_ );

		parentNodes_->push_back( parentSearchKeysRow );

		extraParentNodeGeneratedPerRow_->push_back( false );


		// need root node row ..
		parentSearchKeysRow = new vector< vector<int>* >();

		parentSearchKeys = new vector<int>();

		int uniqueSearchKeysSize = uniqueSearchKeys_->size();

		int maxSearchKey = -1;

		for(int i=0; i<uniqueSearchKeysSize; i++){

			if( uniqueSearchKeys_->at( i ) > maxSearchKey )
				maxSearchKey = uniqueSearchKeys_->at( i );

		}

		// root node should be 1 greater than the max of uniqueSearchKeys_ ..
		maxSearchKey++;

		parentSearchKeys->push_back( maxSearchKey );

		//parentSearchKeys->push_back( -1 );
		//parentSearchKeys->push_back( -1 );
		//parentSearchKeys->push_back( -1 );

		parentSearchKeysRow->push_back( parentSearchKeys );

		parentNodes_->push_back( parentSearchKeysRow );

		// since there is only 1 parent node (the root node),
		// need to indicate that this row needs only 1 childPageID,
		// when passed into "createIndexPages()" ..
		extraParentNodeGeneratedPerRow_->push_back( true );



	} else {

		// handle the case of 5 or more search keys,
		// where parent nodes are created ..

		// handle the leaf nodes ..
		if( parentSearchKeysRow == NULL )
			parentSearchKeysRow = new vector< vector<int>* >();

		if( parentSearchKeys == NULL )
			parentSearchKeys = new vector<int>();

		int nodeCounter = 0;
		for(int i=0; i<currentRowAllSearchKeys.size(); i++){

			// doesn't store on the first searchKey, and
			// only stores each node (4 searchKeys per node) ..
			if( ( i > 0 ) &&
				( ( i % numberSearchKeysPerNode ) == 0 ) ){

				parentSearchKeysRow->push_back( parentSearchKeys );

				parentSearchKeys = new vector<int>();

				nodeCounter = 0;

			}

			parentSearchKeys->push_back( currentRowAllSearchKeys[i] );

			nodeCounter++;

		}

		// if the last node has less then "numberSearchKeysPerNode",
		// still need to store it ..
		if( nodeCounter > 0 )
			parentSearchKeysRow->push_back( parentSearchKeys );

		parentSearchKeys = NULL;

		// add the row to the index (so long as there's a row to add) ..
		if( ( parentNodes_ == NULL ) &&
			( parentSearchKeysRow != NULL ) )
			parentNodes_ = new vector< vector< vector<int>* >* >();

		parentNodes_->push_back( parentSearchKeysRow );

		parentSearchKeysRow = NULL;
		
		//cout << "IndexPages:: createParentNodes() Point A .." << endl << endl;

		if ( extraParentNodeGeneratedPerRow_ == NULL )
			extraParentNodeGeneratedPerRow_ = new vector<bool>();

		extraParentNodeGeneratedPerRow_->push_back( false );


		double rowLevel = 0;

		// you know to stop creating parent node rows when the last row is 4 or less ..
		while( ( atLeastOnce ) ||
		   	   ( currentRowAllSearchKeys.size() > numberSearchKeysPerNode ) ) {	// i.e.: |10 20 30 40|  |50        | ..

			if ( atLeastOnce )
				atLeastOnce = false;

			// DEBUGGING ..
			//cout << "IndexManager:: createParentNodes() rowLevel: " << rowLevel << endl << endl;
			//cout << "IndexManager:: ccreateParentNodes() atLeastOnce: " << atLeastOnce << endl << endl;
			//cout << "IndexManager:: createParentNodes() currentRowAllSearchKeys.size(): " << currentRowAllSearchKeys.size() << endl << endl;

			// initialize "searchKeys" for collection that holds parent nodes for this row ..
			parentSearchKeys = NULL;

			// this should be necessary ..
			if( parentSearchKeys == NULL ){
				parentSearchKeys = new vector<int>();
			}

			// initialize collection that holds this row of "searchKeys" ..
			parentSearchKeysRow = NULL;

			int counter = 0;

			// DEBUGGING ..
			//cout << "IndexManager:: createParentNodes() currentRowAllSearchKeys: ";
			//for(int i=0; i<currentRowAllSearchKeys.size(); i++){

			//	cout << currentRowAllSearchKeys[i] << ", ";

			//}
			//cout << endl << endl;

			int uniqueSearchKeysSize = uniqueSearchKeys_->size();

			for(int i=0; i<uniqueSearchKeysSize; i++){

				// DEBUGGING ..
				//cout << "IndexManager:: createParentNodes() i: " << i << endl << endl;
				//cout << "IndexManager:: createParentNodes() counter: " << counter << endl << endl;
				//cout << "IndexManager:: createParentNodes() uniqueSearchKeys[i("<<i<<")]: " << uniqueSearchKeys_->at(i) << endl << endl;


				// DEBUGGING ..
				//cout << "IndexManager:: createParentNodes() ( ( ( i("<<i<<") % ( pow(5, rowLevel)("<<(int)pow(5, rowLevel)<<") * numberSearchKeysPerNode("<<numberSearchKeysPerNode<<") )("<<( (int)pow(5, rowLevel) * numberSearchKeysPerNode )<<") )("<<( i % ( (int)pow(5, rowLevel) * numberSearchKeysPerNode ) )<<") == 0 )("<<( ( i % ( (int)pow(5, rowLevel) * numberSearchKeysPerNode ) ) == 0 )<<") && ( ( i("<<i<<") % pow(5, rowLevel)("<< (int)pow(5, rowLevel)<<") * fanout("<<fanout<<") )("<<( i % ( (int)pow(5, rowLevel) * fanout ) )<<") != 0 )("<<( ( i % ( (int)pow(5, rowLevel) * fanout ) ) != 0 )<<") && ( i("<<i<<") > 1 )("<<( i > 1 )<<") ): " << ( ( ( i % ( (int)pow(5, rowLevel) * numberSearchKeysPerNode ) ) == 0 ) && ( ( i % ( (int)pow(5, rowLevel) * fanout ) ) != 0 ) && ( i > 1 ) ) << endl << endl;


				// every 4th "searchKey" (after and not including 1st key) is a new parent ..
				// "( i % numberSearchKeysPerNode ) == 0" ..
				// (i.e.: 5th, 9th, 13th, 17th, etc .. ) ..

				// also, skip every 20th key "( i % fanout ) != 0" ..
				// (i.e.: 4th, 8th, 12th, 16th = one parent, 24th, 28th, 32nd, 36th = another parent, etc..)
				//if( ( ( i % numberSearchKeysPerNode ) == 0 ) && ( ( i % fanout ) != 0 ) &&
				//	  ( i > 1 ) ){

				if( ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) == 0 ) && ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) != 0 ) &&
					( i > 1 ) ){

					if( parentSearchKeys == NULL ){
						parentSearchKeys = new vector<int>();

						// DEBUGGING ..
						//cout << "IndexManager:: createParentNodes() I - parentSearchKeys = new vector<int>() .." << endl << endl;

					}

					// DEBUGGING ..
					//cout << "IndexManager:: createParentNodes() I - parentSearchKeys->push_back( uniqueSearchKeys[i] ): " << uniqueSearchKeys_->at(i) << endl << endl;

					parentSearchKeys->push_back( uniqueSearchKeys_->at(i) );
					counter++;

				}


				// DEBUGGING ..
				//cout << "IndexManager:: createParentNodes() ( ( ( counter("<<counter<<") % numberSearchKeysPerNode("<<numberSearchKeysPerNode<<") )("<<( counter % numberSearchKeysPerNode )<<") == 0 )("<<( ( counter % numberSearchKeysPerNode ) == 0 )<<") && ( counter("<<counter<<") > 0 )("<<( counter > 0 )<<") ): " << ( ( ( counter % numberSearchKeysPerNode ) == 0 ) && ( counter > 0 ) ) << endl << endl;

				// every 4th "parent searchKey" finishes a parent node ..
				if( ( ( counter % numberSearchKeysPerNode ) == 0 ) && ( counter > 0 ) ) {

					if( parentSearchKeysRow == NULL ) {
						parentSearchKeysRow = new vector< vector<int>* >();

						// DEBUGGING ..
						//cout << "IndexManager:: createParentNodes() II - parentSearchKeysRow = new vector< vector<int>* >(); .." << endl << endl;

					}

					// DEBUGGING ..
					//cout << "IndexManager:: createParentNodes() II - parentSearchKeysRow->push_back( parentSearchKeys ): ";
					for(int i=0; i<parentSearchKeys->size(); i++){

						cout << parentSearchKeys->at(i) << ", ";

					}
					cout << endl;

					parentSearchKeysRow->push_back( parentSearchKeys );

					// reset these parameters for the next parent-node ..
					counter = 0;
					parentSearchKeys = NULL;

				}


				// DEBUGGING ..
				//cout << "( ( i("<<i<<") % ( (int)pow(indexKeyOffset, rowLevel("<<rowLevel<<"))("<<(int)pow(indexKeyOffset, rowLevel)<<") * fanout("<<fanout<<") ) ) == 0 )("<<( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) == 0 )<<") && ( ( uniqueSearchKeys.size()("<<uniqueSearchKeys.size()<<") - i("<<i<<") )("<<( uniqueSearchKeys.size() - i )<<") <= ( (int)pow(indexKeyOffset, rowLevel("<<rowLevel<<"))("<<(int)pow(indexKeyOffset, rowLevel)<<") * numberSearchKeysPerNode("<<numberSearchKeysPerNode<<") )("<<( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode )<<") )("<<( ( uniqueSearchKeys.size() - i ) <= ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) )<<") && ( i("<<i<<") > 1 )("<<( i > 1 )<<"): " << ( ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) == 0 ) && ( ( uniqueSearchKeys.size() - i ) <= ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) && ( i > 1 ) ) << endl << endl;

				// need to handle following exceptions:
				// 1) 1 leaf node, no parent
				// 		"parentNodes" is size = 1, see much above ..

				// 2) 1 parent for 5 child nodes, + 1 extra node, no parent
				// the last/extra node is a "fanout" node, and
				// you know its the last node because "uniqueSearchKeys" has less than ((5^L)*4) searchKeys left ..
				//if( ( ( i % ( (int)pow(5, rowLevel) * fanout ) ) == 0 ) &&
				//	( ( uniqueSearchKeys.size() - i ) <= ( (int)pow(5, rowLevel) * numberSearchKeysPerNode ) ) &&
				//	( i > 1 ) ){

				// for an index tree w/ 1 level of parents + root level, exception occurs at i == 20 ..
				// for trees w/ more levels, exception occurs at ( i % 20 * (int)pow(5, rowLevel) ) ..
				// (i.e.: rowLevel == 0: 20, 40, 60, 80 / rowLevel == 1: 100, 200, 300 400 / rowLevel == 2: 500, 1000, 1500, 2000) ..
				// (not necessarily intuitive .. i figured it out by hand!) ..
				//int exceptionValueIndex = 20;
				int exceptionValueIndex = ( numberSearchKeysPerNode * fanout );

				if( ( ( i % ( exceptionValueIndex * (int)pow(indexKeyOffset, rowLevel) ) ) == 0 ) &&
					( ( uniqueSearchKeys_->size() - i ) <= ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) &&
					( i > 1 ) ) {

					// parent node was generated for a single child node ..
					if ( extraParentNodeGenerated == false )
						extraParentNodeGenerated = true;


					//cout << "IndexManager:: createParentNodes() III - last/extra node is \"fanout\" node (create parent for it) .." << endl << endl;

					// this should be necessary ..
					if( parentSearchKeys == NULL ){
						parentSearchKeys = new vector<int>();

						//cout << "IndexManager:: createParentNodes() III - parentSearchKeys = new vector<int>() .." << endl << endl;

					}

					// create the parent "searchKey" ..

					// NOTE: this is an exception to the norm, because
					//		 all the "searchKeys" of the first (leftmost) child node must be
					//		 less than its parent node "searchKey" ..

					// the parent "searchKey" should be 1 greater than the child's max value ..
					// (also, the "extra" node is at the end of the collection of keys) ..
					//int parentSearchKey = ( uniqueSearchKeys[ (uniqueSearchKeys.size() - 1) ] + 1 );
					int parentSearchKey = ( uniqueSearchKeys_->back() + 1 );

					parentSearchKeys->push_back( parentSearchKey );

					//cout << "IndexManager:: createParentNodes() III - parentSearchKeys->push_back( parentSearchKey ): " << parentSearchKey << endl << endl;

					// this should NOT be necessary ..
					//if( parentSearchKeysRow == NULL ) {
					//	parentSearchKeysRow = new vector< vector<int>* >();

					//	cout << "III - parentSearchKeysRow = new vector< vector<int>* >() .." << endl << endl;

					//}

					//parentSearchKeysRow->push_back( parentSearchKeys );

					//cout << "III - parentSearchKeysRow->push_back( parentSearchKeys ) .." << endl << endl;

				} // end "if()" ..


			} // end "for()" ..

			//cout << "IndexManager:: createParentNodes() Point A["<<rowLevel<<"].. " << endl << endl;

			// if parent searchKey finishes with less than 4 values,
			// still need to store it in the row ..
			// (also, no reason to store "parentSearchKeys" if it is NULL,
			//  i.e.: # leaf search keys <= 4 .. only |10 20 30 40| .. ) ..
			if( ( counter < numberSearchKeysPerNode ) &&
				( parentSearchKeys != NULL ) ) {

				//cout << "IndexManager:: createParentNodes() Point B["<<rowLevel<<"].. " << endl << endl;

				//cout << "IndexManager:: createParentNodes() IV - (parentSearchKeysRow == NULL): " << (parentSearchKeysRow == NULL) << endl << endl;

				if( parentSearchKeysRow == NULL )
					parentSearchKeysRow = new vector< vector<int>* >();

				// DEBUGGING ..
				//cout << "IndexManager:: createParentNodes() IV - parentSearchKeysRow->push_back( parentSearchKeys ): ";
				//for(int i=0; i<parentSearchKeys->size(); i++){

				//	cout << parentSearchKeys->at(i) << ", ";

				//}
				//cout << endl;

				//cout << "IndexManager:: createParentNodes() IV - parentSearchKeys->size(): " << parentSearchKeys->size() << endl << endl;

				parentSearchKeysRow->push_back( parentSearchKeys );

				//cout << "IndexManager:: createParentNodes() IV - parentSearchKeysRow->size(): " << parentSearchKeysRow->size() << endl << endl;

			}

			//cout << "IndexManager:: createParentNodes() Point C["<<rowLevel<<"].. " << endl << endl;


			// add the row to the index (so long as there's a row to add) ..
			if( ( parentNodes_ == NULL ) &&
				( parentSearchKeysRow != NULL ) )
				parentNodes_ = new vector< vector< vector<int>* >* >();

			//cout << "IndexManager:: createParentNodes() (parentNodes != NULL): " << (parentNodes_ != NULL) << endl << endl;

			//cout << "IndexManager:: createParentNodes() Point D["<<rowLevel<<"].. " << endl << endl;

			if ( parentSearchKeysRow != NULL )
				parentNodes_->push_back( parentSearchKeysRow );

			//cout << "IndexManager:: createParentNodes() (parentSearchKeysRow != NULL): " << (parentSearchKeysRow != NULL) << endl << endl;

			//cout << "IndexManager:: createParentNodes() Point E["<<rowLevel<<"].. " << endl << endl;

			// set the next "row" of parent nodes ..

			// clear the current row of values ..
			currentRowAllSearchKeys.clear();

			// make the parent "searchKeys" the new current "searchKeys" row ..
			// (i.e.: make the new current row out of the previous parent row) ..

			if( parentSearchKeysRow != NULL ){

				//cout << "IndexManager:: createParentNodes() Point F["<<rowLevel<<"].. " << endl << endl;

				//cout << "IndexManager:: createParentNodes() parentSearchKeysRow->size(): " << parentSearchKeysRow->size() << endl << endl;

				for(int i=0; i<parentSearchKeysRow->size(); i++){

					if( parentSearchKeysRow->at(i) != NULL ){

						//cout << "IndexManager:: createParentNodes() Point G["<<rowLevel<<"].. " << endl << endl;

						//cout << "IndexManager:: createParentNodes() parentSearchKeysRow->at(i)->size(): " << parentSearchKeysRow->at(i)->size() << endl << endl;

						for(int j=0; j<parentSearchKeysRow->at(i)->size(); j++){

							if( parentSearchKeysRow->at(i)->at(j) != NULL ){

								//cout << "IndexManager:: createParentNodes() Point H["<<rowLevel<<"].. " << endl << endl;

								currentRowAllSearchKeys.push_back( parentSearchKeysRow->at(i)->at(j) );

							}

						}

					}

				}

			}

			// DEBUGGING ..
			//cout << "IndexManager:: createParentNodes() Point I["<<rowLevel<<"].. " << endl << endl;
			//for(int i=0; i<currentRowAllSearchKeys.size(); i++){

			//	cout << "IndexManager:: createParentNodes() currentRowAllSearchKeys[ i("<<i<<") ]: " << currentRowAllSearchKeys[i] << endl << endl;

			//}

			//loopCounter++;
			//cout << "loopCounter: " << loopCounter << endl << endl;
			//if ( loopCounter > 2 ) {
			//	stopInfiniteLoop = true;
			//}

			//cout << "IndexPages:: createParentNodes() Point B .." << endl << endl;

			// initialize and set variable for each row that tracks whether
			// parent of last set of nodes should have either
			// 1) 1 child PageID or 2) 2 or more child pageIDs ..
			if ( extraParentNodeGeneratedPerRow_ == NULL )
				extraParentNodeGeneratedPerRow_ = new vector<bool>();

			if ( extraParentNodeGenerated == true )
				extraParentNodeGeneratedPerRow_->push_back( true );
			else
				extraParentNodeGeneratedPerRow_->push_back( false );

			// reset this parameter for the next row ..
			extraParentNodeGenerated = false;

			rowLevel++;		//  [ i % pow(5, rowLevel) * numberSearchKeysPerNode  == 0 ] && [ i % pow(5, rowLevel) * fanout != 0 ] ..


		} // end "while()" ..


	} // end "if()" ..


	// DSEBUGGING ..
	//if( extraParentNodeGeneratedPerRow_ != NULL ){
	//	cout << "IndexManager:: createParentNodes() extraParentNodeGeneratedPerRow_->size(): " << extraParentNodeGeneratedPerRow_->size() << endl << endl;
	//} else {
	//	cout << "IndexManager:: createParentNodes() extraParentNodeGeneratedPerRow_ == NULL" << endl << endl;
	//}


	// DEBUGGING ..
/*	if( parentNodes_ != NULL ){

		cout << "IndexManager:: createParentNodes() FINISHED PARENT NODES: " << endl << endl;

		cout << "IndexManager:: createParentNodes() parentNodes->size(): " << parentNodes_->size() << endl << endl;

		for(int i=0; i<parentNodes_->size(); i++){	// "parentNodes" ..

			cout << "IndexManager:: createParentNodes() parentNodes->at( i("<<i<<") )->size(): " << parentNodes_->at(i)->size() << endl << endl;

			for(int j=0; j<parentNodes_->at(i)->size(); j++){	// "parentSearchKeysRow" ..

				cout << "IndexManager:: createParentNodes() parentNodes->at( i("<<i<<") )->at( j("<<j<<") )->size(): " << parentNodes_->at(i)->at(j)->size() << endl << endl;

				for(int k=0; k<parentNodes_->at(i)->at(j)->size(); k++){		// "parentSearchKeys" ..

					cout << parentNodes_->at(i)->at(j)->at(k) << ", ";

				} // end 1 parent node ..

				cout << "	";

			} // end 1 parent node row ..

			cout << endl << endl;

		} // end all parent node rows ..

	} else {

		cout << "IndexManager:: createParentNodes() parentNodes_ == NULL" << endl << endl;

	} // end "if()" ..
*/

} // end "createParentNodes()" ..

void IndexManager:: createIndexPages( vector<int>* allSearchKeys, vector<int>* searchKeyPageIDs, vector<IndexPage*>* indexPages_, IndexHeader* indexHeader, vector< vector< vector<int>* >* >* parentNodes_, vector<bool>* extraParentNodeGeneratedPerRow_, bool isNoDuplicates_ ){

	//cout << "IndexManager:: createIndexPages() start .." << endl << endl;

	// 					   "indexPages_": collection to be filled herein ..
	// 					  "parentNodes_": parent< row < key > > collection used here
	// "extraParentNodeGeneratedPerRow_": tracks rows of index which have exceptional extra-parent nodes

	// need to make "IndexPages" for each node of the index ..
	if( indexPages_ == NULL )
		indexPages_ = new vector<IndexPage*>();

	//cout << "IndexManager:: createIndexPages() indexPages_->size(): " << indexPages_->size() << endl << endl;

	//int numberSearchKeysPerNode = 4;

	int fanout = 5;

	int indexPageCounter = 0;


	//cout << "IndexManager:: createIndexPages() ( parentNodes_ == NULL ): " << ( parentNodes_ == NULL ) << endl << endl;

	if( parentNodes_ != NULL ){

		IndexPage* indexPage;

		int maxPageID = 0;

		vector<int>* childIndexPageIDs;

		// update the "IndexHeader" ..
		//indexHeader = new IndexHeader();

		//indexHeader->setTotalNumberOfIndexPages( indexPages_->size() );

		IndexHeaderEntry* indexHeaderEntry = NULL;

		IndexRows* indexRows = new IndexRows;
		IndexNodeRow* indexNodeRow = NULL;
		IndexNode* indexNode = NULL;

		// number of rows ..
		int parentNodesSize = parentNodes_->size();

		//cout << "IndexManager:: createIndexPages() ( parentNodesSize - 1 ): " << (parentNodesSize - 1) << endl << endl;

		int searchKeyRowSize = 0;

		//cout << "================================================" << endl << endl;

		// start at the root node, work down to 1-level-above the leaf nodes ..
		// NOTE: "parentNodes" has rows of parent nodes, ROOT node stored LAST) ..
		//for(int i=( parentNodesSize - 1 ); i>0; i--){
		for(int i=( parentNodesSize - 1 ); i>=0; i--){
		//for(int i=parentNodesSize; i>=0; i--){

			// make an index that starts with "0" at the root node and increases from there ..
			int rowNumber = ( ( parentNodesSize - 1 ) - i );
			//int rowNumber = ( parentNodesSize - i );

			//int parentNodeIndex = (i-1);

			// DEBUGGING ..
			//cout << "IndexManager:: createIndexPages()        ( parentNodesSize - 1 )["<<i<<"]: " << (parentNodesSize - 1) << endl << endl;
			//cout << "IndexManager:: createIndexPages()( ( parentNodesSize - 1 ) - i )["<<i<<"]: " << ((parentNodesSize - 1) - i) << endl << endl;
			//cout << "IndexManager:: createIndexPages()                      rowNumber["<<i<<"]: " << rowNumber << endl << endl;

			//cout << "IndexManager:: createIndexPages() Point A .." << endl << endl;

			// iterate through EACH parent node in the ROW ..
			searchKeyRowSize = parentNodes_->at(i)->size();

			//cout << "IndexManager:: createIndexPages() Point B .." << endl << endl;

			// track the next row ..
			indexNodeRow = new IndexNodeRow;
			indexRows->addRow( indexNodeRow );

			//cout << "IndexManager:: createIndexPages() Point C .." << endl << endl;

			// iterate through each node in the row ..
			for(int j=0; j<searchKeyRowSize; j++){

				//cout << "IndexManager:: createIndexPages() Point D .." << endl << endl;

				// initialize a new "indexPage" to encapsulate this parent node ..
				indexPage = new IndexPage();

				// set the "searchKeys" for this parent node using the values gathered above ..
				indexPage->setSearchKeys( parentNodes_->at(i)->at(j) );

				int searchKeySize = parentNodes_->at(i)->at(j)->size();

				// DEBUGGING ..
				//cout << "IndexManager:: createIndexPages() indexPage->setSearchKeys( parentNodes_->at(i)->at(j) ): ";
				//for(int k=0; k<searchKeySize; k++){

				//	cout << parentNodes_->at(i)->at(j)->at(k) << ", ";

				//}
				//cout << endl << endl;

				//cout << "IndexManager:: createIndexPages() Point E .." << endl << endl;


				// track the "pageID", "parentID" for this node ..
				indexNode = new IndexNode();
				indexNode->setIndexPageID( indexPageCounter );

				//cout << "IndexManager:: createIndexPages() Point F .." << endl << endl;

				int parentID = -1;

				if( rowNumber == 0 ) {

					indexNode->setParentIndexPageID( -1 );		// root node has no parent ..

					//cout << "IndexManager:: createIndexPages() Point G .." << endl << endl;

				} else if ( rowNumber == 1 ) {

					indexNode->setParentIndexPageID( 0 );		// next row all have root node as parent ..

					parentID = 0;

					//cout << "IndexManager:: createIndexPages() Point H .." << endl << endl;

				} else {

					//cout << "IndexManager:: createIndexPages() Point I .." << endl << endl;

					// look at the previous row,
					// get the "pageID" of the node that is the parent-of-the-current-node ..
					bool parentIDFound = false;

					vector<int>* childIndexPageIDs = NULL;

					// get the previous node row (as collection of IndexNodes) ..
					vector<IndexNode*>* nodeRow = indexRows->getRow( ( rowNumber - 1 ) )->getRow();

					//cout << "nodeRow->size() ["<<i<<"]["<<j<<"]: " << nodeRow->size() << endl << endl;

					int nodeRowSize = nodeRow->size();
					int childIndexPageIDsSize = -1;

					for( int k=0; k<nodeRowSize; k++ ){

						// get the "childIndexPageIDs" for each "IndexNode" in this row ..
						childIndexPageIDs = nodeRow->at( k )->getChildIndexPageIDs();

						//cout << "childIndexPageIDs->size() ["<<i<<"]["<<j<<"]: " << childIndexPageIDs->size() << endl << endl;

						childIndexPageIDsSize = childIndexPageIDs->size();

						for( int m=0; m<childIndexPageIDsSize; m++ ){

							//cout << "( childIndexPageIDs->at( m("<<m<<") )("<<childIndexPageIDs->at( m )<<") == indexPageCounter("<<indexPageCounter<<") ) ["<<i<<"]["<<j<<"]: " << ( childIndexPageIDs->at( m ) == indexPageCounter ) << endl << endl;

							// the child pageID of the parent matches the current pageID ..
							if( childIndexPageIDs->at( m ) == indexPageCounter ){

								// parentID of current node is a pageID from a row above ..
								parentIDFound = true;
								parentID = nodeRow->at( k )->getIndexPageID();
								break;
							}

						} // end "for()" ..

						if( parentIDFound == true )
							break;

					} // end "for()" ..

					//cout << "IndexManager:: createIndexPages() parentIDFound row["<<i<<"] col["<<j<<"]: " << parentIDFound << endl << endl;
					//cout << "     IndexManager:: createIndexPages() parentID row["<<i<<"] col["<<j<<"]: " << parentID << endl << endl;

					if( parentIDFound == true )
						indexNode->setParentIndexPageID( parentID );
					else
						indexNode->setParentIndexPageID( -1 );

				} // end "if()" ..

				//cout << "IndexManager:: createIndexPages() Point J .." << endl << endl;

				// EXCEPTIONAL CASES:
				// #2) 1 parent for 5 child nodes, + 1 extra node, no parent ..
				//int parentNodeLevel = ( parentNodes_->size() - 1 - i );	// root node = level 0 ( @ i == parentNodes->size() - 1 ) ..

				//cout << "IndexManager:: createIndexPages() Point K .." << endl << endl;

				// check whether this "extra" parent node has either 1) 1 child, or 2) more than 1 child ..

				//cout << "IndexManager:: createIndexPages() extraParentNodeGeneratedPerRow_->size(): " << extraParentNodeGeneratedPerRow_->size() << endl << endl;

				bool hasTwoOrMoreChildSearchKeys = true;
				if( extraParentNodeGeneratedPerRow_->at(i) == true )
					hasTwoOrMoreChildSearchKeys = false;

				//cout << "( extraParentNodeGeneratedPerRow->at(i("<<i<<") )("<<extraParentNodeGeneratedPerRow_->at(i)<<") == true )( maxPageID: "<<maxPageID<<" ): " << ( extraParentNodeGeneratedPerRow_->at(i) == true ) << endl << endl;

				bool parentNeedsOnlyOneChildId = false;

				//if( ( parentNodeLevel >= 1 ) && 						// this parent node is at least 1 level below root node ..
				//if( ( parentNodeLevel >= 0 ) && 						// this parent node is at the very least the root node ..
				if( ( rowNumber >= 0 ) && 								// this parent node is at the very least the root node ..
					( j == ( parentNodes_->at(i)->size() - 1 ) ) &&		// last node in the row ..
					( parentNodes_->at(i)->at(j)->size() == 1 ) &&		// node only has 1 searchKey ..
					( hasTwoOrMoreChildSearchKeys == false ) ) {		// parent node was generated for a single child node ..

					parentNeedsOnlyOneChildId = true;

				}

				// DEBUGGING ..
				////cout << "( parentNodeLevel("<<parentNodeLevel<<") >= 0 )( maxPageID: "<<maxPageID<<" ): " << ( parentNodeLevel >= 0 ) << endl << endl;
				//cout << "( rowNumber("<<rowNumber<<") >= 0 )( maxPageID: "<<maxPageID<<" ): " << ( rowNumber >= 0 ) << endl << endl;
				////cout << "extraParentNodeGenerated( maxPageID: "<<maxPageID<<" ): " << extraParentNodeGenerated << endl << endl;
				//cout << "( j("<<j<<") == ( parentNodes_->at(i)->size() - 1 )("<<( parentNodes_->at(i)->size() - 1 )<<") )( maxPageID: "<<maxPageID<<" ): " << ( j == ( parentNodes_->at(i)->size() - 1 ) ) << endl << endl;
				//cout << "( ( parentNodes_->at(i)->at(j)->size()("<<parentNodes_->at(i)->at(j)->size()<<") == 1 ) )( maxPageID: "<<maxPageID<<" ): " << ( parentNodes_->at(i)->at(j)->size() == 1 ) << endl << endl;
				//cout << "hasTwoOrMoreChildSearchKeys( maxPageID: "<<maxPageID<<" ): " << hasTwoOrMoreChildSearchKeys << endl << endl;
				//cout << "parentNeedsOnlyOneChildId( maxPageID: "<<maxPageID<<" ): " << parentNeedsOnlyOneChildId << endl << endl;

				//cout << "================================================" << endl << endl;

				// find out how many child nodes this parent node has ..
				// (i.e.: if parent node has 2 searchKeys, it has 3 child nodes) ..
				int numberOfChildNodes = -1;
				if( parentNeedsOnlyOneChildId ){

					numberOfChildNodes = 1;

				} else {

					numberOfChildNodes = ( parentNodes_->at(i)->at(j)->size() + 1 );

				}

				//cout << "IndexManager:: createIndexPages() numberOfChildNodes: " << numberOfChildNodes << endl << endl;

				// initialize a new collection of child "pageIDs" ..
				childIndexPageIDs = new vector<int>();

				int newPageID = -1;


				// calculate the "pageIDs" for each child of this parent node ..
				for(int k=0; k<numberOfChildNodes; k++){


					// i.e.: if "maxPageID = 0", "newPageID = 1, 2, 3 .."
					//		 if "maxPageID = 3", "newPageID = 4, 5, 6 ..", etc ..
					newPageID = ( ( maxPageID + 1 ) + k );

					//cout << "IndexManager:: createIndexPages() childIndexPageIDs: " << newPageID << endl << endl;

					// collect the child "pageIDs" for use just below ..
					childIndexPageIDs->push_back( newPageID );

				}

				// track the childIDs for this row ..
				indexNode->setChildIndexPageIDs( childIndexPageIDs );

				// track this node by adding this node to the row of nodes ..
				IndexNodeRow* inr = indexRows->getRow( rowNumber );

				if( inr != NULL )
					inr->addIndexNode( indexNode );


				// update the latest "maxPageID" ..
				if( newPageID > maxPageID ){

					maxPageID = newPageID;

				}


				// this indicates processing the "leaf node" row ..
				if( i == 0 ){

					//cout << "IndexManager:: createIndexPages() Point L .." << endl << endl;

					// need to link the leaf nodes with
					// 1) buckets (isNoDuplicates == false), or
					// 2) data pages (isNoDuplicates == true)

					if( isNoDuplicates_ == true ){

						//cout << "IndexManager:: createIndexPages() Point M .." << endl << endl;

						// need to link the leaf nodes with "data pages" ..
						// "childIndexPageIDs" should have pointers to data pages ..
						childIndexPageIDs = new vector<int>();

						// link "allSearchKeys" to "parentNodeSearchKeys" ..

						// this is the searchKeys constructed for the index ..
						vector<int>* parentNodeSearchKeys = parentNodes_->at(i)->at(j);

						// these are loop limits ..
						int parentNodeSearchKeysSize = parentNodeSearchKeys->size();
						int allSearchKeysSize = allSearchKeys->size();

						// loop through the searchKeys for the parent node ..
						for(int k=0; k<parentNodeSearchKeysSize; k++){

							// loop through each searchKey ..
							for(int m=0; m<allSearchKeysSize; m++){

								// compare the parent node searchKey to each searchKey ..
								if( allSearchKeys->at( m ) == parentNodeSearchKeys->at( k ) ){

									childIndexPageIDs->push_back( searchKeyPageIDs->at( m ) );

								}

							}

						}

						// leaf nodes with "data pages" ..
						indexPage->setChildPageIDs( childIndexPageIDs );


					} else {

						//cout << "IndexManager:: createIndexPages() Point N .." << endl << endl;

						// need to link the leaf nodes with "buckets" ..
						// "childIndexPageIDs" should not point to anything,
						// b/c IndexBucketHeader contains link btw search keys and bucketIDs ..
						// this is done via "createIndex() writeIndexBuckets()" ..
					}

					indexPage->setIsLeafNode( true );

				} else {

					//cout << "IndexManager:: createIndexPages() Point O .." << endl << endl;

					// set the child "pageIDs" to be those calculated here ..
					indexPage->setChildPageIDs( childIndexPageIDs );

					indexPage->setIsLeafNode( false );

				}

				// DEBUGGING ..
				//cout << "        ( parentNodesSize == 1 ) ["<<i<<"]["<<j<<"]: " << ( parentNodesSize == 1 ) << endl << endl;
				//cout << "( i("<<i<<") == ( parentNodesSize - 1 ) ) ["<<i<<"]["<<j<<"]: " << ( i == ( parentNodesSize - 1 ) ) << endl << endl;
				//cout << "                      ( i("<<i<<") >= 0 ) ["<<i<<"]["<<j<<"]: " << ( i >= 0 ) << endl << endl;


				// if there is only 1 node in this tree,
				// then for this node which has 1 parent, parentID = -1 ..
				// or,
				// if this node is the root node, parentID = -1 ..

				int pageID = -1;
				int parentPageID = -1;

				if( ( parentNodesSize == 1 ) || 				// there is only 1 node in the index ..
					( i == ( parentNodesSize - 1 ) ) ) {		// current node is the "root" node ..

					pageID = 0;
					
					indexHeader->setRootPageID( pageID );

				} else if( i >= 0 ){							// Handle the non-root nodes ..

					pageID = indexPageCounter;

					parentPageID = parentID;

				}

				indexPage->setPageID( pageID );

				// add the "indexPage" to the collection ..
				indexPages_->push_back( indexPage );

				// prepare to set "header entry" values ..
				indexHeaderEntry = new IndexHeaderEntry();

				// search keys available in this node == "max" - "currently occupied" ..
				int freeSearchKeyCount = ( (fanout - 1) - parentNodes_->at(i)->at(j)->size() );




				// DEBUGGING ..
				//cout << "            IndexManager:: createIndexPages() pageID ["<<i<<"]["<<j<<"]: " << pageID << endl << endl;
				//cout << "      IndexManager:: createIndexPages() parentPageID ["<<i<<"]["<<j<<"]: " << parentPageID << endl << endl;
				//cout << "IndexManager:: createIndexPages() freeSearchKeyCount ["<<i<<"]["<<j<<"]: " << freeSearchKeyCount << endl << endl;

				indexHeaderEntry->setPageID( pageID );
				indexHeaderEntry->setParentPageID( parentPageID );
				indexHeaderEntry->setFreeSearchKeyCount( freeSearchKeyCount );

				//cout << "IndexPages:: createIndexPages() Point A .." << endl << endl;

				vector<bool>* searchKeysAvailable = new vector<bool>();

				for( int k=0; k<( (fanout - 1) - freeSearchKeyCount ); k++ )
					searchKeysAvailable->push_back( false );

				for( int k=0; k<freeSearchKeyCount; k++ )
					searchKeysAvailable->push_back( true );

				// DEBUGGING ..
				//for(int k=0; k<(fanout - 1);k++)
				//	cout << "searchKeysAvailable->at(k)("<<k<<") ["<<i<<"]["<<j<<"]: " << searchKeysAvailable->at(k) << endl << endl;

				indexHeaderEntry->setSearchKeysAvailable( searchKeysAvailable );

				// add the new "indexHeaderEntry" to the "IndexHeader" collection ..
				indexHeader->getHeaderEntries()->push_back( indexHeaderEntry );
				//indexHeader.getHeaderEntries()->push_back( indexHeaderEntry );

				//indexHeader->setTotalNumberOfIndexPages( ( indexHeader->getTotalNumberOfIndexPages() + 1 ) );
				int newTotalNumberOfIndexPages = ( indexHeader->getTotalNumberOfIndexPages() + 1 );
				indexHeader->setTotalNumberOfIndexPages( newTotalNumberOfIndexPages );

				// tracks the current indexPage's "pageID" ..
				indexPageCounter++;


			} // end "for()" ..

		} // end "for()" ..

		// DEBUGGING ..
		//cout << "================================================" << endl << endl;

		//cout << "IndexManager:: createIndexPages() totalNumberOfIndexPages: " << indexHeader->getTotalNumberOfIndexPages() << endl << endl;

		//vector<IndexHeaderEntry*>* indexHeaderEntries = indexHeader->getHeaderEntries();

		//IndexHeaderEntry* indexHeaderEntry1 = NULL;

		//cout << "IndexManager:: createIndexPages() indexHeaderEntries->size(): " << indexHeaderEntries->size() << endl << endl;

		//for(int i=0; i<indexHeaderEntries->size(); i++){

		//	indexHeaderEntry1 = indexHeaderEntries->at(i);

		//	cout << "IndexManager:: createIndexPages() pageID: " << indexHeaderEntry1->getPageID() << endl << endl;

		//	cout << "IndexManager:: createIndexPages() parentPageID: " << indexHeaderEntry1->getParentPageID() << endl << endl;

		//	cout << "IndexManager:: createIndexPages() freeSearchKeyCount: " << indexHeaderEntry1->getFreeSearchKeyCount() << endl << endl;

		//	vector<bool>* searchKeysAvailable = indexHeaderEntry1->getSearchKeysAvailable();

		//	for(int j=0; j<searchKeysAvailable->size(); j++){

		//		cout << "IndexManager:: createIndexPages() searchKeysAvailable["<<j<<"]: " << searchKeysAvailable->at(j) << endl << endl;

		//	}

		//	cout << endl << endl;


		//} // end "for()" ..


/*
		// DEBUGGING ..
		cout << "================================================" << endl << endl;
		cout << "IndexManager:: createIndexPages() indexHeader->getTotalNumberOfIndexPages(): " << indexHeader->getTotalNumberOfIndexPages() << endl << endl;

		// track the "headerEntries" for the "leftIndexPage" ..
		vector<IndexHeaderEntry*>* headerEntries = indexHeader->getHeaderEntries();
		//vector<IndexHeaderEntry*>* headerEntries = indexHeader.getHeaderEntries();
		int headerEntryIndex = -1;


		for(int i=0; i<indexPages_->size(); i++){

			//cout << "IndexManager:: createIndexPages() node: " << i << endl << endl;
			//cout << "IndexManager:: createIndexPages() indexPages[ i("<<i<<") ] searchKeys: ";

			for(int j=0; j<indexPages_->at(i)->getSearchKeys()->size(); j++){

				cout << indexPages_->at(i)->getSearchKeys()->at( j ) << ", ";

			}

			cout << endl << endl;


			//cout << "IndexManager:: createIndexPages() indexPages[ i("<<i<<") ] childPageIDs: ";

			for(int j=0; j<indexPages_->at(i)->getChildPageIDs()->size(); j++){

				cout << indexPages_->at(i)->getChildPageIDs()->at( j ) << ", ";

			}

			cout << endl << endl;

			cout << "IndexManager:: createIndexPages() indexPages[ i("<<i<<") ] isLeafNode: " << indexPages_->at(i)->getIsLeafNode() << endl << endl;

			// reset for next iteration ..
			headerEntryIndex = 0;

			for( int k=0; k<headerEntries->size(); k++ ){

				cout << "IndexManager:: createIndexPages() indexPages[ i("<<i<<") ] ( headerEntries->at(k("<<k<<"))->getPageID() == indexPages_->at(i("<<i<<"))->getPageID() ): " << ( headerEntries->at(k)->getPageID() == indexPages_->at(i)->getPageID() ) << endl << endl;

				if( headerEntries->at(k)->getPageID() == indexPages_->at(i)->getPageID() ){

					headerEntryIndex = k;
					break;

				}

			}

			cout << "            IndexManager:: createIndexPages() headerEntries->at( headerEntryIndex("<<headerEntryIndex<<") )->getPageID(): " << headerEntries->at( headerEntryIndex )->getPageID() << endl << endl;

			cout << "      IndexManager:: createIndexPages() headerEntries->at( headerEntryIndex("<<headerEntryIndex<<") )->getParentPageID(): " << headerEntries->at( headerEntryIndex )->getParentPageID() << endl << endl;

			cout << "IndexManager:: createIndexPages() headerEntries->at( headerEntryIndex("<<headerEntryIndex<<") )->getFreeSearchKeyCount(): " << headerEntries->at( headerEntryIndex )->getFreeSearchKeyCount() << endl << endl;

			cout << "================================================" << endl << endl;


		} // end "for()" ..
*/


	} // end "if()" ..


} // end "createIndexPages()" ..


void IndexManager:: addPageIDsToIndexBucket( vector<int>* searchKeyPageIDs_, IndexBucket* indexBucket_, int startPosition, int endPosition ){

	// helper method for "createIndex()" ..

	//cout << "IndexManager:: addPageIDsToIndexBucket() start .." << endl << endl;

	vector<int>* indexBucketPageIDs = indexBucket_->getPageIDs();
	
	int indexBucketPageIDsSize = -1;
	
	if( indexBucketPageIDs != NULL )
		indexBucketPageIDsSize = indexBucketPageIDs->size();
		
	int searchKeyPageID = -1;
	
	//cout << "IndexManager:: addPageIDsToIndexBucket() ( indexBucketPageIDs == NULL ): " << ( indexBucketPageIDs == NULL ) << endl << endl;
	
	//cout << "IndexManager:: addPageIDsToIndexBucket() indexBucketPageIDsSize: " << indexBucketPageIDsSize << endl << endl;
	
	//cout << "IndexManager:: addPageIDsToIndexBucket() Point A .." << endl << endl;

	// DEBUGGING ..
	//cout << "IndexManager:: addPageIDsToIndexBucket() indexBucketPageIDs->size(): " << indexBucketPageIDs->size() << endl << endl;
	//cout << "IndexManager:: addPageIDsToIndexBucket() startPosition: " << startPosition << endl << endl;
	//cout << "IndexManager:: addPageIDsToIndexBucket() endPosition: " << endPosition << endl << endl;

	// iterate from "startPosition" to "endPosition", getting unique "pageIDs" ..
	bool pageIDAlreadyInBucket;
	for(int j=startPosition; j<=endPosition; j++){

		// check that the pageID is not already in the bucket ..
		pageIDAlreadyInBucket = false;
		
		searchKeyPageID = searchKeyPageIDs_->at(j);
		
		for(int k=0; k<indexBucketPageIDsSize; k++){

			// DEBUGGING ..
			//cout << "IndexManager:: addPageIDsToIndexBucket() ( indexBucketPageIDs->at( k("<<k<<") )("<<indexBucketPageIDs->at(k)<<") == searchKeyPageIDs[ j("<<j<<") ] ]("<<searchKeyPageIDs_[j]<<") ): " << ( indexBucketPageIDs->at(k) == searchKeyPageIDs_[j] ) << endl << endl;

			if( indexBucketPageIDs->at(k) == searchKeyPageID ){

				pageIDAlreadyInBucket = true;
				break;

			}

		} // end "for()" ..

		// DEBUGGING ..
		//cout << "IndexManager:: addPageIDsToIndexBucket() pageIDAlreadyInBucket: " << pageIDAlreadyInBucket << endl << endl;


		// add "pageID" to bucket (but only if its not already there) ..
		if( !pageIDAlreadyInBucket ) {


			// DEBUGGING ..
			//cout << "IndexManager:: addPageIDsToIndexBucket() indexBucketPageIDs->push_back( searchKeyPageIDs_[ j("<<j<<") ] ): " << searchKeyPageIDs_[j] << endl << endl;

			indexBucketPageIDs->push_back( searchKeyPageID );

		}


	} // end "for()" ..
	
	//cout << "IndexManager:: addPageIDsToIndexBucket() Point B .." << endl << endl;

	// cleanup ..
	//delete indexBucketPageIDs;


} // end "addPageIDToIndexBucket()" ..


bool IndexManager:: writeIndexHeader( string tableName_, string indexName_, IndexHeader* header_ ) {

	// "tableName": name of the table associated with the header ..

	//    "header": header to write to file ..

	//cout << "IndexManager:: writeIndexHeader() start .." << endl << endl;

	bool writeHeaderPerformed = false;

	string emptyString = "";

	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {

		//cout << "IndexManager:: writeIndexHeader() Point A .." << endl << endl;

		// handle the directory path of the file ..

		//string directoryPlusFileName = "C:\\Temp\\";
		
		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusFileName = databaseFile->getFileDrive();
		
		//cout << "IndexManager:: writeIndexHeader() Point A1 .." << endl << endl;

		directoryPlusFileName += databaseFile->getFileDirectory();
		
		//cout << "IndexManager:: writeIndexHeader() Point A2 .." << endl << endl;


		// append the file
		directoryPlusFileName += "IndexHeader_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();
		
		//cout << "IndexManager:: writeIndexHeader() Point A3 .." << endl << endl;

		fstream tbl( fileName, ios::in | ios::out | ios::binary );

		//cout << "IndexManager:: writeIndexHeader() Point B .." << endl << endl;

		// this means the file doesn't exist yet,
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){

			tbl.open( fileName, ios::out | ios::binary | ios::app );

			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();

		}

		if ( !tbl.fail() ) {

			//cout << "IndexManager:: writeIndexHeader() Point C .." << endl << endl;

			tbl.seekp( ios::beg );

			// number of fields in the "header" ..
			int numberOfHeaderFields = 1;

			// number of fields in each "headerEntry" of the "header" ..
			int numberOfHeaderEntryFields = 5;
			
			int rootPageID = header_->getRootPageID();

			int totalNumberOfIndexPages = header_->getTotalNumberOfIndexPages();
			
			bool isDuplicatesAllowed = header_->getIsDuplicatesAllowed();

			//cout << "IndexManager:: writeIndexHeader() Point D .." << endl << endl;

			// write the "header" fields to the header file ..

			tbl.write( reinterpret_cast<const char*>( &rootPageID ), intSize );

			tbl.write( reinterpret_cast<const char*>( &totalNumberOfIndexPages ), intSize );
			
			tbl.write( reinterpret_cast<const char*>( &isDuplicatesAllowed ), boolSize );

			//cout << "IndexManager:: writeIndexHeader() Point E .." << endl << endl;

			// write the "headerEntry" fields to the header file ..

			int pid = 0;

			int nosk = 0;

			int nocpi = 0;

			int ppid = 0;

			int fskc = 0;
			
			//cout << "IndexPages:: writeIndexHeader() Point A .." << endl << endl;

			vector<bool>* ska = new vector<bool>();

			//cout << "IndexManager:: writeIndexHeader() Point F .." << endl << endl;

			//cout << "IndexManager:: writeIndexHeader() header_->getTotalNumberOfIndexPages(): " << header_->getTotalNumberOfIndexPages() << ", header_->getHeaderEntries()->size(): " << header_->getHeaderEntries()->size() << endl << endl;

			for(int i=0; i<totalNumberOfIndexPages; i++){


				pid = header_->getHeaderEntries()->at( i )->getPageID();
				tbl.write( reinterpret_cast<const char*>( &pid ), intSize );
				//cout << "IndexManager:: writeIndexHeader() pageID: " << pid << endl << endl;

				nosk = header_->getHeaderEntries()->at( i )->getNumberOfSearchKeys();
				tbl.write( reinterpret_cast<const char*>( &nosk ), intSize );
				//cout << "IndexManager:: writeIndexHeader() numberOfSearchKeys: " << nosk << endl << endl;

				nocpi = header_->getHeaderEntries()->at( i )->getNumberOfChildPageIDs();
				tbl.write( reinterpret_cast<const char*>( &nocpi ), intSize );
				//cout << "IndexManager:: writeIndexHeader() numberOfChildPageIDs: " << nocpi << endl << endl;

				ppid = header_->getHeaderEntries()->at( i )->getParentPageID();
				tbl.write( reinterpret_cast<const char*>( &ppid ), intSize );
				//cout << "IndexManager:: writeIndexHeader() parentPageID: " << ppid << endl << endl;

				fskc = header_->getHeaderEntries()->at( i )->getFreeSearchKeyCount();
				tbl.write( reinterpret_cast<const char*>( &fskc ), intSize );
				//cout << "IndexManager:: writeIndexHeader() freeSearchKeyCount: " << fskc << endl << endl;

				ska = header_->getHeaderEntries()->at( i )->getSearchKeysAvailable();
				bool searchKeyAvailable = true;
				for(int j=0; j<ska->size(); j++) {

					searchKeyAvailable = ska->at( j );

					tbl.write( reinterpret_cast<const char*>( &searchKeyAvailable ), boolSize );

					//cout << "testIndexManager_writeIndexHeader() searchKeyAvailable: " << searchKeyAvailable << endl << endl;

				}

			}

			//cout << "IndexManager:: writeIndexHeader() Point G .." << endl << endl;

			//cout << "IndexManager:: writeIndexHeader() ska == NULL: " << (ska == NULL) << endl << endl;

			// size in bytes of the "rowsAvailable" array found in each "headerEntry" of the "header" ..
			// [ (2 bytes) * (N rows per page) ]
			int searchKeysAvailableSize = ( boolSize * ska->size() );

			//cout << "IndexManager:: writeIndexHeader() Point H .." << endl << endl;

			// size in bytes of a single "headerEntry" of the "header" ..
			// [ [ (4 bytes) * (4 fields) ] +  ]
			int headerPageEntrySize =  ( ( intSize * numberOfHeaderEntryFields ) + searchKeysAvailableSize );

			//cout << "IndexManager:: writeIndexHeader() Point I .." << endl << endl;

			// size in bytes of the header file ..
			// [ (4 bytes) * (2 fields) ] + [ (M data pages) * (single header entry size) ]
			// 8 bytes + M*16 bytes == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			int headerSize = ( ( intSize * numberOfHeaderFields ) + ( totalNumberOfIndexPages * headerPageEntrySize ) );

			//cout << "IndexManager:: writeIndexHeader() Point J .." << endl << endl;

			int fillerSize = 0;

			int gaugeTheFillerSize = blockSize;


			// increase "gaugeTheFillerSize" by increments of "blockSize",
			// until "gaugeTheFillerSize >= headerSize" ..
			// i.e.:         headerSize == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			//		      exit the loop: (48 >= 24), (48 >= 40), (96 >= 56), (96 >= 72), (96 >= 88) ..
			while ( true ){

				if( gaugeTheFillerSize >= headerSize ){

					break;

				} else {

					gaugeTheFillerSize += blockSize;

				}

			} // end "while ( true )" ..


			//cout << "IndexManager:: writeIndexHeader() Point K .." << endl << endl;

			// "fillerSize" should be less than 48 bytes,
			// such that adding "fillerSize" to "headerSize"
			// is an even multiple of "blockSize" ..
			// i.e.:         headerSize == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			//               fillerSize == (48 - 24), (48 - 40), (96 - 56), (96 - 72), (96 - 88), (144 - 104) ..
			fillerSize = ( gaugeTheFillerSize - headerSize );


			if ( fillerSize > 0 ) {

				//string filler = constructPagePadding( fillerSize );
				string filler = getStorageManager()->constructPagePadding( fillerSize );

				int charSize = 1;

				// write the filler string to the DataHeader file,
				// so that the header file size is an even multiple of "blockSize" ..
				for(int i=0; i<filler.size(); i++){

					//cout << filler[i];

					tbl.write( reinterpret_cast<const char*>( &filler[i] ), charSize );

				}

			}

			//cout << "IndexManager:: writeIndexHeader() Point L .." << endl << endl;

			// flag indicating the "writeHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			writeHeaderPerformed = true;


		} else {

	 		string errorString = "\nHeader file failed to open [ SM::writeIndexHeader() ], so Header file was not written.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();


	} else {

	 	string errorString = "\nHeader file table name was not valid [ SM::writeIndexHeader() ], so Header file was not written.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	//cout << "IndexManager:: writeIndexHeader() Point M .." << endl << endl;

	return writeHeaderPerformed;


} // end "writeIndexHeader()" ..


bool IndexManager:: writeIndexBucketHeader( string tableName_, string indexName_, IndexBucketHeader* indexBucketHeader_ ) {

	// "tableName": name of the table associated with the header ..

	//    "header": header to write to file ..


	//cout << "IndexManager:: writeIndexBucketHeader() start .." << endl << endl;

	//cout << "IndexManager:: writeIndexBucketHeader() Point A .." << endl << endl;

	bool writeHeaderPerformed = false;

	string emptyString = "";

	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {


		//cout << "IndexManager:: writeIndexBucketHeader() Point D .." << endl << endl;

		// handle the directory path of the file ..

		//string directoryPlusFileName = "C:\\Temp\\";

		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		// append the file
		directoryPlusFileName += "IndexBucketHeader_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();

		fstream tbl( fileName, ios::in | ios::out |ios::binary );

		// this means the file doesn't exist yet,
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){

			tbl.open( fileName, ios::out | ios::binary | ios::app );

			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();

		}

		if ( !tbl.fail() ) {

			//cout << "IndexManager:: writeIndexBucketHeader() Point E .." << endl << endl;

			tbl.seekp( ios::beg );

			// write the "header" fields to the header file ..

			// "int _totalNumberOfBuckets"
			int tnb = indexBucketHeader_->getTotalNumberBuckets();
			tbl.write( reinterpret_cast<const char*>( &tnb ), intSize );

			//cout << "IndexManager:: writeIndexBucketHeader() tbl.write( tnb ): " << tnb << endl << endl;

			//cout << "IndexManager:: createIndex() writeIndexBucketHeader() Point F .." << endl << endl;

			// write the "headerEntry" fields to the header file ..

/*
			// "map<int, vector<int>* >* _indexBucketIDsForEachSearchKey"
			map<int, vector<int>* >* iBIDsFESK = NULL;
			int sk = 0;						// search key
			int numberBucketIDs;			// # bucketIDs in vector<int>*
			vector<int>* bucketIDs = NULL;
			int bucketIDForSearchKey;		// each bucketID in vector<int>*
			int searchKeySize = 0,
				numberSearchKeySize = 0,
				bucketIDsForSearchKeySize = 0;

			map<int, vector<int>* >::iterator it_map;
			iBIDsFESK = indexBucketHeader_.getBucketIDsForEachSearchKey();
			for( it_map = iBIDsFESK->begin(); it_map != iBIDsFESK->end(); it_map++ ){

				// search key for a set of bucket ids ..
				sk = it_map->first;
				tbl.write( reinterpret_cast< const char * >( &( sk ) ), intSize );
				searchKeySize++;
				cout << "IndexManager:: createIndex() writeIndexBucketHeader() tbl.write( sk ): " << sk << endl << endl;

				// number of bucket ids for this search key ..
				numberBucketIDs = ( it_map->second )->size();
				tbl.write( reinterpret_cast< const char * >( &( numberBucketIDs ) ), intSize );
				numberSearchKeySize++;
				cout << "IndexManager:: createIndex() writeIndexBucketHeader() tbl.write( numberBucketIDs ): " << numberBucketIDs << endl << endl;

				// bucket ids for this search key ..
				//map<int, vector<int>* >::iterator it_vector;
				bucketIDs = it_map->second;

				for( int i=0; i < bucketIDs->size(); i++ ){

					bucketIDForSearchKey = bucketIDs->at(i);
					tbl.write( reinterpret_cast< const char * >( &( bucketIDForSearchKey ) ), intSize );
					bucketIDsForSearchKeySize++;
					cout << "IndexManager:: createIndex() writeIndexBucketHeader() tbl.write( bucketIDForSearchKey ): " << bucketIDForSearchKey << endl << endl;

				}

			} // end "for" ..

			//cout << "IndexManager:: createIndex() writeIndexBucketHeader() Point G .." << endl << endl;

			// "flag" indicating that "header entry" fields are beginning ..
			int startHeaderEntries = -1;
			tbl.write( reinterpret_cast< const char * >( &( startHeaderEntries ) ), intSize );
			cout << "IndexManager:: createIndex() writeIndexBucketHeader() tbl.write( startHeaderEntries ): " << startHeaderEntries << endl << endl;
*/

			//cout << "IndexManager:: createIndex() writeIndexBucketHeader() Point H .." << endl << endl;

			int bi = 0;

			int sk = 0;

			int fpc = 0;

			//bool pa = false;
			//vector<bool>* pointersAvailable = NULL;

			int numberOfHeaderEntryFields = 3;

			//int pointersAvailableSize = 0;

			int totalNumberOfBuckets = indexBucketHeader_->getTotalNumberBuckets();

			//cout << "IndexManager:: writeIndexBucketHeader() totalNumberOfBuckets: " << totalNumberOfBuckets << endl << endl;

			for(int i=0; i<totalNumberOfBuckets; i++){

				bi = indexBucketHeader_->getHeaderEntries()->at( i )->getBucketID();
				tbl.write( reinterpret_cast<const char*>( &bi ), intSize );
				//cout << "IndexManager:: writeIndexBucketHeader() tbl.write( BucketID ): " << bi << endl << endl;

				sk = indexBucketHeader_->getHeaderEntries()->at( i )->getSearchKey();
				tbl.write( reinterpret_cast<const char*>( &sk ), intSize );
				//cout << "IndexManager:: writeIndexBucketHeader() tbl.write( SearchKey ): " << sk << endl << endl;

				fpc = indexBucketHeader_->getHeaderEntries()->at( i )->getFreePointerCount();
				tbl.write( reinterpret_cast<const char*>( &fpc ), intSize );
				//cout << "IndexManager:: writeIndexBucketHeader() tbl.write( fpc ): " << fpc << endl << endl;

				//pointersAvailable = indexBucketHeader_.getHeaderEntries()->at( i )->getPointersAvailable();

				//for(int j=0; j<pointersAvailable->size(); j++) {

				//	pa = pointersAvailable->at( j );

				//	tbl.write( reinterpret_cast< const char * >( &( pa ) ), boolSize );
				//	pointersAvailableSize++;
				//	cout << "writeIndexBucketHeader() tbl.write( pa ): " << pa << endl << endl;

				//}

			}

			//cout << "IndexManager:: writeIndexBucketHeader() Point I .." << endl << endl;

			// 4 bytes ..
			// totalNumberOfBuckets  ..
			int headerFieldsSize = intSize;

			// [4 bytes * num fields * total num buckets]
			//int headerEntriesSize = ( ( intSize * numberOfHeaderEntryFields ) + ( intSize * pointersAvailableSize ) );
			int headerEntriesSize = ( intSize * numberOfHeaderEntryFields * totalNumberOfBuckets );

			// size in bytes of the header file ..
			int headerSize = ( headerFieldsSize + headerEntriesSize );

			//cout << "IndexManager:: writeIndexBucketHeader() headerSize: " << headerSize << endl << endl;


			int fillerSize = 0;

			int gaugeTheFillerSize = blockSize;


			// increase "gaugeTheFillerSize" by increments of "blockSize",
			// until "gaugeTheFillerSize >= headerSize" ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			while ( true ){

				if( gaugeTheFillerSize >= headerSize ){

					break;

				} else {

					gaugeTheFillerSize += blockSize;

				}

			} // end "while ( true )" ..


			//cout << "IndexManager:: writeIndexBucketHeader() Point J .." << endl << endl;

			// "fillerSize" should be less than 48 bytes,
			// such that adding "fillerSize" to "headerSize"
			// is an even multiple of "blockSize" ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			fillerSize = ( gaugeTheFillerSize - headerSize );

			//cout << "IndexManager:: writeIndexBucketHeader() fillerSize: " << fillerSize << endl << endl;


			if ( fillerSize > 0 ) {

				string filler = getStorageManager()->constructPagePadding( fillerSize );

				int charSize = 1;

				// write the filler string to the DataHeader file,
				// so that the header file size is an even multiple of "blockSize" ..
				for(int i=0; i<filler.size(); i++){

					tbl.write( reinterpret_cast<const char*>( &filler[i] ), charSize );

				}

			}

			//cout << "IndexManager:: writeIndexBucketHeader() Point K .." << endl << endl;

			// flag indicating the "writeHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			writeHeaderPerformed = true;


		} else {

	 		string errorString = "\nHeader file failed to open [ SM::writeIndexBucketHeader() ], so Header file was not written.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );

		} // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();


	} else {

	 	string errorString = "\nHeader file table name was not valid [ SM::writeIndexBucketHeader() ], so Header file was not written.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	//cout << "IndexManager:: writeIndexBucketHeader() Point L .." << endl << endl;

	return writeHeaderPerformed;


} // end "writeIndexBucketHeader()" ..


bool IndexManager:: writeIndexBucket( string tableName, string indexName, IndexBucket* indexBucket, int bucketPageID ) {

	// "tableName": name of the table associated with the index ..

	//    "indexBucket": indexBucket to write to file ..

	//cout << "IndexManager:: writeIndexBucket() start .." << endl << endl;
	//cout << "IndexManager:: writeIndexBucket() tableName: " << tableName << endl << endl;
	//cout << "IndexManager:: writeIndexBucket() indexBucket->getPageIDs()->size(): " << indexBucket->getPageIDs()->size() << endl << endl;
	//cout << "IndexManager:: writeIndexBucket() bucketPageID: " << bucketPageID << endl << endl;

	int maxBucketSize = ( blockSize / intSize );

	bool writeBucketPagePerformed = false;

	string emptyString = "";

	string tableNameString = tableName;
	
	//cout << "IndexManager:: writeIndexBucket() Point A .." << endl << endl;

	if ( emptyString.compare( tableNameString ) != 0 ) {

		// handle the directory path of the file ..

		//string directoryPlusFileName = "C:\\Temp\\";

		//cout << "IndexManager:: writeIndexBucket() Point A .." << endl << endl;
		
		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusFileName = databaseFile->getFileDrive();

		//cout << "IndexManager:: writeIndexBucket() Point B .." << endl << endl;

		directoryPlusFileName += databaseFile->getFileDirectory();

		// append the file
		directoryPlusFileName += "IndexBuckets_";
		directoryPlusFileName += tableName;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName;
		char* fileName = (char*)directoryPlusFileName.c_str();

		fstream tbl( fileName, ios::in | ios::out |ios::binary );

		//cout << "IndexManager:: writeIndexBucket() Point D .." << endl << endl;

		// this means the file doesn't exist yet,
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){

			tbl.open( fileName, ios::out | ios::binary | ios::app );

			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();

		}

		//cout << "IndexManager:: writeIndexBucket() Point E .." << endl << endl;

		if ( !tbl.fail() ) {

			//cout << "IndexManager:: writeIndexBucket() Point F .." << endl << endl;

			// make sure file pointer starts at beginning of file ..
			tbl.seekp( ios::beg );

			//cout << "IndexManager:: writeIndexBucket() Point G .." << endl << endl;

			// so locate page-to-write referenced from beginning of file ..
			tbl.seekp( bucketPageID * blockSize );

			//cout << "IndexManager:: writeIndexBucket() Point H .." << endl << endl;

			// number of values in the "bucket" ..
			int numberOfBucketValues = indexBucket->getPageIDs()->size();

			//cout << "IndexManager:: writeIndexBucket() numberOfBucketValues: " << numberOfBucketValues << endl << endl;

			//cout << "IndexManager:: writeIndexBucket() Point I .." << endl << endl;

			// size in bytes of the header file ..
			// [ (4 bytes) * (11 PageIDs) ] + [ (4 bytes) * (1 PageID) ] = 48 bytes
			// 1st 11 bytes are pointers to duplicate searchKeys,
			// last 4 bytes reserved for pointer to next bucket ..
			int bucketSize = ( intSize * numberOfBucketValues );

			int fillerSize = ( blockSize - bucketSize );

			//cout << "IndexManager:: writeIndexBucket() bucketSize: " << bucketSize << endl << endl;

			//cout << "IndexManager:: writeIndexBucket() fillerSize: " << fillerSize << endl << endl;

			// get the pageIDs from the bucket ..
			vector<int>* pageIDs = indexBucket->getPageIDs();

			//cout << "IndexManager:: writeIndexBucket() Point J .." << endl << endl;


			int fieldValue = -1;

			// write the pageIDs in the IndexBucket to file ..
			for (int i=0; i < numberOfBucketValues; i++) {

				// DEBUGGING ..
				//cout << "IndexManager:: writeIndexBucket() pageIDs["<<i<<"]: " << pageIDs->at(i) << endl << endl;

				fieldValue = pageIDs->at(i);

				//cout << "IndexManager:: writeIndexBucket() pageIDs->at(i("<<i<<")): " << pageIDs->at(i) << endl << endl;

				tbl.write( reinterpret_cast<const char*>( &fieldValue ), intSize );

			}


			if ( fillerSize > 0 ) {

				// DEBUGGING ..
				//cout << "IndexManager:: writeDataPage() blockSize: " << blockSize << endl << endl;
				//cout << "IndexManager:: writeDataPage() bucketSize: " << bucketSize << endl << endl;
				//cout << "IndexManager:: writeDataPage() fillerSize ( blockSize("<<blockSize<<") - bucketSize("<<bucketSize<<") ): " << fillerSize << endl << endl;

				string filler = getStorageManager()->constructPagePadding( fillerSize );

				int charSize = 1;

				// write the filler string values to fill up a 48-byte-block-DataPage to file ..
				for(int i=0; i<filler.size(); i++){

					//cout << filler[i];

					tbl.write( reinterpret_cast<const char*>( &filler[i] ), charSize );

				}

				//cout << endl << endl;

			}

			IndexBucketHeader* indexBucketHeader = new IndexBucketHeader();
			bool isReadIndexBucketHeader = readIndexBucketHeader( tableName, indexName, indexBucketHeader );

			//cout << "IndexManager:: writeIndexBucket() isReadIndexBucketHeader: " << isReadIndexBucketHeader << endl << endl;

			if( isReadIndexBucketHeader ){

				//cout << "IndexManager:: writeIndexBucket() Point K .." << endl << endl;

				vector<IndexBucketHeaderEntry*>* ibhes = indexBucketHeader->getHeaderEntries();

				int numberIBHES = ibhes->size();
				//cout << "IndexManager:: writeIndexBucket() numberIBHES: " << numberIBHES << endl << endl;
				for(int i=0; i<numberIBHES; i++) {

					if( ibhes->at(i)->getBucketID() == bucketPageID ){

						ibhes->at(i)->setFreePointerCount( ( maxBucketSize - numberOfBucketValues ) );

						//cout << "IndexManager:: writeIndexBucket() ( maxBucketSize - numberOfBucketValues ): " << ( maxBucketSize - numberOfBucketValues ) << endl << endl;

						break;

					}

				}

			} else {

				//cout << "IndexManager:: writeIndexBucket() Point L .." << endl << endl;

				IndexBucketHeaderEntry* ibhe = new IndexBucketHeaderEntry();

				//cout << "IndexManager:: writeIndexBucket() bucketPageID: " << bucketPageID << endl << endl;

				ibhe->setBucketID( bucketPageID );

				ibhe->setFreePointerCount( ( maxBucketSize - numberOfBucketValues ) );

				//cout << "IndexManager:: writeIndexBucket() ( maxBucketSize - numberOfBucketValues ): " << ( maxBucketSize - numberOfBucketValues ) << endl << endl;

				indexBucketHeader->getHeaderEntries()->push_back( ibhe );

				indexBucketHeader->setTotalNumberBuckets( 1 );

			}

			bool isWriteIndexBucketHeader = writeIndexBucketHeader( tableName, indexName, indexBucketHeader );

			//cout << "IndexManager:: writeIndexBucket() Point M .." << endl << endl;

			// flag indicating the "writeIndexBucket()" operation
			// was performed successfully (header read, file created, file opened) ..
			writeBucketPagePerformed = true;


		} else {

	 		string errorString = "\nFailed to open data file [ SM::writeIndexBucket() ], so index bucket was not written.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..


		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();


	} // end "if ( emptyString.compare( tableNameString ) != 0 )" ..


	//cout << "IndexManager:: writeIndexBucket() Point N .." << endl << endl;

	return writeBucketPagePerformed;


} // end "writeIndexBucket( char* tableName, IndexBucket* indexBucket, int pageID )" ..


bool IndexManager:: writeIndexBuckets( string indexTableName_, string indexName_, vector<IndexBucket*>* indexBuckets_ ){
//map<int, vector<int>* >* IndexManager:: writeIndexBuckets( vector<IndexBucket*>* indexBuckets_, char* indexTableName_, char* indexName_ ){

	int indexBucketSize = ( blockSize / intSize );	// typ 12 ..
	bool wasIndexBucketWritten = false;
	int indexBucketPosition = 0;
	int currentIndexBucketSize = -1;
	int numberPagesForThisBucket = -1;
	int shiftValueForMultiplePageBuckets = 0;
	int shiftValueForSkippedBuckets = 0;

	// bucket #, bucket-pageIDs ..
	//map<int, vector<int>* >* bucketPageIDsPerSearchKey = NULL;

	// bucket-pageIDs (pointers to bucket pages) ..
	vector<int>* bucketPageIDs = NULL;

	// data-pageIDs (pointers to data pages) from the bucket ..
	vector<int>* allCurrentBucketPageIDs = NULL;

	// increment through each "index bucket" ..
	for(int i=0; i<indexBuckets_->size(); i++){

		//if( bucketPageIDsPerSearchKey == NULL )
		//	bucketPageIDsPerSearchKey = new map<int, vector<int>* >();

		if( bucketPageIDs == NULL )
			bucketPageIDs = new vector<int>();

		allCurrentBucketPageIDs = indexBuckets_->at(i)->getPageIDs();

		// find out if the "index bucket" stored in memory needs to be
		// parcelled into multiple file pages ..
		currentIndexBucketSize = allCurrentBucketPageIDs->size();

		// DEBUGGING ..
		//cout << "=====================================================" << endl << endl;
		//cout << "IndexManager:: writeIndexBuckets()current bucket: " << (i+1) << endl << endl;
		//cout << "IndexManager:: writeIndexBuckets()currentIndexBucketSize[i("<<i<<")]: " << currentIndexBucketSize << endl << endl;

		// # pages to write = # times you can subtract "indexBucketSize"
		// from "currentIndexBucketSize" before "result" value < 0 ..
		int result = currentIndexBucketSize;
		int counter = 0;
		while( result > 0 ){

			result -= indexBucketSize;

			counter++;

		}

		// # pages to write == "counter" ..
		numberPagesForThisBucket = counter;

		// DEBUGGING ..
		//cout << "IndexManager:: writeIndexBuckets()numberPagesForThisBucket[i("<<i<<")]: " << numberPagesForThisBucket << endl << endl;
		//cout << "IndexManager:: writeIndexBuckets()if( currentIndexBucketSize("<<currentIndexBucketSize<<") > indexBucketSize("<<indexBucketSize<<") )[i("<<i<<")]: " << ( currentIndexBucketSize > indexBucketSize ) << endl << endl;
		//cout << "IndexManager:: writeIndexBuckets()if( ( 0 < currentIndexBucketSize("<<currentIndexBucketSize<<") )("<<( 0 < currentIndexBucketSize )<<") && ( currentIndexBucketSize("<<currentIndexBucketSize<<") <= indexBucketSize("<<indexBucketSize<<") )("<<( currentIndexBucketSize <= indexBucketSize )<<") )[i("<<i<<")]: " << ( ( 0 < currentIndexBucketSize ) && ( currentIndexBucketSize <= indexBucketSize ) ) << endl << endl;

		// this indicates that this bucket will be written to MULTIPLE pages ..
		if( currentIndexBucketSize > indexBucketSize ){

			// traverse the entire bucket ..
			// must write "bucket pages" in increments of "indexBucketSize" ..
			int bucketPageCounter = 1;

			IndexBucket* intermediateIndexBucket = new IndexBucket();

			while( indexBucketPosition < currentIndexBucketSize ){

				// DEBUGGING ..
				//cout << "--------------" << endl << endl;
				//cout << "IndexManager:: writeIndexBuckets()while( indexBucketPosition("<<indexBucketPosition<<") < currentIndexBucketSize("<<currentIndexBucketSize<<") )[i("<<i<<")]: " << ( indexBucketPosition < currentIndexBucketSize ) << endl << endl;
				//cout << "IndexManager:: writeIndexBuckets()Point A - if( indexBucketPosition("<<indexBucketPosition<<") == ( ( bucketPageCounter("<<bucketPageCounter<<") * indexBucketSize("<<indexBucketSize<<") )("<<( bucketPageCounter * indexBucketSize )<<") - 1 )("<<( ( bucketPageCounter * indexBucketSize ) - 1 )<<") )[i("<<i<<")]: " << ( indexBucketPosition == ( ( bucketPageCounter * indexBucketSize ) - 1 ) ) << endl << endl;
				//cout << "IndexManager:: writeIndexBuckets()Point A - && if( currentIndexBucketSize("<<currentIndexBucketSize<<") >= ( bucketPageCounter("<<bucketPageCounter<<") * indexBucketSize("<<indexBucketSize<<") )("<<( bucketPageCounter * indexBucketSize )<<") )[i("<<i<<")]: " << ( currentIndexBucketSize >= ( bucketPageCounter * indexBucketSize ) ) << endl << endl;
				//cout << "IndexManager:: writeIndexBuckets()Point B - else if( indexBucketPosition("<<indexBucketPosition<<") == ( currentIndexBucketSize("<<currentIndexBucketSize<<") - 1 )("<<( currentIndexBucketSize - 1 )<<") )[i("<<i<<")]: " << ( indexBucketPosition == ( currentIndexBucketSize - 1 ) ) << endl << endl;

				// allocate a new "intermediateIndexBucket" for the next page to be written ..
				// i.e.: when "indexBucketPosition" == 10, 21, 32, 43, etc ..
				// i.e.: 0 1 2 3 4 5 6 7 8 9 10 11			...		 	22			...		 	33  <- "indexBucketPosition"
				//	    | |			...		    | | <- "indexBucket"
				//								 0 1 2 3 4 5 6 7 8 9 10 11
				//	    					    | |			...		    | | <- "indexBucket"
				//								 						 0 1 2 3 4 5 6 7 8 9 10 11
				//	    					    						| |			...		    | | <- "indexBucket"

				// also, "currentIndexBucketSize" large enough that an additional page is needed for this bucket ..
				// i.e.: 0 1 2 3 4 5 6 7 8 9 10 11 12 ..  <- "indexBucketPosition", etc ..
				//	    | |			...		    | | <- "indexBucket", etc ..
				// [ where "( bucketPageCounter * indexBucketSize )" == 12, 24, 36, 48, etc .. ]
				if( ( indexBucketPosition == ( ( bucketPageCounter * ( indexBucketSize - 1 )  ) ) ) &&
					( currentIndexBucketSize >= ( ( bucketPageCounter * ( indexBucketSize - 1 ) ) + 1 ) ) ) {

					// DEBUGGING ..
					//cout << "--------------" << endl << endl;
					//cout << "IndexManager:: writeIndexBuckets()Point C[i("<<i<<")] .." << endl << endl;
					//cout << "IndexManager:: writeIndexBuckets()intermediateIndexBucket->getPageIDs()->push_back( ( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") + 1 ) )[i("<<i<<")]: " << ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets + 1 ) << endl << endl;

					// store the pointer to the next bucket page in the final 4b of this page ..
					intermediateIndexBucket->getPageIDs()->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets + 1 ) );

					// DEBUGGING ..
					//cout << "IndexManager:: writeIndexBuckets()writeIndexBucket(indexTableName("<<indexTableName_<<"), .. ,( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") )("<<( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets )<<"))[i("<<i<<")] .." << endl << endl;

					// writeIndexBucket( indexBucket );
					wasIndexBucketWritten = writeIndexBucket( indexTableName_, indexName_, intermediateIndexBucket, ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );

					//cout << "IndexManager:: writeIndexBuckets() Point D .." << endl << endl;

					// track this bucketPageID ..
					bucketPageIDs->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );

					//cout << "IndexManager:: writeIndexBuckets() Point E .." << endl << endl;

					// reset for new "intermediateIndexBucket" ..
					delete intermediateIndexBucket;
					intermediateIndexBucket = new IndexBucket();

					// DEBUGGING ..
					//cout << "IndexManager:: writeIndexBuckets() new IndexBucket()[i"<<i<<"] .." << endl << endl;

					bucketPageCounter++;

					shiftValueForMultiplePageBuckets++;		// adjustment to "i" b/c of duplicate buckets ..

					// DEBUGGING ..
					//cout << "IndexManager:: writeIndexBuckets() bucketPageCounter++[i("<<i<<")]: " << bucketPageCounter << endl << endl;
					//cout << "IndexManager:: writeIndexBuckets() shiftValueForMultiplePageBuckets++[i("<<i<<")]: " << shiftValueForMultiplePageBuckets << endl << endl;

				} else if( indexBucketPosition == ( currentIndexBucketSize - 1 ) ){

					// DEBUGGING ..
					//cout << "--------------" << endl << endl;
					//cout << "IndexManager:: writeIndexBuckets() Point F[i("<<i<<")] .." << endl << endl;
					//cout << "IndexManager:: writeIndexBuckets() intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition("<<indexBucketPosition<<") )("<<allCurrentBucketPageIDs->at( indexBucketPosition )<<") )[i("<<i<<")]: " << allCurrentBucketPageIDs->at( indexBucketPosition ) << endl << endl;

					// this is the last pageID on the last page for this bucket ..

					// store the pageIDs for this "intermediate index bucket" ..
					intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition ) );

					// DEBUGGING ..
					//cout << "IndexManager:: writeIndexBuckets() writeIndexBucket(indexTableName("<<indexTableName_<<"), .. ,( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") )("<<( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets )<<"))[i("<<i<<")] .." << endl << endl;

					// writeIndexBucket( indexBucket );
					wasIndexBucketWritten = writeIndexBucket( indexTableName_, indexName_, intermediateIndexBucket, ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );

					//cout << "IndexManager:: writeIndexBuckets() Point G .." << endl << endl;

					// track this bucketPageID ..
					bucketPageIDs->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );

					//cout << "IndexManager:: writeIndexBuckets() Point H .." << endl << endl;

					// reset for new "intermediateIndexBucket" ..
					delete intermediateIndexBucket;

					// DEBUGGING ..
					//cout << "IndexManager:: writeIndexBuckets() break out of while loop [i("<<i<<")] .." << endl << endl;
					//cout << "-----------------------------------------------------" << endl << endl;

					break;

				}

				// DEBUGGING ..
				//cout << "IndexManager:: writeIndexBuckets() Point B[i("<<i<<")] .." << endl << endl;
				//cout << "IndexManager:: writeIndexBuckets() intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition("<<indexBucketPosition<<") )[i("<<i<<")]: " << allCurrentBucketPageIDs->at( indexBucketPosition ) <<endl << endl;

				// store the pageIDs for this "intermediate index bucket" ..
				intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition ) );

				// traverse to next pageID in current bucket ..
				indexBucketPosition++;

				// DEBUGGING ..
				//cout << "IndexManager:: writeIndexBuckets() indexBucketPosition++[i("<<i<<")]: " << indexBucketPosition <<endl << endl;


			} // end "while()" ..

			// reset this parameter for the next loop through ..
			indexBucketPosition = 0;

		} else if( ( 0 < currentIndexBucketSize ) && ( currentIndexBucketSize <= indexBucketSize ) ){

			// this indicates that this bucket will be written to a SINGLE page ..

			// DEBUGGING ..
			//cout << "IndexManager:: writeIndexBuckets() Point D[i("<<i<<")] .." << endl << endl;
			//cout << "IndexManager:: writeIndexBuckets() writeIndexBucket(indexTableName("<<indexTableName_<<"), .. ,( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") )("<<( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets )<<"))[i("<<i<<")] .." << endl << endl;

			// writeIndexBucket( indexBucket );
			wasIndexBucketWritten = writeIndexBucket( indexTableName_, indexName_, indexBuckets_->at(i), ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );

			// track this bucketPageID ..
			bucketPageIDs->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );


		} else {

			// if the bucket is skipped (because it is empty),
			// adjust "i" so that this pageID isnt skipped for the
			// next "writeIndexBucket()" operation above ..
			shiftValueForSkippedBuckets++;

		} // end "if( currentIndexBucketSize > indexBucketSize )" ..


		// track the bucket-pageIDs for each unique search key ..
		//bucketPageIDsPerSearchKey->insert( pair<int, vector<int>* >( i, bucketPageIDs ) );


	} // end "for()" ..


	return wasIndexBucketWritten;
	//return bucketPageIDsPerSearchKey;


} // end "writeIndexBuckets()" ..


//bool IndexManager:: writeIndexPage( char* tableName_, char* indexName_, IndexPage* indexPage_, int pageID_, int parentPageID_, bool isNew_ ) {
bool IndexManager:: writeIndexPage( string tableName_, string indexName_, IndexPage* indexPage_, int pageID_, int parentPageID_, bool isNew_, bool isDuplicatesAllowed_ ) {

	// "tableName": name of the table associated with the index ..

	// "indexPage": indexPage to write to file ..

	//    "pageID": pageID associated with this indexPage ..

	//cout << "IndexManager:: writeIndexPage() start .." << endl << endl;
	//cout << "IndexManager:: writeIndexPage() tableName: " << tableName_ << endl << endl;
	//cout << "IndexManager:: writeIndexPage() indexPage_->getChildPageIDs()->size(): " << indexPage_->getChildPageIDs()->size() << endl << endl;
	//cout << "IndexManager:: writeIndexPage() indexPage_->getSearchKeys()->size(): " << indexPage_->getSearchKeys()->size() << endl << endl;

	bool writeIndexPagePerformed = false;

	string emptyString = "";

	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {

		// handle the directory path of the file ..

		//string directoryPlusFileName = "C:\\Temp\\";

		//cout << "IndexManager:: writeIndexPage() Point A .." << endl << endl;

		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		//cout << "IndexManager:: writeIndexPage() Point C .." << endl << endl;


		// append the file ..
		directoryPlusFileName += "Index_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();

		fstream tbl( fileName, ios::in | ios::out | ios::binary );

		//cout << "IndexManager:: writeIndexPage() Point D .." << endl << endl;

		// this means the file doesn't exist yet,
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){

			tbl.open( fileName, ios::out | ios::binary | ios::app );

			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();

		}

		//cout << "IndexManager:: writeIndexPage() Point E .." << endl << endl;

		if ( !tbl.fail() ) {

			//cout << "IndexManager:: writeIndexPage() Point F .." << endl << endl;

			// make sure file pointer starts at beginning of file ..
			tbl.seekp( ios::beg );

			//cout << "IndexManager:: writeIndexPage() Point G .." << endl << endl;

			// so locate page-to-write referenced from beginning of file ..
			tbl.seekp( pageID_ * blockSize );

			//cout << "IndexManager:: writeIndexPage() tbl.seekp( pageID_("<<pageID_<<") * blockSize("<<blockSize<<") ) .." << endl << endl;
			//cout << "IndexManager:: writeIndexPage() Point H .." << endl << endl;

			// get the pageIDs from the bucket ..
			vector<int>* childPageIDs = indexPage_->getChildPageIDs();
			vector<int>* searchKeys = indexPage_->getSearchKeys();
			bool isLeafNode = indexPage_->getIsLeafNode();

			// number of values in the "IndexPage" ..
			int numberOfChildPageIDValues = childPageIDs->size();
			int numberOfSearchKeyValues = searchKeys->size();
			int numberOfIndexPageValues = numberOfChildPageIDValues + numberOfSearchKeyValues;

			//cout << "IndexManager:: writeIndexPage() Point I .." << endl << endl;

			// size in bytes of the header file ..
			// [ (4 bytes) * (6 child page ids) ] + [ (4 bytes) * (5 search keys) ] = 24 + 20 = 44 bytes
			int indexPageSize = ( intSize * numberOfIndexPageValues ) + boolSize;

			int fillerSize = ( blockSize - indexPageSize );

			//cout << "IndexManager:: writeIndexPage() indexPageSize: " << indexPageSize << endl << endl;
			//cout << "IndexManager:: writeIndexPage() fillerSize: " << fillerSize << endl << endl;

			//cout << "IndexManager:: writeIndexPage() Point J .." << endl << endl;

			int fieldValue = -1;

			//cout << "IndexManager:: writeIndexPage() numberOfChildPageIDValues: " << numberOfChildPageIDValues << endl << endl;

			// write the childPageID / searchKey pairs in the IndexPage to file ..
			for (int i=0; i < numberOfChildPageIDValues; i++) {

				// "childPageID" is only written when "i" is "even" ..

				fieldValue = childPageIDs->at(i);

				tbl.write( reinterpret_cast<const char*>( &fieldValue ), intSize );

				// DEBUGGING ..
				//cout << "IndexManager:: writeIndexPage() childPageIDs->at(i("<<i<<")): " << fieldValue << endl << endl;


			}

			//cout << "IndexManager:: writeIndexPage() numberOfSearchKeyValues: " << numberOfSearchKeyValues << endl << endl;

			// write the childPageID / searchKey pairs in the IndexPage to file ..
			for (int i=0; i < numberOfSearchKeyValues; i++) {

				// "pageIDs" (#1 ptr) - (#5 ptr) ..

				fieldValue = searchKeys->at(i);

				tbl.write( reinterpret_cast<const char*>( &fieldValue ), intSize );

				// DEBUGGING ..
				//cout << "IndexManager:: writeIndexPage() searchKeys->at(i("<<i<<")): " << fieldValue << endl << endl;

			}

			// write whether or not this is a "leaf node" ..
			tbl.write( reinterpret_cast<const char*>( &isLeafNode ), boolSize );

			//cout << "IndexManager:: writeIndexPage() isLeafNode: " << isLeafNode << endl << endl;

			// write the final (#6 ptr) "childPageID" to file ..
			//fieldValue = childPageIDs->back();
			//tbl.write( reinterpret_cast< const char * >( &fieldValue ), intSize );


			if ( fillerSize > 0 ) {

				// DEBUGGING ..
				//cout << "IndexManager:: writeDataPage() blockSize: " << blockSize << endl << endl;
				//cout << "IndexManager:: writeDataPage() bucketSize: " << bucketSize << endl << endl;
				//cout << "IndexManager:: writeDataPage() fillerSize ( blockSize("<<blockSize<<") - bucketSize("<<bucketSize<<") ): " << fillerSize << endl << endl;

				string filler = getStorageManager()->constructPagePadding( fillerSize );

				int fillerSize = filler.length();

				int charSize = 1;

				// write the filler string values to fill up a 48-byte-block-DataPage to file ..
				for(int i=0; i<fillerSize; i++){

					//cout << filler[i];

					tbl.write( reinterpret_cast<const char*>( &( filler[i] ) ), charSize );

				}

				//cout << endl << endl;

			}

			// flag indicating the "writeIndexPage()" operation
			// was performed successfully (header read, file created, file opened) ..
			writeIndexPagePerformed = true;

			// update the "index header" ..
			IndexHeader* header = new IndexHeader();
			bool isIndexHeaderRead = readIndexHeader( tableName_, indexName_, header );

			int totalNumberOfIndexPages = 0;

			// if this is a new "indexPage" tracked in the header,
			// update info accordingly below ..
			if( isNew_ == true ) {

				if( isIndexHeaderRead == true )
					totalNumberOfIndexPages = header->getTotalNumberOfIndexPages();
				else
					totalNumberOfIndexPages = 0;

				header->setTotalNumberOfIndexPages( ( totalNumberOfIndexPages + 1 ) );

			}

			// track whether IndexBuckets are used (isDuplicatesAllowed_ == true) ..
			if( isDuplicatesAllowed_ == true || isDuplicatesAllowed_ == false ){

				header->setIsDuplicatesAllowed( isDuplicatesAllowed_ );

			} else {

				// default ..
				header->setIsDuplicatesAllowed( false );

			}

			//cout << "IndexManager writeIndexPage() header->getTotalNumberOfIndexPages(): " << header->getTotalNumberOfIndexPages() << endl << endl;

			// check to see if the header entry exists already ..
			vector<IndexHeaderEntry*>* indexHeaderEntries = header->getHeaderEntries();
			int indexHeaderEntriesSize = indexHeaderEntries->size();
			bool isHeaderEntryFound = false;
			int headerEntryIndex = -1;
			for(int i=0; i<indexHeaderEntriesSize; i++){

				//cout << "IndexManager writeIndexPage() header->getHeaderEntries()->at(i("<<i<<")->getPageID()("<<header->getHeaderEntries()->at(i)->getPageID()<<") == pageID_("<<pageID_<<"): " << ( header->getHeaderEntries()->at(i)->getPageID() == pageID_ ) << endl << endl;

				if( ( indexHeaderEntries->at(i)->getPageID() == pageID_ )&&
				    ( isHeaderEntryFound == false ) ){

					isHeaderEntryFound = true;
					headerEntryIndex = i;
					//break;

				}

			}

			int freeSearchKeyCount = ( (fanout - 1) - numberOfSearchKeyValues );


			//cout << "IndexPages:: writeIndexPage() Point A .." << endl << endl;

			vector<bool>* searchKeysAvailable = new vector<bool>();
			for(int i=0; i<numberOfSearchKeyValues; i++)
				searchKeysAvailable->push_back( false );

			for(int i=0; i<freeSearchKeyCount; i++)
				searchKeysAvailable->push_back( true );

			//cout << "IndexManager:: createIndex() writeIndexPage() searchKeysAvailable->size(): " << searchKeysAvailable->size() << endl << endl;

			if( isHeaderEntryFound == true ){

				//cout << "IndexManager:: writeIndexPage() Point A .." << endl << endl;

				// this means that the entry already exists, so just update it with the latest values ..

				header->getHeaderEntries()->at( headerEntryIndex )->setNumberOfSearchKeys( numberOfSearchKeyValues );
				//cout << "IndexManager:: writeIndexPage() numberOfSearchKeyValues: " << numberOfSearchKeyValues << endl << endl;
				header->getHeaderEntries()->at( headerEntryIndex )->setNumberOfChildPageIDs( numberOfChildPageIDValues );
				//cout << "IndexManager:: writeIndexPage() numberOfChildPageIDValues: " << numberOfChildPageIDValues << endl << endl;
				header->getHeaderEntries()->at( headerEntryIndex )->setParentPageID( parentPageID_ );
				//cout << "IndexManager:: writeIndexPage() parentPageID_: " << parentPageID_ << endl << endl;
				header->getHeaderEntries()->at( headerEntryIndex )->setFreeSearchKeyCount( freeSearchKeyCount );
				//cout << "IndexManager:: writeIndexPage() freeSearchKeyCount: " << freeSearchKeyCount << endl << endl;
				header->getHeaderEntries()->at( headerEntryIndex )->setSearchKeysAvailable( searchKeysAvailable );

			} else {

				//cout << "IndexManager:: writeIndexPage() Point B .." << endl << endl;

				// this means that the entry is new, so need to add a new entry ..

				IndexHeaderEntry* ihe = new IndexHeaderEntry();

				ihe->setPageID( pageID_ );
				//cout << "IndexManager:: writeIndexPage() pageID_: " << pageID_ << endl << endl;
				ihe->setNumberOfSearchKeys( numberOfSearchKeyValues );
				//cout << "IndexManager:: writeIndexPage() numberOfSearchKeyValues: " << numberOfSearchKeyValues << endl << endl;
				ihe->setNumberOfChildPageIDs( numberOfChildPageIDValues );
				//cout << "IndexManager:: writeIndexPage() numberOfChildPageIDValues: " << numberOfChildPageIDValues << endl << endl;
				ihe->setParentPageID( parentPageID_ );
				//cout << "IndexManager:: writeIndexPage() parentPageID_: " << parentPageID_ << endl << endl;
				//int freeSearchKeyCount = ( (fanout - 1) - numberOfSearchKeyValues );
				ihe->setFreeSearchKeyCount( freeSearchKeyCount );
				//cout << "IndexManager:: writeIndexPage() freeSearchKeyCount: " << freeSearchKeyCount << endl << endl;

				ihe->setSearchKeysAvailable( searchKeysAvailable );

				header->getHeaderEntries()->push_back( ihe );

			}

			//cout << "IndexManager:: writeIndexPage() Point C .." << endl << endl;
			
			// these 2 values should be the same ..
			totalNumberOfIndexPages = header->getTotalNumberOfIndexPages();
			int headerEntriesSize = header->getHeaderEntries()->size();
			
			if( totalNumberOfIndexPages != headerEntriesSize ){
				
					header->setTotalNumberOfIndexPages( headerEntriesSize );
				
			}

			bool isWriteIndexHeader = writeIndexHeader( tableName_, indexName_, header );

			//cout << "IndexManager:: writeIndexPage() Point D .." << endl << endl;

		} else {

	 		string errorString = "\nFailed to open data file [ SM::writeIndexPage() ], so index header was not written.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..


		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();


	} // end "if ( emptyString.compare( tableNameString ) != 0 )" ..


	//cout << "SM:: writeIndexPage() end .." << endl;

	return writeIndexPagePerformed;


} // end "writeIndexPage()" ..


//bool IndexManager:: writeIndexPages( char* indexTableName_, char* indexName_, vector<IndexPage*>* indexPages_ , vector<bool>* isNew_ ){
bool IndexManager:: writeIndexPages( string indexTableName_, string indexName_, vector<IndexPage*>* indexPages_ , vector<bool>* isNew_, bool isDuplicatesAllowed_ ){

	//cout << "IndexManager:: writeIndexPages() start .." << endl << endl;

	bool wasIndexPageWritten = false;
	bool wereIndexPagesWritten = true;

	int parentID = -1;


	// only proceed if "index pages" is non-null ..
	if( indexPages_ != NULL ) {

		// loop max values ..
		int indexPagesSize = indexPages_->size();
		int headerEntriesSize = -1;

		IndexHeader* header = new IndexHeader();
		bool isReadIndexHeader = readIndexHeader( indexTableName_, indexName_, header );

		//cout << "IndexManager:: writeIndexPages() header.getTotalNumberOfIndexPages(): " << header->getTotalNumberOfIndexPages() << endl << endl;

		vector<IndexHeaderEntry*>* headerEntries = header->getHeaderEntries();

		// write each "index page" ..
		for(int i=0; i<indexPagesSize; i++){

			// loop max value ..
			headerEntriesSize = headerEntries->size();

			// get the "parentID" for this "pageID" ..
			for(int j=0; j<headerEntries->size(); j++){

				if( headerEntries->at(j)->getPageID() == i ){

					parentID = headerEntries->at(j)->getParentPageID();
					break;

				}

			}

			// write this index page to file ..
			//wasIndexPageWritten = writeIndexPage( indexTableName_, indexName_, indexPages_->at(i), indexPages_->at(i)->getPageID(), parentID, isNew_->at(i) );
			wasIndexPageWritten = writeIndexPage( indexTableName_, indexName_, indexPages_->at(i), indexPages_->at(i)->getPageID(), parentID, isNew_->at(i), isDuplicatesAllowed_ );

			// update parameter ..
			if( wasIndexPageWritten == false )
				wereIndexPagesWritten = false;


		} // end "for()" ..

		//cout << "IndexManager:: writeIndexPages() Point A .." << endl << endl;

		// this is done inside each call to "writeIndexPage()" ..
		// update the "TotalNumberOfIndexPages" field (if necessary) ..
		//if( isReadIndexHeader == true )
			//header.setTotalNumberOfIndexPages( indexPages_->size() );

		//bool isWriteIndexHeader = writeIndexHeader( indexTableName_, indexName_, header );


		if( isReadIndexHeader == false )
			wereIndexPagesWritten = false;


	} else {

		// "indexPages" was null, so return false ..
		wereIndexPagesWritten = false;


	} // end "if()" ..

	//cout << "IndexManager:: writeIndexPages() Point B .." << endl << endl;

	return wereIndexPagesWritten;


} // end "writeIndexPages()" ..


bool IndexManager:: readIndexHeader( string tableName, string indexName_, IndexHeader* header ) {

	//cout << "IndexManager:: readIndexHeader() start .." << endl << endl;
	//cout << "IndexManager:: readIndexHeader() tableName: " << tableName << endl << endl;
	//cout << "IndexManager:: readIndexHeader() indexName: " << indexName_ << endl << endl;

	bool readHeaderPerformed = false;

	string emptyString = "";

	string tableNameString = tableName;

	//cout << "IndexManager:: readIndexHeader() Point A .." << endl << endl;

	// DEBUGGING ..
	//cout << "IndexManager:: readHeader() tableName: " << tableName << endl;

	if ( emptyString.compare( tableNameString ) != 0 ) {

		//cout << "IndexManager:: readIndexHeader() Point B .." << endl << endl;

		//string directoryPlusTableName = "C:\\Temp\\";

		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		directoryPlusFileName += "IndexHeader_";
		directoryPlusFileName += tableName;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();

		//cout << "IndexManager:: readIndexHeader() Point C .." << endl << endl;


		fstream tbl( fileName, ios::in | ios::binary );

		//cout << "IndexManager:: readIndexHeader() Point D .." << endl << endl;


		if ( !tbl.fail() ) {

			tbl.seekp( ios::beg );

			int rootPageID = -1;

			int totalNumberOfIndexPages = -1;
			
			bool isDuplicatesAllowed = false;

			tbl.read( reinterpret_cast<char*>( &rootPageID ), intSize );
			header->setRootPageID( rootPageID );

			tbl.read( reinterpret_cast<char*>( &totalNumberOfIndexPages ), intSize );
			header->setTotalNumberOfIndexPages( totalNumberOfIndexPages );
			
			tbl.read( reinterpret_cast<char*>( &isDuplicatesAllowed ), boolSize );
			header->setIsDuplicatesAllowed( isDuplicatesAllowed );


			//cout << "IndexManager:: readIndexHeader() Point E .." << endl << endl;

			// DEBUGGING ..
			//cout << "IndexManager:: readHeader() header.totalNumberOfDataPages: " << header.totalNumberOfDataPages << endl;

			int pid = 0;

			int ppid = 0;

			int nosk = 0;

			int nocpi = 0;

			int fskc = 0;

			bool ska = false;

			int numberOfSearchKeys = 4;


			header->getHeaderEntries()->clear();

			//cout << "IndexManager:: readIndexHeader() Point F .." << endl << endl;

			for (int i=0; i<totalNumberOfIndexPages; i++){

				//cout << "IndexManager:: readIndexHeader() Point G .." << endl << endl;

				header->getHeaderEntries()->push_back( ( new IndexHeaderEntry() ) );

				//cout << "IndexManager:: readIndexHeader() Point H .." << endl << endl;

				tbl.read( reinterpret_cast<char*>( &pid ), intSize );
				header->getHeaderEntries()->at( i )->setPageID( pid );
				//cout << "IndexManager:: readIndexHeader() pid: " << pid << endl << endl;

				//cout << "IndexManager:: readIndexHeader() Point I .." << endl << endl;

				tbl.read( reinterpret_cast<char*>( &nosk ), intSize );
				header->getHeaderEntries()->at( i )->setNumberOfSearchKeys( nosk );
				//cout << "IndexManager:: readIndexHeader() nosk: " << nosk << endl << endl;

				//cout << "IndexManager:: readIndexHeader() Point J .." << endl << endl;

				tbl.read( reinterpret_cast<char*>( &nocpi ), intSize );
				header->getHeaderEntries()->at( i )->setNumberOfChildPageIDs( nocpi );
				//cout << "IndexManager:: readIndexHeader() nocpi: " << nocpi << endl << endl;

				//cout << "IndexManager:: readIndexHeader() Point K .." << endl << endl;

				tbl.read( reinterpret_cast<char*>( &ppid ), intSize );
				header->getHeaderEntries()->at( i )->setParentPageID( ppid );
				//cout << "IndexManager:: readIndexHeader() ppid: " << ppid << endl << endl;

				//cout << "IndexManager:: readIndexHeader() Point L .." << endl << endl;

				tbl.read( reinterpret_cast<char*>( &fskc ), intSize );
				header->getHeaderEntries()->at( i )->setFreeSearchKeyCount( fskc );
				//cout << "IndexManager:: readIndexHeader() fskc: " << fskc << endl << endl;

				//cout << "IndexManager:: readIndexHeader() Point M .." << endl << endl;

				//cout << "IndexManager:: readIndexHeader() numberOfSearchKeys: " << numberOfSearchKeys << endl << endl;
				//cout << "IndexManager:: readIndexHeader() nosk: " << nosk << endl << endl;

				// need to adjust for overflow index node ..
				if( nosk > numberOfSearchKeys ){
					numberOfSearchKeys = nosk;
					header->getHeaderEntries()->at( i )->getSearchKeysAvailable()->push_back(-1);
				}

				//cout << "IndexManager:: readIndexHeader() header.getHeaderEntries()->at( i )->getSearchKeysAvailable()->size(): " << header.getHeaderEntries()->at( i )->getSearchKeysAvailable()->size() << endl << endl;

				for(int j=0; j<numberOfSearchKeys; j++) {

					//cout << "IndexManager:: readIndexHeader() Point N1 .." << endl << endl;
					tbl.read( reinterpret_cast<char*>( &ska ), boolSize );
					//cout << "IndexManager:: readIndexHeader() Point N2 .." << endl << endl;
					header->getHeaderEntries()->at( i )->getSearchKeysAvailable()->at( j ) = ska;
					//cout << "IndexManager:: readIndexHeader() Point N3 .. ska: " << ska << endl << endl;

					//cout << "IndexManager:: readIndexHeader() Point N .." << endl << endl;

				}

				// reset for next iteration ..
				numberOfSearchKeys = 4;

			}

			//cout << "IndexManager:: readIndexHeader() Point O .." << endl << endl;

			// flag indicating the "readHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			readHeaderPerformed = true;

		} else {

			// workaround: for the initial creation of an index,
			// IndexHeader wont exist, but we dont want an error
			// message, so comment this out and handle in code
			// that invokes this method ..
	 		//string errorString = "\nHeader file failed to open [ SM::readIndexHeader() ], so Header file was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringSM( errorString );

			//cout << "IndexManager:: readIndexHeader() failed to read header .." << endl << endl;

		} // end "if ( !tbl.fail() )" ..

		//cout << "IndexManager:: readIndexHeader() Point P .." << endl << endl;


		tbl.clear();
		tbl.close();


	} else {

	 	string errorString = "\nHeader file table name was not valid [ SM::readIndexHeader() ], so Header file was not read.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	//cout << "IndexManager:: readIndexHeader() Point Q .." << endl << endl;


	if ( !readHeaderPerformed ) {

		//cout << "IndexManager:: readIndexHeader() Point R .." << endl << endl;

		// "IndexHeader" was not found, b/c the file was not opened and/or found, so
		// set the values of the "hp" to default values, and
		// indicate that the "IndexHeader" was _not_ found ..

		header->setTotalNumberOfIndexPages( 0 );

		header->getHeaderEntries()->clear();

	}

	//cout << "IndexManager:: readIndexHeader() Point S .." << endl << endl;


	// DEBUGGING ..
	//cout << "SM:: readIndexHeader() headerPageFound: " << headerPageFound << endl;
	//cout << "SM:: readIndexHeader() end .." << endl;


	return readHeaderPerformed;


} // end "readIndexHeader( char* tableName, DataHeader& header )" ..


bool IndexManager:: readIndexBucketHeader( string tableName_, string indexName_, IndexBucketHeader* indexBucketHeader_ ) {

	//cout << "IndexManager:: readIndexBucketHeader() start .." << endl << endl;

	bool readHeaderPerformed = false;

	string emptyString = "";

	string tableNameString = tableName_;

	// DEBUGGING ..
	//cout << "IndexManager:: readHeader() tableName: " << tableName << endl;

	if ( emptyString.compare( tableNameString ) != 0 ) {

		//string directoryPlusTableName = "C:\\Temp\\";

		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		directoryPlusFileName += "IndexBucketHeader_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();


		fstream tbl( fileName, ios::in | ios::binary );


		if ( !tbl.fail() ) {

			tbl.seekp( ios::beg );

			int totalNumberOfBuckets = -1;

			tbl.read( reinterpret_cast<char*>( &totalNumberOfBuckets ), intSize );

			indexBucketHeader_->setTotalNumberBuckets( totalNumberOfBuckets );


			// DEBUGGING ..
			//cout << "IndexManager:: readHeader() header.totalNumberOfDataPages: " << header.totalNumberOfDataPages << endl;
			//cout << "IndexManager:: readHeader() header.filePageFieldCount: " << header.filePageFieldCount << endl;

/*
			// "map<int, vector<int>* >* _indexBucketIDsForEachSearchKey"
			map<int, vector<int>* >* iBIDsFESK = new map<int, vector<int>* >();
			int sk = 0;						// search key
			int numberBucketIDs;			// # bucketIDs in vector<int>*
			vector<int>* bucketIDs = NULL;

			int nextToken = 0;		// "-1" is a flag indicating that "bucketID" is next ..

			int fieldNumber = 0;

			while( nextToken >= 0 ){

				if( fieldNumber == 0 ){

					tbl.read( reinterpret_cast< char * > ( &( nextToken ) ), intSize );

					if( nextToken >= 0 )
						sk = nextToken;
					else
						break;

					fieldNumber = 1;

				} else if( fieldNumber == 1 ){

					tbl.read( reinterpret_cast< char * > ( &( nextToken ) ), intSize );

					if( nextToken >= 0 )
						numberBucketIDs = nextToken;
					else
						break;

					fieldNumber == 2;

				} else {

					// collect all of the "bucketIDs" for this searck key ..
					for(int i=0; i<numberBucketIDs; i++){

						tbl.read( reinterpret_cast< char * > ( &( nextToken ) ), intSize );

					if( nextToken >= 0 )
						bucketIDs->push_back( nextToken );
					else
						break;

					}

					if( bucketIDs->size() > 0 ) {

						iBIDsFESK->insert( pair<int, vector<int>* >( sk, bucketIDs ) );

						indexBucketHeader_.setBucketIDsForEachSearchKey( iBIDsFESK );

					}

					fieldNumber = 0;

				}

			}
*/

			int bi = 0;

			int sk = 0;

			int fpc = 0;

			//bool pa = false;
			//int bucketSize = 12;

			indexBucketHeader_->getHeaderEntries()->clear();

			for (int i=0; i<indexBucketHeader_->getTotalNumberBuckets(); i++){

				indexBucketHeader_->getHeaderEntries()->push_back( ( new IndexBucketHeaderEntry() ) );

				tbl.read( reinterpret_cast<char*>( &bi ), intSize );
				indexBucketHeader_->getHeaderEntries()->at( i )->setBucketID( bi );

				tbl.read( reinterpret_cast<char*>( &sk ), intSize );
				indexBucketHeader_->getHeaderEntries()->at( i )->setSearchKey( sk );

				tbl.read( reinterpret_cast<char*>( &fpc ), intSize );
				indexBucketHeader_->getHeaderEntries()->at( i )->setFreePointerCount( fpc );

				//for(int j=0; j<bucketSize; j++) {

				//	tbl.read( reinterpret_cast< char * > ( &( pa ) ), boolSize );
				//	indexBucketHeader_.getHeaderEntries()->at( i )->getPointersAvailable()->at( j ) = pa;

				//}

			}

			// flag indicating the "readHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			readHeaderPerformed = true;

		} else {

			// workaround: for the initial creation of an index,
			// IndexHeader wont exist, but we dont want an error
			// message, so comment this out and handle in code
			// that invokes this method ..
	 		//string errorString = "\nHeader file failed to open [ SM::readIndexBucketHeader() ], so Header file was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();


	} else {

	 	string errorString = "\nHeader file table name was not valid [ SM::readIndexBucketHeader() ], so Header file was not read.";

	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringSM( errorString );

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..


	if ( !readHeaderPerformed ) {

		// "IndexBucketHeader" was not found, b/c the file was not opened and/or found, so
		// set the values of the "hp" to default values, and
		// indicate that the "IndexBucketHeader" was _not_ found ..

		indexBucketHeader_->setTotalNumberBuckets( 0 );

		indexBucketHeader_->getHeaderEntries()->clear();

	}


	// DEBUGGING ..
	//cout << "SM:: readIndexHeader() headerPageFound: " << headerPageFound << endl;
	//cout << "SM:: readIndexHeader() end .." << endl;


	return readHeaderPerformed;


} // end "readIndexBucketHeader()" ..

bool IndexManager:: readIndexBucket( string tableName_, string indexName_, IndexBucket* indexBucket_, int bucketPageID_ ) {

	// "tableName": name of the table associated with the index ..

	//    "indexBucket": indexBucket to write to file ..

	//cout << "IndexManager:: readIndexBucket() start .." << endl << endl;
	//cout << "IndexManager:: readIndexBucket() tableName: " << tableName_ << endl << endl;
	////cout << "IndexManager:: readIndexBucket() indexBucket->getPageIDs()->size(): " << indexBucket->getPageIDs()->size() << endl << endl;
	//cout << "IndexManager:: readIndexBucket() bucketPageID: " << bucketPageID_ << endl << endl;

	bool readBucketPagePerformed = false;

	string emptyString = "";

	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {

		IndexBucketHeader* header = new IndexBucketHeader();
		bool readHeaderPerformed = readIndexBucketHeader( tableName_, indexName_, header );

		// DEBUGGING ..
		//cout << "IndexManager:: readDataPage() readHeaderPerformed: " << readHeaderPerformed << endl << endl;

		// if reading header failed, can't read page ..
		if( readHeaderPerformed == false ) {

	 		string errorString = "\nFailed to read Header file [ SM::readIndexBucket() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

			return false;

		}


		// handle the directory path of the file ..

		//string directoryPlusFileName = "C:\\Temp\\";

		//cout << "IndexManager:: readIndexBucket() Point A .." << endl << endl;

		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusFileName = databaseFile->getFileDrive();

		directoryPlusFileName += databaseFile->getFileDirectory();

		//cout << "IndexManager:: readIndexBucket() Point C .." << endl << endl;


		// append the file
		directoryPlusFileName += "IndexBuckets_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();

		//cout << "IndexManager:: readIndexBucket() Point D .." << endl << endl;

		fstream tbl( fileName, ios::in | ios::binary );

		//cout << "IndexManager:: readIndexBucket() Point E .." << endl << endl;

		if ( !tbl.fail() ) {

			//cout << "IndexManager:: readIndexBucket() Point F .." << endl << endl;

			// make sure file pointer starts at beginning of file ..
			tbl.seekg( ios::beg );

			//cout << "IndexManager:: readIndexBucket() Point G .." << endl << endl;

			// so locate page-to-write referenced from beginning of file ..
			tbl.seekg( bucketPageID_ * blockSize );

			//cout << "IndexManager:: readIndexBucket() Point H .." << endl << endl;


			// number of values in the "bucket" ..
			int maxNumberOfBucketIDs = ( blockSize / intSize );

			//int totalNumberOfBuckets = header.getTotalNumberBuckets();

			int freePointerCount = header->getHeaderEntries()->at( bucketPageID_ )->getFreePointerCount();
			int numberOfBucketIDs = (maxNumberOfBucketIDs - freePointerCount);

			/*
			int numberOfBucketValues = -1;
			//vector<bool>* pointersAvailable = NULL;
			for(int i=0; i<totalNumberOfBuckets; i++){
				if ( bucketPageID_ == header.getHeaderEntries()->at( i )->getBucketID() ) {
					numberOfBucketValues = header.getHeaderEntries()->at( i )->getFreePointerCount();
					break;
				}
			}
			*/

			//cout << "IndexManager:: readIndexBucket() Point I .." << endl << endl;


			int fieldValue = -1;

			// write the pageIDs in the IndexBucket to file ..
			//for (int i=0; i < pointersAvailable->size(); i++) {
			//for (int i=0; i < numberOfBucketValues; i++) {
			for (int i=0; i < numberOfBucketIDs; i++) {

				// DEBUGGING ..
				//cout << "IndexManager:: readIndexBucket() pageIDs["<<i<<"]: " << pageIDs->at(i) << endl << endl;

				//if( pointersAvailable->at(i) == false ) {

				tbl.read( reinterpret_cast<char *>( &fieldValue ), intSize );
				indexBucket_->getPageIDs()->push_back( fieldValue );

				//}

			}

			// flag indicating the "writeIndexBucket()" operation
			// was performed successfully (header read, file created, file opened) ..
			readBucketPagePerformed = true;


		} else {

	 		string errorString = "\nFailed to open data file [ SM::readIndexBucket() ], so index bucket was not written to.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		} // end "if ( !tbl.fail() )" ..


		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();


	} // end "if ( emptyString.compare( tableNameString ) != 0 )" ..


	//cout << "SM:: readIndexBucket() end .." << endl;

	return readBucketPagePerformed;


} // end "readIndexBucket()" ..

bool IndexManager:: readIndexPage( string tableName_, string indexName_, IndexPage* indexPage_, int pageID_ ) {


	// "tableName": name of the table associated with the index ..

	// "indexPage": indexPage to read from file ..

	//    "pageID": pageID associated with this indexPage ..

	//cout << "IndexManager:: readIndexPage() start .." << endl << endl;
	//cout << "IndexManager:: readIndexPage() tableName: " << tableName_ << endl << endl;
	//cout << "IndexManager:: readIndexPage() pageID: " << pageID_ << endl << endl;


	bool readPagePerformed = false;

	string emptyString = "";

	string tableNameString = tableName_;


	if ( emptyString.compare( tableNameString ) != 0 ) {


		IndexHeader* header = new IndexHeader();
		bool readHeaderPerformed = readIndexHeader( tableName_, indexName_, header );

		// DEBUGGING ..
		//cout << "IndexManager:: readDataPage() readHeaderPerformed: " << readHeaderPerformed << endl << endl;

		// if reading header failed, can't read page ..
		if( readHeaderPerformed == false ) {

	 		string errorString = "\nFailed to read Header file [ SM::readIndexPage() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

			return false;

		}


		//string directoryPlusTableName = "C:\\Temp\\";

		DatabaseFile* databaseFile = getStorageManager()->getDatabaseFile();

		string directoryPlusTableName = databaseFile->getFileDrive();

		directoryPlusTableName += databaseFile->getFileDirectory();

		directoryPlusTableName += "Index_";
		directoryPlusTableName += tableName_;
		directoryPlusTableName += "_";
		directoryPlusTableName += indexName_;
		char* fileName = (char*)directoryPlusTableName.c_str();

		fstream tbl( fileName,  ios::in | ios::binary );

		//int headerEntryIndex = -1;


		if ( !tbl.fail() ) {


			// DataHeader is written into its own file (by itself)
			// so locate "page-to-read", referenced from beginning of file ..
			tbl.seekg( ( pageID_ * blockSize ) );

			//cout << "IndexManager:: readIndexPage() tbl.seekg( ( pageID_("<<pageID_<<") * blockSize("<<blockSize<<") ) ) .." << endl << endl;


			int numberOfChildPageIDValues = 0;
			int numberOfSearchKeyValues = 0;

			int fieldValue = -1;
			
			int totalNumberOfIndexPages = header->getTotalNumberOfIndexPages();

			//cout << "IndexManager:: readIndexPage() header->getTotalNumberOfIndexPages(): " << header->getTotalNumberOfIndexPages() << endl << endl;

			for(int i=0; i<totalNumberOfIndexPages; i++){

				// DEBUGGING ..
				//cout << "IndexManager:: readIndexPage() pageID("<<pageID_<<") == header->getHeaderEntries()->at( i("<<i<<"]) )->getPageID()("<<header.getHeaderEntries()->at( i )->getPageID()<<"): " << ( ( pageID_ == header.getHeaderEntries()->at( i )->getPageID() ) ) << endl << endl;

				if ( pageID_ == header->getHeaderEntries()->at( i )->getPageID() ){

					numberOfChildPageIDValues = header->getHeaderEntries()->at( i )->getNumberOfChildPageIDs();

					numberOfSearchKeyValues = header->getHeaderEntries()->at( i )->getNumberOfSearchKeys();

					break;

				}

			}

			// DEBUGGING ..
			//for(int i=0; i<header.getTotalNumberOfIndexPages(); i++)
			//	cout << "IndexManager:: readIndexPage() pageID("<<pageID_<<") == header->getHeaderEntries()->at( i("<<i<<"]) )->getPageID()("<<header.getHeaderEntries()->at( i )->getPageID()<<"): " << ( ( pageID_ == header.getHeaderEntries()->at( i )->getPageID() ) ) << endl << endl;


			// DEBUGGING ..
			//cout << "IndexManager:: readIndexPage() numberOfChildPageIDValues: " << numberOfChildPageIDValues << endl << endl;

			// write the childPageID / searchKey pairs in the IndexPage to file ..
			indexPage_->getChildPageIDs()->clear();
			for (int i=0; i < numberOfChildPageIDValues; i++) {

				// "pageIDs" (#1 ptr) - (#5 ptr) ..

				tbl.read( reinterpret_cast<char *>( &fieldValue ), intSize );

				//indexPage_->getChildPageIDs()->at( i ) = fieldValue;
				indexPage_->getChildPageIDs()->push_back( fieldValue );

				// DEBUGGING ..
				//cout << "IndexManager:: readIndexPage() childPageIDs->at(i("<<i<<")): " << fieldValue << endl << endl;


			}

			//cout << "IndexManager:: readIndexPage() numberOfSearchKeyValues: " << numberOfSearchKeyValues << endl << endl;

			// write the childPageID / searchKey pairs in the IndexPage to file ..
			indexPage_->getSearchKeys()->clear();
			for (int i=0; i < numberOfSearchKeyValues; i++) {

				// "searchKeys" (#1 sk) - (#4 sk) ..

				tbl.read( reinterpret_cast<char *>( &fieldValue ), intSize );

				//indexPage_->getSearchKeys()->at( i ) = fieldValue;
				indexPage_->getSearchKeys()->push_back( fieldValue );

				// DEBUGGING ..
				//cout << "IndexManager:: readIndexPage() searchKeys->at(i("<<i<<")): " << fieldValue << endl << endl;

			}
			bool isLeafNode = false;
			tbl.read( reinterpret_cast<char *>( &isLeafNode ), boolSize );
			//bool isLeafNode = fieldValue;
			indexPage_->setIsLeafNode( isLeafNode );

			//cout << "IndexManager:: readIndexPage() isLeafNode: " << isLeafNode << endl << endl;
			//cout << "IndexManager:: readIndexPage() fieldValue: " << fieldValue << endl << endl;

			// "pageID_" value is not read from file,
			// but rather set with whatever value
			// was passed in to this function ..
			indexPage_->setPageID( pageID_ );

			readPagePerformed = true;


		} else {

	 		string errorString = "\nFailed to open data file [ SM::readIndexPage() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		}  // end "if ( !tbl.fail() )" ..

		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();

		// DEBUGGING ..
		//cout << "IndexManager:: readDataPage() maxSlotsAvailable: " << maxSlotsAvailable << endl << endl;


	} // end "if ( emptyString.compare( tableName ) != 0 )" ..


	return readPagePerformed;


} // end "readIndexPage()" ..


bool IndexManager:: readIndexPages( string indexTableName_, string indexName_, vector<IndexPage*>* indexPages_ ){

	//cout << "IndexManager:: writeIndexPages() start .." << endl << endl;

	bool wasIndexPageRead = false;
	bool wereIndexPagesRead = true;

	IndexHeader* header = new IndexHeader();
	bool isReadIndexHeader = readIndexHeader( indexTableName_, indexName_, header );

	//int indexPagesSize = header.getTotalNumberOfIndexPages();

	vector<IndexHeaderEntry*>* headerEntries = header->getHeaderEntries();

	int headerEntriesSize = headerEntries->size();

	int headerEntryPageID = -1;

	IndexPage* indexPage;

	// write each "index page" ..
	//for(int i=0; i<indexPagesSize; i++){
	for(int i=0; i<headerEntriesSize; i++){

		indexPage = new IndexPage();

		headerEntryPageID = headerEntries->at(i)->getPageID();

		// read this index page from file ..
		//wasIndexPageRead = readIndexPage( indexTableName_, indexName_, indexPage, i );
		wasIndexPageRead = readIndexPage( indexTableName_, indexName_, indexPage, headerEntryPageID );

		// update parameter ..
		if( wasIndexPageRead == false )
			wereIndexPagesRead = false;
		else
			indexPages_->push_back( indexPage );

	} // end "for()" ..

	if( isReadIndexHeader == false )
		wereIndexPagesRead = false;

	return wereIndexPagesRead;


} // end "readIndexPages()" ..


int IndexManager:: getIndexSize( string tableName_, string indexName_ ) {


	// "tableName": name of the table associated with the index ..

	// "indexName_": name of the index ..


	//cout << "IndexManager:: getIndexSize() start .." << endl << endl;
	//cout << "IndexManager:: getIndexSize() tableName: " << tableName_ << endl << endl;
	//cout << "IndexManager:: getIndexSize() indexName_: " << indexName_ << endl << endl;

	string emptyString = "";

	string tableNameString = tableName_;

	int indexSize = -1;


	if ( emptyString.compare( tableNameString ) != 0 ) {


		//string directoryPlusTableName = "C:\\Temp\\";

		string directoryPlusTableName = databaseFile->getFileDrive();

		directoryPlusTableName += databaseFile->getFileDirectory();

		directoryPlusTableName += "Index_";
		directoryPlusTableName += tableName_;
		directoryPlusTableName += "_";
		directoryPlusTableName += indexName_;
		char* fileName = (char*)directoryPlusTableName.c_str();

		fstream tbl( fileName,  ios::in | ios::binary );

		if ( !tbl.fail() ) {


			// DataPage is written into its own file
			// so locate "page-to-read", referenced from beginning of file ..
			tbl.seekg( 0, ios::end );

			indexSize = tbl.tellg();

			//cout << "IndexManager:: getIndexSize() indexSize: " << indexSize << endl << endl;


		} else {

	 		string errorString = "\nFailed to open data file [ SM::readIndexPage() ], so data page was not read.";

	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringSM( errorString );

		}  // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();


	} // end "if ( emptyString.compare( tableName ) != 0 )" ..


	return indexSize;


} // end "readIndexPage()" ..



bool IndexManager:: updateIndex( string indexName_, string indexTableName_, string attributeName_, int searchKey_, int childPageID_ ){

 	//cout << "IndexManager:: updateIndex() start .." << endl << endl;
 	//cout << "IndexManager:: updateIndex()   searchKey_: " << searchKey_ << endl << endl;
 	//cout << "IndexManager:: updateIndex() childPageID_: " << childPageID_ << endl << endl;
 	//cout << "IndexManager:: updateIndex() indexTableName_: " << indexTableName_ << endl << endl;
 	//cout << "IndexManager:: updateIndex() indexName_: " << indexName_ << endl << endl;

 	// retrieve Index metadata from the Catalog ..
	getCatalogManager()->readMetaData();

	//cout << "IndexManager:: updateIndex() Point A .." << endl << endl;

 	// do this in code that calls this method ..
 	//vector<string>* attributeNames = getCatalogManager()->getCatalogIndexAttributeNames( tableName );

 	string indexTableName = indexTableName_;
	string indexName = indexName_;
	string attributeName = attributeName_;

 	CatalogIndex* catalogIndex = getCatalogManager()->getCatalogIndex_attributeName( indexTableName, attributeName );

 	bool isNoDuplicates = catalogIndex->getIsNoDuplicates();

 	bool createIndexPerformed = false;

	// check if the table exists in the catalog ..
	bool doesCatalogTableExist = getCatalogManager()->doesCatalogTableNameExist( indexTableName );

	// DEBUGGING ..
	//cout << "IndexManager:: updateIndex() doesCatalogTableExist: " << doesCatalogTableExist << endl << endl;

	if ( doesCatalogTableExist == true ) {

		// check if the attribute already has an index on it ..
		bool doesCatalogTableAttributeNameHaveIndex = getCatalogManager()->doesCatalogTableAttributeNameHaveIndex( indexTableName, attributeName );

		// DEBUGGING ..
		//cout << "IndexManager:: updateIndex() doesCatalogTableAttributeNameHaveIndex: " << doesCatalogTableAttributeNameHaveIndex << endl << endl;

		if ( doesCatalogTableAttributeNameHaveIndex == true ) {

			int tableSize = getCatalogManager()->getTableSize( indexTableName_ );

			// DEBUGGING ..
			//cout << "IndexManager:: updateIndex() tableSize: " << tableSize << endl << endl;
			//cout << "IndexManager:: updateIndex() catalogIndex == NULL?: " << (catalogIndex == NULL) << endl << endl;
			//cout << "IndexManager:: updateIndex() index name: " << catalogIndex->getIndexName() << endl << endl;
			//cout << "IndexManager:: updateIndex() index table name: " << catalogIndex->getIndexTableName() << endl << endl;
			//cout << "IndexManager:: updateIndex() index isNoDuplicates: " << catalogIndex->getIsNoDuplicates() << endl << endl;
			//cout << "IndexManager:: updateIndex() indexSize: " << catalogIndex->getIndexSize() << endl << endl;

			// table exists, index exists, but it may or may not have data yet ..

			if( ( tableSize > 0 ) &&
				( catalogIndex->getIndexSize() == 0 ) ) {

				createIndexPerformed = createIndexWithData( indexName_, indexTableName_, isNoDuplicates, attributeName_ );

				//cout << "IndexManager:: updateIndex() Point A.. " << endl << endl;

 				int newIndexSize = calculateIndexSize( indexTableName_, indexName_ );
 				//int newIndexSize = getIndexSize( indexTableName_, indexName_ );

 				//cout << "IndexManager:: updateIndex() newIndexSize: " << newIndexSize << endl << endl;

 				//catalogIndex->setIndexSize( newIndexSize );

 				getCatalogManager()->getCatalogIndex( indexName_ )->setIndexSize( newIndexSize );
 				
 				//cout << "IndexManager:: updateIndex() Point B.. " << newIndexSize << endl << endl;

 				//catalogIndex.setIndexSize( newIndexSize );
 				//getCatalogManager()->setCatalogIndex( indexName_, catalogIndex );

 				getCatalogManager()->writeMetaData();
 				
 				//cout << "IndexManager:: updateIndex() Point C.. " << newIndexSize << endl << endl;

				//if ( createIndexPerformed == false )
				return createIndexPerformed;

			} // end "if( tableSize > 0 )" ..

		} // end "if()" ..

	} // end "if()" ..

 	//cout << "IndexManager:: updateIndex() Point D .." << endl << endl;

 	// insert into index page ..

 	bool isIndexUpdated = false;

 	bool searchKeyFound = false;

 	vector<int>* returnPageIDs = lookUpLeafNodeID( searchKey_, indexTableName_, indexName_, searchKeyFound );
	int parentPageID = returnPageIDs->at(0);
	int pageID = returnPageIDs->at(1);

	//int leafNodePageID = pageID;

	//cout << "IndexManager:: updateIndex() searchKeyFound: " << searchKeyFound << ", searchKey: " << searchKey_ << ", childPageID: " << childPageID_ << endl << endl;
	//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;
	//cout << "IndexManager:: updateIndex() pageID: " << pageID << endl << endl;
	//cout << "IndexManager:: updateIndex() Point E .." << endl << endl;


	// if "searchKey" is not found, then
	// we insert the searchKey into the node ..
	if ( searchKeyFound == false ){
		
		// find the pageID of the leaf node where the searchKey should be inserted ..
		//pageID = lookUpIDForInsert( indexTableName_, indexName_, searchKey_ );
		
		//cout << "IndexManager:: updateIndex() found pageID: " << pageID << endl << endl;

 		IndexPage* indexPage = NULL;

 		IndexPage* parentIndexPage = NULL;

 		int childLeftMostSearchKey = -1;

 		bool isTraversingParentNodes = true;

 		// locate the index page for insert ..
 		// if pageID == -1, insert
 		bool isParentOverflow = false;
 		bool isReadIndexPage = true;
 		indexPage = new IndexPage();
 		if( pageID > 0 ){
			
			isReadIndexPage = readIndexPage( indexTableName_, indexName_, indexPage, pageID );
 		
 		} else {

			// check whether the parent node will overflow with this addition ..
			// else create a new node
			//IndexPage* parentIndexPage = new IndexPage();
			//isReadIndexPage = readIndexPage( indexTableName_, indexName_, parentIndexPage, parentPageID );
			//if( parentIndexPage->getSearchKeys()->size() == 4 )
			//	isParentOverflow = true;

			// prepare value of new leaf node ..
			indexPage->setIsLeafNode( true );

		}


		//cout << "IndexManager:: updateIndex() - 4821 - Point F .." << endl << endl;

		// add the new "searchKey" to the set ..
		if( indexPage != NULL ){
			
			indexPage->getSearchKeys()->push_back( searchKey_ );
			indexPage->getChildPageIDs()->push_back( childPageID_ );

			//cout << "IndexManager:: updateIndex() searchKey: " << searchKey_ << endl << endl;
			//cout << "IndexManager:: updateIndex() childPageID: " << childPageID_ << endl << endl;
		}


 		// loop through the parent nodes of the index,
 		// inserting "splitting keys" and/or
 		// splitting the parent node whenever the parent node overflows ..
 		while( isTraversingParentNodes ){

 			//cout << "IndexManager:: updateIndex() - 4839 - Point G .." << endl << endl;

 			// DEBUGGING ..
			//vector<int>* searchKeys_pnt = indexPage->getSearchKeys();
			//vector<int>* childPageIDs_pnt = indexPage->getChildPageIDs();

			//for(int i=0; i<searchKeys_pnt->size(); i++)
			//	cout << "IndexManager:: updateIndex() searchKeys_print->at(i("<<i<<")): " << searchKeys_pnt->at(i) << endl << endl;

			//for(int i=0; i<childPageIDs_pnt->size(); i++)
			//	cout << "IndexManager:: updateIndex() childPageIDs_print->at(i("<<i<<")): " << childPageIDs_pnt->at(i) << endl << endl;

			// clause ends on line: 6275 ..
			if ( isReadIndexPage == true ) {

				//cout << "IndexManager:: updateIndex() Point F .." << endl << endl;

				// note whether this node is a "leaf" node or not ..
				bool isLeafNode = indexPage->getIsLeafNode();

				int searchKeysSize = indexPage->getSearchKeys()->size();

				vector<int>* searchKeys = new vector<int>();
				for(int i=0; i<searchKeysSize; i++)
					searchKeys->push_back( indexPage->getSearchKeys()->at(i) );

				// used below, near end of loop, for the parent node ..
				childLeftMostSearchKey = searchKeys->at( 0 );

				// sort the searchKeys set now that there is a new, unsorted addition ..
				sort( searchKeys->begin(), searchKeys->begin() + searchKeysSize );

				// match the "childPageIDs" order with the sorted order of the "searchKeys" ..
				vector<int>* childPageIDs = new vector<int>();

				// loop through each "sorted" searchKey,
				// find its counter-part in the original sort ..
				int searchKeyCounter = 0;

				//cout << "IndexManager:: updateIndex() - 4881 - Point H .." << endl << endl;

				// initialize "childPageIDs" ..
				// ("childPageIDs" size is same size as "searchKeys" for leaf nodes) ..
				int childPageIDSize = -1;
				if (isLeafNode == true)
					childPageIDSize = searchKeysSize;
				else
					childPageIDSize = (searchKeysSize+1);

				while( searchKeyCounter < childPageIDSize ){
					childPageIDs->push_back( -1 );

					searchKeyCounter++;

				}

				// "searchKeys" is sorted, but "childPageIDs" is not yet sorted,
				// and we want each value of "childPageIDs" to match the sorted order of "searchKeys",
				// for each value of "searchKeys", find the index of its value in "indexPage->getSearchKeys()",
				// this index value is used to get the childPageID value in "indexPage->getChildPageIDs()",
				// put this childPageID value into "childPageIDs",
				// now each value of "childPageIDs" will match the sorted order of "searchKeys" ..

				// reinitialize the counter for the upcoming loop ..
				searchKeyCounter = 0;

				int childPageIDCounter = 0;

				int firstSearchKeyPosition = -1;

				while( searchKeyCounter < searchKeysSize ){

					//cout << "IndexManager:: updateIndex() Point I .." << endl << endl;

					for( int i=0; i<searchKeysSize; i++ ){

						//cout << "IndexManager:: updateIndex() Point J .." << endl << endl;
						//cout << "IndexManager:: updateIndex() ( searchKeys->at( searchKeyCounter("<<searchKeyCounter<<") )("<<searchKeys->at( searchKeyCounter )<<") == indexPage->getSearchKeys()->at( i("<<i<<") )("<<indexPage->getSearchKeys()->at( i )<<") ): " << ( searchKeys->at( searchKeyCounter ) == indexPage->getSearchKeys()->at( i ) ) << endl << endl;

						// compare "sorted" searchKeys ("searchKeys") to the original sort ("indexPage->getSearchKeys()") ..
						// (original sort matches the current state of the childPageIDs) ..
						if ( searchKeys->at( searchKeyCounter ) == indexPage->getSearchKeys()->at( i ) ){

							//cout << "IndexManager:: updateIndex() Point L .." << endl << endl;
							
/*
							// handles the first searchKey ..
							// (extra childPageID is handled in special manner) ..
							if( counter == 0 ){

								cout << "IndexManager:: updateIndex() Point K .." << endl << endl;

								if( i > 0 ) {

									cout << "IndexManager:: updateIndex() Point L .." << endl << endl;

									// this indicates that the unsorted searchKey is NOT the first in the series ..
									// (if its NOT first in series, always take RIGHT childPageID) ..

									//(sk: 0-4, cID: 0-5)
									// _A_C_D_ B_		<-i
									//     <- -
									// _A_ B_ C_D_		<-counter

									childPageIDs->at( counter ) = indexPage->getChildPageIDs()->at( i+1 );

								} else if( i == 0 ) {

									cout << "IndexManager:: updateIndex() Point M .." << endl << endl;

									// this indicates that the unsorted searchKey IS the first in the series ..
									// (if it IS first in series, take LEFT childPageID) ..

									//(sk: 0-4, cID: 0-5)
									// _B _A_C_D_	<-i
									// - ->
									// _A _B _C_D_	<-counter

									childPageIDs->at( counter ) = indexPage->getChildPageIDs()->at( i );

									cout << "IndexManager:: updateIndex() childPageIDs->at( counter("<<counter<<") ): " << childPageIDs->at( counter ) << endl << endl;

									cout << "IndexManager:: updateIndex() childPageIDs->size(): " << childPageIDs->size() << endl << endl;

								} // end "if()" ..

							} // end "if()" ..

							// get the "next" value right away ..
							// (make sure indices don't go over the vector "edge") ..
							if( ( (counter+1) < childPageIDs->size() ) &&
								( i+1 < searchKeysSize ) ){

								cout << "IndexManager:: updateIndex() Point N .." << endl << endl;

								childPageIDs->at( counter+1 ) = indexPage->getChildPageIDs()->at( i+1 );

							}

							break;
*/

							// handles the first searchKey ..
							if( searchKeyCounter == 0 ){

								//cout << "IndexManager:: updateIndex() Point M .." << endl << endl;

								firstSearchKeyPosition = i;

								//cout << "IndexManager:: updateIndex() indexPage->getChildPageIDs()->at( i("<<i<<") ): " << indexPage->getChildPageIDs()->at( i ) << endl << endl;

								childPageIDs->at( childPageIDCounter ) = indexPage->getChildPageIDs()->at( i );

								// get the "next" value right away ..
								// (make sure indices don't go over the vector "edge") ..
								// (don't do this for leaf nodes) ..
								if( ( (childPageIDCounter+1) < childPageIDs->size() ) &&
									( i+1 < searchKeysSize ) &&
									( isLeafNode == false ) ){

									//cout << "IndexManager:: updateIndex() Point N .." << endl << endl;
									//cout << "IndexManager:: updateIndex() indexPage->getChildPageIDs()->at( i+1("<<i+1<<") ): " << indexPage->getChildPageIDs()->at( i+1 ) << endl << endl;

									childPageIDCounter++;

									childPageIDs->at( childPageIDCounter ) = indexPage->getChildPageIDs()->at( i+1 );

								}

							// handles remaining searchKeys ..
							} else {

								//cout << "IndexManager:: updateIndex() Point O .." << endl << endl;
								//cout << "IndexManager:: updateIndex() indexPage->getChildPageIDs()->size() - 1: " << (indexPage->getChildPageIDs()->size() - 1) << endl << endl;
								//cout << "IndexManager:: updateIndex() i+1: " << (i+1) << endl << endl;

								// "firstSearchKeyPosition" indicates the first value used from "indexPage->getChildPageIDs()",
								// when "i < firstSearchKeyPosition" then use index of "i",
								// when "i > firstSearchKeyPosition" then use index of "i+1"
								// ( "i+1", since "next" value found right away, above) ..

								if( isLeafNode == false ){


									if( i < firstSearchKeyPosition ){

										//cout << "IndexManager:: updateIndex() Point P .." << endl << endl;
										//cout << "IndexManager:: updateIndex() indexPage->getChildPageIDs()->at( i("<<i<<") ): " << indexPage->getChildPageIDs()->at( i ) << endl << endl;
										//(sk: 0-4, cID: 0-5)
										// _B _A_ D_C_	<-i
										// -B->
										// _A_ _BC_D_	<-counter
										childPageIDs->at( childPageIDCounter ) = indexPage->getChildPageIDs()->at( i );

									} else if( i > firstSearchKeyPosition ) {

										//cout << "IndexManager:: updateIndex() Point Q .." << endl << endl;
										//cout << "IndexManager:: updateIndex() indexPage->getChildPageIDs()->at( i+1("<<i+1<<" ): " << indexPage->getChildPageIDs()->at( i+1 ) << endl << endl;
										//(sk: 0-4, cID: 0-5)
										// _B _A_ D_C_	<-i
										// 	        <-C-
										// _A_ _BC_D_	<-counter
										childPageIDs->at( childPageIDCounter ) = indexPage->getChildPageIDs()->at( i+1 );

									} else {

										// this shouldn't happen( i == firstSearchKeyPosition ),
										// this should be taken care of above already ..
										//cout << "IndexManager:: updateIndex() Point R .." << endl << endl;

									}

								} else {

									//cout << "IndexManager:: updateIndex() indexPage->getChildPageIDs()->at( i("<<i<<") ): " << indexPage->getChildPageIDs()->at( i ) << endl << endl;
									childPageIDs->at( childPageIDCounter ) = indexPage->getChildPageIDs()->at( i );

								} // end "if()" ..

							} // end "if()" ..

							//cout << "IndexManager:: updateIndex() Point S .." << endl << endl;

							break;


						} // end "if()" ..

					} // end "for()" ..

					// increment the counters ..
					searchKeyCounter++;
					childPageIDCounter++;

				} // end "while()" ..

				//cout << "IndexManager:: updateIndex() Point T .." << endl << endl;

				// update the "indexPage" with the sorted "searchKeys" and "childPageIDs" ..
				indexPage->setSearchKeys( searchKeys );
				indexPage->setChildPageIDs( childPageIDs );

				//cout << "IndexManager:: updateIndex() Point U .." << endl << endl;
				//cout << "IndexManager:: updateIndex() ( isLeafNode == true ): " << ( isLeafNode == true ) << endl << endl;
				//cout << "IndexManager:: updateIndex() ( searchKeysSize("<<searchKeysSize<<") <= (fanout - 1)("<<(fanout - 1)<<") ): " << ( searchKeysSize <= (fanout - 1) ) << endl << endl;

				parentIndexPage = new IndexPage();

				IndexHeader* indexHeader = new IndexHeader();
				bool isReadIndexHeader = readIndexHeader( indexTableName_, indexName_, indexHeader );

				int totalNumberOfIndexPages = indexHeader->getTotalNumberOfIndexPages();

				vector<IndexHeaderEntry*>* headerEntries = indexHeader->getHeaderEntries();

				// retrieve parentPageID of the parent node ..
				// (useful whether or not childPageIDFound == true/false) ..
				bool parentParentPageIDFound = false;
				int parentParentPageID = -1;
				for(int i=0; i<totalNumberOfIndexPages; i++){

					//cout << "IndexManager:: updateIndex() ( parentPageID("<<parentPageID<<") == headerEntries->at(i("<<i<<"))->getPageID()("<<headerEntries->at(i)->getPageID()<<") ): " << ( parentPageID == headerEntries->at(i)->getPageID() ) << endl << endl;
					//cout << "IndexManager:: updateIndex() headerEntries->at(i("<<i<<"))->getParentPageID(): " << headerEntries->at(i)->getParentPageID() << endl << endl;
					//cout << "IndexManager:: updateIndex() headerEntries->at(i("<<i<<"))->getFreeSearchKeyCount(): " << headerEntries->at(i)->getFreeSearchKeyCount() << endl << endl;
					//cout << "IndexManager:: updateIndex() headerEntries->at(i("<<i<<"))->getNumberOfSearchKeys(): " << headerEntries->at(i)->getNumberOfSearchKeys() << endl << endl;
					//cout << "IndexManager:: updateIndex() headerEntries->at(i("<<i<<"))->getNumberOfChildPageIDs(): " << headerEntries->at(i)->getNumberOfChildPageIDs() << endl << endl;

					if( parentPageID == headerEntries->at(i)->getPageID() ){

						parentParentPageIDFound = true;
						parentParentPageID = headerEntries->at(i)->getParentPageID();
						break;

					}

				}

				//cout << "IndexManager:: updateIndex() - 5119 - Point V .." << endl << endl;
				//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;
				//cout << "IndexManager:: updateIndex() parentParentPageIDFound: " << parentParentPageIDFound << endl << endl;
				//cout << "IndexManager:: updateIndex() parentParentPageID: " << parentParentPageID << endl << endl;


				if( ( isLeafNode == true ) &&
					( searchKeysSize <= (fanout - 1) ) ) {

					// since this is a leaf node,
					// prepare to write the new indexPage to file ..

					int maxPageID = -1;

					if( pageID < 0 ){

						// if "pageID == -1", then assign a new value ..

						int currentPageID = 0;

						// find the maxPageID out of all the indexPages ..
						for(int i=0; i<totalNumberOfIndexPages; i++){

							currentPageID = headerEntries->at(i)->getPageID();

							if( currentPageID > maxPageID )
								maxPageID = currentPageID;

						}

						// new pageID will be 1 more than the current maxPageID ..
						maxPageID++;

						//cout << "IndexManager:: updateIndex() pageID(1): " << pageID << endl << endl;

						// assign a new pageID, using the maxPageID just here calculated ..
						if( maxPageID > pageID ) {

							pageID = maxPageID;

							indexPage->setPageID( pageID );

						}

					}



					//cout << "IndexManager:: updateIndex() Point W .." << endl << endl;
					//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;
					//cout << "IndexManager:: updateIndex() pageID(2): " << pageID << endl << endl;
					//cout << "IndexManager:: updateIndex() maxPageID: " << maxPageID << endl << endl;

					//vector<int>* searchKeys_print = indexPage->getSearchKeys();
					//vector<int>* childPageIDs_print = indexPage->getChildPageIDs();

					//for(int i=0; i<searchKeys_print->size(); i++)
					//	cout << "IndexManager:: updateIndex() searchKeys_print->at(i("<<i<<")): " << searchKeys_print->at(i) << endl << endl;
					//for(int i=0; i<childPageIDs_print->size(); i++)
					//	cout << "IndexManager:: updateIndex() childPageIDs_print->at(i("<<i<<")): " << childPageIDs_print->at(i) << endl << endl;
					//cout << "IndexManager:: updateIndex() indexPage->getIsLeafNode(): " << indexPage->getIsLeafNode() << endl << endl;
					//cout << "IndexManager:: updateIndex() indexPage->getPageID(): " << indexPage->getPageID() << endl << endl;

					// need to update the "indexPage" with the new additions ..
					// track whether the latest searchKey started a new indexPage ..
					bool isNew = true;
					if( indexPage->getSearchKeys()->size() > 1 )
						isNew = false;

					// write the current indexPage ..
					bool isWriteIndexPage = writeIndexPage( indexTableName_, indexName_, indexPage, pageID, parentPageID, isNew, isNoDuplicates );

					// update the parent indexPage ..

					bool isReadIndexPage = readIndexPage( indexTableName_, indexName_, parentIndexPage, parentPageID );

					//cout << "IndexManager:: updateIndex() Point X .." << endl << endl;

					// look for the "childPageID" in the parent ..
					// (only indicate whether or not it is found) ..
					int numberParentChildPageIDs = parentIndexPage->getChildPageIDs()->size();
					bool childPageIDFound = false;
					for(int i=0; i<numberParentChildPageIDs; i++){

						if( parentIndexPage->getChildPageIDs()->at(i) == pageID ){

							childPageIDFound = true;

						}

					}

					//cout << "IndexManager:: updateIndex() Point Y .." << endl << endl;
					//cout << "IndexManager:: updateIndex() childPageIDFound: " << childPageIDFound << endl << endl;

					// look for the child "searchKey" in the parent ..
					// (indicate whether or not it is found, AND
					//  grab the index value where it was found) ..
					int numberOfSearchKeys = parentIndexPage->getSearchKeys()->size();
					int childSearchKey = indexPage->getSearchKeys()->at(0);
					int childPageID = pageID;
					int parentSearchKeyIndex = -1;
					int childSearchKeyPosition = -1;
					bool childSearchKeyFound = false;
					for(int i=0; i<numberOfSearchKeys; i++){

						if( childSearchKey == parentIndexPage->getSearchKeys()->at(i) ){

							childSearchKeyPosition = i;
							childSearchKeyFound = true;
							break;

						}

					}

					//cout << "IndexManager:: updateIndex() Point Z .." << endl << endl;
					//cout << "IndexManager:: updateIndex() childSearchKey: " << childSearchKey << endl << endl;
					//cout << "IndexManager:: updateIndex() childSearchKeyFound: " << childSearchKeyFound << endl << endl;

					if( childPageIDFound == false ){

						// this means that the parent doesnt yet point to this child indexPage,
						// so need to update the parent indexPage to point to this child indexPage ..

						//cout << "IndexManager:: updateIndex() Point AA .." << endl << endl;

						// add the childSearchKey if it isn't in the parent node yet ..
						// (or if its a "root-parent-with-only-1-node", replace its dummy searchKey value ) ..
						if( childSearchKeyFound == false ){

							//cout << "IndexManager:: updateIndex() Point AB .." << endl << endl;
							//cout << "IndexManager:: updateIndex() ( parentIndexPage->getChildPageIDs()->size()("<<parentIndexPage->getChildPageIDs()->size()<<") <= 1 ): " << ( parentIndexPage->getChildPageIDs()->size() <= 1 ) << endl << endl;
							//cout << "IndexManager:: updateIndex() && ( parentPageID("<<parentPageID<<") == 0 ): " << ( parentPageID == 0 ) << endl << endl;

							if( parentIndexPage->getChildPageIDs()->size() == 1 ){

							//if( ( parentIndexPage->getChildPageIDs()->size() <= 1 ) &&
							//	( parentPageID == 0 ) ){

								// need size of the child leaf node ..
								IndexPage* childIndexPage = new IndexPage();
								int parentLeftMostChildPageID = parentIndexPage->getChildPageIDs()->at(0);

								//cout << "IndexManager:: updateIndex() Point AC .." << endl << endl;

								//cout << "IndexManager:: updateIndex() parentIndexPage->getChildPageIDs()->at(0): " << parentIndexPage->getChildPageIDs()->at(0) << endl << endl;

								bool isReadIndexPage = readIndexPage( indexTableName_, indexName_, childIndexPage, parentLeftMostChildPageID );

								//cout << "IndexManager:: updateIndex() Point AD .." << endl << endl;

								int childSearchKeySize = childIndexPage->getSearchKeys()->size();

								if ( parentPageID == 0 ){

									//		root
									// leaf		leaf(new here?) ..

									// 2 nodes total - parent is the root-node, child is the leaf-node ..
									// (replace the parent-node dummy searchKey with the left-most searchKey of the
									//  second child leaf-node) ..
									// (also, add the pageID for the new leaf-node) ..

									//cout << "IndexManager:: updateIndex() Point AE .." << endl << endl;

									// need to figure out whether the "childPageID" to add to the parent node
									// should be left or right of the existing sole childPageID in the parent node ..

									//				root				OR			root
									// leaf(new here)		leaf ..		      leaf      leaf(new here) ..

									//cout << "IndexManager:: updateIndex() childIndexPage->getSearchKeys()->size(): " << childIndexPage->getSearchKeys()->size() << endl << endl;

									if( childSearchKey < childIndexPage->getSearchKeys()->at(0) ){

										// new "childIndexPage" is LEFT of the existing "childIndexPage" ..

										//cout << "IndexManager:: updateIndex() Point AF .." << endl << endl;

										// parent node gets searchKey from EXISTING (right) node ..
										parentIndexPage->getSearchKeys()->at(0) = childIndexPage->getSearchKeys()->at(0);

										//cout << "IndexManager:: updateIndex() Point AG .." << endl << endl;

										vector<int>::iterator childIterator = parentIndexPage->getChildPageIDs()->begin();

										//cout << "IndexManager:: updateIndex() Point AH .." << endl << endl;

										// new "childPageID" should be LEFT of the existing "childPageID" in the parent ..
										// (insert at the beginning) ..
										parentIndexPage->getChildPageIDs()->insert( childIterator, childPageID );

									} else if( childSearchKey > childIndexPage->getSearchKeys()->at( ( childSearchKeySize - 1 ) ) ){

										// new "childIndexPage" is RIGHT of the existing "childIndexPage" ..

										//cout << "IndexManager:: updateIndex() Point AI .." << endl << endl;

										// parent node gets searchKey from NEW (right) node ..
										parentIndexPage->getSearchKeys()->at(0) = childSearchKey;

										//cout << "IndexManager:: updateIndex() Point AJ .." << endl << endl;

										// new "childPageID" should be RIGHT of the existing "childPageID" in the parent ..
										/// (insert at the end) ..
										parentIndexPage->getChildPageIDs()->push_back( childPageID );

									}


								} else if ( parentPageID > 0 ){

									//cout << "IndexManager:: updateIndex() Point AK .." << endl << endl;

									//		       root
									//     (intermediate nodes) ..
									//	  parent			parent
									// leafx5			 leafx1 .. (new inserts here) ..

									// N nodes total - parent is the intermediate-node, child is leaf-node ..
									// (replace the parent-node dummy searchKey with 1 greater than the left-most searchKey
									//  of the child leaf-node) ..

									if( childIndexPage->getChildPageIDs()->size() >= 4 ) {

										// left leaf node is full, so replace dummy parentSearchKey with new searchKey ..

										parentIndexPage->getSearchKeys()->at(0) = childSearchKey;

										// new "childPageID" should be RIGHT of the existing "childPageID" in the parent ..
										parentIndexPage->getChildPageIDs()->push_back( childPageID );

									} else {

										// left leaf node is NOT full, so keep dummy parentSearchKey ..

										int maxSearchKeyInChild = childIndexPage->getSearchKeys()->at( ( childSearchKeySize - 1 ) );

										// new dummy searchKey in parent is 1 greater than max searchKey in child node ..
										parentIndexPage->getSearchKeys()->at(0) = ( maxSearchKeyInChild + 1 );

										// no need for to update "parentIndexPage->getChildPageIDs()" here,
										// since this parent has only 1 child indexPage ..
										// (i.e. "parentIndexPage->getChildPageIDs()->size() == 1") ..

									}

								}

							} else {

								//		root	       				OR				root
								//  leaf    leaf  (new inserts)   			(intermediate nodes) ..
								//	  										parent			parent
								// 										leafx5			 leafx2 .. (new inserts)

								//cout << "IndexManager:: updateIndex() Point AL .." << endl << endl;

								// insert searchKey and pageID into the parent-node ..

								for(int i=0; i<numberOfSearchKeys; i++){

									if( childSearchKey > parentIndexPage->getSearchKeys()->at(i) ) {

										continue;

									} else if( childSearchKey < parentIndexPage->getSearchKeys()->at(i) ){

										// in general parentSearchKeyIndex is 1 less than the index found at i,
										// because at i, childSearchKey < parentSearchKey ..
										parentSearchKeyIndex = i;
										break;

									}

								} // end "for()" ..

								//cout << "IndexManager:: updateIndex() Point AM .." << endl << endl;
								//cout << "IndexManager:: updateIndex() parentSearchKeyIndex: " << parentSearchKeyIndex << endl << endl;

								// add the childSearchKey to the parent ..
								if( parentSearchKeyIndex >= 0 ){

									//cout << "IndexManager:: updateIndex() Point AN .." << endl << endl;

									// helper variables ..
									int parentChildSearchKey = -1;
									int childPageIDOffset = -1;

									if( parentSearchKeyIndex > 0 ){

										//cout << "IndexManager:: updateIndex() Point AO .." << endl << endl;

										// (ie: parentChildPageIDIndex = parentSearchKeyIndex + 1) ..

										childPageIDOffset = 1;

										// left-most node remains the left-most node (new "childSearchKey" is to the right of it),
										// so the "childSearchKey" just gets inserted directly into the parent node ..
										parentChildSearchKey = childSearchKey;

									} else {

										//cout << "IndexManager:: updateIndex() Point AP .." << endl << endl;

										// "parentSearchKeyIndex == 0", so
										// need to figure out how to update the parent node ..

										// if left-most searchKey
										// of left-most node
										// is greater than "childSearchKey", then
										// indexPage becomes the left-most node, and
										// pageID of indexPage is added to first childPageID slot of the parent node ..
										IndexPage* leftMostChildIndexPage = new IndexPage();
										int leftMostChildPageID = parentIndexPage->getChildPageIDs()->at(0);
										bool isReadIndexPage = readIndexPage( indexTableName_, indexName_, leftMostChildIndexPage, leftMostChildPageID );

										//cout << "IndexManager:: updateIndex() Point AQ .." << endl << endl;

										if( childSearchKey < leftMostChildIndexPage->getSearchKeys()->at(0) ){

											//cout << "IndexManager:: updateIndex() Point AR .." << endl << endl;

											// (ie: parentChildPageIDIndex = parentSearchKeyIndex + 0) ..

											childPageIDOffset = 0;

											// "childSearchKey" is the new left-most node, so
											// the old left-most-node, left-most searchKey into the parent node ..
											parentChildSearchKey = leftMostChildIndexPage->getSearchKeys()->at(0);

										} else {

											//cout << "IndexManager:: updateIndex() Point AS .." << endl << endl;

											// if "childSearchKey" is greater than
											// right-most searchKey
											// of left-most node, then
											// pageID of indexPage is added to 1 more than the
											// childSearchKey position of childPageID slot of the parent node ..
											// (ie: parentChildPageIDIndex = parentSearchKeyIndex + 1) ..

											childPageIDOffset = 1;

											// left-most node remains the left-most node (new "childSearchKey" is right of it), so
											// the "childSearchKey" just gets inserted directly into the parent node ..
											parentChildSearchKey = childSearchKey;

										}

									}

									//cout << "IndexManager:: updateIndex() Point AT .." << endl << endl;

									vector<int>::iterator childIterator = parentIndexPage->getSearchKeys()->begin();

									//cout << "IndexManager:: updateIndex() Point AU .." << endl << endl;

									parentIndexPage->getSearchKeys()->insert( ( childIterator + parentSearchKeyIndex ), parentChildSearchKey );

									//cout << "IndexManager:: updateIndex() Point AV .." << endl << endl;

									// childPageIDs has 1 element more than searchKeys, so adjust for this ..
									// (special case: if insert is at i == 0, childPageIDIndex == 0, too) ..
									int parentChildPageIDIndex = ( parentSearchKeyIndex + childPageIDOffset );

									//cout << "IndexManager:: updateIndex() Point AW .." << endl << endl;

									childIterator = parentIndexPage->getChildPageIDs()->begin();

									//cout << "IndexManager:: updateIndex() Point AX .." << endl << endl;

									parentIndexPage->getChildPageIDs()->insert( ( childIterator + parentChildPageIDIndex ), childPageID );

									//cout << "IndexManager:: updateIndex() Point AY .." << endl << endl;

								} else {

									//cout << "IndexManager:: updateIndex() Point AZ .." << endl << endl;

									// this means that insert is at the end of the vector ..
									parentIndexPage->getSearchKeys()->push_back( childSearchKey );

									//cout << "IndexManager:: updateIndex() Point AAA .." << endl << endl;

									parentIndexPage->getChildPageIDs()->push_back( childPageID );

									//cout << "IndexManager:: updateIndex() Point AAB .." << endl << endl;

								}

							} // end "if()" ..


						} else {

							//cout << "IndexManager:: updateIndex() Point AAC .." << endl << endl;

							// this indicates that the "searchKey" appears in the parent node, but
							// that the "childPageID" does not appear in the parent node ..

							// need to add the "childPageID" to the parent node in the correct order ..

							int insertIndex = ( childSearchKeyPosition + 1 );

							//cout << "IndexManager:: updateIndex() Point AAD .." << endl << endl;

							vector<int>::iterator childIterator = parentIndexPage->getChildPageIDs()->begin();

							//cout << "IndexManager:: updateIndex() Point AAE .." << endl << endl;

							parentIndexPage->getChildPageIDs()->insert( ( childIterator + insertIndex ), childPageID );

							//cout << "IndexManager:: updateIndex() Point AAF .." << endl << endl;

						} // end "if()" ..


						//for(int i=0; i<parentIndexPage->getSearchKeys()->size(); i++){

						//	cout << "IndexManager:: updateIndex() parentIndexPage->getSearchKeys()->at(i("<<i<<")): " << parentIndexPage->getSearchKeys()->at(i) << endl << endl;

						//}

						//for(int i=0; i<parentIndexPage->getChildPageIDs()->size(); i++){

						//	cout << "IndexManager:: updateIndex() parentIndexPage->getChildPageIDs()->at(i("<<i<<")): " << parentIndexPage->getChildPageIDs()->at(i) << endl << endl;

						//}

						//cout << "IndexManager:: updateIndex() parentIndexPage->getIsLeafNode(): " << parentIndexPage->getIsLeafNode() << endl << endl;
						//cout << "IndexManager:: updateIndex() parentIndexPage->getPageID(): " << parentIndexPage->getPageID() << endl << endl;
						//cout << "IndexManager:: updateIndex() Point AAG .." << endl << endl;
						//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;
						//cout << "IndexManager:: updateIndex() parentParentPageID: " << parentParentPageID << endl << endl;

						// write the current indexPage ..
						bool isWriteLeftIndexPage = writeIndexPage( indexTableName_, indexName_, parentIndexPage, parentPageID, parentParentPageID, false, isNoDuplicates );

					} else {

						// childPageID found in the parent ..

						//cout << "IndexManager:: updateIndex() Point AAH .." << endl << endl;

						if( parentIndexPage->getChildPageIDs()->size() <= 1 ){

						//if( ( parentIndexPage->getChildPageIDs()->size() <= 1 ) &&
						//	( parentPageID == 0 ) ){

							//cout << "IndexManager:: updateIndex() Point AAI .." << endl << endl;

							// parent is the root node ..

							// check whether the dummy searchKey in the parent is still 1 greater
							// than the max searchKey in the child ..
							int parentSearchKey = parentIndexPage->getSearchKeys()->at(0);
							bool isParentSearchKeyGreater = true;
							int maxChildSearchKey = -1;
							int currentChildSearchKey = -1;
							int childSearchKeySize = indexPage->getSearchKeys()->size();
							for(int i=0; i<childSearchKeySize; i++){

								currentChildSearchKey = indexPage->getSearchKeys()->at(i);

								// check if root searchKey is no longer greater than child max ..
								if( parentSearchKey < currentChildSearchKey ){

									// update the new max child searchKey ..
									if( maxChildSearchKey < currentChildSearchKey ){

										maxChildSearchKey = currentChildSearchKey;

										if( isParentSearchKeyGreater == true )
											isParentSearchKeyGreater = false;

									}

								}

							} // end "for()" ..

							// replace the dummy searchKey value for a "parent-with-only-1-node" ..
							if ( isParentSearchKeyGreater == false ){

								// update "childSearchKey" with the latest max child searchKey ..
								// (parent searchKey is 1 greater than the max searchKey in the child) ..
								if( childSearchKey <= maxChildSearchKey )
									childSearchKey = ( maxChildSearchKey + 1 );

								parentIndexPage->getSearchKeys()->clear();
								parentIndexPage->getSearchKeys()->push_back( childSearchKey );
								//parentIndexPage->getChildPageIDs()->push_back( childPageID );

								// write the current indexPage ..
								bool isWriteLeftIndexPage = writeIndexPage( indexTableName_, indexName_, parentIndexPage, parentPageID, parentParentPageID, false, isNoDuplicates );

							}

						} // end "if()" ..

					} // end "if()" ..

					//cout << "IndexManager:: updateIndex() Point AAJ .." << endl << endl;


				} // end "if()" ..


				//cout << "IndexManager:: updateIndex() Point AAJ .." << endl << endl;

				// "split" the node if it is full prior to addition of new "searchKey" ..
				// handles when "searchKeys" of node-for-insert has exceeded capacity (5 > 4) ..

				//cout << "IndexManager:: updateIndex() ( isParentOverflow == true ): " << ( isParentOverflow == true ) << endl << endl;
				//cout << "IndexManager:: updateIndex() ( isLeafNode == true ): " << ( isLeafNode == true ) << endl << endl;
				//cout << "IndexManager:: updateIndex() ( searchKeysSize("<<searchKeysSize<<") > (fanout - 1)("<<(fanout - 1)<<") ): " << ( searchKeysSize > (fanout - 1) ) << endl << endl;

				if( ( isParentOverflow == true ) ||
					( ( searchKeysSize > (fanout - 1) ) && ( isLeafNode == true ) ) ) {
				//if ( searchKeysSize > (fanout - 1) ){

					// update the current pageID ..
					//pageID = indexPage->getPageID();

					//cout << "IndexManager:: updateIndex() Point AAK .." << endl << endl;

					// DEBUGGING ..
					//vector<int>* searchKeys_pnt = indexPage->getSearchKeys();
					//vector<int>* childPageIDs_pnt = indexPage->getChildPageIDs();

					//for(int i=0; i<searchKeys_pnt->size(); i++)
					//	cout << "IndexManager:: updateIndex() searchKeys_print->at(i("<<i<<")): " << searchKeys_pnt->at(i) << endl << endl;

					//for(int i=0; i<childPageIDs_pnt->size(); i++)
					//	cout << "IndexManager:: updateIndex() childPageIDs_print->at(i("<<i<<")): " << childPageIDs_pnt->at(i) << endl << endl;

					// split the "overflow" leaf node into 2 nodes, left and right ..
					IndexPage* leftIndexPage = new IndexPage();
					IndexPage* rightIndexPage = new IndexPage();

					// initialize the "isLeafNode" parameter ..
					if ( isLeafNode == true ) {

						leftIndexPage->setIsLeafNode( true );
						rightIndexPage->setIsLeafNode( true );

					} else {

						leftIndexPage->setIsLeafNode( false );
						rightIndexPage->setIsLeafNode( false );

					}

					// if "indexPage" is full, integerDivide(5, 2) = 2 ..
					int leftIndexSize = ( indexPage->getSearchKeys()->size() / 2 );

					//cout << "IndexManager:: updateIndex() indexPage->getSearchKeys()->size(): " << indexPage->getSearchKeys()->size() << endl << endl;
					//cout << "IndexManager:: updateIndex() (i<)leftIndexSize: " << leftIndexSize << endl << endl;

					// put half (first 2) of the full node into the "left" node ..
					for( int i=0; i<leftIndexSize; i++ ){

						leftIndexPage->getSearchKeys()->push_back( indexPage->getSearchKeys()->at(i) );

						leftIndexPage->getChildPageIDs()->push_back( indexPage->getChildPageIDs()->at(i) );

					}

					//cout << "IndexManager:: updateIndex() Point AAL .." << endl << endl;

					int middleSearchKey = -1;

					int rightIndexStart = -1;

					if ( isLeafNode == true ) {

						rightIndexStart = leftIndexSize;	// "2" ..

					} else {

						// always skip the 3rd searchKey of none-leaf nodes ..
						// (there will be 5 sks, pick first 2 to the left, last 2 to the right) ..
						rightIndexStart = (leftIndexSize+1);	// "3" ..

						// add the 3rd "childPageID" to the LEFT node ..
						leftIndexPage->getChildPageIDs()->push_back( indexPage->getChildPageIDs()->at( leftIndexSize ) );

						// track the "middleSearchKey" ..
						middleSearchKey = indexPage->getSearchKeys()->at( leftIndexSize );

					}

					//cout << "IndexManager:: updateIndex() Point AAM .." << endl << endl;
					//cout << "IndexManager:: updateIndex() (i=)rightIndexStart: " << rightIndexStart << endl << endl;
					//cout << "IndexManager:: updateIndex() (i<)indexPage->getSearchKeys()->size(): " << indexPage->getSearchKeys()->size() << endl << endl;
					//cout << "IndexManager:: updateIndex() (i<)indexPage->getChildPageIDs()->size(): " << indexPage->getChildPageIDs()->size() << endl << endl;


					// put the remaining half (last 3/2) into the "right" node ..
					// i = 2/3, i < 5 ..
					for( int i=rightIndexStart; i<indexPage->getSearchKeys()->size(); i++ ){

						// always skip the 3rd searchKey of none-leaf nodes ..
						// (there will be 5 sks, pick first 2 to the left, last 2 to the right) ..
						//if ( ( isLeafNode == true ) ||
						//	 ( ( isLeafNode == false ) && (i != 3) ) )
						rightIndexPage->getSearchKeys()->push_back( indexPage->getSearchKeys()->at(i) );

						rightIndexPage->getChildPageIDs()->push_back( indexPage->getChildPageIDs()->at(i) );

					}

					//cout << "IndexManager:: updateIndex() Point AAN .." << endl << endl;

					if ( isLeafNode == false ) {

						// add the 3rd "childPageID" to the right node ..
						// (6 total childPageIDs, 3 left, 3 right) ..
						rightIndexPage->getChildPageIDs()->push_back( indexPage->getChildPageIDs()->at( searchKeysSize ) );

					}

					//cout << "IndexManager:: updateIndex() Point AAO .." << endl << endl;

					// update the header to account for the adjusted left node and the new right node ..
					IndexHeader* header = new IndexHeader();
					bool isReadIndexHeader = readIndexHeader( indexTableName_, indexName_, header );

					int totalNumberOfIndexPages = header->getTotalNumberOfIndexPages();

					++totalNumberOfIndexPages;
					
					//header->setTotalNumberOfIndexPages( totalNumberOfIndexPages );

					//cout << "IndexManager:: updateIndex() totalNumberOfIndexPages: " << totalNumberOfIndexPages << endl << endl;

					// update the number of indexPages to accomodate the new addition ..
					//header.setTotalNumberOfIndexPages( ++totalNumberOfIndexPages );

					// track the "headerEntries" for the "leftIndexPage" ..
					vector<IndexHeaderEntry*>* headerEntries = header->getHeaderEntries();
					int headerEntryIndex = -1;
					int headerEntriesSize = headerEntries->size();
					for( int i=0; i<headerEntriesSize; i++ ){

						//cout << "IndexManager:: updateIndex() headerEntries->at(i("<<i<<"))->getPageID()("<<headerEntries->at(i)->getPageID()<<") == pageID("<<pageID<<"): " << ( headerEntries->at(i)->getPageID() == pageID ) << endl << endl;

						if( headerEntries->at(i)->getPageID() == pageID ){

							headerEntryIndex = i;
							break;

						}

					}

					//cout << "IndexManager:: updateIndex() Point AAP .." << endl << endl;
					//cout << "IndexManager:: updateIndex() headerEntryIndex: " << headerEntryIndex << endl << endl;

					//// update the header entry values for the "leftIndexPage" ..
					//// parent of "indexPage" should point to "leftIndexPage" (parentID unchanged)..
					//headerEntries->at( headerEntryIndex )->setNumberOfSearchKeys( leftIndexSize );
					//headerEntries->at( headerEntryIndex )->setNumberOfChildPageIDs( leftIndexSize );
					//headerEntries->at( headerEntryIndex )->setFreeSearchKeyCount( ( (fanout-1) - leftIndexSize ) );
					//int searchKeysAvailableSize = headerEntries->at( headerEntryIndex )->getSearchKeysAvailable()->size();
					//headerEntries->at( headerEntryIndex )->getSearchKeysAvailable()->at(0) = false;
					//headerEntries->at( headerEntryIndex )->getSearchKeysAvailable()->at(1) = false;
					//for(int i=2; i<searchKeysAvailableSize; i++)
					//	headerEntries->at( headerEntryIndex )->getSearchKeysAvailable()->at(i) = true;


					int leftIndexPageID = -1;

					//int parentPageID = -1;

					bool leftIndexPageIsNew = false;

					bool isNewRootNode = false;

					if ( isLeafNode == true ) {

						//cout << "IndexManager:: updateIndex() Point AAQ .." << endl << endl;

						// "leftIndexPage" has the "pageID" from "indexPage" ..
						leftIndexPageID = pageID;

						// parent for both "left" and "right" is the same ..
						// (if "headerEntryIndex == -1", parentPageID should have been set further above) ..
						if( headerEntryIndex >= 0 )
							parentPageID = headerEntries->at( headerEntryIndex )->getParentPageID();

						//cout << "IndexManager:: updateIndex() Point AAR .." << endl << endl;
						//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;
						//cout << "IndexManager:: updateIndex() leftIndexPageID: " << leftIndexPageID << endl << endl;

					} else {

						if ( pageID > 0 ){

							//cout << "IndexManager:: updateIndex() Point AAS .." << endl << endl;

							// "leftIndexPage" has the "pageID" from "indexPage" ..
							leftIndexPageID = pageID;

							// parent for both "left" and "right" is the same ..
							// (if "headerEntryIndex == -1", parentPageID should have been set further above) ..
							if( headerEntryIndex >= 0 )
								parentPageID = headerEntries->at( headerEntryIndex )->getParentPageID();

							//cout << "IndexManager:: updateIndex() Point AAT .." << endl << endl;
							//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;
							//cout << "IndexManager:: updateIndex() leftIndexPageID: " << leftIndexPageID << endl << endl;

						} else if( pageID == header->getRootPageID() ) {

							//cout << "IndexManager:: updateIndex() Point AAU .." << endl << endl;

							// this indicates that the "root" node was split ..

							isNewRootNode = true;

							// "pageID" of new indexPage is just the number of index pages minus 1 ..
							leftIndexPageID = ( totalNumberOfIndexPages - 1 );

							// parent needs to be updated ..
							//headerEntries->at( headerEntryIndex )->setParentPageID( pageID );

							leftIndexPageIsNew = true;

							// parent for both "left" and "right" is the same, the "root" node ..
							// ( leftIndexPageID is either "totalNumberOfIndexPages" or "( totalNumberOfIndexPages - 1 )" ) ..
							//parentPageID = pageID;
							parentPageID = (totalNumberOfIndexPages + 1);

							// root node was just split, so stop looping ..
							isTraversingParentNodes = false;

							//cout << "IndexManager:: updateIndex() Point AAV .." << endl << endl;
							//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;
							//cout << "IndexManager:: updateIndex() leftIndexPageID: " << leftIndexPageID << endl << endl;

						} // end "if()" ..

					} // end "if()" ..

					//cout << "IndexManager:: updateIndex() Point AAW .." << endl << endl;

					// "leftIndexPage" has the "pageID" just calculated above ..
					leftIndexPage->setPageID( leftIndexPageID );

					// "rightIndexPage" has a new "pageID" ..
					// "pageID" of new indexPage is just the number of index pages minus 1 ..
					int rightIndexPageID = ( totalNumberOfIndexPages - 1  );
					// "rightIndexPageID" should be 1 greater than "leftIndexPageID" ..
					if ( leftIndexPageID == rightIndexPageID )
						rightIndexPageID++;
					rightIndexPage->setPageID( rightIndexPageID );

					//cout << "IndexManager:: updateIndex() Point AAX .." << endl << endl;
					//cout << "IndexManager:: updateIndex() rightIndexPageID: " << rightIndexPageID << endl << endl;
					//cout << "IndexManager:: updateIndex() rightIndexPage->getPageID(): " << rightIndexPage->getPageID() << endl << endl;

					// update the new "left" and "right" indexPages ..
					bool isWriteLeftIndexPage = writeIndexPage( indexTableName_, indexName_, leftIndexPage, leftIndexPageID, parentPageID, leftIndexPageIsNew, isNoDuplicates );
					bool isWriteRightIndexPage = writeIndexPage( indexTableName_, indexName_, rightIndexPage, rightIndexPageID, parentPageID, true, isNoDuplicates );

					//cout << "IndexManager:: updateIndex() Point AAY .." << endl << endl;

					// update new "root" node if applicable ..
					if ( isNewRootNode ){

						//cout << "IndexManager:: updateIndex() Point AAZ .." << endl << endl;
						//cout << "IndexManager:: updateIndex() rightIndexPage->getPageID(): " << rightIndexPage->getPageID() << endl << endl;

						int rootSearchKey = -1;

						// find the searchKey for the root node ..
						int leftMostChildID = -1;
						IndexPage* childPage = new IndexPage();

						leftMostChildID = rightIndexPage->getChildPageIDs()->at(0);

						bool isLeafNode = false;

						// "drill" down to the correct leaf node ..
						do {

							bool isReadParentIndexPage = readIndexPage( indexTableName_, indexName_, childPage, leftMostChildID );

							leftMostChildID = childPage->getChildPageIDs()->at(0);

							isLeafNode = childPage->getIsLeafNode();

							if ( isLeafNode == true ){

								// the "searchKey" for the new root node is the leftmost leaf node searchKey ..
								rootSearchKey = childPage->getSearchKeys()->at(0);

							} else {

								childPage = new IndexPage();

							}

						} while( isLeafNode == false );

						//cout << "IndexManager:: updateIndex() Point AAAA .." << endl << endl;


						// create the root node ..
						IndexPage* rootPage = new IndexPage();
						int rootPageID = parentPageID;
						//cout << "IndexManager:: updateIndex() rootPageID: " << rootPageID << endl << endl;
						//cout << "IndexManager:: updateIndex() rootSearchKey: " << rootSearchKey << endl << endl;
						//cout << "IndexManager:: updateIndex() leftIndexPage->getPageID(): " << leftIndexPage->getPageID() << endl << endl;
						//cout << "IndexManager:: updateIndex() rightIndexPage->getPageID(): " << rightIndexPage->getPageID() << endl << endl;
						rootPage->setPageID( rootPageID );
						rootPage->getSearchKeys()->push_back( rootSearchKey );
						rootPage->getChildPageIDs()->push_back( leftIndexPage->getPageID() );
						rootPage->getChildPageIDs()->push_back( rightIndexPage->getPageID() );

						// write the new root node to file ..
						bool isWriteRightIndexPage = writeIndexPage( indexTableName_, indexName_, rootPage, rootPageID, -1, false, isNoDuplicates );

						//cout << "IndexManager:: updateIndex() Point - 5961 - AAAB .." << endl << endl;
						
						header->setRootPageID( rootPageID );
						
						//cout << "IndexManager:: updateIndex() rootPageID: " << rootPageID << endl << endl;
						
						//cout << "IndexManager:: updateIndex() Point AAAB1 .." << endl << endl;

						// also, write the updated IndexHeader ..
						bool isReadIndexHeader = writeIndexHeader( indexTableName_, indexName_, header );

						// done updating the index ..
						isTraversingParentNodes = false;
						
						//cout << "IndexManager:: updateIndex() Point AAAB2 .." << endl << endl;

					} else {

						// if not handling the new "root" node,
						// need to add "searchKey" and "childPageID" into parent node ..

						// update the parent for "leftIndexPage" with "rightIndexPage" ..
						//bool isReadParentIndexPage = true;
						if( parentPageID >= 0 )
							isTraversingParentNodes = readIndexPage( indexTableName_, indexName_, parentIndexPage, parentPageID );
						else
							isTraversingParentNodes = false;

						//cout << "IndexManager:: updateIndex() Point AAAC .." << endl << endl;
						//cout << "IndexManager:: updateIndex() rightIndexPage->getSearchKeys()->size(): " << rightIndexPage->getSearchKeys()->size() << endl << endl;
						//cout << "IndexManager:: updateIndex() rightIndexPage->getChildPageIDs()->size(): " << rightIndexPage->getChildPageIDs()->size() << endl << endl;
						//cout << "IndexManager:: updateIndex() rightIndexPageID: " << rightIndexPageID << endl << endl;
						//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;

						// "rightIndexPage" searchKey and pageID should be inserted into parent ..
						// ("middleSearchKey" is inserted into parent if the parent is non-leaf-node) ..
						// ("middleSearchKey" is set much above during "split" into right and left nodes) ..
						int rightSearchKey = -1;
						if( isLeafNode == false )
							rightSearchKey = middleSearchKey;
						else
							rightSearchKey = rightIndexPage->getSearchKeys()->at( 0 );

						int rightPageID = rightIndexPage->getPageID();

						//cout << "IndexManager:: updateIndex() rightSearchKey: " << rightSearchKey << endl << endl;
						//cout << "IndexManager:: updateIndex() rightPageID: " << rightPageID << endl << endl;


						if( parentIndexPage->getChildPageIDs()->size() <= 1 ){

							// this indicates that the dummy root-node searchKey is handled ..

							//cout << "IndexManager:: updateIndex() Point AAAD .." << endl << endl;

							// 2 nodes total - parent is the root-node or intermediate node,
							// child is the leaf-node ..
							// (replace the parent-node dummy searchKey with the left-most searchKey of the
							//  second child leaf-node) ..
							// (also, add the pageID for the new leaf-node) ..

							parentIndexPage->getSearchKeys()->clear();
							parentIndexPage->getSearchKeys()->push_back( rightSearchKey );
							parentIndexPage->getChildPageIDs()->push_back( rightPageID );


						} else {

							// this indicates that the dummy root-node searchKey is NOT handled ..

							//cout << "IndexManager:: updateIndex() Point AAAE .." << endl << endl;

							// add the new "searchKey" to the set ..
							// (indexPage is used for next iteration of loop) ..
							//indexPage = parentIndexPage;
							int parentSearchKeyIndex = -1;
							int searchKeysSize = parentIndexPage->getSearchKeys()->size();
							for(int i=0; i<searchKeysSize; i++){

								//cout << "IndexManager:: updateIndex() ( rightSearchKey("<<rightSearchKey<<") > parentIndexPage->getSearchKeys()->at(i("<<i<<"))("<<parentIndexPage->getSearchKeys()->at(i)<<") ): " << ( rightSearchKey > parentIndexPage->getSearchKeys()->at(i) ) << endl << endl;

								if( rightSearchKey > parentIndexPage->getSearchKeys()->at(i) ){

									continue;

								} else if( rightSearchKey < parentIndexPage->getSearchKeys()->at(i) ){

									// in general parentSearchKeyIndex is 1 less than the index found at i,
									// because at i, rightSearchKey > parentSearchKey ..
									parentSearchKeyIndex = i;
									break;

								}

							} // end "for()" ..


							//cout << "IndexManager:: updateIndex() Point AAAF .." << endl << endl;
							//cout << "IndexManager:: updateIndex() parentSearchKeyIndex: " << parentSearchKeyIndex << endl << endl;

							// add the rightSearchKey to the parent ..
							if( parentSearchKeyIndex >= 0 ){

								//cout << "IndexManager:: updateIndex() Point AAAG .." << endl << endl;

								vector<int>::iterator rightIterator = parentIndexPage->getSearchKeys()->begin();

								//cout << "IndexManager:: updateIndex() Point AAAH .." << endl << endl;

								parentIndexPage->getSearchKeys()->insert( ( rightIterator + parentSearchKeyIndex ), rightSearchKey );

								//cout << "IndexManager:: updateIndex() Point AAAI .." << endl << endl;

								// rightPageIDs has 1 element more than searchKeys, so adjust for this ..
								// (special case: if insert is at i == 0, childPageIDIndex == 0, too) ..
								int parentRightPageIDIndex = ( parentSearchKeyIndex + 1 );

								//cout << "IndexManager:: updateIndex() parentRightPageIDIndex: " << parentRightPageIDIndex << endl << endl;

								//cout << "IndexManager:: updateIndex() Point AAAJ .." << endl << endl;

								rightIterator = parentIndexPage->getChildPageIDs()->begin();

								//cout << "IndexManager:: updateIndex() Point AAAK .." << endl << endl;

								parentIndexPage->getChildPageIDs()->insert( ( rightIterator + parentRightPageIDIndex ), rightPageID );

								//cout << "IndexManager:: updateIndex() Point AAAL .." << endl << endl;

							} else {

								//cout << "IndexManager:: updateIndex() Point AAAM .." << endl << endl;

								// this means that insert is at the end of the vector ..
								parentIndexPage->getSearchKeys()->push_back( rightSearchKey );

								//cout << "IndexManager:: updateIndex() Point AAAN .." << endl << endl;

								parentIndexPage->getChildPageIDs()->push_back( rightPageID );

								//cout << "IndexManager:: updateIndex() Point AAAO .." << endl << endl;

							} // end "if()" ..

						} // end "if()" ..

						// get "parentParentPageID" from the "HeaderEntry" ..
						IndexHeader* indexHeader = new IndexHeader();
						bool isReadIndexHeader = readIndexHeader( indexTableName_, indexName_, indexHeader );
						int totalNumberOfIndexPages = indexHeader->getTotalNumberOfIndexPages();
						vector<IndexHeaderEntry*>* headerEntries = indexHeader->getHeaderEntries();

						int pageID = -1;
						int parentParentPageID = -1;

						for(int i=0; i<totalNumberOfIndexPages; i++){

							pageID = headerEntries->at(i)->getPageID();

							if( parentIndexPage->getPageID() == pageID ){

								parentParentPageID = headerEntries->at(i)->getParentPageID();
								break;

							}

						}

						// write the updated parent node to file ..
						bool isNew = false;
						bool isWriteRightIndexPage = writeIndexPage( indexTableName_, indexName_, parentIndexPage, parentPageID, parentParentPageID, isNew, isNoDuplicates );

						if( isTraversingParentNodes == true ){

							//cout << "IndexManager:: updateIndex() Point AAAP .." << endl << endl;

 							// locate the parent index page for insert for the next iteration ..
 							// if pageID == -1, insert
 							//bool isReadIndexPage = true;
 							//if( pageID > 0 )
							//	isReadIndexPage = readIndexPage( indexTableName_, indexName_, indexPage, parentPageID );
							//else
							//	isTraversingParentNodes = false;

							// DEBUGGING ..
							//vector<int>* searchKeys_pnt = indexPage->getSearchKeys();
							//vector<int>* childPageIDs_pnt = indexPage->getChildPageIDs();

							//for(int i=0; i<searchKeys_pnt->size(); i++)
							//	cout << "IndexManager:: updateIndex() searchKeys_print->at(i): " << searchKeys_pnt->at(i) << endl << endl;

							//for(int i=0; i<childPageIDs_pnt->size(); i++)
							//	cout << "IndexManager:: updateIndex() childPageIDs_print->at(i): " << childPageIDs_pnt->at(i) << endl << endl;

						} // end "if()" ..


						// next round works on the parent node ..
						indexPage = parentIndexPage;


					} // end "if()" ..
					
					//cout << "IndexManager:: updateIndex() Point AAAP1 .." << endl << endl;


					if ( isLeafNode == false ) {

						//cout << "IndexManager:: updateIndex() Point AAAQ .." << endl << endl;

						// need to update the children of both "left" and "right" nodes
						// to have accurate "parentPageID" values ..

						IndexPage* childIndexPage = new IndexPage();
						int childPageID = -1;
						bool isReadParentIndexPage = false;
						bool isWriteLeftIndexPage = false;

						// left node ..
						int leftIndexPageID = leftIndexPage->getPageID();
						vector<int>* leftChildPageIDs = leftIndexPage->getChildPageIDs();
						int leftSize = leftChildPageIDs->size();

						for(int i=0; i<leftSize; i++){

							childPageID = leftChildPageIDs->at(i);

							// read the childIndexPage, and update it with the new "parentPageID" ..
							bool isReadParentIndexPage = readIndexPage( indexTableName_, indexName_, childIndexPage, childPageID );

							bool isWriteLeftIndexPage = writeIndexPage( indexTableName_, indexName_, childIndexPage, childPageID, leftIndexPageID, false, isNoDuplicates );
																									 // IndexPage, pageID, parentPageID ..
						}

						//cout << "IndexManager:: updateIndex() Point AAAR .." << endl << endl;

						// reset these parameters ..
						childPageID = -1;
						isReadParentIndexPage = false;
						isWriteLeftIndexPage = false;

						// right node ..
						int rightIndexPageID = rightIndexPage->getPageID();
						vector<int>* rightChildPageIDs = rightIndexPage->getChildPageIDs();
						int rightSize = rightChildPageIDs->size();

						for(int i=0; i<rightSize; i++){

							childPageID = rightChildPageIDs->at(i);

							// read the childIndexPage, and update it with the new "parentPageID" ..
							bool isReadParentIndexPage = readIndexPage( indexTableName_, indexName_, childIndexPage, childPageID );

							bool isWriteLeftIndexPage = writeIndexPage( indexTableName_, indexName_, childIndexPage, childPageID, rightIndexPageID, false, isNoDuplicates );
																									 // IndexPage, pageID, parentPageID ..
						}

						//cout << "IndexManager:: updateIndex() Point AAAS .." << endl << endl;

					} // end "if()" ..
					
					//cout << "IndexManager:: updateIndex() Point AAAS1 .." << endl << endl;

					// reset this parameter for the next iteration ..
					if( parentIndexPage->getSearchKeys()->size() == 5 ){

						// add the new "searchKey" to the set ..
						// (indexPage is used for next iteration of loop) ..
						indexPage = parentIndexPage;
						pageID = parentPageID;
						parentPageID = parentParentPageID;

						//cout << "IndexManager:: updateIndex() Point AAAT .." << endl << endl;

						isParentOverflow = true;

					} else {
						// since the parent didn't overflow, stop looping ..
						isTraversingParentNodes = false;

						isParentOverflow = false;

						//cout << "IndexManager:: updateIndex() Point AAAU .." << endl << endl;

					}
					
					//cout << "IndexManager:: updateIndex() Point AAAU1 .." << endl << endl;

				} else {

					//cout << "IndexManager:: updateIndex() Point AAAV .." << endl << endl;

					// check whether the parent overflows due to addition,
					// and therefore needs to be updated via another iteration ..

					bool isReadParentIndexPage = false;

					//cout << "IndexManager:: updateIndex() parentPageID: " << parentPageID << endl << endl;

					parentIndexPage = new IndexPage();

					if( parentPageID >= 0 ){

						isReadParentIndexPage = readIndexPage( indexTableName_, indexName_, parentIndexPage, parentPageID );

						if( ( isParentOverflow == false ) &&
							( parentIndexPage->getSearchKeys()->size() == 5 ) ){

							//cout << "IndexManager:: updateIndex() Point AAAW .." << endl << endl;

							// "indexPage" searchKeys should be inserted into parent ..
							//int searchKey = indexPage->getSearchKeys()->at( 0 );
							//int childPageID = indexPage->getChildPageIDs()->at( 0 );

							//cout << "IndexManager:: updateIndex() Point AM3 .." << endl << endl;

							// add the new "searchKey" to the set ..
							// (indexPage is used for next iteration of loop) ..
							indexPage = parentIndexPage;
							pageID = parentPageID;
							parentPageID = parentParentPageID;
							//indexPage->getSearchKeys()->push_back( searchKey );
							//indexPage->getChildPageIDs()->push_back( childPageID );

							//cout << "IndexManager:: updateIndex() Point AAAX .." << endl << endl;

							isParentOverflow = true;


						} else {

							// since the parent didn't overflow, stop looping ..

							isTraversingParentNodes = false;

							isParentOverflow = false;

							//cout << "IndexManager:: updateIndex() Point AAAY .." << endl << endl;

						}


					} else {

						// since the parent wasn't read, stop looping ..

						isTraversingParentNodes = false;

						isParentOverflow = false;

						//cout << "IndexManager:: updateIndex() Point AAAZ .." << endl << endl;

					}

				} // end "if()" ..

				//cout << "IndexManager:: updateIndex() isTraversingParentNodes: " << isTraversingParentNodes << endl << endl;
				//cout << "IndexManager:: updateIndex() Point AAAAA .." << endl << endl;

			} else {
				
				// start of clause on line 4850 ..

				// this indicates that there are no index pages to read ..

				//cout << "IndexManager:: updateIndex() Point AAAAB .." << endl << endl;

				isTraversingParentNodes = false;

				isParentOverflow = false;

			} // end "if" ..

			//cout << "IndexManager:: updateIndex() Point AAAAC .." << endl << endl;

 		} // end "while" ..

 		//cout << "IndexManager:: updateIndex() - 6293 - Point AAAAD .." << endl << endl;
 		
 		
 		if ( isNoDuplicates == false ){
 			
 			//cout << "IndexManager:: updateIndex() - 6298 - Point AAAAE .." << endl << endl;
 			
 			// duplicates are allowed, so update the buckets ..
 			
			IndexBucketHeader* indexBucketHeader = new IndexBucketHeader();
			bool isReadIndexBucketHeader = readIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );

			// get the bucket id,
			// but if there isnt one yet,
			// find max bucket id for use in creating a new one ..
			int maxBucketID = -1;
			int bucketID = -1;
			int headerEntryIndex = -1;
			bool bucketIDFound = false;
			vector<IndexBucketHeaderEntry*>* headerEntries =  indexBucketHeader->getHeaderEntries();
			int totalNumberOfBuckets = indexBucketHeader->getTotalNumberBuckets();
			for(int i=0; i<totalNumberOfBuckets; i++){

				//cout << "IndexManager:: updateIndex() ( searchKey_("<<searchKey_<<") == headerEntries->at(i("<<i<<"))->getSearchKey()("<<headerEntries->at(i)->getSearchKey()<<") ): " << ( searchKey_ == headerEntries->at(i)->getSearchKey() ) << endl << endl;

				if( searchKey_ == headerEntries->at(i)->getSearchKey() ){

					// each bucket has max 12 slots ..
					if( headerEntries->at(i)->getFreePointerCount() > 0 ){

						bucketID = headerEntries->at(i)->getBucketID();
						headerEntryIndex = i;
						bucketIDFound = true;
						break;

					} else {

						// there may be more 12-slot buckets with the same searchKey
						// that have open slots (this one is full, however) ..
						continue;

					}

				}

			}

			if( bucketIDFound == false ){

				// the bucket doesn't exist yet, so find the max bucket id
				// for use in creating a new bucket ..

				for(int i=0; i<totalNumberOfBuckets; i++){

					if( maxBucketID < headerEntries->at(i)->getBucketID() )
						maxBucketID = headerEntries->at(i)->getBucketID();

				}

				// if a max bucket id was found,
				// the new bucket id is just 1 more than this ..
				if( maxBucketID >= 0 ){

					bucketID = ( maxBucketID + 1 );

				}

			}

			//cout << "IndexManager:: updateIndex() Point AAAAF .." << endl << endl;
			//cout << "IndexManager:: updateIndex() totalNumberOfBuckets: " << totalNumberOfBuckets << ", maxBucketID: " << maxBucketID << endl << endl;
			//cout << "IndexManager:: updateIndex() searchKey_: " << searchKey_ << ", bucketID: " << bucketID << endl << endl;

			IndexBucket* indexBucket = new IndexBucket();
			bool isReadIndexBucket = false;
			if( bucketIDFound == true )
				isReadIndexBucket = readIndexBucket( indexTableName_, indexName_, indexBucket, bucketID );

			indexBucket->getPageIDs()->push_back( childPageID_ );

			bool isWriteIndexBucket = writeIndexBucket( indexTableName_, indexName_, indexBucket, bucketID );

			//cout << "IndexManager:: updateIndex() Point AAAAG .." << endl << endl;

			// each bucket has 1-12 "pageIDs", track how many are left to fill ..
			int indexBucketPageIDsSize = indexBucket->getPageIDs()->size();
			int freePointerCount = ( ( blockSize / intSize ) - indexBucketPageIDsSize );

			//cout << "IndexManager:: updateIndex() freePointerCount: " << freePointerCount << endl << endl;

			if( bucketIDFound == false ){

				//cout << "IndexManager:: updateIndex() Point AAAAH .." << endl << endl;
	
				IndexBucketHeaderEntry* indexBucketHeaderEntry = new IndexBucketHeaderEntry();

				indexBucketHeaderEntry->setBucketID( bucketID );
				indexBucketHeaderEntry->setSearchKey( searchKey_ );
				indexBucketHeaderEntry->setFreePointerCount( freePointerCount );

				//cout << "IndexManager:: updateIndex() bucketID: " << bucketID << endl << endl;
				//cout << "IndexManager:: updateIndex() searchKey: " << searchKey_ << endl << endl;

				indexBucketHeader->getHeaderEntries()->push_back( indexBucketHeaderEntry );

			} else {

				//cout << "IndexManager:: updateIndex() Point AAAAI .." << endl << endl;
				//cout << "IndexManager:: updateIndex() bucketID: " << bucketID << endl << endl;
				//cout << "IndexManager:: updateIndex() searchKey: " << searchKey_ << endl << endl;

				indexBucketHeader->getHeaderEntries()->at( headerEntryIndex )->setBucketID( bucketID );
				indexBucketHeader->getHeaderEntries()->at( headerEntryIndex )->setSearchKey( searchKey_ );
				indexBucketHeader->getHeaderEntries()->at( headerEntryIndex )->setFreePointerCount( freePointerCount );

			}

			indexBucketHeader->setTotalNumberBuckets( indexBucketHeader->getHeaderEntries()->size() );

			// update the "IndexBucketHeader" file ..
			bool isWriteBucketHeader = writeIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );
 
 			
 			
 		}

		//cout << "IndexManager:: updateIndex() - 6402 - Point AAAAJ .." << endl << endl;

		// set "return" value ..
		isIndexUpdated = true;

	} else {

		// "searchKey" was found, so either insert a duplicate, or
		// reject the update if duplicates aren't allowed ..

		//cout << "IndexManager:: updateIndex() - 6412 - Point AAAAK .." << endl << endl;
		
		//cout << "IndexManager:: updateIndex() Search Key was Found .." << endl << endl;

		if ( isNoDuplicates == false ) {

			// duplicates are allowed ..

/*
			cout << "IndexManager:: updateIndex() Point AA .." << endl << endl;

			IndexBucketHeader indexBucketHeader;
			bool isReadIndexBucketHeader = readIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );

			int totalNumberOfBuckets = indexBucketHeader.getTotalNumberBuckets();

			cout << "IndexManager:: updateIndex() totalNumberOfBuckets: " << totalNumberOfBuckets << endl << endl;

			vector<IndexBucketHeaderEntry*>* headerEntries = indexBucketHeader.getHeaderEntries();

			int bucketID = -1;
			bool bucketIDFound = false;

			for(int i=0; i<totalNumberOfBuckets; i++){

				cout << "IndexManager:: updateIndex() ( searchKey_("<<searchKey_<<") == headerEntries->at(i("<<i<<")))->getSearchKey()("<<headerEntries->at(i)->getSearchKey()<<") ): " << ( searchKey_ == headerEntries->at(i)->getSearchKey() ) << endl << endl;

				if( searchKey_ == headerEntries->at(i)->getSearchKey() ){

					// found the bucket to insert into ..
					bucketID = headerEntries->at(i)->getBucketID();
					bucketIDFound = true;
					break;

				}

			}

			cout << "IndexManager:: updateIndex() Point AB .." << endl << endl;

			if( bucketIDFound == true ){

				// insert the pageID into the bucket ..
				IndexBucket* indexBucket = new IndexBucket();
				bool isReadIndexBucket = readIndexBucket( indexTableName_, indexName_, indexBucket, bucketID );

				cout << "IndexManager:: updateIndex() Point AC .." << endl << endl;

				indexBucket->getPageIDs()->push_back( childPageID_ );

				bool isWriteIndexBucket = writeIndexBucket( indexTableName_, indexName_, indexBucket, bucketID );

			}

*/

			//cout << "IndexManager:: updateIndex()- 6468 - Point AAAAL .." << endl << endl;

			IndexBucketHeader* indexBucketHeader = new IndexBucketHeader();

			bool isReadIndexBucketHeader = readIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );

			// get the bucket id,
			// but if there isnt one yet,
			// find max bucket id for use in creating a new one ..
			int maxBucketID = -1;
			int bucketID = -1;
			int headerEntryIndex = -1;
			bool bucketIDFound = false;
			vector<IndexBucketHeaderEntry*>* headerEntries =  indexBucketHeader->getHeaderEntries();
			int totalNumberOfBuckets = indexBucketHeader->getTotalNumberBuckets();
			for(int i=0; i<totalNumberOfBuckets; i++){

				//cout << "IndexManager:: updateIndex() ( searchKey_("<<searchKey_<<") == headerEntries->at(i("<<i<<"))->getSearchKey()("<<headerEntries->at(i)->getSearchKey()<<") ): " << ( searchKey_ == headerEntries->at(i)->getSearchKey() ) << endl << endl;

				if( searchKey_ == headerEntries->at(i)->getSearchKey() ){

					// each bucket has max 12 slots ..
					if( headerEntries->at(i)->getFreePointerCount() > 0 ){

						bucketID = headerEntries->at(i)->getBucketID();
						headerEntryIndex = i;
						bucketIDFound = true;
						break;

					} else {

						// there may be more 12-slot buckets with the same searchKey
						// that have open slots (this one is full, however) ..
						continue;

					}

				}

			}

			//cout << "IndexManager:: updateIndex() Point AAAAM .." << endl << endl;

			// the bucket doesn't exist yet, so find the max bucket id
			// for use in creating a new bucket ..
			if( bucketIDFound == false ){

				for(int i=0; i<totalNumberOfBuckets; i++){

					if( maxBucketID < headerEntries->at(i)->getBucketID() )
						maxBucketID = headerEntries->at(i)->getBucketID();

				}

				// if a max bucket id was found,
				// the new bucket id is just 1 more than this ..
				if( maxBucketID >= 0 ){

					bucketID = ( maxBucketID + 1 );

				}

			}

			//cout << "IndexManager:: updateIndex() Point AAAAN .." << endl << endl;
			//cout << "IndexManager:: updateIndex() maxBucketID: " << maxBucketID << endl << endl;
			//cout << "IndexManager:: updateIndex() bucketID: " << bucketID << endl << endl;

			IndexBucket* indexBucket = new IndexBucket();
			bool isReadIndexBucket = false;
			if( bucketIDFound == true )
				isReadIndexBucket = readIndexBucket( indexTableName_, indexName_, indexBucket, bucketID );

			indexBucket->getPageIDs()->push_back( childPageID_ );

			bool isWriteIndexBucket = writeIndexBucket( indexTableName_, indexName_, indexBucket, bucketID );

			//cout << "IndexManager:: updateIndex() Point AAAAO .." << endl << endl;

			// each bucket has 1-12 "pageIDs", track how many are left to fill ..
			int indexBucketPageIDsSize = indexBucket->getPageIDs()->size();
			int freePointerCount = ( ( blockSize / intSize ) - indexBucketPageIDsSize );

			if( bucketIDFound == false ){

				//cout << "IndexManager:: updateIndex() Point AAAAP .." << endl << endl;

				IndexBucketHeaderEntry* indexBucketHeaderEntry = new IndexBucketHeaderEntry();

				//cout << "IndexManager:: updateIndex() bucketID: " << bucketID << endl << endl;
				//cout << "IndexManager:: updateIndex() searchKey: " << searchKey_ << endl << endl;

				indexBucketHeaderEntry->setBucketID( bucketID );
				indexBucketHeaderEntry->setSearchKey( searchKey_ );
				indexBucketHeaderEntry->setFreePointerCount( freePointerCount );

				indexBucketHeader->getHeaderEntries()->push_back( indexBucketHeaderEntry );

			} else {

				//cout << "IndexManager:: updateIndex() Point AAAAQ .." << endl << endl;
				//cout << "IndexManager:: updateIndex() bucketID: " << bucketID << endl << endl;
				//cout << "IndexManager:: updateIndex() searchKey: " << searchKey_ << endl << endl;

				indexBucketHeader->getHeaderEntries()->at( headerEntryIndex )->setBucketID( bucketID );
				indexBucketHeader->getHeaderEntries()->at( headerEntryIndex )->setSearchKey( searchKey_ );
				indexBucketHeader->getHeaderEntries()->at( headerEntryIndex )->setFreePointerCount( freePointerCount );

			}

			indexBucketHeader->setTotalNumberBuckets( indexBucketHeader->getHeaderEntries()->size() );

			// update the "IndexBucketHeader" file ..
			bool isWriteBucketHeader = writeIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );

			// duplicate searchKey was INSERTED,
			// set the "return" value ..
			isIndexUpdated = true;

		} else {
			
			//cout << "IndexManager:: updateIndex() - 6606 - Point AAAAR .." << endl << endl;

			// duplicate searchKey REJECTED,
			// set the "return" value ..
			isIndexUpdated = false;

		} // end "if" ..

	} // end "if" ..


 	// if overflow, adjust indexPages B+ tree ..

 	// if duplicates, create index buckets ..

 	// write adjustments to file ..

 	// update the header files ..

 	// update the catalog file ..

 	int newIndexSize = calculateIndexSize( indexTableName_, indexName_ );
 	//int newIndexSize = getIndexSize( indexTableName_, indexName_ );

 	//cout << "IndexManager:: updateIndex() newIndexSize: " << newIndexSize << endl << endl;

 	//catalogIndex->setIndexSize( newIndexSize );

 	getCatalogManager()->getCatalogIndex( indexName_ )->setIndexSize( newIndexSize );

 	//catalogIndex.setIndexSize( newIndexSize );
 	//getCatalogManager()->setCatalogIndex( indexName_, catalogIndex );

 	getCatalogManager()->writeMetaData();

 	//cout << "IndexManager:: updateIndex() Point AAAAS .." << endl << endl;
 	
 	//cout << "IndexManager:: updateIndex() isIndexUpdated: " << isIndexUpdated << endl << endl;

 	return isIndexUpdated;

} // end "updateIndex()" ..


void IndexManager:: printIndex( string indexTableName_, string indexName_ ){

	//cout << "IndexManager:: printIndex() start .." << endl << endl;
	//cout << "IndexManager:: printIndex() indexTableName: " << indexTableName_ << endl << endl;
	//cout << "IndexManager:: printIndex() indexName: " << indexName_ << endl << endl;

	IndexPage* indexPage = NULL;

	bool isReadIndexPage = false;

	bool isTraverseIndex = true;

	bool isNoDuplicates = getCatalogManager()->getCatalogIndex( indexName_ )->getIsNoDuplicates();

	// start at the root node ..
	int rootPageID = 0;

	queue<int>* pageIDs = new queue<int>();

	pageIDs->push( rootPageID );

	while ( isTraverseIndex ) {

		//cout << "IndexManager:: printIndex() isTraverseIndex: " << isTraverseIndex << endl << endl;

		//cout << "IndexManager:: printIndex() Point A .." << endl << endl;
		//cout << "IndexManager:: printIndex() pageIDs->front(): " << pageIDs->front() << endl << endl;

		indexPage = new IndexPage();

		if( pageIDs->size() > 0 ) {

			if( pageIDs->front() >= 0 )
				isReadIndexPage = readIndexPage( indexTableName_, indexName_, indexPage, pageIDs->front() );
			else
				isTraverseIndex = false;

		} else {
			isTraverseIndex = false;
		}

		//cout << "IndexManager:: printIndex() Point B .." << endl << endl;
		//cout << "IndexManager:: printIndex() isReadIndexPage: " << isReadIndexPage << endl << endl;
		//cout << "IndexManager:: printIndex() isTraverseIndex: " << isTraverseIndex << endl << endl;

		//cout << "IndexManager:: printIndex() indexPage->getIsLeafNode(): " << indexPage->getIsLeafNode() << endl << endl;

		if ( isReadIndexPage == true ) {

			cout << "  pointers: ";

			cout << "index page id: " << indexPage->getPageID() << ", type: ";

			if(	indexPage->getPageID() == 0 )
				cout << "root" << endl;
			else if( indexPage->getIsLeafNode() == false )
				cout << "internal node" << endl;
			else if( indexPage->getIsLeafNode() == true )
				cout << "leaf node" << endl;

			cout << "  keys: ";
			vector<int>* searchKeys = indexPage->getSearchKeys();
			int searchKeysSize = searchKeys->size();
			for(int i=0; i<searchKeysSize; i++){

				cout << searchKeys->at(i);

				if( (i+1) < searchKeysSize )
					cout << ", ";

			} // end "for()" ..

			cout << "  pointers: ";

			// handle any node that isn't a duplicate, or 
			// intermediate nodes ..
			if ( ( isNoDuplicates == true ) ||
				 ( indexPage->getIsLeafNode() == false ) ) {

				vector<int>* childPageIDs = indexPage->getChildPageIDs();
				int childPageIDSize = childPageIDs->size();
				for(int i=0; i<childPageIDSize; i++){

					//if( childPageIDs->at(i) >= 0 )
					cout << childPageIDs->at(i);

					if( (i+1) < childPageIDSize )
						cout << ", ";

				}

				// pop off the "front" element of the queue ..
				pageIDs->pop();

				// add "childPageIDs" for non-leaf nodes ..
				if( indexPage->getIsLeafNode() == false ){
					
					// store the nodes to traverse ..
					for(int i=0; i<childPageIDSize; i++){

						pageIDs->push( childPageIDs->at(i) );

					}	
					
				}

			// handle leaf-nodes that are duplicates ..
			} else {

				IndexBucketHeader* indexBucketHeader = new IndexBucketHeader();

				bool isReadIndexBucketHeader = readIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );

				vector<IndexBucketHeaderEntry*>* ibhe = indexBucketHeader->getHeaderEntries();

				int numberOfEntries = indexBucketHeader->getTotalNumberBuckets();

				// check each "bucketHeaderEntry" for bucketIDs that
				// match the searchKeys for this node ..
				int bucketID = -1;
				vector<int>* searchKeys = indexPage->getSearchKeys();
				int searchKeysSize = searchKeys->size();
				vector<int>* bucketIDs = new vector<int>();

				// only output the bucketIDs that match the
				// searchKeys for this node ..
				for(int i=0; i<searchKeysSize; i++){

					// matched "bucketHeaderEntry"-searchKey with node-searchKey ..
					// (load "bucketIDs" with another bucketID) ..
					for(int j=0; j<numberOfEntries; j++){

						bucketID = ibhe->at(j)->getBucketID();

						//cout << "IndexManager:: printIndex() ( ibhe->at(j("<<j<<"))->getSearchKey()("<<ibhe->at(j)->getSearchKey()<<") == searchKeys->at(i("<<i<<"))("<<searchKeys->at(i)<<") ): " << ( ibhe->at(j)->getSearchKey() == searchKeys->at(i) ) << endl << endl;

						if( ibhe->at(j)->getSearchKey() == searchKeys->at(i) ){

							// matched "bucketHeaderEntry"-searchKey with node-searchKey ..
							// (load "bucketIDs" with another bucketID) ..

							// (note: order of bucketIDs depends on order of "searchKeys") ..

							bucketIDs->push_back( bucketID );

						}

					} // end "for()" ..

				} // end "for()" ..

				// "numberOfEntries" becomes the number of "bucketIDs" ..
				numberOfEntries = bucketIDs->size();

				// print out the bucketIDs ..
				int bucketIDSize = bucketIDs->size();
				for(int i=0; i<bucketIDSize; i++){

					cout << bucketIDs->at(i);

					if( (i+1) < numberOfEntries )
						cout << ", ";

				}

				// pop off the "front" element of the queue ..
				pageIDs->pop();

			} // end "if()" ..
			cout << endl << endl;
			cout << "=======================================" << endl << endl;

			// reset this parameter ..
			isReadIndexPage = false;

		} else {

			//cout << "IndexManager:: printIndex() Point C .." << endl << endl;

			// the "indexPage" failed to be read ..
			isTraverseIndex = false;

		} // end "if()" ..

	} // end "while()" ..

} // end "printIndex()" ..


vector<int>* IndexManager:: lookUpLeafNodeID( int searchKey_, string indexTableName_, string indexName_, bool &searchKeyFound_ ){

	// if the leaf node is found,
	// this method returns a pagePageID and pageID ..
	// else it returns empty collection ..

	//cout << "IndexManager:: lookUpLeafNodeID() start .." << endl << endl;
	//cout << "IndexManager:: lookUpLeafNodeID() searchKey: " << searchKey_ << endl << endl;

	// traverse the index, find the correct pageID ..
	// start with the "root" node, pageID = 0 ..
	IndexPage* indexPage = NULL;

	searchKeyFound_ = false;

	bool isTraverseIndex = true;

	bool isReadIndexPage = false;

	// start at the "root" node ..
	vector<int>* returnPageIDs = new vector<int>();

	int parentPageID = 0;

	int pageID = 0;
	
	// pageID of the root node should have been stored in the IndexHeader ..
	IndexHeader* indexHeader = new IndexHeader();
	
	bool isReadIndexHeader = readIndexHeader( indexTableName_, indexName_, indexHeader );

	if( isReadIndexHeader == true ){
		
		pageID = indexHeader->getRootPageID();
		
	}

	//int counter = 0;

	//cout << "IndexManager:: lookUpLeafNodeID() Point A .." << endl << endl;

	while ( isTraverseIndex ) {

		//cout << "IndexManager:: lookUpLeafNodeID() Point B .." << endl << endl;

		//cout << "IndexManager:: lookUpLeafNodeID() pageID: " << pageID << endl << endl;

		// stop looping ..
		if( pageID < 0 )
			break;

		indexPage = new IndexPage();

		isReadIndexPage = readIndexPage( indexTableName_, indexName_, indexPage, pageID );

		if( isIndexDebug ){

			cout << "DEBUG: bucket: " << indexPage->getPageID() << endl;

		}

		//cout << "IndexManager:: lookUpLeafNodeID() Point C .." << endl << endl;
		//cout << "IndexManager:: lookUpLeafNodeID() isReadIndexPage: " << isReadIndexPage << endl << endl;

		if ( ( isReadIndexPage == true ) &&
			 ( pageID >= 0 ) ) {

			vector<int>* searchKeys = indexPage->getSearchKeys();
			int searchKeysSize = searchKeys->size();

			//cout << "IndexManager:: lookUpLeafNodeID() searchKeysSize: " << searchKeysSize << endl << endl;

			// stop looping if there aren't any search keys ..
			if( searchKeysSize == 0 )
				break;

			for(int i=0; i<searchKeysSize; i++){

				//cout << "IndexManager:: lookUpLeafNodeID() Point D .." << endl << endl;

				// "pageID",
				// either used above in "readIndexPage()"
				// or it contains the final pageID for this searchKey ..

				//cout << "IndexManager:: lookUpLeafNodeID() pageID: " << pageID << endl << endl;
				//cout << "IndexManager:: lookUpLeafNodeID() indexPage->getIsLeafNode(): " << indexPage->getIsLeafNode() << endl << endl;

				//cout << "IndexManager:: lookUpLeafNodeID()  ( searchKey_("<<searchKey_<<") < searchKeys->at(i("<<i<<"))("<<searchKeys->at(i)<<") ): " << ( searchKey_ < searchKeys->at(i) ) << endl << endl;
				//cout << "IndexManager:: lookUpLeafNodeID() ( searchKey_("<<searchKey_<<") == searchKeys->at(i("<<i<<"))("<<searchKeys->at(i)<<") ): " << ( searchKey_ == searchKeys->at(i) ) << endl << endl;
				//cout << "IndexManager:: lookUpLeafNodeID()  ( searchKey_("<<searchKey_<<") > searchKeys->at(i("<<i<<"))("<<searchKeys->at(i)<<") ): " << ( searchKey_ > searchKeys->at(i) ) << endl << endl;

				if ( searchKey_ < searchKeys->at(i) ) {

					//cout << "IndexManager:: lookUpLeafNodeID() Point E .." << endl << endl;


					if( indexPage->getIsLeafNode() ) {

						//cout << "IndexManager:: lookUpLeafNodeID() Point F .." << endl << endl;

						// leaf-node for insert is located ..
						// if leaf node is full, and
						// if it is on the first searchKey,
						// need a new node (pageID = -1),
						// else returns the current value of pageID ..
						if( ( searchKeysSize == (fanout - 1) ) &&
							( i == 0 ) )
							pageID = -1;

						isTraverseIndex = false;

					} else {

						// for non-leaf-node, left childPageID points to
						// next indexPage to search ..
						// else, need to allocate a new indexPage ..

						//cout << "IndexManager:: lookUpLeafNodeID() Point G .." << endl << endl;

						// parent ID is the pageID of the previous indexPage ..
						parentPageID = pageID;

						if( indexPage->getChildPageIDs()->size() > 0 ){
							//cout << "IndexManager:: lookUpLeafNodeID() Point H .." << endl << endl;
							pageID = indexPage->getChildPageIDs()->at(i);

						} else {
							//cout << "IndexManager:: lookUpLeafNodeID() Point I .." << endl << endl;
							pageID = -1;

						}

					}

					break;

				} else if( searchKey_ == searchKeys->at(i) ){

					//cout << "IndexManager:: lookUpLeafNodeID() Point J .." << endl << endl;

					// searchKey found ..

					if( indexPage->getIsLeafNode() ){

						//cout << "IndexManager:: lookUpLeafNodeID() Point K .." << endl << endl;

						// parent ID is the pageID of the previous indexPage ..
						//parentPageID = pageID;

						// the left childPageID points to found dataPage ..
						//if( indexPage->getChildPageIDs()->size() > 0 )
						//	pageID = indexPage->getChildPageIDs()->at(i);
						//else
						//	pageID = -1;

						// the left childPageID is the pageID for this searchKey ..
						// found, so stop traversing the index ..
						isTraverseIndex = false;
						searchKeyFound_ = true;

					} else {

						//cout << "IndexManager:: lookUpLeafNodeID() Point L .." << endl << endl;

						// parent ID is the pageID of the previous indexPage ..
						parentPageID = pageID;

						// keep traversing the index ..
						// (pageID should be for the "right" branch) ..
						// else, need to allocate a new pageID ..

						if( indexPage->getChildPageIDs()->size() > (i+1) ){
							//cout << "IndexManager:: lookUpLeafNodeID() Point M .." << endl << endl;
							pageID = indexPage->getChildPageIDs()->at(i+1);
						} else {
							//cout << "IndexManager:: lookUpLeafNodeID() Point N .." << endl << endl;
							pageID = -1;
						}

					}

					break;

				} else if ( searchKey_ > searchKeys->at(i) ){

					//cout << "IndexManager:: lookUpLeafNodeID() Point O .." << endl << endl;

					//// parent ID is the pageID of the previous indexPage ..
					////parentPageID = pageID;

					// if the entire node is searched,
					// then the far right child page id points
					// to the next node to check ..
					if( i == (searchKeysSize - 1) ){

						//cout << "IndexManager:: lookUpLeafNodeID() Point P .." << endl << endl;

						if( indexPage->getIsLeafNode() ) {
							
							//cout << "IndexManager:: lookUpLeafNodeID() Point Q .." << endl << endl;
							//cout << "IndexManager:: lookUpLeafNodeID() ( searchKeysSize("<<searchKeysSize<<") == (fanout - 1)("<<(fanout - 1)<<") ): " << ( searchKeysSize == (fanout - 1) ) << endl << endl;

							// if leaf node is full, need a new node,
							// else returns this pageID ..
							if( searchKeysSize == (fanout - 1) )
								pageID = -1;

							isTraverseIndex = false;

						} else {
							
							//cout << "IndexManager:: lookUpLeafNodeID() Point R .." << endl << endl;

							IndexPage* childIndexPage = new IndexPage();
							int childPageID = indexPage->getChildPageIDs()->at(0);
							bool isReadIndexPage = readIndexPage( indexTableName_, indexName_, childIndexPage, childPageID );

							// number of child nodes for this parent node ..
							int childPageIDsSize = indexPage->getChildPageIDs()->size();

							//cout << "IndexManager:: lookUpLeafNodeID() Point R1 .." << endl << endl;

							int parentSearchKey = indexPage->getSearchKeys()->at(0);
							int childIndexPageSize = childIndexPage->getSearchKeys()->size();
							int farRightChildSearchKey = childIndexPage->getSearchKeys()->at( childIndexPageSize - 1 );

							//cout << "IndexManager:: lookUpLeafNodeID() ( searchKeysSize("<<searchKeysSize<<") == 1 ): " << ( searchKeysSize == 1 ) << endl << endl;
							//cout << "IndexManager:: lookUpLeafNodeID() ( childPageIDsSize("<<childPageIDsSize<<") == 1 ): " << ( childPageIDsSize == 1 ) << endl << endl;
							//cout << "IndexManager:: lookUpLeafNodeID() ( isReadIndexPage == true ): " << ( isReadIndexPage == true ) << endl << endl;
							//cout << "IndexManager:: lookUpLeafNodeID() ( childIndexPage->getIsLeafNode() == true ): " << ( childIndexPage->getIsLeafNode() == true ) << endl << endl;
							//cout << "IndexManager:: lookUpLeafNodeID() ( parentSearchKey("<<parentSearchKey<<") == ( farRightChildSearchKey("<<farRightChildSearchKey<<") + 1 ) ): " << ( parentSearchKey == ( farRightChildSearchKey + 1 ) ) << endl << endl;

							// UPDATE INDEX: handle when the FIRST index page is being filled ..
							// ("create index" performed WITHOUT records, followed by new inserts) ..
							// (the 1st index node to be filled has a parent node,
							//  this parent node has 1 searchKey,
							//  this parent node searchKey is always 1 greater than the max searchKey in this single child leaf node,
							//  also, no matter the value of the "searchKey-for-insert",
							//  at this stage it always goes into the first index page) ..

							// CREATE INDEX: handle when the NEW index page is being filled ..
							// ("create index" performed on EXISTING records, followed by new inserts) ..
							// (root node has 5 child nodes, an additional searchKey is inserted,
							//	root node is not split, instead a new root node is made, new parent
							//	node with searchKey 1 greater than inserted searchKey also made) ..

							// "Dummy" parent node always has just 1 searchKey, 1 childPageID ..

							if( ( searchKeysSize == 1 ) &&
								( childPageIDsSize == 1 ) &&
								( isReadIndexPage == true ) &&
								( childIndexPage->getIsLeafNode() == true ) &&
								( parentSearchKey == ( farRightChildSearchKey + 1 ) ) ){

								//cout << "IndexManager:: lookUpLeafNodeID() Point R2 .." << endl << endl;

								// parentPageID is the pageID of the previous indexPage ..
								// (corresponds to "create index" above) ..
								if( pageID > 0 )
									parentPageID = pageID;

								// check whether the child index node is full yet ..
								int childSearchKeySize = childIndexPage->getSearchKeys()->size();

								// if the child node is not yet full,
								// it is eligible for another searchKey ..
								// (else if first, left leaf-node is filled, start filling next leaf-node) ..
								// (else, pageID == -1 and a new indexPage will be created) ..
								// (note: i == (searchKeysSize - 1) from above,
								//  searchKeysSize == 1 from above, so i == 0) ..
								if( childSearchKeySize < (fanout - 1) ){

									//cout << "IndexManager:: lookUpLeafNodeID() Point R3 .." << endl << endl;

									// still need to fill first left-most node ..
									pageID = indexPage->getChildPageIDs()->at(i);

								} else if( indexPage->getChildPageIDs()->size() > 1 ) {

									//cout << "IndexManager:: lookUpLeafNodeID() Point R4 .." << endl << endl;

									// so long as right branch exists, use it ..
									pageID = indexPage->getChildPageIDs()->at(i+1);

								} else {

									//cout << "IndexManager:: lookUpLeafNodeID() Point R5 .." << endl << endl;

									// need new node to put searchKey into ..
									pageID = -1;

								}


							// handle "non-first" index pages ..
							} else {

								//cout << "IndexManager:: lookUpLeafNodeID() Point R6 .." << endl << endl;

								// parent ID is the pageID of the previous indexPage ..
								if( pageID > 0 )
									parentPageID = pageID;

								// searchKey may be next in series ..
								// (pageID should be for the "right" branch) ..

								//cout << "indexPage->getChildPageIDs()->size()("<<indexPage->getChildPageIDs()->size()<<") > (i+1)("<<(i+1)<<"): " << ( indexPage->getChildPageIDs()->size() > (i+1) ) << endl << endl;

								if( indexPage->getChildPageIDs()->size() > (i+1) ){
									
									//cout << "IndexManager:: lookUpLeafNodeID() Point S .." << endl << endl;
									pageID = indexPage->getChildPageIDs()->at( (i+1) );

									//cout << "IndexManager:: lookUpLeafNodeID() pageID(i+1)("<<(i+1)<<"): " << pageID << endl << endl;
									//cout << "IndexManager:: lookUpLeafNodeID() indexPage->getChildPageIDs()->size(): " << indexPage->getChildPageIDs()->size() << endl << endl;
									//for(int i=0; i<indexPage->getChildPageIDs()->size(); i++)
									//	cout << "IndexManager:: lookUpLeafNodeID() indexPage->getChildPageIDs()->at(i("<<i<<")): " << indexPage->getChildPageIDs()->at(i) << endl << endl;

								} else {
									//cout << "IndexManager:: lookUpLeafNodeID() Point T .." << endl << endl;
									pageID = -1;
								}

							} // end "if()" ..

						} // end "if()" ..

					} // end "if()" ..


					//cout << "IndexManager:: lookUpLeafNodeID() Point U .." << endl << endl;

					continue;

				} // end "if()" ..


			} // end "for()" ..


		} else {

			//cout << "IndexManager:: lookUpLeafNodeID() Point V .." << endl << endl;

			//pageID = -1;

			isTraverseIndex = false;

		} // end "if()" ..


	} // end "while()" ..

	//cout << "IndexManager:: lookUpLeafNodeID() parentPageID: " << parentPageID << endl << endl;
	//cout << "IndexManager:: lookUpLeafNodeID() pageID: " << pageID << endl << endl;

	returnPageIDs->push_back( parentPageID );
	returnPageIDs->push_back( pageID );

	return returnPageIDs;

} // end "lookUpLeafNodeID()" ..


//vector<int>* IndexManager:: lookUpIDs( int searchKey_, char* indexTableName_, char* indexName_, char* attributeName_ ){
vector<int>* IndexManager:: lookUpIDs( string indexTableName_, string indexName_, int searchKey_ ){

	// if there are _no_ duplicates,
	// this method returns a collection with single page id ..
	// if there are duplicates,
	// this method returns a collection of bucket ids ..

	//cout << "IndexManager:: lookUpIDs() start .." << endl << endl;

	bool searchKeyFound = false;

	vector<int>* returnPageIDs = lookUpLeafNodeID( searchKey_, indexTableName_, indexName_, searchKeyFound );
	int pageID = returnPageIDs->at(1);

	//cout << "IndexManager:: lookUpIDs() Point A .." << endl << endl;

	// correct IndexPage "pageID"(s) should have been found above ..

	// return the "pageIDs" found in the index ..
	vector<int>* pageIDs = new vector<int>();

	if ( searchKeyFound == true ) {

		//cout << "IndexManager:: lookUpIDs() Point B .." << endl << endl;

		// a) retrieve Index metadata from the Catalog ..
		getCatalogManager()->readMetaData();

		string indexTableName = indexTableName_;
		string indexName = indexName_;
		//string attributeName = attributeName_;

		//CatalogIndex* catalogIndex = getCatalogManager()->getCatalogIndex_attributeName( indexTableName, attributeName );
		CatalogIndex* catalogIndex = getCatalogManager()->getCatalogIndex_indexName( indexTableName, indexName );

		//cout << "IndexManager:: lookUpIDs() Point C .." << endl << endl;

		// using the IndexPage pageID, need to retrieve the dataPage pageID ..
		IndexPage* indexPage = new IndexPage();
		bool isReadIndexPage =  readIndexPage( indexTableName, indexName, indexPage, pageID );
		int dataPagePageID = -1;
		if( indexPage != NULL ){
				
			dataPagePageID = indexPage->getChildPageID( searchKey_ );
				
		}
			
		//cout << "IndexManager:: lookUpIDs() dataPagePageID: " << dataPagePageID << endl << endl;

		bool isNoDuplicates = catalogIndex->getIsNoDuplicates();

		if( isNoDuplicates == true ){

			//cout << "IndexManager:: lookUpIDs() Point D .." << endl << endl;

			// a single pageID has the searchKey value ..
			if( dataPagePageID >= 0 )
				pageIDs->push_back( dataPagePageID );

		} else {

			//cout << "IndexManager:: lookUpIDs() Point E .." << endl << endl;

			// go through the buckets,
			// get out the pageIDs and
			// the bucketID for the next bucket ..

			IndexBucket* indexBucket = new IndexBucket();
			vector<int>* bucketPageIDs = NULL;

			bool areBucketsLeft = true;

			int currentBucketSize = -1;
			int fullBucketSize = (blockSize / intSize);		// typ 12 ..
			int numberOfPageIDs = -1;

			int nextBucketID = -1;

			// a single pageID has the searchKey value ..
			if( dataPagePageID >= 0 )
				nextBucketID = dataPagePageID;

			while( areBucketsLeft && 
				   ( nextBucketID > 0 ) ){

				//if( isIndexDebug ){
				//
				//	cout << "DEBUG: index bucket id: " << nextBucketID << endl;
				//
				//}

				bool isReadIndexBucket = readIndexBucket( indexTableName_, indexName_, indexBucket, nextBucketID );

				//cout << "IndexManager:: lookUpIDs() Point F .." << endl << endl;

				bucketPageIDs = indexBucket->getPageIDs();
				currentBucketSize = bucketPageIDs->size();

				if( currentBucketSize == fullBucketSize ){

					//cout << "IndexManager:: lookUpIDs() Point G .." << endl << endl;

					// this indicates that the bucket is full, and
					// the last "pageID" is really a "bucketID" pointing to the next bucket in the chain ..

					numberOfPageIDs = (currentBucketSize - 1);

					nextBucketID = bucketPageIDs->at( (currentBucketSize - 1) );

				} else {

					//cout << "IndexManager:: lookUpIDs() Point H .." << endl << endl;

					// this indicates that the bucket is less than full, and
					// that the final "pageID" value is NOT a "bucketID" ..

					numberOfPageIDs = currentBucketSize;

					areBucketsLeft = false;

				}

				// get out all the "pageIDs" from this bucket ..
				for(int i=0; i<numberOfPageIDs; i++){

					pageIDs->push_back( bucketPageIDs->at(i) );

				}

				//cout << "IndexManager:: lookUpIDs() Point I .." << endl << endl;

			}

/*
			// SLOWER METHOD:
			// turns out that the header only has the 1st bucketID for each chain of buckets!
			// so .. i don't think that this works!
			// use the header to get all the bucketIDs ..
			IndexBucketHeader indexBucketHeader;
			bool isIndexBucketHeaderRead = readIndexBucketHeader( indexTableName_, indexName_, indexBucketHeader );

			vector<IndexBucketHeaderEntry*>* headerEntries = indexBucketHeader.getHeaderEntries();

			vector<int>* bucketIDs = new vector<int>();

			for(int i=0; i<headerEntries->size(); i++){

				if( headerEntries->at(i)->getSearchKey() == searchKey_ ){

					bucketIDs->push_back( headerEntries->at(i)->getBucketID() );

				}

			}

			// go through all the buckets using the bucketIDs,
			// get all the pageIDs ..
			IndexBucket* indexBucket = NULL;
			bool isReadIndexBucket = false;
			vector<int>* bucketPageIDs = new vector<int>();
			for(int i=0; i<bucketIDs->size(); i++){

				indexBucket = new IndexBucket();

				isReadIndexBucket = readIndexBucket( indexTableName_, indexName_, indexBucket, bucketIDs->at(i) );

				bucketPageIDs = indexBucket->getPageIDs();

				for(int j=0; j<bucketPageIDs->size(); j++)
					pageIDs->push_back( bucketPageIDs->at(j) );

			}

*/

		} // end "if()" ..

	} // end "if()" ..

	return pageIDs;

} // end "lookUpIDs()" ..


vector<int>* IndexManager:: lookUpIDsForIndices( string indexTableName_, vector<string>* indexNames_ ){

	// compare columns in the table (but only the columns that have indices)
	// and find values which "match" or are "equal" across _all_ the given columns ..
	// return the pageIDs that correspond to these values ..

	int indexNamesSize = indexNames_->size();

	IndexHeader** header = new IndexHeader*[ indexNamesSize ];
	
	for(int i=0; i<indexNamesSize; i++){
		
		header[i] = new IndexHeader();
		
	}

	bool isReadIndexHeader = false;

	for(int i=0; i<indexNamesSize; i++){

		isReadIndexHeader = readIndexHeader( indexTableName_, indexNames_->at(i), header[i] );

		if( isReadIndexHeader == false )
			break;

	}

	// if *all* headers weren't read, return prematurely and return null ..
	if( isReadIndexHeader == false )
		return NULL;

	//header[i].getHeaderEntries();

	vector<IndexHeaderEntry*>* headerEntries = NULL;
	int headerEntriesSize = -1;
	int headerEntryPageID = -1;
	IndexPage* indexPage;

	bool wasIndexPageRead = false;
	bool isDuplicatesAllowed = false;

	for(int i=0; i<indexNamesSize; i++){
		
		// check to see if the index allows duplicates ..
		isDuplicatesAllowed = header[i]->getIsDuplicatesAllowed();

		headerEntries = header[i]->getHeaderEntries();

		headerEntriesSize = headerEntries->size();

		// write each "index page" ..
		//for(int j=0; j<indexPagesSize; j++){
		for(int j=0; j<headerEntriesSize; j++){

			indexPage = new IndexPage();

			headerEntryPageID = headerEntries->at(j)->getPageID();

			// read this index page from file ..
			//wasIndexPageRead = readIndexPage( indexTableName_, indexName_, indexPage, j );
			wasIndexPageRead = readIndexPage( indexTableName_, indexNames_->at(i), indexPage, headerEntryPageID );

			// compare indexPage value to other values ..

			if( isDuplicatesAllowed == true ){

				// in order to get the "searchKey" stored by the index,
				// must retrieve 1 bucket and get its first "searchKey" value ..



				//bool isReadIndexBucket = readIndexBucket( char* tableName_, char* indexName_, IndexBucket* indexBucket_, int bucketPageID_ );

			}


		}

	}

	return NULL;


}


int IndexManager:: lookUpIDForInsert( string indexTableName_, string indexName_, int searchKey_ ){
	
	//cout << "IndexManager:: lookUpIDForInsert() start .. " << endl << endl;
	
	bool isIDFound = true;
	
	// default root page id ..
	int rootPageID = 0;
	
	// get the "rootPageID" from the "IndexHeader" ..
	IndexHeader* indexHeader = new IndexHeader();
	
	bool isReadIndexHeader = readIndexHeader( indexTableName_, indexName_, indexHeader );

	if( isReadIndexHeader == true ){
		
		rootPageID = indexHeader->getRootPageID();
		
	}
	
	//cout << "IndexManager:: lookUpIDForInsert() rootPageID: " << rootPageID << endl << endl;
	
	//cout << "IndexManager:: lookUpIDForInsert() Point A .. " << endl << endl;
	
	// assign the "rootPageID" to the "pageID" ..
	int pageID = rootPageID;
	
	// read the root "IndexPage" ..
	IndexPage* indexPage = new IndexPage();
	
	bool wasIndexPageRead = readIndexPage( indexTableName_, indexName_, indexPage, pageID );
	
	//cout << "IndexManager:: lookUpIDForInsert() Point B .. " << endl << endl;
	
	vector<int>* searchKeys = NULL;
	int searchKeysSize = -1;
	int childPageIDIndex = -1;
	
	vector<int>* childPageIDs = NULL;
	int childPageIDsSize = -1;
	
	bool isSearchKeySizeEqualToChildPageIDSize = false;
	
	// DEBUGGING ..
	int loopCounter = 0;
	
	// find the IndexPage that has 
	while( indexPage->getIsLeafNode() == false ){
		
		if( wasIndexPageRead = true ){
			
			// get the childPageID of the (next) IndexPage to search ..
			childPageIDs = indexPage->getChildPageIDs();
			childPageIDsSize = childPageIDs->size();
		
			searchKeys = indexPage->getSearchKeys();
			searchKeysSize = searchKeys->size();
			
			//cout << "IndexManager:: lookUpIDForInsert() searchKeysSize: " << searchKeysSize << endl << endl;
			
			//for(int i=0; i<searchKeysSize; i++){
			//	cout << "IndexManager:: lookUpIDForInsert() searchKeys->at( i("<< i <<") ): " << searchKeys->at( i ) << endl << endl;
			//}
			
			// check if # of searchKeys matches # of childPageIDs ..
			if( searchKeysSize == childPageIDsSize ){
				isSearchKeySizeEqualToChildPageIDSize = true;	
			} else {
				isSearchKeySizeEqualToChildPageIDSize = false;	
			}
		
			for(int i=0; i<searchKeysSize; i++){
				
				//cout << "IndexManager:: lookUpIDForInsert() searchKeys->at( i("<< i <<") ): " << searchKeys->at( i ) << endl << endl;
				
				// childPageID is left of this searchKey ..
				if( searchKey_ < searchKeys->at( i ) ){
					
					childPageIDIndex = i;
					break;
					 
				// childPageID is right of this searchKey ..
				} else if( searchKey_ == searchKeys->at( i ) ){
					
					// avoid overflow of vector ..
					if( isSearchKeySizeEqualToChildPageIDSize == true )
						childPageIDIndex = i;
					else
						childPageIDIndex = i+1;
						
					break;
					
				} else if( searchKey_ > searchKeys->at( i ) ){
					
					// for the last childPageID in the collection
					// just use the greater pageID ..
					if( i == ( searchKeysSize - 1 ) ){
						
						// avoid overflow of vector ..
						if( isSearchKeySizeEqualToChildPageIDSize == true )
							childPageIDIndex = i;
						else
							childPageIDIndex = i+1;
						
						break;
						
					}
					
					// the other 2 conditions should find 
					// the childPageID ..
					continue;
					
				} // end "if()" ..
				
			} // end "for()" ..
			
		} else {
			
			// a read error should halt search ..
			isIDFound = false;
			break;	
			
		} // end "if()" ..
		
		//cout << "IndexManager:: lookUpIDForInsert() Point C[" << loopCounter << "] .. " << endl << endl;
		
		//cout << "IndexManager:: lookUpIDForInsert() childPageIDIndex: " << childPageIDIndex << ", childPageIDs->size(): " << childPageIDs->size() << endl << endl;
		
		pageID = childPageIDs->at( childPageIDIndex );
		
		indexPage = new IndexPage();
		
		// read the next index page to search ..
		wasIndexPageRead = readIndexPage( indexTableName_, indexName_, indexPage, pageID );
		
		//cout << "IndexManager:: lookUpIDForInsert() Point D[" << loopCounter << "] .. " << endl << endl;
		
		loopCounter++;
		
	} // end "while()" ..
	
	
	// return the pageID of the retrieved leaf node ..
	int returnPageID = -1;
	
	if( isIDFound == true )
		returnPageID = indexPage->getPageID();
		
	//cout << "IndexManager:: lookUpIDForInsert() returnPageID: " << returnPageID << endl << endl;
	
	return returnPageID;
	
	
}
