#include "Tester.h"

Tester:: Tester(){
	
	parentNodes_ = NULL;
	
	extraParentNodeGeneratedPerRow_ = NULL;
	
}

Tester:: ~Tester(){
	
	if( parentNodes_ != NULL ) delete parentNodes_;
	
	if( extraParentNodeGeneratedPerRow_ != NULL ) delete extraParentNodeGeneratedPerRow_;
	
}

void Tester:: writeMetaData(){
	
	CatalogTable* catalogTable = new CatalogTable();
	
	catalogTable->setTableName("xyz");
	catalogTable->setTableSize( 0 );
	catalogTable->addAttribute("aa");
	catalogTable->addAttribute("bb");
	catalogTable->addAttribute("cc");
	
	CatalogIndex* catalogIndex = new CatalogIndex();
	
	catalogIndex->setIndexName("I");
	catalogIndex->setIndexTableName("xyz");
	catalogIndex->setIndexSize( 0 );
	catalogIndex->setIsNoDuplicates( false ); // "NO DUPLICATES" option NOT invoked ..
	catalogIndex->setAttribute("aa");

	MessageManager* mm = new MessageManager();
	CatalogFile* catalogFile = new CatalogFile();
	catalogFile->setFileDrive("C:\\");
	catalogFile->setFileDirectory("Temp\\");
	catalogFile->setFileName("Catalog");
	catalogFile->setFileExtension(".bin");
	
	CatalogManager catalogManager( mm, catalogFile );
	
	catalogManager.addCatalogTable( catalogTable );		// calls "writeMetaData()" ..
	catalogManager.addCatalogIndex( catalogIndex );		// calls "writeMetaData()" ..
	
	bool writeMetaDataSuccessful = catalogManager.writeMetaData();
	
	cout << "writeMetaDataSuccessful: " << writeMetaDataSuccessful << endl << endl;
	
} // end "writeMetaData()" ..


void Tester:: readMetaData(){
	
	MessageManager* mm = new MessageManager();
	CatalogFile* catalogFile = new CatalogFile();
	catalogFile->setFileDrive("C:\\");
	catalogFile->setFileDirectory("Temp\\");
	catalogFile->setFileName("Catalog");
	catalogFile->setFileExtension(".bin");
	
	CatalogManager catalogManager( mm, catalogFile );
	
	bool readMetaDataSuccessful = catalogManager.readMetaData();
	
	cout << "readMetaDataSuccessful: " << readMetaDataSuccessful << endl << endl;
	
	// DEBUGGING ..
	cout << "CatalogManager:: readMetaData(): " << endl << endl;
	
	CatalogTable* catalogTable = catalogManager.getCatalogTable("xyz");
	
	cout << "CatalogManager:: catalogTable.getTableName(): " << catalogTable->getTableName() << endl << endl;
	cout << "CatalogManager:: catalogTable.getNumberOfAttributes(): " << catalogTable->getNumberOfAttributes() << endl << endl;
	cout << "CatalogManager:: catalogTable.getTableSize(): " << catalogTable->getTableSize() << endl << endl;
	
	vector<CatalogTableAttribute*>* cta = catalogTable->getAttributes();
	for(int i=0; i<cta->size(); i++)
		cout << "CatalogManager:: cta[i("<<i<<")].getTableAttributeName(): " << cta->at(i)->getTableAttributeName() << endl << endl;
	
	CatalogIndex* catalogIndex = catalogManager.getCatalogIndex("I");
	
	cout << "CatalogManager:: catalogIndex.getIndexName(): " << catalogIndex->getIndexName() << endl << endl;
	cout << "CatalogManager:: catalogIndex.getIndexTableName(): " << catalogIndex->getIndexTableName() << endl << endl;
	cout << "CatalogManager:: catalogIndex.getIndexSize(): " << catalogIndex->getIndexSize() << endl << endl;
	cout << "CatalogManager:: catalogIndex.getIsNoDuplicates(): " << catalogIndex->getIsNoDuplicates() << endl << endl;
	CatalogIndexAttribute* cia = catalogIndex->getAttribute();
	cout << "CatalogManager:: cia.getIndexAttributeName(): " << cia->getIndexAttributeName() << endl << endl;

	
} // end "readMetaData()" ..


void Tester:: testPowerOperator(){
	
	// "^" is a "binary" operator in C/C++ ..
	
	int rowLevels = 0;
	
	cout << "(5^rowLevel("<<rowLevels<<")): " << (5^rowLevels) << endl << endl;
	
	rowLevels = 1;
	
	cout << "(5^rowLevel("<<rowLevels<<")): " << (5^rowLevels) << endl << endl;
	
	rowLevels = 2;
	
	cout << "(5^rowLevel("<<rowLevels<<")): " << (5^rowLevels) << endl << endl;
	
		
	// must use <cmath> function "double z = pow(double x, double y)" for "z = x^y" ..
	
	double rowLevel = 0;
	
	cout << "pow( 5, rowLevel("<<rowLevel<<") ): " << pow(5, rowLevel) << endl << endl;
	
	rowLevel = 1;
	
	cout << "pow( 5, rowLevel("<<rowLevel<<") ): " << pow(5, rowLevel) << endl << endl;
	
	rowLevel = 2;
	
	cout << "pow( 5, rowLevel("<<rowLevel<<") ): " << pow(5, rowLevel) << endl << endl;
	
} // end "testPowerOperator()" ..


void Tester:: writeIndexBuckets(){
	
	// test data ..
	vector<IndexBucket*> indexBuckets;
	
	IndexBucket* indexBucket = new IndexBucket();
	
	vector<int>* pageIDs = new vector<int>();
	
	// index bucket #1 = 12 pageIDs (1 page) ..
	pageIDs->push_back(0);
	pageIDs->push_back(1);
	pageIDs->push_back(2);
	pageIDs->push_back(3);
	pageIDs->push_back(4);
	pageIDs->push_back(5);
	pageIDs->push_back(6);
	pageIDs->push_back(7);
	pageIDs->push_back(8);
	pageIDs->push_back(9);
	pageIDs->push_back(10);
	pageIDs->push_back(11);

	
	indexBucket->setPageIDs( pageIDs );
	
	indexBuckets.push_back( indexBucket );
	
	
	// index bucket #2 = 5 pageIDs (1 page) ..
	
	indexBucket = new IndexBucket();
	
	pageIDs = new vector<int>();
	
	pageIDs->push_back(12);		// 0
	pageIDs->push_back(13);		// 1
	pageIDs->push_back(14);		// 2
	pageIDs->push_back(15);		// 3
	pageIDs->push_back(16);		// 4
	
	indexBucket->setPageIDs( pageIDs );
	
	indexBuckets.push_back( indexBucket );
	
	
	
	// index bucket #3 = 17 pageIDs (2 pages) ..
	
	indexBucket = new IndexBucket();
	
	pageIDs = new vector<int>();
	
	pageIDs->push_back(17);		// 0
	pageIDs->push_back(18);		// 1
	pageIDs->push_back(19);		// 2
	pageIDs->push_back(20);		// 3
	pageIDs->push_back(21);		// 4
	pageIDs->push_back(22);		// 5
	pageIDs->push_back(23);		// 6
	pageIDs->push_back(24);		// 7
	pageIDs->push_back(25);		// 8
	pageIDs->push_back(26);		// 9
	pageIDs->push_back(27);		// 10
	pageIDs->push_back(28);		// 11
	pageIDs->push_back(29);		// 12
	pageIDs->push_back(30);		// 13
	pageIDs->push_back(31);		// 14
	pageIDs->push_back(32);		// 15
	pageIDs->push_back(33);		// 16
	
	indexBucket->setPageIDs( pageIDs );
	
	indexBuckets.push_back( indexBucket );
	
	
	// index bucket #4 = 29 pageIDs (3 pages) ..
	
	indexBucket = new IndexBucket();
	
	pageIDs = new vector<int>();
	
	pageIDs->push_back(34);		// 0
	pageIDs->push_back(35);		// 1
	pageIDs->push_back(36);		// 2
	pageIDs->push_back(37);		// 3
	pageIDs->push_back(38);		// 4
	pageIDs->push_back(39);		// 5
	pageIDs->push_back(40);		// 6
	pageIDs->push_back(41);		// 7
	pageIDs->push_back(42);		// 8
	pageIDs->push_back(43);		// 9
	pageIDs->push_back(44);		// 10
	pageIDs->push_back(45);		// 11
	pageIDs->push_back(46);		// 12
	pageIDs->push_back(47);		// 13
	pageIDs->push_back(48);		// 14
	pageIDs->push_back(49);		// 15
	pageIDs->push_back(50);		// 16
	pageIDs->push_back(51);		// 17
	pageIDs->push_back(52);		// 18
	pageIDs->push_back(53);		// 19
	pageIDs->push_back(54);		// 20
	pageIDs->push_back(55);		// 21
	pageIDs->push_back(56);		// 22
	pageIDs->push_back(57);		// 23
	pageIDs->push_back(58);		// 24
	pageIDs->push_back(59);		// 25
	pageIDs->push_back(60);		// 26
	pageIDs->push_back(61);		// 27
	pageIDs->push_back(62);		// 28
	
	indexBucket->setPageIDs( pageIDs );
	
	indexBuckets.push_back( indexBucket );
	
	
	// index bucket #5 = 1 pageIDs (1 page) ..
	
	indexBucket = new IndexBucket();
	
	pageIDs = new vector<int>();
	
	pageIDs->push_back(63);		// 0

	
	indexBucket->setPageIDs( pageIDs );
	
	indexBuckets.push_back( indexBucket );
	
	
	// index bucket #6 = 0 pageIDs (1 page?) ..
	
	// empty "pageIDs" goes into "indexBucket" ..
	indexBucket = new IndexBucket();
	pageIDs = new vector<int>();

	// this wont work ..
	indexBucket->setPageIDs( pageIDs );
	indexBuckets.push_back( indexBucket );

	
	// index bucket #7 = 6 pageIDs (1 page) ..
	
	indexBucket = new IndexBucket();
	
	pageIDs = new vector<int>();
	
	pageIDs->push_back(64);		// 0
	pageIDs->push_back(65);		// 1
	pageIDs->push_back(66);		// 2
	pageIDs->push_back(67);		// 3
	pageIDs->push_back(68);		// 4
	pageIDs->push_back(69);		// 5
	
	indexBucket->setPageIDs( pageIDs );
	
	indexBuckets.push_back( indexBucket );
	
	
	// Create the layers ..
	string fileDrive = "C:\\";
	string fileDirectory = "Temp\\";
	CatalogFile* catalogFile = new CatalogFile( fileDrive, fileDirectory, "catalog", ".bin" );
	DatabaseFile* databaseFile = new DatabaseFile( fileDrive, fileDirectory );
	MessageManager* mm = new MessageManager();
	CatalogManager* cm = new CatalogManager( mm, catalogFile );
	StorageManager* sm = new StorageManager( cm, databaseFile );
	IndexManager* im = new IndexManager( sm );
	
	
	// header data ..
	
	char* indexTableName = "abc";
	
	char* indexName = "I";
	
	IndexBucketHeader indexBucketHeader;
	
	bool isReadBucketHeader = readIndexBucketHeader( indexTableName, indexName, indexBucketHeader );
	
	if( indexBucketHeader.getTotalNumberBuckets() < indexBuckets.size() )
		indexBucketHeader.setTotalNumberBuckets( indexBuckets.size() );
		
	
	int numberBucketIDsPerSearchKey = 5;
	int numberSearchKeys = 4;
	
	map<int, vector<int>* >* _indexBucketIDsForEachSearchKey = new map<int, vector<int>* >();
	
	vector<int>* bucketIDs = new vector<int>();
	
	int increment = 0;
	
	for(int i=0; i<numberSearchKeys; i++) {
	
		bucketIDs = new vector<int>();
	
		for(int j=0; j<numberBucketIDsPerSearchKey; j++)
			bucketIDs->push_back( (j + increment) );
		
		_indexBucketIDsForEachSearchKey->insert( pair<int, vector<int>* >( i, bucketIDs ) );
	
		increment++;
	
		//delete bucketIDs;
		
	}
		
	cout << "testStorageManager_writeIndexBucketHeader() Point B .." << endl << endl;
	
	indexBucketHeader.setBucketIDsForEachSearchKey( _indexBucketIDsForEachSearchKey );
	
	vector<IndexBucketHeaderEntry*>* ibhev = new vector<IndexBucketHeaderEntry*>();
	
	IndexBucketHeaderEntry* ibhe = NULL;
	
	//vector<bool>* pointersAvailable = NULL;
	
	for(int i=0; i<numberSearchKeys; i++) {
		
		// this isnt necessarily realistic, but test is sufficient ..
		for(int j=0; j<numberBucketIDsPerSearchKey; j++) {
		
			ibhe = new IndexBucketHeaderEntry();
		
			ibhe->setBucketID( ( j + increment ) );
			
			// 5 used, 7 unused ..
			ibhe->setFreePointerCount( 7 );
			
			//pointersAvailable = new vector<bool>();
			
			// total of 12 slots max per bucket ..
			//for(int k=0; k<5; k++)
			//	pointersAvailable->push_back(false);
			//for(int k=0; k<7; k++)
			//	pointersAvailable->push_back(true);
			
			//ibhe->setPointersAvailable( pointersAvailable );
				
			ibhev->push_back( ibhe );
			
		}
		
		increment++;
		
	}
	
	
	indexBucketHeader.setHeaderEntries( ibhev );
		
	
	// algorithm ..
	
	int indexBucketSize = 12;
	bool wasIndexBucketWritten = false;
	int indexBucketPosition = 0;
	int currentIndexBucketNumberOfPageIDs = -1;
	int numberPagesForThisBucket = -1;
	int shiftValueForMultiplePageBuckets = 0;	// if current "indexBucket" has multiple pages, then use this to increment "i" ..
	int shiftValueForSkippedBuckets = 0;		// if current "indexBucket" is empty, then use this to decrement "i" ..
	
	// bucket #, bucket-pageIDs ..
	map<int, vector<int>* >* bucketPageIDsPerSearchKey = NULL;
	
	// bucket-pageIDs (pointers to bucket pages) ..
	vector<int>* bucketPageIDs = NULL;
					
	// data-pageIDs (pointers to data pages) from the bucket ..
	vector<int>* allCurrentBucketPageIDs = NULL;
					
	// increment through each "index bucket" ..
	for(int i=0; i<indexBuckets.size(); i++){
		
		if( bucketPageIDsPerSearchKey == NULL )
			bucketPageIDsPerSearchKey = new map<int, vector<int>* >();
			
		if( bucketPageIDs == NULL )
			bucketPageIDs = new vector<int>();
						
		allCurrentBucketPageIDs = indexBuckets[i]->getPageIDs();
						
		// find out if the "index bucket" stored in memory needs to be 
		// parcelled into multiple file pages ..
		currentIndexBucketNumberOfPageIDs = allCurrentBucketPageIDs->size();
		
		cout << "=====================================================" << endl << endl;
		
		cout << "current bucket: " << ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) << endl << endl;
		
		cout << "currentIndexBucketSize[i("<<i<<")]: " << currentIndexBucketNumberOfPageIDs << endl << endl;
						
		// # pages to write = # times you can subtract "indexBucketSize"
		// from "currentIndexBucketNumberOfPageIDs" before "result" value < 0 ..
		int result = currentIndexBucketNumberOfPageIDs;
		int counter = 0;
		while( result > 0 ){
								
			result -= indexBucketSize;
							
			counter++;
								
		}
						
		// # pages to write == "counter" .. 
		numberPagesForThisBucket = counter;
		
		cout << "numberPagesForThisBucket[i("<<i<<")]: " << numberPagesForThisBucket << endl << endl;
		
		cout << "if( currentIndexBucketSize("<<currentIndexBucketNumberOfPageIDs<<") > indexBucketSize("<<indexBucketSize<<") )[i("<<i<<")]: " << ( currentIndexBucketNumberOfPageIDs > indexBucketSize ) << endl << endl;
						
		cout << "if( ( 0 < currentIndexBucketSize("<<currentIndexBucketNumberOfPageIDs<<") )("<<( 0 < currentIndexBucketNumberOfPageIDs )<<") && ( currentIndexBucketSize("<<currentIndexBucketNumberOfPageIDs<<") <= indexBucketSize("<<indexBucketSize<<") )("<<( currentIndexBucketNumberOfPageIDs <= indexBucketSize )<<") )[i("<<i<<")]: " << ( ( 0 < currentIndexBucketNumberOfPageIDs ) && ( currentIndexBucketNumberOfPageIDs <= indexBucketSize ) ) << endl << endl;
						
		// this indicates that this bucket will be written to MULTIPLE pages ..
		if( currentIndexBucketNumberOfPageIDs > indexBucketSize ){

			// traverse the entire bucket ..
			// must write "bucket pages" in increments of "indexBucketSize" ..
			int bucketPageCounter = 1;
							
			IndexBucket* intermediateIndexBucket = new IndexBucket();
							
			while( indexBucketPosition < currentIndexBucketNumberOfPageIDs ){
				
				cout << "--------------" << endl << endl;
				
				cout << "while( indexBucketPosition("<<indexBucketPosition<<") < currentIndexBucketSize("<<currentIndexBucketNumberOfPageIDs<<") )[i("<<i<<")]: " << ( indexBucketPosition < currentIndexBucketNumberOfPageIDs ) << endl << endl;
								
				cout << "Point A - if( indexBucketPosition("<<indexBucketPosition<<") == ( ( bucketPageCounter("<<bucketPageCounter<<") * indexBucketSize("<<indexBucketSize<<") )("<<( bucketPageCounter * indexBucketSize )<<") - 1 )("<<( ( bucketPageCounter * indexBucketSize ) - 1 )<<") )[i("<<i<<")]: " << ( indexBucketPosition == ( ( bucketPageCounter * indexBucketSize ) - 1 ) ) << endl << endl;
								
				cout << "Point A - && if( currentIndexBucketSize("<<currentIndexBucketNumberOfPageIDs<<") >= ( bucketPageCounter("<<bucketPageCounter<<") * indexBucketSize("<<indexBucketSize<<") )("<<( bucketPageCounter * indexBucketSize )<<") )[i("<<i<<")]: " << ( currentIndexBucketNumberOfPageIDs >= ( bucketPageCounter * indexBucketSize ) ) << endl << endl;
								
				cout << "Point C - else if( indexBucketPosition("<<indexBucketPosition<<") == ( currentIndexBucketSize("<<currentIndexBucketNumberOfPageIDs<<") - 1 )("<<( currentIndexBucketNumberOfPageIDs - 1 )<<") )[i("<<i<<")]: " << ( indexBucketPosition == ( currentIndexBucketNumberOfPageIDs - 1 ) ) << endl << endl;
								
				// allocate a new "intermediateIndexBucket" for the next page to be written ..
				// i.e.: when "indexBucketPosition" == 11, 22, 33, 44, etc ..
				// i.e.: 0 1 2 3 4 5 6 7 8 9 10 11			...		 	22			...		 	33  <- "indexBucketPosition"
				//	    | |			...		    | | <- "indexBucket"
				//								 0 1 2 3 4 5 6 7 8 9 10 11
				//	    					    | |			...		    | | <- "indexBucket"
				//								 						 0 1 2 3 4 5 6 7 8 9 10 11
				//	    					    						| |			...		    | | <- "indexBucket"
								
				// also, "currentIndexBucketSize" large enough that an additional page is needed for this bucket ..
				// i.e.: 0 1 2 3 4 5 6 7 8 9 10 11 12 ..  <- "indexBucketPosition", etc ..
				//	    | |			...		    | | <- "indexBucket", etc ..
				// [ where "( bucketPageCounter * indexBucketSize )" == 12, 24, 36, 48, etc .. ]
				
				if( ( indexBucketPosition == ( ( bucketPageCounter * ( indexBucketSize - 1 )  ) ) ) &&
					( currentIndexBucketNumberOfPageIDs >= ( ( bucketPageCounter * ( indexBucketSize - 1 ) ) + 1 ) ) ) {
									
					cout << "--------------" << endl << endl;
									
					cout << "Point A[i("<<i<<")] .." << endl << endl;
					
					cout << "intermediateIndexBucket->getPageIDs()->push_back( ( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") + 1 ) )[i("<<i<<")]: " << ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets + 1 ) << endl << endl;
									
					// store the pointer to the next bucket page in the final 4b of this page ..
					intermediateIndexBucket->getPageIDs()->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets + 1 ) );
									
					cout << "writeIndexBucket(indexTableName("<<indexTableName<<"), .. ,( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") )("<<( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets )<<"))[i("<<i<<")] .." << endl << endl;
											
					// writeIndexBucket( indexBucket );
					wasIndexBucketWritten = im->writeIndexBucket( indexTableName, indexName, intermediateIndexBucket, ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );
									
					// track this bucketPageID ..		
					bucketPageIDs->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );
						
					// reset for new "intermediateIndexBucket" ..
					delete intermediateIndexBucket;
					intermediateIndexBucket = new IndexBucket();
					
					cout << "new IndexBucket()[i"<<i<<"] .." << endl << endl;
									
					bucketPageCounter++;
									
					shiftValueForMultiplePageBuckets++;		// adjustment to "i" b/c of duplicate buckets ..
					
					cout << "bucketPageCounter++[i("<<i<<")]: " << bucketPageCounter << endl << endl;
					
					cout << "shiftValueForMultiplePageBuckets++[i("<<i<<")]: " << shiftValueForMultiplePageBuckets << endl << endl;
									
				} else if( indexBucketPosition == ( currentIndexBucketNumberOfPageIDs - 1 ) ){
					
					cout << "--------------" << endl << endl;
					
					cout << "Point C[i("<<i<<")] .." << endl << endl;
					
					cout << "intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition("<<indexBucketPosition<<") )("<<allCurrentBucketPageIDs->at( indexBucketPosition )<<") )[i("<<i<<")]: " << allCurrentBucketPageIDs->at( indexBucketPosition ) << endl << endl;
									
					// this is the last pageID on the last page for this bucket ..
									
					// store the pageIDs for this "intermediate index bucket" ..
					intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition ) );
									
					cout << "writeIndexBucket(indexTableName("<<indexTableName<<"), .. ,( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") )("<<( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets )<<"))[i("<<i<<")] .." << endl << endl;
									
					// writeIndexBucket( indexBucket );
					wasIndexBucketWritten = im->writeIndexBucket( indexTableName, indexName, intermediateIndexBucket, ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );
									
					// track this bucketPageID ..		
					bucketPageIDs->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );
									
					// reset for new "intermediateIndexBucket" ..
					delete intermediateIndexBucket;
					
					cout << "break out of while loop [i("<<i<<")] .." << endl << endl;
					
					cout << "-----------------------------------------------------" << endl << endl;
									
					break;
									
				}
								
				cout << "Point B[i("<<i<<")] .." << endl << endl;
				
				cout << "intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition("<<indexBucketPosition<<") )[i("<<i<<")]: " << allCurrentBucketPageIDs->at( indexBucketPosition ) <<endl << endl;
								
				// store the pageIDs for this "intermediate index bucket" ..
				intermediateIndexBucket->getPageIDs()->push_back( allCurrentBucketPageIDs->at( indexBucketPosition ) );
									
				// traverse to next pageID in current bucket ..
				indexBucketPosition++;
				
				cout << "indexBucketPosition++[i("<<i<<")]: " << indexBucketPosition <<endl << endl;
								
								
			} // end "while()" ..
							
			// reset this parameter for the next loop through ..
			indexBucketPosition = 0;
							
		} else if( ( 0 < currentIndexBucketNumberOfPageIDs ) && ( currentIndexBucketNumberOfPageIDs <= indexBucketSize ) ){
			
			// this indicates that this bucket will be written to a SINGLE page ..
			
			cout << "Point D[i("<<i<<")] .." << endl << endl;
						
			cout << "writeIndexBucket(indexTableName("<<indexTableName<<"), .. ,( i("<<i<<") + shiftValueForMultiplePageBuckets("<<shiftValueForMultiplePageBuckets<<") - shiftValueForSkippedBuckets("<<shiftValueForSkippedBuckets<<") )("<<( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets )<<"))[i("<<i<<")] .." << endl << endl;						
						
			// writeIndexBucket( indexBucket );
			wasIndexBucketWritten = im->writeIndexBucket( indexTableName, indexName, indexBuckets[i], ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );
						
			// track this bucketPageID ..		
			bucketPageIDs->push_back( ( i + shiftValueForMultiplePageBuckets - shiftValueForSkippedBuckets ) );
						
				
		} else {
			
			// if the bucket is skipped (because it is empty), 
			// adjust "i" so that this pageID isnt skipped for the 
			// next "writeIndexBucket()" operation above ..
			shiftValueForSkippedBuckets++;
			
		} // end "if( currentIndexBucketSize > indexBucketSize )" ..
					
					
		// track the bucket-pageIDs for each unique search key ..
		bucketPageIDsPerSearchKey->insert( pair<int, vector<int>* >( i, bucketPageIDs ) );
		
		
	} // end "for()" ..
	
	// cleanup ..
	delete indexBucket;
	delete pageIDs;
	
	
} // end "writeIndexBuckets()" ..


void Tester:: createIndexSearchKeys(){
	
	
/*
 * Data set #1 ..
 */
/*

					int uniqueSearchKeysArray[] = {};
					
/*
 * Data set #2 ..
 */
/*

					int uniqueSearchKeysArray[] = {5};
					
/*
 * Data set #3 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10};
					
					
/*
 * Data set #4 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20};
			
/*
 * Data set #5 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25};		
	
/*
 * Data set #6 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30};
												  
/*
 * Data set #7 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40};
	
	
/*
 * Data set #8 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60};
	
/*
 * Data set #9 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80};
	
/*
 * Data set #10 ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100};
	
	
/*
 * smaller set of data (exceptional case #2) ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105};
									
/*
 * smaller set of data (exceptional case #2) ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110};	
									
									
/*
 * smaller set of data (exceptional case #2) ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115};	
									
/*
 * smaller set of data (exceptional case #2) ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120};			  
												  
				
/*
 * Data set #11 ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140};
												  
/*
 * Data set #12 ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220};

/*
 * Data set #13 ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230};
							  
/*
 * Data set #14 ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475};
												  		
/*
 * Data set #15 ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500};
												  		
/*
 * bigger set of data (exceptional case #2) ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505};
					
/*
 * bigger set of data (exceptional case #2) ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520};
					
/*
 * bigger set of data (exceptional case #2) ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520,
												  525};

/*
 * bigger set of data (exceptional case #2) ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520,
												  525, 530, 535, 540,
												  545};
		  
/*
 * bigger set of data (exceptional case #2) ..
 */


					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520,
												  525, 530, 535, 540,
												  545, 550, 555, 560,
												  565};
		  
/*
 * bigger set of data (exceptional case #2) ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520,
												  525, 530, 535, 540,
												  545, 550, 555, 560,
												  565, 570, 575, 580,
												  585};
		  
/*
 * bigger set of data (exceptional case #2) ..
 */
/*

					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520,
												  525, 530, 535, 540,
												  545, 550, 555, 560,
												  565, 570, 575, 580,
												  585, 590, 595, 600};
								
								
								
/*
 * Data set #16 ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520,
												  525, 530, 535, 540,
												  545, 550, 555, 560,
												  565, 570, 575, 580,
												  585, 590, 595, 600,
												  
												  605};
												  

/*
 * Data set #17 ..
 */
/*
					int uniqueSearchKeysArray[] = {5, 10, 15, 20,
												  25, 30, 35, 40,
												  45, 50, 55, 60,
												  65, 70, 75, 80,
												  85, 90, 95, 100,
												  
												  105, 110, 115, 120,
												  125, 130, 135, 140,
												  145, 150, 155, 160,
												  165, 170, 175, 180,
												  185, 190, 195, 200,
												  
												  205, 210, 215, 220,
												  225, 230, 235, 240,
												  245, 250, 255, 260,
												  265, 270, 275, 280,
												  285, 290, 295, 300,
												  
												  305, 310, 315, 320,
												  325, 330, 335, 340,
												  345, 350, 355, 360,
												  365, 370, 375, 380,
												  385, 390, 395, 400,
												  
												  405, 410, 415, 420,
												  425, 430, 435, 440,
												  445, 450, 455, 460,
												  465, 470, 475, 480,
												  485, 490, 495, 500,
												  
												  505, 510, 515, 520,
												  525, 530, 535, 540,
												  545, 550, 555, 560,
												  565, 570, 575, 580,
												  585, 590, 595, 600,
												  
												  605, 610, 615, 620,
												  625, 630, 635, 640,
												  645, 650, 655, 660,
												  665, 670, 675, 680,
												  685, 690, 695, 700,
												  
												  705, 710, 715, 720,
												  725, 730, 735, 740,
												  745, 750, 755, 760,
												  765, 770, 775, 780,
												  785, 790, 795, 800,
												  
												  805, 810, 815, 820,
												  825, 830, 835, 840,
												  845, 850, 855, 860,
												  865, 870, 875, 880,
												  885, 890, 895, 900,
												  
												  905, 910, 915, 920,
												  925, 930, 935, 940,
												  945, 950, 955, 960,
												  965, 970, 975, 980,
												  985, 990, 995, 1000,
												  
												  1005, 1010, 1015, 1020,
												  1025, 1030, 1035, 1040,
												  1045, 1050, 1055, 1060,
												  1065, 1070, 1075, 1080,
												  1085, 1090, 1095, 1100,
												  
												  1105, 1110, 1115, 1120,
												  1125, 1130, 1135, 1140,
												  1145, 1150, 1155, 1160,
												  1165, 1170, 1175, 1180,
												  1185, 1190, 1195, 1200,
												  
												  1205, 1210, 1215, 1220,
												  1225, 1230, 1235, 1240,
												  1245, 1250, 1255, 1260,
												  1265, 1270, 1275, 1280,
												  1285, 1290, 1295, 1300,
												  
												  1305, 1310, 1315, 1320,
												  1325, 1330, 1335, 1340,
												  1345, 1350, 1355, 1360,
												  1365, 1370, 1375, 1380,
												  1385, 1390, 1395, 1300,
									
												  1405, 1410, 1415, 1420,
												  1425, 1430, 1435, 1440,
												  1445, 1450, 1455, 1460,
												  1465, 1470, 1475, 1480,
												  1485, 1490, 1495, 1500,
												  
												  1505, 1510, 1515, 1520,
												  1525, 1530, 1535, 1540,
												  1545, 1550, 1555, 1560,
												  1565, 1570, 1575, 1580,
												  1585, 1590, 1595, 1600,
												  
												  1605, 1610, 1615, 1620,
												  1625, 1630, 1635, 1640,
												  1645, 1650, 1655, 1660,
												  1665, 1670, 1675, 1680,
												  1685, 1690, 1695, 1700,
												  
												  1705, 1710, 1715, 1720,
												  1725, 1730, 1735, 1740,
												  1745, 1750, 1755, 1760,
												  1765, 1770, 1775, 1780,
												  1785, 1790, 1795, 1800,
												  
												  1805, 1810, 1815, 1820,
												  1825, 1830, 1835, 1840,
												  1845, 1850, 1855, 1860,
												  1865, 1870, 1875, 1880,
												  1885, 1890, 1895, 1900,
												  
												  1905, 1910, 1915, 1920,
												  1925, 1930, 1935, 1940,
												  1945, 1950, 1955, 1960,
												  1965, 1970, 1975, 1980,
												  1985, 1990, 1995, 2000,
												  
												  2005, 2010, 2015, 2020,
												  2025, 2030, 2035, 2040,
												  2045, 2050, 2055, 2060,
												  2065, 2070, 2075, 2080,
												  2085, 2090, 2095, 2100,
												  
												  2105, 2110, 2115, 2120,
												  2125, 2130, 2135, 2140,
												  2145, 2150, 2155, 2160,
												  2165, 2170, 2175, 2180,
												  2185, 2190, 2195, 2200,
												  
												  2205, 2210, 2215, 2220,
												  2225, 2230, 2235, 2240,
												  2245, 2250, 2255, 2260,
												  2265, 2270, 2275, 2280,
												  2285, 2290, 2295, 2300,
												  
												  2305, 2310, 2315, 2320,
												  2325, 2330, 2335, 2340,
												  2345, 2350, 2355, 2360,
												  2365, 2370, 2375, 2380,
												  2385, 2390, 2395, 2400,
												  
												  2405, 2410, 2415, 2420,
												  2425, 2430, 2435, 2440,
												  2445, 2450, 2455, 2460,
												  2465, 2470, 2475, 2480,
												  2485, 2490, 2495, 2500,
												  
												  2505, 2510, 2515, 2520,
												  2525, 2530, 2535, 2540,
												  2545, 2550, 2555, 2560,
												  2565, 2570, 2575, 2580,
												  2585, 2590, 2595, 2600,
												  
												  2605};
									
/*
 * end of data sets ..
 */
												  
					vector<int> uniqueSearchKeys(uniqueSearchKeysArray, ( uniqueSearchKeysArray + ( sizeof(uniqueSearchKeysArray) / sizeof(int) ) ) );
	
					vector<int> currentRowAllSearchKeys = uniqueSearchKeys;
	
					// contains all the rows of searchKeys ..
					// first row are the leaf searchKeys ..
					// remaining rows (if there) are parent searchKeys ..
					//vector< vector< vector<int>* >* >* _parentNodes = NULL;
					
					// contains all the "searchKeys" for a row ..
					vector< vector<int>* >* parentSearchKeysRow = NULL;
					
					// contains the "searchKeys" for a single parent node ..
					vector<int>* parentSearchKeys = NULL;
					
					//NOTE: parentNodes -> parentSearchKeysRow -> parentSearchKeys
					
					int numberSearchKeysPerNode = 4;			// i.e.: |10 20 30 40|
					
					int fanout = (numberSearchKeysPerNode + 1);	// i.e.: |P|10|P|20|P|30|P|40|P|
					
					int indexKeyOffset = 5;				// every "indexKeyOffset^L" searchKey is a pointer value for a parent node above ..
														// where "L" is the row number (leaf node is row number = 1) ..
														// i.e.: leaf node: every (6^1 = 6th) leaf node search key
														//       next level closer to root node: every (6^2 = 36th) leaf node search key
					
					bool atLeastOnce = true;
					
					bool extraParentNodeGenerated = false;	// true if last parent node in row is 6th node,
															// and a super-parent node needs to be generated for it ..
															// parent node was generated for a single child node ..
													
					//vector< bool >* _extraParentNodeGeneratedPerRow = NULL;	// For each row of nodes, in last set of nodes
																			// if number of nodes = 1, then 
																			// its parent is generated as an exceptional case, so
																			// indicate that the parent should have only 1 child pageID
																			// instead of the usual 2 or more .. 
															
					 
					double rowLevel = 0;
					
					// you know to stop creating parent node rows when the last row is "sizeOfLastRow" or less ..
					while( ( atLeastOnce ) || 
						   ( currentRowAllSearchKeys.size() > numberSearchKeysPerNode ) ) {	// i.e.: |10 20 30 40|  |50        | ..
					
						if ( atLeastOnce )
							atLeastOnce = false;
						
						// DEBUGGING ..	
						cout << "rowLevel: " << rowLevel << endl << endl;
						cout << "atLeastOnce: " << atLeastOnce << endl << endl;
						cout << "currentRowAllSearchKeys.size(): " << currentRowAllSearchKeys.size() << endl << endl;
						
						// initialize "searchKeys" for collection that holds parent nodes for this row ..
						parentSearchKeys = NULL;
						
						// initialize collection that holds this row of "searchKeys" ..
						parentSearchKeysRow = NULL;
						
						int counter = 0;
						
						// DEBUGGING ..	
						cout << "currentRowAllSearchKeys: ";
						for(int i=0; i<currentRowAllSearchKeys.size(); i++){
							
							cout << currentRowAllSearchKeys[i] << ", ";
							
						}
						cout << endl << endl;
						
						for(int i=0; i<uniqueSearchKeys.size(); i++){
								
							// DEBUGGING ..	
							cout << "i: " << i << endl << endl;
							cout << "counter: " << counter << endl << endl;
							cout << "uniqueSearchKeys[i("<<i<<")]: " << uniqueSearchKeys[i] << endl << endl;

						
							// DEBUGGING ..	
							cout << "( ( ( i("<<i<<") % ( pow(indexKeyOffset, rowLevel)("<<(int)pow(indexKeyOffset, rowLevel)<<") * numberSearchKeysPerNode("<<numberSearchKeysPerNode<<") )("<<( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode )<<") )("<<( i % ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) )<<") == 0 )("<<( ( i % ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) == 0 )<<") && ( ( i("<<i<<") % pow(5, rowLevel)("<< (int)pow(indexKeyOffset, rowLevel)<<") * fanout("<<fanout<<") )("<<( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) )<<") != 0 )("<<( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) != 0 )<<") && ( i("<<i<<") > 1 )("<<( i > 1 )<<") ): " << ( ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) == 0 ) && ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) != 0 ) && ( i > 1 ) ) << endl << endl;
							
						
							// every 4th "searchKey" (after and not including 1st key) is a new parent ..
							// "( i % numberSearchKeysPerNode ) == 0" ..
							// (i.e.: 5th, 9th, 13th, 17th, etc .. ) ..
							
							// also, skip every 20th key "( i % fanout ) != 0" ..
							// (i.e.: 4th, 8th, 12th, 16th = one parent, 24th, 28th, 32nd, 36th = another parent, etc..)
							//if( ( ( i % numberSearchKeysPerNode ) == 0 ) && ( ( i % fanout ) != 0 ) && 
							//	  ( i > 1 ) ){
							
							if( ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) == 0 ) && ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) != 0 ) && 
								  ( i > 1 ) ){
								  
								if( parentSearchKeys == NULL ){
									parentSearchKeys = new vector<int>();
								
									// DEBUGGING ..	
									cout << "I - parentSearchKeys = new vector<int>() .." << endl << endl;
							
								}
							
								// DEBUGGING ..	
								cout << "I - parentSearchKeys->push_back( uniqueSearchKeys[i] ): " << uniqueSearchKeys[i] << endl << endl;
							
								parentSearchKeys->push_back( uniqueSearchKeys[i] );
								counter++;
								
							}
							
							
							// DEBUGGING ..	
							cout << "( ( ( counter("<<counter<<") % numberSearchKeysPerNode("<<numberSearchKeysPerNode<<") )("<<( counter % numberSearchKeysPerNode )<<") == 0 )("<<( ( counter % numberSearchKeysPerNode ) == 0 )<<") && ( counter("<<counter<<") > 0 )("<<( counter > 0 )<<") ): " << ( ( ( counter % numberSearchKeysPerNode ) == 0 ) && ( counter > 0 ) ) << endl << endl;
							
							// every 4th "parent searchKey" finishes a parent node ..
							if( ( ( counter % numberSearchKeysPerNode ) == 0 ) && ( counter > 0 ) ) {
								
								if( parentSearchKeysRow == NULL ) {
									parentSearchKeysRow = new vector< vector<int>* >();
									
									// DEBUGGING ..	
									cout << "II - parentSearchKeysRow = new vector< vector<int>* >(); .." << endl << endl;	
									
								}
					
								// DEBUGGING ..	
								cout << "II - parentSearchKeysRow->push_back( parentSearchKeys ): ";
								for(int i=0; i<parentSearchKeys->size(); i++){
									
									cout << parentSearchKeys->at(i) << ", ";
									
								}
								cout << endl;
					
								parentSearchKeysRow->push_back( parentSearchKeys );	
								
								// reset these parameters for the next parent-node ..
								counter = 0;
								parentSearchKeys = NULL;
								
							}
							
							
							// DEBUGGING ..	
							cout << "( ( i("<<i<<") % ( (int)pow(indexKeyOffset, rowLevel("<<rowLevel<<"))("<<(int)pow(indexKeyOffset, rowLevel)<<") * fanout("<<fanout<<") ) ) == 0 )("<<( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) == 0 )<<") && ( ( uniqueSearchKeys.size()("<<uniqueSearchKeys.size()<<") - i("<<i<<") )("<<( uniqueSearchKeys.size() - i )<<") <= ( (int)pow(indexKeyOffset, rowLevel("<<rowLevel<<"))("<<(int)pow(indexKeyOffset, rowLevel)<<") * numberSearchKeysPerNode("<<numberSearchKeysPerNode<<") )("<<( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode )<<") )("<<( ( uniqueSearchKeys.size() - i ) <= ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) )<<") && ( i("<<i<<") > 1 )("<<( i > 1 )<<"): " << ( ( ( i % ( (int)pow(indexKeyOffset, rowLevel) * fanout ) ) == 0 ) && ( ( uniqueSearchKeys.size() - i ) <= ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) && ( i > 1 ) ) << endl << endl;
							
							// need to handle following exceptions:
							// 1) 1 leaf node, no parent
							// 		"parentNodes" is empty .. done ..
						
							// 2) 1 parent for 5 child nodes, + 1 extra node, no parent
							// the last/extra node is a "fanout" node, and
							// you know its the last node because "uniqueSearchKeys" has less than ((5^L)*4) searchKeys left ..
							//if( ( ( i % ( (int)pow(5, rowLevel) * fanout ) ) == 0 ) && 
							//	( ( uniqueSearchKeys.size() - i ) <= ( (int)pow(5, rowLevel) * numberSearchKeysPerNode ) ) &&
							//	( i > 1 ) ){
							
							// for an index tree w/ 1 level of parents + root level, exception occurs at i == 20 ..
							// for trees w/ more levels, exception occurs at ( i % 20 * (int)pow(5, rowLevel) ) ..
							// (i.e.: rowLevel == 0: 20, 40, 60, 80 / rowLevel == 1: 100, 200, 300 400 / rowLevel == 2: 500, 1000, 1500, 2000) ..
							// (not necessarily intuitive .. i figured it out by hand!) ..
							//int exceptionValueIndex = 20;
							int exceptionValueIndex = ( numberSearchKeysPerNode * fanout );
							
							if( ( ( i % ( exceptionValueIndex * (int)pow(indexKeyOffset, rowLevel) ) ) == 0 ) && 
							    ( ( uniqueSearchKeys.size() - i ) <= ( (int)pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode ) ) &&
							    ( i > 1 ) ) {
							
								// parent node was generated for a single child node ..
								if ( extraParentNodeGenerated == false )
									extraParentNodeGenerated = true;

							
								cout << "III - last/extra node is \"fanout\" node (create parent for it) .." << endl << endl;
							
								// this should be necessary ..
								if( parentSearchKeys == NULL ){
									parentSearchKeys = new vector<int>();
									
									cout << "III - parentSearchKeys = new vector<int>() .." << endl << endl;
									
								}
							
								// create the parent "searchKey" ..
								
								// NOTE: this is an exception to the norm, because
								//		 all the "searchKeys" of the first (leftmost) child node must be 
								//		 less than its parent node "searchKey" ..
							
								// the parent "searchKey" should be 1 greater than the child's max value ..
								// (also, the "extra" node is at the end of the collection of keys) ..
								//int parentSearchKey = ( uniqueSearchKeys[ (uniqueSearchKeys.size() - 1) ] + 1 );
								int parentSearchKey = ( uniqueSearchKeys.back() + 1 );
							
								parentSearchKeys->push_back( parentSearchKey );
								
								cout << "III - parentSearchKeys->push_back( parentSearchKey ): " << parentSearchKey << endl << endl;
								
								// this should NOT be necessary ..
								//if( parentSearchKeysRow == NULL ) {
								//	parentSearchKeysRow = new vector< vector<int>* >();
									
								//	cout << "III - parentSearchKeysRow = new vector< vector<int>* >() .." << endl << endl;
									
								//}
									
								//parentSearchKeysRow->push_back( parentSearchKeys );	
								
								//cout << "III - parentSearchKeysRow->push_back( parentSearchKeys ) .." << endl << endl;
							
							}
							
						
						} // end "for()" ..
						
						cout << "Point A["<<rowLevel<<"].. " << endl << endl;
						
						// if parent searchKey finishes with less than 4 values,
						// still need to store it in the row ..
						// (also, no reason to store "parentSearchKeys" if it is NULL,
						//  i.e.: # leaf search keys <= 4 .. only |10 20 30 40| .. ) ..
						if( ( counter < numberSearchKeysPerNode ) &&
						    ( parentSearchKeys != NULL ) ) {
						    	
						    cout << "Point B["<<rowLevel<<"].. " << endl << endl;
						    
						    cout << "IV - (parentSearchKeysRow == NULL): " << (parentSearchKeysRow == NULL) << endl << endl;
							
							if( parentSearchKeysRow == NULL )
								parentSearchKeysRow = new vector< vector<int>* >();
					
							// DEBUGGING ..	
							cout << "IV - parentSearchKeysRow->push_back( parentSearchKeys ): ";
							for(int i=0; i<parentSearchKeys->size(); i++){
									
								cout << parentSearchKeys->at(i) << ", ";
									
							}
							cout << endl;
							
							cout << "IV - parentSearchKeys->size(): " << parentSearchKeys->size() << endl << endl;
					
							parentSearchKeysRow->push_back( parentSearchKeys );
							
							cout << "IV - parentSearchKeysRow->size(): " << parentSearchKeysRow->size() << endl << endl;
							
						}
							
						cout << "Point C["<<rowLevel<<"].. " << endl << endl;
						
						
						// add the row to the index (so long as there's a row to add) ..
						if( ( parentNodes_ == NULL ) && 
							( parentSearchKeysRow != NULL ) )
							parentNodes_ = new vector< vector< vector<int>* >* >();

						cout << "(parentNodes != NULL): " << (parentNodes_ != NULL) << endl << endl;

						cout << "Point D["<<rowLevel<<"].. " << endl << endl;

						if ( parentSearchKeysRow != NULL )
							parentNodes_->push_back( parentSearchKeysRow );
	
						cout << "(parentSearchKeysRow != NULL): " << (parentSearchKeysRow != NULL) << endl << endl;
	
						cout << "Point E["<<rowLevel<<"].. " << endl << endl;
						
						// set the next "row" of parent nodes ..
						
						// clear the current row of values ..
						currentRowAllSearchKeys.clear();
						
						// make the parent "searchKeys" the new current "searchKeys" row ..
						// (i.e.: make the new current row out of the previous parent row) ..

						if( parentSearchKeysRow != NULL ){
							
							cout << "Point F["<<rowLevel<<"].. " << endl << endl;
							
							cout << "parentSearchKeysRow->size(): " << parentSearchKeysRow->size() << endl << endl;
							
							for(int i=0; i<parentSearchKeysRow->size(); i++){
								
								if( parentSearchKeysRow->at(i) != NULL ){
									
									cout << "Point G["<<rowLevel<<"].. " << endl << endl;
									
									cout << "parentSearchKeysRow->at(i)->size(): " << parentSearchKeysRow->at(i)->size() << endl << endl;
							
									for(int j=0; j<parentSearchKeysRow->at(i)->size(); j++){
										
										if( parentSearchKeysRow->at(i)->at(j) != NULL ){
											
											cout << "Point H["<<rowLevel<<"].. " << endl << endl;
							
											currentRowAllSearchKeys.push_back( parentSearchKeysRow->at(i)->at(j) );
							
										}
							
									}	
									
								}
							
							}
							
						}

						cout << "Point I["<<rowLevel<<"].. " << endl << endl;
						
						for(int i=0; i<currentRowAllSearchKeys.size(); i++){
							
							cout << "currentRowAllSearchKeys[ i("<<i<<") ]: " << currentRowAllSearchKeys[i] << endl << endl;	
							
						}
						
						//loopCounter++;
						//cout << "loopCounter: " << loopCounter << endl << endl;
						//if ( loopCounter > 2 ) {
						//	stopInfiniteLoop = true;
						//}
						
						// initialize and set variable for each row that tracks whether 
						// parent of last set of nodes should have either
						// 1) 1 child PageID or 2) 2 or more child pageIDs ..
						if ( extraParentNodeGeneratedPerRow_ == NULL )
							extraParentNodeGeneratedPerRow_ = new vector<bool>();
							
						if ( extraParentNodeGenerated == true )
							extraParentNodeGeneratedPerRow_->push_back( true );
						else
							extraParentNodeGeneratedPerRow_->push_back( false );
						
						// reset this parameter for the next row ..
						extraParentNodeGenerated = false;
						
						rowLevel++;		//  [ i % pow(indexKeyOffset, rowLevel) * numberSearchKeysPerNode  == 0 ] && [ i % pow(5, rowLevel) * fanout != 0 ] ..
						
						
					} // end "while()" ..


					// DEBUGGING ..	
					if( parentNodes_ != NULL ){
						
						cout << "FINISHED PARENT NODES: " << endl << endl;
						
						cout << "parentNodes->size(): " << parentNodes_->size() << endl << endl;
						
						for(int i=0; i<parentNodes_->size(); i++){	// "parentNodes" ..
							
							cout << "parentNodes->at( i("<<i<<") )->size(): " << parentNodes_->at(i)->size() << endl << endl;
							
							for(int j=0; j<parentNodes_->at(i)->size(); j++){	// "parentSearchKeysRow" ..
								
								cout << "parentNodes->at( i("<<i<<") )->at( j("<<j<<") )->size(): " << parentNodes_->at(i)->at(j)->size() << endl << endl;
							
								for(int k=0; k<parentNodes_->at(i)->at(j)->size(); k++){		// "parentSearchKeys" ..
								
									cout << parentNodes_->at(i)->at(j)->at(k) << ", ";
								
								} // end 1 parent node ..
							
								cout << "	";
							
							} // end 1 parent node row ..
						
							cout << endl << endl;
							
						} // end all parent node rows ..
					
					} // end "if()" ..
		
	
} // end "createIndexSearchKeys()" ..


void Tester:: createIndexPages(){
	
	// Here is what is created so far:
					
	//            index nodes: "parentNodes" -> "parentSearchKeysRow" -> "parentSearchKeys"
					
	//             leaf nodes: "uniqueSearchKeys" ..
					
	// buckets for leaf nodes: "indexBuckets" ..
	
	cout << "Tester:: testStorageManager_createIndexPages() start .." << endl << endl;
	
	vector<IndexPage*>* indexPages_ = NULL;
	
	//vector< vector< vector<int>* >* >* parentNodes_ = new vector< vector< vector<int>* >* >();
	//vector<bool>* extraParentNodeGeneratedPerRow_ = new vector<bool>();
	
	int fanout = 5;
	
	//vector<int>* indexPageIDs = new vector<int>();
	
	int indexPageCounter = 0;
		//previousRowMaxPageID = 0;
	
	
	// need to make "IndexPages" for each node of the index ..
	if( indexPages_ == NULL ) 
		indexPages_ = new vector<IndexPage*>();
		
		
	//cout << "Tester:: createIndexPages() parentNodes_ == NULL: " << ( parentNodes_ == NULL ) << endl << endl;
		
					
	if( parentNodes_ != NULL ){


		IndexPage* indexPage;

		int maxPageID = 0;

		vector<int>* childIndexPageIDs;
		
		//cout << "Tester:: createIndexPages() Point A .." << endl << endl;
		
		
		// update the "IndexHeader" ..
		IndexHeader* indexHeader = new IndexHeader();

		//indexHeader->setTotalNumberOfIndexPages( indexPages_->size() );

		IndexHeaderEntry* indexHeaderEntry = NULL;
		
		IndexRows* indexRows = new IndexRows;
		IndexNodeRow* indexNodeRow = NULL;
		IndexNode* indexNode = NULL;
		
		//cout << "Tester:: createIndexPages() Point B .." << endl << endl;
		

		cout << "================================================" << endl << endl;

		// number of rows ..
		int parentNodesSize = parentNodes_->size();
		
		//cout << "Tester:: createIndexPages() Point C .." << endl << endl;
		
		cout << "i = (parentNodesSize - 1): " << (parentNodesSize - 1) << endl << endl;

		// start at the root node, work down to 1-level-above the leaf nodes ..
		// NOTE: "parentNodes" has rows of parent nodes, with ROOT node stored LAST) ..
		for(int i=(parentNodesSize - 1); i>=0; i--){

			//cout << "Tester:: createIndexPages() Point D["<<i<<"] .." << endl << endl;

			// make an index that starts with "0" at the root node and increases from there ..
			int rowNumber = ( ( parentNodesSize - 1 ) - i );
			
			//cout << "              parentNodesSize["<<i<<"]: " << parentNodesSize << endl << endl;
			cout << "        ( parentNodesSize - 1 )["<<i<<"]: " << (parentNodesSize - 1) << endl << endl;
			cout << "( ( parentNodesSize - 1 ) - i )["<<i<<"]: " << ((parentNodesSize - 1) - i) << endl << endl;
			cout << "                      rowNumber["<<i<<"]: " << rowNumber << endl << endl;


			// iterate through EACH parent node in the ROW ..
			int searchKeyRowSize = parentNodes_->at(i)->size();
			
			//cout << "Tester:: createIndexPages() Point E["<<i<<"] .." << endl << endl;
			
			// track the next row ..
			indexNodeRow = new IndexNodeRow;
			indexRows->addRow( indexNodeRow );
			//indexRows->addRow( new IndexNodeRow() );
			
			for( int j=0; j<searchKeyRowSize; j++ ){
				
			  	//cout << "Tester:: createIndexPages() Point F["<<i<<"]["<<j<<"] .." << endl << endl;
				
				// initialize a new "indexPage" to encapsulate this parent node ..
				indexPage = new IndexPage();

				// set the "searchKeys" for this parent node using the values gathered above ..
				indexPage->setSearchKeys( parentNodes_->at(i)->at(j) );
				
				//cout << "Tester:: createIndexPages() Point G["<<i<<"]["<<j<<"] .." << endl << endl;
				
				// track the "pageID", "parentID" for this node ..
				indexNode = new IndexNode();
				indexNode->setIndexPageID( indexPageCounter );
				
				//cout << "Tester:: createIndexPages() Point H["<<i<<"]["<<j<<"] .." << endl << endl;
				//cout << "Tester:: createIndexPages() Point I["<<i<<"]["<<j<<"] .." << endl << endl;
				
				int parentID = -1;
				
				cout << "rowNumber ["<<i<<"]["<<j<<"]: " << rowNumber << endl << endl;
				
				if( rowNumber == 0 ) 
					indexNode->setParentIndexPageID( -1 );		// root node has no parent ..
				else if ( rowNumber == 1 ) {
					
					indexNode->setParentIndexPageID( 0 );		// next row all have root node as parent ..
				
					parentID = 0;
				
				} else {
					
					// look at the previous row, 
					// get the "pageID" of the node that is the parent-of-the-current-node ..
					bool parentIDFound = false;
					
					vector<int>* childIndexPageIDs = NULL;

					// get the previous node row (as collection of IndexNodes) ..
					vector<IndexNode*>* nodeRow = indexRows->getRow( ( rowNumber - 1 ) )->getRow();
					
					cout << "nodeRow->size() ["<<i<<"]["<<j<<"]: " << nodeRow->size() << endl << endl;
					
					for( int k=0; k<nodeRow->size(); k++ ){
						
						// get the "childIndexPageIDs" for each "IndexNode" in this row ..
						childIndexPageIDs = nodeRow->at( k )->getChildIndexPageIDs();
						
						cout << "childIndexPageIDs->size() ["<<i<<"]["<<j<<"]: " << childIndexPageIDs->size() << endl << endl;
						
						for( int m=0; m<childIndexPageIDs->size(); m++ ){
							
							cout << "( childIndexPageIDs->at( m("<<m<<") )("<<childIndexPageIDs->at( m )<<") == indexPageCounter("<<indexPageCounter<<") ) ["<<i<<"]["<<j<<"]: " << ( childIndexPageIDs->at( m ) == indexPageCounter ) << endl << endl;
							
							if( childIndexPageIDs->at( m ) == indexPageCounter ){
								
								// parentID of current node is a pageID from a row above ..
								parentIDFound = true;
								parentID = nodeRow->at( k )->getIndexPageID();
								break;
							}
							
						} // end "for()" ..
						
						if( parentIDFound == true )
							break;
						
					} // end "for()" ..
					
					cout << "parentIDFound ["<<i<<"]["<<j<<"]:" << parentIDFound << endl << endl;
					cout << "parentID ["<<i<<"]["<<j<<"]: " << parentID << endl << endl;
					
					if( parentIDFound == true )
						indexNode->setParentIndexPageID( parentID );
					else
						indexNode->setParentIndexPageID( -1 );
						
				} // end "if()" ..
				
				//cout << "Tester:: createIndexPages() Point J["<<i<<"]["<<j<<"] .." << endl << endl;

				// EXCEPTIONAL CASES:
				// #2) 1 parent for 5 child nodes, + 1 extra node, no parent ..	
				int parentNodeLevel = ( parentNodes_->size() - 1 - i );	// root node = level 0 ( @ i == parentNodes->size() - 1 ) ..

				// check whether this "extra" parent node has either 1) 1 child, or 2) more than 1 child ..
								
				//cout << "Tester:: createIndexPages() Point K["<<i<<"]["<<j<<"] .." << endl << endl;
								
				bool hasTwoOrMoreChildSearchKeys = true;
				if( extraParentNodeGeneratedPerRow_->at( i ) == true )
					hasTwoOrMoreChildSearchKeys = false;

				//cout << "( extraParentNodeGeneratedPerRow->at(i("<<i<<") )("<<extraParentNodeGeneratedPerRow_->at(i)<<") == true )( maxPageID: "<<maxPageID<<" ): " << ( extraParentNodeGeneratedPerRow_->at(i) == true ) << endl << endl;
							
				bool parentNeedsOnlyOneChildId = false;
					
				if( ( parentNodeLevel >= 1 ) && 						// this parent node is at least 1 level below root node ..
					( j == ( parentNodes_->at(i)->size() - 1 ) ) &&		// last node in the row ..
					( parentNodes_->at(i)->at(j)->size() == 1 ) &&		// node only has 1 searchKey ..
					( hasTwoOrMoreChildSearchKeys == false ) ) {		// parent node was generated for a single child node ..
																		
					parentNeedsOnlyOneChildId = true;
								
				}
								
				//cout << "( parentNodeLevel("<<parentNodeLevel<<") >= 1 )( maxPageID: "<<maxPageID<<" ): " << ( parentNodeLevel >= 1 ) << endl << endl;
				//cout << "( j("<<j<<") == ( parentNodes->at(i)->size() - 1 )("<<( parentNodes_->at(i)->size() - 1 )<<") )( maxPageID: "<<maxPageID<<" ): " << ( j == ( parentNodes_->at(i)->size() - 1 ) ) << endl << endl;
				////cout << "extraParentNodeGenerated( maxPageID: "<<maxPageID<<" ): " << extraParentNodeGenerated << endl << endl;
				//cout << "hasTwoOrMoreChildSearchKeys( maxPageID: "<<maxPageID<<" ): " << hasTwoOrMoreChildSearchKeys << endl << endl;
				//cout << "parentNeedsOnlyOneChildId( maxPageID: "<<maxPageID<<" ): " << parentNeedsOnlyOneChildId << endl << endl;
							
				//cout << "================================================" << endl << endl;
							
				// find out how many child nodes this parent node has ..
				// (i.e.: if parent node has 2 searchKeys, it has 3 child nodes) ..
				int numberOfChildNodes = -1;
				if( parentNeedsOnlyOneChildId ){
								
					numberOfChildNodes = 1;
								
				} else {
								
					numberOfChildNodes = ( parentNodes_->at(i)->at(j)->size() + 1 );
								
				}

				//cout << "Tester:: createIndexPages() Point L["<<i<<"]["<<j<<"] .." << endl << endl;
							
				// initialize a new collection of child "pageIDs" ..
				childIndexPageIDs = new vector<int>();
				
				int newPageID = -1;

							
				// calculate the "pageIDs" for each child of this parent node ..
				for(int m=0; m<numberOfChildNodes; m++){
								
					// i.e.: if "maxPageID = 0", "newPageID = 1, 2, 3 .."
					//		 if "maxPageID = 3", "newPageID = 4, 5, 6 ..", etc ..
					newPageID = ( ( maxPageID + 1 ) + m );
								
					// collect the child "pageIDs" for use just below ..
					childIndexPageIDs->push_back( newPageID );
								
				}
				
				//cout << "Tester:: createIndexPages() Point M1["<<i<<"]["<<j<<"] .." << endl << endl;
				
				// track the childIDs for this row ..
				indexNode->setChildIndexPageIDs( childIndexPageIDs );
				
				//cout << "Tester:: createIndexPages() Point M2["<<i<<"]["<<j<<"] .." << endl << endl;
					
				// track this node by adding this node to the row of nodes ..
				//indexRows->getRow( rowNumber )->addIndexNode( indexNode );
				
				IndexNodeRow* inr = indexRows->getRow( rowNumber );
				
				//cout << "Tester:: createIndexPages() Point M3["<<i<<"]["<<j<<"] .." << endl << endl;
				
				if( inr != NULL )
					inr->addIndexNode( indexNode );
				
							
				//cout << "Tester:: createIndexPages() Point N["<<i<<"]["<<j<<"] .." << endl << endl;
							
				// update the latest "maxPageID" ..
				if( newPageID > maxPageID ){
					
					maxPageID = newPageID;
								
				}
				
				//cout << "Tester:: createIndexPages() Point N1["<<i<<"]["<<j<<"] .." << endl << endl;
				
							
				// set the child "pageIDs" to be those calculated here ..
				indexPage->setChildPageIDs( childIndexPageIDs );
				
				//cout << "Tester:: createIndexPages() Point N2["<<i<<"]["<<j<<"] .." << endl << endl;
							
				// add the "indexPage" to the collection ..
				indexPages_->push_back( indexPage );
				
				
				//cout << "Tester:: createIndexPages() Point O["<<i<<"]["<<j<<"] .." << endl << endl;
				
				
				int freeSearchKeyCount = 0;
				
				indexHeaderEntry = new IndexHeaderEntry();
				
				// if there is only 1 node in this tree, 
				// then for this node which has no parent, parentID = -1 ..
				// or,
				// if this node is the root node, parentID = -1 ..
				
				cout << "        ( parentNodesSize == 1 ) ["<<i<<"]["<<j<<"]: " << ( parentNodesSize == 1 ) << endl << endl;
				
				cout << "( i == ( parentNodesSize - 1 ) ) ["<<i<<"]["<<j<<"]: " << ( i == ( parentNodesSize - 1 ) ) << endl << endl;
				
				cout << "                      ( i >= 0 ) ["<<i<<"]["<<j<<"]: " << ( i >= 0 ) << endl << endl;	
				
				int pageID = -1;
				int parentPageID = -1;
				
				if( ( parentNodesSize == 1 ) || 				// there is only 1 node in the index ..
					( i == ( parentNodesSize - 1 ) ) ) {		// current node is the "root" node ..
					
					pageID = 0;

				} else if( i >= 0 ){							// Handle the non-root nodes ..
					
					//int parentID = indexRows->findParentPageID( indexPageCounter, rowNumber );
					
					pageID = indexPageCounter;
					
					parentPageID = parentID;
					
				} else {
					
					cout << "indexHeaderEntries not getting set .." << endl << endl;	
					
				}
				
				freeSearchKeyCount = ( (fanout - 1) - parentNodes_->at(i)->at(j)->size() );
				//freeSearchKeyCount = ( fanout - parentNodes_->at(i)->at(j)->size() );
				
				cout << "            pageID ["<<i<<"]["<<j<<"]: " << pageID << endl << endl;
				cout << "      parentPageID ["<<i<<"]["<<j<<"]: " << parentPageID << endl << endl;
				cout << "freeSearchKeyCount ["<<i<<"]["<<j<<"]: " << freeSearchKeyCount << endl << endl;
					
				indexHeaderEntry->setPageID( pageID );
				indexHeaderEntry->setParentPageID( parentPageID );
				indexHeaderEntry->setFreeSearchKeyCount( freeSearchKeyCount );
					
				vector<bool>* searchKeysAvailable = new vector<bool>();
					
				for( int k=0; k<( (fanout - 1) - freeSearchKeyCount ); k++ )
					searchKeysAvailable->push_back( false );
					
				for( int k=0; k<freeSearchKeyCount; k++ )
					searchKeysAvailable->push_back( true );
					
				for(int k=0; k<(fanout - 1);k++)
					cout << "searchKeysAvailable->at(k)("<<k<<") ["<<i<<"]["<<j<<"]: " << searchKeysAvailable->at(k) << endl << endl;
						
				indexHeaderEntry->setSearchKeysAvailable( searchKeysAvailable );
				

				// add the new "indexHeaderEntry" to the "IndexHeader" collection ..
				indexHeader->getHeaderEntries()->push_back( indexHeaderEntry );

				// tracks the current indexPage's "pageID" ..
				indexPageCounter++;
				
				//cout << "Tester:: createIndexPages() Point P["<<i<<"]["<<j<<"] .." << endl << endl;
				

			} // end "for()" of each node in row ..
			
			
			// set this parameter prior to going to next row / iteration ..
			//previousRowMaxPageID = maxPageID;
			
			//cout << "Tester:: createIndexPages() Point Q ["<<i<<"].." << endl << endl;


		} // end "for()" of each row ..
		
		
		//cout << "Tester:: createIndexPages() Point R .." << endl << endl;
		
		indexHeader->setTotalNumberOfIndexPages( indexPages_->size() );
		
		
		// DEBUGGING ..
		cout << "================================================" << endl << endl;
		
		cout << "totalNumberOfIndexPages: " << indexHeader->getTotalNumberOfIndexPages() << endl << endl;
					
		vector<IndexHeaderEntry*>* indexHeaderEntries = indexHeader->getHeaderEntries();
		
		IndexHeaderEntry* indexHeaderEntry1 = NULL;
		
		cout << "indexHeaderEntries->size(): " << indexHeaderEntries->size() << endl << endl;
					
		for(int i=0; i<indexHeaderEntries->size(); i++){
			
			indexHeaderEntry1 = indexHeaderEntries->at(i);

			cout << "pageID: " << indexHeaderEntry1->getPageID() << endl << endl;
			
			cout << "parentPageID: " << indexHeaderEntry1->getParentPageID() << endl << endl;
			
			cout << "freeSearchKeyCount: " << indexHeaderEntry1->getFreeSearchKeyCount() << endl << endl;
			
			vector<bool>* searchKeysAvailable = indexHeaderEntry1->getSearchKeysAvailable();
				
			for(int j=0; j<searchKeysAvailable->size(); j++){
							
				cout << "searchKeysAvailable["<<j<<"]: " << searchKeysAvailable->at(j) << endl << endl;
							
			}
						
			cout << endl << endl;
						
						
		} // end "for()" ..
		


		// DEBUGGING ..
		cout << "================================================" << endl << endl;
					
		for(int i=0; i<indexPages_->size(); i++){
						
			cout << "node: " << i << endl << endl;
						
						
			cout << "indexPages[ i("<<i<<") ] searchKeys: ";
						
			for(int j=0; j<indexPages_->at(i)->getSearchKeys()->size(); j++){
							
				cout << indexPages_->at(i)->getSearchKeys()->at( j ) << ", ";
							
			}
						
			cout << endl << endl;
						
						
			cout << "indexPages[ i("<<i<<") ] childPageIDs: ";
						
			for(int j=0; j<indexPages_->at(i)->getChildPageIDs()->size(); j++){
							
				cout << indexPages_->at(i)->getChildPageIDs()->at( j ) << ", ";
							
			}
						
			cout << endl << endl;
						
						
			cout << "================================================" << endl << endl;
						
						
		} // end "for()" ..
						
						
	} // end "if()" ..
	
	
} // end "createIndexPages" .


bool Tester:: writeIndexBucket() {
	
	// "tableName": name of the table associated with the index ..
	
	//    "indexBucket": indexBucket to write to file ..
	
	int blockSize = 48;
	
	int intSize = 4;
	
	int maxBucketSize = (blockSize / intSize);
	
	int bucketSize = 8;	// max = 12
	
	cout << "StorageManager:: writeIndexBucket() start .." << endl << endl;
	
 	char* tableName = "abc";
 	char* indexName = "I";
 	IndexBucket* indexBucket = new IndexBucket();
 	vector<int>* pageIDs = new vector<int>();
 	for(int i=0; i<bucketSize; i++)
 		pageIDs->push_back( i );
 	indexBucket->setPageIDs( pageIDs );
 	int bucketPageID = 0;
	
	cout << "StorageManager:: writeIndexBucket() tableName: " << tableName << endl << endl;
	cout << "StorageManager:: writeIndexBucket() indexBucket->getPageIDs()->size(): " << indexBucket->getPageIDs()->size() << endl << endl;
	cout << "StorageManager:: writeIndexBucket() bucketPageID: " << bucketPageID << endl << endl;
	
	bool writeBucketPagePerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName;

	if ( emptyString.compare( tableNameString ) != 0 ) {
		
		// handle the directory path of the file ..
	
		string directoryPlusFileName = "C:\\Temp\\";
		
	

		// append the file
		directoryPlusFileName += "IndexBuckets_";
		directoryPlusFileName += tableName;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName;
		char* fileName = (char*)directoryPlusFileName.c_str();
		
		fstream tbl( fileName, ios::in | ios::out |ios::binary );
		
		cout << "StorageManager:: writeIndexBucket() Point D .." << endl << endl;
		
		// this means the file doesn't exist yet, 
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){
			
			tbl.open( fileName, ios::out | ios::binary | ios::app );
			
			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();	
			
		}
		
		cout << "StorageManager:: writeIndexBucket() Point E .." << endl << endl;
		
		if ( !tbl.fail() ) {
			
			cout << "StorageManager:: writeIndexBucket() Point F .." << endl << endl;
	
			// make sure file pointer starts at beginning of file ..
			tbl.seekp( ios::beg );
			
			cout << "StorageManager:: writeIndexBucket() Point G .." << endl << endl;

			// so locate page-to-write referenced from beginning of file ..
			tbl.seekp( bucketPageID * blockSize );
			
			cout << "StorageManager:: writeIndexBucket() Point H .." << endl << endl;
		
			// number of values in the "bucket" ..
			int numberOfBucketValues = indexBucket->getPageIDs()->size();
			
			cout << "StorageManager:: writeIndexBucket() numberOfBucketValues: " << numberOfBucketValues << endl << endl;
	
			cout << "StorageManager:: writeIndexBucket() Point I .." << endl << endl;
	
			// size in bytes of the header file ..
			// [ (4 bytes) * (11 PageIDs) ] + [ (4 bytes) * (1 PageID) ] = 48 bytes
			// 1st 11 bytes are pointers to duplicate searchKeys, 
			// last 4 bytes reserved for pointer to next bucket ..
			int bucketSize = ( intSize * numberOfBucketValues );
			
			int fillerSize = ( blockSize - bucketSize );
			
			cout << "StorageManager:: writeIndexBucket() bucketSize: " << bucketSize << endl << endl;
			
			cout << "StorageManager:: writeIndexBucket() fillerSize: " << fillerSize << endl << endl;
				 
			// get the pageIDs from the bucket ..
			vector<int>* pageIDs = indexBucket->getPageIDs();
			
			cout << "StorageManager:: writeIndexBucket() Point J .." << endl << endl;

		
			int fieldValue = -1;
	
			// write the pageIDs in the IndexBucket to file ..
			for (int i=0; i < numberOfBucketValues; i++) {
			
				// DEBUGGING ..
				//cout << "StorageManager:: writeIndexBucket() pageIDs["<<i<<"]: " << pageIDs->at(i) << endl << endl;
			
				fieldValue = pageIDs->at(i);
				
				cout << "StorageManager:: writeIndexBucket() pageIDs->at(i("<<i<<")): " << pageIDs->at(i) << endl << endl;
			
				tbl.write( reinterpret_cast< const char * >( &fieldValue ), intSize );	

			}
	
	
			if ( fillerSize > 0 ) {
		
				// DEBUGGING ..
				//cout << "StorageManager:: writeDataPage() blockSize: " << blockSize << endl << endl;
				//cout << "StorageManager:: writeDataPage() bucketSize: " << bucketSize << endl << endl;
				//cout << "StorageManager:: writeDataPage() fillerSize ( blockSize("<<blockSize<<") - bucketSize("<<bucketSize<<") ): " << fillerSize << endl << endl;
	
				string filler = createFillerString( fillerSize );
		
				int charSize = 1;
		
				// write the filler string values to fill up a 48-byte-block-DataPage to file ..
				for(int i=0; i<filler.size(); i++){
			
					cout << filler[i];
		
					tbl.write( reinterpret_cast< const char * >( &( filler[i] ) ), charSize );
			
				}
		
				//cout << endl << endl;
		
			}
			
			IndexBucketHeader indexBucketHeader;
			bool isReadIndexBucketHeader = readIndexBucketHeader( tableName, indexName, indexBucketHeader );
	
			cout << "writeIndexBucket() isReadIndexBucketHeader: " << isReadIndexBucketHeader << endl << endl;
	
			if( isReadIndexBucketHeader ){
				
				cout << "StorageManager:: writeIndexBucket() Point K .." << endl << endl;
				
				vector<IndexBucketHeaderEntry*>* ibhes = indexBucketHeader.getHeaderEntries();
				
				int numberIBHES = ibhes->size();
				cout << "writeIndexBucket() numberIBHES: " << numberIBHES << endl << endl;
				for(int i=0; i<numberIBHES; i++) {
					
					if( ibhes->at(i)->getBucketID() == bucketPageID ){
						
						ibhes->at(i)->setFreePointerCount( ( maxBucketSize - numberOfBucketValues ) );
						
						cout << "writeIndexBucket() ( maxBucketSize - numberOfBucketValues ): " << ( maxBucketSize - numberOfBucketValues ) << endl << endl;
						
						break;
						
					}
					
				}
				
			} else {
				
				cout << "StorageManager:: writeIndexBucket() Point L .." << endl << endl;
				
				IndexBucketHeaderEntry* ibhe = new IndexBucketHeaderEntry();
				
				cout << "writeIndexBucket() bucketPageID: " << bucketPageID << endl << endl;
				
				ibhe->setBucketID( bucketPageID );
				
				ibhe->setFreePointerCount( ( maxBucketSize - numberOfBucketValues ) );
				
				cout << "writeIndexBucket() ( maxBucketSize - numberOfBucketValues ): " << ( maxBucketSize - numberOfBucketValues ) << endl << endl;
				
				indexBucketHeader.getHeaderEntries()->push_back( ibhe );
				
				indexBucketHeader.setTotalNumberBuckets( 1 );
				
			}
			
			bool isWriteIndexBucketHeader = writeIndexBucketHeader( tableName, indexName, indexBucketHeader );
	
	
			// flag indicating the "writeIndexBucket()" operation
			// was performed successfully (header read, file created, file opened) ..
			writeBucketPagePerformed = true;
		
		
		} else {
		
	 		string errorString = "\nFailed to open data file [ SM::writeIndexBucket() ], so index bucket was not written to.";
	 	
	 		cout << errorString << endl;
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringQM( errorString );	
		
		} // end "if ( !tbl.fail() )" ..
	
	
		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();
		
		
	} // end "if ( emptyString.compare( tableNameString ) != 0 )" ..

	
	//cout << "SM:: writeIndexBucket() end .." << endl;
	
	return writeBucketPagePerformed;
	

} // end "writeIndexBucket()" ..


//bool Tester:: writeIndexPage() {
bool Tester:: writeIndexPage( char* tableName_, char* indexName_, IndexPage* indexPage_, int pageID_, int parentPageID_, bool isNew_ ) {
	
	// "tableName": name of the table associated with the index ..
	
	// "indexPage": indexPage to write to file ..
	
	//    "pageID": pageID associated with this indexPage ..
	
	if( indexPage_ == NULL ) {
	
		tableName_ = "abc";
		indexName_ = "I";
		indexPage_ = new IndexPage();
		vector<int>* searchKeys_ = new vector<int>();
		vector<int>* childPageIDs_ = new vector<int>();
		for(int i=0; i<3; i++)				// max = 4
			searchKeys_->push_back( i );

		for(int i=0; i<4; i++)				// max = 5
			childPageIDs_->push_back( i );

		indexPage_->setSearchKeys( searchKeys_ );
		indexPage_->setChildPageIDs( childPageIDs_ );
		pageID_ = 0;
		parentPageID_ = -1;
	
	}
	
	
	
	int blockSize = 48;
	
	int intSize = 4;
	
	int fanout = 5;
	
	cout << "StorageManager:: writeIndexBucket() start .." << endl << endl;
	cout << "StorageManager:: writeIndexBucket() tableName: " << tableName_ << endl << endl;
	cout << "StorageManager:: writeIndexBucket() indexPage_->getChildPageIDs()->size(): " << indexPage_->getChildPageIDs()->size() << endl << endl;
	cout << "StorageManager:: writeIndexBucket() indexPage_->getSearchKeys()->size(): " << indexPage_->getSearchKeys()->size() << endl << endl;
	
	bool writeIndexPagePerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {
		
		// handle the directory path of the file ..
	
		string directoryPlusFileName = "C:\\Temp\\";
	

		// append the file
		directoryPlusFileName += "Index_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();
		
		fstream tbl( fileName, ios::in | ios::out | ios::binary );
		
		cout << "StorageManager:: writeIndexPage() Point D .." << endl << endl;
		
		// this means the file doesn't exist yet, 
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){
			
			tbl.open( fileName, ios::out | ios::binary | ios::app );
			
			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();	
			
		}
		
		cout << "StorageManager:: writeIndexPage() Point E .." << endl << endl;
		
		if ( !tbl.fail() ) {

			cout << "StorageManager:: writeIndexPage() Point F .." << endl << endl;
	
			// make sure file pointer starts at beginning of file ..
			tbl.seekp( ios::beg );
			
			cout << "StorageManager:: writeIndexPage() Point G .." << endl << endl;

			// so locate page-to-write referenced from beginning of file ..
			tbl.seekp( pageID_ * blockSize );
			
			cout << "StorageManager:: writeIndexPage() Point H .." << endl << endl;
		
			// number of values in the "IndexPage" ..
			int numberOfChildPageIDValues = indexPage_->getChildPageIDs()->size();
			int numberOfSearchKeyValues = indexPage_->getSearchKeys()->size();
			int numberOfIndexPageValues = numberOfChildPageIDValues + numberOfSearchKeyValues;
			
			
			cout << "StorageManager:: writeIndexPage() numberOfChildPageIDValues: " << numberOfChildPageIDValues << endl << endl;
	
			cout << "StorageManager:: writeIndexPage() Point I .." << endl << endl;
	
			// size in bytes of the header file ..
			// [ (4 bytes) * (6 child page ids) ] + [ (4 bytes) * (5 search keys) ] = 24 + 20 = 44 bytes
			int indexPageSize = ( intSize * numberOfIndexPageValues );
			
			int fillerSize = ( blockSize - indexPageSize );
			
			cout << "StorageManager:: writeIndexPage() indexPageSize: " << indexPageSize << endl << endl;
			cout << "StorageManager:: writeIndexPage() fillerSize: " << fillerSize << endl << endl;
				 
			// get the pageIDs from the bucket ..
			vector<int>* childPageIDs = indexPage_->getChildPageIDs();
			
			vector<int>* searchKeys = indexPage_->getSearchKeys();
			
			
			cout << "StorageManager:: writeIndexPage() Point J .." << endl << endl;
		
			int fieldValue = -1;
	
			// write the childPageID / searchKey pairs in the IndexPage to file ..
			for (int i=0; i < numberOfChildPageIDValues; i++) {
					
				// "childPageID" is only written when "i" is "even" ..
					
				fieldValue = childPageIDs->at(i);
				
				tbl.write( reinterpret_cast< const char * >( &fieldValue ), intSize );

				// DEBUGGING ..
				cout << "StorageManager:: writeIndexPage() childPageIDs->at(i("<<i<<")): " << childPageIDs->at(i) << endl << endl;
			

			}
			
			// write the childPageID / searchKey pairs in the IndexPage to file ..
			for (int i=0; i < numberOfSearchKeyValues; i++) {

				// "pageIDs" (#1 ptr) - (#5 ptr) ..
					
				fieldValue = searchKeys->at(i);
				
				tbl.write( reinterpret_cast< const char * >( &fieldValue ), intSize );

				// DEBUGGING ..
				cout << "StorageManager:: writeIndexPage() searchKeys->at(i("<<i<<")): " << searchKeys->at(i) << endl << endl;
			

			}
			
			// write the final (#6 ptr) "childPageID" to file ..
			//fieldValue = childPageIDs->back();	
			//tbl.write( reinterpret_cast< const char * >( &fieldValue ), intSize );
	
			cout << "StorageManager:: writeIndexPage() Point K .." << endl << endl;
	
			if ( fillerSize > 0 ) {
				
				cout << "StorageManager:: writeIndexPage() Point L .." << endl << endl;
		
				// DEBUGGING ..
				//cout << "StorageManager:: writeDataPage() blockSize: " << blockSize << endl << endl;
				//cout << "StorageManager:: writeDataPage() bucketSize: " << bucketSize << endl << endl;
				//cout << "StorageManager:: writeDataPage() fillerSize ( blockSize("<<blockSize<<") - bucketSize("<<bucketSize<<") ): " << fillerSize << endl << endl;
	
				string filler = createFillerString( fillerSize );
				
				cout << "StorageManager:: writeIndexPage() Point M .." << endl << endl;
		
				int charSize = 1;
		
				// write the filler string values to fill up a 48-byte-block-DataPage to file ..
				for(int i=0; i<filler.size(); i++){
			
					cout << filler[i];
		
					tbl.write( reinterpret_cast< const char * >( &( filler[i] ) ), charSize );
			
				}
				
				cout << "\nStorageManager:: writeIndexPage() Point N .." << endl << endl;
		
				//cout << endl << endl;
		
			}
	
			// flag indicating the "writeIndexPage()" operation
			// was performed successfully (header read, file created, file opened) ..
			writeIndexPagePerformed = true;
		
			// update the "index header" ..
			IndexHeader header;
			
			cout << "StorageManager:: writeIndexPage() Point O .." << endl << endl;
			
			bool isIndexHeaderRead = readIndexHeader( tableName_, indexName_, header );
			
			int totalNumberOfIndexPages = 0;
			
			cout << "StorageManager:: writeIndexPage() isIndexHeaderRead: " << isIndexHeaderRead << endl << endl;
			
			cout << "StorageManager:: writeIndexPage() Point P .." << endl << endl;
			
			// if this is a new "indexPage" tracked in the header, 
			// update info accordingly below ..	
			if( isNew_ == true ) {
		
				if( isIndexHeaderRead == true )
					totalNumberOfIndexPages = header.getTotalNumberOfIndexPages();
				else
					totalNumberOfIndexPages = 0;
		
				header.setTotalNumberOfIndexPages( ( totalNumberOfIndexPages + 1 ) );
		
			}
			
			cout << "StorageManager:: writeIndexPage() Point Q .." << endl << endl;
			
			IndexHeaderEntry* ihe = new IndexHeaderEntry();
			
			ihe->setPageID( pageID_ );
			ihe->setNumberOfSearchKeys( numberOfSearchKeyValues );
			ihe->setNumberOfChildPageIDs( numberOfChildPageIDValues );
			ihe->setParentPageID( parentPageID_ );
			ihe->setFreeSearchKeyCount( ( (fanout - 1) - numberOfSearchKeyValues) );
			
			vector<bool>* searchKeysAvailable = new vector<bool>();
			for(int i=0; i<numberOfSearchKeyValues; i++)
				searchKeysAvailable->push_back( false );
				
			for(int i=0; i<( (fanout - 1) - numberOfSearchKeyValues); i++)
				searchKeysAvailable->push_back( true );
			
			ihe->setSearchKeysAvailable( searchKeysAvailable );
			
			header.getHeaderEntries()->push_back( ihe );
			
			cout << "StorageManager:: writeIndexPage() Point R .." << endl << endl;
			
			writeIndexHeader( tableName_, "I", header );
			
			cout << "StorageManager:: writeIndexPage() Point S .." << endl << endl;
		
		} else {
		
	 		string errorString = "\nFailed to open data file [ SM::writeIndexPage() ], so index bucket was not written to.";
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringQM( errorString );	
		
		} // end "if ( !tbl.fail() )" ..
	
	
		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();
		
		
	} // end "if ( emptyString.compare( tableNameString ) != 0 )" ..

	
	//cout << "SM:: writeIndexPage() end .." << endl;
	
	return writeIndexPagePerformed;
	

} // end "writeIndexPage()" ..


bool Tester:: writeIndexPages(){
	
	cout << "Tester:: testStorageManager_writeIndexPages() start .." << endl << endl;
	
	char* tableName_ = "abc";
	
	char* indexName_ = "I";
	
	IndexPage* indexPage_ = NULL;
	vector<int>* searchKeys_ = NULL;
	vector<int>* childPageIDs_ = NULL;
	
	vector<IndexPage*>* indexPages_ = new vector<IndexPage*>();
	
	for(int i=0; i<5; i++){
		
		indexPage_ = new IndexPage();
		searchKeys_ = new vector<int>();
		childPageIDs_ = new vector<int>();
		
		for(int j=0; j<3; j++)				// max = 4
			searchKeys_->push_back( j );

		for(int j=0; j<4; j++)				// max = 5
			childPageIDs_->push_back( j );

		indexPage_->setSearchKeys( searchKeys_ );
		indexPage_->setChildPageIDs( childPageIDs_ );
		
		indexPages_->push_back( indexPage_ );
		
	}
	
	bool wasIndexPageWritten = false;
	bool wereIndexPagesWritten = true;
	
	for(int i=0; i<indexPages_->size(); i++){
						
		wasIndexPageWritten = writeIndexPage( tableName_, indexName_, indexPages_->at( i ), i, -1, true );
		
		if( wasIndexPageWritten == false )
			wereIndexPagesWritten = false;
						
	}
	
	return wereIndexPagesWritten;
	
} // end "writeIndexPages()" ..


bool Tester:: writeIndexBucketHeader( char* tableName_, char* indexName_, IndexBucketHeader& indexBucketHeader_ ) {
	
	// "tableName": name of the table associated with the header ..
	
	//    "header": header to write to file ..
	
	cout << "testStorageManager_writeIndexBucketHeader() start .." << endl << endl;
	
	int blockSize = 48;
	
	int intSize = 4;
	
	int boolSize = 2;

	cout << "testStorageManager_writeIndexBucketHeader() Point A .." << endl << endl;
	
	// start test input data ..
	
	if( &indexBucketHeader_ == NULL ){
		
		tableName_ = "abc";
		indexName_ = "I";
		
		int numberBucketIDsPerSearchKey = 5;
		int numberSearchKeys = 4;
	
		indexBucketHeader_.setTotalNumberBuckets( (numberBucketIDsPerSearchKey * numberSearchKeys) );
	
		map<int, vector<int>* >* _indexBucketIDsForEachSearchKey = new map<int, vector<int>* >();
	
		vector<int>* bucketIDs = new vector<int>();
	
		int increment = 0;
	
		for(int i=0; i<numberSearchKeys; i++) {
	
			bucketIDs = new vector<int>();
	
			for(int j=0; j<numberBucketIDsPerSearchKey; j++)
				bucketIDs->push_back( (j + increment) );
		
			_indexBucketIDsForEachSearchKey->insert( pair<int, vector<int>* >( i, bucketIDs ) );
	
			increment++;
	
			//delete bucketIDs;
		
		}
		
		cout << "testStorageManager_writeIndexBucketHeader() Point B .." << endl << endl;
	
		indexBucketHeader_.setBucketIDsForEachSearchKey( _indexBucketIDsForEachSearchKey );
	
		vector<IndexBucketHeaderEntry*>* ibhev = new vector<IndexBucketHeaderEntry*>();
	
		IndexBucketHeaderEntry* ibhe = NULL;
	
		//vector<bool>* pointersAvailable = NULL;
	
		for(int i=0; i<numberSearchKeys; i++) {
		
			for(int j=0; j<numberBucketIDsPerSearchKey; j++) {
		
				ibhe = new IndexBucketHeaderEntry();
		
				ibhe->setBucketID( ( j + increment ) );
			
				ibhe->setFreePointerCount( 5 );
			
				//pointersAvailable = new vector<bool>();

				//for(int k=0; k<5; k++)
				//	pointersAvailable->push_back(true);
			
				//ibhe->setPointersAvailable( pointersAvailable );
				
				ibhev->push_back( ibhe );
			
			}
		
			increment++;
		
		}
	
	
		indexBucketHeader_.setHeaderEntries( ibhev );
		
		
	} // end "if()" ..
	
	
	cout << "testStorageManager_writeIndexBucketHeader() Point C .." << endl << endl;
	
	bool writeHeaderPerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {
		
	
	cout << "testStorageManager_writeIndexBucketHeader() Point D .." << endl << endl;
	
		// handle the directory path of the file ..
	
		string directoryPlusFileName = "C:\\Temp\\";
	
		// append the file
		directoryPlusFileName += "IndexBucketHeader_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();
		
		fstream tbl( fileName, ios::in | ios::out |ios::binary );
		
		// this means the file doesn't exist yet, 
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){
			
			tbl.open( fileName, ios::out | ios::binary | ios::app );
			
			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();	
			
		}
		
		if ( !tbl.fail() ) {
	
			cout << "testStorageManager_writeIndexBucketHeader() Point E .." << endl << endl;
	
			tbl.seekp( ios::beg );


			vector<IndexBucketHeaderEntry*>* _headerEntries; 			


			// write the "header" fields to the header file ..


			// "int _totalNumberOfBuckets"
			int tnb = indexBucketHeader_.getTotalNumberBuckets();
			tbl.write( reinterpret_cast< const char * >( &( tnb ) ), intSize );
			
			cout << "testStorageManager_writeIndexBucketHeader() tbl.write( tnb ): " << tnb << endl << endl;
		
			cout << "testStorageManager_writeIndexBucketHeader() Point F .." << endl << endl;
	
			// write the "headerEntry" fields to the header file ..
			
			
			// "map<int, vector<int>* >* _indexBucketIDsForEachSearchKey"
			map<int, vector<int>* >* iBIDsFESK = NULL;
			int sk = 0;						// search key
			int numberBucketIDs;			// # bucketIDs in vector<int>*
			vector<int>* bucketIDs = NULL;
			int bucketIDForSearchKey;		// each bucketID in vector<int>*
			int searchKeySize = 0,
				numberSearchKeySize = 0,
				bucketIDsForSearchKeySize = 0;
			
			map<int, vector<int>* >::iterator it_map;
			iBIDsFESK = indexBucketHeader_.getBucketIDsForEachSearchKey();
			for( it_map = iBIDsFESK->begin(); it_map != iBIDsFESK->end(); it_map++ ){
					
				// search key for a set of bucket ids ..
				sk = it_map->first;
				tbl.write( reinterpret_cast< const char * >( &( sk ) ), intSize );
				searchKeySize++;
				cout << "testStorageManager_writeIndexBucketHeader() tbl.write( sk ): " << sk << endl << endl;
					
				// number of bucket ids for this search key ..
				numberBucketIDs = ( it_map->second )->size();
				tbl.write( reinterpret_cast< const char * >( &( numberBucketIDs ) ), intSize );
				numberSearchKeySize++;
				cout << "testStorageManager_writeIndexBucketHeader() tbl.write( numberBucketIDs ): " << numberBucketIDs << endl << endl;
					
				// bucket ids for this search key ..
				//map<int, vector<int>* >::iterator it_vector;
				bucketIDs = it_map->second;
					
				for( int i=0; i < bucketIDs->size(); i++ ){
						
					bucketIDForSearchKey = bucketIDs->at(i);
					tbl.write( reinterpret_cast< const char * >( &( bucketIDForSearchKey ) ), intSize );
					bucketIDsForSearchKeySize++;
					cout << "testStorageManager_writeIndexBucketHeader() tbl.write( bucketIDForSearchKey ): " << bucketIDForSearchKey << endl << endl;
						
				}

			} // end "for" ..
	
			cout << "testStorageManager_writeIndexBucketHeader() Point G .." << endl << endl;
	
			// "flag" indicating that "header entry" fields are beginning ..
			int startHeaderEntries = -1;
			tbl.write( reinterpret_cast< const char * >( &( startHeaderEntries ) ), intSize );
			cout << "testStorageManager_writeIndexBucketHeader() tbl.write( startHeaderEntries ): " << startHeaderEntries << endl << endl;
	
			cout << "testStorageManager_writeIndexBucketHeader() Point H .." << endl << endl;
	
			int bi = 0;
			
			int fpc = 0;
			
			bool pa = false;
			
			vector<bool>* pointersAvailable = NULL;
			
			int numberOfSimpleHeaderEntryFields = 2;
			
			int pointersAvailableSize = 0;
			
			int totalNumberOfBuckets = indexBucketHeader_.getTotalNumberBuckets();

			for(int i=0; i<totalNumberOfBuckets; i++){
		
				bi = indexBucketHeader_.getHeaderEntries()->at( i )->getBucketID();
				tbl.write( reinterpret_cast< const char * >( &( bi ) ), intSize );
				cout << "testStorageManager_writeIndexBucketHeader() tbl.write( bi ): " << bi << endl << endl;
		
				fpc = indexBucketHeader_.getHeaderEntries()->at( i )->getFreePointerCount();
				tbl.write( reinterpret_cast< const char * >( &( fpc ) ), intSize );
				cout << "testStorageManager_writeIndexBucketHeader() tbl.write( fpc ): " << fpc << endl << endl;
			
				//pointersAvailable = indexBucketHeader_.getHeaderEntries()->at( i )->getPointersAvailable();

				//for(int j=0; j<pointersAvailable->size(); j++) {
		
				//	pa = pointersAvailable->at( j );
		
				//	tbl.write( reinterpret_cast< const char * >( &( pa ) ), boolSize );
				//	pointersAvailableSize++;
				//	cout << "writeIndexBucketHeader() tbl.write( pa ): " << pa << endl << endl;

				//}	
		
			}
			
			cout << "testStorageManager_writeIndexBucketHeader() Point I .." << endl << endl;
			
			// 4 bytes + [ 4 bytes * [keys + num + bucketIDs] ]
			int headerFieldsSize = intSize + ((searchKeySize + numberSearchKeySize + bucketIDsForSearchKeySize) * intSize);
		
			// [4 bytes * num fields] + [4 bytes * pointers available]
			//int headerEntriesSize =  ( ( intSize * numberOfSimpleHeaderEntryFields ) + ( intSize * pointersAvailableSize ) );
			int headerEntriesSize =  ( intSize * numberOfSimpleHeaderEntryFields * totalNumberOfBuckets );
	
			// size in bytes of the header file ..
			int headerSize = ( headerFieldsSize + headerEntriesSize );
			
	
			int fillerSize = 0;
		
			int gaugeTheFillerSize = blockSize;
		
		
			// increase "gaugeTheFillerSize" by increments of "blockSize",
			// until "gaugeTheFillerSize >= headerSize" ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			while ( true ){
				
				if( gaugeTheFillerSize >= headerSize ){

					break;
					
				} else {
					
					gaugeTheFillerSize += blockSize;
				
				}
					
			} // end "while ( true )" ..
		
		
			cout << "testStorageManager_writeIndexBucketHeader() Point J .." << endl << endl;
		
			// "fillerSize" should be less than 48 bytes, 
			// such that adding "fillerSize" to "headerSize"
			// is an even multiple of "blockSize" ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			fillerSize = ( gaugeTheFillerSize - headerSize );
		
	
			if ( fillerSize > 0 ) {
			
				string filler = createFillerString( fillerSize );
			
				int charSize = 1;
			
				// write the filler string to the DataHeader file,
				// so that the header file size is an even multiple of "blockSize" ..
				for(int i=0; i<filler.size(); i++){
				
					tbl.write( reinterpret_cast< const char * >( &( filler[i] ) ), charSize );
			
				}
			
			}
			
			cout << "testStorageManager_writeIndexBucketHeader() Point K .." << endl << endl;
		
			// flag indicating the "writeHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			writeHeaderPerformed = true;


		} else {
		
	 		string errorString = "\nHeader file failed to open [ SM::writeIndexHeader() ], so Header file was not written.";
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringQM( errorString );
			
		} // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();
		
		
	} else {
		
	 	string errorString = "\nHeader file table name was not valid [ SM::writeIndexHeader() ], so Header file was not written.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		//getMessageManager()->setErrorStringQM( errorString );
		
	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	
	return writeHeaderPerformed;
	

} // end "writeIndexBucketHeader()" ..


//bool Tester:: writeIndexHeader() {
bool Tester:: writeIndexHeader( char* tableName_, char* indexName_, IndexHeader& header_ ) {
	
	// "tableName": name of the table associated with the header ..
	
	//    "header": header to write to file ..
	
	
	// start set input ..
	
	cout << "testStorageManager_writeIndexHeader() start .." << endl << endl;
	
	int fanout = 5;
	
	int blockSize = 48;
	
	int intSize = 4;
	
	int boolSize = 1;
	
	cout << "testStorageManager_writeIndexHeader() Point A .." << endl << endl;
	
	if( tableName_ == "" ) {
		
		tableName_ = "abc";
		indexName_ = "I";
		
		IndexHeader header_;
		int numberEntries = 5;
		int numberSearchKeysAvailable = 4;
		vector<IndexHeaderEntry*>* ihev = new vector<IndexHeaderEntry*>();
		IndexHeaderEntry* ihe = NULL;
		vector<bool>* keysAvailable = new vector<bool>();
		for(int i=0; i<numberEntries; i++){
	
			ihe = new IndexHeaderEntry();
		
			ihe->setPageID( i );
		
			ihe->setNumberOfSearchKeys( (fanout - 1) );
		
			ihe->setNumberOfChildPageIDs( fanout );
		
			ihe->setParentPageID( (2*i) );
		
			ihe->setFreeSearchKeyCount( (numberSearchKeysAvailable - i) );
		
			for(int j=0; j<numberSearchKeysAvailable; j++)
				keysAvailable->push_back( true );
		
			ihe->setSearchKeysAvailable( keysAvailable );
		
			ihev->push_back( ihe );
		
		}
	
		cout << "testStorageManager_writeIndexHeader() Point B .." << endl << endl;
	
		header_.setHeaderEntries( ihev );
		
		cout << "testStorageManager_writeIndexHeader() header_.getHeaderEntries()->size(): " << header_.getHeaderEntries()->size() << endl << endl;
	
		header_.setTotalNumberOfIndexPages( header_.getHeaderEntries()->size() );
		
	}
	
	// end set input ..
	
	bool writeHeaderPerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {
		
		cout << "testStorageManager_writeIndexHeader() Point C .." << endl << endl;
	
		// handle the directory path of the file ..
	
		string directoryPlusFileName = "C:\\Temp\\";
	

		// append the file
		directoryPlusFileName += "IndexHeader_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();
		
		fstream tbl( fileName, ios::in | ios::out |ios::binary );
		
		cout << "testStorageManager_writeIndexHeader() Point D .." << endl << endl;
		
		// this means the file doesn't exist yet, 
		// so attempt to create the file ready for output ..
		if( tbl.fail() ){
			
			tbl.open( fileName, ios::out | ios::binary | ios::app );
			
			// needed in order to "clear" errorflags left over from first "fail" ..
			tbl.clear();	
			
		}
		
		if ( !tbl.fail() ) {
			
			cout << "testStorageManager_writeIndexHeader() Point E .." << endl << endl;
	
			tbl.seekp( ios::beg );
		
			// number of fields in the "header" ..
			int numberOfHeaderFields = 1;
		
			// number of fields in each "headerEntry" of the "header" ..
			int numberOfHeaderEntryFields = 5; 
			
			int totalNumberOfIndexPages = header_.getTotalNumberOfIndexPages();
				
			cout << "testStorageManager_writeIndexHeader() Point F .." << endl << endl;
			
			// write the "header" fields to the header file ..
			tbl.write( reinterpret_cast< const char * >( &( totalNumberOfIndexPages ) ), intSize );
		
			cout << "testStorageManager_writeIndexHeader() Point G .." << endl << endl;
	
			// write the "headerEntry" fields to the header file ..
	
			int pid = 0;
			
			int nosk = 0;
			
			int nocpi = 0;
	
			int ppid = 0;
	
			int fskc = 0;
		
			vector<bool>* ska = NULL;
	
			cout << "testStorageManager_writeIndexHeader() Point H .." << endl << endl;

			for(int i=0; i<header_.getTotalNumberOfIndexPages(); i++){
		

				pid = header_.getHeaderEntries()->at( i )->getPageID();
				tbl.write( reinterpret_cast< const char * >( &( pid ) ), intSize );
				cout << "testStorageManager_writeIndexHeader() pid: " << pid << endl << endl;
		
				nosk = header_.getHeaderEntries()->at( i )->getNumberOfSearchKeys();
				tbl.write( reinterpret_cast< const char * >( &( nosk ) ), intSize );
				cout << "testStorageManager_writeIndexHeader() nosk: " << nosk << endl << endl;
				
				nocpi = header_.getHeaderEntries()->at( i )->getNumberOfChildPageIDs();
				tbl.write( reinterpret_cast< const char * >( &( nocpi ) ), intSize );
				cout << "testStorageManager_writeIndexHeader() nocpi: " << nocpi << endl << endl;
		
				ppid = header_.getHeaderEntries()->at( i )->getParentPageID();
				tbl.write( reinterpret_cast< const char * >( &( ppid ) ), intSize );
				cout << "testStorageManager_writeIndexHeader() ppid: " << ppid << endl << endl;
		
				fskc = header_.getHeaderEntries()->at( i )->getFreeSearchKeyCount();
				tbl.write( reinterpret_cast< const char * >( &( fskc ) ), intSize );
				cout << "testStorageManager_writeIndexHeader() fskc: " << fskc << endl << endl;
			
				ska = header_.getHeaderEntries()->at( i )->getSearchKeysAvailable();
				bool searchKeyAvailable = true;
				for(int j=0; j<ska->size(); j++) {
		
					searchKeyAvailable = ska->at( j );
		
					tbl.write( reinterpret_cast< const char * >( &( searchKeyAvailable ) ), boolSize );
					
					cout << "testStorageManager_writeIndexHeader() searchKeyAvailable: " << searchKeyAvailable << endl << endl;

				}
		
			}
			
			cout << "testStorageManager_writeIndexHeader() Point I .." << endl << endl;
	
			// size in bytes of the "rowsAvailable" array found in each "headerEntry" of the "header" ..
			// [ (2 bytes) * (N rows per page) ]
			int searchKeysAvailableSize = ( boolSize * ska->size() );
		
			// size in bytes of a single "headerEntry" of the "header" ..
			// [ [ (4 bytes) * (4 fields) ] +  ]
			int headerPageEntrySize =  ( ( intSize * numberOfHeaderEntryFields ) + searchKeysAvailableSize );
	
			// size in bytes of the header file ..
			// [ (4 bytes) * (2 fields) ] + [ (M data pages) * (single header entry size) ]
			// 8 bytes + M*16 bytes == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			int headerSize = ( ( intSize * numberOfHeaderFields ) + ( totalNumberOfIndexPages * headerPageEntrySize ) );
			
	
			int fillerSize = 0;
		
			int gaugeTheFillerSize = blockSize;
		
		
			// increase "gaugeTheFillerSize" by increments of "blockSize",
			// until "gaugeTheFillerSize >= headerSize" ..
			// i.e.:         headerSize == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			//		      exit the loop: (48 >= 24), (48 >= 40), (96 >= 56), (96 >= 72), (96 >= 88) ..
			while ( true ){
				
				if( gaugeTheFillerSize >= headerSize ){

					break;
					
				} else {
					
					gaugeTheFillerSize += blockSize;
				
				}
					
			} // end "while ( true )" ..
		
		
			cout << "testStorageManager_writeIndexHeader() Point J .." << endl << endl;
		
			// "fillerSize" should be less than 48 bytes, 
			// such that adding "fillerSize" to "headerSize"
			// is an even multiple of "blockSize" ..
			// i.e.:         headerSize == 24, 40, 56, 72, 88, 104, 120, 136, 152, 168, 184, 200 ..
			//       gaugeTheFillerSize == 48, 96, 144, 192, 240 ..
			//               fillerSize == (48 - 24), (48 - 40), (96 - 56), (96 - 72), (96 - 88), (144 - 104) ..
			fillerSize = ( gaugeTheFillerSize - headerSize );
		
	
			if ( fillerSize > 0 ) {
			
				string filler = createFillerString( fillerSize );
			
				int charSize = 1;
			
				// write the filler string to the DataHeader file,
				// so that the header file size is an even multiple of "blockSize" ..
				for(int i=0; i<filler.size(); i++){
				
					//cout << filler[i];
				
					tbl.write( reinterpret_cast< const char * >( &( filler[i] ) ), charSize );
			
				}
			
			}
			
			cout << "testStorageManager_writeIndexHeader() Point K .." << endl << endl;
		
			// flag indicating the "writeHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			writeHeaderPerformed = true;


		} else {
		
	 		string errorString = "\nHeader file failed to open [ SM::writeIndexHeader() ], so Header file was not written.";
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringQM( errorString );
			
		} // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();
		
		
	} else {
		
	 	string errorString = "\nHeader file table name was not valid [ SM::writeIndexHeader() ], so Header file was not written.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		//getMessageManager()->setErrorStringQM( errorString );
		
	} // end "if ( emptyString.compare( tableName ) != 0 )" ..

	
	return writeHeaderPerformed;
	

} // end "writeIndexHeader()" ..


bool Tester:: readIndexPage( char* tableName_, char* indexName_, IndexPage* indexPage_, int pageID_ ) {
	
	
	if( indexPage_ == NULL ) {
	
		tableName_ = "abc";
	
		indexName_ = "I";
	
		indexPage_ = new IndexPage();
	
		pageID_ = 0;
		
	}
	
	int blockSize = 48;
	
	int intSize = 4;
	
	// "tableName": name of the table associated with the index ..
	
	// "indexPage": indexPage to read from file ..
	
	//    "pageID": pageID associated with this indexPage ..
	
	cout << "Tester:: readIndexPage() start .." << endl << endl;
	cout << "Tester:: readIndexPage() tableName: " << tableName_ << endl << endl;
	cout << "Tester:: readIndexPage() pageID: " << pageID_ << endl << endl;
	
	
	bool readPagePerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName_;
	
	
	if ( emptyString.compare( tableNameString ) != 0 ) {
	
	
		IndexHeader header;
		bool readHeaderPerformed = readIndexHeader( tableName_, indexName_, header );
		
		// DEBUGGING ..
		//cout << "StorageManager:: readDataPage() readHeaderPerformed: " << readHeaderPerformed << endl << endl;
	
		// if reading header failed, can't read page ..
		if( readHeaderPerformed == false ) {
			
	 		string errorString = "\nFailed to read Header file [ SM::readIndexPage() ], so data page was not read.";
	 	
	 		cout << errorString << endl << endl;
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringSM( errorString );
			
			return false;
			
		}
	
	
		string directoryPlusTableName = "C:\\Temp\\";
		directoryPlusTableName += "Index_";
		directoryPlusTableName += tableName_;
		directoryPlusTableName += "_";
		directoryPlusTableName += indexName_;
		char* fileName = (char*)directoryPlusTableName.c_str();
	
		fstream tbl( fileName,  ios::in | ios::binary );
	
		cout << "testStorageManager_readIndexPage() tbl.fail(): " << tbl.fail() << endl << endl;
	
		if ( !tbl.fail() ) {
	

			// DataHeader is written into its own file (by itself) 
			// so locate "page-to-read", referenced from beginning of file ..
			tbl.seekg( ( pageID_ * blockSize ) );
	
	
			int numberOfChildPageIDValues = 0;
			int numberOfSearchKeyValues = 0;
			
			int fieldValue = -1;
			
			cout << "header.getTotalNumberOfIndexPages(): " << header.getTotalNumberOfIndexPages() << endl << endl;
			
			for(int i=0; i<header.getTotalNumberOfIndexPages(); i++){
		
				// DEBUGGING ..
				//cout << "StorageManager:: readDataPage() pageID("<<pageID<<") == header._headerEntries["<<i<<"].pageID("<<header._headerEntries[i].pageID<<"): " << ( pageID == header._headerEntries[i].pageID ) << endl << endl;
		
				if ( pageID_ == header.getHeaderEntries()->at( i )->getPageID() ){
		
					numberOfChildPageIDValues = header.getHeaderEntries()->at( i )->getNumberOfChildPageIDs();
		
					numberOfSearchKeyValues = header.getHeaderEntries()->at( i )->getNumberOfSearchKeys();

					break;
			
				}
		
			}
			
			cout << "Tester:: readIndexPage() numberOfChildPageIDValues: " << numberOfChildPageIDValues << endl << endl;
			
			cout << "Tester:: readIndexPage() numberOfSearchKeyValues: " << numberOfSearchKeyValues << endl << endl;


			//cout << "Tester:: readIndexPage() indexPage_->getChildPageIDs()->size(): " << indexPage_->getChildPageIDs()->size() << endl << endl;

			// write the childPageID / searchKey pairs in the IndexPage to file ..
			for (int i=0; i < numberOfChildPageIDValues; i++) {
					
				// "childPageID" is only written when "i" is "even" ..
					
				tbl.read( reinterpret_cast< char * > ( &( fieldValue ) ), intSize );
					
				indexPage_->getChildPageIDs()->at( i ) = fieldValue;
				
				// DEBUGGING ..
				cout << "Tester:: readIndexPage() childPageIDs->at(i("<<i<<")): " << fieldValue << endl << endl;
			

			}
			
			cout << "Tester:: readIndexPage() Point A .. " << endl << endl;
			
			// write the childPageID / searchKey pairs in the IndexPage to file ..
			for (int i=0; i < numberOfSearchKeyValues; i++) {

				// "pageIDs" (#1 ptr) - (#5 ptr) ..
				
				tbl.read( reinterpret_cast< char * > ( &( fieldValue ) ), intSize );
					
				indexPage_->getSearchKeys()->at( i ) = fieldValue;
				
				// DEBUGGING ..
				cout << "Tester:: readIndexPage() searchKeys->at(i("<<i<<")): " << fieldValue << endl << endl;
			
			}
		
			cout << "Tester:: readIndexPage() Point A .. " << endl << endl;
			
			readPagePerformed = true;
		
		
		} else {
			
	 		string errorString = "\nFailed to open data file [ SM::readIndexPage() ], so data page was not read.\n";
	 	
	 		cout << errorString;
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringQM( errorString );
			
		}  // end "if ( !tbl.fail() )" ..

		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();
	
		// DEBUGGING ..
		//cout << "StorageManager:: readDataPage() maxSlotsAvailable: " << maxSlotsAvailable << endl << endl;
		
		
	} // end "if ( emptyString.compare( tableName ) != 0 )" ..
			
			
	return readPagePerformed;
	
	
} // end "readIndexPage()" ..


bool Tester:: readIndexPages( vector<IndexPage*>* indexPages_ ){
	
	cout << "Tester:: readIndexPages() start .." << endl << endl;
	
	char* tableName_ = "abc";
	
	char* indexName_ = "I";
	
	IndexPage* indexPage = NULL;
	
	bool wasIndexPageRead = false;
	bool wereIndexPagesRead = true;
	
	cout << "Tester:: readIndexPages() Point A .." << endl << endl;
	
	IndexHeader header;
	bool readHeaderPerformed = readIndexHeader( tableName_, indexName_, header );
		
	int numberOfIndexPages = header.getTotalNumberOfIndexPages();
	
	cout << "Tester:: readIndexPages() Point B .." << endl << endl;
	
	int pageID = -1;
	
	for(int i=0; i<numberOfIndexPages; i++){
						
		cout << "Tester:: readIndexPages() Point C["<<i<<"] .." << endl << endl;
						
		indexPage = new IndexPage();
		
		pageID = header.getHeaderEntries()->at(i)->getPageID();
						
		wasIndexPageRead = readIndexPage( tableName_, indexName_, indexPage, pageID );
		
		cout << "Tester:: readIndexPages() Point D["<<i<<"] .." << endl << endl;
		
		indexPages_->push_back( indexPage );
		
		if( wasIndexPageRead == false )
			wereIndexPagesRead = false;
						
	}
	
	cout << "Tester:: readIndexPages() wereIndexPagesRead: " << wereIndexPagesRead << endl << endl;
	
	return wereIndexPagesRead;
	
} // end "writeIndexPages()" ..


bool Tester:: readIndexHeader( char* tableName, char* indexName_, IndexHeader& header ) {
	

	bool readHeaderPerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName;
	
	int intSize = 4;
	
	int boolSize = 1;
	
	// DEBUGGING ..
	//cout << "StorageManager:: readHeader() tableName: " << tableName << endl;
	
	if ( emptyString.compare( tableNameString ) != 0 ) {
	
		string directoryPlusFileName = "C:\\Temp\\";
		
		directoryPlusFileName += "IndexHeader_";
		directoryPlusFileName += tableName;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();
		

		fstream tbl( fileName, ios::in | ios::binary );
		
		cout << "testStorageManager_readIndexHeader tbl.fail(): " << tbl.fail() << endl << endl;

		if ( !tbl.fail() ) {
	
			tbl.seekp( ios::beg );
			
			int totalNumberOfIndexPages = -1;
			
			tbl.read( reinterpret_cast< char * > ( &( totalNumberOfIndexPages ) ), intSize );
			
			cout << "readIndexHeader totalNumberOfIndexPages: " << totalNumberOfIndexPages << endl << endl;
			
			header.setTotalNumberOfIndexPages( totalNumberOfIndexPages );
	
	
			// DEBUGGING ..
			//cout << "StorageManager:: readHeader() header.totalNumberOfDataPages: " << header.totalNumberOfDataPages << endl;
			//cout << "StorageManager:: readHeader() header.filePageFieldCount: " << header.filePageFieldCount << endl;
	
	
			int pid = 0;
	
			int ppid = 0;
			
			int nosk = 0;
			
			int nocpi = 0;
	
			int fskc = 0;
	
			bool ska = false;
			
			int numberOfSearchKeys = 4;
	
	
			header.getHeaderEntries()->clear();
			
			for (int i=0; i<header.getTotalNumberOfIndexPages(); i++){
				
				header.getHeaderEntries()->push_back( ( new IndexHeaderEntry() ) );
		
				tbl.read( reinterpret_cast< char * > ( &( pid ) ), intSize );
				header.getHeaderEntries()->at( i )->setPageID( pid );
				cout << "readIndexHeader pageID: " << pid << endl << endl;
				
				tbl.read( reinterpret_cast< char * > ( &( nosk ) ), intSize );
				header.getHeaderEntries()->at( i )->setNumberOfSearchKeys( nosk );
				cout << "readIndexHeader numberOfSearchKeys: " << nosk << endl << endl;
				
				tbl.read( reinterpret_cast< char * > ( &( nocpi ) ), intSize );
				header.getHeaderEntries()->at( i )->setNumberOfChildPageIDs( nocpi );
				cout << "readIndexHeader numberOfChildPageIDs: " << nocpi << endl << endl;
		
				tbl.read( reinterpret_cast< char * > ( &( ppid ) ), intSize );
				header.getHeaderEntries()->at( i )->setParentPageID( ppid );
				cout << "readIndexHeader parentPageID: " << ppid << endl << endl;
		
				tbl.read( reinterpret_cast< char * > ( &( fskc ) ), intSize );
				header.getHeaderEntries()->at( i )->setFreeSearchKeyCount( fskc );
				cout << "readIndexHeader freeSearchKeyCount: " << fskc << endl << endl;
		
				cout << "readIndexHeader numberOfSearchKeys: " << numberOfSearchKeys << endl << endl;
		
				/*tbl.read( reinterpret_cast< char * > ( &( ska ) ), boolSize );
				cout << "readIndexHeader searchKeyAvailable: " << ska << endl << endl;
				
				tbl.read( reinterpret_cast< char * > ( &( ska ) ), boolSize );
				cout << "readIndexHeader searchKeyAvailable: " << ska << endl << endl;
				
				tbl.read( reinterpret_cast< char * > ( &( ska ) ), boolSize );
				cout << "readIndexHeader searchKeyAvailable: " << ska << endl << endl;
				
				tbl.read( reinterpret_cast< char * > ( &( ska ) ), boolSize );
				cout << "readIndexHeader searchKeyAvailable: " << ska << endl << endl;
				*/
		
				for(int j=0; j<numberOfSearchKeys; j++) {
					
					tbl.read( reinterpret_cast< char * > ( &( ska ) ), boolSize );
					cout << "readIndexHeader searchKeyAvailable: " << ska << endl << endl;
					header.getHeaderEntries()->at( i )->getSearchKeysAvailable()->at( j ) = ska;
					
				}
		
			}
			
			// flag indicating the "readHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			readHeaderPerformed = true;
			
		} else {
			
	 		string errorString = "\nHeader file failed to open [ SM::readIndexHeader() ], so Header file was not read (ignore if file is in process of being created).";
	 	
	 		cout << errorString << endl << endl;
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringQM( errorString );	
			
		} // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();
		
		
	} else {
		
	 	string errorString = "\nHeader file table name was not valid [ SM::readIndexHeader() ], so Header file was not read.";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		//getMessageManager()->setErrorStringQM( errorString );	

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..
	
	
	if ( !readHeaderPerformed ) {
		
		// "IndexHeader" was not found, b/c the file was not opened and/or found, so
		// set the values of the "hp" to default values, and
		// indicate that the "IndexHeader" was _not_ found ..
			
		header.setTotalNumberOfIndexPages( 0 );
			
		header.getHeaderEntries()->clear();	
		
	}
	
	
	// DEBUGGING ..
	//cout << "SM:: readIndexHeader() headerPageFound: " << headerPageFound << endl;
	//cout << "SM:: readIndexHeader() end .." << endl;
	
	
	return readHeaderPerformed;
	
	
} // end "readIndexHeader( char* tableName, DataHeader& header )" .. 


bool Tester:: readIndexBucketHeader( char* tableName_, char* indexName_, IndexBucketHeader& indexBucketHeader_ ) {
	

	bool readHeaderPerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName_;
	
	// DEBUGGING ..
	//cout << "StorageManager:: readHeader() tableName: " << tableName << endl;
	
	if ( emptyString.compare( tableNameString ) != 0 ) {
	
		int intSize = 4;
		
		int boolSize = 1;
	
		string directoryPlusFileName = "C:\\Temp\\";
		
		directoryPlusFileName += "IndexBucketHeader_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();
		

		fstream tbl( fileName, ios::in | ios::binary );
		

		if ( !tbl.fail() ) {
	
			tbl.seekp( ios::beg );
			
			int totalNumberOfBuckets = -1;
			
			tbl.read( reinterpret_cast< char * > ( &( totalNumberOfBuckets ) ), intSize );
			
			indexBucketHeader_.setTotalNumberBuckets( totalNumberOfBuckets );
	
	
			// DEBUGGING ..
			//cout << "StorageManager:: readHeader() header.totalNumberOfDataPages: " << header.totalNumberOfDataPages << endl;
			//cout << "StorageManager:: readHeader() header.filePageFieldCount: " << header.filePageFieldCount << endl;
	
	
			// "map<int, vector<int>* >* _indexBucketIDsForEachSearchKey"
			map<int, vector<int>* >* iBIDsFESK = new map<int, vector<int>* >();
			int sk = 0;						// search key
			int numberBucketIDs;			// # bucketIDs in vector<int>*
			vector<int>* bucketIDs = NULL;
				
			int nextToken = 1;
			
			int fieldNumber = 0;
				
			while( nextToken > 0 ){
				
				if( fieldNumber == 0 ){
					
					tbl.read( reinterpret_cast< char * > ( &( nextToken ) ), intSize );
					
					sk = nextToken;
					
					fieldNumber = 1;
					
				} else if( fieldNumber == 1 ){
					
					tbl.read( reinterpret_cast< char * > ( &( nextToken ) ), intSize );
					
					numberBucketIDs = nextToken;
					
					fieldNumber == 2;
					
				} else {
					
					// collect all of the "bucketIDs" for this searck key ..
					for(int i=0; i<numberBucketIDs; i++){
						
						tbl.read( reinterpret_cast< char * > ( &( nextToken ) ), intSize );
						
						bucketIDs->push_back( nextToken );
						
					}
					
					if( bucketIDs->size() > 0 )
						iBIDsFESK->insert( pair<int, vector<int>* >( sk, bucketIDs ) );
						
					indexBucketHeader_.setBucketIDsForEachSearchKey( iBIDsFESK );
					
					fieldNumber = 0;
					
				}
				
			}
			
			int bi = 0;
			
			int fpc = 0;
			
			bool pa = false;
			
			int bucketSize = 12;
	
			indexBucketHeader_.getHeaderEntries()->clear();
			
			for (int i=0; i<indexBucketHeader_.getTotalNumberBuckets(); i++){
				
				indexBucketHeader_.getHeaderEntries()->push_back( ( new IndexBucketHeaderEntry() ) );
		
				tbl.read( reinterpret_cast< char * > ( &( bi ) ), intSize );
				indexBucketHeader_.getHeaderEntries()->at( i )->setBucketID( bi );
		
				tbl.read( reinterpret_cast< char * > ( &( fpc ) ), intSize );
				indexBucketHeader_.getHeaderEntries()->at( i )->setFreePointerCount( fpc );
		
				//for(int j=0; j<bucketSize; j++) {
		
				//	tbl.read( reinterpret_cast< char * > ( &( pa ) ), boolSize );
				//	indexBucketHeader_.getHeaderEntries()->at( i )->getPointersAvailable()->at( j ) = pa;

				//}	
		
			}
			
			// flag indicating the "readHeader()" operation
			// was performed successfully (tableName ok, file opened) ..
			readHeaderPerformed = true;
			
		} else {
			
	 		string errorString = "\nHeader file failed to open [ SM::readIndexBucketHeader() ], so Header file was not read.";
	 	
	 		cout << errorString << endl;
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringQM( errorString );	
			
		} // end "if ( !tbl.fail() )" ..


		tbl.clear();
		tbl.close();
		
		
	} else {
		
	 	string errorString = "\nHeader file table name was not valid [ SM::readIndexBucketHeader() ], so Header file was not read.";
	 	
	 	cout << errorString << endl;
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		//getMessageManager()->setErrorStringQM( errorString );	

	} // end "if ( emptyString.compare( tableName ) != 0 )" ..
	
	
	if ( !readHeaderPerformed ) {
		
		// "IndexBucketHeader" was not found, b/c the file was not opened and/or found, so
		// set the values of the "hp" to default values, and
		// indicate that the "IndexBucketHeader" was _not_ found ..
			
		indexBucketHeader_.setTotalNumberBuckets( 0 );
			
		indexBucketHeader_.getHeaderEntries()->clear();	
		
	}
	
	
	// DEBUGGING ..
	//cout << "SM:: readIndexHeader() headerPageFound: " << headerPageFound << endl;
	//cout << "SM:: readIndexHeader() end .." << endl;
	
	
	return readHeaderPerformed;
	
	
} // end "readIndexBucketHeader()" .. 


bool Tester:: readIndexBucket( char* tableName_, char* indexName_, IndexBucket* indexBucket_, int bucketPageID_ ) {
	
	// "tableName": name of the table associated with the index ..
	
	//    "indexBucket": indexBucket to write to file ..
	
	cout << "StorageManager:: readIndexBucket() start .." << endl << endl;
	cout << "StorageManager:: readIndexBucket() tableName: " << tableName_ << endl << endl;
	//cout << "StorageManager:: readIndexBucket() indexBucket->getPageIDs()->size(): " << indexBucket->getPageIDs()->size() << endl << endl;
	cout << "StorageManager:: readIndexBucket() bucketPageID: " << bucketPageID_ << endl << endl;
	
	bool readBucketPagePerformed = false;
	
	string emptyString = "";
	
	string tableNameString = tableName_;

	if ( emptyString.compare( tableNameString ) != 0 ) {
		
		
		IndexBucketHeader header;
		bool readHeaderPerformed = readIndexBucketHeader( tableName_, indexName_, header );
		
		// DEBUGGING ..
		//cout << "StorageManager:: readDataPage() readHeaderPerformed: " << readHeaderPerformed << endl << endl;
	
		// if reading header failed, can't read page ..
		if( readHeaderPerformed == false ) {
			
	 		string errorString = "\nFailed to read Header file [ SM::readIndexBucket() ], so data page was not read.";
	 	
	 		cout << errorString << endl;
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringSM( errorString );
			
			return false;
			
		}
		
		
		// handle the directory path of the file ..
	
		string directoryPlusFileName = "C:\\Temp\\";
		
		cout << "StorageManager:: readIndexBucket() Point A .." << endl << endl;
	
		int blockSize = 48;
		
		int intSize = 4;
		
		int maxNumberOfBucketIDs = (blockSize / intSize);
	
		// append the file
		directoryPlusFileName += "IndexBuckets_";
		directoryPlusFileName += tableName_;
		directoryPlusFileName += "_";
		directoryPlusFileName += indexName_;
		char* fileName = (char*)directoryPlusFileName.c_str();
		
		cout << "StorageManager:: readIndexBucket() Point D .." << endl << endl;
		
		fstream tbl( fileName, ios::in | ios::binary );
		
		cout << "StorageManager:: readIndexBucket() Point E .." << endl << endl;
		
		if ( !tbl.fail() ) {
			
			cout << "StorageManager:: readIndexBucket() Point F .." << endl << endl;
	
			// make sure file pointer starts at beginning of file ..
			tbl.seekg( ios::beg );
			
			cout << "StorageManager:: readIndexBucket() Point G .." << endl << endl;

			// so locate page-to-write referenced from beginning of file ..
			tbl.seekg( bucketPageID_ * blockSize );
			
			cout << "StorageManager:: readIndexBucket() Point H .." << endl << endl;

			// number of values in the "bucket" ..
			int totalNumberOfBuckets = header.getTotalNumberBuckets();
			cout << "StorageManager:: readIndexBucket() totalNumberOfBuckets: " << totalNumberOfBuckets << endl << endl;			


			int freePointerCount = header.getHeaderEntries()->at( bucketPageID_ )->getFreePointerCount();
			int numberOfBucketIDs = (maxNumberOfBucketIDs - freePointerCount);
			
			
			cout << "StorageManager:: readIndexBucket() Point I .." << endl << endl;
			
			//cout << "StorageManager:: readIndexBucket() pointersAvailable->size(): " << pointersAvailable->size() << endl << endl;

		
			int fieldValue = -1;
	
			// write the pageIDs in the IndexBucket to file ..
			for (int i=0; i < numberOfBucketIDs; i++) {
			
				// DEBUGGING ..
				//cout << "StorageManager:: writeIndexBucket() pageIDs["<<i<<"]: " << pageIDs->at(i) << endl << endl;

				tbl.read( reinterpret_cast< char * > ( &( fieldValue ) ), intSize );
				indexBucket_->getPageIDs()->push_back( fieldValue );
				cout << "StorageManager:: readIndexBucket() fieldValue: " << fieldValue << endl << endl;

			}
	
			// flag indicating the "writeIndexBucket()" operation
			// was performed successfully (header read, file created, file opened) ..
			readBucketPagePerformed = true;
		
		
		} else {
		
	 		string errorString = "\nFailed to open data file [ SM::readIndexBucket() ], so index bucket was not written to.";
	 	
	 		cout << errorString << endl;
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			//getMessageManager()->setErrorStringSM( errorString );	
		
		} // end "if ( !tbl.fail() )" ..
	
	
		// DEBUGGING ..
		//checkFileState( tbl );

		tbl.clear();
		tbl.close();
		
		
	} // end "if ( emptyString.compare( tableNameString ) != 0 )" ..

	
	//cout << "SM:: readIndexBucket() end .." << endl;
	
	return readBucketPagePerformed;
	

} // end "readIndexBucket()" ..


void Tester:: test_write(){
	
///*
	//fstream dataTable1( "C:\\Temp\\test", ios::in | ios::out | ios::binary );
	fstream dataTable1;
	dataTable1.open( "C:\\Temp\\test", ios::out | ios::binary | ios::app );
	
	//cout << "Tester:: test_write() dataTable.fail(): " << dataTable1.fail() << endl << endl;

	dataTable1.seekp( 0 );
	
	char pad = '%';
	
	for(int i=0; i<48; i++){

		dataTable1.write( reinterpret_cast<const char*>(&pad), sizeof(char) );

	}

	dataTable1.clear();
	dataTable1.close();
//*/

	//fstream dataTable2( "C:\\Temp\\test", ios::in | ios::out | ios::binary );
	fstream dataTable2;
	dataTable2.open( "C:\\Temp\\test", ios::in | ios::out | ios::binary );
	
	//cout << "Tester:: test_write() dataTable.eof(): " << dataTable2.eof() << ", dataTable.fail(): " << dataTable2.fail() << ", dataTable.bad(): " << dataTable2.bad() << endl;
	
	dataTable2.seekp( 0 );
	
	//cout << "Tester:: test_write() dataTable.eof(): " << dataTable2.eof() << ", dataTable.fail(): " << dataTable2.fail() << ", dataTable.bad(): " << dataTable2.bad() << endl;
	
	int fieldValue = 10;
	
	dataTable2.write( reinterpret_cast<const char*>(&fieldValue), sizeof(int) );
	
	//cout << "Tester:: test_write() dataTable.eof(): " << dataTable2.eof() << ", dataTable.fail(): " << dataTable2.fail() << ", dataTable.bad(): " << dataTable2.bad() << endl;
	
	fieldValue = 20;
	
	dataTable2.write( reinterpret_cast<const char*>(&fieldValue), sizeof(int) );
	
	//cout << "Tester:: test_write() dataTable.eof(): " << dataTable2.eof() << ", dataTable.fail(): " << dataTable2.fail() << ", dataTable.bad(): " << dataTable2.bad() << endl;
	
	fieldValue = 30;
	
	dataTable2.write( reinterpret_cast<const char*>(&fieldValue), sizeof(int) );
	
	//cout << "Tester:: test_write() dataTable.eof(): " << dataTable2.eof() << ", dataTable.fail(): " << dataTable2.fail() << ", dataTable.bad(): " << dataTable2.bad() << endl;
	
	fieldValue = 40;
	
	dataTable2.write( reinterpret_cast<const char*>(&fieldValue), sizeof(int) );
	
	//cout << "Tester:: test_write() dataTable.eof(): " << dataTable2.eof() << ", dataTable.fail(): " << dataTable2.fail() << ", dataTable.bad(): " << dataTable2.bad() << endl;
	
	//for(int i=0; i<32; i++){

	//	dataTable2.write( (const char *) ( &( pad ) ), 2 );

	//}
	
	dataTable2.clear();
	dataTable2.close();

	
}

void Tester:: test_read(){
	
	//fstream dataTable( "C:\\Temp\\test",  ios::in | ios::binary );
	fstream dataTable;
	dataTable.open( "C:\\Temp\\test", ios::in | ios::binary );
	
	//cout << "Tester:: test_read() dataTable.eof(): " << dataTable.eof() << ", dataTable.fail(): " << dataTable.fail() << ", dataTable.bad(): " << dataTable.bad() << endl;
	
	dataTable.seekg( 0 );
	
	int intValue = -1;
	
	for(int i=0; i<4; i++){
		
		dataTable.read( reinterpret_cast<char *>(&intValue), sizeof(int) );
		
		cout << "intValue: " << intValue << endl;
		
		//cout << "dataTable.eof(): " << dataTable.eof() << ", dataTable.fail(): " << dataTable.fail() << ", dataTable.bad(): " << dataTable.bad() << endl;
		
	}
	
	char charValue = '!';
	
	for(int i=0; i<32; i++){
		
		dataTable.read( reinterpret_cast<char *>(&charValue), sizeof(char) );
		
		cout << i << ") charValue: " << charValue << endl;
		
		//cout << "dataTable.eof(): " << dataTable.eof() << ", dataTable.fail(): " << dataTable.fail() << ", dataTable.bad(): " << dataTable.bad() << endl;
		
	}

	dataTable.clear();
	dataTable.close();
	
}

int Tester:: getNumberOfUniqueRows(){
	
	int numberOfUniqueRows = -1;
	
	// NOTE: if "commit" used, left-hand numbers are flushed to file, else
	//		 right-hand numbers represents the rows written to file ..
		
	// 5, 10 ..
	//numberOfUniqueRows = 2;
		
	// 5, 10, 15, 20 .. (5 flushed to file) ..
	//numberOfUniqueRows = 4;
		
	// 5, 10, 15, 20 .. 25 .. (5 & 10 flushed to file) ..
	//numberOfUniqueRows = 5;
		
	// 5, 10, 15, 20 .. 30 .. (5 - 15 flushed to file) ..
	//numberOfUniqueRows = 6;
		
	// 5, 10, 15, 20 .. 35 .. (5 - 20 flushed to file) ..
	//numberOfUniqueRows = 7;
		
	// 5, 10, 15, 20 .. 40 .. (5 - 25 flushed to file) ..
	//numberOfUniqueRows = 8;
		
	// 5, 10, 15, 20 .. 45 .. (5 - 30 flushed to file) ..
	//numberOfUniqueRows = 9;
		
	// 5, 10, 15, 20 .. 60 .. (5 - 45 flushed to file) ..
	//numberOfUniqueRows = 12;
		
	// 5, 10, 15, 20 .. 80 .. (5 - 65 flushed to file) ..
	//numberOfUniqueRows = 16;
		
	// 5, 10, 15, 20 .. 100 .. (5 - 85 flushed to file) ..
	//numberOfUniqueRows = 20;
		
	// 5, 10, 15, 20 .. 105 .. (5 - 90 flushed to file) ..
	numberOfUniqueRows = 21;
		
	// 5, 10, 15, 20 .. 115 .. (5 - 100 flushed to file) ..
	//numberOfUniqueRows = 23;
		
	// 5, 10, 15, 20 .. 120 .. (5 - 105 flushed to file) ..
	// (parent root-node splits) ..
	//numberOfUniqueRows = 24;
		
	// 5, 10, 15, 20 .. 125 .. (5 - 110 flushed to file) ..
	//numberOfUniqueRows = 25;
		
	// 5, 10, 15, 20 .. 140 .. (5 - 125 flushed to file) ..
	//numberOfUniqueRows = 28;
		
	// 5, 10, 15, 20 .. 160 .. (5 - 145 flushed to file) ..
	//numberOfUniqueRows = 32;
		
	// 5, 10, 15, 20 .. 180 .. (5 - 165 flushed to file) ..
	// (parent node splits) ..
	//numberOfUniqueRows = 36;
		
	// 5, 10, 15, 20 .. 200 .. (5 - 185 flushed to file) ..
	//numberOfUniqueRows = 40;
		
	// 5, 10, 15, 20 .. 215 .. (5 - 200 flushed to file) ..
	//numberOfUniqueRows = 43;
		
	// 5, 10, 15, 20 .. 220 .. (5 - 205 flushed to file) ..
	//numberOfUniqueRows = 44;
		
	// 5, 10, 15, 20 .. 240 .. (5 - 225 flushed to file) ..
	// (parent node splits) ..
	//numberOfUniqueRows = 48;
		
	// 5, 10, 15, 20 .. 260 .. (5 - 245 flushed to file) ..
	//numberOfUniqueRows = 52;
		
	// 5, 10, 15, 20 .. 275 .. (5 - 260 flushed to file) ..
	//numberOfUniqueRows = 55;
		
	// 5, 10, 15, 20 .. 280 .. (5 - 265 flushed to file) ..
	//numberOfUniqueRows = 56;
		
	// 5, 10, 15, 20 .. 300 .. (5 - 285 flushed to file) ..
	// (parent node splits) ..
	//numberOfUniqueRows = 60;
		
	// 5, 10, 15, 20 .. 320 .. (5 - 305 flushed to file) ..
	//numberOfUniqueRows = 64;
		
	// 5, 10, 15, 20 .. 340 .. (5 - 325 flushed to file) ..
	//numberOfUniqueRows = 68;
		
	// 5, 10, 15, 20 .. 345 .. (5 - 330 flushed to file) ..
	//numberOfUniqueRows = 69;
		
	// 5, 10, 15, 20 .. 360 .. (5 - 345 flushed to file) ..
	// (parent node splits) ..
	// (parent root-node splits) ..
	//numberOfUniqueRows = 72;
		
	// 5, 10, 15, 20 .. 1260 .. (5 - 1245 flushed to file) ..
	// (parent node splits) ..
	// (parent root-node splits) ..
	// (console fails, has errors) ..
	//numberOfUniqueRows = 252;
	
	return numberOfUniqueRows;
	
}

void Tester:: test_insert( UserInterface* ui, string tableName, int numberOfUniqueRows, int numberRepeatRows, int multiple, int numberOfAttributes ){
	
	vector<string>* insertValues = new vector<string>();
	
	for(int i=1; i<=numberOfUniqueRows; i++){
			
		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			for(int k=0; k<numberOfAttributes; k++){
				
				// "a" column has the index ..
				if( k == 0 )
					insertValues->push_back( intToStr( ( multiple * i ) ) );
				else
					insertValues->push_back( intToStr( i ) );
				
			}

			// remaining columns are unimportant, but need dummy values ..
			//insertValues->push_back( intToStr( ( 20 * multiple * i ) ) );
			
			//insertValues->push_back( intToStr( ( 30 * multiple * i ) ) );
			
			//insertValues->push_back( intToStr( ( 40 * multiple * i ) ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues->clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
	}
	
}

void Tester:: test_createTables_insert_select(){
	
	// 8 columns per table max ..
	int maxColumnsPerTable = 8;
		
	// 3 buffer pages ..
	int bufferPoolSize = 3;

	UserInterface* ui = new UserInterface( maxColumnsPerTable, bufferPoolSize );
	
	// change the path to "/home/stalbot/CS525/Temp" ..
	vector<string>* inputTokens = new vector<string>();
	
	inputTokens->push_back( "set" );
	inputTokens->push_back( "path" );
	inputTokens->push_back( "home" );
	inputTokens->push_back( "stalbot" );
	inputTokens->push_back( "CS525" );
	inputTokens->push_back( "Temp" );
	
	ui->changePath( inputTokens, false );	// inputTokens, isWindows
		
	/*
	 * TABLE I ..
	 */
		
	string tableName = "xyz";
	vector<string>* attributeNames1 = new vector<string>();
	attributeNames1->push_back("a");
	attributeNames1->push_back("b");
	attributeNames1->push_back("c");
	attributeNames1->push_back("d");
	int numberOfAttributes = 4;
		
	ui->getQueryManager()->createTable( tableName, attributeNames1, numberOfAttributes );
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
	int numberOfUniqueRows = getNumberOfUniqueRows();
		
	// each unique row is repeated this many times ..
	int numberRepeatRows = 3;
		
	// each unique row is separated by a multiple ..
	int multiple = 2;
	
	// insert into the table ..
	test_insert( ui, tableName, numberOfUniqueRows, numberRepeatRows, multiple, numberOfAttributes );
		
	cout << "tester Point A .." << endl << endl;
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	cout << "tester Point B .." << endl << endl;
	
	/*
	 * TABLE II ..
	 */
		
	tableName = "abc";
	vector<string>* attributeNames2 = new vector<string>();
	attributeNames2->push_back("aaa");
	attributeNames2->push_back("bbb");
	numberOfAttributes = 2;
		
	ui->getQueryManager()->createTable( tableName, attributeNames2, numberOfAttributes );
		
	cout << "tester Point C .." << endl << endl;
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
	numberOfUniqueRows = getNumberOfUniqueRows();
		
	// each unique row is repeated this many times ..
	numberRepeatRows = 6;
		
	// each unique row is separated by a multiple ..
	multiple = 1;
	
	// insert into the table ..
	test_insert( ui, tableName, numberOfUniqueRows, numberRepeatRows, multiple, numberOfAttributes );
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	/*
	 * TABLE III ..
	 */
		
	tableName = "123";
	vector<string>* attributeNames3 = new vector<string>();
	attributeNames3->push_back("q");
	attributeNames3->push_back("w");
	attributeNames3->push_back("e");
	attributeNames3->push_back("r");
	attributeNames3->push_back("t");
	attributeNames3->push_back("y");
	numberOfAttributes = 6;
		
	ui->getQueryManager()->createTable( tableName, attributeNames3, numberOfAttributes );
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
	numberOfUniqueRows = getNumberOfUniqueRows();
		
	// each unique row is repeated this many times ..
	numberRepeatRows = 2;
		
	// each unique row is separated by a multiple ..
	multiple = 3;
	
	// insert into the table ..
	test_insert( ui, tableName, numberOfUniqueRows, numberRepeatRows, multiple, numberOfAttributes );
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	cout << "tester Point D .." << endl << endl;

	/*
	 * SELECT ..
	 */
	
	// doing "select * from X", so shouldn't need last 2 data structures ..
	
	vector<string>* selectColumnNames = NULL;
	multimap<string, string>* whereColumnNamePairs = new multimap<string, string>();
	map<string, int>* whereColumnNameValuePairs = new map<string, int>();
	
	//whereColumnNamePairs->insert( pair<string, string>(  ) );
	
	vector< map<int, DataPage* >* >* returnPages = NULL;
	
	map<int, DataPage* >* bpDataPages = NULL;
	
	map<int, DataPage* >* fileDataPages = NULL;
	
	// CHANGE this to augment the # of loops ..
	int numberOfSelectRuns = 1;
	
	for(int i=0; i<numberOfSelectRuns; i++){
		
		/*
		 * TABLE I ..
		 */
		
		tableName = "xyz";
		selectColumnNames = attributeNames1;
		
		cout << "tester Point E .." << endl << endl;
	
		returnPages = ui->getQueryManager()->selectFromX( tableName, selectColumnNames, whereColumnNamePairs, whereColumnNameValuePairs );
	
		bpDataPages = returnPages->at(0);
	
		fileDataPages = returnPages->at(1);
		
		ui->setTableName( tableName );
		ui->printResults( bpDataPages, fileDataPages, selectColumnNames );
		
		cout << "tester Point F .." << endl << endl;
		
		/*
		 * TABLE II ..
		 */
		
		tableName = "abc";
		selectColumnNames = attributeNames2;
	
		returnPages = ui->getQueryManager()->selectFromX( tableName, selectColumnNames, whereColumnNamePairs, whereColumnNameValuePairs );
	
		bpDataPages = returnPages->at(0);
	
		fileDataPages = returnPages->at(1);
	
		ui->setTableName( tableName );
		ui->printResults( bpDataPages, fileDataPages, selectColumnNames );		
		
		/*
		 * TABLE III ..
		 */
		
		tableName = "123";
		selectColumnNames = attributeNames3;
	
		returnPages = ui->getQueryManager()->selectFromX( tableName, selectColumnNames, whereColumnNamePairs, whereColumnNameValuePairs );
	
		bpDataPages = returnPages->at(0);
	
		fileDataPages = returnPages->at(1);
	
		ui->setTableName( tableName );
		ui->printResults( bpDataPages, fileDataPages, selectColumnNames );	
		
	}
	

	
	
}

void Tester:: test_createTable_createIndex_insert_printIndex(){
	
	// 8 columns per table max ..
	int maxColumnsPerTable = 8;
		
	// 3 buffer pages ..
	int bufferPoolSize = 3;

	UserInterface* ui = new UserInterface( maxColumnsPerTable, bufferPoolSize );

	// change the path to "/home/stalbot/CS525/Temp" ..
	vector<string>* inputTokens = new vector<string>();
	
	inputTokens->push_back( "set" );
	inputTokens->push_back( "path" );
	inputTokens->push_back( "home" );
	inputTokens->push_back( "stalbot" );
	inputTokens->push_back( "CS525" );
	inputTokens->push_back( "Temp" );
	
	ui->changePath( inputTokens, false );	// inputTokens, isWindows

	string tableName = "xyz";
	vector<string>* attributeNames = new vector<string>();
	attributeNames->push_back("a");
	attributeNames->push_back("b");
	attributeNames->push_back("c");
	int numberOfAttributes = 3;
		
	vector<string>* insertValues = new vector<string>();
		
	ui->getQueryManager()->createTable( tableName, attributeNames, numberOfAttributes );
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
	string indexName = "I";
	bool isNoDuplicates = false;
	string attributeName = "a";
		
	ui->getQueryManager()->createIndex( indexName, tableName, isNoDuplicates, attributeName );
	
	cout << "Tester:: Point A .." << endl << endl;
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
	// number of pages (each row in a page is a repeat, but each page is unique) ..
	int numberOfUniqueRows = getNumberOfUniqueRows();
		
	// each unique row is separated by a multiple ..
	int multiple = 5;
		
	// each unique row is repeated this many times ..
	int numberRepeatRows = 4;
	
	// insert into the table ..
	test_insert( ui, tableName, numberOfUniqueRows, numberRepeatRows, multiple, numberOfAttributes );
		
/*
	for(int i=1; i<=numberOfUniqueRows; i++){
			
		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			// "a" column has the index ..
			insertValues->push_back( intToStr( ( i * multiple ) ) );
			
			// remaining columns are unimportant, but need dummy values ..
			insertValues->push_back( intToStr( ( i ) ) );
			
			insertValues->push_back( intToStr( ( i ) ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues->clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
	}
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
		
*/
		
	// optionally, add values to split leaf nodes ..

	int newValue = 105;

	// repeat each unique row ..
	for(int j=0; j<numberRepeatRows; j++){
			
		// "a" column has the index ..
		insertValues->push_back( intToStr( newValue ) );
		insertValues->push_back( intToStr( 1 ) );
		insertValues->push_back( intToStr( 1 ) );
			
		ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
		// prepare for next row ..
		insertValues->clear();
				
		if( ui->getMessageManager()->hasErrorMessage() == true )
			cout << ui->getMessageManager()->getErrorMessage() << endl;
		else
			cout << ui->getMessageManager()->getResultMessage() << endl;
					
		ui->getMessageManager()->resetErrorStatus();
				
	}

			
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();

		
	// optionally, add values to split internal nodes ..
/*
	int newValue = 293;
		
	// 3(numberOfInserts) + 2(existing in internal node) = 5(enough to initiate split in internal node) ..
	int numberOfInserts = 3;

	for(int i=0; i<numberOfInserts; i++){

		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			// "a" column has the index ..
			insertValues.push_back( intToStr( newValue ) );
			insertValues.push_back( intToStr( 1 ) );
			insertValues.push_back( intToStr( 1 ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues.clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
		// 83, 103, 123 ..
		// 73, 93, 113 ..
		// 23, 43, 63 ..
		// 13, 33, 53 ..
		newValue += 20;
			
	}
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
*/


	// print the index result ..
	ui->getIndexManager()->printIndex( tableName, indexName );
	
	
}


void Tester:: test_createTable_insert_createIndex_printIndex(){
	
	cout << "Point A .." << endl << endl;
	
	// 8 columns per table max ..
	int maxColumnsPerTable = 8;
		
	// 3 buffer pages ..
	int bufferPoolSize = 3;

	UserInterface* ui = new UserInterface( maxColumnsPerTable, bufferPoolSize );
	
	// change the path to "/home/stalbot/CS525/Temp" ..
	vector<string>* inputTokens = new vector<string>();
	
	inputTokens->push_back( "set" );
	inputTokens->push_back( "path" );
	inputTokens->push_back( "home" );
	inputTokens->push_back( "stalbot" );
	inputTokens->push_back( "CS525" );
	inputTokens->push_back( "Temp" );
	
	ui->changePath( inputTokens, false );	// inputTokens, isWindows
	
	cout << "Point B .." << endl << endl;
		
	string tableName = "xyz";
	vector<string>* attributeNames = new vector<string>();
	attributeNames->push_back("a");
	attributeNames->push_back("b");
	attributeNames->push_back("c");
	int numberOfAttributes = 3;
		
	vector<string>* insertValues = new vector<string>();
		
	ui->getQueryManager()->createTable( tableName, attributeNames, numberOfAttributes );
	
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
	string indexName = "I";
	bool isNoDuplicates = false; 
	string attributeName = "a";
		
	// change the order of index creation to be  
	// 1) prior to insert, 2) after records are inserted ..
	// (accomplish by commenting this line out, uncommenting duplicate line below) ..
/*
	ui->getQueryManager()->createIndex( indexName, tableName, isNoDuplicates, attributeName );
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
*/
		
	// number of pages (each row in a page is a repeat, but each page is unique) ..
	int numberOfUniqueRows = getNumberOfUniqueRows();
		
	// each unique row is separated by a multiple ..
	int multiple = 5;
		
	// each unique row is repeated this many times ..
	int numberRepeatRows = 4;
	
	// insert into the table ..
	test_insert( ui, tableName, numberOfUniqueRows, numberRepeatRows, multiple, numberOfAttributes );
	
/*
	cout << "Point C .." << endl << endl;
		
	for(int i=1; i<=numberOfUniqueRows; i++){
			
		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			// "a" column has the index ..
			insertValues->push_back( intToStr( ( i * multiple ) ) );
			
			// remaining columns are unimportant, but need dummy values ..
			insertValues->push_back( intToStr( ( i ) ) );
			
			insertValues->push_back( intToStr( ( i ) ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues->clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
	}
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
		
*/
		
	// optionally, add values to split leaf nodes ..
/*
	int newValue = 105;

	// repeat each unique row ..
	for(int j=0; j<numberRepeatRows; j++){
			
		// "a" column has the index ..
		insertValues.push_back( intToStr( newValue ) );
		insertValues.push_back( intToStr( 1 ) );
		insertValues.push_back( intToStr( 1 ) );
			
		ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
		// prepare for next row ..
		insertValues.clear();
				
		if( ui->getMessageManager()->hasErrorMessage() == true )
			cout << ui->getMessageManager()->getErrorMessage() << endl;
		else
			cout << ui->getMessageManager()->getResultMessage() << endl;
					
		ui->getMessageManager()->resetErrorStatus();
				
	}

			
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
*/
		
	// optionally, add values to split internal nodes ..
/*
	int newValue = 293;
		
	// 3(numberOfInserts) + 2(existing in internal node) = 5(enough to initiate split in internal node) ..
	int numberOfInserts = 3;

	for(int i=0; i<numberOfInserts; i++){

		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			// "a" column has the index ..
			insertValues.push_back( intToStr( newValue ) );
			insertValues.push_back( intToStr( 1 ) );
			insertValues.push_back( intToStr( 1 ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues.clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
		// 83, 103, 123 ..
		// 73, 93, 113 ..
		// 23, 43, 63 ..
		// 13, 33, 53 ..
		newValue += 20;
			
	}
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
*/


	// change the order of index creation to be  
	// 1) prior to insert, 2) after records are inserted ..
	// (accomplish by commenting this line out, uncommenting duplicate line above) ..

	ui->getQueryManager()->createIndex( indexName, tableName, isNoDuplicates, attributeName );
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();


	// if the index is created just above, then
	// add more records ..
	
	int numberOfAdditionalUniqueRows = 4;
	
	// 5, 10, 15, 20 .. 105 .. (5 - 90 flushed to file) ..
	//int numberOfUniqueRows = .. from above
	// 110, 115, 120 ..
	int numberOfMoreUniqueRows = ( (numberOfUniqueRows+1) + numberOfAdditionalUniqueRows );
		
	for(int i=(numberOfUniqueRows+1); i<=numberOfMoreUniqueRows; i++){
			
		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			// "a" column has the index ..
			insertValues->push_back( intToStr( ( i * multiple ) ) );
			
			// remaining columns are unimportant, but need dummy values ..
			insertValues->push_back( intToStr( ( i ) ) );
			
			insertValues->push_back( intToStr( ( i ) ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues->clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
	}
		
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	

	// print the index result ..
	ui->getIndexManager()->printIndex( tableName, indexName );
	
	
}



void Tester:: test_createTable_insert_createIndex_selectData(){
	
	cout << "Point A .." << endl << endl;
	
	// 8 columns per table max ..
	int maxColumnsPerTable = 8;
		
	// 3 buffer pages ..
	int bufferPoolSize = 3;

	UserInterface* ui = new UserInterface( maxColumnsPerTable, bufferPoolSize );
	
	// change the path to "/home/stalbot/CS525/Temp" ..
	vector<string>* inputTokens = new vector<string>();
	
	inputTokens->push_back( "set" );
	inputTokens->push_back( "path" );
	inputTokens->push_back( "home" );
	inputTokens->push_back( "stalbot" );
	inputTokens->push_back( "CS525" );
	inputTokens->push_back( "Temp" );
	
	ui->changePath( inputTokens, false );	// inputTokens, isWindows
	
	cout << "Point B .." << endl << endl;
		
	string tableName = "xyz";
	vector<string>* attributeNames = new vector<string>();
	attributeNames->push_back("a");
	attributeNames->push_back("b");
	attributeNames->push_back("c");
	int numberOfAttributes = 3;
		
	vector<string>* insertValues = new vector<string>();
		
	ui->getQueryManager()->createTable( tableName, attributeNames, numberOfAttributes );
	
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
	string indexName = "I";
	bool isNoDuplicates = false; 
	string indexAttributeName = "a";
		
	// change the order of index creation to be  
	// 1) prior to insert, 2) after records are inserted ..
	// (accomplish by commenting this line out, uncommenting duplicate line below) ..
/*
	ui->getQueryManager()->createIndex( indexName, tableName, isNoDuplicates, indexAttributeName );
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();
		
*/
		
	// number of pages (each row in a page is a repeat, but each page is unique) ..
	int numberOfUniqueRows = getNumberOfUniqueRows();
		
	// each unique row is separated by a multiple ..
	int multiple = 5;
		
	// each unique row is repeated this many times ..
	int numberRepeatRows = 4;
	
	// insert into the table ..
	test_insert( ui, tableName, numberOfUniqueRows, numberRepeatRows, multiple, numberOfAttributes );
	
/*
	cout << "Point C .." << endl << endl;
		
	for(int i=1; i<=numberOfUniqueRows; i++){
			
		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			// "a" column has the index ..
			insertValues->push_back( intToStr( ( i * multiple ) ) );
			
			// remaining columns are unimportant, but need dummy values ..
			insertValues->push_back( intToStr( ( i ) ) );
			
			insertValues->push_back( intToStr( ( i ) ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues->clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
	}
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
		
*/
		
	// optionally, add values to split leaf nodes ..
/*
	int newValue = 105;

	// repeat each unique row ..
	for(int j=0; j<numberRepeatRows; j++){
			
		// "a" column has the index ..
		insertValues.push_back( intToStr( newValue ) );
		insertValues.push_back( intToStr( 1 ) );
		insertValues.push_back( intToStr( 1 ) );
			
		ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
		// prepare for next row ..
		insertValues.clear();
				
		if( ui->getMessageManager()->hasErrorMessage() == true )
			cout << ui->getMessageManager()->getErrorMessage() << endl;
		else
			cout << ui->getMessageManager()->getResultMessage() << endl;
					
		ui->getMessageManager()->resetErrorStatus();
				
	}

			
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
*/
		
	// optionally, add values to split internal nodes ..
/*
	int newValue = 293;
		
	// 3(numberOfInserts) + 2(existing in internal node) = 5(enough to initiate split in internal node) ..
	int numberOfInserts = 3;

	for(int i=0; i<numberOfInserts; i++){

		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			// "a" column has the index ..
			insertValues.push_back( intToStr( newValue ) );
			insertValues.push_back( intToStr( 1 ) );
			insertValues.push_back( intToStr( 1 ) );
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues.clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
		// 83, 103, 123 ..
		// 73, 93, 113 ..
		// 23, 43, 63 ..
		// 13, 33, 53 ..
		newValue += 20;
			
	}
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
*/


	// change the order of index creation to be  
	// 1) prior to insert, 2) after records are inserted ..
	// (accomplish by commenting this line out, uncommenting duplicate line above) ..

	ui->getQueryManager()->createIndex( indexName, tableName, isNoDuplicates, indexAttributeName );
		
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();


	// if the index is created just above, then
	// add more records ..
	
	int numberOfAdditionalUniqueRows = 4;
	
	// 5, 10, 15, 20 .. 105 .. (5 - 90 flushed to file) ..
	//int numberOfUniqueRows = .. from above
	// 110, 115, 120 ..
	int numberOfMoreUniqueRows = ( (numberOfUniqueRows+1) + numberOfAdditionalUniqueRows );
		
	for(int i=(numberOfUniqueRows+1); i<=numberOfMoreUniqueRows; i++){
			
		// repeat each unique row ..
		for(int j=0; j<numberRepeatRows; j++){
			
			for(int k=0; k<numberOfAttributes; k++){
				
				// "a" column has the index ..
				if( k == 0 )
					insertValues->push_back( intToStr( ( multiple * i ) ) );
				else
					insertValues->push_back( intToStr( i ) );
				
			}
			
			/*
			// "a" column has the index ..
			insertValues->push_back( intToStr( ( i * multiple ) ) );
			
			// remaining columns are unimportant, but need dummy values ..
			insertValues->push_back( intToStr( ( i ) ) );
			
			insertValues->push_back( intToStr( ( i ) ) );
			*/
			
			ui->getQueryManager()->insertIntoTableValues( tableName, insertValues, numberOfAttributes );
		
			// prepare for next row ..
			insertValues->clear();
				
			if( ui->getMessageManager()->hasErrorMessage() == true )
				cout << ui->getMessageManager()->getErrorMessage() << endl;
			else
				cout << ui->getMessageManager()->getResultMessage() << endl;
					
			ui->getMessageManager()->resetErrorStatus();
				
		}
			
	}
		
		
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName, indexName );
	
	// set the values of the "SELECT .. WHERE" SQL statement ..
	//ui->setTableName( tableName );
	ui->addTableName( tableName );
	
	vector<string>* selectColumnNames = new vector<string>();
	selectColumnNames->push_back( indexAttributeName );
	ui->setSelectColumnNames( selectColumnNames );
	
	multimap<string, string>* whereColumnNamePairs = new multimap<string, string>();
	//whereColumnNamePairs->insert( pair<string, string>( indexAttributeName, inputField ) );
	//ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	map<string, int>* whereColumnNameValuePairs = new map<string, int>();
	whereColumnNameValuePairs->insert( pair<string, int>( indexAttributeName, strToInt( "105" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	
}

void Tester:: printMessages( UserInterface* ui ){
	
	// output message ..
	if( ui->getMessageManager()->hasErrorMessage() == true )
		cout << ui->getMessageManager()->getErrorMessage() << endl;
	else
		cout << ui->getMessageManager()->getResultMessage() << endl;
	ui->getMessageManager()->resetErrorStatus();	
	
}



void Tester:: test_testData(){
	
	// STEP 1: create the tables ..
	
	// 8 columns per table max ..
	int maxColumnsPerTable = 8;
		
	// 3 buffer pages ..
	int bufferPoolSize = 3;

	UserInterface* ui = new UserInterface( maxColumnsPerTable, bufferPoolSize );
	
	// change the path to "/home/stalbot/CS525/Temp" ..
	vector<string>* inputTokens = new vector<string>();
	
	inputTokens->push_back( "set" );
	inputTokens->push_back( "path" );
	inputTokens->push_back( "home" );
	inputTokens->push_back( "stalbot" );
	inputTokens->push_back( "CS525" );
	inputTokens->push_back( "Temp" );
	
	ui->changePath( inputTokens, false );	// inputTokens, isWindows
	
	//cout << "Tester:: test_testData_createTables() Point B .." << endl << endl;
		
	string tableName1 = "T1";
	vector<string>* attributeNames1 = new vector<string>();
	attributeNames1->push_back("A11");
	attributeNames1->push_back("A12");
	attributeNames1->push_back("A13");
	int numberOfAttributes1 = 3;
		
	vector<string>* insertValues1 = new vector<string>();
		
	ui->getQueryManager()->createTable( tableName1, attributeNames1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
		
	//cout << "Tester:: test_testData_createTables() Point C .." << endl << endl;
		
	string tableName2 = "T2";
	vector<string>* attributeNames2 = new vector<string>();
	attributeNames2->push_back("A21");
	attributeNames2->push_back("A22");
	attributeNames2->push_back("A23");
	attributeNames2->push_back("A24");
	attributeNames2->push_back("A25");
	int numberOfAttributes2 = 5;
		
	vector<string>* insertValues2 = new vector<string>();
		
	ui->getQueryManager()->createTable( tableName2, attributeNames2, numberOfAttributes2 );
	
	// output message ..
	printMessages( ui );
		
	// TABLE: "T1" ..
		
	// row: 1, 1, 1
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );

	// prepare for next row ..
	insertValues1->clear();

	// row: 1, 2, 2
	insertValues1->push_back( "1" );
	insertValues1->push_back( "2" );
	insertValues1->push_back( "2" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 1, 1, 3
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "3" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 1, 2, 4
	insertValues1->push_back( "1" );
	insertValues1->push_back( "2" );
	insertValues1->push_back( "4" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 2, 1, 1
	insertValues1->push_back( "2" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();

	// row: 2, 2, 2
	insertValues1->push_back( "2" );
	insertValues1->push_back( "2" );
	insertValues1->push_back( "2" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..indexAttributeName1
	insertValues1->clear();

	// row: 2, 1, 3
	insertValues1->push_back( "2" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "3" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();

	// row: 2, 2, 4
	insertValues1->push_back( "2" );
	insertValues1->push_back( "2" );
	insertValues1->push_back( "4" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();

	// row: 3, 1, 1
	insertValues1->push_back( "3" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();


	// TABLE: "T2" ..
		
	// row: 1, 2, 3, 4, 5
	insertValues2->push_back( "1" );
	insertValues2->push_back( "2" );
	insertValues2->push_back( "3" );
	insertValues2->push_back( "4" );
	insertValues2->push_back( "5" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName2, insertValues2, numberOfAttributes2 );
	
	// output message ..
	printMessages( ui );

	// prepare for next row ..
	insertValues2->clear();


	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	



	// STEP 2: create index, print index, set debug on, select data ..

	string indexName1 = "I11";
	bool isNoDuplicates = false; 
	string indexAttributeName1 = "A11";

	// change the order of index creation to be  
	// 1) prior to insert, 2) after records are inserted ..
	// (accomplish by commenting this line out, uncommenting duplicate line above) ..

	ui->getQueryManager()->createIndex( indexName1, tableName1, isNoDuplicates, indexAttributeName1 );
		
	// output message ..
	printMessages( ui );
	



	// print the index result ..
	ui->getIndexManager()->printIndex( tableName1, indexName1 );
	
	// output message ..
	printMessages( ui );
	
	// "SET INDEXDEBUG ON" ..
	ui->getIndexManager()->setIsIndexDebug( true );
	
	

	// set the values of the "SELECT .. WHERE" SQL statement ..
	// TABLE "T1" ..
	//ui->setTableName( tableName );
	ui->addTableName( tableName1 );
	
	vector<string>* selectColumnNames = new vector<string>();
	//selectColumnNames->push_back( indexAttributeName );
	ui->setSelectColumnNames( selectColumnNames );
	
	string whereAttribute = "A11";
	
	multimap<string, string>* whereColumnNamePairs = new multimap<string, string>();
	//whereColumnNamePairs->insert( pair<string, string>( whereAttribute, inputField ) );
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	map<string, int>* whereColumnNameValuePairs = new map<string, int>();
	whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	
	// STEP 3: insert into index, commit, print index ..
	
	// row: 4, 1, 1
	insertValues1->push_back( "4" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();

	// row: 5, 1, 1
	insertValues1->push_back( "5" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName1, indexName1 );
	
	// output message ..
	printMessages( ui );
	
	
	// STEP 4: print the same select statement as before ..
	
	ui->clearTableNames();

	// TABLE "T1" ..
	ui->addTableName( tableName1 );
	
	ui->setSelectColumnNames( selectColumnNames );
	
	whereAttribute = "A11";
	
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	whereColumnNameValuePairs = new map<string, int>();
	whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	
	// STEP 5: insert more data, print index, print table ..
	
	// row: 6, 1, 1
	insertValues1->push_back( "6" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 7, 1, 1
	insertValues1->push_back( "7" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 8, 1, 1
	insertValues1->push_back( "8" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName1, indexName1 );
	
	// output message ..
	printMessages( ui );
	
	ui->clearTableNames();

	// TABLE "T1" ..
	ui->addTableName( tableName1 );
	
	ui->setSelectColumnNames( selectColumnNames );
	
	whereAttribute = "A11";
	
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	whereColumnNameValuePairs = new map<string, int>();
	//whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	


	// STEP 6: insert more data, print index, print table ..
	
	// row: 9, 1, 1
	insertValues1->push_back( "9" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 10, 1, 1
	insertValues1->push_back( "10" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 11, 1, 1
	insertValues1->push_back( "11" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName1, indexName1 );
	
	// output message ..
	printMessages( ui );
	
	ui->clearTableNames();

	// TABLE "T1" ..
	ui->addTableName( tableName1 );
	
	ui->setSelectColumnNames( selectColumnNames );
	
	whereAttribute = "A11";
	
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	whereColumnNameValuePairs = new map<string, int>();
	//whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	

	// STEP 7: insert more data, print index, print table ..
	
	// row: 12, 1, 1
	insertValues1->push_back( "12" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 13, 1, 1
	insertValues1->push_back( "13" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 14, 1, 1
	insertValues1->push_back( "14" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName1, indexName1 );
	
	// output message ..
	printMessages( ui );
	
	ui->clearTableNames();

	// TABLE "T1" ..
	ui->addTableName( tableName1 );
	
	ui->setSelectColumnNames( selectColumnNames );
	
	whereAttribute = "A11";
	
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	whereColumnNameValuePairs = new map<string, int>();
	//whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );



	// STEP 8: insert more data, print index, print table ..
	
	// row: 15, 1, 1
	insertValues1->push_back( "15" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 16, 1, 1
	insertValues1->push_back( "16" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 17, 1, 1
	insertValues1->push_back( "17" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName1, indexName1 );
	
	// output message ..
	printMessages( ui );
	
	ui->clearTableNames();

	// TABLE "T1" ..
	ui->addTableName( tableName1 );
	
	ui->setSelectColumnNames( selectColumnNames );
	
	whereAttribute = "A11";
	
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	whereColumnNameValuePairs = new map<string, int>();
	//whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );




	// STEP 9: insert more data, print index, print table ..
	
	// row: 18, 1, 1
	insertValues1->push_back( "18" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 19, 1, 1
	insertValues1->push_back( "19" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// row: 20, 1, 1
	insertValues1->push_back( "20" );
	insertValues1->push_back( "1" );
	insertValues1->push_back( "1" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName1, insertValues1, numberOfAttributes1 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues1->clear();
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName1, indexName1 );
	
	// output message ..
	printMessages( ui );
	
	ui->clearTableNames();

	// TABLE "T1" ..
	ui->addTableName( tableName1 );
	
	ui->setSelectColumnNames( selectColumnNames );
	
	whereAttribute = "A11";
	
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	whereColumnNameValuePairs = new map<string, int>();
	//whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );



	// STEP 10: create index attempt, error: already exists ..

	string indexName2 = "I12";
	isNoDuplicates = false; 
	indexAttributeName1 = "A11";

	// change the order of index creation to be  
	// 1) prior to insert, 2) after records are inserted ..
	// (accomplish by commenting this line out, uncommenting duplicate line above) ..

	ui->getQueryManager()->createIndex( indexName2, tableName1, isNoDuplicates, indexAttributeName1 );
		
	// output message ..
	printMessages( ui );

	
	
	// STEP 11: create index, print index ..

	string indexName3 = "I21";
	isNoDuplicates = true; 
	string indexAttributeName2 = "A21";

	// change the order of index creation to be  
	// 1) prior to insert, 2) after records are inserted ..
	// (accomplish by commenting this line out, uncommenting duplicate line above) ..

	ui->getQueryManager()->createIndex( indexName3, tableName2, isNoDuplicates, indexAttributeName2 );
		
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName2, indexName3 );
	


	// STEP 12: insert more data, print index, print table ..
	
	// row: 2, 2, 3, 4, 5
	insertValues2->push_back( "2" );
	insertValues2->push_back( "2" );
	insertValues2->push_back( "3" );
	insertValues2->push_back( "4" );
	insertValues2->push_back( "5" );
			
	ui->getQueryManager()->insertIntoTableValues( tableName2, insertValues2, numberOfAttributes2 );
	
	// output message ..
	printMessages( ui );
	
	// prepare for next row ..
	insertValues2->clear();
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );
	
	// print the index result ..
	ui->getIndexManager()->printIndex( tableName2, indexName3 );
	
	// output message ..
	printMessages( ui );
	
	ui->clearTableNames();

	// TABLE "T2" ..
	ui->addTableName( tableName2 );
	
	ui->setSelectColumnNames( selectColumnNames );
	
	whereAttribute = "A21";
	
	ui->setWhereColumnNamePairs( whereColumnNamePairs );
	
	whereColumnNameValuePairs = new map<string, int>();
	//whereColumnNameValuePairs->insert( pair<string, int>( whereAttribute, strToInt( "3" ) ) );
	ui->setWhereColumnNameValuePairs( whereColumnNameValuePairs );

	// print the database table using the "SELECT" statement above ..
	ui->printDatabaseTable();
	
	// output message ..
	printMessages( ui );
	
	// flush the remaining data pages to file ..
	ui->getQueryManager()->commit();
	
	// output message ..
	printMessages( ui );


}


string Tester:: createFillerString( int fillerSize ){
	
	// create the "filler" string that is used to fill up
	// the portion of a 48-byte page in the file
	// which is not occupied by data, so that
	// the file size is always an increment of 48 bytes
	
	// i.e.: data = 36 bytes, filler = 12 bytes, total = 48 bytes ..
	
	string filler;
	
	for ( int i=0; i < fillerSize; i++ ) {
		
		filler.push_back( '$' );
		
	}
	
	return filler;

		
} // end "constructPagePadding( int fillerSize )" ..

